
/*!
 *  \file   beagle/ENSO/GenerationalOp.hpp
 *  \brief  Definition of the class ENSO::GenerationalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_ENSO_GenerationalOp_hpp
#define Beagle_ENSO_GenerationalOp_hpp

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/Container.hpp"
#include "beagle/GenerationalOp.hpp"

namespace Beagle
{
namespace ENSO
{

/*!
 *  \class GenerationalOp beagle/ENSO/GenerationalOp.hpp "beagle/ENSO/GenerationalOp.hpp"
 *  \brief Generational replacement strategy operator class for ENSO.
 *  \ingroup ENSOF
 */
class GenerationalOp : public Beagle::GenerationalOp
{

public:

	//! ENSO::GenerationalOp allocator type.
	typedef AllocatorT<ENSO::GenerationalOp,Beagle::GenerationalOp::Alloc>
	Alloc;
	//! ENSO::GenerationalOp handle type.
	typedef PointerT<ENSO::GenerationalOp,Beagle::GenerationalOp::Handle>
	Handle;
	//! ENSO::GenerationalOp bag type.
	typedef ContainerT<ENSO::GenerationalOp,Beagle::GenerationalOp::Bag>
	Bag;

	GenerationalOp(std::string inName="ENSO-GenerationalOp");
	virtual ~GenerationalOp()
	{ }

	virtual void registerParams(System& ioSystem);
	virtual void operate(Deme& ioDeme, Context& ioContext);

protected:

    /*!
     *  \brief Binary predicate functor that tests the ordering of two individuals
     *      based on the age of their symmetry break and lattice indices.  Sorting
     *      the population using this predicate groups together all the individuals
     *      whose symmetry age is greater than the threshold at the end of the array,
     *      while all the other individuals are grouped according to their lattice
     *      indices.
     *  \par Note:
     *      This class is compliant with the STL Strict Weak Ordering.
     *      See http://www.sgi.com/tech/stl/StrictWeakOrdering.html
     */
    class IsLessIndivPredicate
    {
    public:
        IsLessIndivPredicate(unsigned int inGeneration, float inMinMaturity, float inIncMaturity) :
                mGeneration(inGeneration),
                mMinMaturity(inMinMaturity),
                mIncMaturity(inIncMaturity)
        { }

        ~IsLessIndivPredicate()
        { }

        /*!
         *  \brief Test the ordering of two individuals based on their symmetry age and lattice indices.
         *      Sorting individuals using this operator partitions them into two groups: (1) age less
         *      than or equal to the threshold, and (2) age greater than the threshold.  Within each group,
         *      the individuals are further grouped into the same lattice index.
         */
        inline bool operator()(const Pointer& inLeftIndiv, const Pointer& inRightIndiv) const
        {
            Beagle_StackTraceBeginM();
            const ENSO::Individual& lLeftIndiv = castObjectT<const ENSO::Individual&>(*inLeftIndiv);
            const ENSO::Individual& lRightIndiv = castObjectT<const ENSO::Individual&>(*inRightIndiv);
            bool lLeftIsMature = lLeftIndiv.isMature(mGeneration, mMinMaturity, mIncMaturity);
            bool lRightIsMature = lRightIndiv.isMature(mGeneration, mMinMaturity, mIncMaturity);
            if ((!lLeftIsMature && !lRightIsMature) || (lLeftIsMature && lRightIsMature))
                return lLeftIndiv[0]->getLatticeIndex() < lRightIndiv[0]->getLatticeIndex();
            else
                return lRightIsMature;
            Beagle_StackTraceEndM("bool ENSO::GenerationalOp::IsLessIndivPredicate::operator()(const Pointer& inLeftIndiv, const Pointer& inRightIndiv) const");
        }

    private:
        unsigned int    mGeneration;
        float           mMinMaturity;
        float           mIncMaturity;
    };

	UInt::Handle        mNumOffSymBrkMut;      //!< Number of offspring to create in each symmetry break mutation.
	Float::Handle       mMinMaturity;          //!< Minimum maturity for comparing fitness of symmetrically different individuals.
	Float::Handle       mIncMaturity;          //!< Increment in min maturity for comparing fitness of symmetrically different individuals.
	Bool::Handle        mFitInvalidate;        //!< Whether to invalidate fitness of all individuals in the population.

};

}
}

#endif // Beagle_ENSO_GenerationalOp_hpp

