
/*!
 *  \file   beagle/ENSO/Individual.hpp
 *  \brief  Definition of the class ENSO::Individual.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_ENSO_Individual_hpp
#define Beagle_ENSO_Individual_hpp

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Individual.hpp"
#include "beagle/ENSO/Tree.hpp"


namespace Beagle
{
namespace ENSO
{

/*!
 *  \class Individual beagle/ENSO/Individual.hpp "beagle/ENSO/Individual.hpp"
 *  \brief An individual inherits from a uniform bag of genotypes and has a fitness member.
 *  \ingroup ENSOF
 */
class Individual : public Beagle::Individual
{

public:

	//! ENSO::Individual allocator type.
	typedef AllocatorT<ENSO::Individual,Beagle::Individual::Alloc> Alloc;
	//! ENSO::Individual handle type.
	typedef PointerT<ENSO::Individual,Beagle::Individual::Handle> Handle;
	//! ENSO::Individual bag type.
	typedef ContainerT<ENSO::Individual,Beagle::Individual::Bag> Bag;

	explicit Individual(unsigned int inN=0, ENSO::Tree::Handle inModel=NULL);

    virtual void                readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext);
	virtual void				writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;
	virtual const std::string&	getType() const;
	const ENSO::Tree::Handle&	operator[](unsigned int inN) const;
	ENSO::Tree::Handle&			operator[](unsigned int inN);
	virtual double              getDistanceGenotype(const Beagle::Individual& inRightIndividual) const;

    void                                    makeGraph();
    const std::vector<Parameters::Handle>&  getGraph() const;

    bool    isMature(unsigned int inGeneration, float inMinMaturity, float inIncMaturity) const;
    bool    allowSymmetryBreak(unsigned int inGeneration, float inMinSymmetryBreak, float inIncSymmetryBreak) const;
    bool    allowIsEdgeMutation(unsigned int inGeneration, float inMinIsEdgeMut, float inIncIsEdgeMut) const;

protected:

    std::vector<Parameters::Handle> mGraph;             //!< The phenotype graph constructed from the genotype tree.

};

}
}


/*!
 *  \brief Constant handle to the nth genotype of the ENSO::Individual.
 *  \param inN Index of the ENSO::Tree to get
 *  \return Constant pointer (ENSO::Tree::Handle) to the genotype at the index N.
 */
inline const Beagle::ENSO::Tree::Handle& Beagle::ENSO::Individual::operator[](unsigned int inN) const
{
	Beagle_StackTraceBeginM();
	Beagle_UpperBoundCheckAssertM(inN,size()-1);
	return castHandleT<ENSO::Tree>(std::vector<Pointer>::operator[](inN));
	Beagle_StackTraceEndM("const ENSO::Tree::Handle& ENSO::Individual::operator[](unsigned int inN) const");
}


/*!
 *  \brief Handle to the nth genotype of the ENSO::Individual.
 *  \param inN Index of the ENSO::Tree to get.
 *  \return Pointer (ENSO::Tree::Handle) to the genotype at the index N.
 */
inline Beagle::ENSO::Tree::Handle& Beagle::ENSO::Individual::operator[](unsigned int inN)
{
	Beagle_StackTraceBeginM();
	Beagle_UpperBoundCheckAssertM(inN,size()-1);
	return castHandleT<ENSO::Tree>(std::vector<Pointer>::operator[](inN));
	Beagle_StackTraceEndM("ENSO::Tree::Handle& ENSO::Individual::operator[](unsigned int inN)");
}


/*!
 *  \brief Get the phenotype graph.
 */
inline const std::vector<Beagle::ENSO::Parameters::Handle>& Beagle::ENSO::Individual::getGraph() const
{
	Beagle_StackTraceBeginM();
    return mGraph;
	Beagle_StackTraceEndM("const std::vector<Parameters::Handle>& ENSO::Individual::getGraph() const");
}


/*!
 *  \brief Checks if the individual is mature enough for competing with others of different symmetries.
 */
inline bool Beagle::ENSO::Individual::isMature(unsigned int inGeneration, float inMinMaturity, float inIncMaturity) const
{
	Beagle_StackTraceBeginM();
    return inGeneration - (*this)[0]->getLastSymmetryBreakGeneration() > inMinMaturity + (*this)[0]->size()*inIncMaturity;
	Beagle_StackTraceEndM("bool ENSO::Individual::isMature(unsigned int inGeneration, float inMinMaturity, float inIncMaturity) const");
}


/*!
 *  \brief Checks if enough generations have elapsed to allow another symmetry break.
 */
inline bool Beagle::ENSO::Individual::allowSymmetryBreak(unsigned int inGeneration, float inMinSymmetryBreak, float inIncSymmetryBreak) const
{
	Beagle_StackTraceBeginM();
    return inGeneration - (*this)[0]->getLastSymmetryBreakGeneration() > inMinSymmetryBreak + (*this)[0]->size()*inIncSymmetryBreak;
	Beagle_StackTraceEndM("bool ENSO::Individual::allowSymmetryBreak(unsigned int inGeneration, float inMinSymmetryBreak, float inIncSymmetryBreak) const");
}


/*!
 *  \brief Checks if enough generations have elapsed to allow another edge/non-edge mutation.
 */
inline bool Beagle::ENSO::Individual::allowIsEdgeMutation(unsigned int inGeneration, float inMinIsEdgeMut, float inIncIsEdgeMut) const
{
	Beagle_StackTraceBeginM();
    return inGeneration - (*this)[0]->getLastIsEdgeMutationGeneration() > inMinIsEdgeMut + (*this)[0]->size()*inIncIsEdgeMut;
	Beagle_StackTraceEndM("bool ENSO::Individual::allowIsEdgeMutation(unsigned int inGeneration, float inMinIsEdgeMut, float inIncIsEdgeMut) const");
}

#endif // Beagle_ENSO_Individual_hpp
