
/*!
 *  \file   beagle/ENSO/MutationBreakSymmetryOp.hpp
 *  \brief  Definition of the type ENSO::MutationBreakSymmetryOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_ENSO_MutationBreakSymmetryOp_hpp
#define Beagle_ENSO_MutationBreakSymmetryOp_hpp

#include <string>

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/MutationOp.hpp"


namespace Beagle
{
namespace ENSO
{

/*!
 *  \class MutationBreakSymmetryOp beagle/ENSO/MutationBreakSymmetryOp.hpp "beagle/ENSO/MutationBreakSymmetryOp.hpp"
 *  \brief Mutation that minimally breaks the symmetry of an ENSO genotype
 *         tree by utilizing the subgroup (orbital partition) lattice.
 *  \ingroup ENSOF
 */
class MutationBreakSymmetryOp : public Beagle::MutationOp
{

public:

	//! ENSO::MutationBreakSymmetryOp allocator type.
	typedef AllocatorT<MutationBreakSymmetryOp,Beagle::MutationOp::Alloc>
	Alloc;
	//! ENSO::MutationBreakSymmetryOp handle type.
	typedef PointerT<MutationBreakSymmetryOp,Beagle::MutationOp::Handle>
	Handle;
	//! ENSO::MutationBreakSymmetryOp bag type.
	typedef ContainerT<MutationBreakSymmetryOp,Beagle::MutationOp::Bag>
	Bag;

    /*!
     *  \struct Element beagle/ENSO/MutationBreakSymmetryOp.hpp "beagle/ENSO/MutationBreakSymmetryOp.hpp"
     *  \brief Element of lattice of subgroups (orbital partitions).
     *  \ingroup ENSOF
     */
    struct Element
    {
        //! Each cell in the orbital partition has a color corresponding to which
        //! vertices or edges of the phenotype graph are contained in that cell,
        //! and these colors are stored as vectors of bits.
        std::vector< std::vector<bool> >    mColors;

        std::vector<unsigned int>           mChildren;  //!< Indices of childen of this element in the lattice.
        unsigned int                        mGroupSize; //!< Size (number of elements) of the group.

        explicit Element()
        { }
    };

	explicit MutationBreakSymmetryOp(std::string inMutationPbName="enso.mutbrksym.indpb",
                                     std::string inName="ENSO-MutationBreakSymmetryOp");
	virtual ~MutationBreakSymmetryOp()
	{ }

	virtual void    registerParams(Beagle::System& ioSystem);
	virtual bool    mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext);
	virtual void    init(Beagle::System& ioSystem);

    bool            breakSymmetry(ENSO::Tree& ioTree, Beagle::Context& ioContext) const;
    bool            breakSymmetry(ENSO::Tree& ioTree, unsigned int inNewLatticeIndex, Beagle::Context& ioContext) const;

protected:

    static std::vector<Element>     mLattice;           //!< Subgroup (orbital partition) lattice.  
    String::Handle                  mLatticeFile;       //!< Name of file containing the subgroup lattice.
	UInt::Handle                    mNumVertices;       //!< Number of vertices in phenotype graph.
	Float::Handle                   mMinSymmetryBreak;  //!< Minimum generation interval between symmetry breaks.
	Float::Handle                   mIncSymmetryBreak;  //!< Increment in minimum interval between symmetry breaks.
    Bool::Handle                    mPruneLattice;      //!< Whether to prune the subgroup lattice.
    Bool::Handle                    mUndirectedGraph;   //!< Break symmetry assuming undirected graphs.
    Float::Handle                   mRestoreProba;      //!< Probability that symmetry is restored.

    void readLattice(const std::string& inLatticeFile);
    void readColors(PACC::XML::ConstIterator inIter, std::vector< std::vector<bool> >& outColors);
    void readChildren(PACC::XML::ConstIterator inIter, std::vector<unsigned int>& outChildren);
    void writeLattice(const std::string& inLatticeFile);

};

}
}

#endif // Beagle_ENSO_MutationBreakSymmetryOp_hpp
