
/*!
 *  \file   beagle/ENSO/Parameters.hpp
 *  \brief  Definition of class ENSO::Parameters; similar to Beagle::Map
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_ENSO_Parameters_hpp
#define Beagle_ENSO_Parameters_hpp

#include <string>
#include <map>

#include "PACC/XML.hpp"

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/PointerT.hpp"
#include "beagle/Container.hpp"
#include "beagle/ContainerT.hpp"
#include "beagle/HashString.hpp"
#include "beagle/Genotype.hpp"

#if defined(BEAGLE_HAVE_STD_HASHMAP) | defined(BEAGLE_HAVE_STDEXT_HASHMAP)
#include <hash_map>
#endif // defined(BEAGLE_HAVE_STD_HASHMAP) | defined(BEAGLE_HAVE_STDEXT_HASHMAP)

#ifdef BEAGLE_HAVE_GNUCXX_HASHMAP
#include <ext/hash_map>
#endif // BEAGLE_HAVE_GNUCXX_HASHMAP


namespace Beagle
{
namespace ENSO
{

/*!
 *  \class Parameters beagle/ENSO/Parameters.hpp "beagle/ENSO/Parameters.hpp"
 *  \brief Associative container using a string as key accessors to the stored
 *    Genotype handles; similar to the Beagle::Map class.
 *  \ingroup ENSOF
 */

#if defined(BEAGLE_HAVE_STD_HASHMAP)
class Parameters : public Genotype, public std::hash_map<std::string,Genotype::Handle,HashString>
#elif defined(BEAGLE_HAVE_GNUCXX_HASHMAP)
class Parameters : public Genotype, public __gnu_cxx::hash_map<std::string,Genotype::Handle,HashString>
#elif defined(BEAGLE_HAVE_STDEXT_HASHMAP)
class Parameters : public Genotype, public stdext::hash_map<std::string,Genotype::Handle,HashString>
#else // No hash_map found
class Parameters : public Genotype, public std::map< std::string,Genotype::Handle,std::less<std::string> >
#endif
{

public:

	//! Parameters allocator type.
	typedef AllocatorT<Parameters,Genotype::Alloc> Alloc;
	//! Parameters handle type.
	typedef PointerT<Parameters,Genotype::Handle> Handle;
	//! Parameters bag type.
	typedef ContainerT<Parameters,Genotype::Bag> Bag;

	explicit Parameters(bool inValid=false);
	virtual ~Parameters()
	{ }

    Parameters(const Parameters& inRightParams);  // Copy constructor

    virtual void                antialias();

    virtual unsigned int        getSize() const;
	virtual const std::string&  getType() const;
    virtual Parameters&         operator=(const Parameters& inRightParams);
    virtual void                copy(const Member& inOriginal, System& ioSystem);
    virtual bool                isEqual(const Object& inRightObj) const;
    virtual bool                isLess(const Object& inRightObj) const;
    virtual void                readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext);
    virtual void                writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;

    virtual void                randomize(Beagle::Context& ioContext);

    bool    isValid() const;
    void    setInvalid();
    void    setValid();

protected:

    bool        mValid;         //!< Whether the parameters are valid.

};

}
}


/*!
 *  \brief Return validity of the parameters.
 *  \return True if parameters is valid, false if not.
 */
inline bool Beagle::ENSO::Parameters::isValid() const
{
    Beagle_StackTraceBeginM();
    return mValid;
    Beagle_StackTraceEndM("bool ENSO::Parameters::isValid() const");
}


/*!
 *  \brief Invalidate the parameters.
 */
inline void Beagle::ENSO::Parameters::setInvalid()
{
    Beagle_StackTraceBeginM();
    mValid = false;
    Beagle_StackTraceEndM("void ENSO::Parameters::setInvalid()");
}


/*!
 *  \brief Validate the parameters.
 */
inline void Beagle::ENSO::Parameters::setValid()
{
    Beagle_StackTraceBeginM();
    mValid = true;
    Beagle_StackTraceEndM("void ENSO::Parameters::setValid()");
}

#endif // Beagle_ENSO_Parameters_hpp
