
/*!
 *  \file   beagle/ENSO/Tree.hpp
 *  \brief  Definition of the type ENSO::Tree.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_ENSO_Tree_hpp
#define Beagle_ENSO_Tree_hpp

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/Container.hpp"
#include "beagle/Genotype.hpp"
#include "beagle/ENSO/Parameters.hpp"

#include <algorithm>


namespace Beagle
{
namespace ENSO
{

/*!
 *  \struct Node beagle/ENSO/Tree.hpp "beagle/ENSO/Tree.hpp"
 *  \brief ENSO tree node structure.
 *  \ingroup ENSOF
 */
struct Node
{
    //! Color of this node; also encodes the vertex/edge objects having this color
    //! by setting the corresponding bits.
    std::vector<bool>           mColor;

    unsigned int                mParent;        //!< Index of parent node in the genotype vector.
    std::vector<unsigned int>   mChildren;      //!< Indices of child nodes in the genotype vector.

    Parameters                  mParameters;    //!< Parameters for vertices/edges.

	explicit Node(const std::vector<bool>& inColor=std::vector<bool>(), unsigned int inParent=UINT_MAX);
	bool operator==(const Node& inRightNode) const;
};


/*!
 *  \class Tree beagle/ENSO/Tree.hpp "beagle/ENSO/Tree.hpp"
 *  \brief The ENSO genotype tree class.
 *  \ingroup ENSOF
 */
class Tree : public Genotype, public std::vector<Node>
{

public:

	//! ENSO::Tree allocator type.
	typedef AllocatorT<Tree, Genotype::Alloc>
	Alloc;
	//! ENSO::Tree handle type.
	typedef PointerT<Tree, Genotype::Handle>
	Handle;
	//! ENSO::Tree bag type.
	typedef ContainerT<Tree, Genotype::Bag>
	Bag;

	explicit Tree(unsigned int inNumVertices=0);
	virtual ~Tree()
	{ }

	virtual void                copy(const Member& inOriginal, System& ioSystem);
	virtual unsigned int        getSize() const;
	virtual const std::string&  getType() const;

	virtual bool        isEqual(const Object& inRightObj) const;
	virtual bool        isLess(const Object& inRightObj) const;
	virtual void        readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext);
	virtual void        writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;

	unsigned int        getNumVertices() const;
    void                setNumVertices(unsigned int inNumVertices);

    void                getLeaves(std::vector<unsigned int>& outLeaves) const;
    void                getVertexLeaves(std::vector<unsigned int>& outLeaves) const;
    void                getEdgeLeaves(std::vector<unsigned int>& outLeaves) const;
    void                getVertexNodes(std::vector<unsigned int>& outNodes) const;
    void                getEdgeNodes(std::vector<unsigned int>& outNodes) const;
    void                getParallelEdgeLeaves(std::vector<unsigned int>& outLeaves) const;

    void                sortNodes(std::vector<unsigned int>& ioNodes) const;

    void                recordSymmetryState(unsigned int inGeneration, unsigned int inLatticeIndex);
    bool                restoreSymmetryState(unsigned int inGeneration);
    unsigned int        getLastSymmetryBreakGeneration() const;
	unsigned int        getLatticeIndex() const;

    void                setLastIsEdgeMutationGeneration(unsigned int inGeneration);
    unsigned int        getLastIsEdgeMutationGeneration() const;

protected:

    /*!
     *  \brief Binary predicate functor that tests the ordering of two Tree nodes.
     *  \par Note:
     *    This class is compliant with the STL Strict Weak Ordering.
     *    See http://www.sgi.com/tech/stl/StrictWeakOrdering.html
     */
    class IsLessNodePredicate
    {
    public:
        IsLessNodePredicate(const Tree& inTree) : mTree(inTree)
        { }
        ~IsLessNodePredicate()
        { }

        /*!
         *  \brief Test the ordering of two tree nodes with the given indices.
         *  \param inLeftIndex Index to the left node to compare.
         *  \param inRightIndex Index to the right node to compare.
         */
        inline bool operator()(unsigned int inLeftIndex, unsigned int inRightIndex) const
        {
            Beagle_StackTraceBeginM();
            return (mTree[inLeftIndex].mColor < mTree[inRightIndex].mColor);
            Beagle_StackTraceEndM("bool ENSO::Tree::IsLessNodePredicate::operator()(unsigned int inLeftIndex, unsigned int inRightIndex) const");
        }

    private:
        const Tree&     mTree;          //!< The tree whose nodes are being compared.
    };

    unsigned int                mNumVertices;                   //!< Number of vertices in phenotype graph.
    unsigned int                mLastSymmetryBreakGeneration;   //!< Generation at which symmetry was last broken.
    unsigned int                mLastIsEdgeMutationGeneration;  //!< Generation at which last edge/non-edge mutation occured.
    std::vector<unsigned int>   mLatticePath;                   //!< Lattice indices at which symmetry breaks occured.
    std::vector<unsigned int>   mSizePath;                      //!< Sizes of the tree at which symmetry breaks occured.

private:

	void        readVectorContent(PACC::XML::ConstIterator inIter, std::vector<unsigned int>& outVector);
	void        writeVectorContent(PACC::XML::Streamer& ioStreamer, const std::vector<unsigned int>& inVector) const;

};

}
}


/*!
 *  \return The number of nodes in the tree.
 */
inline unsigned int Beagle::ENSO::Tree::getSize() const
{
	Beagle_StackTraceBeginM();
	return size();
	Beagle_StackTraceEndM("unsigned int ENSO::Tree::getSize() const");
}


/*!
 * \return The number of vertices in the phenotype graph.
 */
inline unsigned int Beagle::ENSO::Tree::getNumVertices() const
{
	Beagle_StackTraceBeginM();
	return mNumVertices;
	Beagle_StackTraceEndM("unsigned int ENSO::Tree::getNumVertices() const");
}


/*!
 *  \brief Set the number of vertices in the phenotype graph.
 */
inline void Beagle::ENSO::Tree::setNumVertices(unsigned int inNumVertices)
{
	Beagle_StackTraceBeginM();
	mNumVertices = inNumVertices;
	Beagle_StackTraceEndM("void ENSO::Tree::setNumVertices(unsigned int inNumVertices)");
}


/*!
 *  \brief Sort a vector of node indices based on the color of nodes.
 *  \param ioNodes Contains the indices of nodes to sort.
 */
inline void Beagle::ENSO::Tree::sortNodes(std::vector<unsigned int>& ioNodes) const
{
	Beagle_StackTraceBeginM();
    std::stable_sort(ioNodes.begin(), ioNodes.end(), IsLessNodePredicate(*this));
	Beagle_StackTraceEndM("void Beagle::ENSO::Tree::sortNodes(std::vector<unsigned int>& ioNodes) const");
}


/*!
 *  \brief Get the generation at which the last symmetry break occured.
 */
inline unsigned int Beagle::ENSO::Tree::getLastSymmetryBreakGeneration() const
{
	Beagle_StackTraceBeginM();
    return mLastSymmetryBreakGeneration;
	Beagle_StackTraceEndM("unsigned int ENSO::Tree::getLastSymmetryBreakGeneration() const");
}


/*!
 * \return The index of current subgroup lattice element.
 */
inline unsigned int Beagle::ENSO::Tree::getLatticeIndex() const
{
	Beagle_StackTraceBeginM();
	return mLatticePath.back();
	Beagle_StackTraceEndM("unsigned int ENSO::Tree::getLatticeIndex() const");
}


/*!
 *  \brief Set the generation at which the last edge/non-edge mutation occured.
 */
inline void Beagle::ENSO::Tree::setLastIsEdgeMutationGeneration(unsigned int inGeneration)
{
       Beagle_StackTraceBeginM();
       mLastIsEdgeMutationGeneration = inGeneration;
       Beagle_StackTraceEndM("void ENSO::Tree::setLastIsEdgeMutationGeneration(unsigned int inGeneration)");
}


/*!
 *  \brief Get the generation at which the last symmetry break occured.
 */
inline unsigned int Beagle::ENSO::Tree::getLastIsEdgeMutationGeneration() const
{
	Beagle_StackTraceBeginM();
    return mLastIsEdgeMutationGeneration;
	Beagle_StackTraceEndM("unsigned int ENSO::Tree::getLastIsEdgeMutationGeneration() const");
}

#endif // Beagle_ENSO_Tree_hpp
