
/*!
 *  \file   beagle/ENSO/src/EvaluationOp.cpp
 *  \brief  Source code of class ENSO::EvaluationOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct an ENSO evaluation operator.
 *  \param inPhenotypeEvalOp Evaluation operator for phenotype.
 *  \param inName Name of the operator.
 */
ENSO::EvaluationOp::EvaluationOp(Beagle::EvaluationOp::Handle inPhenotypeEvalOp, const std::string& inName) :
		Beagle::EvaluationOp(inName),
        mPhenotypeEvalOp(inPhenotypeEvalOp)
{ }


/*!
 *  \brief Register the parameters of the ENSO evaluation operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::EvaluationOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Beagle::EvaluationOp::registerParams(ioSystem);
    if (mPhenotypeEvalOp == NULL) {
        std::ostringstream lOSS;
        lOSS << "Phenotype evaluation operator is not specified in ENSO::EvaluationOp.";
        throw Beagle_RunTimeExceptionM(lOSS.str());
    }
	mPhenotypeEvalOp->registerParams(ioSystem);
	Beagle_StackTraceEndM("void ENSO::EvaluationOp::registerParams(Beagle::System&)");
}


/*!
 *  \brief Initialize this operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::EvaluationOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Beagle::EvaluationOp::init(ioSystem);
    if (mPhenotypeEvalOp == NULL) {
        std::ostringstream lOSS;
        lOSS << "Phenotype evaluation operator is not specified in ENSO::EvaluationOp.";
        throw Beagle_RunTimeExceptionM(lOSS.str());
    }
	mPhenotypeEvalOp->init(ioSystem);
	Beagle_StackTraceEndM("void ENSO::EvaluationOp::init(Beagle::System&)");
}


/*!
 *  \brief Evaluate the fitness of the given ENSO individual.
 *  \param inIndividual Current individual to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness value of the ENSO individual.
 */
Fitness::Handle ENSO::EvaluationOp::evaluate(Beagle::Individual& inIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	ENSO::Individual& lIndividual = castObjectT<ENSO::Individual&>(inIndividual);
	return evaluate(lIndividual, ioContext);

	Beagle_StackTraceEndM("Fitness::Handle ENSO::EvaluationOp::evaluate(Beagle::Individual& inIndividual, Beagle::Context& ioContext)");
}
