
/*!
 *  \file   beagle/ENSO/src/GenerationalOp.cpp
 *  \brief  Source code of class ENSO::GenerationalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Build generational operator.
 *  \param inName Name of the generational operator.
 */
ENSO::GenerationalOp::GenerationalOp(std::string inName) :
		Beagle::GenerationalOp(inName)
{ }


/*!
 *  \brief Register the parameters of this operator.
 *  \param ioSystem Reference to the evolutionary system.
 */
void ENSO::GenerationalOp::registerParams(System& ioSystem)
{
	Beagle_StackTraceBeginM();
    Beagle::GenerationalOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of offspring to create in each symmetry break mutation.",
		    "UInt",
		    "1",
		    "Number of offspring with different parameter values to create in each symmetry break mutation."
		);
		mNumOffSymBrkMut = castHandleT<UInt>(
		                    ioSystem.getRegister().insertEntry("enso.numoff.symbrkmut", new UInt(1), lDescription));
	}
    {
        Register::Description lDescription(
            "Minimum maturity after the last symmetry break mutation for competing with other individuals.",
            "Float",
            "0.0",
            "Fitnesses of symmetrically different individuals are compared only if their symmetry ages are above this maturity."
        );
        mMinMaturity = castHandleT<Float>(
                            ioSystem.getRegister().insertEntry("enso.sel.minmaturity", new Float(0.0), lDescription));
    }
    {
        Register::Description lDescription(
            "Increment in minimum maturity after the last symmetry break mutation for competing with other individuals.",
            "Float",
            "0.0",
            "Increment per genotype node in minimum maturity for comparing fitnesses of symmetrically different individuals."
        );
        mIncMaturity = castHandleT<Float>(
                            ioSystem.getRegister().insertEntry("enso.sel.incmaturity", new Float(0.0), lDescription));
    }
    {
        Register::Description lDescription(
            "Whether to invalidate fitness of all individuals in the population.",
            "Bool",
            "0",
            "If true, invalidate fitness of all individuals in the population regardless of whether they were changed."
        );
        mFitInvalidate = castHandleT<Bool>(
                            ioSystem.getRegister().insertEntry("enso.sel.fitinvalidate", new Bool(0), lDescription));
    }
	Beagle_StackTraceEndM("void ENSO::GenerationalOp::registerParams(System&)");
}


/*!
 *  \brief Apply the operation on a deme in the given context.
 *  \param ioDeme Reference to the deme on which the operation takes place.
 *  \param ioContext Evolutionary context of the operation.
 */
void ENSO::GenerationalOp::operate(Deme& ioDeme, Context& ioContext)
{
	Beagle_StackTraceBeginM();
	Beagle_NonNullPointerAssertM(getRootNode());
	Beagle_NonNullPointerAssertM(mElitismKeepSize);
	Beagle_ValidateParameterM(mElitismKeepSize->getWrappedValue() <= 1,
	                          "ec.elite.keepsize",
	                          "The elistism keepsize must be equal to 0 or 1!");

	Beagle_LogTraceM(
	    ioContext.getSystem().getLogger(),
	    "replacement-strategy", "Beagle::ENSO::GenerationalOp",
	    std::string("Processing using generational replacement strategy the ")+
	    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
	);
	Beagle_LogObjectM(
	    ioContext.getSystem().getLogger(),
	    Logger::eTrace,
	    "replacement-strategy", "Beagle::ENSO::GenerationalOp",
	    (*this)
	);

	RouletteT<unsigned int> lRoulette;
	buildRoulette(lRoulette, ioContext);

    Beagle::Individual::Bag lOffsprings;
	const Factory& lFactory = ioContext.getSystem().getFactory();
    History::Handle lHistory = castHandleT<History>(ioContext.getSystem().haveComponent("History"));
    std::string lIndividualType = ioDeme[0]->getType();
    Beagle::Individual::Alloc::Handle lIndividualAlloc =
        castHandleT<Beagle::Individual::Alloc>(lFactory.getAllocator(lIndividualType));

    // Sort the deme based on lattice index and last symmetry break generation.
    const float lMinMaturity = mMinMaturity->getWrappedValue();
    const float lIncMaturity = mIncMaturity->getWrappedValue();
    std::sort(ioDeme.begin(), ioDeme.end(), IsLessIndivPredicate(ioContext.getGeneration(), lMinMaturity, lIncMaturity));

    if(mElitismKeepSize->getWrappedValue() > 0) {
		// Select the elite individuals from each nich of the sorted deme.
        for(unsigned int i=0; i<ioDeme.size();) {
            unsigned int lEliteIndex = i;
            ENSO::Individual::Handle lIndiv = castHandleT<ENSO::Individual>(ioDeme[i]);
            if (lIndiv->isMature(ioContext.getGeneration(), lMinMaturity, lIncMaturity)) {
                for(++i; i<ioDeme.size(); ++i) {
                    if (ioDeme[lEliteIndex]->isLess(*ioDeme[i]))
                        lEliteIndex = i;
                }
            }
            else {
                for(++i; i<ioDeme.size() && (*lIndiv)[0]->getLatticeIndex()==(*castHandleT<ENSO::Individual>(ioDeme[i]))[0]->getLatticeIndex(); ++i) {
                    if (ioDeme[lEliteIndex]->isLess(*ioDeme[i]))
                        lEliteIndex = i;
                }
            }
            lOffsprings.push_back(ioDeme[lEliteIndex]);
		}

		// The offsprings are actually copies of the elite individuals.
        for(unsigned int i=0; i<lOffsprings.size(); ++i) {
			Beagle::Individual::Handle lEliteIndiv = castHandleT<Beagle::Individual>(lIndividualAlloc->allocate());
            lEliteIndiv->copy(*lOffsprings[i], ioContext.getSystem());
            if(lHistory != NULL) {
                HistoryID::Handle lHID = castHandleT<HistoryID>(lOffsprings[i]->getMember("HistoryID"));
                std::vector<HistoryID> lParent;
                if(lHID != NULL) lParent.push_back(*lHID);
                lHistory->allocateNewID(*lEliteIndiv);
                lHistory->trace(ioContext, lParent, lEliteIndiv, getName(), "elitism");
            }
            lOffsprings[i] = lEliteIndiv;
        }
    }

    for(unsigned int i=lOffsprings.size(); i<ioDeme.size();) {
		unsigned int lIndexBreeder = lRoulette.select(ioContext.getSystem().getRandomizer());
		BreederNode::Handle lSelectedBreeder=getRootNode();
		for(unsigned int j=0; j<lIndexBreeder; ++j)
			lSelectedBreeder=lSelectedBreeder->getNextSibling();
		Beagle_NonNullPointerAssertM(lSelectedBreeder);
		Beagle_NonNullPointerAssertM(lSelectedBreeder->getBreederOp());
        Beagle::Individual::Handle lBredIndiv =
		    lSelectedBreeder->getBreederOp()->breed(ioDeme, lSelectedBreeder->getFirstChild(), ioContext);
		Beagle_NonNullPointerAssertM(lBredIndiv);

        // If the breeder performed a symmetry break mutation, create many offspring from it by initializing
        // the parameters of the new nodes (which are at the end of the genotype tree vector) randomly.
        const std::string& lOperatorName = lSelectedBreeder->getBreederOp()->getName();
        if ((lOperatorName == "ENSO-MutationBreakSymmetryOp" || lOperatorName == "ENSO-MutationRandomColorSplitOp") &&
            (lBredIndiv->getFitness() == NULL || lBredIndiv->getFitness()->isValid() == false)) {
            for (unsigned int j=0; j<mNumOffSymBrkMut->getWrappedValue() && i<ioDeme.size(); j++, i++) {
                ENSO::Individual::Handle lMutatedIndiv = castHandleT<ENSO::Individual>(lIndividualAlloc->allocate());
                Beagle_NonNullPointerAssertM(lMutatedIndiv);
                lMutatedIndiv->copy(*lBredIndiv, ioContext.getSystem());

                if(lHistory != NULL) {
                    HistoryID::Handle lHID = castHandleT<HistoryID>(lBredIndiv->getMember("HistoryID"));
                    std::vector<HistoryID> lParent;
                    if(lHID != NULL) lParent.push_back(*lHID);
                    lHistory->allocateNewID(*lMutatedIndiv);
                    lHistory->trace(ioContext, lParent, lMutatedIndiv, getName(), "selection");
                }

                ENSO::Tree& lMutatedTree = *(*lMutatedIndiv)[0];
                for (unsigned int k=1; k<lMutatedTree.size(); k++) {
                    if (lMutatedTree[k].mParameters.isValid() == false) {
                        lMutatedTree[k].mParameters.randomize(ioContext);
                        lMutatedTree[k].mParameters.setValid();
                    }
                } 
                lOffsprings.push_back(lMutatedIndiv);

                if(lHistory != NULL) {
                    HistoryID::Handle lHID = castHandleT<HistoryID>(lMutatedIndiv->getMember("HistoryID"));
                    std::vector<HistoryID> lParent;
                    if(lHID != NULL) lParent.push_back(*lHID);
                    lHistory->incrementHistoryVar(*lMutatedIndiv);
                    lHistory->trace(ioContext, lParent, lMutatedIndiv, getName(), "mutation");
                }
            }
            
        }
        else {
            lOffsprings.push_back(lBredIndiv);
            i++;
        }
	}

	for(unsigned int j=0; j<lOffsprings.size(); ++j) {
        ioDeme[j] = lOffsprings[j];
        if (mFitInvalidate->getWrappedValue()) ioDeme[j]->getFitness()->setInvalid();
    }

    // Invalidate hall of fame if fitness is invalidated.
    if (mFitInvalidate->getWrappedValue()) {
        HallOfFame::Handle lHoF = ioDeme.getHallOfFame();
        if (lHoF != NULL) ioDeme.removeMember(lHoF->getName());
        lHoF = ioContext.getVivarium().getHallOfFame();
        if (lHoF != NULL) ioContext.getVivarium().removeMember(lHoF->getName());
    }

	Beagle_StackTraceEndM("void ENSO::GenerationalOp::operate(Deme& ioDeme, Context& ioContext)");
}


