
/*!
 *  \file   beagle/ENSO/src/Individual.cpp
 *  \brief  Source code of class ENSO::Individual.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"
#include "beagle/GA.hpp"

using namespace Beagle;


/*!
 *  \brief Contruct a ENSO individual.
 *  \param inN Initial number of trees.
 *  \param inModel Initial value of tree handles.
 */
ENSO::Individual::Individual(unsigned int inN, ENSO::Tree::Handle inModel) :
		Beagle::Individual(inN,inModel),
        mGraph()
{ }


/*!
 *  \brief Get exact type of the individual.
 *  \return Exact type of the individual.
 */
const std::string& ENSO::Individual::getType() const
{
	Beagle_StackTraceBeginM();
	const static std::string lType("ENSO-Individual");
	return lType;
	Beagle_StackTraceEndM("const std::string& ENSO::Individual::getType() const");
}


/*!
 *  \brief Read an individual from an XML node. 
 *  \param inIter XML iterator to read the individual from.
 *  \param ioContext Evolutionary context to use to read the individual.
 *
 *  Override the base class implementation to instantiate the phenotype graph.
 */
void ENSO::Individual::readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();
    Beagle::Individual::readWithContext(inIter, ioContext);
    makeGraph();
	Beagle_StackTraceEndM("void ENSO::Individual::readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext)");
}


/*!
 *  \brief Write content of an individual into a XML streamer.
 *  \param ioStreamer XML streamer to write the individual into.
 *  \param inIndent Whether XML output should be indented.
 *
 *  Override the base class implementation to cause indentation
 *  of the various items being written.
 */
void ENSO::Individual::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();
	// Write members of the individual
	writeMembers(ioStreamer, true);
	// Write fitness
	writeFitness(ioStreamer, true);
	// Write genotypes
	writeGenotypes(ioStreamer, true);
	Beagle_StackTraceEndM("void ENSO::Individual::writeContent(PACC::XML::Streamer&,bool) const");
}


/*!
 *  \brief Make the phenotype graph encoded by the genotype tree.
 */
void ENSO::Individual::makeGraph()
{
	Beagle_StackTraceBeginM();

	ENSO::Tree& lTree = *((*this)[0]);

    mGraph.clear();
    mGraph.resize(lTree.getNumVertices()*lTree.getNumVertices(), NULL);

    std::vector<unsigned int> lLeaves;
    lTree.getLeaves(lLeaves);
    for (unsigned int i=0; i<lLeaves.size(); i++) {
        ENSO::Node& lLeaf = lTree[lLeaves[i]];

        if (lLeaf.mParameters.isValid() == false) {
            throw Beagle_RunTimeExceptionM("Cannot make graph with invalid parameters!");
        }

        // Make a copy of the parameters for storing in the graph.
        ENSO::Parameters::Handle lParameters = new ENSO::Parameters;
        *lParameters = lLeaf.mParameters;

        // Store the index of this leaf in the parameters map.
        (*lParameters)["index"] = new GA::IntegerVector(1, lLeaves[i]);

        // If this leaf node represents vertices or edges, then we enter the parameter
        // handle for them in the graph.
        ENSO::Parameters::iterator lIter = lParameters->find("isedge");
        if ((lIter == lParameters->end()) ||                                // Vertex
            ((*castHandleT<GA::BitString>(lIter->second))[0] == true)) {    // Edge
            for (unsigned int j=0; j<lLeaf.mColor.size(); j++) {
                if (lLeaf.mColor[j] == true) mGraph[j] = lParameters;
            }
        }
    }

	Beagle_StackTraceEndM("void ENSO::Individual::makeGraph()");
}


/*!
 *  \brief Find the genotype distance between this individual and another individual.
 *  \return Distance between the two individuals.
 */
double ENSO::Individual::getDistanceGenotype(const Beagle::Individual& inRightIndividual) const
{
	Beagle_StackTraceBeginM();

	const ENSO::Individual& lRightIndividual = castObjectT<const ENSO::Individual&>(inRightIndividual);

	const ENSO::Tree& lTree1 = *((*this)[0]);
	const ENSO::Tree& lTree2 = *(lRightIndividual[0]);

    std::vector<unsigned int> lNodes1;
    std::vector<unsigned int> lNodes2;

    // Distance based on all the nodes in the tree.
//     lNodes1.resize(lTree1.size());
//     lNodes2.resize(lTree2.size());
//     for (unsigned int i=0; i<lNodes1.size(); i++) lNodes1[i] = i;
//     for (unsigned int i=0; i<lNodes2.size(); i++) lNodes2[i] = i;

    // Distance based on leaf nodes only.
    lTree1.getLeaves(lNodes1);
    lTree2.getLeaves(lNodes2);

    // Sort the node indices according to node color.
    lTree1.sortNodes(lNodes1);
    lTree2.sortNodes(lNodes2);

    unsigned int lIndex1 = 0;
    unsigned int lIndex2 = 0;
    unsigned int lDistance = 0;
    while (lIndex1 < lNodes1.size() && lIndex2 < lNodes2.size()) {
        const ENSO::Node& lNode1 = lTree1[lNodes1[lIndex1]];
        const ENSO::Node& lNode2 = lTree2[lNodes2[lIndex2]];
        if (lNode1.mColor == lNode2.mColor) {
            lIndex1++;
            lIndex2++;
        }
        else if (lNode1.mColor < lNode2.mColor) {
            lDistance += lNode1.mParameters.getSize();
            lIndex1++;
        }
        else {
            lDistance += lNode2.mParameters.getSize();
            lIndex2++;
        }
    }

    while (lIndex1 < lNodes1.size()) {
        const ENSO::Node& lNode1 = lTree1[lNodes1[lIndex1]];
        lDistance += lNode1.mParameters.getSize();
        lIndex1++;
    }

    while (lIndex2 < lNodes2.size()) {
        const ENSO::Node& lNode2 = lTree2[lNodes2[lIndex2]];
        lDistance += lNode2.mParameters.getSize();
        lIndex2++;
    }

    return (double)lDistance;
    
	Beagle_StackTraceEndM("double ENSO::Individual::getDistanceGenotype(const Beagle::Individual& inRightIndividual) const");
}
