
/*!
 *  \file   beagle/ENSO/src/InitFullySymmetricOp.cpp
 *  \brief  Source code of class ENSO::InitFullySymmetricOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"
#include "beagle/GA.hpp"

#include <sstream>

using namespace Beagle;


/*!
 *  \brief Construct an ENSO genotype tree whose phenotype graph is fully symmetric.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
ENSO::InitFullySymmetricOp::InitFullySymmetricOp(std::string inReproProbaName, std::string inName) :
		Beagle::ENSO::InitializationOp(inReproProbaName, inName),
        mBreakSymmetryOp(new ENSO::MutationBreakSymmetryOp)
{ }


/*!
 *  \brief Register the parameters of the intialization operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::InitFullySymmetricOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
    mBreakSymmetryOp->registerParams(ioSystem);
	ENSO::InitializationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Probability that the newly initialized phenotype graph is a null graph.",
		    "Float",
		    "0.0",
		    "Probability that the newly initialized phenotype graph is a null graph, having no edges."
		);
		mNullGraphProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry("enso.init.nullpb", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Index of subgroup lattice corresponding to initial symmetry.",
		    "UInt",
		    "0",
		    "Index of subgroup lattice corresponding to initial symmetry; if non-zero initial graphs are not fully symmetric."
		);
		mLatticeIndex = castHandleT<UInt>(
		                    ioSystem.getRegister().insertEntry("enso.init.latticeindex", new UInt(0), lDescription));
	}
	{
		std::ostringstream lOSS;
		lOSS << "If a lattice index for initial symmetry is specified, then this vector specifies the sorted indices ";
        lOSS << "(separated by comma, colon or semicolon) of colors in the lattice element to be made edges.";
		Register::Description lDescription(
		    "Sorted indices of edge colors in the lattice element to be made edges.",
		    "UIntArray",
		    "",
		    lOSS.str()
		);
		mEdgeIndices = castHandleT<UIntArray>(
		                    ioSystem.getRegister().insertEntry("enso.init.edges", new UIntArray, lDescription));
	}
	{
		std::ostringstream lOSS;
		lOSS << "If a lattice index for initial symmetry is specified, then this vector specifies the sorted indices ";
        lOSS << "(separated by comma, colon or semicolon) of colors in the lattice element to be made nonedges.";
		Register::Description lDescription(
		    "Sorted indices of edge colors in the lattice element to be made non-edges.",
		    "UIntArray",
		    "",
		    lOSS.str()
		);
		mNonedgeIndices = castHandleT<UIntArray>(
		                    ioSystem.getRegister().insertEntry("enso.init.nonedges", new UIntArray, lDescription));
	}
	Beagle_StackTraceEndM("void ENSO::InitFullySymmetricOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize this operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::InitFullySymmetricOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	mBreakSymmetryOp->init(ioSystem);
	ENSO::InitializationOp::init(ioSystem);
	Beagle_StackTraceEndM("void ENSO::InitFullySymmetricOp::init(Beagle::System&)");
}


/*!
 *  \brief Initialize a genotype.
 *  \param outTree Tree to initialize.
 *  \param ioContext Evolution context.
 */
void ENSO::InitFullySymmetricOp::initTree(ENSO::Tree& outTree, Context& ioContext) const
{
	Beagle_StackTraceBeginM();

	const unsigned int lNumVertices = mNumVertices->getWrappedValue();
    outTree.setNumVertices(lNumVertices);

    // Clear the tree and create a root node having two children, one representing the
    // vertices and the other the edges of the phenotype graph.
	outTree.clear();
    outTree.resize(3);

    // The root node.
    Node& lRoot = outTree[0];
    lRoot.mColor.clear();
    lRoot.mColor.resize(lNumVertices*lNumVertices, true);
    lRoot.mChildren.clear();
    lRoot.mChildren.push_back(1);
    lRoot.mChildren.push_back(2);

    // Child nodes representing vertices and edges.
    Node& lVertices = outTree[1];
    lVertices.mParent = 0;
    lVertices.mChildren.clear();
    Node& lEdges = outTree[2];
    lEdges.mParent = 0;
    lEdges.mChildren.clear();

    // Set the child node colors.
    lVertices.mColor.clear();
    lVertices.mColor.resize(lNumVertices*lNumVertices, false);
    lEdges.mColor.clear();
    lEdges.mColor.resize(lNumVertices*lNumVertices, true);
    for (unsigned int i=0; i<lNumVertices; i++) {
        unsigned int diagonal = i*lNumVertices + i;
        lVertices.mColor[diagonal] = true;
        lEdges.mColor[diagonal] = false;
    }

    outTree.recordSymmetryState(ioContext.getGeneration(), 0);

    initParameters(outTree, ioContext);

	unsigned int lLatticeIndex = mLatticeIndex->getWrappedValue();
    if (lLatticeIndex > 0) {
        mBreakSymmetryOp->breakSymmetry(outTree, lLatticeIndex, ioContext);
        
        // Randomize the parameters.
        for (unsigned int k=1; k<outTree.size(); k++) {
            if (outTree[k].mParameters.isValid() == false) {
                outTree[k].mParameters.randomize(ioContext);
                outTree[k].mParameters.setValid();
            }
        }

        // If both vertex and edge colors were split, then the size of the tree was
        // increased by the number of colors in the specified lattice index.  Furthermore,
        // the nodes for these colors are in the same order in the genotype tree.  If the
        // vertex color was not split, then nodes for only the edge colors were added to
        // the tree.  This information is used below to make the nodes in the tree correspond
        // with the colors in the lattice element when setting the isedge flag of the
        // parameters.
        if (outTree[2].mChildren.size() > 0) {
            unsigned int h = 0, i = 0;
            unsigned int j = ((outTree[1].mChildren.size() > 0) ? 0 : 1) + outTree[1].mChildren.size();
            unsigned int k = 3 + outTree[1].mChildren.size();
            for (; k < outTree.size(); k++, j++) {
                if (h < mEdgeIndices->size() && (*mEdgeIndices)[h] == j) {
                    (*castHandleT<GA::BitString>(outTree[k].mParameters["isedge"]))[0] = true;
                    h++;
                }
                else if (i < mNonedgeIndices->size() && (*mNonedgeIndices)[i] == j) {
                    (*castHandleT<GA::BitString>(outTree[k].mParameters["isedge"]))[0] = false;
                    i++;
                }
            }
        }
    }

	Beagle_LogVerboseM(
	    ioContext.getSystem().getLogger(),
	    "initialization", "Beagle::ENSO::InitFullySymmetricOp",
	    std::string("Initializing a fully symmetric phenotype graph for the ")+
	    uint2ordinal(ioContext.getGenotypeIndex()+1)+std::string(" genotype tree.")
	);

	Beagle_StackTraceEndM("void ENSO::InitFullySymmetricOp::initTree(ENSO::Tree& outTree, Context& ioContext) const");
}


/*!
 *  \brief Initialize the genes in the genotype tree.
 *  \param ioTree Tree whose genes to initialize.
 *  \param ioContext Evolution context.
 */
void ENSO::InitFullySymmetricOp::initParameters(ENSO::Tree& ioTree, Context& ioContext) const
{
	Beagle_StackTraceBeginM();

    // Initialize edges of phenotype graph such that the graph is either null or complete.
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();
	const float lNullGraphProba = mNullGraphProba->getWrappedValue();
    if (lRandomizer.rollUniform(0.0, 1.0) < lNullGraphProba) {
        ioTree[2].mParameters["isedge"] = new GA::BitString(1, false);
    }
    else {
        ioTree[2].mParameters["isedge"] = new GA::BitString(1, true);
    }

    ioTree[1].mParameters.setValid();
    ioTree[2].mParameters.setValid();
    
	Beagle_StackTraceEndM("void ENSO::InitFullySymmetricOp::initParameters(ENSO::Tree& ioTree, Context& ioContext) const");
}

