
/*!
 *  \file   beagle/ENSO/src/InitializationOp.cpp
 *  \brief  Source code of class ENSO::InitializationOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a plain new initialization operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register.
 *  \param inName Name of the initialization operator.
 */
ENSO::InitializationOp::InitializationOp(std::string inReproProbaName, std::string inName) :
		Beagle::InitializationOp(inReproProbaName, inName)
{ }


/*!
 *  \brief Register the parameters of the generic ENSO intialization operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::InitializationOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Beagle::InitializationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of vertices in the phenotype graph.",
		    "UInt",
		    "0",
		    "Number of vertices in the phenotype graph."
		);
		mNumVertices = castHandleT<UInt>(
		                    ioSystem.getRegister().insertEntry("enso.init.numvertices", new UInt(0), lDescription));
	}
	Beagle_StackTraceEndM("void ENSO::InitializationOp::registerParams(Beagle::System&)");
}


/*!
 *  \brief Initialize the genotype trees of an individual.
 *  \param outIndividual Individual to initialize.
 *  \param ioContext Evolution context.
 *  \throw Beagle::RunTimeException If the number of inputs/outputs are incorrectly set.
 */
void ENSO::InitializationOp::initIndividual(Beagle::Individual& outIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	ENSO::Individual& lIndividual = castObjectT<ENSO::Individual&>(outIndividual);

	// We just have one genotype in the individual.
    const Factory& lFactory = ioContext.getSystem().getFactory();
    Genotype::Alloc::Handle lGenotypeAlloc =
        castHandleT<Genotype::Alloc>(lFactory.getConceptAllocator("Genotype"));
    lIndividual.clear();
    lIndividual.push_back(castHandleT<Genotype>(lGenotypeAlloc->allocate()));

	Tree::Handle lOldTreeHandle = castHandleT<Tree>(ioContext.getGenotypeHandle());
	unsigned int lOldTreeIndex = ioContext.getGenotypeIndex();

	for(unsigned int i=0; i<lIndividual.size(); ++i) {
		ioContext.setGenotypeHandle(lIndividual[i]);
		ioContext.setGenotypeIndex(i);
		initTree(*lIndividual[i], ioContext);
	}

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "initialization", "Beagle::ENSO::InitializationOp",
	    std::string("Initialized individual: ")+lIndividual.serialize()
	);

	ioContext.setGenotypeIndex(lOldTreeIndex);
	ioContext.setGenotypeHandle(lOldTreeHandle);

	Beagle_StackTraceEndM("void ENSO::InitializationOp::initIndividual(Beagle::Individual& outIndividual, Beagle::Context& ioContext)");
}

