
/*!
 *  \file   beagle/ENSO/src/MutationBreakSymmetryOp.cpp
 *  \brief  Implementation of the type ENSO::MutationBreakSymmetryOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

#include <string>
#include <fstream>

#ifdef BEAGLE_HAVE_LIBZ
#include <gzstream.h>
#endif // BEAGLE_HAVE_LIBZ


using namespace Beagle;

std::vector<ENSO::MutationBreakSymmetryOp::Element> ENSO::MutationBreakSymmetryOp::mLattice;


/*!
 *  \brief Construct the symmetry break mutation operator for the ENSO genotype tree.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
ENSO::MutationBreakSymmetryOp::MutationBreakSymmetryOp(std::string inMutationPbName,
                                                       std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the symmetry break mutation operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::MutationBreakSymmetryOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Minimal symmetry break mutation probability for an individual. ";
		lOSS << "This mutation breaks symmetry by traversing the subgroup lattice ";
		lOSS << "from top to bottom.";
		Register::Description lDescription(
		    "Individual minimal symmetry break mutation probability.",
		    "Float",
		    "0.05",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
    {
        Register::Description lDescription(
            "Name of file containing the subgroup (ortibal partition) lattice.",
            "String",
            "",
            "Name of file containing the subgroup (ortibal partition) lattice."
        );
        mLatticeFile = castHandleT<String>(
                             ioSystem.getRegister().insertEntry("enso.mutbrksym.lattice", new String(""), lDescription));
    }
	{
		Register::Description lDescription(
		    "Number of vertices in the phenotype graph.",
		    "UInt",
		    "0",
		    "Number of vertices in the phenotype graph."
		);
		mNumVertices = castHandleT<UInt>(
		                    ioSystem.getRegister().insertEntry("enso.init.numvertices", new UInt(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum number of generations between symmetry break mutations.",
		    "Float",
		    "0.0",
		    "Minimum number of generations between symmetry break mutations."
		);
		mMinSymmetryBreak = castHandleT<Float>(
		                    ioSystem.getRegister().insertEntry("enso.mutbrksym.minsymbrk", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Increment per genotype node in the minimum number of generations between symmetry break mutations.",
		    "Float",
		    "0.0",
		    "Increment per genotype node in the minimum number of generations between symmetry break mutations."
		);
		mIncSymmetryBreak = castHandleT<Float>(
		                    ioSystem.getRegister().insertEntry("enso.mutbrksym.incsymbrk", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to prune the subgroup lattice based on the parity of vertices in a partition.",
		    "Bool",
		    "0",
		    "Whether to prune the subgroup lattice if the number of vertices in a partition is odd."
		);
		mPruneLattice = castHandleT<Bool>(
		                    ioSystem.getRegister().insertEntry("enso.mutbrksym.prunelattice", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether symmetry breaks should guarantee that edges (i,j) and (j,i) have the same color.",
		    "Bool",
		    "0",
		    "Whether symmetry breaks should guarantee that edges (i,j) and (j,i) have the same color to represent undirected graphs."
		);
		mUndirectedGraph = castHandleT<Bool>(
		                    ioSystem.getRegister().insertEntry("enso.mutbrksym.undirected", new Bool(0), lDescription));
	}
    {
        Register::Description lDescription(
            "Probability that the mutation will restore symmetry instead of breaking it.",
            "Float",
            "0.0",
            "Probability that the mutation will restore symmetry instead of breaking it."
        );
        mRestoreProba = castHandleT<Float>(
                             ioSystem.getRegister().insertEntry("enso.mutbrksym.restorepb", new Float(0.0), lDescription));
    }
	Beagle_StackTraceEndM("void ENSO::MutationBreakSymmetryOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply symmetry break mutation to an ENSO individual.
 *  \param ioIndividual ENSO individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool ENSO::MutationBreakSymmetryOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	ENSO::Individual& lIndividual = castObjectT<ENSO::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::ENSO::MutationBreakSymmetryOp",
	    std::string("Mutating individual with ENSO::MutationBreakSymmetryOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::ENSO::MutationBreakSymmetryOp",
	    std::string("Individual before ENSO symmetry break mutation: ")+
	    ioIndividual.serialize()
	);

	ENSO::Tree& lTree = *lIndividual[0];
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    const float lMinSymmetryBreak = mMinSymmetryBreak->getWrappedValue();
    const float lIncSymmetryBreak = mIncSymmetryBreak->getWrappedValue();

    unsigned int lOldLatticeIndex = lTree.getLatticeIndex();
    unsigned int lNewLatticeIndex = UINT_MAX;
	bool lMutated = false;

    if (lIndividual.allowSymmetryBreak(ioContext.getGeneration(), lMinSymmetryBreak, lIncSymmetryBreak)) {
        // Try to restore symmetry first.
        if (lRandomizer.rollUniform(0.0, 1.0) < mRestoreProba->getWrappedValue()) {
            lMutated = lTree.restoreSymmetryState(ioContext.getGeneration());
            lNewLatticeIndex = lTree.getLatticeIndex();
        }

        // If not mutated yet, try breaking symmetry.
        if (lMutated == false) {
            lMutated = breakSymmetry(lTree, ioContext);
            lNewLatticeIndex = lTree.getLatticeIndex();
        }

        // Try symmetry restoration again if symmetry breaking didn't work.
        if (lMutated == false && mRestoreProba->getWrappedValue() > 0.0) {
            lMutated = lTree.restoreSymmetryState(ioContext.getGeneration());
            lNewLatticeIndex = lTree.getLatticeIndex();
        }
    }

	if (lMutated) {
		// Mutation successful; log messages and return.
		std::ostringstream lOSS;
		lOSS << "Changed symmetry by moving from element " << lOldLatticeIndex;
		lOSS << " to element " << lNewLatticeIndex << " on the color lattice";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationBreakSymmetryOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationBreakSymmetryOp",
		    std::string("Individual after ENSO symmetry break mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationBreakSymmetryOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool ENSO::MutationBreakSymmetryOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Break symmetry of the genotype tree by moving to a random child of the subgroup lattice.
 *  \param ioTree The genotype tree.
 *  \param ioContext Context of the evolution.
 *  \return True if the tree is mutated, false otherwise.
 */
bool ENSO::MutationBreakSymmetryOp::breakSymmetry(ENSO::Tree& ioTree, Beagle::Context& ioContext) const
{
	Beagle_StackTraceBeginM();

	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    // Break symmetry by selecting a random child of the current lattice element.
    Element& lElement = mLattice[ioTree.getLatticeIndex()];
    if (lElement.mChildren.size() > 0) {
        unsigned int lChildIndex = lRandomizer.rollInteger(0, lElement.mChildren.size()-1);
        unsigned int lNewLatticeIndex = lElement.mChildren[lChildIndex];
        return breakSymmetry(ioTree, lNewLatticeIndex, ioContext);
    }
    else {
        return false;
    }

	Beagle_StackTraceEndM("void ENSO::MutationBreakSymmetryOp::breakSymmetry(ENSO::Tree& ioTree, Beagle::Context& ioContext) const");
}


/*!
 *  \brief Break symmetry of the genotype tree by moving to a new point on the subgroup lattice.
 *  \param ioTree The genotype tree.
 *  \param inNewLatticeIndex Index of the new lattice point.
 *  \param ioContext Context of the evolution.
 *  \return True if the tree is mutated, false otherwise.
 */
bool ENSO::MutationBreakSymmetryOp::breakSymmetry(ENSO::Tree& ioTree, unsigned int inNewLatticeIndex, Beagle::Context& ioContext) const
{
	Beagle_StackTraceBeginM();

	bool lMutated = false;
    Element& lChildElement = mLattice[inNewLatticeIndex];
    std::vector< std::vector<bool> > lChildColors(lChildElement.mColors);  // make a copy since it is modified below.

    unsigned int lNumVertices = mNumVertices->getWrappedValue();

    bool lVertexParity = false;
    if (mPruneLattice->getWrappedValue()) {
        // Check if the child element is in the pruned part of the lattice.
        // Symmetry is not broken if the number of vertices in any of its colors is odd.
        for (unsigned int j=0; j<lChildColors.size() && lVertexParity==false; j++) {
            std::vector<bool>& lChildColor = lChildColors[j];
            for (unsigned int k=0; k<lChildColor.size(); k+=lNumVertices+1) {
                lVertexParity = lVertexParity != (bool)(lChildColor[k]);
            }
        }
    }

    bool lSameEdgeColor = true;
    if (mUndirectedGraph->getWrappedValue()) {
        lSameEdgeColor = false;
        // Check if the child element has at least one color with edges (i,j) and (j,i).
        // Symmetry is not broken if such a color does not exist.
        for (unsigned int j=0; j<lChildColors.size() && lSameEdgeColor==false; j++) {
            std::vector<bool>& lChildColor = lChildColors[j];
            lSameEdgeColor = true;
            for (unsigned int k=0; k<lNumVertices && lSameEdgeColor==true; k++) {
                for (unsigned int i=k+1; i<lNumVertices && lSameEdgeColor==true; i++) {
                    lSameEdgeColor = lChildColor[k*lNumVertices+i] == lChildColor[i*lNumVertices+k];
                }
            }
        }
    }

    if (lVertexParity == false && lSameEdgeColor == true) {
        std::vector<unsigned int> lLeaves;
        ioTree.getLeaves(lLeaves);
        for (unsigned int i=0; i<lLeaves.size(); i++) {
            const std::vector<bool> lLeafColor(ioTree[lLeaves[i]].mColor);  // make a copy since ioTree is modified below.
            for (unsigned int j=0; j<lChildColors.size(); j++) {
                std::vector<bool>& lChildColor = lChildColors[j];  // make lChildColor a reference since it is cleared below.

                // Find out if the two colors are equal and if they intersect.
                bool lEqual = true;
                bool lIntersect = false;
                for (unsigned int k=0; k<lChildColor.size() && (lEqual==true || lIntersect==false); k++) {
                    lEqual = lEqual && (lChildColor[k] == lLeafColor[k]);
                    lIntersect = lIntersect || (lChildColor[k] && lLeafColor[k]);
                }

                // If the two colors are the same, then the child relationship does not hold.
                // Otherwise, the color is a child of the leaf color if the two colors intersect.
                // Since a color can have only one parent, this color is cleared to optimize
                // the above loop that finds intersection.
                if (lEqual) {
                    lChildColor.clear();
                }
                else if (lIntersect) {
                    ioTree.push_back(ENSO::Node(lChildColor, lLeaves[i]));
                    ioTree.back().mParameters = ioTree[lLeaves[i]].mParameters;
                    ioTree.back().mParameters.setInvalid();
                    ioTree[lLeaves[i]].mChildren.push_back(ioTree.size()-1);
                    lChildColor.clear();
                }
            }
        }

        lMutated = true;
        ioTree.recordSymmetryState(ioContext.getGeneration(), inNewLatticeIndex);
    }

    return lMutated;

	Beagle_StackTraceEndM("void ENSO::MutationBreakSymmetryOp::breakSymmetry(ENSO::Tree& ioTree, unsigned int inNewLatticeIndex, Beagle::Context& ioContext) const");
}


/*!
 *  \brief Initialize this operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::MutationBreakSymmetryOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

    std::string lLatticeFile = mLatticeFile->getWrappedValue();
    if (mLattice.size() == 0 && lLatticeFile != "") {
        std::ostringstream lOSS;
        lOSS << "Reading subgroup lattice from file '" << lLatticeFile << "'";
        Beagle_LogDebugM(
            ioSystem.getLogger(),
            "mutation", "Beagle::ENSO::MutationBreakSymmetryOp",
            lOSS.str()
        );

        readLattice(lLatticeFile);
    }

	Beagle_StackTraceEndM("void ENSO::MutationBreakSymmetryOp::init(Beagle::System&)");
}


/*!
 *  \brief Read the subgroup (orbital partition) lattice from file.
 *  \param inLatticeFile XML file containing the subgroup lattice.
 */
void ENSO::MutationBreakSymmetryOp::readLattice(const std::string& inLatticeFile)
{
	Beagle_StackTraceBeginM();

#ifdef BEAGLE_HAVE_LIBZ
    igzstream lLatticeFileStream(inLatticeFile.c_str());
#else // BEAGLE_HAVE_LIBZ
    std::ifstream lLatticeFileStream(inLatticeFile.c_str());
#endif // BEAGLE_HAVE_LIBZ

    PACC::XML::Document lParser;
    lParser.parse(lLatticeFileStream, inLatticeFile);
    lLatticeFileStream.close();

    PACC::XML::ConstIterator lIter = lParser.getFirstDataTag();
	if((lIter->getType() != PACC::XML::eData) || (lIter->getValue() != "Lattice")) {
		throw Beagle_IOExceptionNodeM(*lIter, "tag <Lattice> expected!");
	}

    mLattice.clear();
	std::string lSizeText = lIter->getAttribute("size");
	if(lSizeText.empty()==false)
		mLattice.resize(str2uint(lSizeText));

	std::string lNumVerticesText = lIter->getAttribute("numVertices");
	if(lNumVerticesText.empty()==false) {
		if (str2uint(lNumVerticesText) != mNumVertices->getWrappedValue()) {
            throw Beagle_IOExceptionNodeM(*lIter, "Number of graph vertices in lattice file is inconsistent with configuration file!");
        }
    }

	for(PACC::XML::ConstIterator lChild=lIter->getFirstChild(); lChild; ++lChild) {
		if((lChild->getType() == PACC::XML::eData) && (lChild->getValue() == "Element")) {
            std::string lIndexText = lChild->getAttribute("index");
            if(lIndexText.empty()) {
                std::ostringstream lOSS;
                lOSS << "Element with unknown index in subgroup lattice!";
                throw Beagle_IOExceptionNodeM(*lChild, lOSS.str());
            }
            Element& lElement = mLattice[str2uint(lIndexText)];

            std::string lGroupSizeText = lChild->getAttribute("groupSize");
            if(lGroupSizeText.empty()==false)
                lElement.mGroupSize = str2uint(lGroupSizeText);

            for(PACC::XML::ConstIterator lChild2=lChild->getFirstChild(); lChild2; ++lChild2) {
                if(lChild2->getType() == PACC::XML::eData) {
                    if(lChild2->getValue() == "Colors") {
                        readColors(lChild2, lElement.mColors);
                    }
                    else if(lChild2->getValue() == "Children") {
                        readChildren(lChild2, lElement.mChildren);
                    }
                }
            }
        }
	}

	Beagle_StackTraceEndM("void ENSO::MutationBreakSymmetryOp::readLattice(const std::string& inLatticeFile)");
}


/*!
 *  \brief Read the vector of colors for a lattice element from an XML subtree.
 *  \param inIter XML iterator to use to read the color vector from.
 *  \param outColors The colors will be read into this vector.
 */
void ENSO::MutationBreakSymmetryOp::readColors(PACC::XML::ConstIterator inIter, std::vector< std::vector<bool> >& outColors)
{
	Beagle_StackTraceBeginM();

    outColors.clear();
	std::string lSizeText = inIter->getAttribute("size");
	if(lSizeText.empty()==false)
		outColors.resize(str2uint(lSizeText));

	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if((lChild->getType() == PACC::XML::eData) && (lChild->getValue() == "Color")) {
            std::string lIndexText = lChild->getAttribute("index");
            if(lIndexText.empty()) {
                std::ostringstream lOSS;
                lOSS << "Element with unknown index in color of subgroup lattice!";
                throw Beagle_IOExceptionNodeM(*lChild, lOSS.str());
            }
            std::vector<bool>& lColor = outColors[str2uint(lIndexText)];

            PACC::XML::ConstIterator lChild2 = lChild->getFirstChild();
            if((!lChild2) || (lChild2->getType()!=PACC::XML::eString))
                throw Beagle_IOExceptionNodeM(*lChild2, "Expected bit string for color in subgroup lattice!");

            lColor.clear();
            std::istringstream lISS(lChild2->getValue());
            while(lISS.eof() == false) {
                char lBuf = '\0';
                lISS >> lBuf;
                if(lBuf=='\0') break;
                if((lBuf != '0') && (lBuf != '1')) {
                    throw Beagle_IOExceptionNodeM(*lChild2, "Bit string value is not 0 or 1 in color!");
                }
                lColor.push_back((lBuf == '1') ? true : false);
            }
        }
    }

	Beagle_StackTraceEndM("void ENSO::MutationBreakSymmetryOp::readColors(PACC::XML::ConstIterator inIter, std::vector< std::vector<bool> >& outColors)");
}


/*!
 *  \brief Read the children of a lattice element from an XML subtree.
 *  \param inIter XML iterator to use to read the children vector from.
 *  \param outChildren Indices of children will be read into this vector.
 */
void ENSO::MutationBreakSymmetryOp::readChildren(PACC::XML::ConstIterator inIter, std::vector<unsigned int>& outChildren)
{
	Beagle_StackTraceBeginM();

    PACC::XML::ConstIterator lChild = inIter->getFirstChild();
	if(lChild->getType()!=PACC::XML::eString)
		throw Beagle_IOExceptionNodeM(*lChild,"Expected vector of unsigned ints for children of element!");

	outChildren.clear();
	std::istringstream lISS(lChild->getValue());
	while(lISS.good()) {
		unsigned int lValue;
		lISS >> lValue;
		outChildren.push_back(lValue);
		if(lISS.good()==false) break;
		int lDelim=lISS.get();
		if((lISS.good()==false) || (lDelim==-1)) break;
	}

	Beagle_StackTraceEndM("void ENSO::MutationBreakSymmetryOp::readChildren(PACC::XML::ConstIterator inIter, std::vector<unsigned int>& outChildren)");
}


/*!
 *  \brief Write the subgroup (orbital partition) lattice to file.
 *  \param inLatticeFile The file to write the subgroup lattice in XML format.
 */
void ENSO::MutationBreakSymmetryOp::writeLattice(const std::string& inLatticeFile)
{
	Beagle_StackTraceBeginM();

    std::string lFilename = inLatticeFile;
#ifdef BEAGLE_HAVE_LIBZ
    lFilename += ".gz";
	std::ostream* lOS = new ogzstream(lFilename.c_str());
#else // BEAGLE_HAVE_LIBZ
	std::ostream* lOS = new std::ofstream(lFilename.c_str());
#endif // BEAGLE_HAVE_LIBZ

	try {
		PACC::XML::Streamer lStreamer(*lOS);
		lStreamer.openTag("Lattice");
		lStreamer.insertAttribute("size", uint2str(mLattice.size()));
		lStreamer.insertAttribute("numVertices", uint2str(mNumVertices->getWrappedValue()));
        for (unsigned int i=0; i<mLattice.size(); i++) {
            Element& lElement = mLattice[i];
            lStreamer.openTag("Element");
            lStreamer.insertAttribute("index", uint2str(i));
            lStreamer.insertAttribute("groupSize", uint2str(lElement.mGroupSize));
            std::vector< std::vector<bool> >& lColors = lElement.mColors;
            lStreamer.openTag("Colors");
            lStreamer.insertAttribute("size", uint2str(lColors.size()));
            for (unsigned int j=0; j<lColors.size(); j++) {
                lStreamer.openTag("Color", false);
                lStreamer.insertAttribute("index", uint2str(j));
                std::ostringstream lOSS;
                for (unsigned int k=0; k<lColors[j].size(); k++) lOSS << (lColors[j][k] ? '1' : '0');
                lStreamer.insertStringContent(lOSS.str());
                lStreamer.closeTag();
            }
            lStreamer.closeTag();
            std::vector<unsigned int>& lChildren = lElement.mChildren;
            if (lChildren.size() > 0) {
                lStreamer.openTag("Children", false);
                lStreamer.insertAttribute("size", uint2str(lChildren.size()));
                std::ostringstream lOSS;
                for (unsigned int j=0; j<lChildren.size(); j++) {
                    lOSS << lChildren[j];
                    if (j != (lChildren.size()-1)) lOSS << ';';
                }
                lStreamer.insertStringContent(lOSS.str());
                lStreamer.closeTag();
            }
            lStreamer.closeTag();
        }
        lStreamer.closeTag();
        *lOS << std::endl;
	} catch (...) {
		delete lOS;
		throw;
	}
	delete lOS;
	lOS = NULL;

	Beagle_StackTraceEndM("void ENSO::MutationBreakSymmetryOp::writeLattice(const std::string& inLatticeFile)");
}

