
/*!
 *  \file   beagle/ENSO/src/MutationComparatorsOp.cpp
 *  \brief  Implementation of the type ENSO::MutationComparatorsOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

using namespace Beagle;


/*!
 *  \brief Construct the parameter mutation operator for the ENSO genotype tree.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
ENSO::MutationComparatorsOp::MutationComparatorsOp(std::string inMutationPbName,
                                                   std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Apply comparator mutation to an ENSO individual.
 *  \param ioIndividual ENSO individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool ENSO::MutationComparatorsOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	ENSO::Individual& lIndividual = castObjectT<ENSO::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::ENSO::MutationComparatorsOp",
	    std::string("Mutating individual with ENSO::MutationComparatorsOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::ENSO::MutationComparatorsOp",
	    std::string("Individual before ENSO parameter mutation: ")+
	    ioIndividual.serialize()
	);

	ENSO::Tree& lTree = *lIndividual[0];
	bool lMutated = addComparators(lTree, ioContext);

	if (lMutated) {
		// Mutation successful; log messages and return.
        GA::IntegerVector& lNodes = castObjectT<GA::IntegerVector&>(*(lTree[0].mParameters["nodes"]));
        std::ostringstream lOSS;
        lOSS << "Added node " << lNodes.back() << " to the list of comparators";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationComparatorsOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationComparatorsOp",
		    std::string("Individual after ENSO parameter mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationComparatorsOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool ENSO::MutationComparatorsOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Add comparators to the genotype tree if possible.
 *  \param ioTree The genotype tree.
 *  \param ioContext Context of the evolution.
 *  \return True if the tree is mutated, false otherwise.
 */
bool ENSO::MutationComparatorsOp::addComparators(ENSO::Tree& ioTree, Beagle::Context& ioContext) const
{
	Beagle_StackTraceBeginM();

	const unsigned int lNumVertices = ioTree.getNumVertices();

    // Find the nodes representing parallel edges (comparisons that can be performed in parallel)
    // and choose such a random node to add to the list of comparators.
    std::vector<unsigned int> lNodeIndices;
    ioTree.getParallelEdgeLeaves(lNodeIndices);
	bool lMutated = false;
    if (lNodeIndices.size() > 0) {
        unsigned int lChosenIndex = ioContext.getSystem().getRandomizer().rollInteger(0, lNodeIndices.size()-1);
        GA::IntegerVector& lComparators = castObjectT<GA::IntegerVector&>(*(ioTree[0].mParameters["comparators"]));
        GA::IntegerVector& lLevels = castObjectT<GA::IntegerVector&>(*(ioTree[0].mParameters["levels"]));
        GA::IntegerVector& lNodes = castObjectT<GA::IntegerVector&>(*(ioTree[0].mParameters["nodes"]));

        // Make sure the last node added is not the same as the chosen node.
        if (lNodes.size() == 0 || (unsigned int)lNodes.back() != lNodeIndices[lChosenIndex]) {
            const std::vector<bool>& lColor = ioTree[lNodeIndices[lChosenIndex]].mColor;
            GA::IntegerVector lNewComparators;
            for (unsigned int j = 0; j < lNumVertices; j++) {
                // Add comparators from the upper triangle of the adjacency matrix.
                for (unsigned int k = j+1; k < lNumVertices; k++) {
                    if (lColor[j*lNumVertices+k]) {
                        lNewComparators.push_back(j);
                        lNewComparators.push_back(k);
                    }
                }
                // Add comparators from the lower triangle of the adjacency matrix only if they
                // are not specified in the upper triangle.
                for (unsigned int k = 0; k < j; k++) {
                    if (lColor[j*lNumVertices+k] && !lColor[k*lNumVertices+j]) {
                        lNewComparators.push_back(k);
                        lNewComparators.push_back(j);
                    }
                }
            }

            // Make sure each new comparator is not a repetition, i.e. if the same comparator has
            // been added before, then the new comparator is added only if comparators in between
            // have at least one of the comparator lines in common.
            unsigned int lRepeated = UINT_MAX;
            for (unsigned int j = 0; j < lNewComparators.size() && lRepeated == UINT_MAX; j += 2) {
                for (unsigned int k = 0; k < lComparators.size(); k += 2) {
                    if (lNewComparators[j] == lComparators[k] && lNewComparators[j+1] == lComparators[k+1]) {
                        lRepeated = k;
                    }
                    else if (lRepeated < UINT_MAX && 
                             (lNewComparators[j] == lComparators[k] || lNewComparators[j] == lComparators[k+1] ||
                              lNewComparators[j+1] == lComparators[k] || lNewComparators[j+1] == lComparators[k+1])) {
                        lRepeated = UINT_MAX;
                    }
                }
            }

            // Add the new comparators if none of them is a repetition.
            if (lRepeated == UINT_MAX) {
                lNodes.push_back(lNodeIndices[lChosenIndex]);
                lLevels.push_back(lComparators.size());
                lComparators.insert(lComparators.end(), lNewComparators.begin(), lNewComparators.end());
                lMutated = true;
            }
            // If any of the new set of comparators is a repetition of the last set of comparators,
            // then remove the last set and add the new set.
            else if (lRepeated > (unsigned int)lLevels.back()) {
                lNodes.pop_back();
                lNodes.push_back(lNodeIndices[lChosenIndex]);
                lComparators.resize(lLevels.back());
                lComparators.insert(lComparators.end(), lNewComparators.begin(), lNewComparators.end());
                lMutated = true;
            }
        }
    }

    return lMutated;

	Beagle_StackTraceEndM("bool ENSO::MutationComparatorsOp::addComparators(ENSO::Tree& ioTree, Beagle::Context& ioContext) const");
}
