
/*!
 *  \file   beagle/ENSO/src/MutationParametersOp.cpp
 *  \brief  Implementation of the type ENSO::MutationParametersOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

using namespace Beagle;


/*!
 *  \brief Construct the parameter mutation operator for the ENSO genotype tree.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
ENSO::MutationParametersOp::MutationParametersOp(std::string inMutationPbName,
                                                 std::string inParametersKeyName,
                                                 std::string inApplyToVertexName,
                                                 std::string inMutateValueProbaName,
                                                 std::string inMutateGaussSigmaName,
                                                 std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName),
        mParametersKeyName(inParametersKeyName),
        mApplyToVertexName(inApplyToVertexName),
		mMutateValueProbaName(inMutateValueProbaName),
		mMutateGaussSigmaName(inMutateGaussSigmaName)
{ }


/*!
 *  \brief Register the parameters of the parameter mutation operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::MutationParametersOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Parameter mutation probability for an individual. This mutation ";
		lOSS << "perturbs parameters in the ENSO genotype tree.";
		Register::Description lDescription(
		    "Individual parameter mutation probability.",
		    "Float",
		    "0.05",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	{
		std::ostringstream lOSS;
		lOSS << "Key of the entry in the parameters map for which values are ";
		lOSS << "modified by the parameter mutation operator.";
		Register::Description lDescription(
		    "Name (key) of the parameters entry to be mutated.",
		    "String",
		    "",
		    lOSS.str()
		);
		mParametersKey = castHandleT<String>(
		                     ioSystem.getRegister().insertEntry(mParametersKeyName, new String(""), lDescription));
	}
	{
		std::ostringstream lOSS;
		lOSS << "Whether this parameter mutation operator should be applied ";
		lOSS << "to the vertices (1) or edges (0) of the phenotype graph.";
		Register::Description lDescription(
		    "Whether this parameter mutation should be applied to vertices or edges.",
		    "Bool",
		    "0",
		    lOSS.str()
		);
		mApplyToVertex = castHandleT<Bool>(
		                     ioSystem.getRegister().insertEntry(mApplyToVertexName, new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Probability that any single parameter value will be mutated.",
		    "Float",
		    "0.05",
		    "Probability that any single value in the selected parameters map entry will be mutated."
		);
		mMutateValueProba = castHandleT<Float>(
                             ioSystem.getRegister().insertEntry(mMutateValueProbaName, new Float(0.05f), lDescription));
	}
	{
		Register::Description lDescription(
		    "Sigma of Gaussian mutation applied to floating point parameters.",
		    "Float",
		    "1.0",
		    "The standard deviation of Gaussian mutation applied to floating point parameters."
		);
		mMutateGaussSigma = castHandleT<Float>(
                             ioSystem.getRegister().insertEntry(mMutateGaussSigmaName, new Float(1.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum number of generations between edge/non-edge mutations.",
		    "Float",
		    "0.0",
		    "Minimum number of generations between edge/non-edge mutations."
		);
		mMinIsEdgeMut = castHandleT<Float>(
		                    ioSystem.getRegister().insertEntry("enso.mutparams.minisedge", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Increment per genotype node in the minimum number of generations between edge/non-edge mutations.",
		    "Float",
		    "0.0",
		    "Increment per genotype node in the minimum number of generations between edge/non-edge mutations."
		);
		mIncIsEdgeMut = castHandleT<Float>(
		                    ioSystem.getRegister().insertEntry("enso.mutparams.incisedge", new Float(0.0), lDescription));
	}
	Beagle_StackTraceEndM("void ENSO::MutationParametersOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply parameter mutation to an ENSO individual.
 *  \param ioIndividual ENSO individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool ENSO::MutationParametersOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	ENSO::Individual& lIndividual = castObjectT<ENSO::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::ENSO::MutationParametersOp",
	    std::string("Mutating individual with ENSO::MutationParametersOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::ENSO::MutationParametersOp",
	    std::string("Individual before ENSO parameter mutation: ")+
	    ioIndividual.serialize()
	);

	ENSO::Tree& lTree = *lIndividual[0];
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    const std::string lParametersKey = mParametersKey->getWrappedValue();
    const bool lApplyToVertex = mApplyToVertex->getWrappedValue();
    const float lMutateValueProba = mMutateValueProba->getWrappedValue();
	const float lMutateGaussSigma = mMutateGaussSigma->getWrappedValue();
    const float lMinIsEdgeMut = mMinIsEdgeMut->getWrappedValue();
    const float lIncIsEdgeMut = mIncIsEdgeMut->getWrappedValue();

    std::vector<unsigned int> lNodeIndices;
    if (lApplyToVertex == true) {
        lTree.getVertexLeaves(lNodeIndices);
    }
    else {
        lTree.getEdgeLeaves(lNodeIndices);
    }

	bool lMutated = false;
    std::ostringstream lOSS;
    lOSS << "Mutated parameters (key = " << lParametersKey << ") of nodes with colors:";

    for (unsigned int j=0; j<lNodeIndices.size(); j++) {
        ENSO::Node& lNode = lTree[lNodeIndices[j]];
        ENSO::Parameters& lParameters = lNode.mParameters;
        ENSO::Parameters::iterator lParamsIter = lParameters.find(lParametersKey);
        if (lParamsIter != lParameters.end() &&
            (lParametersKey != "isedge" || lIndividual.allowIsEdgeMutation(ioContext.getGeneration(), lMinIsEdgeMut, lIncIsEdgeMut))) {
            Genotype::Handle& lEntry = lParamsIter->second;
            const std::string& lEntryType = lEntry->getType();
            if (lEntryType == "GA-BitString") {
                GA::BitString::Handle lBitString = castHandleT<GA::BitString>(lEntry);
                for (unsigned int i=0; i<lBitString->size(); i++) {
                    if (lRandomizer.rollUniform(0.0, 1.0) < lMutateValueProba) {
                        (*lBitString)[i] = !(*lBitString)[i];
                        if (lParametersKey == "isedge") {
                            if ((*lBitString)[i]) lParameters.randomize(ioContext);
                            lTree.setLastIsEdgeMutationGeneration(ioContext.getGeneration());
                        }
                        lMutated = true;
                        lOSS << " ";
                        for (unsigned int k=0; k<lNode.mColor.size(); k++) lOSS << (lNode.mColor[k] ? '1' : '0');
                    }
                }
            }
            else if (lEntryType == "GA-FloatVector") {
                if (lApplyToVertex || (*castHandleT<GA::BitString>(lParameters["isedge"]))[0]) {
                    GA::FloatVector::Handle lFloatVector = castHandleT<GA::FloatVector>(lEntry);
                    for (unsigned int i=0; i<lFloatVector->size(); i++) {
                        if (lRandomizer.rollUniform(0.0, 1.0) < lMutateValueProba) {
                            (*lFloatVector)[i] = lRandomizer.rollGaussian((*lFloatVector)[i], lMutateGaussSigma);
                            lMutated = true;
                            lOSS << " ";
                            for (unsigned int k=0; k<lNode.mColor.size(); k++) lOSS << (lNode.mColor[k] ? '1' : '0');
                        }
                    }
                }
            }
            else {
                std::ostringstream lOSS1;
                lOSS1 << "ENSO::MutationParametersOp not defined for parameters of type " << lEntryType;
                throw Beagle_RunTimeExceptionM(lOSS1.str());
            }
        }
    }

	if (lMutated) {
		// Mutation successful; log messages and return.
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationParametersOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationParametersOp",
		    std::string("Individual after ENSO parameter mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationParametersOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool ENSO::MutationParametersOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Read a mutation operator from XML subtree.
 *  \param inIter XML iterator to use to read mutation operator.
 *  \param ioSystem Evolutionary system.
 */
void ENSO::MutationParametersOp::readWithSystem(PACC::XML::ConstIterator inIter, System& ioSystem)
{
	Beagle_StackTraceBeginM();
	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!=getName())) {
		std::ostringstream lOSS;
		lOSS << "tag <" << getName() << "> expected!" << std::flush;
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}
	std::string lMutationPbReadName = inIter->getAttribute("mutationpb");
	if(lMutationPbReadName.empty() == false) mMutationPbName = lMutationPbReadName;
	std::string lParametersKeyReadName = inIter->getAttribute("paramskey");
	if(lParametersKeyReadName.empty() == false) mParametersKeyName = lParametersKeyReadName;
	std::string lApplyToVertexReadName = inIter->getAttribute("applyvert");
	if(lApplyToVertexReadName.empty() == false) mApplyToVertexName = lApplyToVertexReadName;
	std::string lMutateValueProbaReadName = inIter->getAttribute("valpb");
	if(lMutateValueProbaReadName.empty() == false) mMutateValueProbaName = lMutateValueProbaReadName;
	std::string lMutateGaussSigmaReadName = inIter->getAttribute("sigma");
	if(lMutateGaussSigmaReadName.empty() == false) mMutateGaussSigmaName = lMutateGaussSigmaReadName;
	Beagle_StackTraceEndM("void ENSO::MutationParametersOp::readWithSystem(PACC::XML::ConstIterator, System&)");
}


/*!
 *  \brief Write mutation operator into XML streamer.
 *  \param ioStreamer XML streamer to write mutation operator into.
 *  \param inIndent Whether XML output should be indented.
 */
void ENSO::MutationParametersOp::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();
	Beagle::MutationOp::writeContent(ioStreamer, inIndent);
	ioStreamer.insertAttribute("paramskey", mParametersKeyName);
	ioStreamer.insertAttribute("applyvert", mApplyToVertexName);
	ioStreamer.insertAttribute("valpb", mMutateValueProbaName);
	ioStreamer.insertAttribute("sigma", mMutateGaussSigmaName);
	Beagle_StackTraceEndM("void ENSO::MutationParametersOp::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}
