
/*!
 *  \file   beagle/ENSO/src/MutationRandomColorSplitOp.cpp
 *  \brief  Implementation of the type ENSO::MutationRandomColorSplitOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

using namespace Beagle;


/*!
 *  \brief Construct the random color split mutation operator for the ENSO genotype tree.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
ENSO::MutationRandomColorSplitOp::MutationRandomColorSplitOp(std::string inMutationPbName,
                                                             std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the random color split mutation operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::MutationRandomColorSplitOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		Register::Description lDescription(
		    "Individual random color split mutation probability.",
		    "Float",
		    "0.05",
		    "Random color split mutation probability for an individual."
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of vertices in the phenotype graph.",
		    "UInt",
		    "0",
		    "Number of vertices in the phenotype graph."
		);
		mNumVertices = castHandleT<UInt>(
		                    ioSystem.getRegister().insertEntry("enso.init.numvertices", new UInt(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum number of generations between symmetry break mutations.",
		    "Float",
		    "0.0",
		    "Minimum number of generations between symmetry break mutations."
		);
		mMinSymmetryBreak = castHandleT<Float>(
		                    ioSystem.getRegister().insertEntry("enso.mutbrksym.minsymbrk", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Increment per genotype node in the minimum number of generations between symmetry break mutations.",
		    "Float",
		    "0.0",
		    "Increment per genotype node in the minimum number of generations between symmetry break mutations."
		);
		mIncSymmetryBreak = castHandleT<Float>(
		                    ioSystem.getRegister().insertEntry("enso.mutbrksym.incsymbrk", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to prune the subgroup lattice based on the parity of vertices in a partition.",
		    "Bool",
		    "0",
		    "Whether to prune the subgroup lattice if the number of vertices in a partition is odd."
		);
		mPruneLattice = castHandleT<Bool>(
		                    ioSystem.getRegister().insertEntry("enso.mutbrksym.prunelattice", new Bool(0), lDescription));
	}
    {
        Register::Description lDescription(
            "Probability that the mutation will restore symmetry instead of breaking it.",
            "Float",
            "0.0",
            "Probability that the mutation will restore symmetry instead of breaking it."
        );
        mRestoreProba = castHandleT<Float>(
                             ioSystem.getRegister().insertEntry("enso.mutbrksym.restorepb", new Float(0.0), lDescription));
    }
	Beagle_StackTraceEndM("void ENSO::MutationRandomColorSplitOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply random color split mutation to an ENSO individual.
 *  \param ioIndividual ENSO individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool ENSO::MutationRandomColorSplitOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	ENSO::Individual& lIndividual = castObjectT<ENSO::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::ENSO::MutationRandomColorSplitOp",
	    std::string("Mutating individual with ENSO::MutationRandomColorSplitOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::ENSO::MutationRandomColorSplitOp",
	    std::string("Individual before ENSO random color split mutation: ")+
	    ioIndividual.serialize()
	);

	ENSO::Tree& lTree = *lIndividual[0];
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    const float lMinSymmetryBreak = mMinSymmetryBreak->getWrappedValue();
    const float lIncSymmetryBreak = mIncSymmetryBreak->getWrappedValue();

    // Mutate by randomly splitting the color of nodes in the tree.
	bool lMutated = false;
    unsigned int lOldTreeSize = lTree.size();
    if (lIndividual.allowSymmetryBreak(ioContext.getGeneration(), lMinSymmetryBreak, lIncSymmetryBreak)) {
        // Try to restore previous colors first.
        if (lRandomizer.rollUniform(0.0, 1.0) < mRestoreProba->getWrappedValue()) {
            lMutated = lTree.restoreSymmetryState(ioContext.getGeneration());
        }

        // If not mutated yet, try splitting colors.
        if (lMutated == false) {
            std::vector< std::vector<bool> > lChildColors;
            splitNodeColors(lTree, lChildColors, ioContext);

            // The child is pruned if the number of vertices in any of its colors is odd.
            bool lVertexParity = false;
            unsigned int lNumVertices = mNumVertices->getWrappedValue();
            if (mPruneLattice->getWrappedValue()) {
                for (unsigned int j=0; j<lChildColors.size() && lVertexParity==false; j++) {
                    std::vector<bool>& lChildColor = lChildColors[j];
                    for (unsigned int k=0; k<lChildColor.size(); k+=lNumVertices+1) {
                        lVertexParity = lVertexParity != (bool)(lChildColor[k]);
                    }
                }
            }

            if (lVertexParity == false) {
                std::vector<unsigned int> lLeaves;
                lTree.getLeaves(lLeaves);
                for (unsigned int i=0; i<lLeaves.size(); i++) {
                    const std::vector<bool> lLeafColor(lTree[lLeaves[i]].mColor);  // make a copy since lTree is modified below.
                    for (unsigned int j=0; j<lChildColors.size(); j++) {
                        std::vector<bool>& lChildColor = lChildColors[j];  // make lChildColor a reference since it is cleared below.

                        // Find out if the two colors are equal and if they intersect.
                        bool lEqual = true;
                        bool lIntersect = false;
                        for (unsigned int k=0; k<lChildColor.size() && (lEqual==true || lIntersect==false); k++) {
                            lEqual = lEqual && (lChildColor[k] == lLeafColor[k]);
                            lIntersect = lIntersect || (lChildColor[k] && lLeafColor[k]);
                        }

                        // If the two colors are the same, then the child relationship does not hold.
                        // Otherwise, the color is a child of the leaf color if the two colors intersect.
                        // Since a color can have only one parent, this color is cleared to optimize
                        // the above loop that finds intersection.
                        if (lEqual) {
                            lChildColor.clear();
                        }
                        else if (lIntersect) {
                            lTree.push_back(ENSO::Node(lChildColor, lLeaves[i]));
                            lTree.back().mParameters = lTree[lLeaves[i]].mParameters;
                            lTree.back().mParameters.setInvalid();
                            lTree[lLeaves[i]].mChildren.push_back(lTree.size()-1);
                            lChildColor.clear();
                        }
                    }
                }

                lMutated = true;
                lTree.recordSymmetryState(ioContext.getGeneration(), 0);
            }
        }

        // Try symmetry restoration again if random color spliting didn't work.
        if (lMutated == false && mRestoreProba->getWrappedValue() > 0.0) {
            lMutated = lTree.restoreSymmetryState(ioContext.getGeneration());
        }
    }

    unsigned int lNewTreeSize = lTree.size();
	if (lMutated) {
		// Mutation successful; log messages and return.
		std::ostringstream lOSS;
		lOSS << "Genotype tree size changed from " << lOldTreeSize << " to  " << lNewTreeSize;
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationRandomColorSplitOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationRandomColorSplitOp",
		    std::string("Individual after ENSO random color split mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::ENSO::MutationRandomColorSplitOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool ENSO::MutationRandomColorSplitOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Split a random number of node colors based on the number of ones they contain.
 *  \param inTree The genotype tree.
 *  \param outChildColors The newly created child colors.
 *  \param ioContext Evolutionary context used.
 */
void ENSO::MutationRandomColorSplitOp::splitNodeColors(ENSO::Tree& inTree, std::vector< std::vector<bool> >& outChildColors, Beagle::Context& ioContext) const
{
    Beagle_StackTraceBeginM();

    outChildColors.clear();
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    std::vector<unsigned int> lNodeIndices;
    inTree.getLeaves(lNodeIndices);

    // Return immediately if no node colors can be split further, i.e. each leaf node
    // contains only one vertex or edge of the phenotype graph.
    unsigned int lNumVertices = mNumVertices->getWrappedValue();
    if (lNodeIndices.size() == lNumVertices*lNumVertices) {
        return;
    }

    // Split the colors of a random number of leaf nodes of the genotype tree.
    unsigned int lNumColorsToSplit = lRandomizer.rollInteger(1, lNodeIndices.size());
    for (unsigned int i=0; i<lNumColorsToSplit; i++) {
        // Select a node in proportion to the number of ones in its color.
        RouletteT<unsigned int> lRoulette;
        for (unsigned int i=0; i<lNodeIndices.size(); i++) {
            const std::vector<bool>& lColor = inTree[lNodeIndices[i]].mColor;
            int lNumOnes = std::count(lColor.begin(), lColor.end(), true);
            lRoulette.insert(i, (double)lNumOnes);
        }
        unsigned int lSelection = lRoulette.select(lRandomizer);
        unsigned int lNodeIndex = lNodeIndices[lSelection];
        const std::vector<bool>& lNodeColor = inTree[lNodeIndex].mColor;

        // Find the indices of ones in the color.
        std::vector<unsigned int> lIndicesOfOnes;
        for (unsigned int j=0; j<lNodeColor.size(); j++) {
            if (lNodeColor[j]) lIndicesOfOnes.push_back(j);
        }

        // Shuffle the indices of ones, and split them randomly.
        std::random_shuffle(lIndicesOfOnes.begin(), lIndicesOfOnes.end(), lRandomizer);
        std::vector<unsigned int> lPartitionIndices;
        if (lIndicesOfOnes.size() > 1) {
            unsigned int lLowerPartitionIndex = 0;
            unsigned int lUpperPartitionIndex = lIndicesOfOnes.size() - 2;
            do {
                unsigned int lPartitionIndex = lRandomizer.rollInteger(lLowerPartitionIndex, lUpperPartitionIndex);
                lPartitionIndices.push_back(lPartitionIndex);
                lLowerPartitionIndex = lPartitionIndex + 1;
            } while (lLowerPartitionIndex <= lUpperPartitionIndex && lRandomizer.getBoolean());
            lPartitionIndices.push_back(lIndicesOfOnes.size() - 1);
        }

        // Create the child colors.
        for (unsigned int j=0, lLowerPartitionIndex=0; j<lPartitionIndices.size(); j++) {  
            unsigned int lUpperPartitionIndex = lPartitionIndices[j];
            std::vector<bool> lChildColor(lNodeColor.size(), false);
            for (unsigned int k=lLowerPartitionIndex; k<=lUpperPartitionIndex; k++) {
                lChildColor[lIndicesOfOnes[k]] = true;
            }
            outChildColors.push_back(lChildColor);
            lLowerPartitionIndex = lUpperPartitionIndex + 1;
        }

        // Make sure the node that was split will not be chosen again in future iterations.
        lNodeIndices.erase(lNodeIndices.begin()+lSelection);
    }

    Beagle_StackTraceEndM("void ENSO::MutationRandomColorSplitOp::splitNodeColors(ENSO::Tree&, std::vector< std::vector<bool> >&, Beagle::Context&) const");
}

