
/*!
 *  \file   beagle/ENSO/src/NeuralNetworkEvalOp.cpp
 *  \brief  Implementation of the class ENSO::NeuralNetworkEvalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"
#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a new ENSO robot evaluation operator.
 *  \param inName Name of the operator.
 */
ENSO::NeuralNetworkEvalOp::NeuralNetworkEvalOp(Beagle::EvaluationOp::Handle inPhenotypeEvalOp, const std::string& inName) :
		ENSO::EvaluationOp(inPhenotypeEvalOp, inName)
{ }


/*!
 *  \brief Evaluate the fitness of the neural network encoded by the ENSO individual.
 *  \param ioIndividual Individual to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness measure.
 */
Fitness::Handle ENSO::NeuralNetworkEvalOp::evaluate(ENSO::Individual& ioIndividual, Context& ioContext)
{
	Beagle_StackTraceBeginM();

    ioIndividual.makeGraph();
    const std::vector<ENSO::Parameters::Handle>& lGraph = ioIndividual.getGraph();

    NN::Graph::Handle lNetwork = new NN::Graph;
    NN::GraphState::Handle lNetworkState = castHandleT<NN::GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));

    unsigned int lNumModules = ioIndividual[0]->getNumVertices();
    unsigned int lNumModInputs = lNetworkState->getNumberInputs()/lNumModules;
    unsigned int lNumModOutputs = lNetworkState->getNumberOutputs()/lNumModules;
    unsigned int lNumModHelpers = lNetworkState->getNumberHelpers()/lNumModules;

    // Create modules.
    for (unsigned int i = 0; i < lNumModules; i++) {
        ENSO::Parameters::Handle lParamsHandle = lGraph[i*lNumModules + i];
        Beagle_NonNullPointerAssertM(lParamsHandle);
        ENSO::Parameters& lParams = *lParamsHandle;
        GA::FloatVector::Handle lBiases = castHandleT<GA::FloatVector>(lParams["biases"]);
        GA::FloatVector::Handle lSlopes = castHandleT<GA::FloatVector>(lParams["slopes"]);
        GA::FloatVector::Handle lWeights = castHandleT<GA::FloatVector>(lParams["weights"]);

        // Create module input nodes.
        for (unsigned int j = 0; j < lNumModInputs; j++) {
            unsigned int lNodeId = lNetworkState->getFirstInput() + j*lNumModules + i;
            unsigned int lParamsIndex = j;
            NN::Node& lNode = lNetwork->addNode(NN::Node(lNodeId, (*lBiases)[lParamsIndex], (*lSlopes)[lParamsIndex]));
            lNetwork->getMain().addInput(lNodeId);
        }

        // Create module output nodes.
        for (unsigned int j = 0; j < lNumModOutputs; j++) {
            unsigned int lNodeId = lNetworkState->getFirstOutput() + j*lNumModules + i;
            unsigned int lParamsIndex = lNumModInputs + j;
            NN::Node& lNode = lNetwork->addNode(NN::Node(lNodeId, (*lBiases)[lParamsIndex], (*lSlopes)[lParamsIndex]));
            lNetwork->getMain().addOutput(lNodeId);
            for (unsigned int k = 0; k < lNumModHelpers; k++) {
                unsigned int lSourceId = lNetworkState->getFirstHelper() + k*lNumModules + i;
                unsigned int lParamsIndex = j*lNumModHelpers + k;
                lNode.addEdge(NN::Edge(lSourceId, (*lWeights)[lParamsIndex], false));
            }
        }

        // Create module hidden nodes.
        for (unsigned int j = 0; j < lNumModHelpers; j++) {
            unsigned int lNodeId = lNetworkState->getFirstHelper() + j*lNumModules + i;
            unsigned int lParamsIndex = lNumModInputs + lNumModOutputs + j;
            NN::Node& lNode = lNetwork->addNode(NN::Node(lNodeId, (*lBiases)[lParamsIndex], (*lSlopes)[lParamsIndex]));
            lNetwork->getMain().addHidden(lNodeId);
            for (unsigned int k = 0; k < lNumModInputs; k++) {
                unsigned int lSourceId = lNetworkState->getFirstInput() + k*lNumModules + i;
                unsigned int lParamsIndex = lNumModOutputs*lNumModHelpers + j*lNumModInputs + k;
                lNode.addEdge(NN::Edge(lSourceId, (*lWeights)[lParamsIndex], false));
            }
        }
    }

    // Create edges between modules.
    for (unsigned int i1 = 0; i1 < lNumModules; i1++) {
        for (unsigned int i2 = 0; i2 < lNumModules; i2++) {
            if (i1 == i2) continue;
            ENSO::Parameters::Handle lParamsHandle = lGraph[i1*lNumModules + i2];
            if (lParamsHandle == NULL) continue;
            ENSO::Parameters& lParams = *lParamsHandle;
            GA::FloatVector::Handle lWeights = castHandleT<GA::FloatVector>(lParams["weights"]);

            for (unsigned int j = 0; j < lNumModHelpers; j++) {
                unsigned int lNodeId = lNetworkState->getFirstHelper() + j*lNumModules + i1;
                NN::Node& lNode = lNetwork->getNode(lNodeId);
                for (unsigned int k = 0; k < lNumModInputs; k++) {
                    unsigned int lSourceId = lNetworkState->getFirstInput() + k*lNumModules + i2;
                    unsigned int lParamsIndex = j*lNumModInputs + k;
                    lNode.addEdge(NN::Edge(lSourceId, (*lWeights)[lParamsIndex], false));
                }
            }
            
        }
    }

    NN::FitnessIndexed::Handle lFitness = castHandleT<NN::EvaluationOp>(mPhenotypeEvalOp)->evaluate(*lNetwork, ioContext);
    Beagle_NonNullPointerAssertM(lFitness);
    FitnessSimple::Handle lFitnessSimple = new FitnessSimple((*lFitness)[0]);
    return lFitnessSimple;

	Beagle_StackTraceEndM("Fitness::Handle ENSO::NeuralNetworkEvalOp::evaluate(ENSO::Individual& ioIndividual, Context& ioContext)");
}
