
/*!
 *  \file   beagle/ENSO/src/NeuralNetworkInitOp.cpp
 *  \brief  Implementation of the class ENSO::NeuralNetworkInitOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"
#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct an ENSO genotype tree encoding of a neural network.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
ENSO::NeuralNetworkInitOp::NeuralNetworkInitOp(std::string inReproProbaName, std::string inName) :
		ENSO::InitFullySymmetricOp(inReproProbaName, inName),
        mInitNetwork()
{ }


/*!
 *  \brief Register the parameters of the intialization operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::NeuralNetworkInitOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	ENSO::InitFullySymmetricOp::registerParams(ioSystem);
    {
        Register::Description lDescription(
            "Name of file containing the initial neural network individual.",
            "String",
            "",
            "Neural network individual for initializing the parameters of the ENSO genotype tree."
        );
        mNetworkFile = castHandleT<String>(
                             ioSystem.getRegister().insertEntry("enso.initnn.network", new String(""), lDescription));
    }
	{
		Register::Description lDescription(
		    "Minimum value of initial edge weight",
		    "Double",
		    "-10.0",
		    "Edge weight is constrained to be greater than or equal to this value when a new edge is created."
		);
		mMinEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial edge weight",
		    "Double",
		    "10.0",
		    "Edge weight is constrained to be less than this value when a new edge is created."
		);
		mMaxEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node bias",
		    "Double",
		    "-10.0",
		    "Node bias is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node bias",
		    "Double",
		    "10.0",
		    "Node bias is constrained to be less than this value when a new node is created."
		);
		mMaxNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node slope",
		    "Double",
		    "1.0",
		    "Node slope is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemin", new Double(1.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node slope",
		    "Double",
		    "5.0",
		    "Node slope is constrained to be less than this value when a new node is created."
		);
		mMaxNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemax", new Double(5.0), lDescription));
	}
	Beagle_StackTraceEndM("void ENSO::NeuralNetworkInitOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize this operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::NeuralNetworkInitOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	ENSO::InitFullySymmetricOp::init(ioSystem);

    const std::string lNetworkFile = mNetworkFile->getWrappedValue();
    if (lNetworkFile == "") return;

	std::ifstream lIFS(lNetworkFile.c_str());
	PACC::XML::Document lParser(lIFS, lNetworkFile);
	lIFS.close();
	PACC::XML::ConstFinder lFinder(lParser.getFirstDataTag());
	PACC::XML::ConstIterator lIter = lFinder.find("//Genotype");
	if (!lIter) return;

    Context lContext;
	lContext.setSystemHandle(&ioSystem);
	mInitNetwork.readWithContext(lIter, lContext);

    if (mInitNetwork.size() > 0) {
        Beagle_LogDebugM(
            ioSystem.getLogger(),
            "init", "ENSO::NeuralNetworkInitOp",
            std::string("Read initial neural network from file '")+lNetworkFile+"'"
        );
    }

	Beagle_StackTraceEndM("void ENSO::NeuralNetworkInitOp::init(Beagle::System&)");
}


/*!
 *  \brief Initialize the neural network genes in the genotype tree.
 *  \param ioTree Tree whose genes to initialize.
 *  \param ioContext Evolution context.
 */
void ENSO::NeuralNetworkInitOp::initParameters(ENSO::Tree& ioTree, Context& ioContext) const
{
	Beagle_StackTraceBeginM();
	ENSO::InitFullySymmetricOp::initParameters(ioTree, ioContext);

    NN::GraphState::Handle lNetworkState = castHandleT<NN::GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    const double lMinBias = mMinNodeBias->getWrappedValue();
    const double lMaxBias = mMaxNodeBias->getWrappedValue();
    const double lMinSlope = mMinNodeSlope->getWrappedValue();
    const double lMaxSlope = mMaxNodeSlope->getWrappedValue();
    const double lMinWeight = mMinEdgeWeight->getWrappedValue();
    const double lMaxWeight = mMaxEdgeWeight->getWrappedValue();

    unsigned int lNumModules = ioTree.getNumVertices();
    unsigned int lNumModInputs = lNetworkState->getNumberInputs()/lNumModules;
    unsigned int lNumModOutputs = lNetworkState->getNumberOutputs()/lNumModules;
    unsigned int lNumModHelpers = lNetworkState->getNumberHelpers()/lNumModules;
    unsigned int lNumModUnits = lNumModInputs + lNumModOutputs + lNumModHelpers;
    unsigned int lNumModLinks = lNumModOutputs*lNumModHelpers + lNumModHelpers*lNumModInputs;
    unsigned int lNumInterModLinks = lNumModHelpers*lNumModInputs;

    // Initialize parameters of node representing phenotype graph vertices.
    GA::FloatVector::Handle lBiases = new GA::FloatVector(lNumModUnits);
    GA::FloatVector::Handle lSlopes = new GA::FloatVector(lNumModUnits);
    GA::FloatVector::Handle lWeights = new GA::FloatVector(lNumModLinks);

    if (mInitNetwork.size() > 0) {    // Initialize node parameters using network from file.
        // Initialize parameters of input units.
        for (unsigned int j = 0; j < lNumModInputs; j++) {
            unsigned int lNodeId = lNetworkState->getFirstInput() + j*lNumModules;
            unsigned int lParamsIndex = j;
            const NN::Node& lNode = mInitNetwork.getNode(lNodeId);
            (*lBiases)[lParamsIndex] = lNode.getBias();
            (*lSlopes)[lParamsIndex] = lNode.getSlope();
        }

        // Initialize parameters of output units.
        for (unsigned int j = 0; j < lNumModOutputs; j++) {
            unsigned int lNodeId = lNetworkState->getFirstOutput() + j*lNumModules;
            unsigned int lParamsIndex = lNumModInputs + j;
            const NN::Node& lNode = mInitNetwork.getNode(lNodeId);
            (*lBiases)[lParamsIndex] = lNode.getBias();
            (*lSlopes)[lParamsIndex] = lNode.getSlope();
            for (unsigned int k = 0; k < lNumModHelpers; k++) {
                unsigned int lSourceId = lNetworkState->getFirstHelper() + k*lNumModules;
                unsigned int lParamsIndex = j*lNumModHelpers + k;
                (*lWeights)[lParamsIndex] = lNode.getEdge(lSourceId).getWeight();
            }
        }

        // Initialize parameters of hidden units.
        for (unsigned int j = 0; j < lNumModHelpers; j++) {
            unsigned int lNodeId = lNetworkState->getFirstHelper() + j*lNumModules;
            unsigned int lParamsIndex = lNumModInputs + lNumModOutputs + j;
            const NN::Node& lNode = mInitNetwork.getNode(lNodeId);
            (*lBiases)[lParamsIndex] = lNode.getBias();
            (*lSlopes)[lParamsIndex] = lNode.getSlope();
            for (unsigned int k = 0; k < lNumModInputs; k++) {
                unsigned int lSourceId = lNetworkState->getFirstInput() + k*lNumModules;
                unsigned int lParamsIndex = lNumModOutputs*lNumModHelpers + j*lNumModInputs + k;
                (*lWeights)[lParamsIndex] = lNode.getEdge(lSourceId).getWeight();
            }
        }
    }
    else {    // Initialize node parameters randomly.
        for (unsigned int i = 0; i < lNumModUnits; i++) {
			(*lBiases)[i] = lRandomizer.rollUniform(lMinBias, lMaxBias);
			(*lSlopes)[i] = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
        }

        for (unsigned int i = 0; i < lNumModLinks; i++) {
			(*lWeights)[i] = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
        }
    }

    ioTree[1].mParameters["biases"] = lBiases;
    ioTree[1].mParameters["slopes"] = lSlopes;
    ioTree[1].mParameters["weights"] = lWeights;

    // Initialize edge parameters randomly.
    lWeights = new GA::FloatVector(lNumInterModLinks);
    for (unsigned int i = 0; i < lNumInterModLinks; i++) {
        (*lWeights)[i] = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
    }

    ioTree[2].mParameters["weights"] = lWeights;
    
	Beagle_StackTraceEndM("void ENSO::NeuralNetworkInitOp::initParameters(ENSO::Tree& ioTree, Context& ioContext) const");
}
