
/*!
 *  \file   beagle/ENSO/src/NeuralNetworkParams.cpp
 *  \brief  Implementation of the type ENSO::NeuralNetworkParams.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"
#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Register the parameters of the component storing neural network parameter properties.
 *  \param ioSystem Evolutionary system used for initialization.
 */
void ENSO::NeuralNetworkParams::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	ParameterProperties::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Minimum value of initial edge weight",
		    "Double",
		    "-10.0",
		    "Edge weight is constrained to be greater than or equal to this value when a new edge is created."
		);
		mMinEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial edge weight",
		    "Double",
		    "10.0",
		    "Edge weight is constrained to be less than this value when a new edge is created."
		);
		mMaxEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node bias",
		    "Double",
		    "-10.0",
		    "Node bias is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node bias",
		    "Double",
		    "10.0",
		    "Node bias is constrained to be less than this value when a new node is created."
		);
		mMaxNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node slope",
		    "Double",
		    "1.0",
		    "Node slope is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemin", new Double(1.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node slope",
		    "Double",
		    "5.0",
		    "Node slope is constrained to be less than this value when a new node is created."
		);
		mMaxNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemax", new Double(5.0), lDescription));
	}
	Beagle_StackTraceEndM("void ENSO::NeuralNetworkParams::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialization of the component storing neural network parameter properties.
 *  \param ioSystem Evolutionary system used for initialization.
 */
void ENSO::NeuralNetworkParams::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

    GA::FloatVector::Handle lBiases = new GA::FloatVector(2);
    GA::FloatVector::Handle lSlopes = new GA::FloatVector(2);
    GA::FloatVector::Handle lWeights = new GA::FloatVector(2);

    (*lBiases)[0] = mMinNodeBias->getWrappedValue();
    (*lBiases)[1] = mMaxNodeBias->getWrappedValue();
    (*lSlopes)[0] = mMinNodeSlope->getWrappedValue();
    (*lSlopes)[1] = mMaxNodeSlope->getWrappedValue();
    (*lWeights)[0] = mMinEdgeWeight->getWrappedValue();
    (*lWeights)[1] = mMaxEdgeWeight->getWrappedValue();

    (*this)["biases"] = lBiases;
    (*this)["slopes"] = lSlopes;
    (*this)["weights"] = lWeights;

	Beagle_StackTraceEndM("void ENSO::NeuralNetworkParams::init(Beagle::System& ioSystem)");
}


/*!
 *  \brief Read the component storing neural network parameter properties from an XML subtree.
 *  \param inIter XML iterator used to read the component.
 *  \param ioSystem Evolutionary system.
 *  \throw Beagle::IOException If reading failed.
 */
void ENSO::NeuralNetworkParams::readWithSystem(PACC::XML::ConstIterator inIter, Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!=getName()))
		throw Beagle_IOExceptionNodeM(*inIter, "tag <"+getName()+"> expected!");

	Context::Alloc::Handle lContextAlloc =
	    castHandleT<Context::Alloc>(ioSystem.getFactory().getConceptAllocator("Context"));
	Context::Handle lContext = castHandleT<Context>(lContextAlloc->allocate());
	lContext->setSystemHandle(&ioSystem);

	clear();
	Factory lFactory = ioSystem.getFactory();
	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if((lChild->getType()==PACC::XML::eData) && (lChild->getValue()=="Entry")) {
			std::string lKey = lChild->getAttribute("key");
			if(lKey.empty()) {
				throw Beagle_IOExceptionNodeM(*lChild, "no key attribute for NeuralNetworkParams entry!");
            }
			PACC::XML::ConstIterator lChild2 = lChild->getFirstChild();

            std::string lGenotypeName = lChild2->getValue();
            std::string lGenotypeType = lChild2->getAttribute("type");
            Genotype::Alloc::Handle lGenotypeAlloc = NULL;
            if(lGenotypeType.empty()) {
                lGenotypeAlloc = castHandleT<Genotype::Alloc>(lFactory.getConceptAllocator(lGenotypeName));
                if(lGenotypeAlloc == NULL) {
                    std::ostringstream lOSS;
                    lOSS << "Genotype of name '" << lGenotypeName << "' can't be read; ";
                    lOSS << "Its type is not given and a valid concept allocator is not associated to it!";
                    throw Beagle_IOExceptionNodeM(*lChild2, lOSS.str());
                }
            }
            else {
                lGenotypeAlloc = castHandleT<Genotype::Alloc>(lFactory.getAllocator(lGenotypeType));
                if(lGenotypeAlloc == NULL) {
                    std::ostringstream lOSS;
                    lOSS << "It seems that type '" << lGenotypeType << "' associated to genotype '";
                    lOSS << lGenotypeName << "' is not valid!";
                    throw Beagle_IOExceptionNodeM(*lChild2, lOSS.str());
                }
            }
            Genotype::Handle lGenotype = castHandleT<Genotype>(lGenotypeAlloc->allocate());
            std::pair<iterator, bool> lInsertResult = insert(value_type(lKey, lGenotype));
            if(lInsertResult.second == false) {
                std::ostringstream lOSS;
                lOSS << "NeuralNetworkParams map already contains an entry with key '" << lKey; "'";
                throw Beagle_IOExceptionNodeM(*lChild2, lOSS.str());
            }
            lInsertResult.first->second->readWithContext(lChild2, *lContext);
		}
	}

	Beagle_StackTraceEndM("void ENSO::NeuralNetworkParams::readWithSystem(PACC::XML::ConstIterator inIter, Beagle::System& ioSystem)");
}


/*!
 *  \brief Write the parameter properties into a Beagle XML stream.
 *  \param ioStreamer XML streamer to write the properties into.
 *  \param inIndent Whether XML output should be indented.
 */
void ENSO::NeuralNetworkParams::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();

	ioStreamer.openTag("NeuralNetworkParams", inIndent);

	for(const_iterator lIter = begin(); lIter != end(); ++lIter) {
		ioStreamer.openTag("Entry", inIndent);
		ioStreamer.insertAttribute("key", lIter->first);
		lIter->second->write(ioStreamer, false);
		ioStreamer.closeTag();
	}

	ioStreamer.closeTag();

	Beagle_StackTraceEndM("void ENSO::NeuralNetworkParams::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}
