
/*!
 *  \file   beagle/ENSO/src/PackageBase.cpp
 *  \brief  Source code of class ENSO::PackageBase.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

using namespace Beagle;


/*!
 *  \brief Construct package with basic ENSO objects.
 */
ENSO::PackageBase::PackageBase() :
		Beagle::Package("ENSO-PackageBase")
{ }


/*!
 *  \brief Configure system by adding necessary components into it.
 *  \param ioSystem System to configure.
 */
void ENSO::PackageBase::configure(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Factory& lFactory = ioSystem.getFactory();

	// Add available basic types to the factory
	lFactory.insertAllocator("Beagle::ENSO::GenerationalOp", new ENSO::GenerationalOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::Individual", new ENSO::Individual::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::InitFullySymmetricOp", new ENSO::InitFullySymmetricOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::MutationBreakSymmetryOp", new ENSO::MutationBreakSymmetryOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::MutationComparatorsOp", new ENSO::MutationComparatorsOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::MutationParametersOp", new ENSO::MutationParametersOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::MutationRandomColorSplitOp", new ENSO::MutationRandomColorSplitOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::NeuralNetworkEvalOp", new ENSO::NeuralNetworkEvalOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::NeuralNetworkInitOp", new ENSO::NeuralNetworkInitOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::NeuralNetworkParams", new ENSO::NeuralNetworkParams::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::PackageBase", new ENSO::PackageBase::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::Parameters", new ENSO::Parameters::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::ParameterProperties", new ENSO::ParameterProperties::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::SelectDebTournamentOp", new ENSO::SelectDebTournamentOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::SelectTournamentOp", new ENSO::SelectTournamentOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::SortingNetworkInitOp", new ENSO::SortingNetworkInitOp::Alloc);
	lFactory.insertAllocator("Beagle::ENSO::Tree", new ENSO::Tree::Alloc);

	// Set aliases
	lFactory.aliasAllocator("Beagle::ENSO::GenerationalOp", "ENSO-GenerationalOp");
	lFactory.aliasAllocator("Beagle::ENSO::Individual", "ENSO-Individual");
	lFactory.aliasAllocator("Beagle::ENSO::InitFullySymmetricOp", "ENSO-InitFullySymmetricOp");
	lFactory.aliasAllocator("Beagle::ENSO::MutationBreakSymmetryOp", "ENSO-MutationBreakSymmetryOp");
	lFactory.aliasAllocator("Beagle::ENSO::MutationComparatorsOp", "ENSO-MutationComparatorsOp");
	lFactory.aliasAllocator("Beagle::ENSO::MutationParametersOp", "ENSO-MutationParametersOp");
	lFactory.aliasAllocator("Beagle::ENSO::MutationRandomColorSplitOp", "ENSO-MutationRandomColorSplitOp");
	lFactory.aliasAllocator("Beagle::ENSO::NeuralNetworkEvalOp", "ENSO-NeuralNetworkEvalOp");
	lFactory.aliasAllocator("Beagle::ENSO::NeuralNetworkInitOp", "ENSO-NeuralNetworkInitOp");
	lFactory.aliasAllocator("Beagle::ENSO::NeuralNetworkParams", "ENSO-NeuralNetworkParams");
	lFactory.aliasAllocator("Beagle::ENSO::PackageBase", "ENSO-PackageBase");
	lFactory.aliasAllocator("Beagle::ENSO::Parameters", "ENSO-Parameters");
	lFactory.aliasAllocator("Beagle::ENSO::ParameterProperties", "ENSO-ParameterProperties");
	lFactory.aliasAllocator("Beagle::ENSO::SelectDebTournamentOp", "ENSO-SelectDebTournamentOp");
	lFactory.aliasAllocator("Beagle::ENSO::SelectTournamentOp", "ENSO-SelectTournamentOp");
	lFactory.aliasAllocator("Beagle::ENSO::SortingNetworkInitOp", "ENSO-SortingNetworkInitOp");
	lFactory.aliasAllocator("Beagle::ENSO::Tree", "ENSO-Tree");

	// Set some basic concept-type associations in factory
	lFactory.setConcept("Fitness", "FitnessSimple");
	lFactory.setConcept("Individual", "ENSO-Individual");
	lFactory.setConcept("InitializationOp", "ENSO-InitFullySymmetricOp");
	lFactory.setConcept("MutationOp", "ENSO-MutationBreakSymmetryOp");
	lFactory.setConcept("Genotype", "ENSO-Tree");
	lFactory.setConcept("SelectionOp", "ENSO-SelectTournamentOp");
	lFactory.setConcept("StatsCalculateOp", "StatsCalcFitnessSimpleOp");

	// Set some default parameters
	ioSystem.getRegister().setEntryDefaultValue("ec.pop.size",
	        new UIntArray(1,500),
	        "500");
	ioSystem.getRegister().setEntryDefaultValue("ec.sel.tournsize",
	        new UInt(2),
	        "2");

	Beagle_StackTraceEndM("void ENSO::PackageBase::configure(System&)");
}


/*!
 *  \brief List packages on which current package depends.
 *  \return Bag with the packages instanciation on which current package depends.
 *
 *  ENSO::PackageBase depends on Beagle::PackageBase.
 */
Beagle::Package::Bag ENSO::PackageBase::listDependencies(void)
{
	Beagle_StackTraceBeginM();
	Package::Bag lDependencies;
	lDependencies.push_back(new Beagle::PackageBase());
	return lDependencies;
	Beagle_StackTraceEndM("Beagle::Package::Bag ENSO::PackageBase::listDependencies(void)");
}
