
/*!
 *  \file   beagle/ENSO/src/Parameters.cpp
 *  \brief  Source code of class ENSO::Parameters; similar to Beagle::Map
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"
#include "beagle/GA.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Default construct a parameters object.
 *  \param inValid If the parameters is valid or not.
 */
ENSO::Parameters::Parameters(bool inValid) :
        mValid(inValid)
{ }


/*!
 *  \brief Copy constructor.
 *  \param inRightParams The other parameters object.
 */
ENSO::Parameters::Parameters(const ENSO::Parameters& inRightParams) :
        Genotype(inRightParams),
#if defined(BEAGLE_HAVE_STD_HASHMAP)
        std::hash_map<std::string,Genotype::Handle,HashString>(inRightParams),
#elif defined(BEAGLE_HAVE_GNUCXX_HASHMAP)
        __gnu_cxx::hash_map<std::string,Genotype::Handle,HashString>(inRightParams),
#elif defined(BEAGLE_HAVE_STDEXT_HASHMAP)
        stdext::hash_map<std::string,Genotype::Handle,HashString>(inRightParams),
#else // No hash_map found
        std::map< std::string,Genotype::Handle,std::less<std::string> >(inRightParams),
        mValid(inRightParams.mValid)
#endif
{
	Beagle_StackTraceBeginM();
    antialias();
	Beagle_StackTraceEndM("ENSO::Parameters::Parameters(const ENSO::Parameters& inRightParams)");
}


/*!
 *  \brief The assignment operator, with deep copies.
 *  \param inRightParams The other parameters object.
 */
ENSO::Parameters& ENSO::Parameters::operator=(const ENSO::Parameters& inRightParams)
{
	Beagle_StackTraceBeginM();

    // First call the assignment operators for the base types.
    Genotype::operator=(inRightParams);
#if defined(BEAGLE_HAVE_STD_HASHMAP)
    std::hash_map<std::string,Genotype::Handle,HashString>::operator=(inRightParams);
#elif defined(BEAGLE_HAVE_GNUCXX_HASHMAP)
    __gnu_cxx::hash_map<std::string,Genotype::Handle,HashString>::operator=(inRightParams);
#elif defined(BEAGLE_HAVE_STDEXT_HASHMAP)
    stdext::hash_map<std::string,Genotype::Handle,HashString>::operator=(inRightParams);
#else // No hash_map found
    std::map< std::string,Genotype::Handle,std::less<std::string> >::operator=(inRightParams);
#endif

    mValid = inRightParams.mValid;
    antialias();

    return *this;

	Beagle_StackTraceEndM("ENSO::Parameters& ENSO::Parameters::operator=(const ENSO::Parameters& inRightParams)");
}


/*!
 *  \brief Antialias the pointers in the map by making deep copies of them.
 */
void ENSO::Parameters::antialias()
{
	Beagle_StackTraceBeginM();

	for(iterator lIter = begin(); lIter != end(); ++lIter) {
        // lEntry is a reference so that we can assign newly allocated handle to it.
        Genotype::Handle& lEntry = lIter->second;
        const std::string& lEntryType = lEntry->getType();
        if (lEntryType == "GA-BitString") {
            GA::BitString::Handle lBitString = castHandleT<GA::BitString>(lEntry);
            GA::BitString::Handle lNewBitString = new GA::BitString;
            *lNewBitString = *lBitString;
            lEntry = lNewBitString;
        }
        else if (lEntryType == "GA-IntegerVector") {
            GA::IntegerVector::Handle lIntegerVector = castHandleT<GA::IntegerVector>(lEntry);
            GA::IntegerVector::Handle lNewIntegerVector = new GA::IntegerVector;
            *lNewIntegerVector = *lIntegerVector;
            lEntry = lNewIntegerVector;
        }
        else if (lEntryType == "GA-FloatVector") {
            GA::FloatVector::Handle lFloatVector = castHandleT<GA::FloatVector>(lEntry);
            GA::FloatVector::Handle lNewFloatVector = new GA::FloatVector;
            *lNewFloatVector = *lFloatVector;
            lEntry = lNewFloatVector;
        }
        else {
            std::ostringstream lOSS;
            lOSS << "'ENSO::Parameters::operator=()' not defined for entry of type '" << lEntryType << "'";
            throw Beagle_RunTimeExceptionM(lOSS.str());
        }
    }

	Beagle_StackTraceEndM("void ENSO::Parameters::antialias()")
}


/*!
 *  \brief Copy the given Parameters into current.
 *  \param inOriginal Parameters to copy.
 *  \param ioSystem Evolutionary system to use for making the copy.
 */
void ENSO::Parameters::copy(const Member& inOriginal, System& ioSystem)
{
	Beagle_StackTraceBeginM();
	const ENSO::Parameters& lOriginal = castObjectT<const ENSO::Parameters&>(inOriginal);
	(*this) = lOriginal;
	Beagle_StackTraceEndM("void ENSO::Parameters::copy(const Member&,System&)");
}


/*!
 *  \brief Get exact type of the genotype.
 *  \return Exact type of the genotype.
 */
const std::string& ENSO::Parameters::getType() const
{
	Beagle_StackTraceBeginM();
	const static std::string lType("ENSO-Parameters");
	return lType;
	Beagle_StackTraceEndM("const std::string& ENSO::Parameters::getType() const")
}


/*!
 *  \brief Compare equality of two maps.
 *  \param inRightObj Parameters to compare to this.
 *  \return True if the maps are equal, false if not.
 *  \throw Beagle::BadCastException If compared map Objects are not of the same type.
 *  \par Note:
 *    Returns true if the size of maps is equal and if the maps are identical
 *    when compared entry-by-entry, and returns false otherwise.
 */
bool ENSO::Parameters::isEqual(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const ENSO::Parameters& lRightParameters = castObjectT<const ENSO::Parameters&>(inRightObj);
	bool lEqual = (size() == lRightParameters.size());
	for(const_iterator lLeftIter = begin(); lEqual && (lLeftIter != end()); lLeftIter++) {
        const_iterator lRightIter = lRightParameters.find(lLeftIter->first);
		lEqual = (lRightIter != lRightParameters.end()) && lLeftIter->second->isEqual(*(lRightIter->second));
	}
    return lEqual;
	Beagle_StackTraceEndM("bool ENSO::Parameters::isEqual(const Object& inRightObj) const");
}


/*!
 *  \brief Evaluate the order of two maps.
 *  \param inRightObj Parameters to compare to this.
 *  \return True if left map is less than the right one, false if not.
 *  \throw Beagle::BadCastException If compared map Objects are not of the same type.
 *  \par Note:
 *    Returns true if every entry of this parameter map is contained in the right
 *    map and this map is less than the right map when compared entry-by-entry.
 *    Otherwise, returns false.
 */
bool ENSO::Parameters::isLess(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const ENSO::Parameters& lRightParameters = castObjectT<const ENSO::Parameters&>(inRightObj);
	bool lLess = (size() <= lRightParameters.size());
	for(const_iterator lLeftIter = begin(); lLess && (lLeftIter != end()); lLeftIter++) {
        const_iterator lRightIter = lRightParameters.find(lLeftIter->first);
		lLess = (lRightIter != lRightParameters.end()) && lLeftIter->second->isLess(*(lRightIter->second));
	}
	return lLess;
	Beagle_StackTraceEndM("bool ENSO::Parameters::isLess(const Object& inRightObj) const");
}


/*!
 *  \brief Read a map from a XML node.
 *  \param inIter XML iterator to read the Parameters from.
 *  \param ioContext Context to use to read the Parameters.
 *  \throw Beagle::IOException If reading from the stream problems occured.
 */
void ENSO::Parameters::readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType() != PACC::XML::eData) || (inIter->getValue() != "Genotype")) {
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Genotype> expected!");
	}

	std::string lGenotypeType = inIter->getAttribute("type");
	if((lGenotypeType.empty() == false) && (lGenotypeType != getType())) {
		std::ostringstream lOSS;
		lOSS << "type given '" << lGenotypeType << "' mismatch type of the genotype '";
		lOSS << getType() << "'!";
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}

    std::string lValid = inIter->getAttribute("valid");
    if(lValid == "yes") setValid();
    else setInvalid();

	clear();
	Factory lFactory = ioContext.getSystem().getFactory();
	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if((lChild->getType()==PACC::XML::eData) && (lChild->getValue()=="Entry")) {
			std::string lKey = lChild->getAttribute("key");
			if(lKey.empty()) {
				throw Beagle_IOExceptionNodeM(*lChild, "no key attribute for Parameters entry!");
            }
			PACC::XML::ConstIterator lChild2 = lChild->getFirstChild();

            std::string lGenotypeName = lChild2->getValue();
            std::string lGenotypeType = lChild2->getAttribute("type");
            Genotype::Alloc::Handle lGenotypeAlloc = NULL;
            if(lGenotypeType.empty()) {
                lGenotypeAlloc = castHandleT<Genotype::Alloc>(lFactory.getConceptAllocator(lGenotypeName));
                if(lGenotypeAlloc == NULL) {
                    std::ostringstream lOSS;
                    lOSS << "Genotype of name '" << lGenotypeName << "' can't be read; ";
                    lOSS << "Its type is not given and a valid concept allocator is not associated to it!";
                    throw Beagle_IOExceptionNodeM(*lChild2, lOSS.str());
                }
            }
            else {
                lGenotypeAlloc = castHandleT<Genotype::Alloc>(lFactory.getAllocator(lGenotypeType));
                if(lGenotypeAlloc == NULL) {
                    std::ostringstream lOSS;
                    lOSS << "It seems that type '" << lGenotypeType << "' associated to genotype '";
                    lOSS << lGenotypeName << "' is not valid!";
                    throw Beagle_IOExceptionNodeM(*lChild2, lOSS.str());
                }
            }
            Genotype::Handle lGenotype = castHandleT<Genotype>(lGenotypeAlloc->allocate());
            std::pair<iterator, bool> lInsertResult = insert(value_type(lKey, lGenotype));
            if(lInsertResult.second == false) {
                std::ostringstream lOSS;
                lOSS << "Parameters map already contains an entry with key '" << lKey; "'";
                throw Beagle_IOExceptionNodeM(*lChild2, lOSS.str());
            }
            lInsertResult.first->second->readWithContext(lChild2, ioContext);
		}
	}

	Beagle_StackTraceEndM("void ENSO::Parameters::readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext)");
}


/*!
 *  \brief Write a map into a XML streamer.
 *  \param ioStreamer XML streamer to read the Parameters to.
 *  \param inIndent Whether XML output should be indented.
 */
void ENSO::Parameters::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();

    if(isValid()) {
        ioStreamer.insertAttribute("valid", "yes");
    }
    else {
        ioStreamer.insertAttribute("valid", "no");
    }
	for(const_iterator lIter = begin(); lIter != end(); ++lIter) {
		ioStreamer.openTag("Entry", inIndent);
		ioStreamer.insertAttribute("key", lIter->first);
		lIter->second->write(ioStreamer, false);
		ioStreamer.closeTag();
	}

	Beagle_StackTraceEndM("void ENSO::Parameters::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}


/*!
 *  \return Number of parameters in the map entries.
 */
unsigned int ENSO::Parameters::getSize() const
{
	Beagle_StackTraceBeginM();

    unsigned int lSize = 0;
	for(const_iterator lIter = begin(); lIter != end(); ++lIter) {
        const Genotype::Handle& lEntry = lIter->second;
        const std::string& lEntryType = lEntry->getType();
        if (lEntryType == "GA-BitString") {
            GA::BitString::Handle lBitString = castHandleT<GA::BitString>(lEntry);
            lSize += lBitString->size();
        }
        else if (lEntryType == "GA-IntegerVector") {
            GA::IntegerVector::Handle lIntegerVector = castHandleT<GA::IntegerVector>(lEntry);
            lSize += lIntegerVector->size();
        }
        else if (lEntryType == "GA-FloatVector") {
            GA::FloatVector::Handle lFloatVector = castHandleT<GA::FloatVector>(lEntry);
            lSize += lFloatVector->size();
        }
        else {
            std::ostringstream lOSS;
            lOSS << "'ENSO::Parameters::getSize()' not defined for entry of type '" << lEntryType << "'";
            throw Beagle_RunTimeExceptionM(lOSS.str());
        }
    }

	return lSize;

	Beagle_StackTraceEndM("unsigned int ENSO::Parameters::getSize() const");
}


/*!
 *  \brief Randomize the entries in this parameter map.
 */
void ENSO::Parameters::randomize(Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

    ParameterProperties::Handle lParamProps = castHandleT<ParameterProperties>(ioContext.getSystem().getComponent("ENSO-ParameterProperties"));
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	for(iterator lIter = begin(); lIter != end(); ++lIter) {
        Genotype::Handle& lEntry = lIter->second;
        const std::string& lEntryType = lEntry->getType();
        if (lEntryType == "GA-BitString") {
            if (lIter->first != "isedge") {
                GA::BitString::Handle lBitString = castHandleT<GA::BitString>(lEntry);
                for(unsigned int i=0; i<lBitString->size(); i++) {
                    (*lBitString)[i] = lRandomizer.getBoolean();
                }
            }
        }
        else if (lEntryType == "GA-IntegerVector") {
            GA::IntegerVector::Handle lIntegerVector = castHandleT<GA::IntegerVector>(lEntry);
            GA::IntegerVector::Handle lIntegerProps = castHandleT<GA::IntegerVector>((*lParamProps)[lIter->first]);
            int lMin = (*lIntegerProps)[0];
            int lMax = (*lIntegerProps)[1];
            for(unsigned int i=0; i<lIntegerVector->size(); i++) {
                (*lIntegerVector)[i] = lRandomizer.rollInteger(lMin, lMax);
            }
        }
        else if (lEntryType == "GA-FloatVector") {
            GA::FloatVector::Handle lFloatVector = castHandleT<GA::FloatVector>(lEntry);
            GA::FloatVector::Handle lFloatProps = castHandleT<GA::FloatVector>((*lParamProps)[lIter->first]);
            float lMin = (*lFloatProps)[0];
            float lMax = (*lFloatProps)[1];
            for(unsigned int i=0; i<lFloatVector->size(); i++) {
                (*lFloatVector)[i] = lRandomizer.rollUniform(lMin, lMax);
            }
        }
        else {
            std::ostringstream lOSS;
            lOSS << "'ENSO::Parameters::randomize()' not defined for entry of type '" << lEntryType << "'";
            throw Beagle_RunTimeExceptionM(lOSS.str());
        }
    }

    Beagle_StackTraceEndM("void ENSO::Parameters::randomize(Beagle::Context& ioContext)");
}
