
/*!
 *  \file   beagle/ENSO/src/SelectDebTournamentOp.cpp
 *  \brief  Source code of class SelectDebTournamentOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a tournament selection operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register.
 *  \param inName Name of the tournament selection operator.
 */
ENSO::SelectDebTournamentOp::SelectDebTournamentOp(std::string inReproProbaName, std::string inName) :
		Beagle::SelectTournamentOp(inReproProbaName, inName)
{ }


/*!
 *  \brief Register the parameters of this operator.
 *  \param ioSystem System to use to initialize the operator.
 */
void ENSO::SelectDebTournamentOp::registerParams(System& ioSystem)
{
	Beagle_StackTraceBeginM();

	Beagle::SelectTournamentOp::registerParams(ioSystem);
	Register::Description lDescription(
	    "Threshold used in comparing genotype distance.",
	    "Float",
	    "1.0",
	    "Threshold used in comparing genotype distance before fitnesses of two individuals are compared."
	);
	mThreshold = castHandleT<Float>(
	                          ioSystem.getRegister().insertEntry("enso.sel.threshold", new Float(1.0), lDescription));

	Beagle_StackTraceEndM("void ENSO::SelectDebTournamentOp::registerParams(System&)");
}


/*!
 *  \brief Select an individual using the tournament selection method.
 *  \param ioPool Individual pool to use for selection.
 *  \param ioContext Context of the evolution.
 *  \return Index of the choosen individual in the pool.
 */
unsigned int ENSO::SelectDebTournamentOp::selectIndividual(Beagle::Individual::Bag& ioPool, Context& ioContext)
{
	Beagle_StackTraceBeginM();
	Beagle_ValidateParameterM(mNumberParticipants->getWrappedValue() > 0,
	                          "ec.sel.tournsize", ">0");

    float lThreshold = mThreshold->getWrappedValue();

	const Stats& lStats = *(ioContext.getVivarium().getStats());
	Measure lFitnessMeasure;
	for (unsigned int i=0; i<lStats.size(); i++) {
        if (lStats[i].mID == "fitness") {
            lFitnessMeasure = lStats[i];
            break;
        }
	}

	unsigned int lChosenIndividual =
	    ioContext.getSystem().getRandomizer().rollInteger(0,((unsigned int)ioPool.size())-1);
    ENSO::Individual::Handle lChosenIndivHandle = castHandleT<ENSO::Individual>(ioPool[lChosenIndividual]);
	ENSO::Tree::Handle lChosenTree = (*lChosenIndivHandle)[0];
	unsigned int lChosenIndivAge = ioContext.getGeneration() - lChosenTree->getLastSymmetryBreakGeneration();

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "selection", "ENSO::SelectDebTournamentOp",
	    std::string("Starting by choosing the ")+
	    uint2ordinal(lChosenIndividual+1)+" individual"
	);

	for(unsigned int j=1; j<mNumberParticipants->getWrappedValue(); j++) {
		unsigned int lTriedIndividual =
		    ioContext.getSystem().getRandomizer().rollInteger(0,((unsigned int)ioPool.size())-1);
        ENSO::Individual::Handle lTriedIndivHandle = castHandleT<ENSO::Individual>(ioPool[lTriedIndividual]);
		ENSO::Tree::Handle lTriedTree = (*lTriedIndivHandle)[0];
		unsigned int lTriedIndivAge = ioContext.getGeneration() - lTriedTree->getLastSymmetryBreakGeneration();

        double lIndivDistance = lChosenIndivHandle->getDistanceGenotype(*lTriedIndivHandle);
        double lFitnessDistance = lChosenIndivHandle->getFitness()->getDistance(*(lTriedIndivHandle->getFitness()));

		if (lIndivDistance < lThreshold) {
			if (lFitnessDistance > ioContext.getSystem().getRandomizer().rollGaussian(0.0, lFitnessMeasure.mStd) &&
				ioPool[lChosenIndividual]->isLess(*ioPool[lTriedIndividual])) {
                lChosenIndividual = lTriedIndividual;
                Beagle_LogDebugM(
                    ioContext.getSystem().getLogger(),
                    "selection", "ENSO::SelectDebTournamentOp",
                    std::string("Trying the ")+uint2ordinal(lTriedIndividual+1)+
                    " individual -> choosing it"
                );
            } else {
                Beagle_LogDebugM(
                    ioContext.getSystem().getLogger(),
                    "selection", "ENSO::SelectDebTournamentOp",
                    std::string("Trying the ")+uint2ordinal(lTriedIndividual+1)+
                    " individual -> the previously choosen one is better"
                );
            }
            break;
        }
    }

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "selection",
	    "ENSO::SelectDebTournamentOp",
	    std::string("Selecting the ")+uint2ordinal(lChosenIndividual+1)+" individual"
	);
	Beagle_LogObjectDebugM(
	    ioContext.getSystem().getLogger(),
	    "selection",
	    "ENSO::SelectDebTournamentOp",
	    *ioPool[lChosenIndividual]
	);

	return lChosenIndividual;
	Beagle_StackTraceEndM("unsigned int ENSO::SelectDebTournamentOp::selectIndividual(Beagle::Individual::Bag& ioPool, Context& ioContext)");
}
