
/*!
 *  \file   beagle/ENSO/src/SelectTournamentOp.cpp
 *  \brief  Source code of class SelectTournamentOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a tournament selection operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register.
 *  \param inName Name of the tournament selection operator.
 */
ENSO::SelectTournamentOp::SelectTournamentOp(std::string inReproProbaName, std::string inName) :
		Beagle::SelectTournamentOp(inReproProbaName, inName)
{ }


/*!
 *  \brief Register the parameters of this operator.
 *  \param ioSystem System to use to initialize the operator.
 */
void ENSO::SelectTournamentOp::registerParams(System& ioSystem)
{
	Beagle_StackTraceBeginM();

	Beagle::SelectTournamentOp::registerParams(ioSystem);
    {
        Register::Description lDescription(
            "Minimum maturity after the last symmetry break mutation for competing with other individuals.",
            "Float",
            "0.0",
            "Fitnesses of symmetrically different individuals are compared only if their symmetry ages are above this maturity."
        );
        mMinMaturity = castHandleT<Float>(
                            ioSystem.getRegister().insertEntry("enso.sel.minmaturity", new Float(0.0), lDescription));
    }
    {
        Register::Description lDescription(
            "Increment in minimum maturity after the last symmetry break mutation for competing with other individuals.",
            "Float",
            "0.0",
            "Increment per genotype node in minimum maturity for comparing fitnesses of symmetrically different individuals."
        );
        mIncMaturity = castHandleT<Float>(
                            ioSystem.getRegister().insertEntry("enso.sel.incmaturity", new Float(0.0), lDescription));
    }
	Beagle_StackTraceEndM("void ENSO::SelectTournamentOp::registerParams(System&)");
}


/*!
 *  \brief Select an individual using the tournament selection method.
 *  \param ioPool Individual pool to use for selection.
 *  \param ioContext Context of the evolution.
 *  \return Index of the choosen individual in the pool.
 */
unsigned int ENSO::SelectTournamentOp::selectIndividual(Beagle::Individual::Bag& ioPool, Context& ioContext)
{
	Beagle_StackTraceBeginM();
	Beagle_ValidateParameterM(mNumberParticipants->getWrappedValue() > 0,
	                          "ec.sel.tournsize", ">0");

    Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    const float lMinMaturity = mMinMaturity->getWrappedValue();
    const float lIncMaturity = mIncMaturity->getWrappedValue();

    unsigned int lChosenIndividual = lRandomizer.rollInteger(0,((unsigned int)ioPool.size())-1);
	ENSO::Individual::Handle lChosenIndivHandle = castHandleT<ENSO::Individual>(ioPool[lChosenIndividual]);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "selection", "ENSO::SelectTournamentOp",
	    std::string("Starting by choosing the ")+
	    uint2ordinal(lChosenIndividual+1)+" individual"
	);

    // In the ioPool sorted by ENSO::GenerationalOp, find the min and
    // max index range of individuals for tournament selection.
    unsigned int lMinSelectIndiv = UINT_MAX;
    unsigned int lMaxSelectIndiv = UINT_MAX;
    if (lChosenIndivHandle->isMature(ioContext.getGeneration(), lMinMaturity, lIncMaturity)) {
        for (unsigned int i=1; ; i++) {
            if (lChosenIndividual < i || 
                castHandleT<ENSO::Individual>(ioPool[lChosenIndividual-i])->isMature(ioContext.getGeneration(), lMinMaturity, lIncMaturity) == false) {
                lMinSelectIndiv = (lChosenIndividual+1)-i;
                break;
            }
        }
        lMaxSelectIndiv = ((unsigned int)ioPool.size())-1;
    }
    else {
        for (unsigned int i=1; ; i++) {
            if (lChosenIndividual < i ||
                (*lChosenIndivHandle)[0]->getLatticeIndex() != (*castHandleT<ENSO::Individual>(ioPool[lChosenIndividual-i]))[0]->getLatticeIndex()) {
                lMinSelectIndiv = (lChosenIndividual+1)-i;
                break;
            }
        }
        for (unsigned int i=1; ; i++) {
            if (lChosenIndividual+i == ioPool.size() ||
                (*lChosenIndivHandle)[0]->getLatticeIndex() != (*castHandleT<ENSO::Individual>(ioPool[lChosenIndividual+i]))[0]->getLatticeIndex()) {
                lMaxSelectIndiv = (lChosenIndividual+i)-1;
                break;
            }
        }
    }

    for(unsigned int j=1; j<mNumberParticipants->getWrappedValue(); j++) {
        unsigned int lTriedIndividual = lRandomizer.rollInteger(lMinSelectIndiv, lMaxSelectIndiv);
		if(ioPool[lChosenIndividual]->isLess(*ioPool[lTriedIndividual])) {
			lChosenIndividual = lTriedIndividual;
			Beagle_LogDebugM(
			    ioContext.getSystem().getLogger(),
			    "selection", "ENSO::SelectTournamentOp",
			    std::string("Trying the ")+uint2ordinal(lTriedIndividual+1)+
			    " individual -> choosing it"
			);
		} else {
			Beagle_LogDebugM(
			    ioContext.getSystem().getLogger(),
			    "selection", "ENSO::SelectTournamentOp",
			    std::string("Trying the ")+uint2ordinal(lTriedIndividual+1)+
			    " individual -> the previously choosen one is better"
			);
		}
	}

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "selection",
	    "ENSO::SelectTournamentOp",
	    std::string("Selecting the ")+uint2ordinal(lChosenIndividual+1)+" individual"
	);
	Beagle_LogObjectDebugM(
	    ioContext.getSystem().getLogger(),
	    "selection",
	    "ENSO::SelectTournamentOp",
	    *ioPool[lChosenIndividual]
	);

	return lChosenIndividual;
	Beagle_StackTraceEndM("unsigned int ENSO::SelectTournamentOp::selectIndividual(Beagle::Individual::Bag& ioPool, Context& ioContext)");
}
