
/*!
 *  \file   beagle/ENSO/src/SortingNetworkInitOp.cpp
 *  \brief  Implementation of the class ENSO::SortingNetworkInitOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

using namespace Beagle;


/*!
 *  \brief Construct an ENSO genotype tree encoding of a sorting network.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
ENSO::SortingNetworkInitOp::SortingNetworkInitOp(std::string inReproProbaName, std::string inName) :
		ENSO::InitFullySymmetricOp(inReproProbaName, inName),
        mComparatorsOp(new ENSO::MutationComparatorsOp)
{ }


/*!
 *  \brief Register the parameters of the intialization operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::SortingNetworkInitOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
    mComparatorsOp->registerParams(ioSystem);
	ENSO::InitFullySymmetricOp::registerParams(ioSystem);
	Beagle_StackTraceEndM("void ENSO::SortingNetworkInitOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize this operator.
 *  \param ioSystem System of the evolution.
 */
void ENSO::SortingNetworkInitOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	mComparatorsOp->init(ioSystem);
	ENSO::InitFullySymmetricOp::init(ioSystem);
	Beagle_StackTraceEndM("void ENSO::SortingNetworkInitOp::init(Beagle::System&)");
}


/*!
 *  \brief Initialize a genotype.
 *  \param outTree Tree to initialize.
 *  \param ioContext Evolution context.
 */
void ENSO::SortingNetworkInitOp::initTree(ENSO::Tree& outTree, Context& ioContext) const
{
	Beagle_StackTraceBeginM();

	const unsigned int lNumVertices = mNumVertices->getWrappedValue();
    outTree.setNumVertices(lNumVertices);

    // Clear the tree and create a root node having two children, one representing the
    // vertices and the other the edges of the phenotype graph.
	outTree.clear();
    outTree.resize(3);

    // The root node.
    Node& lRoot = outTree[0];
    lRoot.mColor.clear();
    lRoot.mColor.resize(lNumVertices*lNumVertices, true);
    lRoot.mChildren.clear();
    lRoot.mChildren.push_back(1);
    lRoot.mChildren.push_back(2);

    // Child nodes representing vertices and edges.
    Node& lVertices = outTree[1];
    lVertices.mParent = 0;
    lVertices.mChildren.clear();
    Node& lEdges = outTree[2];
    lEdges.mParent = 0;
    lEdges.mChildren.clear();

    // Set the child node colors.
    lVertices.mColor.clear();
    lVertices.mColor.resize(lNumVertices*lNumVertices, false);
    lEdges.mColor.clear();
    lEdges.mColor.resize(lNumVertices*lNumVertices, true);
    for (unsigned int i=0; i<lNumVertices; i++) {
        unsigned int diagonal = i*lNumVertices + i;
        lVertices.mColor[diagonal] = true;
        lEdges.mColor[diagonal] = false;
    }

    outTree.recordSymmetryState(ioContext.getGeneration(), 0);

    initParameters(outTree, ioContext);

    // Break symmetry to correspond to user-specified lattice index.
	unsigned int lLatticeIndex = mLatticeIndex->getWrappedValue();
    if (lLatticeIndex > 0) {
        mBreakSymmetryOp->breakSymmetry(outTree, lLatticeIndex, ioContext);
    }

    // Break symmetry further if necessary so that there is at least one edge color
    // having not more than one edge with the same end point, i.e. it is possible
    // for that color to represent a parallel stage of the sorting network.
    std::vector<unsigned int> lNodeIndices;
    do {
        outTree.getParallelEdgeLeaves(lNodeIndices);
        if (lNodeIndices.size() == 0) {
            while (!mBreakSymmetryOp->breakSymmetry(outTree, ioContext)) {}
        }
    } while (lNodeIndices.size() == 0);

    // Initialize the first stage of the sorting network with the edges of such a node
    // chosen randomly.
    if (!mComparatorsOp->addComparators(outTree, ioContext)) {
        std::ostringstream lOSS;
        lOSS << "Unable to initialize the first stage of the sorting network.";
        throw Beagle_RunTimeExceptionM(lOSS.str());
    }

	Beagle_LogVerboseM(
	    ioContext.getSystem().getLogger(),
	    "initialization", "Beagle::ENSO::SortingNetworkInitOp",
	    std::string("Initializing sorting network for the ")+
	    uint2ordinal(ioContext.getGenotypeIndex()+1)+std::string(" genotype tree.")
	);

	Beagle_StackTraceEndM("void ENSO::SortingNetworkInitOp::initTree(ENSO::Tree& outTree, Context& ioContext) const");
}


/*!
 *  \brief Initialize the vector of comparators of the genotype tree.
 *  \param ioTree Tree to initialize.
 *  \param ioContext Evolution context.
 */
void ENSO::SortingNetworkInitOp::initParameters(ENSO::Tree& ioTree, Context& ioContext) const
{
	Beagle_StackTraceBeginM();

    // The comparators are listed as pairs of vertices, and parallel stages
    // are indices into the vector of comparators.  The nodes vector stores
    // the indices of the corresponding edge nodes in the genotype tree.
    ioTree[0].mParameters["comparators"] = new GA::IntegerVector();
    ioTree[0].mParameters["stages"] = new GA::IntegerVector();
    ioTree[0].mParameters["nodes"] = new GA::IntegerVector();

    ioTree[2].mParameters["isedge"] = new GA::BitString();

    ioTree[0].mParameters.setValid();
    ioTree[1].mParameters.setValid();
    ioTree[2].mParameters.setValid();
    
	Beagle_StackTraceEndM("void ENSO::SortingNetworkInitOp::initParameters(ENSO::Tree& ioTree, Context& ioContext) const");
}
