
/*!
 *  \file   beagle/ENSO/src/Tree.cpp
 *  \brief  Implementation of the type ENSO::Tree.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/ENSO.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct a node in the ENSO genotype tree.
 *  \param inColor Color of this node.
 *  \param inParent Index of parent node.
 */
ENSO::Node::Node(const std::vector<bool>& inColor, unsigned int inParent) :
		mColor(inColor),
		mParent(inParent),
        mChildren(),
        mParameters()
{ }


/*!
 *  \brief Compare equality of two ENSO tree nodes.
 *  \param inRightNode Right node to be compare to.
 *  \return True if nodes are identical, false if not.
 */
bool ENSO::Node::operator==(const ENSO::Node& inRightNode) const
{
	Beagle_StackTraceBeginM();
    return (mColor == inRightNode.mColor) &&
        (mParent == inRightNode.mParent) &&
        (mChildren == inRightNode.mChildren) &&
        (mParameters.isEqual(inRightNode.mParameters));
	Beagle_StackTraceEndM("bool ENSO::Node::operator==(const ENSO::Node&) const");
}


/*!
 *  \brief Construct an ENSO genotype tree.
 *  \param inNumVertices Number of vertices in the phenotype graph.
 */
ENSO::Tree::Tree(unsigned int inNumVertices) :
        std::vector<Node>(),
		mNumVertices(inNumVertices),
        mLastSymmetryBreakGeneration(0),
        mLastIsEdgeMutationGeneration(0),
        mLatticePath(),
        mSizePath()
{ }


/*!
 *  \brief Copy ENSO tree into current.
 *  \param inOriginal ENSO tree to copy.
 *  \param ioSystem Evolutionary system to use for making the copy.
 */
void ENSO::Tree::copy(const Member& inOriginal, System& ioSystem)
{
	Beagle_StackTraceBeginM();
	const ENSO::Tree& lOriginal = castObjectT<const ENSO::Tree&>(inOriginal);
	(*this) = lOriginal;
	Beagle_StackTraceEndM("void ENSO::Tree::copy(const Member&, System&)");
}


/*!
 *  \brief Get exact type of the genotype.
 *  \return Exact type of the genotype.
 */
const std::string& ENSO::Tree::getType() const
{
	Beagle_StackTraceBeginM();
	const static std::string lType("ENSO-Tree");
	return lType;
	Beagle_StackTraceEndM("const std::string& ENSO::Tree::getType() const")
}


/*!
 *  \brief Compare the equality of two ENSO genotype trees.
 *  \param inRightObj Right genotype tree to be compared to.
 *  \return True if the genotype trees are equal, false otherwise.
 *  \throw Beagle::BadCastException If compared genotype tree objects are not of the same type.
 */
bool ENSO::Tree::isEqual(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const ENSO::Tree& lRightTree = castObjectT<const ENSO::Tree&>(inRightObj);
	if(size() != lRightTree.size()) return false;
	return std::equal(begin(), end(), lRightTree.begin());
	Beagle_StackTraceEndM("bool ENSO::Tree::isEqual(const Object& inRightObj) const");
}


/*!
 *  \brief Check if this genotype tree is less than the other.
 *  \param inRightObj Right genotype tree to be compared to.
 *  \return True if this genotype tree has fewer nodes than the right one, false otherwise.
 *  \throw Beagle::BadCastException If compared genotype tree objects are not of the same type.
 */
bool ENSO::Tree::isLess(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const ENSO::Tree& lRightTree = castObjectT<const ENSO::Tree&>(inRightObj);
	return (size() < lRightTree.size());
	Beagle_StackTraceEndM("bool ENSO::Tree::isLess(const Object& inRightObj) const");
}


/*!
 *  \brief Read an ENSO genotype tree from an XML subtree.
 *  \param inIter XML iterator to use to read the genotype tree from.
 *  \param ioContext Context to use to read the genotype tree.
 */
void ENSO::Tree::readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType() != PACC::XML::eData) || (inIter->getValue() != "Genotype")) {
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Genotype> expected!");
	}
	std::string lGenotypeType = inIter->getAttribute("type");
	if((lGenotypeType.empty() == false) && (lGenotypeType != getType())) {
		std::ostringstream lOSS;
		lOSS << "type given '" << lGenotypeType << "' mismatch type of the genotype '";
		lOSS << getType() << "'!";
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}

    clear();
	std::string lSizeText = inIter->getAttribute("size");
	if(lSizeText.empty()==false)
		resize(str2uint(lSizeText));

	std::string lNumVerticesText = inIter->getAttribute("numVertices");
	if(lNumVerticesText.empty()==false)
		mNumVertices = str2uint(lNumVerticesText);

	std::string lLastSymmetryBreakGenerationText = inIter->getAttribute("genLastSymmetryBreak");
	if(lLastSymmetryBreakGenerationText.empty()==false)
		mLastSymmetryBreakGeneration = str2uint(lLastSymmetryBreakGenerationText);

	std::string lLastIsEdgeMutationGenerationText = inIter->getAttribute("genLastIsEdgeMutation");
	if(lLastIsEdgeMutationGenerationText.empty()==false)
		mLastIsEdgeMutationGeneration = str2uint(lLastIsEdgeMutationGenerationText);

	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if(lChild->getType() == PACC::XML::eData) {
            if (lChild->getValue() == "LatticePath") {
                readVectorContent(lChild->getFirstChild(), mLatticePath);
            }
            else if (lChild->getValue() == "SizePath") {
                readVectorContent(lChild->getFirstChild(), mSizePath);
            }
            else if (lChild->getValue() == "Node") {
                std::string lIndexText = lChild->getAttribute("index");
                if(lIndexText.empty()) {
                    std::ostringstream lOSS;
                    lOSS << "Node with unknown index in genotype tree!";
                    throw Beagle_IOExceptionNodeM(*lChild, lOSS.str());
                }
                Node& lNode = operator[](str2uint(lIndexText));

                std::string lParentText = lChild->getAttribute("parent");
                if(lParentText.empty()==false)
                    lNode.mParent = str2uint(lParentText);

                for(PACC::XML::ConstIterator lChild2=lChild->getFirstChild(); lChild2; ++lChild2) {
                    if(lChild2->getType() == PACC::XML::eData) {
                        if(lChild2->getValue() == "Color") {
                            PACC::XML::ConstIterator lChild3 = lChild2->getFirstChild();
                            if((!lChild3) || (lChild3->getType()!=PACC::XML::eString))
                                throw Beagle_IOExceptionNodeM(*lChild3, "Expected bit string for the color!");

                            lNode.mColor.clear();
                            std::istringstream lISS(lChild3->getValue());
                            while(lISS.eof() == false) {
                                char lBuf = '\0';
                                lISS >> lBuf;
                                if(lBuf=='\0') break;
                                if((lBuf != '0') && (lBuf != '1')) {
                                    throw Beagle_IOExceptionNodeM(*lChild3, "Bit string value is not 0 or 1!");
                                }
                                lNode.mColor.push_back((lBuf == '1') ? true : false);
                            }
                        }
                        else if(lChild2->getValue() == "Children") {
                            readVectorContent(lChild2->getFirstChild(), lNode.mChildren);
                        }
                        else if(lChild2->getValue() == "Genotype") {
                            lNode.mParameters.readWithContext(lChild2, ioContext);
                        }
                    }
                }
            }
        }
	}

	Beagle_StackTraceEndM("void ENSO::Tree::readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext)");
}


/*!
 *  \brief Read an unsigned integer vector from an XML subtree.
 *  \param inIter XML iterator to use to read the genotype tree from.
 *  \param outVector The vector to read the integers into.
 */
void ENSO::Tree::readVectorContent(PACC::XML::ConstIterator inIter, std::vector<unsigned int>& outVector)
{
	Beagle_StackTraceBeginM();

    if(inIter->getType()!=PACC::XML::eString)
        throw Beagle_IOExceptionNodeM(*inIter,"Expected vector of unsigned integers!");

    outVector.clear();
    std::istringstream lISS(inIter->getValue());
    while(lISS.good()) {
        unsigned int lValue;
        lISS >> lValue;
        outVector.push_back(lValue);
        if(lISS.good()==false) break;
        int lDelim=lISS.get();
        if((lISS.good()==false) || (lDelim==-1)) break;
    }

	Beagle_StackTraceEndM("void ENSO::Tree::readVectorContent(PACC::XML::ConstIterator inIter, std::vector<unsigned int>& outVector)");
}


/*!
 *  \brief Write an ENSO genotype tree into a XML streamer.
 *  \param ioStreamer XML streamer to write the genotype tree into.
 *  \param inIndent Whether XML output should be indented.
 */
void ENSO::Tree::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();

    ioStreamer.insertAttribute("size", uint2str(size()));
    ioStreamer.insertAttribute("numVertices", uint2str(mNumVertices));
    ioStreamer.insertAttribute("genLastSymmetryBreak", uint2str(mLastSymmetryBreakGeneration));
    ioStreamer.insertAttribute("genLastIsEdgeMutation", uint2str(mLastIsEdgeMutationGeneration));

    if (mLatticePath.size() > 0) {
        ioStreamer.openTag("LatticePath", false);
        writeVectorContent(ioStreamer, mLatticePath);
        ioStreamer.closeTag();
    }

    if (mSizePath.size() > 0) {
        ioStreamer.openTag("SizePath", false);
        writeVectorContent(ioStreamer, mSizePath);
        ioStreamer.closeTag();
    }

	for(unsigned int i = 0; i < size(); ++i) {
        ioStreamer.openTag("Node", inIndent);
        ioStreamer.insertAttribute("index", uint2str(i));
        const Node& lNode = operator[](i);
        if (lNode.mParent < UINT_MAX)
            ioStreamer.insertAttribute("parent", uint2str(lNode.mParent));

        ioStreamer.openTag("Color", false);
        ioStreamer.insertAttribute("size", uint2str(lNode.mColor.size()));
        std::ostringstream lOSS;
        for(unsigned int j=0; j<lNode.mColor.size(); j++) {
            lOSS << (lNode.mColor[j] ? '1' : '0');
        }
        ioStreamer.insertStringContent(lOSS.str());
        ioStreamer.closeTag();

        if (lNode.mChildren.size() > 0) {
            ioStreamer.openTag("Children", false);
            writeVectorContent(ioStreamer, lNode.mChildren);
            ioStreamer.closeTag();
        }

        if (lNode.mParameters.size() > 0) {
            lNode.mParameters.write(ioStreamer, inIndent);
        }

        ioStreamer.closeTag();
	}

	Beagle_StackTraceEndM("void ENSO::Tree::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}


/*!
 *  \brief Write an unsigned integer vector into a XML streamer.
 *  \param ioStreamer XML streamer to write the genotype tree into.
 *  \param inVector The vector to be written.
 */
void ENSO::Tree::writeVectorContent(PACC::XML::Streamer& ioStreamer, const std::vector<unsigned int>& inVector) const
{
	Beagle_StackTraceBeginM();

    std::ostringstream lOSS;
    for(unsigned int i=0; i<inVector.size(); ++i) {
        lOSS << inVector[i];
        if(i != (inVector.size()-1)) lOSS << ';';
    }
    ioStreamer.insertStringContent(lOSS.str());

	Beagle_StackTraceEndM("void ENSO::Tree::writeVectorContent(PACC::XML::Streamer& ioStreamer, const std::vector<unsigned int>& inVector) const");
}


/*!
 *  \brief Get all the leaves of the genotype tree.
 *  \param outLeaves Contains the indices of leaves when call returns.
 */
void ENSO::Tree::getLeaves(std::vector<unsigned int>& outLeaves) const
{
	Beagle_StackTraceBeginM();

    outLeaves.clear();
	for(unsigned int i = 0; i < size(); ++i) {
        const Node& lNode = operator[](i);
        if (lNode.mChildren.size() == 0) {
            outLeaves.push_back(i);
        }
    }

	Beagle_StackTraceEndM("void ENSO::Tree::getLeaves(std::vector<unsigned int>& outLeaves) const");
}


/*!
 *  \brief Get all the vertex leaves of the genotype tree.
 *  \param outLeaves Contains the indices of vertex leaves when call returns.
 */
void ENSO::Tree::getVertexLeaves(std::vector<unsigned int>& outLeaves) const
{
	Beagle_StackTraceBeginM();

    outLeaves.clear();
	for(unsigned int i = 0; i < size(); ++i) {
        const Node& lNode = operator[](i);
        if ((lNode.mChildren.size() == 0) && (lNode.mParameters.find("isedge") == lNode.mParameters.end())) {
            outLeaves.push_back(i);
        }
    }

	Beagle_StackTraceEndM("void ENSO::Tree::getVertexLeaves(std::vector<unsigned int>& outLeaves) const");
}


/*!
 *  \brief Get all the edge leaves of the genotype tree.
 *  \param outLeaves Contains the indices of edge leaves when call returns.
 */
void ENSO::Tree::getEdgeLeaves(std::vector<unsigned int>& outLeaves) const
{
	Beagle_StackTraceBeginM();

    outLeaves.clear();
	for(unsigned int i = 0; i < size(); ++i) {
        const Node& lNode = operator[](i);
        if ((lNode.mChildren.size() == 0) && (lNode.mParameters.find("isedge") != lNode.mParameters.end())) {
            outLeaves.push_back(i);
        }
    }

	Beagle_StackTraceEndM("void ENSO::Tree::getEdgeLeaves(std::vector<unsigned int>& outLeaves) const");
}


/*!
 *  \brief Get all the vertex nodes of the genotype tree.
 *  \param outNodes Contains the indices of vertex nodes when call returns.
 */
void ENSO::Tree::getVertexNodes(std::vector<unsigned int>& outNodes) const
{
	Beagle_StackTraceBeginM();

    outNodes.clear();
	for(unsigned int i = 1; i < size(); ++i) {
        const Node& lNode = operator[](i);
        if (lNode.mParameters.find("isedge") == lNode.mParameters.end()) {
            outNodes.push_back(i);
        }
    }

	Beagle_StackTraceEndM("void ENSO::Tree::getVertexNodes(std::vector<unsigned int>& outNodes) const");
}


/*!
 *  \brief Get all the edge nodes of the genotype tree.
 *  \param outNodes Contains the indices of edge nodes when call returns.
 */
void ENSO::Tree::getEdgeNodes(std::vector<unsigned int>& outNodes) const
{
	Beagle_StackTraceBeginM();

    outNodes.clear();
	for(unsigned int i = 1; i < size(); ++i) {
        const Node& lNode = operator[](i);
        if (lNode.mParameters.find("isedge") != lNode.mParameters.end()) {
            outNodes.push_back(i);
        }
    }

	Beagle_StackTraceEndM("void ENSO::Tree::getEdgeNodes(std::vector<unsigned int>& outNodes) const");
}


/*!
 *  \brief Get all the leaves of the genotype tree that represent parallel edges
 *      (edges with no common vertex).
 *  \param outLeaves Contains the indices of parallel edge leaves when call returns.
 */
void ENSO::Tree::getParallelEdgeLeaves(std::vector<unsigned int>& outLeaves) const
{
	Beagle_StackTraceBeginM();

    outLeaves.clear();
	for(unsigned int i = 0; i < size(); ++i) {
        const Node& lNode = operator[](i);
        if ((lNode.mChildren.size() == 0) && (lNode.mParameters.find("isedge") != lNode.mParameters.end())) {
            const std::vector<bool>& lColor = lNode.mColor;
            // Treat the node color as an adjacency matrix and verify that it is
            // its rows and columns contain not more than a single one.
            unsigned int lNumOnes = 0;
            for (unsigned int j = 0; j < mNumVertices && lNumOnes <= 1; ++j) { // rows
                lNumOnes = std::count(lColor.begin()+mNumVertices*j, lColor.begin()+mNumVertices*(j+1), true);
            }
            for (unsigned int j = 0; j < mNumVertices && lNumOnes <= 1; ++j) { // columns
                lNumOnes = 0;
                for (unsigned int k = 0; k < mNumVertices && lNumOnes <= 1; ++k) {
                    if (lColor[k*mNumVertices + j]) lNumOnes++;
                }
            }
            if (lNumOnes <= 1) {
                outLeaves.push_back(i);
            }
        }
    }

	Beagle_StackTraceEndM("void ENSO::Tree::getParallelEdgeLeaves(std::vector<unsigned int>& outLeaves) const");
}


/*!
 *  \brief Record the current symmetry state of the genotype tree.
 *  \param inGeneration The current generation.
 *  \param inLatticeIndex The current index on the subgroup lattice.
 */
void ENSO::Tree::recordSymmetryState(unsigned int inGeneration, unsigned int inLatticeIndex)
{
	Beagle_StackTraceBeginM();

    if (mSizePath.size() == 0 || mSizePath.back() != size()) {
        mLastSymmetryBreakGeneration = inGeneration;
        mLastIsEdgeMutationGeneration = inGeneration;
    }

    mLatticePath.push_back(inLatticeIndex);
    mSizePath.push_back(size());

	Beagle_StackTraceEndM("void ENSO::Tree::recordSymmetryState(unsigned int inGeneration, unsigned int inLatticeIndex)");
}


/*!
 *  \brief Restore the genotype tree to the previous symmetry state.
 *  \param inGeneration The current generation.
 *  \return True if symmetry was successfully restored, false otherwise.
 */
bool ENSO::Tree::restoreSymmetryState(unsigned int inGeneration)
{
	Beagle_StackTraceBeginM();

    if (mSizePath.size() < 2) {
        return false;
    }
    else {
        mLatticePath.pop_back();
        mSizePath.pop_back();

        if (mSizePath.back() != size()) {
            mLastSymmetryBreakGeneration = inGeneration;
            mLastIsEdgeMutationGeneration = inGeneration;
        }

        for (unsigned int i=size()-1; i>=mSizePath.back(); i--) {
            unsigned int lParent = (*this)[i].mParent;
            (*this)[lParent].mChildren.pop_back();
            (*this)[lParent].mParameters.setInvalid();
        }
        resize(mSizePath.back());

        return true;
    }

	Beagle_StackTraceEndM("bool ENSO::Tree::restoreSymmetryState(unsigned int inGeneration)");
}
