/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2007 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GA/src/PackageIntegerVector.cpp
 *  \brief  Source code of class GA::PackageIntegerVector.
 *  \author Christian Gagne
 *  $Revision: 1.4 $
 *  $Date: 2007/09/17 17:25:49 $
 */

#include "beagle/GA.hpp"

using namespace Beagle;


/*!
 *  \brief Construct package with integer vector GA objects.
 *  \param inInitSize Size of vectors generated by initialization.
 */
GA::PackageIntegerVector::PackageIntegerVector(unsigned int inInitSize) :
		Package("GA-PackageIntegerVector"),
		mInitSize(inInitSize)
{ }


/*!
 *  \brief Configure system by adding necessary components into it.
 *  \param ioSystem System to configure.
 */
void GA::PackageIntegerVector::configure(System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Factory& lFactory = ioSystem.getFactory();

	// Add available basic types to the factory
	lFactory.insertAllocator("Beagle::GA::CrossoverIndicesIntVecOp", new GA::CrossoverIndicesIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::CrossoverOnePointIntVecOp", new GA::CrossoverOnePointIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::CrossoverTwoPointsIntVecOp", new GA::CrossoverTwoPointsIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::CrossoverUniformIntVecOp", new GA::CrossoverUniformIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::InitIndicesIntVecOp", new GA::InitIndicesIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::InitIntVecOp", new GA::InitIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::InitQRIntVecOp", new GA::InitQRIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::IntegerVector", new GA::IntegerVector::Alloc);
	lFactory.insertAllocator("Beagle::GA::MutationQRUniformIntVecOp", new GA::MutationQRUniformIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::MutationShuffleIntVecOp", new GA::MutationShuffleIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::MutationUniformIntVecOp", new GA::MutationUniformIntVecOp::Alloc);
	lFactory.insertAllocator("Beagle::GA::PackageIntegerVector", new GA::PackageIntegerVector::Alloc);

	// Aliases to short name of operators
	lFactory.aliasAllocator("Beagle::GA::CrossoverIndicesIntVecOp", "GA-CrossoverIndicesIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::CrossoverOnePointIntVecOp", "GA-CrossoverOnePointIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::CrossoverTwoPointsIntVecOp", "GA-CrossoverTwoPointsIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::CrossoverUniformIntVecOp", "GA-CrossoverUniformIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::InitIndicesIntVecOp", "GA-InitIndicesIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::InitIntVecOp", "GA-InitIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::InitQRIntVecOp", "GA-InitQRIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::IntegerVector", "GA-IntegerVector");
	lFactory.aliasAllocator("Beagle::GA::MutationQRUniformIntVecOp", "GA-MutationQRUniformIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::MutationShuffleIntVecOp", "GA-MutationShuffleIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::MutationUniformIntVecOp", "GA-MutationUniformIntVecOp");
	lFactory.aliasAllocator("Beagle::GA::PackageIntegerVector", "GA-PackageIntegerVector");

	// Set some basic concept-type associations in factory
	lFactory.setConcept("CrossoverOp", "GA-CrossoverUniformIntVecOp");
	lFactory.setConcept("Genotype", "GA-IntegerVector");
	lFactory.setConcept("InitializationOp", "GA-InitIntVecOp");
	lFactory.setConcept("MutationOp", "GA-MutationUniformIntVecOp");

	// Add components
	if(ioSystem.haveComponent("QuasiRandom") == NULL) {
		ioSystem.addComponent(new QuasiRandom);
	}

	// Set default initialization size
	ioSystem.getRegister().setEntryDefaultValue("ga.init.vectorsize",
	        new UInt(mInitSize),
	        uint2str(mInitSize));

	Beagle_StackTraceEndM("void GA::PackageIntegerVector::configureSystem(System&)");
}


/*!
 *  \brief List packages which current package depends.
 *  \return Bag with the packages instanciation on which current package depends.
 *
 *  PackageIntegerVector depends on PackageBase.
 */
Package::Bag GA::PackageIntegerVector::listDependencies(void)
{
	Beagle_StackTraceBeginM();
	Package::Bag lDependencies;
	lDependencies.push_back(new PackageBase());
	return lDependencies;
	Beagle_StackTraceEndM("Package::Bag GA::PackageIntegerVector::listDependencies(void)");
}
