
/*!
 *  \file   beagle/NN/FitnessIndexed.hpp
 *  \brief  Definition of the class NN/FitnessIndexed.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_NN_FitnessIndexed_hpp
#define Beagle_NN_FitnessIndexed_hpp

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/AllocatorT.hpp"
#include "beagle/PointerT.hpp"
#include "beagle/ContainerT.hpp"
#include "beagle/Fitness.hpp"

namespace Beagle
{
namespace NN
{


/*!
 *  \class FitnessIndexed beagle/NN/FitnessIndexed.hpp "beagle/NN/FitnessIndexed.hpp"
 *  \brief Fitness class containing an array of indexed fitness values.
 *  \ingroup NNF
 */
class FitnessIndexed : public Fitness, public std::vector<float>
{

public:

	//! NN::FitnessIndexed allocator type.
	typedef AllocatorT<FitnessIndexed,Fitness::Alloc> Alloc;
	//! NN::FitnessIndexed handle type.
	typedef PointerT<FitnessIndexed,Fitness::Handle> Handle;
	//! NN::FitnessIndexed bag type.
	typedef ContainerT<FitnessIndexed,Fitness::Bag> Bag;

    FitnessIndexed();
	explicit FitnessIndexed(unsigned int inSize, float inValue);
	explicit FitnessIndexed(const std::vector<float>& inFitness);
	virtual ~FitnessIndexed()
	{ }

    virtual void                copy(const Member& inOriginal, System& ioSystem);
	virtual float               getDistance(const Fitness& inRightFitness) const;
    virtual unsigned int        getNumberOfObjectives() const;
	virtual float               getObjective(unsigned int inObjIndex) const;
    virtual const std::string&  getType() const;
	virtual bool                isDominated(const Fitness& inRightFitness) const;
	virtual bool                isEqual(const Object& inRightObj) const;
	virtual bool                isLess(const Object& inRightObj) const;

	virtual void                read(PACC::XML::ConstIterator inIter);
	virtual void                writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;

	float        getValue() const;
	void         setValue(float inValue);

	float        getValue(unsigned int inIndex) const;
	void         setValue(unsigned int inIndex, float inValue);

	static unsigned int getIndex();
	static void         setIndex(unsigned int inIndex);

protected:

	static unsigned int smIndex;   //!< Index for selecting a fitness value.

};

}
}


/*!
 *  \brief  Get number of objectives of this fitness.
 *  \return 1, an indexed fitness is made of a single objective.
 */
inline unsigned int Beagle::NN::FitnessIndexed::getNumberOfObjectives() const
{
	Beagle_StackTraceBeginM();
	return 1;
	Beagle_StackTraceEndM("unsigned int Beagle::NN::FitnessIndexed::getNumberOfObjectives() const");
}


/*!
 *  \return The fitness value selected by the current index.
 */
inline float Beagle::NN::FitnessIndexed::getValue() const
{
	Beagle_StackTraceBeginM();
	return (*this)[smIndex];
	Beagle_StackTraceEndM("float Beagle::NN::FitnessIndexed::getValue() const");
}


/*!
 *  \param inValue The fitness value to be set for the current index.
 */
inline void Beagle::NN::FitnessIndexed::setValue(float inValue)
{
	Beagle_StackTraceBeginM();
	(*this)[smIndex] = inValue;
	Beagle_StackTraceEndM("void Beagle::NN::FitnessIndexed::setValue(float inValue)");
}


/*!
 *  \return The fitness value at the given index.
 */
inline float Beagle::NN::FitnessIndexed::getValue(unsigned int inIndex) const
{
	Beagle_StackTraceBeginM();
	return (*this)[inIndex];
	Beagle_StackTraceEndM("float NN::FitnessIndexed::getValue(unsigned int inIndex) const");
}


/*!
 *  \param inIndex The index at which to set the value.
 *  \param inValue The fitness value.
 */
inline void Beagle::NN::FitnessIndexed::setValue(unsigned int inIndex, float inValue)
{
	Beagle_StackTraceBeginM();
	(*this)[inIndex] = inValue;
	Beagle_StackTraceEndM("void Beagle::NN::FitnessIndexed::setValue(unsigned int inIndex, float inValue)");
}


/*!
 *  \return The current fitness value index.
 */
inline unsigned int Beagle::NN::FitnessIndexed::getIndex()
{
	Beagle_StackTraceBeginM();
	return smIndex;
	Beagle_StackTraceEndM("static unsigned int NN::FitnessIndexed::getIndex()");
}


/*!
 *  \param inIndex The current index of fitness value.
 */
inline void Beagle::NN::FitnessIndexed::setIndex(unsigned int inIndex)
{
	Beagle_StackTraceBeginM();
	smIndex = inIndex;
	Beagle_StackTraceEndM("static void Beagle::NN::FitnessIndexed::setIndex(unsigned int inIndex)");
}


#endif // Beagle_NN_FitnessIndexed_hpp

