
/*!
 *  \file   beagle/NN/Graph.hpp
 *  \brief  Definition of the type NN::Graph.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_NN_Graph_hpp
#define Beagle_NN_Graph_hpp

#include <string>
#include <map>

#if defined(BEAGLE_HAVE_STD_HASHMAP) | defined(BEAGLE_HAVE_STDEXT_HASHMAP)
#include <hash_map>
#endif // defined(BEAGLE_HAVE_STD_HASHMAP) | defined(BEAGLE_HAVE_STDEXT_HASHMAP)

#ifdef BEAGLE_HAVE_GNUCXX_HASHMAP
#include <ext/hash_map>
#endif // BEAGLE_HAVE_GNUCXX_HASHMAP

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/Container.hpp"
#include "beagle/Genotype.hpp"
#include "beagle/NN/MapT.hpp"
#include "beagle/NN/Node.hpp"
#include "beagle/NN/Module.hpp"
#include "beagle/Randomizer.hpp"

namespace Beagle
{
namespace NN
{


/*!
 *  \class Graph beagle/NN/Graph.hpp "beagle/NN/Graph.hpp"
 *  \brief Genotype graph class for Neural Networks.
 *  \ingroup NNF
 */
class Graph : public Genotype, public MapT<Node>
{

public:

	//! NN::Graph allocator type.
	typedef AllocatorT<Graph, Genotype::Alloc>
	Alloc;
	//! NN::Graph handle type.
	typedef PointerT<Graph, Genotype::Handle>
	Handle;
	//! NN::Graph bag type.
	typedef ContainerT<Graph, Genotype::Bag>
	Bag;

	//! Type for specifying a pair of unsigned ints.
	typedef std::pair<unsigned int, unsigned int> UIntPair;

	//! Type specifying the activations of a set of nodes.
#if defined(BEAGLE_HAVE_STD_HASHMAP)
	typedef std::hash_map<unsigned int, double> Activations;
#elif defined(BEAGLE_HAVE_GNUCXX_HASHMAP)
	typedef __gnu_cxx::hash_map<unsigned int, double> Activations;
#elif defined(BEAGLE_HAVE_STDEXT_HASHMAP)
	typedef stdext::hash_map<unsigned int, double> Activations;
#else // No hash_map found
	typedef std::map<unsigned int, double> Activations;
#endif

	explicit Graph();
	virtual ~Graph()
	{ }

	virtual void                copy(const Member& inOriginal, System& ioSystem);
	virtual unsigned int        getSize() const;
	virtual const std::string&  getType() const;

	virtual bool         isEqual(const Object& inRightObj) const;
	virtual bool         isLess(const Object& inRightObj) const;
	virtual void         readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext);
	virtual void         writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;

	bool            hasNode(unsigned int inNodeId) const;
	const Node&     getNode(unsigned int inNodeId) const;
	Node&           getNode(unsigned int inNodeId);
	Node&           addNode(const Node& inNode);
	bool            delNode(const Node& inNode);
	bool            delNode(unsigned int inNodeId);
	unsigned int    getNumberNodes() const;
	const Node&     getRandomNode(Beagle::Randomizer& ioRandomizer) const;
	Node&           getRandomNode(Beagle::Randomizer& ioRandomizer);
	UIntPair        getRandomNode(Beagle::Randomizer& ioRandomizer, bool inIncludeMain, bool inIncludeInput) const;
	UIntPair        getRandomNode(Beagle::Randomizer& ioRandomizer, unsigned int inModuleKey, bool inIncludeMain, bool inIncludeInput) const;

	unsigned int    getNumberEdges() const;
	UIntPair        getRandomEdge(Beagle::Randomizer& ioRandomizer) const;
	UIntPair        getRandomEdge(Beagle::Randomizer& ioRandomizer, unsigned int inModuleKey) const;

	const Module&   getMain() const;
	Module&         getMain();

	bool            hasModule(unsigned int inModuleKey) const;
	const Module&   getModule(unsigned int inModuleKey) const;
	Module&         getModule(unsigned int inModuleKey);
	Module&         addModule(const Module& inModule);
	bool            delModule(const Module& inModule);
	bool            delModule(unsigned int inModuleKey);
	unsigned int    getNumberModules() const;
	const Module&   getRandomModule(Beagle::Randomizer& ioRandomizer) const;
	Module&         getRandomModule(Beagle::Randomizer& ioRandomizer);
	const MapT<Module>&  getModules() const;

	void            getActivations(Activations& outActivations) const;
	double          getActivation(unsigned int inNodeId) const;
	double          getPreviousActivation(unsigned int inNodeId) const;
	unsigned int    getActivationCount() const;
	void            activate(const Activations& inActivations);
	void            resetActivation();

	void            writeGraphviz(std::ostream& ioStream, const Beagle::Context& inContext) const;
	std::string     getNodeColor(unsigned int inNodeId, double inMinValue, double inMaxValue) const;
	std::string     getEdgeColor(unsigned int inSourceId, unsigned int inTargetId, double inMinValue, double inMaxValue) const;

protected:

	Module       mMain;               //!< The main module which connects to other modules
	MapT<Module> mModules;            //!< The modules in the graph besides main
	unsigned int mActivationCount;    //!< Number of times the graph was activated

	void         setActivation(unsigned int inNodeId, double inActivation);
	void         activate(unsigned int inNodeId, double inActivationSum);

	unsigned int hue2rgb(double inM1, double inM2, double inH) const;
	std::string  hsl2rgb(double inH, double inS, double inL) const;

};

}
}


/*!
 *  \return Number of nodes in the graph.
 */
inline unsigned int Beagle::NN::Graph::getSize() const
{
	Beagle_StackTraceBeginM();
	return getNumberNodes();
	Beagle_StackTraceEndM("unsigned int NN::Graph::getSize() const");
}


/*!
 *  \return True if the node specified by the id exists, false otherwise.
 */
inline bool Beagle::NN::Graph::hasNode(unsigned int inNodeId) const
{
	Beagle_StackTraceBeginM();
	return MapT<Node>::hasKey(inNodeId);
	Beagle_StackTraceEndM("bool NN::Graph::hasNode(unsigned int inNodeId) const");
}


/*!
 *  \return The handle to the node specified by the id.
 */
inline const Beagle::NN::Node& Beagle::NN::Graph::getNode(unsigned int inNodeId) const
{
	Beagle_StackTraceBeginM();
	return MapT<Node>::getData(inNodeId);
	Beagle_StackTraceEndM("const NN::Node& NN::Graph::getNode(unsigned int inNodeId) const");
}


/*!
 *  \return The handle to the node specified by the id.
 */
inline Beagle::NN::Node& Beagle::NN::Graph::getNode(unsigned int inNodeId)
{
	Beagle_StackTraceBeginM();
	return MapT<Node>::getData(inNodeId);
	Beagle_StackTraceEndM("NN::Node& NN::Graph::getNode(unsigned int inNodeId)");
}


/*!
 *  \brief Add a node to the graph.
 *  \param inNode The node to be added.
 *  \return Reference to either the added node or an existing node with the same id.
 */
inline Beagle::NN::Node& Beagle::NN::Graph::addNode(const Node& inNode)
{
	Beagle_StackTraceBeginM();
	return MapT<Node>::addData(inNode.getIdentifier(), inNode);
	Beagle_StackTraceEndM("NN::Node& NN::Graph::addNode(const NN::Node& inNode)");
}


/*!
 *  \brief Delete a node from the graph.
 *  \param inNode The node to be deleted.
 *  \return True if the node was deleted, false otherwise.
 */
inline bool Beagle::NN::Graph::delNode(const Node& inNode)
{
	Beagle_StackTraceBeginM();
	return MapT<Node>::delData(inNode.getIdentifier());
	Beagle_StackTraceEndM("bool NN::Graph::delNode(const NN::Node& inNode)");
}


/*!
 *  \brief Delete a node from the graph.
 *  \param inNodeId The identifier of the node to be deleted.
 *  \return True if the node was deleted, false otherwise.
 */
inline bool Beagle::NN::Graph::delNode(unsigned int inNodeId)
{
	Beagle_StackTraceBeginM();
	return MapT<Node>::delData(inNodeId);
	Beagle_StackTraceEndM("bool NN::Graph::delNode(unsigned int inNodeId)");
}


/*!
 * \return The number of nodes in the graph.
 */
inline unsigned int Beagle::NN::Graph::getNumberNodes() const
{
	Beagle_StackTraceBeginM();
	return size();
	Beagle_StackTraceEndM("unsigned int NN::Graph::getNumberNodes() const");
}


/*!
 *  \return The handle to a random node chosen with uniform probability.
 */
inline const Beagle::NN::Node& Beagle::NN::Graph::getRandomNode(Beagle::Randomizer& ioRandomizer) const
{
	Beagle_StackTraceBeginM();
	return MapT<Node>::getRandomData(ioRandomizer);
	Beagle_StackTraceEndM("const NN::Node& NN::Graph::getRandomNode(Beagle::Randomizer& ioRandomizer) const");
}


/*!
 *  \return The handle to a random node chosen with uniform probability.
 */
inline Beagle::NN::Node& Beagle::NN::Graph::getRandomNode(Beagle::Randomizer& ioRandomizer)
{
	Beagle_StackTraceBeginM();
	return MapT<Node>::getRandomData(ioRandomizer);
	Beagle_StackTraceEndM("NN::Node& NN::Graph::getRandomNode(Beagle::Randomizer& ioRandomizer)");
}


/*!
 *  \return The number of edges in the graph.
 */
inline unsigned int Beagle::NN::Graph::getNumberEdges() const
{
	Beagle_StackTraceBeginM();
	unsigned int lCount = 0;
	for (const_iterator lIter = begin(); lIter != end(); ++lIter) {
		lCount += lIter->second.getNumberEdges();
	}
	return lCount;
	Beagle_StackTraceEndM("unsigned int NN::Graph::getNumberEdges() const");
}


/*!
 *  \return The handle to the main module.
 */
inline const Beagle::NN::Module& Beagle::NN::Graph::getMain() const
{
	Beagle_StackTraceBeginM();
	return mMain;
	Beagle_StackTraceEndM("const NN::Module& NN::Graph::getMain() const");
}


/*!
 *  \return The handle to the main module.
 */
inline Beagle::NN::Module& Beagle::NN::Graph::getMain()
{
	Beagle_StackTraceBeginM();
	return mMain;
	Beagle_StackTraceEndM("NN::Module& NN::Graph::getMain()");
}


/*!
 *  \return True if the module specified by the key exists, false otherwise.
 */
inline bool Beagle::NN::Graph::hasModule(unsigned int inModuleKey) const
{
	Beagle_StackTraceBeginM();
	return mModules.hasKey(inModuleKey);
	Beagle_StackTraceEndM("bool NN::Graph::hasModule(unsigned int inModuleKey) const");
}


/*!
 *  \return The handle to the module specified by the key.
 */
inline const Beagle::NN::Module& Beagle::NN::Graph::getModule(unsigned int inModuleKey) const
{
	Beagle_StackTraceBeginM();
	return (inModuleKey < UINT_MAX) ? mModules.getData(inModuleKey) : mMain;
	Beagle_StackTraceEndM("const NN::Module& NN::Graph::getModule(unsigned int inModuleKey) const");
}


/*!
 *  \return The handle to the module specified by the key.
 */
inline Beagle::NN::Module& Beagle::NN::Graph::getModule(unsigned int inModuleKey)
{
	Beagle_StackTraceBeginM();
	return (inModuleKey < UINT_MAX) ? mModules.getData(inModuleKey) : mMain;
	Beagle_StackTraceEndM("NN::Module& NN::Graph::getModule(unsigned int inModuleKey)");
}


/*!
 *  \brief Add a module to the graph.
 *  \param inModule The module to be added.
 *  \return Reference to either the added module or an existing module with the same key.
 */
inline Beagle::NN::Module& Beagle::NN::Graph::addModule(const Module& inModule)
{
	Beagle_StackTraceBeginM();
	return mModules.addData(inModule.getKey(), inModule);
	Beagle_StackTraceEndM("NN::Module& NN::Graph::addModule(const NN::Module& inModule)");
}


/*!
 *  \brief Delete a module from the graph.
 *  \param inModule The module to be deleted.
 *  \return True if the module was deleted, false otherwise.
 */
inline bool Beagle::NN::Graph::delModule(const Module& inModule)
{
	Beagle_StackTraceBeginM();
	return mModules.delData(inModule.getKey());
	Beagle_StackTraceEndM("bool NN::Graph::delModule(const NN::Module& inModule)");
}


/*!
 *  \brief Delete a module from the graph.
 *  \param inModuleKey The key of the module to be deleted.
 *  \return True if the module was deleted, false otherwise.
 */
inline bool Beagle::NN::Graph::delModule(unsigned int inModuleKey)
{
	Beagle_StackTraceBeginM();
	return mModules.delData(inModuleKey);
	Beagle_StackTraceEndM("bool NN::Graph::delModule(unsigned int inModuleKey)");
}


/*!
 * \return The number of modules in the graph.
 */
inline unsigned int Beagle::NN::Graph::getNumberModules() const
{
	Beagle_StackTraceBeginM();
	return mModules.size();
	Beagle_StackTraceEndM("unsigned int NN::Graph::getNumberModules() const");
}


/*!
 *  \return The handle to a random module chosen with uniform probability.
 */
inline const Beagle::NN::Module& Beagle::NN::Graph::getRandomModule(Beagle::Randomizer& ioRandomizer) const
{
	Beagle_StackTraceBeginM();
	return mModules.getRandomData(ioRandomizer);
	Beagle_StackTraceEndM("const NN::Module& NN::Graph::getRandomModule(Beagle::Randomizer& ioRandomizer) const");
}


/*!
 *  \return The handle to a random module chosen with uniform probability.
 */
inline Beagle::NN::Module& Beagle::NN::Graph::getRandomModule(Beagle::Randomizer& ioRandomizer)
{
	Beagle_StackTraceBeginM();
	return mModules.getRandomData(ioRandomizer);
	Beagle_StackTraceEndM("NN::Module& NN::Graph::getRandomModule(Beagle::Randomizer& ioRandomizer)");
}


/*!
 * \return The module map of the graph.
 */
inline const Beagle::NN::MapT<Beagle::NN::Module>& Beagle::NN::Graph::getModules() const
{
	Beagle_StackTraceBeginM();
	return mModules;
	Beagle_StackTraceEndM("const NN::MapT<NN::Module>& NN::Graph::getModules() const");
}


/*!
 *  \return The current activation of the specified module.
 */
inline double Beagle::NN::Graph::getActivation(unsigned int inNodeId) const
{
	Beagle_StackTraceBeginM();
	return getNode(inNodeId).getActivation(mActivationCount);
	Beagle_StackTraceEndM("double NN::Graph::getActivation(unsigned int inNodeId) const");
}


/*!
 *  \return The previous activation of the specified node.
 */
inline double Beagle::NN::Graph::getPreviousActivation(unsigned int inNodeId) const
{
	Beagle_StackTraceBeginM();
	return getNode(inNodeId).getActivation(mActivationCount-1);
	Beagle_StackTraceEndM("double NN::Graph::getPreviousActivation(unsigned int inNodeId) const");
}


/*!
 *  \return The current activation count
 */
inline unsigned int Beagle::NN::Graph::getActivationCount() const
{
	Beagle_StackTraceBeginM();
	return mActivationCount;
	Beagle_StackTraceEndM("unsigned int NN::Graph::getActivationCount() const");
}


/*!
 *  \brief Resets the graph and all nodes as not activated.
 */
inline void Beagle::NN::Graph::resetActivation()
{
	Beagle_StackTraceBeginM();
	for (iterator lIter = begin(); lIter != end(); ++lIter) {
		lIter->second.resetActivation();
	}
	mActivationCount = 0;
	Beagle_StackTraceEndM("void NN::Graph::resetActivation()");
}


/*!
 *  \brief Set the current activation of the specified node.
 *  \param inNodeId Node whose activation is to be set.
 *  \param inActivation The activation of the node.
 */
inline void Beagle::NN::Graph::setActivation(unsigned int inNodeId, double inActivation)
{
	Beagle_StackTraceBeginM();
	getNode(inNodeId).setActivation(inActivation, mActivationCount);
	Beagle_StackTraceEndM("void NN::Graph::setActivation(unsigned int inNodeId, double inActivation)");
}


/*!
 *  \brief Activate a node, given the sum of its source activations.
 *  \param inNodeId Node to be activated.
 *  \param inActivationSum The sum of source activations.
 */
inline void Beagle::NN::Graph::activate(unsigned int inNodeId, double inActivationSum)
{
	Beagle_StackTraceBeginM();
	getNode(inNodeId).activate(inActivationSum, mActivationCount);
	Beagle_StackTraceEndM("void Beagle::NN::Graph::activate(unsigned int inNodeId, double inActivationSum)");
}


#endif // Beagle_NN_Graph_hpp
