
/*!
 *  \file   beagle/NN/GraphState.hpp
 *  \brief  Definition of the type NN::GraphState.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */


#ifndef Beagle_NN_GraphState_hpp
#define Beagle_NN_GraphState_hpp

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/ContainerT.hpp"
#include "beagle/UInt.hpp"
#include "beagle/NN/MapT.hpp"
#include "beagle/NN/Graph.hpp"


namespace Beagle
{
namespace NN
{


/*!
 *  \class GraphState beagle/NN/GraphState.hpp "beagle/NN/GraphState.hpp"
 *  \brief The class keeping track of global evolution state of NN graphs.
 *  \ingroup NNF
 */
class GraphState : public Component
{

public:

	//! NN::GraphState allocator type.
	typedef AllocatorT<GraphState,Component::Alloc>
	Alloc;
	//! NN::GraphState handle type.
	typedef PointerT<GraphState,Component::Handle>
	Handle;
	//! NN::GraphState bag type.
	typedef ContainerT<GraphState,Component::Bag>
	Bag;

	GraphState(void);
	virtual ~GraphState()
	{ }

	virtual void registerParams(Beagle::System& ioSystem);
	virtual void init(Beagle::System& ioSystem);
	virtual void readWithSystem(PACC::XML::ConstIterator inIter, Beagle::System& ioSystem);
	virtual void write(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;

	unsigned int getNewNodeIdentifier(const Graph& inGraph, unsigned int inSourceId, unsigned int inTargetId);

	unsigned int getNumberInputs() const;
	unsigned int getNumberOutputs() const;
	unsigned int getNumberHelpers() const;

	unsigned int getFirstInput() const;
	unsigned int getLastInput() const;
	unsigned int getFirstOutput() const;
	unsigned int getLastOutput() const;
	unsigned int getFirstHelper() const;
	unsigned int getLastHelper() const;

	bool isInput(unsigned int inNodeId) const;
	bool isOutput(unsigned int inNodeId) const;
	bool isHelper(unsigned int inNodeId) const;

protected:

	/*!
	 *  \class UIntArray beagle/NN/GraphState.hpp "beagle/NN/GraphState.hpp"
	 *  \brief An array of unsigned integers.
	 *  \ingroup NNF
	 */
class UIntArray : public Beagle::UIntArray
	{

	public:

		virtual void read(PACC::XML::ConstIterator inIter);
		virtual void write(PACC::XML::Streamer& ioStreamer, bool inIndent=true, unsigned int inKey=UINT_MAX) const;

	};

	/*!
	 *  The architecture of the graph stores the node ancestry, i.e., which edges
	 *  are split to create which nodes.  A three-level data structure similar to
	 *  a graph is used for this purpose.  The first level contains target node ids,
	 *  while the second level contains source node ids of all incoming edges that
	 *  were split to produce new nodes.  The ids of such newly created nodes are
	 *  stored in the third level as an array.
	 */
	typedef MapT< MapT< GraphState::UIntArray > > Architecture;

	//! Specifies a range of node ids by the first and last of the range [nid1, nid2]
	typedef std::pair<unsigned int, unsigned int> NodeRange;

	UInt::Handle mNumInputs;        //!< Number of input nodes in the graph.
	UInt::Handle mNumOutputs;       //!< Number of output nodes in the graph.
	UInt::Handle mNumHelpers;       //!< Number of helper nodes in the graph.

	unsigned int mNodeCounter;      //!< Counter for generating node identifiers.
	NodeRange    mInputRange;       //!< Range of input node ids.
	NodeRange    mOutputRange;      //!< Range of output node ids.
	NodeRange    mHelperRange;      //!< Range of helper node ids.
	Architecture mArchitecture;     //!< Ancestry of node creation by splitting edges.

};

}
}


/*!
 *  \return The number of input nodes.
 */
inline unsigned int Beagle::NN::GraphState::getNumberInputs() const
{
	Beagle_StackTraceBeginM();
	return mNumInputs->getWrappedValue();
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getNumberInputs() const");
}


/*!
 *  \return The number of output nodes.
 */
inline unsigned int Beagle::NN::GraphState::getNumberOutputs() const
{
	Beagle_StackTraceBeginM();
	return mNumOutputs->getWrappedValue();
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getNumberOutputs() const");
}


/*!
 *  \return The number of helper nodes.
 */
inline unsigned int Beagle::NN::GraphState::getNumberHelpers() const
{
	Beagle_StackTraceBeginM();
	return mNumHelpers->getWrappedValue();
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getNumberHelpers() const");
}


/*!
 *  \return The node id of the first input node.
 */
inline unsigned int Beagle::NN::GraphState::getFirstInput() const
{
	Beagle_StackTraceBeginM();
	return mInputRange.first;
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getFirstInput() const");
}


/*!
 *  \return The node id of the last input node.
 */
inline unsigned int Beagle::NN::GraphState::getLastInput() const
{
	Beagle_StackTraceBeginM();
	return mInputRange.second;
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getLastInput() const");
}


/*!
 *  \return The node id of the first output node.
 */
inline unsigned int Beagle::NN::GraphState::getFirstOutput() const
{
	Beagle_StackTraceBeginM();
	return mOutputRange.first;
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getFirstOutput() const");
}


/*!
 *  \return The node id of the last output node.
 */
inline unsigned int Beagle::NN::GraphState::getLastOutput() const
{
	Beagle_StackTraceBeginM();
	return mOutputRange.second;
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getLastOutput() const");
}


/*!
 *  \return The node id of the first helper node.
 */
inline unsigned int Beagle::NN::GraphState::getFirstHelper() const
{
	Beagle_StackTraceBeginM();
	return mHelperRange.first;
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getFirstHelper() const");
}


/*!
 *  \return The node id of the last helper node.
 */
inline unsigned int Beagle::NN::GraphState::getLastHelper() const
{
	Beagle_StackTraceBeginM();
	return mHelperRange.second;
	Beagle_StackTraceEndM("unsigned int NN::GraphState::getLastHelper() const");
}


/*!
 *  \return Whether the given node id is an input.
 */
inline bool Beagle::NN::GraphState::isInput(unsigned int inNodeId) const
{
	Beagle_StackTraceBeginM();
	return mInputRange.first <= inNodeId && inNodeId <= mInputRange.second;
	Beagle_StackTraceEndM("bool Beagle::NN::GraphState::isInput(unsigned int inNodeId) const");
}


/*!
 *  \return Whether the given node id is an output.
 */
inline bool Beagle::NN::GraphState::isOutput(unsigned int inNodeId) const
{
	Beagle_StackTraceBeginM();
	return mOutputRange.first <= inNodeId && inNodeId <= mOutputRange.second;
	Beagle_StackTraceEndM("bool Beagle::NN::GraphState::isOutput(unsigned int inNodeId) const");
}


/*!
 *  \return Whether the given node id is an helper.
 */
inline bool Beagle::NN::GraphState::isHelper(unsigned int inNodeId) const
{
	Beagle_StackTraceBeginM();
	return mHelperRange.first <= inNodeId && inNodeId <= mHelperRange.second;
	Beagle_StackTraceEndM("bool Beagle::NN::GraphState::isHelper(unsigned int inNodeId) const");
}


#endif // Beagle_NN_GraphState_hpp
