
/*!
 *  \file   beagle/NN/MapT.hpp
 *  \brief  Definition of class NN::MapT.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_NN_MapT_hpp
#define Beagle_NN_MapT_hpp

#include <map>
#include <string>

#if defined(BEAGLE_HAVE_STD_HASHMAP) | defined(BEAGLE_HAVE_STDEXT_HASHMAP)
#include <hash_map>
#endif // defined(BEAGLE_HAVE_STD_HASHMAP) | defined(BEAGLE_HAVE_STDEXT_HASHMAP)

#ifdef BEAGLE_HAVE_GNUCXX_HASHMAP
#include <ext/hash_map>
#endif // BEAGLE_HAVE_GNUCXX_HASHMAP

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Randomizer.hpp"

namespace Beagle
{
namespace NN
{


/*!
 *  \class MapT beagle/NN/MapT.hpp "beagle/NN/MapT.hpp"
 *  \brief Hash map using unsigned int as key to access instances of type T.
 *  \ingroup NNF
 */
#if defined(BEAGLE_HAVE_STD_HASHMAP)
template <class T>
class MapT : public std::hash_map<unsigned int, T>
#elif defined(BEAGLE_HAVE_GNUCXX_HASHMAP)
template <class T>
class MapT : public __gnu_cxx::hash_map<unsigned int, T>
#elif defined(BEAGLE_HAVE_STDEXT_HASHMAP)
template <class T>
class MapT : public stdext::hash_map<unsigned int, T>
#else // No hash_map found
template <class T>
class MapT : public std::map<unsigned int, T>
#endif
{

public:

	explicit MapT(const std::string& inName="Map");
	virtual ~MapT()
	{ }

	virtual bool isEqual(const MapT<T>& inRightObj) const;
	virtual bool isLess(const MapT<T>& inRightObj) const;
	virtual bool isLessEqual(const MapT<T>& inRightObj) const;
	virtual void read(PACC::XML::ConstIterator inIter);
	virtual void write(PACC::XML::Streamer& ioStreamer, bool inIndent=true, unsigned int inKey=UINT_MAX) const;

	bool hasKey(unsigned int inKey) const;
	const T&   getData(unsigned int inKey) const;
	const T&   getRandomData(Beagle::Randomizer& ioRandomizer) const;
	T&   getData(unsigned int inKey);
	T&   getRandomData(Beagle::Randomizer& ioRandomizer);
	T&   addData(unsigned int inKey, const T& inData);
	bool delData(unsigned int inKey);

	const std::string& getName() const;
	void         setName(const std::string& inName);

protected:

	std::string mName;            //!< Name of the map.

	virtual void readAttributes(PACC::XML::ConstIterator inIter);
	virtual void writeAttributes(PACC::XML::Streamer& ioStreamer, bool inIndent=true, unsigned int inKey=UINT_MAX) const;

};

}
}


/*!
 *  \brief Construct a map from an object allocator
 *  \param inTypeAlloc Allocator for elements
 */
template <class T>
Beagle::NN::MapT<T>::MapT(const std::string& inName) :
		mName(inName)
{ }


/*!
 *  \brief Compare equality of two maps.
 *  \param inRightMap Map to be compared to.
 *  \return True if the maps are equal, false otherwise.
 *  \throw Beagle::BadCastException If compared map objects are not of the same type.
 *
 *  We are interested only in the equality of the keys, so returns true if both
 *  maps have the same keys.
 */
template <class T>
bool Beagle::NN::MapT<T>::isEqual(const Beagle::NN::MapT<T>& inRightMap) const
{
	Beagle_StackTraceBeginM();
	bool lEqual = (MapT<T>::size() == inRightMap.size());
	for(typename MapT<T>::const_iterator lLeftIter = MapT<T>::begin();
	        lEqual && (lLeftIter != MapT<T>::end()); lLeftIter++) {
		lEqual = inRightMap.hasKey(lLeftIter->first);
	}
	return lEqual;
	Beagle_StackTraceEndM("bool NN::MapT<T>::isEqual(const Beagle::NN::MapT<T>& inRightMap) const");
}


/*!
 *  \brief Check if this map is a proper subset of the other.
 *  \param inRightMap Map to be compared to.
 *  \return True if left map is a proper subset of the right one, false otherwise.
 *  \throw Beagle::BadCastException If compared map objects are not of the same type.
 *
 *  We are interested only in the set of keys, so returns true if the set of
 *  keys in the left map is a proper subset of the set of keys in the right map.
 */
template <class T>
bool Beagle::NN::MapT<T>::isLess(const Beagle::NN::MapT<T>& inRightMap) const
{
	Beagle_StackTraceBeginM();
	bool lLess = (MapT<T>::size() < inRightMap.size());
	for(typename MapT<T>::const_iterator lLeftIter = MapT<T>::begin();
	        lLess && (lLeftIter != MapT<T>::end()); lLeftIter++) {
		lLess = inRightMap.hasKey(lLeftIter->first);
	}
	return lLess;
	Beagle_StackTraceEndM("bool NN::MapT<T>::isLess(const Beagle::NN::MapT<T>& inRightMap) const");
}


/*!
 *  \brief Check if this map is a subset of the other.
 *  \param inRightMap Map to be compared to.
 *  \return True if left map is a subset of the right one, false otherwise.
 *  \throw Beagle::BadCastException If compared map objects are not of the same type.
 *
 *  We are interested only in the set of keys, so returns true if the set of
 *  keys in the left map is a subset of the set of keys in the right map.
 */
template <class T>
bool Beagle::NN::MapT<T>::isLessEqual(const Beagle::NN::MapT<T>& inRightMap) const
{
	Beagle_StackTraceBeginM();
	bool lLessEqual = (MapT<T>::size() <= inRightMap.size());
	for(typename MapT<T>::const_iterator lLeftIter = MapT<T>::begin();
	        lLessEqual && (lLeftIter != MapT<T>::end()); lLeftIter++) {
		lLessEqual = inRightMap.hasKey(lLeftIter->first);
	}
	return lLessEqual;
	Beagle_StackTraceEndM("bool NN::MapT<T>::isLessEqual(const Beagle::NN::MapT<T>& inRightMap) const");
}


/*!
 *  \brief Read a map from an XML node.
 *  \param inIter XML iterator to read the Map from.
 *  \throw Beagle::IOException If reading from the stream failed.
 */
template <class T>
void Beagle::NN::MapT<T>::read(PACC::XML::ConstIterator inIter)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType() != PACC::XML::eData) || (inIter->getValue() != getName())) {
		std::ostringstream lOSS;
		lOSS << "tag <" << getName() << "> expected!";
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}

	MapT<T>::clear();

	readAttributes(inIter);

	// Get size attribute and resize map for performance.
#if defined(BEAGLE_HAVE_STD_HASHMAP) | defined(BEAGLE_HAVE_GNUCXX_HASHMAP) | defined(BEAGLE_HAVE_STDEXT_HASHMAP)
	std::string lSizeText = inIter->getAttribute("size");
	if(lSizeText.empty()==false)
		MapT<T>::resize(str2uint(lSizeText));
#endif

	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if(lChild->getType()==PACC::XML::eData) {
			std::string lKeyText = lChild->getAttribute("key");
			if(lKeyText.empty())
				throw Beagle_IOExceptionNodeM(*lChild, "no key attribute for Map entry!");
			unsigned int lKey = str2uint(lKeyText);
			T& lData = addData(lKey, T());
			lData.read(lChild);
		}
	}

	Beagle_StackTraceEndM("void NN::MapT<T>::read(PACC::XML::ConstIterator inIter)");
}


/*!
 *  \brief Write a map into an XML streamer.
 *  \param ioStreamer XML streamer to write the Map to.
 *  \param inIndent Whether XML output should be indented.
 *  \param inKey Key attribute to be used if this map itself is an entry in another map.
 */
template <class T>
void Beagle::NN::MapT<T>::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const
{
	Beagle_StackTraceBeginM();

	ioStreamer.openTag(getName(), inIndent);
	writeAttributes(ioStreamer, inIndent, inKey);
	ioStreamer.insertAttribute("size", uint2str(MapT<T>::size()));
	for(typename MapT<T>::const_iterator lIter = MapT<T>::begin(); lIter != MapT<T>::end(); ++lIter) {
		lIter->second.write(ioStreamer, inIndent, lIter->first);
	}
	ioStreamer.closeTag();

	Beagle_StackTraceEndM("void NN::MapT<T>::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const");
}


/*!
 *  \brief Read any extra attributes from the XML node.
 *  \param inIter XML iterator to read the attributes from.
 */
template <class T>
inline void Beagle::NN::MapT<T>::readAttributes(PACC::XML::ConstIterator inIter)
{ }


/*!
 *  \brief Write any extra attributes into the XML streamer.
 *  \param ioStreamer XML streamer to write the attributes to.
 *  \param inIndent Whether XML output should be indented.
 *  \param inKey Key attribute if this map itself is an entry in another map.
 */
template <class T>
inline void Beagle::NN::MapT<T>::writeAttributes(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const
{
	Beagle_StackTraceBeginM();
	if (inKey < UINT_MAX)
		ioStreamer.insertAttribute("key", uint2str(inKey));
	Beagle_StackTraceEndM("void NN::MapT<T>::writeAttributes(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const");
}


/*!
 *  \brief Check if the map has the given key.
 *  \param inKey The key to be checked.
 *  \return True if the key is in the map; false otherwise.
 */
template <class T>
inline bool Beagle::NN::MapT<T>::hasKey(unsigned int inKey) const
{
	Beagle_StackTraceBeginM();
	return (MapT<T>::find(inKey) != MapT<T>::end());
	Beagle_StackTraceEndM("bool NN::MapT<T>::hasKey(unsigned int inKey) const");
}


/*!
 *  \brief Retrieve the data associated with the given key.
 *  \param inKey The key of the data.
 *  \return The data associated with the key.
 */
template <class T>
const T& Beagle::NN::MapT<T>::getData(unsigned int inKey) const
{
	Beagle_StackTraceBeginM();
	typename MapT<T>::const_iterator lIter = MapT<T>::find(inKey);
	if (lIter == MapT<T>::end()) {
		std::ostringstream lOSS;
		lOSS << "Key <" << inKey << "> does not exist in the map!";
		throw Beagle_RunTimeExceptionM(lOSS.str());
	}
	return lIter->second;
	Beagle_StackTraceEndM("const T& NN::MapT<T>::getData(unsigned int inKey) const");
}


/*!
 *  \brief Retrieve the data for an entry chosen with uniform random probability.
 *  \param ioRandomizer Randomizer used to select element.
 *  \return The data associated with a randomly chosen entry.
 */
template <class T>
const T& Beagle::NN::MapT<T>::getRandomData(Beagle::Randomizer& ioRandomizer) const
{
	Beagle_StackTraceBeginM();
	Beagle_AssertM(MapT<T>::size() > 0);
	unsigned int lIndex = ioRandomizer.rollInteger(0, MapT<T>::size()-1);
	typename MapT<T>::const_iterator lIter = MapT<T>::begin();
	for (unsigned int i = 0;
	        i < lIndex; i++) {
		lIter++;
	}
	return lIter->second;
	Beagle_StackTraceEndM("const T& NN::MapT<T>::getRandomData(Beagle::Randomizer& ioRandomizer) const");
}


/*!
 *  \brief Retrieve the data associated with the given key.
 *  \param inKey The key of the data.
 *  \return The data associated with the key.
 */
template <class T>
T& Beagle::NN::MapT<T>::getData(unsigned int inKey)
{
	Beagle_StackTraceBeginM();
	typename MapT<T>::iterator lIter = MapT<T>::find(inKey);
	if (lIter == MapT<T>::end()) {
		std::ostringstream lOSS;
		lOSS << "Key <" << inKey << "> does not exist in the map!";
		throw Beagle_RunTimeExceptionM(lOSS.str());
	}
	return lIter->second;
	Beagle_StackTraceEndM("T& NN::MapT<T>::getData(unsigned int inKey)");
}


/*!
 *  \brief Retrieve the data for an entry chosen with uniform random probability.
 *  \param ioRandomizer Randomizer used to select element.
 *  \return The data associated with a randomly chosen entry.
 */
template <class T>
T& Beagle::NN::MapT<T>::getRandomData(Beagle::Randomizer& ioRandomizer)
{
	Beagle_StackTraceBeginM();
	Beagle_AssertM(MapT<T>::size() > 0);
	unsigned int lIndex = ioRandomizer.rollInteger(0, MapT<T>::size()-1);
	typename MapT<T>::iterator lIter = MapT<T>::begin();
	for (unsigned int i = 0;
	        i < lIndex; i++) {
		lIter++;
	}
	return lIter->second;
	Beagle_StackTraceEndM("T& NN::MapT<T>::getRandomData(Beagle::Randomizer& ioRandomizer)");
}


/*!
 *  \brief Add a new entry in the map.
 *  \param inKey Key of the data to be added.
 *  \param inData The new data to be added.
 *  \return Reference to either the added data or existing data having the same key.
 */
template <class T>
inline T& Beagle::NN::MapT<T>::addData(unsigned int inKey, const T& inData)
{
	Beagle_StackTraceBeginM();
	return MapT<T>::insert(typename MapT<T>::value_type(inKey, inData)).first->second;
	Beagle_StackTraceEndM("T& Beagle::NN::MapT<T>::addData(unsigned int inKey, const T& inData)");
}


/*!
 *  \brief Delete the data specified by the key.
 *  \param inKey Key of the data to be deleted.
 *  \return True if the data was deleted, false otherwise.
 */
template <class T>
inline bool Beagle::NN::MapT<T>::delData(unsigned int inKey)
{
	Beagle_StackTraceBeginM();
	return MapT<T>::erase(inKey) > 0;
	Beagle_StackTraceEndM("bool NN::MapT<T>::delData(unsigned int inKey)");
}


/*!
 *  \brief Returns the name of the map.
 */
template <class T>
inline const std::string& Beagle::NN::MapT<T>::getName() const
{
	Beagle_StackTraceBeginM();
	return mName;
	Beagle_StackTraceEndM("const string& NN::MapT<T>::getName() const");
}


/*!
 *  \brief Sets the name of the map.
 *  \param inName Name of the map.
 */
template <class T>
inline void Beagle::NN::MapT<T>::setName(const std::string& inName)
{
	Beagle_StackTraceBeginM();
	mName = inName;
	Beagle_StackTraceEndM("void NN::MapT<T>::setName(const string& inName)");
}


#endif // Beagle_NN_MapT_hpp

