
/*!
 *  \file   beagle/NN/Module.hpp
 *  \brief  Definition of the type NN::Module.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_NN_Module_hpp
#define Beagle_NN_Module_hpp

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/Container.hpp"
#include "beagle/NN/NodeSet.hpp"

namespace Beagle
{
namespace NN
{


/*!
 *  \class Module beagle/NN/Module.hpp "beagle/NN/Module.hpp"
 *  \brief Class for keeping track of NN graph modules.
 *  \ingroup NNF
 *
 *  A module is defined by specifying the ids of its input, output, hidden
 *  and externally linked nodes.  Since there can be only one instance of a
 *  node in a graph, the smallest output node id can be used as the key for
 *  storing the modules of a graph in a hash table.
 */
class Module
{

public:

	explicit Module(unsigned int inKey=UINT_MAX);
	explicit Module(const Module& inModule1, const Module& inModule2);
	virtual ~Module()
	{ }

	virtual bool isEqual(const Module& inRightModule) const;
	virtual bool isLess(const Module& inRightModule) const;
	virtual bool isLessEqual(const Module& inRightModule) const;

	virtual void read(PACC::XML::ConstIterator inIter);
	virtual void write(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;
	virtual void write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const;

	void         setKey(unsigned int inKey);
	void         addInput(unsigned int inNodeId);
	void         addOutput(unsigned int inNodeId);
	void         addHidden(unsigned int inNodeId);

	unsigned int   getKey() const;
	const NodeSet& getInputSet() const;
	const NodeSet& getOutputSet() const;
	const NodeSet& getHiddenSet() const;

	void                 addExport(unsigned int inSourceId, unsigned int inTargetId);
	const MapT<NodeSet>& getExportMap() const;
	void                 setExportMap(const MapT<NodeSet>& inExportMap);

	bool         provides(const MapT<NodeSet>& inExportMap) const;
	unsigned int getSize(bool inIncludeInput) const;
	unsigned int getRandomNode(Beagle::Randomizer& ioRandomizer, bool inIncludeInput) const;

protected:

	unsigned int mKey;         //!< The key of this module, used for hashing, etc.

	NodeSet mInputSet;         //!< Node ids of module input nodes
	NodeSet mOutputSet;        //!< Node ids of module output nodes
	NodeSet mHiddenSet;        //!< Node ids of module hidden nodes

	MapT<NodeSet> mExportMap;  //!< Nodes having edges to external nodes

};

}
}


/*!
 * \brief Set the key for the module.
 */
inline void Beagle::NN::Module::setKey(unsigned int inKey)
{
	Beagle_StackTraceBeginM();
	mKey = inKey;
	Beagle_StackTraceEndM("void NN::Module::setKey(unsigned int inKey)");
}


/*!
 * \brief Add the given node id to the set of input node ids.
 */
inline void Beagle::NN::Module::addInput(unsigned int inNodeId)
{
	Beagle_StackTraceBeginM();
	mInputSet.insert(inNodeId);
	Beagle_StackTraceEndM("void NN::Module::addInput(unsigned int inNodeId)");
}


/*!
 * \brief Add the given node id to the set of output node ids.
 */
inline void Beagle::NN::Module::addOutput(unsigned int inNodeId)
{
	Beagle_StackTraceBeginM();
	mOutputSet.insert(inNodeId);
	Beagle_StackTraceEndM("void NN::Module::addOutput(unsigned int inNodeId)");
}


/*!
 * \brief Add the given node id to the set of hidden node ids.
 */
inline void Beagle::NN::Module::addHidden(unsigned int inNodeId)
{
	Beagle_StackTraceBeginM();
	mHiddenSet.insert(inNodeId);
	Beagle_StackTraceEndM("void NN::Module::addHidden(unsigned int inNodeId)");
}


/*!
 * \return The key of this module, used for hashing, etc.
 */
inline unsigned int Beagle::NN::Module::getKey() const
{
	Beagle_StackTraceBeginM();
	return mKey;
	Beagle_StackTraceEndM("unsigned int NN::Module::getKey() const");
}


/*!
 * \return The input set of the module.
 */
inline const Beagle::NN::NodeSet& Beagle::NN::Module::getInputSet() const
{
	Beagle_StackTraceBeginM();
	return mInputSet;
	Beagle_StackTraceEndM("const NN::NodeSet& NN::Module::getInputSet() const");
}


/*!
 * \return The output set of the module.
 */
inline const Beagle::NN::NodeSet& Beagle::NN::Module::getOutputSet() const
{
	Beagle_StackTraceBeginM();
	return mOutputSet;
	Beagle_StackTraceEndM("const NN::NodeSet& NN::Module::getOutputSet() const");
}


/*!
 * \return The hidden set of the module.
 */
inline const Beagle::NN::NodeSet& Beagle::NN::Module::getHiddenSet() const
{
	Beagle_StackTraceBeginM();
	return mHiddenSet;
	Beagle_StackTraceEndM("const NN::NodeSet& NN::Module::getHiddenSet() const");
}


/*!
 * \brief Add the given source and target node ids to the export map.
 */
inline void Beagle::NN::Module::addExport(unsigned int inSourceId, unsigned int inTargetId)

{
	Beagle_StackTraceBeginM();
	if (mHiddenSet.hasKey(inSourceId) || mOutputSet.hasKey(inSourceId)) {
		mExportMap.addData(inSourceId, NodeSet()).insert(inTargetId);
	} else {
		throw Beagle_RunTimeExceptionM("Cannot add unknown node "+uint2str(inSourceId)+" to export map of module "+uint2str(mKey));
	}
	Beagle_StackTraceEndM("void NN::Module::addExport(unsigned int inSourceId, unsigned int inTargetId)");
}


/*!
 * \return The export map of the module.
 */
inline const Beagle::NN::MapT<Beagle::NN::NodeSet>& Beagle::NN::Module::getExportMap() const
{
	Beagle_StackTraceBeginM();
	return mExportMap;
	Beagle_StackTraceEndM("const NN::MapT<NN::NodeSet>& NN::Module::getExportMap() const");
}


/*!
 * \brief Set the export map of the module.
 */
inline void Beagle::NN::Module::setExportMap(const MapT<NodeSet>& inExportMap)
{
	Beagle_StackTraceBeginM();
	mExportMap = inExportMap;
    mExportMap.setName("Export");
	Beagle_StackTraceEndM("void NN::Module::setExportMap(const MapT<NodeSet>& inExportMap)");
}


/*!
 * \param inIncludeInput Whether to include the input set in calculating size.
 * \return The size of the module in terms of the number of nodes.
 */
inline unsigned int Beagle::NN::Module::getSize(bool inIncludeInput) const
{
	Beagle_StackTraceBeginM();
	unsigned int lSize = mOutputSet.size() + mHiddenSet.size();
	if (inIncludeInput)
		lSize += mInputSet.size();
	return lSize;
	Beagle_StackTraceEndM("unsigned int NN::Module::getSize(bool inIncludeInput) const");
}


#endif // Beagle_NN_Module_hpp
