
/*!
 *  \file   beagle/NN/Node.hpp
 *  \brief  Definition of the type NN::Node.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_NN_Node_hpp
#define Beagle_NN_Node_hpp

#include <cmath>
#include <cfloat>

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/Container.hpp"
#include "beagle/NN/MapT.hpp"
#include "beagle/NN/Edge.hpp"
#include "beagle/Randomizer.hpp"


namespace Beagle
{
namespace NN
{


/*!
 *  \class Node beagle/NN/Node.hpp "beagle/NN/Node.hpp"
 *  \brief NN graph node class.
 *  \ingroup NNF
 *
 *  Each node stores a map of edges incident on it, keyed by source node id.
 */
class Node : public MapT<Edge>
{

public:

	//! Color values used in search algorithms
	enum Color {eWhite, eGray, eBlack};

	explicit Node(unsigned int inIdentifier=UINT_MAX, double inBias=0.0, double inSlope=1.0);
	virtual ~Node()
	{ }

	virtual bool isEqual(const Node& inRightObj) const;
	virtual bool isLess(const Node& inRightObj) const;

	unsigned int getIdentifier() const;
	unsigned int getModuleKey() const;
	double       getBias() const;
	double       getSlope() const;
	unsigned int getColor() const;
	double       getActivation(unsigned int inIndex) const;

	void         setModuleKey(unsigned int inModuleKey);
	void         setBias(double inBias);
	void         setSlope(double inSlope);
	void         setColor(unsigned int inColor);
	void         setActivation(double inActivation, unsigned int inIndex);

	bool         hasEdge(unsigned int inSourceId) const;
	unsigned int getNumberEdges() const;
	const Edge&  getEdge(unsigned int inSourceId) const;
	const Edge&  getRandomEdge(Beagle::Randomizer& ioRandomizer) const;
	Edge&  getEdge(unsigned int inSourceId);
	Edge&  getRandomEdge(Beagle::Randomizer& ioRandomizer);
	Edge&        addEdge(const Edge& inEdge);
	bool         delEdge(const Edge& inEdge);
	bool         delEdge(unsigned int inSourceId);

	void         activate(double inActivationSum, unsigned int inIndex);
	void         resetActivation();

protected:

	unsigned int mIdentifier;   //!< Identifier of the node
	unsigned int mModuleKey;    //!< The key of the module that the node belongs to
	double       mBias;         //!< Bias of the node
	double       mSlope;        //!< Slope of the node

	unsigned int mColor;           //!< Color used by search algorithms
	double       mActivation[2];   //!< Current and last activation of this node

	virtual void readAttributes(PACC::XML::ConstIterator inIter);
	virtual void writeAttributes(PACC::XML::Streamer& ioStreamer, bool inIndent=true, unsigned int inKey=UINT_MAX) const;

	double       sigmoid(double inX) const;

};

}
}


/*!
 *  \return The identifier of the node.
 */
inline unsigned int Beagle::NN::Node::getIdentifier() const
{
	Beagle_StackTraceBeginM();
	return mIdentifier;
	Beagle_StackTraceEndM("unsigned int NN::Node::getIdentifier() const");
}


/*!
 *  \return The key of the module to which the node belongs.
 */
inline unsigned int Beagle::NN::Node::getModuleKey() const
{
	Beagle_StackTraceBeginM();
	return mModuleKey;
	Beagle_StackTraceEndM("unsigned int NN::Node::getModuleKey() const");
}


/*!
 *  \return The bias of the node.
 */
inline double Beagle::NN::Node::getBias() const
{
	Beagle_StackTraceBeginM();
	return mBias;
	Beagle_StackTraceEndM("double NN::Node::getBias() const");
}


/*!
 *  \return The slope of the node.
 */
inline double Beagle::NN::Node::getSlope() const
{
	Beagle_StackTraceBeginM();
	return mSlope;
	Beagle_StackTraceEndM("double NN::Node::getSlope() const");
}


/*!
 *  \return The color of the node.
 */
inline unsigned int Beagle::NN::Node::getColor() const
{
	Beagle_StackTraceBeginM();
	return mColor;
	Beagle_StackTraceEndM("unsigned int NN::Node::getColor() const");
}


/*!
 *  \brief Get the activation of the node stored at index inIndex%2. 
 *  \param inIndex The activation count of the parent graph.
 *  \return The activation of the node.
 */
inline double Beagle::NN::Node::getActivation(unsigned int inIndex) const
{
	Beagle_StackTraceBeginM();
	return mActivation[inIndex%2];
	Beagle_StackTraceEndM("double NN::Node::getActivation(unsigned int inIndex) const");
}


/*!
 *  \param inModuleKey The key of the module to which the node belongs.
 */
inline void Beagle::NN::Node::setModuleKey(unsigned int inModuleKey)
{
	Beagle_StackTraceBeginM();
	mModuleKey = inModuleKey;
	Beagle_StackTraceEndM("void NN::Node::setModuleKey(unsigned int inModuleKey)");
}


/*!
 *  \param inBias The bias of the node.
 */
inline void Beagle::NN::Node::setBias(double inBias)
{
	Beagle_StackTraceBeginM();
	mBias = inBias;
	Beagle_StackTraceEndM("void NN::Node::setBias(double inBias)");
}


/*!
 *  \param inSlope The slope of the node.
 */
inline void Beagle::NN::Node::setSlope(double inSlope)
{
	Beagle_StackTraceBeginM();
	mSlope = inSlope;
	Beagle_StackTraceEndM("void NN::Node::setSlope(double inSlope)");
}


/*!
 *  \param inColor The color of the node.
 */
inline void Beagle::NN::Node::setColor(unsigned int inColor)
{
	Beagle_StackTraceBeginM();
	mColor = inColor;
	Beagle_StackTraceEndM("void NN::Node::setColor(unsigned int inColor)");
}


/*!
 *  \brief Store the activation in the activation array at inIndex%2.
 *  \param inActivation The activation of the node.
 *  \param inIndex The activation count of the parent graph.
 */
inline void Beagle::NN::Node::setActivation(double inActivation, unsigned int inIndex)
{
	Beagle_StackTraceBeginM();
	mActivation[inIndex%2] = inActivation;
	Beagle_StackTraceEndM("void NN::Node::setActivation(double inActivation, unsigned int inIndex)");
}


/*!
 *  \return True if an incoming edge specified by the source node id exists, false otherwise.
 */
inline bool Beagle::NN::Node::hasEdge(unsigned int inSourceId) const
{
	Beagle_StackTraceBeginM();
	return MapT<Edge>::hasKey(inSourceId);
	Beagle_StackTraceEndM("bool NN::Node::hasEdge(unsigned int inSourceId) const");
}


/*!
 *  \return The number of edges incident on this node.
 */
inline unsigned int Beagle::NN::Node::getNumberEdges() const
{
	Beagle_StackTraceBeginM();
	return size();
	Beagle_StackTraceEndM("unsigned int NN::Node::getNumberEdges() const");
}


/*!
 *  \return The reference to the incoming edge specified by the source node id.
 */
inline const Beagle::NN::Edge& Beagle::NN::Node::getEdge(unsigned int inSourceId) const
{
	Beagle_StackTraceBeginM();
	return MapT<Edge>::getData(inSourceId);
	Beagle_StackTraceEndM("const NN::Edge& NN::Node::getEdge(unsigned int inSourceId) const");
}


/*!
 *  \return The reference to a random incoming edge chosen with uniform probability.
 */
inline const Beagle::NN::Edge& Beagle::NN::Node::getRandomEdge(Beagle::Randomizer& ioRandomizer) const
{
	Beagle_StackTraceBeginM();
	return MapT<Edge>::getRandomData(ioRandomizer);
	Beagle_StackTraceEndM("const NN::Edge& NN::Node::getRandomEdge(Beagle::Randomizer& ioRandomizer) const");
}


/*!
 *  \return The reference to the incoming edge specified by the source node id.
 */
inline Beagle::NN::Edge& Beagle::NN::Node::getEdge(unsigned int inSourceId)
{
	Beagle_StackTraceBeginM();
	return MapT<Edge>::getData(inSourceId);
	Beagle_StackTraceEndM("NN::Edge& NN::Node::getEdge(unsigned int inSourceId)");
}


/*!
 *  \return The reference to a random incoming edge chosen with uniform probability.
 */
inline Beagle::NN::Edge& Beagle::NN::Node::getRandomEdge(Beagle::Randomizer& ioRandomizer)
{
	Beagle_StackTraceBeginM();
	return MapT<Edge>::getRandomData(ioRandomizer);
	Beagle_StackTraceEndM("NN::Edge& NN::Node::getRandomEdge(Beagle::Randomizer& ioRandomizer)");
}


/*!
 *  \brief Add an incoming edge to the node.
 *  \param inEdge The edge to be added.
 *  \return Reference to either the added edge or an existing edge with the same source node id.
 */
inline Beagle::NN::Edge& Beagle::NN::Node::addEdge(const Edge& inEdge)
{
	Beagle_StackTraceBeginM();
	return MapT<Edge>::addData(inEdge.getSourceIdentifier(), inEdge);
	Beagle_StackTraceEndM("NN::Edge& NN::Node::addEdge(const NN::Edge& inEdge)");
}


/*!
 *  \brief Delete an incoming edge from the node.
 *  \param inEdge The edge to be deleted.
 *  \return True if the edge was deleted, false otherwise.
 */
inline bool Beagle::NN::Node::delEdge(const Edge& inEdge)
{
	Beagle_StackTraceBeginM();
	return MapT<Edge>::delData(inEdge.getSourceIdentifier());
	Beagle_StackTraceEndM("bool NN::Node::delEdge(const NN::Edge& inEdge)");
}


/*!
 *  \brief Delete an incoming edge from the node.
 *  \param inSourceId The source identifier of the edge to be deleted.
 *  \return True if the edge was deleted, false otherwise.
 */
inline bool Beagle::NN::Node::delEdge(unsigned int inSourceId)
{
	Beagle_StackTraceBeginM();
	return MapT<Edge>::delData(inSourceId);
	Beagle_StackTraceEndM("bool NN::Node::delEdge(unsigned int inSourceId)");
}


/*!
 *  \brief Compute the sigmoid function using the bias and slope of the node.
 *  \param inX The point at which the sigmoid is to be calculated.
 *  \return The value of the sigmoid function at the given point.
 */
inline double Beagle::NN::Node::sigmoid(double inX) const
{
	Beagle_StackTraceBeginM();
	return 1/(1+exp(-mSlope*(inX+mBias)));
	Beagle_StackTraceEndM("double NN::Node::sigmoid(double inX) const");
}


/*!
 *  \brief Compute the activation of this node.
 *  \param inActivationSum The linear combination of input activations.
 *  \param inIndex The activation count of the parent graph.
 */
inline void Beagle::NN::Node::activate(double inActivationSum, unsigned int inIndex)
{
	Beagle_StackTraceBeginM();
	setActivation(sigmoid(inActivationSum), inIndex);
	Beagle_StackTraceEndM("void NN::Node::activate(double inActivationSum, unsigned int inIndex)");
}


/*!
 *  \brief Resets the node as not activated.
 */
inline void Beagle::NN::Node::resetActivation()
{
	Beagle_StackTraceBeginM();
	mActivation[0] = DBL_MAX;
	mActivation[1] = DBL_MAX;
	Beagle_StackTraceEndM("void NN::Node::resetActivation()");
}


#endif // Beagle_NN_Node_hpp
