
/*!
 *  \file   beagle/NN/NodeSet.hpp
 *  \brief  Definition of the type NN::NodeSet.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_NN_NodeSet_hpp
#define Beagle_NN_NodeSet_hpp

#include <set>

#if defined(BEAGLE_HAVE_STD_HASHSET) | defined(BEAGLE_HAVE_STDEXT_HASHSET)
#include <hash_set>
#endif // defined(BEAGLE_HAVE_STD_HASHSET) | defined(BEAGLE_HAVE_STDEXT_HASHSET)

#ifdef BEAGLE_HAVE_GNUCXX_HASHSET
#include <ext/hash_set>
#endif // BEAGLE_HAVE_GNUCXX_HASHSET

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/Container.hpp"
#include "beagle/NamedObject.hpp"

namespace Beagle
{
namespace NN
{


/*!
 *  \class NodeSet beagle/NN/NodeSet.hpp "beagle/NN/NodeSet.hpp"
 *  \brief Set of NN graph node ids.
 *  \ingroup NNF
 */
#if defined(BEAGLE_HAVE_STD_HASHSET)
class NodeSet : public NamedObject, public std::hash_set<unsigned int>
#elif defined(BEAGLE_HAVE_GNUCXX_HASHSET)
class NodeSet : public NamedObject, public __gnu_cxx::hash_set<unsigned int>
#elif defined(BEAGLE_HAVE_STDEXT_HASHSET)
class NodeSet : public NamedObject, public stdext::hash_set<unsigned int>
#else // No hash_map found
class NodeSet : public NamedObject, public std::set<unsigned int>
#endif
{

public:

	//! NN::NodeSet allocator type.
	typedef AllocatorT<NodeSet, NamedObject::Alloc>
	Alloc;
	//! NN::NodeSet handle type.
	typedef PointerT<NodeSet, NamedObject::Handle>
	Handle;
	//! NN::NodeSet bag type.
	typedef ContainerT<NodeSet, NamedObject::Bag>
	Bag;

	explicit NodeSet(const std::string& inName="NodeSet");
	virtual ~NodeSet()
	{ }

	virtual bool isEqual(const Object& inRightObj) const;
	virtual bool isLess(const Object& inRightObj) const;
	virtual bool isLessEqual(const NodeSet& inRightNodeSet) const;
	virtual void read(PACC::XML::ConstIterator inIter);
	virtual void write(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;
	virtual void write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const;

	bool         hasKey(unsigned int inKey) const;
	unsigned int getMin() const;

	void         merge(const NodeSet& inNodeSet);

};

}
}


/*!
 *  \brief Check if the given the set has the given key.
 *  \param inKey The key to be checked.
 *  \return True if the key is in the set; false otherwise.
 */
inline bool Beagle::NN::NodeSet::hasKey(unsigned int inKey) const
{
	Beagle_StackTraceBeginM();
	return (find(inKey) != end());
	Beagle_StackTraceEndM("bool NN::NodeSet::hasKey(unsigned int inKey) const");
}


#endif // Beagle_NN_NodeSet_hpp
