
/*!
 *  \file   beagle/NN/src/BreadthFirstSearch.cpp
 *  \brief  Implementation of the type NN::BreadthFirstSearch.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a breadth-first search object.
 */
NN::BreadthFirstSearch::BreadthFirstSearch() :
		mContinue(true)
{ }


/*!
 *  \brief Perform a breadth-first search of the given graph.
 *  \param ioGraph The graph to search in.
 *  \param inTargetSet The set of target nodes from which search begins.
 *
 *  The search proceeds backwards from the target nodes towards inputs.
 */
void NN::BreadthFirstSearch::search(Graph& ioGraph, const NodeSet& inTargetSet)
{
	Beagle_StackTraceBeginM();

	// Initialize the color of all nodes in the graph to white.
	for (Graph::iterator lIter = ioGraph.begin(); lIter != ioGraph.end(); ++lIter) {
		Node& lNode = lIter->second;
		lNode.setColor(Node::eWhite);
		eventInitialize(lNode);
	}

	// Queue for managing the set of gray nodes.
	std::queue<unsigned int> lQ;

	// Verify the target nodes and add them to the queue.
	for (NodeSet::const_iterator lIter = inTargetSet.begin(); lIter != inTargetSet.end(); ++lIter) {
		unsigned int lTargetId = *lIter;

		// First make sure the target node is in the graph.
		if (!ioGraph.hasNode(lTargetId)) {
			std::ostringstream lOSS;
			lOSS << "Invalid target node [" << lTargetId << "] for breadth-first search!";
			throw Beagle_RunTimeExceptionM(lOSS.str());
		}

		// Discover the target node and add it to the queue.
		Node& lTarget = ioGraph.getNode(lTargetId);
		lTarget.setColor(Node::eGray);
		lQ.push(lTargetId);
		eventDiscover(lTarget);
	}

	while (lQ.size() > 0 && mContinue) {
		// Examine the node at the front of the queue and remove it.
		Node& lCurrent = ioGraph.getNode(lQ.front());
		lQ.pop();
		eventBegin(lCurrent);

		// Examine all incoming edges of the current node.
		for (Node::const_iterator lNodeIter = lCurrent.begin(); lNodeIter != lCurrent.end() && mContinue; ++lNodeIter) {
			unsigned int lAdjacentId = lNodeIter->second.getSourceIdentifier();

			// Make sure this source node exists in the graph.
			if (!ioGraph.hasNode(lAdjacentId)) {
				std::ostringstream lOSS;
				lOSS << "Invalid node [" << lAdjacentId << "] encountered during breadth-first search!";
				throw Beagle_RunTimeExceptionM(lOSS.str());
			}

			Node& lAdjacent = ioGraph.getNode(lAdjacentId);

			// If current edge is a tree edge...
			if (lAdjacent.getColor() == Node::eWhite) {
				lAdjacent.setColor(Node::eGray);
				lQ.push(lAdjacentId);
				eventDiscover(lAdjacent);
			}
		}

		// Finish examining the current node.
		lCurrent.setColor(Node::eBlack);
		eventFinish(lCurrent);
	}

	Beagle_StackTraceEndM("void NN::BreadthFirstSearch::search(NN::Graph& ioGraph, NodeSet& inTargetSet)");
}

