
/*!
 *  \file   beagle/NN/src/ChangeFitnessIndexOp.cpp
 *  \brief  Source code of class ChangeFitnessIndexOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a ChangeFitnessIndexOp.
 *  \param inName Name of the operator.
 */
NN::ChangeFitnessIndexOp::ChangeFitnessIndexOp(std::string inName) :
		BreederOp(inName)
{ }


/*!
 *  \brief Register the parameters of the operator.
 *  \param ioSystem System of the evolution.
 */
void NN::ChangeFitnessIndexOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Beagle::BreederOp::registerParams(ioSystem);
	{
		std::ostringstream lOSS;
		lOSS << "Probability that the fitness index will be changed.";
		Register::Description lDescription(
		    "Fitness index changing prob.",
		    "Float",
		    "0.0",
		    lOSS.str()
		);
		mChangeIndexProba = castHandleT<Float>(
		                        ioSystem.getRegister().insertEntry("nn.changeindex.proba", new Float(0.0f), lDescription));
	}
	Beagle_StackTraceEndM("void NN::ChangeFitnessIndexOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Change the fitness index in a breeding pool and return a bred individual.
 *  \param inBreedingPool Breeding pool to use for the breeding operation.
 *  \param inChild Node handle associated to child node in the breeder tree.
 *  \param ioContext Evolutionary context of the breeding operation.
 *  \return The bred individual.
 */
Beagle::Individual::Handle NN::ChangeFitnessIndexOp::breed(Beagle::Individual::Bag& inBreedingPool,
        BreederNode::Handle inChild,
        Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	changeIndex(*inBreedingPool[0], ioContext);

	Beagle_NonNullPointerAssertM(inChild);
	Beagle_NonNullPointerAssertM(inChild->getBreederOp());

	Beagle::Individual::Handle lBredIndividual =
	    inChild->getBreederOp()->breed(inBreedingPool, inChild->getFirstChild(), ioContext);

	// Change index back to 0 before we return.
	FitnessIndexed::setIndex(0);

	return lBredIndividual;

	Beagle_StackTraceEndM("Individual::Handle NN::ChangeFitnessIndexOp::breed(Beagle::Individual::Bag& inBreedingPool, BreederNode::Handle inChild, Beagle::Context& ioContext)");
}


/*!
 *  \return Return selection probability of breeder operator.
 *  \param inChild Child node in the breeder tree.
 */
float NN::ChangeFitnessIndexOp::getBreedingProba(BreederNode::Handle inChild)
{
	Beagle_StackTraceBeginM();
	Beagle_NonNullPointerAssertM(inChild);
	Beagle_NonNullPointerAssertM(inChild->getBreederOp());
	return inChild->getBreederOp()->getBreedingProba(inChild->getFirstChild());
	Beagle_StackTraceEndM("float NN::ChangeFitnessIndexOp::getBreedingProba(BreederNode::Handle inChild)");
}


/*!
 *  \brief Changes the fitness index of individuals in ioDeme.
 *  \param ioDeme The deme to operate on.
 *  \param ioContext Evolutionary context.
 */
void NN::ChangeFitnessIndexOp::operate(Beagle::Deme& ioDeme, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	changeIndex(*ioDeme[0], ioContext);

	Beagle_StackTraceEndM("void NN::ChangeFitnessIndexOp::operate(Beagle::Deme& ioDeme, Beagle::Context& ioContext)");
}


/*!
 *  \brief Changes the fitness index of individuals.
 *  \param ioIndividual The prototype individual from which to get fitness.
 *  \param ioContext Evolutionary context.
 */
bool NN::ChangeFitnessIndexOp::changeIndex(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	// Extract the specified index from the operator name.
	unsigned int lIndex;
	std::string::size_type lLocation = getName().find('+');
	if (lLocation != std::string::npos) {
		std::string lIndexText = getName().substr(lLocation);
		lIndex = str2uint(lIndexText);
	} else if (ioContext.getDemeIndex() == 0 &&
	           ioContext.getSystem().getRandomizer().rollUniform(0.0, 1.0) < mChangeIndexProba->getWrappedValue()) {
		FitnessIndexed::Handle lFitness = castHandleT<FitnessIndexed>(ioIndividual.getFitness());
		lIndex = lFitness->size() - 1;
	} else {
		lIndex = 0;
	}
	FitnessIndexed::setIndex(lIndex);

	Beagle_LogTraceM(
	    ioContext.getSystem().getLogger(),
	    "changefitnessindex", "Beagle::ChangeFitnessIndexOp",
	    std::string("Set fitness index of individuals to ")+uint2str(lIndex)
	);

	return true;

	Beagle_StackTraceEndM("bool NN::ChangeFitnessIndexOp::changeIndex(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}

