
/*!
 *  \file   beagle/NN/src/CrossoverModuleOp.cpp
 *  \brief  Source code of class NN::CrossoverModuleOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <cfloat>
#include <cmath>
#include <float.h>
#include <algorithm>
#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a module based crossover operator.
 *  \param inMatingPbName Mating probability parameter name.
 *  \param inName Name of the operator.
 */
NN::CrossoverModuleOp::CrossoverModuleOp(std::string inMatingPbName,
                                         std::string inName) :
		CrossoverOp(inMatingPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the prediction based crossover operator.
 *  \param ioSystem System of the evolution.
 */
void NN::CrossoverModuleOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		Register::Description lDescription(
		    "Individual module crossover prob.",
		    "Float",
		    "0.0",
		    "Module based crossover probability of a single individual."
		);
		mMatingProba = castHandleT<Float>(
		                   ioSystem.getRegister().insertEntry(mMatingProbaName, new Float(0.0f), lDescription));
	}
    Beagle::CrossoverOp::registerParams(ioSystem);
	Beagle_StackTraceEndM("void NN::CrossoverModuleOp::registerParams(System& ioSystem)");
}


/*!
 *  \brief Mate two NN individuals using module crossover.
 *  \param ioIndiv1   First individual to mate.
 *  \param ioContext1 Evolutionary context of the first individual.
 *  \param ioIndiv2   Second individual to mate.
 *  \param ioContext2 Evolutionary context of the second individual.
 *  \return True if the individuals are effectively mated, false if not.
 */
bool NN::CrossoverModuleOp::mate(Beagle::Individual& ioIndiv1, Beagle::Context& ioContext1,
                                 Beagle::Individual& ioIndiv2, Beagle::Context& ioContext2)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndiv1 = castObjectT<NN::Individual&>(ioIndiv1);
	NN::Individual& lIndiv2 = castObjectT<NN::Individual&>(ioIndiv2);

	Beagle_LogDebugM(
	    ioContext1.getSystem().getLogger(),
	    "crossover", "Beagle::NN::CrossoverModuleOp",
	    std::string("The first individual mated is (before module crossover): ")+
	    lIndiv1.serialize()
	);
	Beagle_LogDebugM(
	    ioContext1.getSystem().getLogger(),
	    "crossover", "Beagle::NN::CrossoverModuleOp",
	    std::string("The second individual mated is (before module crossover): ")+
	    lIndiv2.serialize()
	);

	if (lIndiv1.size() > 1 || lIndiv2.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::CrossoverModuleOp not implemented for individuals of size > 1");

	Graph& lGraph1 = *lIndiv1[0];
	Graph& lGraph2 = *lIndiv2[0];

	Graph lGCopy2(lGraph2);

	bool lMated1 = crossover(lGraph1, lGraph2, ioContext1);
	bool lMated2 = crossover(lGCopy2, lGraph1, ioContext1);

	if (lMated1 || lMated2) {
		// Mating successful, log messages and return
		Beagle_LogDebugM(
		    ioContext1.getSystem().getLogger(),
		    "crossover", "Beagle::NN::CrossoverModuleOp",
		    std::string("The first individual mated is (after module crossover): ")+
		    lIndiv1.serialize()
		);
		Beagle_LogDebugM(
		    ioContext1.getSystem().getLogger(),
		    "crossover", "Beagle::NN::CrossoverModuleOp",
		    std::string("The second individual mated is (after module crossover): ")+
		    lIndiv2.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext1.getSystem().getLogger(),
		    "crossover", "Beagle::NN::CrossoverModuleOp",
		    std::string("Crossover attempt failed; ")+
		    std::string("resuming from crossover without modifying the individuals")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::CrossoverModuleOp::mate(Individual& ioIndiv1, Context& ioContext1, Individual& ioIndiv2, Context& ioContext2)");
}


/*!
 *  \brief Crossover modules from a one graph to another
 *  \param inGraph1 The graph from which modules are crossed over
 *  \param ioGraph2 The graph to which modules are crossed over
 *  \return True if at least one module was crossed over, false if not
 */
bool NN::CrossoverModuleOp::crossover(const Graph& inGraph1, Graph& ioGraph2, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	bool lCrossed = false;

	// Insertion of a graph1 module into graph2 is possible only if the graph1 module
	// can satisfy all node references by external edges in the graph2 modules that
	// will be replaced as a result of the insertion.
	const MapT<Module>& lModules1 = inGraph1.getModules();
	for (MapT<Module>::const_iterator lModIter1 = lModules1.begin(); lModIter1 != lModules1.end(); ++lModIter1) {
		const Module& lModule1 = lModIter1->second;

		// Check if the graph1 module has external edges, evidence that it is being used
		// in graph1, and is therefore potentially useful.
		if (lModule1.getExportMap().size() == 0)
			continue;

		const NodeSet& lOutSet1 = lModule1.getOutputSet();
		const NodeSet& lHidSet1 = lModule1.getHiddenSet();

		// Check if the nodes in the graph2 modules being replaced is a subset of the
		// nodes in the graph1 module.
		NodeSet lModSet1(lOutSet1);
		lModSet1.merge(lHidSet1);
		NodeSet lModSet2;
		for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
			if (ioGraph2.hasNode(*lOutIter1)) {
				unsigned int lModKey2 = ioGraph2.getNode(*lOutIter1).getModuleKey();
				lModSet2.merge(ioGraph2.getModule(lModKey2).getOutputSet());
				lModSet2.merge(ioGraph2.getModule(lModKey2).getHiddenSet());
			}
		}
		if (!lModSet2.isLessEqual(lModSet1))
			continue;

		//     // Check if the edges in the graph2 modules being replaced is a subset of the
		//     // edges in the graph1 module.
		//     bool lEdgesSub2 = true;
		//     for (NodeSet::const_iterator lModIter2 = lModSet2.begin(); lEdgesSub2 && lModIter2 != lModSet2.end(); ++lModIter2) {
		//       lEdgesSub2 = ioGraph2.getNode(*lModIter2).NN::MapT<Edge>::isLessEqual(inGraph1.getNode(*lModIter2));
		//     }
		//     if (!lEdgesSub2) continue;

		MapT<NodeSet> lExpMap2;
		for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
			if (ioGraph2.hasNode(*lOutIter1)) {
				unsigned int lModKey2 = ioGraph2.getNode(*lOutIter1).getModuleKey();
				const MapT<NodeSet>& lModExp2 = ioGraph2.getModule(lModKey2).getExportMap();
				for (MapT<NodeSet>::const_iterator lExpIter2 = lModExp2.begin(); lExpIter2 != lModExp2.end(); ++lExpIter2) {
					lExpMap2.addData(lExpIter2->first, lExpIter2->second);
				}
			}
		}

		// Check if modules in both graphs have the same export maps.
		const MapT<NodeSet>& lExpMap1 = lModule1.getExportMap();
		bool lSameExport = lExpMap1.isEqual(lExpMap2);
		for (MapT<NodeSet>::const_iterator lExpIter1 = lExpMap1.begin(); lSameExport && lExpIter1 != lExpMap1.end(); ++lExpIter1) {
			lSameExport = lExpIter1->second.isEqual(lExpMap2.getData(lExpIter1->first));
		}
		if (!lSameExport)
			continue;

		if (lModule1.provides(lExpMap2)) {
			std::ostringstream lOSS;
			lOSS << "Module [ " << lModule1.getKey() << " ] replaced modules [ ";

			// First remove the graph2 modules being replaced.
			for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
				if (ioGraph2.hasNode(*lOutIter1)) {
					unsigned int lModKey2 = ioGraph2.getNode(*lOutIter1).getModuleKey();
					const Module& lModule2 = ioGraph2.getModule(lModKey2);
					const NodeSet& lOutSet2 = lModule2.getOutputSet();
					const NodeSet& lHidSet2 = lModule2.getHiddenSet();
					for (NodeSet::const_iterator lOutIter2 = lOutSet2.begin(); lOutIter2 != lOutSet2.end(); ++lOutIter2) {
						ioGraph2.delNode(*lOutIter2);
					}
					for (NodeSet::const_iterator lHidIter2 = lHidSet2.begin(); lHidIter2 != lHidSet2.end(); ++lHidIter2) {
						ioGraph2.delNode(*lHidIter2);
					}
					ioGraph2.delModule(lModule2);

					lOSS << lModKey2 << " ";
				}
			}
			lOSS << "]";

			// Now add the module from graph1 to graph2, and replace the export map
			// of the module with the merged export map of the graph2 modules.
			for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
				ioGraph2.addNode(inGraph1.getNode(*lOutIter1));
			}
			for (NodeSet::const_iterator lHidIter1 = lHidSet1.begin(); lHidIter1 != lHidSet1.end(); ++lHidIter1) {
				ioGraph2.addNode(inGraph1.getNode(*lHidIter1));
			}
			ioGraph2.addModule(lModule1).setExportMap(lExpMap2);

			lCrossed = true;

			// Log message of module crossover
			Beagle_LogTraceM(
			    ioContext.getSystem().getLogger(),
			    "crossover", "Beagle::NN::CrossoverModuleOp",
			    lOSS.str()
			);
		}
	}

	if (lCrossed) {
		// Log message of completion of module crossover
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "crossover", "Beagle::NN::CrossoverModuleOp",
		    "Module crossover completed"
		);

		return true;
	} else {
		return false;
	}

	Beagle_StackTraceEndM("bool NN::CrossoverModuleOp::crossover(const Graph& inGraph1, Graph& ioGraph2, Beagle::Context& ioContext)");
}

