
/*!
 *  \file   beagle/NN/src/Edge.cpp
 *  \brief  Implementation of the type NN::Edge.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct an NN graph edge.
 *  \param inSourceId Identifier of the source node
 *  \param inWeight Weight of the edge
 */
NN::Edge::Edge(unsigned int inSourceId, double inWeight, bool inRecurrent) :
		mSourceId(inSourceId),
		mWeight(inWeight),
		mRecurrent(inRecurrent)
{ }


/*!
 *  \brief Compare equality of two NN graph edges.
 *  \param inRightEdge Right edge to be compared to.
 *  \return True if the edges have the same source node id, false otherwise.
 */
bool NN::Edge::isEqual(const Edge& inRightEdge) const
{
	Beagle_StackTraceBeginM();
	return (mSourceId == inRightEdge.mSourceId);
	Beagle_StackTraceEndM("bool NN::Edge::isEqual(const Edge& inRightEdge) const");
}


/*!
 *  \brief Check if this edge is less than the other.
 *  \param inRightEdge Right edge to be compared to.
 *  \return True if the source node id of this edge is less, false otherwise.
 */
bool NN::Edge::isLess(const Edge& inRightEdge) const
{
	Beagle_StackTraceBeginM();
	return (mSourceId < inRightEdge.mSourceId);
	Beagle_StackTraceEndM("bool NN::Edge::isLess(const Edge& inRightEdge) const");
}


/*!
 *  \brief Read an NN graph edge from an XML subtree.
 *  \param inIter XML iterator to use to read the edge from.
 *  \throw Beagle::IOException If reading from the stream failed.
 */
void NN::Edge::read(PACC::XML::ConstIterator inIter)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!="Edge"))
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Edge> expected!");

	// Get attributes and set member variables.
	std::string lSourceIdText = inIter->getAttribute("key");
	if(lSourceIdText.empty()==false)
		mSourceId = str2uint(lSourceIdText);
	std::string lWeightText = inIter->getAttribute("weight");
	if(lWeightText.empty()==false)
		mWeight = str2dbl(lWeightText);
	std::string lRecurrentText = inIter->getAttribute("recur");
	if(lRecurrentText.empty()==false) {
		std::istringstream lISS(lRecurrentText);
		lISS >> mRecurrent;
	} else {
		mRecurrent = false;
	}

	Beagle_StackTraceEndM("void NN::Edge::read(PACC::XML::ConstIterator inIter)");
}


/*!
 *  \brief Write a NN graph edge into an XML streamer.
 *  \param ioStreamer XML streamer to write the edge into.
 *  \param inIndent Whether XML output should be indented.
 *  \param inKey Key attribute if this edge is used as an entry in a map.
 */
void NN::Edge::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const
{
	Beagle_StackTraceBeginM();

	ioStreamer.openTag("Edge", inIndent);
	if (inKey < UINT_MAX)
		ioStreamer.insertAttribute("key", uint2str(inKey));
	else
		ioStreamer.insertAttribute("key", uint2str(mSourceId));
	ioStreamer.insertAttribute("weight", dbl2str(mWeight));
	if (mRecurrent) {
		std::ostringstream lOSS;
		lOSS << mRecurrent << std::flush;
		ioStreamer.insertAttribute("recur", lOSS.str());
	}
	ioStreamer.closeTag();

	Beagle_StackTraceEndM("void NN::Edge::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const");
}


