
/*!
 *  \file   beagle/NN/src/EvaluationOp.cpp
 *  \brief  Source code of class NN::EvaluationOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a evaluation operator.
 *  \param inName Name of the operator.
 */
NN::EvaluationOp::EvaluationOp(const std::string& inName) :
		Beagle::EvaluationOp(inName)
{ }


/*!
 *  \brief Register the parameters of the evaluation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::EvaluationOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Beagle::EvaluationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of modules in the network",
		    "UInt",
		    "4",
		    "Number of modules in the network."
		);
		mNumModules = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.number", new UInt(4), lDescription));
	}
	{
		Register::Description lDescription(
		    "The number of different module instances in the genotype",
		    "UInt",
		    "1",
		    "The genotype modules instances are copied to produce the other modules in the phenotype."
		);
		mNumModInst = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.numinst", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of rows into which the modules are arranged",
		    "UInt",
		    "1",
		    "Number of rows into which the modules are arranged."
		);
		mNumModRows = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.numrow", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of input nodes per module",
		    "UInt",
		    "2",
		    "Number of input nodes per module."
		);
		mNumModInputs = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.numin", new UInt(2), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of output nodes per module",
		    "UInt",
		    "1",
		    "Number of output nodes per module."
		);
		mNumModOutputs = castHandleT<UInt>(
		                  ioSystem.getRegister().insertEntry("nn.module.numout", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of helper nodes per module",
		    "UInt",
		    "2",
		    "Number of helper nodes per module."
		);
		mNumModHelpers = castHandleT<UInt>(
		                  ioSystem.getRegister().insertEntry("nn.module.numhelp", new UInt(2), lDescription));
	}
	Beagle_StackTraceEndM("void NN::EvaluationOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Evaluate the fitness of the given NN individual.
 *  \param inIndividual Current individual to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness value of the NN individual.
 */
Fitness::Handle NN::EvaluationOp::evaluate(Beagle::Individual& inIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(inIndividual);
	GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));

	if (inIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::EvaluationOp not implemented for individuals of size > 1");


    unsigned int lNumModules = mNumModules->getWrappedValue();
    unsigned int lNumModInst = mNumModInst->getWrappedValue();

    // Construct phenotype if different from genotype.
    NN::Graph::Handle lPhenotype;
    if (lNumModules > lNumModInst) {
        const Factory& lFactory = ioContext.getSystem().getFactory();
		std::string lGenotypeType = lIndividual[0]->getType();
		Genotype::Alloc::Handle lGenotypeAlloc =
			castHandleT<Genotype::Alloc>(lFactory.getAllocator(lGenotypeType));
		lPhenotype = castHandleT<NN::Graph>(lGenotypeAlloc->allocate());
		lPhenotype->copy(*inIndividual[0], ioContext.getSystem());

        Module& lMain = lPhenotype->getMain();

        unsigned int lNumModInputs = mNumModInputs->getWrappedValue();
        unsigned int lNumModOutputs = mNumModOutputs->getWrappedValue();
        unsigned int lNumModHelpers = mNumModHelpers->getWrappedValue();
        unsigned int lNumModRows = mNumModRows->getWrappedValue();
        unsigned int lNumModCols = lNumModules/lNumModRows;
        for (unsigned int lRowi = 0; lRowi < lNumModRows; lRowi++) {
            for (unsigned int lColi = 0; lColi < lNumModCols; lColi++) {
                unsigned int i = lRowi*lNumModCols + lColi;
                if (i < lNumModInst) continue;
                unsigned int lInstancei = i%lNumModInst;

                // Create module input nodes.
                for (unsigned int j = 0; j < lNumModInputs; j++) {
                    unsigned int lInstanceId = lGraphState->getFirstInput() + j*lNumModules + lInstancei;
                    const NN::Node& lInstance = lPhenotype->getNode(lInstanceId);
                    unsigned int lNodeId = lInstanceId + i - lInstancei;
                    NN::Node& lNode = lPhenotype->addNode(NN::Node(lNodeId, lInstance.getBias(), lInstance.getSlope()));
                    for (NN::Node::const_iterator lNodeIter = lInstance.begin(); lNodeIter != lInstance.end(); ++lNodeIter) {
                        const NN::Edge& lEdge = lNodeIter->second;
                        unsigned int lSourceId = lEdge.getSourceIdentifier() + i - lInstancei;
                        lNode.addEdge(NN::Edge(lSourceId, lEdge.getWeight(), lEdge.isRecurrent()));
                    }
                    lMain.addInput(lNodeId);
                }

                // Create module hidden nodes.
                for (unsigned int j = 0; j < lNumModHelpers; j++) {
                    unsigned int lInstanceId = lGraphState->getFirstHelper() + j*lNumModules + lInstancei;
                    const NN::Node& lInstance = lPhenotype->getNode(lInstanceId);
                    unsigned int lNodeId = lInstanceId + i - lInstancei;
                    NN::Node& lNode = lPhenotype->hasNode(lNodeId) ? lPhenotype->getNode(lNodeId) : lPhenotype->addNode(NN::Node(lNodeId));
                    lNode.setBias(lInstance.getBias());
                    lNode.setSlope(lInstance.getSlope());
                    for (NN::Node::const_iterator lNodeIter = lInstance.begin(); lNodeIter != lInstance.end(); ++lNodeIter) {
                        const NN::Edge& lEdge = lNodeIter->second;
                        if (lEdge.getSourceIdentifier() < lGraphState->getFirstInput() + lNumModInputs*lNumModules) {
                            unsigned int lCount = lEdge.getSourceIdentifier()/lNumModules;
                            unsigned int lRowInst = (lEdge.getSourceIdentifier() - lCount*lNumModules)/lNumModCols;
                            unsigned int lColInst = (lEdge.getSourceIdentifier() - lCount*lNumModules)%lNumModCols;
                            unsigned int lRowNode = (lRowInst + lRowi)%lNumModRows;
                            unsigned int lColNode = (lColInst + lColi)%lNumModCols;
                            unsigned int lSourceId = (lRowNode*lNumModCols + lColNode) + lCount*lNumModules;
                            lNode.addEdge(NN::Edge(lSourceId, lEdge.getWeight(), lEdge.isRecurrent()));
                        }
                        else if (lEdge.getSourceIdentifier() > lGraphState->getLastInput()) {
                            unsigned int lSourceId = lEdge.getSourceIdentifier() + i - lInstancei;
                            lNode.addEdge(NN::Edge(lSourceId, lEdge.getWeight(), lEdge.isRecurrent()));
                        }
                    }
                    lMain.addHidden(lNodeId);
                }

                // Create module output nodes.
                for (unsigned int j = 0; j < lNumModOutputs; j++) {
                    unsigned int lInstanceId = lGraphState->getFirstOutput() + j*lNumModules + lInstancei;
                    const NN::Node& lInstance = lPhenotype->getNode(lInstanceId);
                    unsigned int lNodeId = lInstanceId + i - lInstancei;
                    NN::Node& lNode = lPhenotype->hasNode(lNodeId) ? lPhenotype->getNode(lNodeId) : lPhenotype->addNode(NN::Node(lNodeId, lInstance.getBias(), lInstance.getSlope()));
                    for (NN::Node::const_iterator lNodeIter = lInstance.begin(); lNodeIter != lInstance.end(); ++lNodeIter) {
                        const NN::Edge& lEdge = lNodeIter->second;
                        unsigned int lSourceId = lEdge.getSourceIdentifier() + i - lInstancei;
                        lNode.addEdge(NN::Edge(lSourceId, lEdge.getWeight(), lEdge.isRecurrent()));
                    }
                    lMain.addOutput(lNodeId);
                }
            }
        }
    }

	NN::Graph& lGraph = (lNumModules > lNumModInst) ? *lPhenotype : *lIndividual[0];
	lGraph.resetActivation();

	return evaluate(lGraph, ioContext);

	Beagle_StackTraceEndM("Fitness::Handle NN::EvaluationOp::evaluate(Beagle::Individual& inIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Activate the nodes of the graph for the given input activations.
 *  \param ioGraph The graph to set the activations in.
 *  \param inActivations Map containing the activations of input nodes.
 *  \param ioContext Evolutionary context.
 *
 *  If the graph is being activated for the first time, it is activated enough
 *  number of times to propagate the inputs all the way to the outputs.
 *  Otherwise, the graph is activated only once.
 */
void NN::EvaluationOp::activate(NN::Graph& ioGraph, const NN::Graph::Activations& inActivations, Context& ioContext)
{
	Beagle_StackTraceBeginM();

	if (ioGraph.getActivationCount() > 0) {
		// If this is not the first time activating the graph, activate it only once.
		ioGraph.activate(inActivations);
	} else {
		// In order to make sure all nodes are activated, we use the number of nodes
		// in the graph as an estimate of the number of times to activate.
		unsigned int lNumNodes = ioGraph.getNumberNodes();
		for (unsigned int i = 0; i < lNumNodes; i++) {
			ioGraph.activate(inActivations);
		}
	}

	Beagle_StackTraceEndM("NN::EvaluationOp::activate(NN::Graph& ioGraph, const NN::Graph::Activations& inActivations, Context& ioContext)");
}
