
/*!
 *  \file   beagle/NN/src/FitnessIndexed.cpp
 *  \brief  Source code of class FitnessIndexed.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <cmath>
#include <cfloat>
#include <algorithm>

using namespace Beagle;


unsigned int NN::FitnessIndexed::smIndex = 0;


/*!
 *  \brief Default constructor for an indexed fitness object.
 */
NN::FitnessIndexed::FitnessIndexed() :
		Fitness(false),
		std::vector<float>()
{ }


/*!
 *  \brief Construct an indexed fitness object.
 *  \param inSize Size of the indexed fitness object.
 *  \param inValue Default initial value of the indexed fitness measures.
 */
NN::FitnessIndexed::FitnessIndexed(unsigned int inSize, float inValue) :
		Fitness(true),
		std::vector<float>(inSize, inValue)
{ }


/*!
 *  \brief Construct a valid indexed fitness object.
 *  \param inFitness Value of the fitness.
 */
NN::FitnessIndexed::FitnessIndexed(const std::vector<float>& inFitness) :
		Fitness(true),
		std::vector<float>(inFitness)
{ }


/*!
 *  \brief Copy fitness into current.
 *  \param inOriginal Fitness to copy.
 *  \param ioSystem Evolutionary system to use for making the copy.
 */
void NN::FitnessIndexed::copy(const Member& inOriginal, System& ioSystem)
{
	Beagle_StackTraceBeginM();
	const FitnessIndexed& lOriginal = castObjectT<const FitnessIndexed&>(inOriginal);
	(*this) = lOriginal;
	Beagle_StackTraceEndM("void NN::FitnessIndexed::copy(const Member&,System&)");
}


/*!
 *  \brief  Get distance between two indexed fitness measures.
 *  \param  inRightFitness Right fitness to compare.
 *  \return Absolute difference value between the indexed measures.
 */
float NN::FitnessIndexed::getDistance(const Fitness& inRightFitness) const
{
	Beagle_StackTraceBeginM();
	const NN::FitnessIndexed& lRightFitness = castObjectT<const NN::FitnessIndexed&>(inRightFitness);
	if((isValid()==false) || (lRightFitness.isValid()==false)) {
		std::ostringstream lOSS;
		lOSS << "Can't return the distance between two indexed fitness values as one or both ";
		lOSS << "are invalid!";
		throw Beagle_RunTimeExceptionM(lOSS.str());
	}
	return std::fabs(getValue() - lRightFitness.getValue());
	Beagle_StackTraceEndM("float NN::FitnessIndexed::getDistance(const Fitness& inRightFitness) const");
}


/*!
 *  \brief  Get the objective value of the indexed fitness measure.
 *  \param  inObjIndex Objective index to get value.
 *  \return Indexed fitness objective value.
 */
float NN::FitnessIndexed::getObjective(unsigned int inObjIndex) const
{
	Beagle_StackTraceBeginM();
	if(isValid() == false) {
		throw Beagle_RunTimeExceptionM("Can't return an objective value of an invalid fitness!");
	}
	if(inObjIndex >= size()) {
		std::ostringstream lOSS;
		lOSS << "Objective index given (" << inObjIndex;
		lOSS << ") exceed the number of objectives of the fitness (";
		lOSS << size() << ")!";
		throw Beagle_RunTimeExceptionM(lOSS.str());
	}
	return getValue(inObjIndex);
	Beagle_StackTraceEndM("float NN::FitnessIndexed::getObjective(unsigned int inObjIndex) const");
}


/*!
 *  \brief Get exact type of the fitness.
 *  \return Exact type of the fitness.
 */
const std::string& NN::FitnessIndexed::getType() const
{
	Beagle_StackTraceBeginM();
	const static std::string lType("NN-FitnessIndexed");
	return lType;
	Beagle_StackTraceEndM("const std::string& NN::FitnessIndexed::getType() const")
}


/*!
 *  \brief Test order of two indexed fitness measure.
 *  \param inRightObj Right measure to compare.
 *  \return True if the actual fitness is less than the right one, false if not.
 */
bool NN::FitnessIndexed::isDominated(const Fitness& inRightFitness) const
{
	Beagle_StackTraceBeginM();
	const NN::FitnessIndexed& lRightFitness = castObjectT<const NN::FitnessIndexed&>(inRightFitness);
	if(isValid() != lRightFitness.isValid())
		return false;
	if((isValid()==false) && (lRightFitness.isValid()==false))
		return false;
	return getValue() < lRightFitness.getValue();
	Beagle_StackTraceEndM("bool NN::FitnessIndexed::isDominated(const Object& inRightObj) const");
}


/*!
 *  \brief Test equality of two indexed fitness measure.
 *  \param inRightObj Right measure to compare.
 *  \return True if the fitness are equal, false if not.
 */
bool NN::FitnessIndexed::isEqual(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const NN::FitnessIndexed& lRightFitness = castObjectT<const NN::FitnessIndexed&>(inRightObj);
	if(isValid() != lRightFitness.isValid())
		return false;
	if((isValid()==false) && (lRightFitness.isValid()==false))
		return true;
	return getValue() == lRightFitness.getValue();
	Beagle_StackTraceEndM("bool NN::FitnessIndexed::isEqual(const Object& inRightObj) const");
}


/*!
 *  \brief Test order of two indexed fitness measure.
 *  \param inRightObj Right measure to compare.
 *  \return True if the actual fitness is less from the right one, false if not.
 */
bool NN::FitnessIndexed::isLess(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const NN::FitnessIndexed& lRightFitness = castObjectT<const NN::FitnessIndexed&>(inRightObj);
	if(isValid() != lRightFitness.isValid())
		return false;
	if((isValid()==false) && (lRightFitness.isValid()==false))
		return false;
	return getValue() < lRightFitness.getValue();
	Beagle_StackTraceEndM("bool NN::FitnessIndexed::isLess(const Object& inRightObj) const");
}


/*!
 *  \brief Read an indexed fitness from a XML subtree.
 *  \param inIter XML iterator to use to read the fitness value.
 */
void NN::FitnessIndexed::read(PACC::XML::ConstIterator inIter)
{
	Beagle_StackTraceBeginM();
	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!="Fitness"))
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Fitness> expected!");

	std::string lValid = inIter->getAttribute("valid");
	if(lValid.empty() || (lValid == "yes")) {
	
		// Check type of fitness read
		std::string lType = inIter->getAttribute("type");
		if((lType.empty() == false) && (lType != getType())) {
			std::ostringstream lOSS;
			lOSS << "type given '" << lType << "' mismatch type of the fitness '";
			lOSS << getType() << "'!";
			throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
		}

		// Read values
		clear();
		std::string lSizeText = inIter->getAttribute("size");
		if(lSizeText.empty()==false)
			resize(str2uint(lSizeText), -FLT_MAX);

		for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
			if((lChild->getType()!=PACC::XML::eData) || (lChild->getValue()!="Value"))
				continue;
			unsigned int i = 0;
			std::string lIText = lChild->getAttribute("i");
			if(lIText.empty()==false)
				i = str2uint(lIText);

			PACC::XML::ConstIterator lChild2 = lChild->getFirstChild();
			if(!lChild2)
				throw Beagle_IOExceptionNodeM(*lChild, "needed a float value in the <Value> tag!");
			if(lChild2->getType() != PACC::XML::eString)
				throw Beagle_IOExceptionNodeM(*lChild2, "needed a float value in the <Value> tag!");
			(*this)[i] = str2dbl(lChild2->getValue());
		}
		setValid();		
	} else if(lValid == "no") {
		clear();
		setInvalid();
	}
	else {
		throw Beagle_IOExceptionNodeM(*inIter, "bad fitness validity value!");
	}

	Beagle_StackTraceEndM("void NN::FitnessIndexed::read(PACC::XML::ConstIterator)");
}


/*!
 *  \brief Write an indexed fitness into an XML streamer.
 *  \param ioStreamer XML streamer to use to write the fitness value.
 *  \param inIndent Whether output should be indented.
 */
void NN::FitnessIndexed::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();
	if(isValid()) {
		ioStreamer.insertAttribute("size", uint2str(size()));
		for(unsigned int i=0; i<size(); ++i) {
			if ((*this)[i] > -FLT_MAX) {
				ioStreamer.openTag("Value", false);
				ioStreamer.insertAttribute("i", uint2str(i));
				ioStreamer.insertStringContent(dbl2str((*this)[i]));
				ioStreamer.closeTag();
			}
		}
	} else {
		ioStreamer.insertAttribute("valid", "no");
	}
	Beagle_StackTraceEndM("void NN::FitnessIndexed::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}

