
/*!
 *  \file   beagle/NN/src/Graph.cpp
 *  \brief  Implementation of the type NN::Graph.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <queue>
#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct an NN graph.
 */
NN::Graph::Graph() :
		NN::MapT<Node>("Graph"),
		mModules("Modules"),
		mActivationCount(0)
{ }


/*!
 *  \brief Copy NN graph into current.
 *  \param inOriginal NN graph to copy.
 *  \param ioSystem Evolutionary system to use for making the copy.
 */
void NN::Graph::copy(const Member& inOriginal, System& ioSystem)
{
	Beagle_StackTraceBeginM();
	const NN::Graph& lOriginal = castObjectT<const NN::Graph&>(inOriginal);
	(*this) = lOriginal;
	Beagle_StackTraceEndM("void NN::Graph::copy(const Member&,System&)");
}


/*!
 *  \brief Get exact type of the genotype.
 *  \return Exact type of the genotype.
 */
const std::string& NN::Graph::getType() const
{
	Beagle_StackTraceBeginM();
	const static std::string lType("NN-Graph");
	return lType;
	Beagle_StackTraceEndM("const std::string& NN::Graph::getType() const")
}


/*!
 *  \brief Compare the equality of two NN genotype graphs.
 *  \param inRightObj Right genotype graph to be compared to.
 *  \return True if the genotype graphs are equal, false otherwise.
 *  \throw Beagle::BadCastException If compared genotype graph objects are not of the same type.
 */
bool NN::Graph::isEqual(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const NN::Graph& lRightGraph = castObjectT<const NN::Graph&>(inRightObj);
	return MapT<Node>::isEqual(lRightGraph);
	Beagle_StackTraceEndM("bool NN::Graph::isEqual(const Object& inRightObj) const");
}


/*!
 *  \brief Check if this genotype graph is less than the other.
 *  \param inRightObj Right genotype graph to be compared to.
 *  \return True if this genotype graph is less than the right one, false otherwise.
 *  \throw Beagle::BadCastException If compared genotype graph objects are not of the same type.
 */
bool NN::Graph::isLess(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const NN::Graph& lRightGraph = castObjectT<const NN::Graph&>(inRightObj);
	return MapT<Node>::isLess(lRightGraph);
	Beagle_StackTraceEndM("bool NN::Graph::isLess(const Object& inRightObj) const");
}


/*!
 *  \brief Read an NN genotype graph from an XML subtree.
 *  \param inIter XML iterator to use to read the genotype graph from.
 *  \param ioContext Context to use to read the genotype graph.
 */
void NN::Graph::readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType() != PACC::XML::eData) || (inIter->getValue() != "Genotype")) {
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Genotype> expected!");
	}
	std::string lGenotypeType = inIter->getAttribute("type");
	if((lGenotypeType.empty() == false) && (lGenotypeType != getType())) {
		std::ostringstream lOSS;
		lOSS << "type given '" << lGenotypeType << "' mismatch type of the genotype '";
		lOSS << getType() << "'!";
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}

	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if(lChild->getType() == PACC::XML::eData) {
			if(lChild->getValue() == "Graph") {
				MapT<Node>::read(lChild);
			} else if(lChild->getValue() == "Module") {
				mMain.read(lChild);
			} else if(lChild->getValue() == "Modules") {
				mModules.read(lChild);
			} else {
				throw Beagle_IOExceptionNodeM(*lChild, "Unexpected tag in genotype!");
			}
		}
	}

	Beagle_StackTraceEndM("void NN::Graph::readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext)");
}


/*!
 *  \brief Write an NN genotype graph into a XML streamer.
 *  \param ioStreamer XML streamer to write the genotype graph into.
 *  \param inIndent Whether XML output should be indented.
 */
void NN::Graph::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();

	MapT<Node>::write(ioStreamer, true);
	mMain.write(ioStreamer, true);
	mModules.write(ioStreamer, true);

	Beagle_StackTraceEndM("void NN::Graph::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}


/*!
 *  \brief Choose a random node in proportion to module sizes.
 *  \param ioRandomizer The randomizer to use.
 *  \param inIncludeMain Whether to include the main module while choosing the node.
 *  \param inIncludeInput Whether to include input nodes while choosing the node.
 *  \return The randomly chosen module key and node id.
 */
NN::Graph::UIntPair NN::Graph::getRandomNode(Beagle::Randomizer& ioRandomizer, bool inIncludeMain, bool inIncludeInput) const
{
	Beagle_StackTraceBeginM();

	// Calculate the total sizes of the modules.
	unsigned int lTotalSize = 0;
	for (MapT<Module>::const_iterator lIter = mModules.begin(); lIter != mModules.end(); ++lIter) {
		lTotalSize += lIter->second.getSize(inIncludeInput);
	}
	if (inIncludeMain)
		lTotalSize += mMain.getSize(inIncludeInput);

	unsigned int lIndex = ioRandomizer.rollInteger(0, lTotalSize-1);

	// Find the module for the chosen index.
	MapT<Module>::const_iterator lIter;
	for (lIter = mModules.begin(); lIter != mModules.end() && lIndex >= lIter->second.getSize(inIncludeInput); ++lIter) {
		lIndex -= lIter->second.getSize(inIncludeInput);
	}
	const Module& lModule = (lIter != mModules.end()) ? lIter->second : mMain;
	unsigned int lModuleKey = (lIter != mModules.end()) ? lIter->first : UINT_MAX;

	unsigned int lNodeId = lModule.getRandomNode(ioRandomizer, inIncludeInput);

	// Note that getting the module key from the node will not always give the key
	// of the module chosen above because the node could be a module input node,
	// which have the module key of the main module.
	return UIntPair(lModuleKey, lNodeId);

	Beagle_StackTraceEndM("NN::Graph::UIntPair NN::Graph::getRandomNode(Beagle::Randomizer& ioRandomizer, bool inIncludeMain, bool inIncludeInput) const");
}


/*!
 *  \brief Choose a random node from a specified module, including possibly the main module.
 *  \param ioRandomizer The randomizer to use.
 *  \param inModuleKey The key of the module from which to choose the node.
 *  \param inIncludeMain Whether to include the main module while choosing the node.
 *  \param inIncludeInput Whether to include input nodes while choosing the node.
 *  \return The randomly chosen module key and node id.
 */
NN::Graph::UIntPair NN::Graph::getRandomNode(Beagle::Randomizer& ioRandomizer, unsigned int inModuleKey, bool inIncludeMain, bool inIncludeInput) const
{
	Beagle_StackTraceBeginM();

	if (inModuleKey == UINT_MAX)
		return UIntPair(UINT_MAX, mMain.getRandomNode(ioRandomizer, inIncludeInput));

	// Calculate the total sizes of the modules.
	const Module& lSpecModule = getModule(inModuleKey);
	unsigned int lTotalSize = lSpecModule.getSize(inIncludeInput);
	if (inIncludeMain)
		lTotalSize += mMain.getSize(inIncludeInput);

	unsigned int lIndex = ioRandomizer.rollInteger(0, lTotalSize-1);
	const Module& lModule = (lIndex < lSpecModule.getSize(inIncludeInput)) ? lSpecModule : mMain;
	unsigned int lModuleKey = (lIndex < lSpecModule.getSize(inIncludeInput)) ? inModuleKey : UINT_MAX;

	unsigned int lNodeId = lModule.getRandomNode(ioRandomizer, inIncludeInput);

	// Note that getting the module key from the node will not always give the key
	// of the module chosen above because the node could be a module input node,
	// which have the module key of the main module.
	return UIntPair(lModuleKey, lNodeId);

	Beagle_StackTraceEndM("NN::Graph::UIntPair NN::Graph::getRandomNode(Beagle::Randomizer& ioRandomizer, unsigned int inModuleKey, bool inIncludeMain, bool inIncludeInput) const");
}


/*!
 *  \return The pair (TargetId, SourceId) of a random edge chosen with uniform probability.
 */
NN::Graph::UIntPair NN::Graph::getRandomEdge(Beagle::Randomizer& ioRandomizer) const
{
	Beagle_StackTraceBeginM();

	// Choose an edge index from the total number of edges in the graph.
	const unsigned int lGraphNumberEdges = getNumberEdges();

	unsigned int lEdgeIndex = ioRandomizer.rollInteger(0, lGraphNumberEdges-1);
	unsigned int lEdgeCount = 0;

	// Find out which node in the graph the edge is incident on.
	unsigned int lTargetId = UINT_MAX;
	for (NN::Graph::const_iterator lGraphIter = begin(); lGraphIter != end() && lTargetId == UINT_MAX; ++lGraphIter) {
		const unsigned int lNodeNumberEdges = lGraphIter->second.getNumberEdges();
		if(lEdgeIndex < lEdgeCount+lNodeNumberEdges) {
			lTargetId = lGraphIter->second.getIdentifier();
		} else {
			lEdgeCount += lNodeNumberEdges;
		}
	}

	const NN::Node& lNode = getNode(lTargetId);

	// Find out which incoming edge of the node is the chosen one.
	unsigned int lSourceId = UINT_MAX;
	for (NN::Node::const_iterator lNodeIter = lNode.begin(); lNodeIter != lNode.end() && lSourceId == UINT_MAX; ++lNodeIter) {
		if(lEdgeIndex < lEdgeCount+1) {
			lSourceId = lNodeIter->second.getSourceIdentifier();
		} else {
			lEdgeCount += 1;
		}
	}

	return UIntPair(lTargetId, lSourceId);

	Beagle_StackTraceEndM("NN::Graph::UIntPair NN::Graph::getRandomEdge(Beagle::Randomizer& ioRandomizer) const");
}


/*!
 *  \brief Choose a random edge from a specified module.
 *  \param ioRandomizer The randomizer to use.
 *  \param inModuleKey The key of the module from which to choose the edge.
 *  \return The (TargetId, SourceId) pair of the randomly chosen edge.
 */
NN::Graph::UIntPair NN::Graph::getRandomEdge(Beagle::Randomizer& ioRandomizer, unsigned int inModuleKey) const
{
	Beagle_StackTraceBeginM();

	const Module& lModule = getModule(inModuleKey);
	const NodeSet& lOutSet = lModule.getOutputSet();
	const NodeSet& lHidSet = lModule.getHiddenSet();

	// Calculate the number of edges in the module.
	unsigned int lNumEdges = 0;
	for (NodeSet::const_iterator lOutIter = lOutSet.begin(); lOutIter != lOutSet.end(); ++lOutIter) {
		lNumEdges += getNode(*lOutIter).getNumberEdges();
	}
	for (NodeSet::const_iterator lHidIter = lHidSet.begin(); lHidIter != lHidSet.end(); ++lHidIter) {
		lNumEdges += getNode(*lHidIter).getNumberEdges();
	}

	unsigned int lIndex = ioRandomizer.rollInteger(0, lNumEdges-1);

	// Find the node on which the chosen edge is incident on.
	NodeSet::const_iterator lIter;
	unsigned int lTargetId;
	for (lIter = lOutSet.begin(); lIter != lOutSet.end() && lIndex >= getNode(*lIter).getNumberEdges(); ++lIter) {
		lIndex -= getNode(*lIter).getNumberEdges();
	}
	if (lIter != lOutSet.end()) {
		lTargetId = *lIter;
	} else {
		for (lIter = lHidSet.begin(); lIter != lHidSet.end() && lIndex >= getNode(*lIter).getNumberEdges(); ++lIter) {
			lIndex -= getNode(*lIter).getNumberEdges();
		}
		if (lIter != lHidSet.end()) {
			lTargetId = *lIter;
		} else {
			throw Beagle_RunTimeExceptionM("Couldn't pick the target node id of random edge");
		}
	}

	unsigned int lSourceId = getNode(lTargetId).getRandomEdge(ioRandomizer).getSourceIdentifier();

	return UIntPair(lTargetId, lSourceId);

	Beagle_StackTraceEndM("NN::Graph::UIntPair NN::Graph::getRandomEdge(Beagle::Randomizer& ioRandomizer, unsigned int inModuleKey) const");
}


/*!
 *  \brief Propagate activation in the graph by one link.
 *  \param inActivations Map containing the activations of input nodes.
 */
void NN::Graph::activate(const Activations& inActivations)
{
	Beagle_StackTraceBeginM();

	// Increment activation count of the graph first.
	mActivationCount++;

	// Set the activations of the input nodes of the graph.  Note that these activations
	// will affect the activations of the other nodes only in the next call to this
	// function when the activation count will be incremented.
	for (Activations::const_iterator lIter = inActivations.begin(); lIter != inActivations.end(); ++lIter) {
		setActivation(lIter->first, lIter->second);
	}

	// Activate all non-input nodes of the graph.
	for (Graph::iterator lGraphIter = begin(); lGraphIter != end(); ++lGraphIter) {
		Node& lNode = lGraphIter->second;
		if (inActivations.find(lNode.getIdentifier()) == inActivations.end()) {
			double lActivationSum = 0.0;
			bool lActivateNode = false;
			for (Node::const_iterator lNodeIter = lNode.begin(); lNodeIter != lNode.end(); ++lNodeIter) {
				const Edge& lEdge = lNodeIter->second;
				const double lSourceActivation = getPreviousActivation(lEdge.getSourceIdentifier());
				if (lSourceActivation < DBL_MAX) {
					lActivationSum += lSourceActivation*lEdge.getWeight();
					lActivateNode = true;
				}
			}
			// The node is activated only if at least one of its sources was previously activated.
			if (lActivateNode)
				activate(lNode.getIdentifier(), lActivationSum);
		}
	}

	Beagle_StackTraceEndM("void NN::Graph::activate(const Activations& inActivations)");
}


/*!
 *  \brief Get the activations of the nodes specified in given map.
 *  \param outActivations Map returned with activations of the specified nodes.
 */
void NN::Graph::getActivations(Activations& outActivations) const
{
	Beagle_StackTraceBeginM();

	for (Activations::iterator lIter = outActivations.begin(); lIter != outActivations.end(); ++lIter) {
		lIter->second = getActivation(lIter->first);
	}

	Beagle_StackTraceEndM("void NN::Graph::getActivations(Activations& outActivations) const");
}


/*!
 *  \brief Write the graph in the graphviz dot format.
 *  \param ioStream The stream to write the graph to.
 *  \param inContext The context of evolution.
 */
void NN::Graph::writeGraphviz(std::ostream& ioStream, const Beagle::Context& inContext) const
{
	Beagle_StackTraceBeginM();

	GraphState::Handle lGraphState = castHandleT<GraphState>(inContext.getSystem().getComponent("NN-GraphState"));

	const Register& lRegister = inContext.getSystem().getRegister();
	const double lEdgeMaxVal = (castHandleT<Double>(lRegister.getEntry("nn.edge.initwtmax")))->getWrappedValue();
	const double lEdgeMinVal = (castHandleT<Double>(lRegister.getEntry("nn.edge.initwtmin")))->getWrappedValue();

	std::ostringstream lOSSHead;
	std::ostringstream lOSSBody;
	std::ostringstream lOSSTail;
	std::ostringstream lOSSSources;
	std::ostringstream lOSSSinks;

	lOSSHead << "digraph G {" << std::endl;
	lOSSHead << "rankdir=LR; ranksep=0.22; nodesep=0.06; center=true;" << std::endl;
	lOSSSources << "{ rank = source; ";
	lOSSSinks << "{ rank = sink; ";
	for (NN::Graph::const_iterator lGraphIter = begin(); lGraphIter != end(); ++lGraphIter) {
		const NN::Node& lNode = lGraphIter->second;
		unsigned int lNodeId = lNode.getIdentifier();

		// Shape of the node is determined by the node type.
		lOSSBody << lNodeId << " [";
		if (lGraphState->isInput(lNodeId)) {  // Input node
			lOSSSources << lNodeId << "; ";
			lOSSBody << "shape=box,width=0.20,height=0.20";
		} else if (lGraphState->isOutput(lNodeId)) {  // Output node
			lOSSSinks << lNodeId << "; ";
			lOSSBody << "shape=diamond,width=0.28,height=0.28";
		} else {  // Hidden node
			lOSSBody << "shape=ellipse,width=0.22,height=0.22";
		}

		lOSSBody << ",label=\"\"";
		if (lGraphState->isHelper(lNodeId))
			lOSSBody << ",color=\"#00FF00\"";
		lOSSBody << ",style=filled,fillcolor=\"" << getNodeColor(lNodeId, -1.0, 1.0) << "\"";
		lOSSBody << "];" << std::endl;  // End of node attributes.

		for (NN::Node::const_iterator lNodeIter = lNode.begin(); lNodeIter != lNode.end(); ++lNodeIter) {
			const NN::Edge& lEdge = lNodeIter->second;
			unsigned int lSourceId = lEdge.getSourceIdentifier();

			// Style of the edges is depends on whether the edge is recurrent.
			lOSSBody << lSourceId << " -> " << lNodeId << " [style=";
			if (lEdge.isRecurrent()) {
				lOSSBody << "dashed";
			} else {
				lOSSBody << "solid";
			}
			lOSSBody << ",color=\"" << getEdgeColor(lSourceId, lNodeId, lEdgeMinVal, lEdgeMaxVal) << "\"";
			lOSSBody << ",arrowsize=0.5";
			lOSSBody << "];" << std::endl;  // End of edge attributes.
		}
	}

	lOSSSources << "}" << std::endl;
	lOSSSinks << "}" << std::endl;
	lOSSTail << "}" << std::endl;

	// Write the different sections to the output stream.
	ioStream << lOSSHead.str();
	ioStream << lOSSSources.str();
	ioStream << lOSSSinks.str();
	ioStream << lOSSBody.str();
	ioStream << lOSSTail.str();

	Beagle_StackTraceEndM("void NN::Graph::writeGraphviz(std::ostream& ioStream, const Beagle::Context& inContext) const");
}


/*!
 *  \brief Produce an rgb color string for the specified node.
 */
std::string NN::Graph::getNodeColor(unsigned int inNodeId, double inMinValue, double inMaxValue) const
{
	Beagle_StackTraceBeginM();

	// Color of the node is determined by the node activation.
	// If the node activation is invalid (DBL_MAX), the node appears gray.
	// The activation for lNodeL is divided by 2 because lightness value of 0.5 is normal.
	double lNodeActivation = getActivation(inNodeId);
	double lNormL = (lNodeActivation < 0) ? inMinValue : inMaxValue;
	double lNodeH = (lNodeActivation < 0) ? 0.66 : 0.0;
	double lNodeS = (lNodeActivation < DBL_MAX) ? 1.0 : 0.0;
	double lNodeL = (lNodeActivation < DBL_MAX) ? (lNodeActivation/(2.0*lNormL)) : 0.5;

	return hsl2rgb(lNodeH, lNodeS, lNodeL);

	Beagle_StackTraceEndM("std::string NN::Graph::getNodeColor(unsigned int inNodeId, double inMinValue, double inMaxValue) const");
}


/*!
 *  \brief Produce an rgb color string for the specified edge.
 */
std::string NN::Graph::getEdgeColor(unsigned int inSourceId, unsigned int inTargetId, double inMinValue, double inMaxValue) const
{
	Beagle_StackTraceBeginM();

	// Edge color is determined by the weight of the edge.
	double lEdgeWeight = getNode(inTargetId).getEdge(inSourceId).getWeight();
	double lNormL = (lEdgeWeight < 0) ? inMinValue : inMaxValue;
	double lEdgeH = (lEdgeWeight < 0) ? 0.66 : 0.0;
	double lEdgeS = 1.0;
	double lEdgeL = lEdgeWeight/(2.0*lNormL);

	return hsl2rgb(lEdgeH, lEdgeS, lEdgeL);

	Beagle_StackTraceEndM("std::string NN::Graph::getEdgeColor(unsigned int inSourceId, unsigned int inTargetId, double inMinValue, double inMaxValue) const");
}


/*!
 *  \brief Helper function used by hsl2rgb() for computing an r/g/b value.
 */
unsigned int NN::Graph::hue2rgb(double inM1, double inM2, double inH) const
{
	Beagle_StackTraceBeginM();

	if (inH < 0.0)
		inH = inH+1.0;
	if (inH > 1.0)
		inH = inH-1.0;

	double lC;
	if (inH*6.0 < 1.0)
		lC = inM1+(inM2-inM1)*inH*6.0;
	else if (inH*2.0 < 1.0)
		lC = inM2;
	else if (inH*3.0 < 2.0)
		lC = inM1+(inM2-inM1)*(2.0/3.0-inH)*6.0;
	else
		lC = inM1;

	return (unsigned int)(lC*255);

	Beagle_StackTraceEndM("unsigned int NN::Graph::hue2rgb(double inM1, double inM2, double inH) const");
}


/*!
 *  \brief Convert an hsl color value to the rgb color string.
 *  \param inH Hue
 *  \param inS Saturation
 *  \param inL Lightness
 *  \return The rgb color string
 *
 *  Conversion algorithm reference: http://www.w3.org/TR/css3-color/#hsl-color
 *  The h, s and l values are assumed to be normalized to lie in [0,1].
 *  The rgb color string consists of a # character followed by three hexadecimal
 *  numbers between 00 and FF.
 */
std::string NN::Graph::hsl2rgb(double inH, double inS, double inL) const
{
	Beagle_StackTraceBeginM();

	double lM2 = (inL<=0.5) ? inL*(inS+1.0) : inL+inS-inL*inS;
	double lM1 = inL*2.0-lM2;
	unsigned int lR = hue2rgb(lM1, lM2, inH+1.0/3.0);
	unsigned int lG = hue2rgb(lM1, lM2, inH);
	unsigned int lB = hue2rgb(lM1, lM2, inH-1.0/3.0);

	char lRGB[20];
	sprintf(lRGB, "#%02X%02X%02X", lR, lG, lB);
	return lRGB;

	Beagle_StackTraceEndM("std::string NN::Graph::hsl2rgb(double inH, double inS, double inL) const");
}

