
/*!
 *  \file   beagle/NN/src/GraphState.cpp
 *  \brief  Implementation of the type NN::GraphState.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct the global graph evolution state.
 */
NN::GraphState::GraphState() :
		Component("NN-GraphState"),
		mNodeCounter(0),
		mInputRange(0, 0),
		mOutputRange(0, 0),
		mHelperRange(0, 0),
		mArchitecture("Architecture")
{ }


/*!
 *  \brief Register the parameters of the global graph evolution state.
 *  \param ioSystem Evolutionary system used for initialization.
 */
void NN::GraphState::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Component::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of input nodes in graph",
		    "UInt",
		    "1",
		    "Number of input nodes in the genotype graph."
		);
		mNumInputs = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.graph.numin", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of output nodes in graph",
		    "UInt",
		    "1",
		    "Number of output nodes in the genotype graph."
		);
		mNumOutputs = castHandleT<UInt>(
		                  ioSystem.getRegister().insertEntry("nn.graph.numout", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of helper nodes in graph",
		    "UInt",
		    "0",
		    "Number of helper nodes in the genotype graph."
		);
		mNumHelpers = castHandleT<UInt>(
		                  ioSystem.getRegister().insertEntry("nn.graph.numhelp", new UInt(0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::GraphState::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialization of the global graph evolution state.
 *  \param ioSystem Evolutionary system used for initialization.
 */
void NN::GraphState::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	if (mNumInputs==NULL || mNumOutputs==NULL || mNumHelpers==NULL)
		throw Beagle_RunTimeExceptionM(std::string("NN::GraphState has not been initialized.").
		                               append(" Consider the NN::GraphState::registerParams() method."));

	unsigned int lNumInputs = mNumInputs->getWrappedValue();
	unsigned int lNumOutputs = mNumOutputs->getWrappedValue();
	unsigned int lNumHelpers = mNumHelpers->getWrappedValue();

	Beagle_ValidateParameterM(lNumInputs>0, "nn.graph.numin", ">0");
	Beagle_ValidateParameterM(lNumOutputs>0, "nn.graph.numout", ">0");

	// Set the range of input, output and helper node identifiers.
	// Note that if there are no helpers, the second element of helper range
	// will be one less than the first.
	mInputRange.first = 0;
	mInputRange.second = lNumInputs - 1;
	mOutputRange.first = lNumInputs;
	mOutputRange.second = lNumInputs + lNumOutputs - 1;
	mHelperRange.first = lNumInputs + lNumOutputs;
	mHelperRange.second = lNumInputs + lNumOutputs + lNumHelpers - 1;

	// Initialize the node counter to the next valid node identifier.
	mNodeCounter = lNumInputs + lNumOutputs + lNumHelpers;

	Beagle_StackTraceEndM("void NN::GraphState::init(Beagle::System& ioSystem)");
}


/*!
 *  \brief Read an unsigned integer array from an XML subtree.
 *  \param inIter XML iterator to use to read the array from.
 *  \throw Beagle::IOException If reading from the stream failed.
 */
void NN::GraphState::UIntArray::read(PACC::XML::ConstIterator inIter)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!="Array"))
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Array> expected!");

	Beagle::UIntArray::read(inIter->getFirstChild());

	Beagle_StackTraceEndM("void NN::GraphState::UIntArray::read(PACC::XML::ConstIterator inIter)");
}


/*!
 *  \brief Write a NN graph edge into an XML streamer.
 *  \param ioStreamer XML streamer to write the edge into.
 *  \param inIndent Whether XML output should be indented.
 *  \param inKey Key attribute if this edge is used as an entry in a map.
 */
void NN::GraphState::UIntArray::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const
{
	Beagle_StackTraceBeginM();

	ioStreamer.openTag("Array", false);
	if (inKey < UINT_MAX)
		ioStreamer.insertAttribute("key", uint2str(inKey));
	Beagle::UIntArray::write(ioStreamer, false);
	ioStreamer.closeTag();

	Beagle_StackTraceEndM("void NN::GraphState::UIntArray::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const");
}


/*!
 *  \brief Read a primitive super set from a XML subtree.
 *  \param inIter XML iterator used to read the super set from.
 *  \param ioSystem Evolutionary system.
 *  \throw Beagle::IOException If reading failed.
 */
void NN::GraphState::readWithSystem(PACC::XML::ConstIterator inIter, Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!=getName()))
		throw Beagle_IOExceptionNodeM(*inIter, "tag <"+getName()+"> expected!");

	std::string lNodeCounterText = inIter->getAttribute("idcounter");
	if(lNodeCounterText.empty())
		throw Beagle_IOExceptionNodeM(*inIter, "attribute \"idcounter\" of the tag <"+getName()+"> expected!");
	else
		mNodeCounter = str2uint(lNodeCounterText);

	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if(lChild->getType() == PACC::XML::eData) {
			if(lChild->getValue() == "Inputs") {
				mInputRange.first = str2uint(lChild->getAttribute("first"));
				mInputRange.second = str2uint(lChild->getAttribute("last"));
			} else if(lChild->getValue() == "Outputs") {
				mOutputRange.first = str2uint(lChild->getAttribute("first"));
				mOutputRange.second = str2uint(lChild->getAttribute("last"));
			} else if(lChild->getValue() == "Helpers") {
				mHelperRange.first = str2uint(lChild->getAttribute("first"));
				mHelperRange.second = str2uint(lChild->getAttribute("last"));
			} else if(lChild->getValue() == "Architecture") {
				mArchitecture.read(lChild);
			} else {
				throw Beagle_IOExceptionNodeM(*lChild, "Unexpected tag in <"+getName()+">!");
			}
		}
	}

	Beagle_StackTraceEndM("void NN::GraphState::readWithSystem(PACC::XML::ConstIterator inIter, Beagle::System& ioSystem)");
}


/*!
 *  \brief Write a primitive super set into a Beagle XML stream.
 *  \param ioStreamer XMl streamer to write the super set into.
 *  \param inIndent Whether XML output should be indented.
 */
void NN::GraphState::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();

	ioStreamer.openTag("GraphState", inIndent);
	ioStreamer.insertAttribute("idcounter", uint2str(mNodeCounter));
	ioStreamer.openTag("Inputs", inIndent);
	ioStreamer.insertAttribute("first", uint2str(mInputRange.first));
	ioStreamer.insertAttribute("last", uint2str(mInputRange.second));
	ioStreamer.closeTag();
	ioStreamer.openTag("Outputs", inIndent);
	ioStreamer.insertAttribute("first", uint2str(mOutputRange.first));
	ioStreamer.insertAttribute("last", uint2str(mOutputRange.second));
	ioStreamer.closeTag();
	ioStreamer.openTag("Helpers", inIndent);
	ioStreamer.insertAttribute("first", uint2str(mHelperRange.first));
	ioStreamer.insertAttribute("last", uint2str(mHelperRange.second));
	ioStreamer.closeTag();
	mArchitecture.write(ioStreamer, inIndent);
	ioStreamer.closeTag();

	Beagle_StackTraceEndM("void NN::GraphState::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}


/*!
 *  \brief Determine the identifier for a new node created by splitting an edge.
 *  \param inGraph The graph in which the new node is being created.
 *  \param inSourceId The id of the source node of the edge.
 *  \param inTargetId The id of the target node of the edge.
 *  \return The identifier for the new node.
 */
unsigned int NN::GraphState::getNewNodeIdentifier(const Graph& inGraph, unsigned int inSourceId, unsigned int inTargetId)
{
	Beagle_StackTraceBeginM();

	// If this edge was split before, check if we can use any of the node ids in the array for that edge.
	unsigned int lNodeId = UINT_MAX;
	if (mArchitecture.hasKey(inTargetId) && mArchitecture.getData(inTargetId).hasKey(inSourceId)) {
		GraphState::UIntArray& lNodeIdArray = mArchitecture.getData(inTargetId).getData(inSourceId);
		for (unsigned int i = 0; i < lNodeIdArray.size() && lNodeId == UINT_MAX; i++) {
			// We can use the first node id that do not already exist in the graph.
			if (!inGraph.hasNode(lNodeIdArray[i]))
				lNodeId = lNodeIdArray[i];
		}
	}

	// Otherwise, a new node id needs to be created.
	if (lNodeId == UINT_MAX) {
		lNodeId = mNodeCounter++;
		// If this edge was not split before, then create new entries in the node ancestry data structure,
		// and add the new node id.
		mArchitecture.addData(inTargetId, MapT< GraphState::UIntArray >()).addData(inSourceId, GraphState::UIntArray()).push_back(lNodeId);
	}

	return lNodeId;

	Beagle_StackTraceEndM("unsigned int NN::GraphState::getNewNodeIdentifier(const NN::Graph& inGraph, unsigned int inSourceId, unsigned int inTargetId)");
}


