
/*!
 *  \file   beagle/NN/src/InitControllerOp.cpp
 *  \brief  Source code of class NN::InitControllerOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <sstream>

using namespace Beagle;


/*!
 *  \brief Construct a NN genotype graph fully connected intialization operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
NN::InitControllerOp::InitControllerOp(std::string inReproProbaName, std::string inName) :
		Beagle::NN::InitializationOp(inReproProbaName, inName)
{ }


/*!
 *  \brief Register the parameters of the generic NN intialization operator.
 *  \param ioSystem System of the evolution.
 */
void NN::InitControllerOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	NN::InitializationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of modules in the network",
		    "UInt",
		    "4",
		    "Number of modules in the network."
		);
		mNumModules = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.number", new UInt(4), lDescription));
	}
	{
		Register::Description lDescription(
		    "The number of different module instances in the genotype",
		    "UInt",
		    "1",
		    "The genotype modules instances are copied to produce the other modules in the phenotype."
		);
		mNumModInst = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.numinst", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of rows into which the modules are arranged",
		    "UInt",
		    "1",
		    "Number of rows into which the modules are arranged."
		);
		mNumModRows = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.numrow", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of input nodes per module",
		    "UInt",
		    "2",
		    "Number of input nodes per module."
		);
		mNumModInputs = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.numin", new UInt(2), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of output nodes per module",
		    "UInt",
		    "1",
		    "Number of output nodes per module."
		);
		mNumModOutputs = castHandleT<UInt>(
		                  ioSystem.getRegister().insertEntry("nn.module.numout", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of helper nodes per module",
		    "UInt",
		    "2",
		    "Number of helper nodes per module."
		);
		mNumModHelpers = castHandleT<UInt>(
		                  ioSystem.getRegister().insertEntry("nn.module.numhelp", new UInt(2), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to make output nodes in modules separate instances",
		    "Bool",
		    "0",
		    "Module output nodes will be different instances if this parameter is set."
		);
		mModOutInst = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.module.outinst", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to use recurrency between outputs and hidden nodes",
		    "Bool",
		    "0",
		    "Each output is connected to a hidden node of the module if this parameter is set."
		);
		mModRecurrency = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.module.recur", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to use recurrency between hidden nodes",
		    "Bool",
		    "0",
		    "Recurrency between hidden nodes in the module is enabled if this parameter is set."
		);
		mModSimpleRecurrency = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.module.simplerecur", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to use self loop edges on hidden nodes when recurrency between hidden nodes is enabled",
		    "Bool",
		    "1",
		    "Each hidden node is connected to itself if simple recurrency is enabled and this parameter is set."
		);
		mModSelfEdge = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.module.selfedge", new Bool(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to use edges to other hidden nodes when recurrency between hidden nodes is enabled",
		    "Bool",
		    "1",
		    "Each hidden node is connected to other hidden nodes in the module if simple recurrency is enabled and this parameter is set."
		);
		mModInterEdge = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.module.interedge", new Bool(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "The number of hidden nodes in a module to exclude from recurrency",
		    "UInt",
		    "0",
		    "When creating recurrent edges in a module, this many hidden nodes are excluded from such edges."
		);
		mNumModExcludeRecur = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.numexclrecur", new UInt(0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::InitControllerOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize a genotype.
 *  \param outGraph Graph to initialize.
 *  \param ioContext Evolution context.
 *  \return The number of nodes in the genotype graph.
 */
unsigned int NN::InitControllerOp::initGraph(NN::Graph& outGraph, Context& ioContext) const
{
	Beagle_StackTraceBeginM();

	outGraph.clear();

	GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	const double lMinBias = mMinNodeBias->getWrappedValue();
	const double lMaxBias = mMaxNodeBias->getWrappedValue();
	const double lMinSlope = mMinNodeSlope->getWrappedValue();
	const double lMaxSlope = mMaxNodeSlope->getWrappedValue();
	const double lMinWeight = mMinEdgeWeight->getWrappedValue();
	const double lMaxWeight = mMaxEdgeWeight->getWrappedValue();

    const bool lModRecurrency = mModRecurrency->getWrappedValue();
    const bool lModSimpleRecurrency = mModSimpleRecurrency->getWrappedValue();
    const bool lModSelfEdge = mModSelfEdge->getWrappedValue();
    const bool lModInterEdge = mModInterEdge->getWrappedValue();
    const bool lModOutInst = mModOutInst->getWrappedValue();

    unsigned int lNumModules = mNumModules->getWrappedValue();
    unsigned int lNumModRows = mNumModRows->getWrappedValue();
    unsigned int lNumModCols = lNumModules/lNumModRows;
    unsigned int lNumModInst = mNumModInst->getWrappedValue();
    unsigned int lNumModInputs = mNumModInputs->getWrappedValue();
    unsigned int lNumModOutputs = mNumModOutputs->getWrappedValue();
    unsigned int lNumModHelpers = mNumModHelpers->getWrappedValue();
    unsigned int lNumModExcludeRecur = mNumModExcludeRecur->getWrappedValue();

    Module& lMain = outGraph.getMain();

    for (unsigned int i = 0; i < lNumModInst; i++) {
        // Create module input nodes.
        for (unsigned int j = 0; j < lNumModInputs; j++) {
            unsigned int lNodeId = lGraphState->getFirstInput() + i + j*lNumModules;
            double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
            double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
            outGraph.addNode(NN::Node(lNodeId, lBias, lSlope));
            lMain.addInput(lNodeId);
        }

        // Create module hidden nodes.
        for (unsigned int j = 0; j < lNumModHelpers; j++) {
            unsigned int lNodeId = lGraphState->getFirstHelper() + i + j*lNumModules;
            double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
            double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
            outGraph.addNode(NN::Node(lNodeId, lBias, lSlope));
            lMain.addHidden(lNodeId);
        }
        
        // Create module output nodes.
        for (unsigned int j = 0; j < lNumModOutputs; j++) {
            unsigned int lNodeId = lGraphState->getFirstOutput() + i + j*lNumModules;
            double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
            double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
            outGraph.addNode(NN::Node(lNodeId, lBias, lSlope));
            lMain.addOutput(lNodeId);
        }
        
        // Create the input to hidden connections.
        for (unsigned int j = 0; j < lNumModHelpers; j++) {
            unsigned int lHiddenId = lGraphState->getFirstHelper() + i + j*lNumModules;
            for (unsigned int k = 0; k < lNumModInputs*lNumModules; k++) {
                unsigned int lInputId = lGraphState->getFirstInput() + k;
                double lWeight = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
                outGraph.getNode(lHiddenId).addEdge(NN::Edge(lInputId, lWeight, false));
            }
        }

        // Create the hidden to output connections.
        for (unsigned int j = 0; j < lNumModOutputs; j++) {
            unsigned int lOutputId = lGraphState->getFirstOutput() + i + j*lNumModules;
            for (unsigned int k = 0; k < lNumModHelpers; k++) {
                unsigned int lHiddenId = lGraphState->getFirstHelper() + i + k*lNumModules;
                double lWeight = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
                outGraph.getNode(lOutputId).addEdge(NN::Edge(lHiddenId, lWeight, false));
            }
        }

        // Create recurrent connections between output and hidden nodes.
        if (lModRecurrency) {
            unsigned int lNumModRecurHelpers = lNumModHelpers - lNumModExcludeRecur;
            for (unsigned int j = 0; j < lNumModRecurHelpers; j++) {
                unsigned int lHiddenId = lGraphState->getFirstHelper() + i + j*lNumModules;
                for (unsigned int k = 0; k < lNumModOutputs; k++) {
                    unsigned int lOutputId = lGraphState->getFirstOutput() + i + k*lNumModules;
                    double lWeight = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
                    outGraph.getNode(lHiddenId).addEdge(NN::Edge(lOutputId, lWeight, true));
                }
            }
        }

        // Create recurrent connections between hidden nodes.
        if (lModSimpleRecurrency) {
            unsigned int lNumModRecurHelpers = lNumModHelpers - lNumModExcludeRecur;
            for (unsigned int j = 0; j < lNumModRecurHelpers; j++) {
                unsigned int lHiddenId = lGraphState->getFirstHelper() + i + j*lNumModules;
                for (unsigned int k = 0; k < lNumModRecurHelpers; k++) {
                    if ((j == k && lModSelfEdge) || (j != k && lModInterEdge)) {
                        unsigned int lSourceId = lGraphState->getFirstHelper() + i + k*lNumModules;
                        double lWeight = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
                        outGraph.getNode(lHiddenId).addEdge(NN::Edge(lSourceId, lWeight, true));
                    }
                }
            }
        }
    }

    // Create remaining input nodes.
    for (unsigned int i = lGraphState->getFirstInput() + lNumModInputs*lNumModules; i <= lGraphState->getLastInput(); i++) {
        double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
        double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
        outGraph.addNode(NN::Node(i, lBias, lSlope));
        lMain.addInput(i);
    }

    // Create remaining hidden nodes.
    for (unsigned int i = lNumModInst; i < lNumModules; i++) {
        for (unsigned int j = 0; j < lNumModHelpers; j++) {
            unsigned int lNodeId = lGraphState->getFirstHelper() + i + j*lNumModules;
            double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
            double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
            outGraph.addNode(NN::Node(lNodeId, lBias, lSlope));
            lMain.addHidden(lNodeId);
        }
    }
    
    // Create remaining output nodes.
    for (unsigned int i = lGraphState->getFirstOutput() + lNumModOutputs*(lModOutInst ? lNumModInst : lNumModules); i <= lGraphState->getLastOutput(); i++) {
        double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
        double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
        outGraph.addNode(NN::Node(i, lBias, lSlope));
        lMain.addOutput(i);
    }

    // Create the new input to hidden connections.
    for (unsigned int i = 0; i < lNumModules; i++) {
        for (unsigned int j = 0; j < lNumModHelpers; j++) {
            unsigned int lHiddenId = lGraphState->getFirstHelper() + i + j*lNumModules;
            for (unsigned int k = lGraphState->getFirstInput() + lNumModInputs*lNumModules; k <= lGraphState->getLastInput(); k++) {
                unsigned int lInputId = lGraphState->getFirstInput() + k;
                double lWeight = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
                outGraph.getNode(lHiddenId).addEdge(NN::Edge(lInputId, lWeight, false));
            }
        }
    }

    // Create the new hidden to output connections.
    unsigned int lNumExtraOutputs = (lGraphState->getLastOutput() - lGraphState->getFirstOutput() + 1) - lNumModOutputs*lNumModules;
    unsigned int lNumExtraModOutputs = lNumExtraOutputs/lNumModules;
    for (unsigned int i = 0; i < lNumModules; i++) {
        for (unsigned int j = 0; j < lNumModHelpers; j++) {
            unsigned int lHiddenId = lGraphState->getFirstHelper() + i + j*lNumModules;
            for (unsigned int k = 0; k < lNumExtraModOutputs; k++) {
                unsigned int lOutputId = lGraphState->getFirstOutput() + lNumModOutputs*lNumModules + i + k*lNumModules;
                double lWeight = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
                outGraph.getNode(lOutputId).addEdge(NN::Edge(lHiddenId, lWeight, false));
            }
        }
    }

	Beagle_LogVerboseM(
	    ioContext.getSystem().getLogger(),
	    "initialization", "Beagle::NN::InitControllerOp",
	    std::string("Initializing the ")+
	    uint2ordinal(ioContext.getGenotypeIndex()+1)+std::string(" genotype graph.")
	);

	return outGraph.getNumberNodes();

	Beagle_StackTraceEndM("unsigned int NN::InitControllerOp::initGraph(NN::Graph& outGraph, Context& ioContext) const");
}

