
/*!
 *  \file   beagle/NN/src/InitFullyConnectedOp.cpp
 *  \brief  Source code of class NN::InitFullyConnectedOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <sstream>

using namespace Beagle;


/*!
 *  \brief Construct a NN genotype graph fully connected intialization operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
NN::InitFullyConnectedOp::InitFullyConnectedOp(std::string inReproProbaName, std::string inName) :
		Beagle::NN::InitializationOp(inReproProbaName, inName)
{ }


/*!
 *  \brief Initialize a genotype.
 *  \param outGraph Graph to initialize.
 *  \param ioContext Evolution context.
 *  \return The number of nodes in the genotype graph.
 */
unsigned int NN::InitFullyConnectedOp::initGraph(NN::Graph& outGraph, Context& ioContext) const
{
	Beagle_StackTraceBeginM();

	outGraph.clear();

	GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	const double lMinBias = mMinNodeBias->getWrappedValue();
	const double lMaxBias = mMaxNodeBias->getWrappedValue();
	const double lMinSlope = mMinNodeSlope->getWrappedValue();
	const double lMaxSlope = mMaxNodeSlope->getWrappedValue();
	const double lMinWeight = mMinEdgeWeight->getWrappedValue();
	const double lMaxWeight = mMaxEdgeWeight->getWrappedValue();

	// The first deme has the main graph, while the remaining demes have modules
	// for the helper nodes.
	if (ioContext.getDemeIndex() == 0) {
		// Create input and output nodes and add them to the main module.
		Module& lMain = outGraph.getMain();
		for (unsigned int i = lGraphState->getFirstInput(); i <= lGraphState->getLastInput(); i++) {
			double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
			double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
			outGraph.addNode(NN::Node(i, lBias, lSlope));
			lMain.addInput(i);
		}
		for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++) {
			double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
			double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
			outGraph.addNode(NN::Node(i, lBias, lSlope));
			lMain.addOutput(i);
		}

		// Create edges connecting each output node to all the input nodes.
		for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++) {
			NN::Node& lOutputNode = outGraph.getNode(i);
			for (unsigned int j = lGraphState->getFirstInput(); j <= lGraphState->getLastInput(); j++) {
				double lWeight = lRandomizer.rollUniform(lMinWeight, lMaxWeight);
				lOutputNode.addEdge(NN::Edge(j, lWeight, false));
			}
		}

		// Create helper modules.
		for (unsigned int lHelperId = lGraphState->getFirstHelper(); lHelperId <= lGraphState->getLastHelper(); lHelperId++) {
			NN::Module& lModule = outGraph.addModule(NN::Module(lHelperId));

			// Create a new node in the graph for the helper and add the node id to the module.
			double lBias = lRandomizer.rollUniform(mMinNodeBias->getWrappedValue(), mMaxNodeBias->getWrappedValue());
			double lSlope = lRandomizer.rollUniform(mMinNodeSlope->getWrappedValue(), mMaxNodeSlope->getWrappedValue());
			NN::Node& lHelper = outGraph.addNode(NN::Node(lHelperId, lBias, lSlope));
			lHelper.setModuleKey(lHelperId);
			lModule.addOutput(lHelperId);

			// The graph inputs and outputs are the module inputs.  Create connections to the
			// module output from the module inputs and add their node ids to the module.
			for (unsigned int i = lGraphState->getFirstInput(); i <= lGraphState->getLastInput(); i++) {
				double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
				lHelper.addEdge(NN::Edge(i, lWeight, false));
				lModule.addInput(i);
			}
			for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++) {
				double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
				lHelper.addEdge(NN::Edge(i, lWeight, false));
				lModule.addInput(i);

				// Connect the helper node to the output also.
				NN::Node& lOutputNode = outGraph.getNode(i);
				lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
				lOutputNode.addEdge(NN::Edge(lHelperId, lWeight, true));
				lModule.addExport(lHelperId, i);
			}
		}
	} else {
		unsigned int lHelperId = lGraphState->getFirstHelper() + ioContext.getDemeIndex() - 1;
		NN::Module& lModule = outGraph.addModule(NN::Module(lHelperId));

		// Create a node in the graph for the helper.
		double lBias = lRandomizer.rollUniform(mMinNodeBias->getWrappedValue(), mMaxNodeBias->getWrappedValue());
		double lSlope = lRandomizer.rollUniform(mMinNodeSlope->getWrappedValue(), mMaxNodeSlope->getWrappedValue());
		NN::Node& lHelper = outGraph.addNode(NN::Node(lHelperId, lBias, lSlope));
		lHelper.setModuleKey(lHelperId);
		lModule.addOutput(lHelperId);

		// The graph inputs and outputs are the module inputs.  Create those node and
		// add edges connecting them to the module output.
		for (unsigned int i = lGraphState->getFirstInput(); i <= lGraphState->getLastInput(); i++) {
			double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
			double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
			NN::Node& lInput = outGraph.addNode(NN::Node(i, lBias, lSlope));
			lInput.setModuleKey(lHelperId);
			lModule.addInput(i);

			double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
			lHelper.addEdge(NN::Edge(i, lWeight, false));
		}
		for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++) {
			double lBias = lRandomizer.rollUniform(lMinBias, lMaxBias);
			double lSlope = lRandomizer.rollUniform(lMinSlope, lMaxSlope);
			NN::Node& lInput = outGraph.addNode(NN::Node(i, lBias, lSlope));
			lInput.setModuleKey(lHelperId);
			lModule.addInput(i);

			double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
			lHelper.addEdge(NN::Edge(i, lWeight, false));
		}
	}

	Beagle_LogVerboseM(
	    ioContext.getSystem().getLogger(),
	    "initialization", "Beagle::NN::InitFullyConnectedOp",
	    std::string("Connecting each output node to all the input nodes to initialize the ")+
	    uint2ordinal(ioContext.getGenotypeIndex()+1)+std::string(" genotype graph.")
	);

	return outGraph.getNumberNodes();

	Beagle_StackTraceEndM("unsigned int NN::InitFullyConnectedOp::initGraph(NN::Graph& outGraph, Context& ioContext) const");
}

