
/*!
 *  \file   beagle/NN/src/InitializationOp.cpp
 *  \brief  Source code of class NN::InitializationOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a plain new initialization operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register.
 *  \param inName Name of the initialization operator.
 */
NN::InitializationOp::InitializationOp(std::string inReproProbaName, std::string inName) :
		Beagle::InitializationOp(inReproProbaName, inName)
{ }


/*!
 *  \brief Register the parameters of the generic NN intialization operator.
 *  \param ioSystem System of the evolution.
 */
void NN::InitializationOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Beagle::InitializationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Minimum value of initial edge weight",
		    "Double",
		    "-10.0",
		    "Edge weight is constrained to be greater than or equal to this value when a new edge is created."
		);
		mMinEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial edge weight",
		    "Double",
		    "10.0",
		    "Edge weight is constrained to be less than this value when a new edge is created."
		);
		mMaxEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node bias",
		    "Double",
		    "-10.0",
		    "Node bias is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node bias",
		    "Double",
		    "10.0",
		    "Node bias is constrained to be less than this value when a new node is created."
		);
		mMaxNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node slope",
		    "Double",
		    "1.0",
		    "Node slope is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemin", new Double(1.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node slope",
		    "Double",
		    "5.0",
		    "Node slope is constrained to be less than this value when a new node is created."
		);
		mMaxNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemax", new Double(5.0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::InitializationOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize the genotype graphs of an individual.
 *  \param outIndividual Individual to initialize.
 *  \param ioContext Evolution context.
 *  \throw Beagle::RunTimeException If the number of inputs/outputs are incorrectly set.
 */
void NN::InitializationOp::initIndividual(Beagle::Individual& outIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(outIndividual);

	// We just have one genotype in the individual.
    const Factory& lFactory = ioContext.getSystem().getFactory();
    Genotype::Alloc::Handle lGenotypeAlloc =
        castHandleT<Genotype::Alloc>(lFactory.getConceptAllocator("Genotype"));
    lIndividual.clear();
    lIndividual.push_back(castHandleT<Genotype>(lGenotypeAlloc->allocate()));

	Graph::Handle lOldGraphHandle = castHandleT<Graph>(ioContext.getGenotypeHandle());
	unsigned int lOldGraphIndex = ioContext.getGenotypeIndex();

	for(unsigned int i=0; i<lIndividual.size(); ++i) {
		ioContext.setGenotypeHandle(lIndividual[i]);
		ioContext.setGenotypeIndex(i);
		initGraph(*lIndividual[i], ioContext);
	}

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "initialization", "Beagle::NN::InitializationOp",
	    std::string("Initialized individual: ")+lIndividual.serialize()
	);

	ioContext.setGenotypeIndex(lOldGraphIndex);
	ioContext.setGenotypeHandle(lOldGraphHandle);

	Beagle_StackTraceEndM("void NN::InitializationOp::initIndividual(Beagle::Individual& outIndividual, Beagle::Context& ioContext)");
}

