
/*!
 *  \file   beagle/NN/src/Module.cpp
 *  \brief  Implementation of the type NN::Module.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct an NN graph module.
 */
NN::Module::Module(unsigned int inKey) :
		mKey(inKey),
		mInputSet("Input"),
		mOutputSet("Output"),
		mHiddenSet("Hidden"),
		mExportMap("Export")
{ }


/*!
 *  \brief Construct an NN graph module by merging the given modules.
 *  \param inModule1 The first module to be merged
 *  \param inModule2 The second module to be merged
 */
NN::Module::Module(const Module& inModule1, const Module& inModule2) :
		mInputSet("Input"),
		mOutputSet("Output"),
		mHiddenSet("Hidden"),
		mExportMap("Export")
{
	Beagle_StackTraceBeginM();
	for (NodeSet::const_iterator lIter = inModule1.mInputSet.begin(); lIter != inModule1.mInputSet.end(); lIter++) {
		mInputSet.insert(*lIter);
	}
	for (NodeSet::const_iterator lIter = inModule2.mInputSet.begin(); lIter != inModule2.mInputSet.end(); lIter++) {
		mInputSet.insert(*lIter);
	}
	for (NodeSet::const_iterator lIter = inModule1.mOutputSet.begin(); lIter != inModule1.mOutputSet.end(); lIter++) {
		mOutputSet.insert(*lIter);
	}
	for (NodeSet::const_iterator lIter = inModule2.mOutputSet.begin(); lIter != inModule2.mOutputSet.end(); lIter++) {
		mOutputSet.insert(*lIter);
	}
	for (NodeSet::const_iterator lIter = inModule1.mHiddenSet.begin(); lIter != inModule1.mHiddenSet.end(); lIter++) {
		mHiddenSet.insert(*lIter);
	}
	for (NodeSet::const_iterator lIter = inModule2.mHiddenSet.begin(); lIter != inModule2.mHiddenSet.end(); lIter++) {
		mHiddenSet.insert(*lIter);
	}
	for (MapT<NodeSet>::const_iterator lIter = inModule1.mExportMap.begin(); lIter != inModule1.mExportMap.end(); lIter++) {
		mExportMap.addData(lIter->first, lIter->second);
	}
	for (MapT<NodeSet>::const_iterator lIter = inModule2.mExportMap.begin(); lIter != inModule2.mExportMap.end(); lIter++) {
		mExportMap.addData(lIter->first, lIter->second);
	}
	mKey = mOutputSet.getMin();
	Beagle_StackTraceEndM("NN::Module::Module(const Module& inModule1, const Module& inModule2)");
}


/*!
 *  \brief Compare equality of two NN graph modules.
 *  \param inRightModule Right module to be compared to.
 *  \return True if the modules have the same input, output and hidden node ids, false otherwise.
 */
bool NN::Module::isEqual(const Module& inRightModule) const
{
	Beagle_StackTraceBeginM();
	return (mInputSet.isEqual(inRightModule.mInputSet) &&
	        mOutputSet.isEqual(inRightModule.mOutputSet) &&
	        mHiddenSet.isEqual(inRightModule.mHiddenSet));
	Beagle_StackTraceEndM("bool NN::Module::isEqual(const Module& inRightModule) const");
}


/*!
 *  \brief Check if this module is a proper subset of the other.
 *  \param inRightModule Right module to be compared to.
 *  \return True if the input, output and hidden node ids are a proper subset, false otherwise.
 */
bool NN::Module::isLess(const Module& inRightModule) const
{
	Beagle_StackTraceBeginM();
	return (mInputSet.isLess(inRightModule.mInputSet) &&
	        mOutputSet.isLess(inRightModule.mOutputSet) &&
	        mHiddenSet.isLess(inRightModule.mHiddenSet));
	Beagle_StackTraceEndM("bool NN::Module::isLess(const Module& inRightModule) const");
}


/*!
 *  \brief Check if this module is a subset of the other.
 *  \param inRightModule Right module to be compared to.
 *  \return True if the input, output and hidden node ids are a subset, false otherwise.
 */
bool NN::Module::isLessEqual(const Module& inRightModule) const
{
	Beagle_StackTraceBeginM();
	return (mInputSet.isLessEqual(inRightModule.mInputSet) &&
	        mOutputSet.isLessEqual(inRightModule.mOutputSet) &&
	        mHiddenSet.isLessEqual(inRightModule.mHiddenSet));
	Beagle_StackTraceEndM("bool NN::Module::isLessEqual(const Module& inRightModule) const");
}


/*!
 *  \brief Check if this module provides all the source nodes in the given export map.
 *  \param inExportMap Map of source nodes to check.
 *  \return True if the module provides all the source nodes, false otherwise.
 */
bool NN::Module::provides(const MapT<NodeSet>& inExportMap) const
{
	Beagle_StackTraceBeginM();
	bool lProvides = (inExportMap.size() <= mHiddenSet.size() + mOutputSet.size());
	for(MapT<NodeSet>::const_iterator lLeftIter = inExportMap.begin(); lProvides && (lLeftIter != inExportMap.end()); lLeftIter++) {
		lProvides = mHiddenSet.hasKey(lLeftIter->first) || mOutputSet.hasKey(lLeftIter->first);
	}
	return lProvides;
	Beagle_StackTraceEndM("bool NN::Module::provides(const MapT<NodeSet>& inExportMap) const");
}


/*!
 *  \brief Read an NN graph module from an XML subtree.
 *  \param inIter XML iterator to use to read the module from.
 *  \throw Beagle::IOException If reading from the stream failed.
 */
void NN::Module::read(PACC::XML::ConstIterator inIter)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!="Module"))
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Module> expected!");

	std::string lKeyText = inIter->getAttribute("key");
	if(lKeyText.empty()==false)
		mKey = str2uint(lKeyText);
	else
		mKey = UINT_MAX;

	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if(lChild->getType() == PACC::XML::eData) {
			if(lChild->getValue() == "Input") {
				mInputSet.read(lChild);
			} else if(lChild->getValue() == "Output") {
				mOutputSet.read(lChild);
			} else if(lChild->getValue() == "Hidden") {
				mHiddenSet.read(lChild);
			} else if(lChild->getValue() == "Export") {
				mExportMap.read(lChild);
			} else {
				throw Beagle_IOExceptionNodeM(*lChild, "Unexpected tag in graph module!");
			}
		}
	}

	Beagle_StackTraceEndM("void NN::Module::read(PACC::XML::ConstIterator inIter)");
}


/*!
 *  \brief Write a NN graph module into an XML streamer.
 *  \param ioStreamer XML streamer to write the module into.
 *  \param inIndent Whether XML output should be indented.
 */
void NN::Module::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();
	write(ioStreamer, inIndent, UINT_MAX);
	Beagle_StackTraceEndM("void NN::Module::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}


/*!
 *  \brief Write a NN graph module into an XML streamer.
 *  \param ioStreamer XML streamer to write the module into.
 *  \param inIndent Whether XML output should be indented.
 *  \param inKey Key attribute if this module is used as an entry in a map.
 */
void NN::Module::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const
{
	Beagle_StackTraceBeginM();

	ioStreamer.openTag("Module", inIndent);
	if (inKey < UINT_MAX)
		ioStreamer.insertAttribute("key", uint2str(inKey));
	else if (mKey < UINT_MAX)
		ioStreamer.insertAttribute("key", uint2str(mKey));
	mInputSet.write(ioStreamer, false);
	mOutputSet.write(ioStreamer, false);
	mHiddenSet.write(ioStreamer, false);
	mExportMap.write(ioStreamer, false);
	ioStreamer.closeTag();

	Beagle_StackTraceEndM("void NN::Module::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const");
}


/*!
 *  \brief Returns the id of a randomly chosen node from the module.
 *  \param inIncludeInput If true input nodes may also be chosen,
 *         otherwise only output and hidden nodes are chosen.
 */
unsigned int NN::Module::getRandomNode(Beagle::Randomizer& ioRandomizer, bool inIncludeInput) const
{
	Beagle_StackTraceBeginM();

	unsigned int lNumNodes = mOutputSet.size() + mHiddenSet.size();
	if (inIncludeInput)
		lNumNodes += mInputSet.size();

	unsigned int lIndex = ioRandomizer.rollInteger(0, lNumNodes-1);
	const NodeSet *lNodeSet;
	if (lIndex < mOutputSet.size()) {
		lNodeSet = &mOutputSet;
	} else {
		lIndex -= mOutputSet.size();
		if (lIndex < mHiddenSet.size()) {
			lNodeSet = &mHiddenSet;
		} else {
			lIndex -= mHiddenSet.size();
			lNodeSet = &mInputSet;
		}
	}

	NodeSet::const_iterator lIter = lNodeSet->begin();
	for (unsigned int i = 0; i < lIndex; i++) {
		lIter++;
	}

	return *lIter;

	Beagle_StackTraceEndM("unsigned int NN::Module::getRandomNode(Beagle::Randomizer& ioRandomizer, bool inIncludeInput) const");
}
