
/*!
 *  \file   beagle/NN/src/MutationAddEdgeOp.cpp
 *  \brief  Implementation of the type NN::MutationAddEdgeOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct add-edge mutation operator for the NN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
NN::MutationAddEdgeOp::MutationAddEdgeOp(std::string inMutationPbName,
                                         std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the add-edge mutation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::MutationAddEdgeOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Add-edge mutation probability for an individual. ";
		lOSS << "An add-edge mutation adds a new edge between two ";
		lOSS << "randomly selected nodes of the NN genotype graph.";
		Register::Description lDescription(
		    "Individual add-edge mutation prob.",
		    "Float",
		    "0.05",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Whether new edges can be recurrent",
		    "Bool",
		    "0",
		    "Flag indicating whether newly created edges can be recurrent."
		);
		mMutateRecurrent = castHandleT<Bool>(
		                       ioSystem.getRegister().insertEntry("nn.edge.recur", new Bool(false), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial edge weight",
		    "Double",
		    "-10.0",
		    "Edge weight is constrained to be greater than or equal to this value when a new edge is created."
		);
		mMinEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial edge weight",
		    "Double",
		    "10.0",
		    "Edge weight is constrained to be less than this value when a new edge is created."
		);
		mMaxEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmax", new Double(10.0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::MutationAddEdgeOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply add-edge mutation to an NN individual.
 *  \param ioIndividual NN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 *
 *  The new edge is added between two randomly chosen nodes.  Recurrent edges
 *  are formed if creation of recurrent edges is allowed and if the new edge
 *  results in the creation of a cycle.
 */
bool NN::MutationAddEdgeOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationAddEdgeOp",
	    std::string("Mutating individual with NN::MutationAddEdgeOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationAddEdgeOp",
	    std::string("Individual before NN add-edge mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::MutationAddEdgeOp not implemented for individuals of size > 1");

	NN::Graph& lGraph = *lIndividual[0];
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	//   // The source node can be any node in any module.
	//   NN::Node& lSource = lGraph.getRandomNode(lRandomizer);
	//   unsigned int lSourceModuleKey = lSource.getModuleKey();
	//   unsigned int lSourceId = lSource.getIdentifier();

	//   // For deme 0, choose the target node from the main module; for other demes
	//   // choose the target node from the same module as the source.  In both cases
	//   // inputs are excluded.
	//   unsigned int lTargetModuleKey = (ioContext.getDemeIndex() == 0) ? UINT_MAX : lSourceModuleKey;
	//   unsigned int lTargetId = lGraph.getModule(lTargetModuleKey).getRandomNode(lRandomizer, false);

	// The source node can be any node in any module.
	NN::Graph::UIntPair lSourcePair = lGraph.getRandomNode(lRandomizer, true, true);
	unsigned int lSourceModuleKey = lSourcePair.first;
	unsigned int lSourceId = lSourcePair.second;

	// The target node cannot be module inputs.  If the source is an input, the
	// target can only be from the source module; otherwise the target can be from
	// the same module as source or from main.
	bool lIncludeMain = !lGraph.getModule(lSourceModuleKey).getInputSet().hasKey(lSourceId);
	NN::Graph::UIntPair lTargetPair = lGraph.getRandomNode(lRandomizer, lSourceModuleKey, lIncludeMain, false);
	unsigned int lTargetModuleKey = lTargetPair.first;
	unsigned int lTargetId = lTargetPair.second;

	//   // Choose module for mutation based on generation.
	//   GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	//   unsigned int lNumHelpers = lGraphState->getNumberHelpers();
	//   unsigned int lHelperId = lGraphState->getFirstHelper() + ioContext.getGeneration()%(lNumHelpers+1);
	//   unsigned int lSourceModuleKey, lTargetModuleKey;
	//   unsigned int lSourceId, lTargetId;
	//   if (lHelperId > lGraphState->getLastHelper()) {  // Mutation in the main module.
	//     // The source node can be any node in any module.
	//     NN::Node& lSource = lGraph.getRandomNode(lRandomizer);
	//     lSourceModuleKey = lSource.getModuleKey();
	//     lSourceId = lSource.getIdentifier();
	//     lTargetModuleKey = UINT_MAX;
	//   }
	//   else {  // Mutation in the helper module.
	//     /// Choose source node from the helper module.
	//     lSourceModuleKey = lHelperId;
	//     lSourceId = lGraph.getRandomNode(lRandomizer, lSourceModuleKey, false, true).second;
	//     lTargetModuleKey = lHelperId;
	//   }
	//   // Choose a target node, excluding inputs.
	//   lTargetId = lGraph.getModule(lTargetModuleKey).getRandomNode(lRandomizer, false);

	// If making an external connection, restrict source node to the helper node.
	//   if (lTargetModuleKey != lSourceModuleKey) lSourceId = lSourceModuleKey;

	// An edge is created from the chosen source to target if
	//   1. The edge does not already exist and
	//   2. The recurrent flag is set or path does not exist from target to source.
	// The new edge is recurrent if a path already exists from target to source.  So
	// condition 2 ensures that recurrent edges are created only when recurrency is
	// allowed.
	NN::Node& lTarget = lGraph.getNode(lTargetId);
	bool lRecurrentAllowed = mMutateRecurrent->getWrappedValue();
	NN::SearchReachable lSearchReachable;
	bool lRecurrent = false;
	if (!lTarget.hasEdge(lSourceId) &&
	        (!(lRecurrent = lSearchReachable.reachable(lGraph, lSourceId, lTargetId)) || lRecurrentAllowed)) {

		double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
		lTarget.addEdge(NN::Edge(lSourceId, lWeight, lRecurrent));

		// If the target module is different from the source module, add the source
		// and target node ids to the export map of the source module.
		if (lTargetModuleKey != lSourceModuleKey)
			lGraph.getModule(lSourceModuleKey).addExport(lSourceId, lTargetId);

		// Mutation successful, log messages and return.
		std::ostringstream lOSS;
		lOSS << "Added a new edge (recurrency = " << lRecurrent;
		lOSS << ") from node [" << lSourceId << "] (module " << lSourceModuleKey;
		lOSS << ") to node [" << lTargetId << "] (module " << lTargetModuleKey << ")";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddEdgeOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddEdgeOp",
		    std::string("Individual after NN add-edge mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddEdgeOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::MutationAddEdgeOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}
