
/*!
 *  \file   beagle/NN/src/MutationAddModuleOp.cpp
 *  \brief  Implementation of the type NN::MutationAddModuleOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct add module mutation operator for the NN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
NN::MutationAddModuleOp::MutationAddModuleOp(std::string inMutationPbName,
                                             std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the add module mutation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::MutationAddModuleOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Add module mutation probability for an individual. ";
		lOSS << "An add module mutation creates a new module using one of the unused helper nodes.";
		Register::Description lDescription(
		    "Individual add module mutation prob.",
		    "Float",
		    "0.0",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.0f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Minimum value of initial edge weight",
		    "Double",
		    "-10.0",
		    "Edge weight is constrained to be greater than or equal to this value when a new edge is created."
		);
		mMinEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial edge weight",
		    "Double",
		    "10.0",
		    "Edge weight is constrained to be less than this value when a new edge is created."
		);
		mMaxEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node bias",
		    "Double",
		    "-10.0",
		    "Node bias is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node bias",
		    "Double",
		    "10.0",
		    "Node bias is constrained to be less than this value when a new node is created."
		);
		mMaxNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node slope",
		    "Double",
		    "1.0",
		    "Node slope is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemin", new Double(1.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node slope",
		    "Double",
		    "5.0",
		    "Node slope is constrained to be less than this value when a new node is created."
		);
		mMaxNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemax", new Double(5.0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::MutationAddModuleOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply add module mutation to an NN individual.
 *  \param ioIndividual NN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool NN::MutationAddModuleOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationAddModuleOp",
	    std::string("Mutating individual with NN::MutationAddModuleOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationAddModuleOp",
	    std::string("Individual before NN add module mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::MutationAddModuleOp not implemented for individuals of size > 1");

	NN::Graph& lGraph = *lIndividual[0];

	GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	// Select a random helper node to use as the module output.
	unsigned int lHelperId = lRandomizer.rollInteger(lGraphState->getFirstHelper(), lGraphState->getLastHelper());

	// If this helper node is not already in the graph, we can create a new module
	// with the helper node as module output.
	if (lGraph.hasNode(lHelperId) == false) {
		NN::Module& lModule = lGraph.addModule(NN::Module(lHelperId));

		// Create a new node in the graph for the helper and add the node id to the module.
		double lBias = lRandomizer.rollUniform(mMinNodeBias->getWrappedValue(), mMaxNodeBias->getWrappedValue());
		double lSlope = lRandomizer.rollUniform(mMinNodeSlope->getWrappedValue(), mMaxNodeSlope->getWrappedValue());
		NN::Node& lHelper = lGraph.addNode(NN::Node(lHelperId, lBias, lSlope));
		lHelper.setModuleKey(lHelperId);
		lModule.addOutput(lHelperId);

		// The graph inputs and outputs are the module inputs.  Create connections to the
		// module output from the module inputs and add their node ids to the module.
		for (unsigned int i = lGraphState->getFirstInput(); i <= lGraphState->getLastInput(); i++) {
			double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
			lHelper.addEdge(NN::Edge(i, lWeight, false));
			lModule.addInput(i);
		}
		for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++) {
			double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
			lHelper.addEdge(NN::Edge(i, lWeight, false));
			lModule.addInput(i);
		}

		// Mutation successful, log messages and return
		std::ostringstream lOSS;
		lOSS << "Module created with helper node [" << lHelperId << "] as output.";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddModuleOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddModuleOp",
		    std::string("Individual after NN add module mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddModuleOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::MutationAddModuleOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}

