
/*!
 *  \file   beagle/NN/src/MutationAddNodeOp.cpp
 *  \brief  Implementation of the type NN::MutationAddNodeOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct add-node mutation operator for the NN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
NN::MutationAddNodeOp::MutationAddNodeOp(std::string inMutationPbName,
                                         std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the add-node mutation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::MutationAddNodeOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Add-node mutation probability for an individual. ";
		lOSS << "An add-node mutation adds a new node to the genotype graph by spliting ";
		lOSS << "a randomly selected edge, creating that edge if it doesn't exist.";
		Register::Description lDescription(
		    "Individual add-node mutation prob.",
		    "Float",
		    "0.05",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Minimum value of initial edge weight",
		    "Double",
		    "-10.0",
		    "Edge weight is constrained to be greater than or equal to this value when a new edge is created."
		);
		mMinEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial edge weight",
		    "Double",
		    "10.0",
		    "Edge weight is constrained to be less than this value when a new edge is created."
		);
		mMaxEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node bias",
		    "Double",
		    "-10.0",
		    "Node bias is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node bias",
		    "Double",
		    "10.0",
		    "Node bias is constrained to be less than this value when a new node is created."
		);
		mMaxNodeBias = castHandleT<Double>(
		                   ioSystem.getRegister().insertEntry("nn.node.initbiasmax", new Double(10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Minimum value of initial node slope",
		    "Double",
		    "1.0",
		    "Node slope is constrained to be greater than or equal to this value when a new node is created."
		);
		mMinNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemin", new Double(1.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial node slope",
		    "Double",
		    "5.0",
		    "Node slope is constrained to be less than this value when a new node is created."
		);
		mMaxNodeSlope = castHandleT<Double>(
		                    ioSystem.getRegister().insertEntry("nn.node.initslopemax", new Double(5.0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::MutationAddNodeOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply add-node mutation to an NN individual.
 *  \param ioIndividual NN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 *
 *  The new node is added by splitting a randomly chosen edge.
 */
bool NN::MutationAddNodeOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationAddNodeOp",
	    std::string("Mutating individual with NN::MutationAddNodeOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationAddNodeOp",
	    std::string("Individual before NN add-node mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::MutationAddNodeOp not implemented for individuals of size > 1");

	NN::Graph& lGraph = *lIndividual[0];
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	// If an existing edge is randomly chosen, it biases the graphs created to be
	// elongated in the input-output direction.  So we choose the source and target
	// nodes of an edge instead.

	//   // The source node can be any node in any module.
	//   NN::Node& lSource = lGraph.getRandomNode(lRandomizer);
	//   unsigned int lSourceModuleKey = lSource.getModuleKey();
	//   unsigned int lSourceId = lSource.getIdentifier();

	//   // For deme 0, choose the target node from the main module; for other demes
	//   // choose the target node from the same module as the source.  In both cases
	//   // inputs are excluded.
	//   unsigned int lTargetModuleKey = (ioContext.getDemeIndex() == 0) ? UINT_MAX : lSourceModuleKey;
	//   unsigned int lTargetId = lGraph.getModule(lTargetModuleKey).getRandomNode(lRandomizer, false);

	// The source node can be any node in any module.
	NN::Graph::UIntPair lSourcePair = lGraph.getRandomNode(lRandomizer, true, true);
	unsigned int lSourceModuleKey = lSourcePair.first;
	unsigned int lSourceId = lSourcePair.second;

	// The target node cannot be module inputs.  If the source is an input, the
	// target can only be from the source module; otherwise the target can be from
	// the same module as source or from main.
	bool lIncludeMain = !lGraph.getModule(lSourceModuleKey).getInputSet().hasKey(lSourceId);
	NN::Graph::UIntPair lTargetPair = lGraph.getRandomNode(lRandomizer, lSourceModuleKey, lIncludeMain, false);
	unsigned int lTargetModuleKey = lTargetPair.first;
	unsigned int lTargetId = lTargetPair.second;

	//   // Choose module for mutation based on generation.
	//   GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	//   unsigned int lNumHelpers = lGraphState->getNumberHelpers();
	//   unsigned int lHelperId = lGraphState->getFirstHelper() + ioContext.getGeneration()%(lNumHelpers+1);
	//   unsigned int lSourceModuleKey, lTargetModuleKey;
	//   unsigned int lSourceId, lTargetId;
	//   if (lHelperId > lGraphState->getLastHelper()) {  // Mutation in the main module.
	//     // The source node can be any node in any module.
	//     NN::Node& lSource = lGraph.getRandomNode(lRandomizer);
	//     lSourceModuleKey = lSource.getModuleKey();
	//     lSourceId = lSource.getIdentifier();
	//     lTargetModuleKey = UINT_MAX;
	//   }
	//   else {  // Mutation in the helper module.
	//     /// Choose source node from the helper module.
	//     lSourceModuleKey = lHelperId;
	//     lSourceId = lGraph.getRandomNode(lRandomizer, lSourceModuleKey, false, true).second;
	//     lTargetModuleKey = lHelperId;
	//   }
	//   // Choose a target node, excluding inputs.
	//   lTargetId = lGraph.getModule(lTargetModuleKey).getRandomNode(lRandomizer, false);

	// Now, if the chosen edge exists, we can successfully perform the add-node mutation.
	NN::Node& lTarget = lGraph.getNode(lTargetId);
	if (lTarget.hasEdge(lSourceId)) {
		NN::Edge& lEdge = lTarget.getEdge(lSourceId);

		// Create a new node and add it to the genotype graph.
        GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
		unsigned int lNewNodeId = lGraphState->getNewNodeIdentifier(lGraph, lSourceId, lTargetId);
		double lBias = lRandomizer.rollUniform(mMinNodeBias->getWrappedValue(), mMaxNodeBias->getWrappedValue());
		double lSlope = lRandomizer.rollUniform(mMinNodeSlope->getWrappedValue(), mMaxNodeSlope->getWrappedValue());
		NN::Node& lNewNode = lGraph.addNode(NN::Node(lNewNodeId, lBias, lSlope));

		// The new node belongs to the same module as the target node.
		lNewNode.setModuleKey(lTargetModuleKey);
		lGraph.getModule(lTargetModuleKey).addHidden(lNewNodeId);

		// Create a new edge incident on the new node.  This edge is not recurrent because it
		// is the only path to the new node.
		double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
		lNewNode.addEdge(NN::Edge(lSourceId, lWeight, false));

		// Create a new edge leaving the new node.  This edge has the same recur flag as the
		// edge we split.
		lTarget.addEdge(NN::Edge(lNewNodeId, lEdge.getWeight(), lEdge.isRecurrent()));

		// Finally, delete the edge that was split.
		lTarget.delEdge(lEdge);

		// Mutation successful, log messages and return
		std::ostringstream lOSS;
		lOSS << "Added a new node [" << lNewNodeId << "] by splitting edge from node [";
		lOSS << lSourceId << "] to node [" << lTargetId << "]";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddNodeOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddNodeOp",
		    std::string("Individual after NN add-node mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationAddNodeOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::MutationAddNodeOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}
