
/*!
 *  \file   beagle/NN/src/MutationEdgeWeightOp.cpp
 *  \brief  Implementation of the type NN::MutationEdgeWeightOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct an edge-weight mutation operator for the NN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
NN::MutationEdgeWeightOp::MutationEdgeWeightOp(std::string inMutationPbName,
                                               std::string inMutateGaussSigmaName,
                                               std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName),
		mMutateGaussSigmaName(inMutateGaussSigmaName)
{ }


/*!
 *  \brief Register the parameters of the edge-weight mutation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::MutationEdgeWeightOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Edge-weight mutation probability for an individual. ";
		lOSS << "An edge-weight mutation applies Gaussian perturbation ";
		lOSS << "to the weights of edges in an NN genotype graph.";
		Register::Description lDescription(
		    "Individual edge-weight mutation prob.",
		    "Float",
		    "0.05",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	{
		std::ostringstream lOSS;
		lOSS << "Probability that the weight of any single edge in the graph will be mutated.";
		Register::Description lDescription(
		    "Single edge-weight mutation prob.",
		    "Float",
		    "0.05",
		    lOSS.str()
		);
		mSingleMutationProba = castHandleT<Float>(
		                           ioSystem.getRegister().insertEntry("nn.mutedgewt.sngpb", new Float(0.05f), lDescription));
	}
	{
		Register::Description lDescription(
		    "Sigma of Gaussian mutation.",
		    "Float",
		    "1.0",
		    "The standard deviation of Gaussian noise applied to the edge weight during mutation."
		);
		mMutateGaussSigma = castHandleT<Float>(
		                        ioSystem.getRegister().insertEntry(mMutateGaussSigmaName, new Float(1.0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::MutationEdgeWeightOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply edge-weight mutation to an NN individual.
 *  \param ioIndividual NN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 *
 *  The weight of a randomly chosen edge is mutated by applying Gaussian noise.
 *  The sigma of the Gaussian is specified as a parameter of evolution, while
 *  the mean is the current value of the weight.
 */
bool NN::MutationEdgeWeightOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationEdgeWeightOp",
	    std::string("Mutating individual with NN::MutationEdgeWeightOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationEdgeWeightOp",
	    std::string("Individual before NN edge-weight mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::MutationEdgeWeightOp not implemented for individuals of size > 1");

	NN::Graph& lGraph = *lIndividual[0];
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	bool lMutated = false;
	std::ostringstream lOSS;
	lOSS << "Mutated edge weights (source node, target node):";

	//   // Choose edges to mutate.  In deme 0, the edges are chosen from the main
	//   // module, while in the other demes the edges are chosen from the module
	//   // the individuals in the deme contain.
	//   const NN::Module& lModule = (ioContext.getDemeIndex() == 0) ? lGraph.getMain() : lGraph.getModules().begin()->second;
	//   NN::NodeSet lNodeSet(lModule.getOutputSet());
	//   lNodeSet.merge(lModule.getHiddenSet());

	// Get all non-input nodes in the graph.
	NN::NodeSet lNodeSet(lGraph.getMain().getOutputSet());
	lNodeSet.merge(lGraph.getMain().getHiddenSet());
	for (MapT<Module>::const_iterator lIter = lGraph.getModules().begin(); lIter != lGraph.getModules().end(); ++lIter) {
		lNodeSet.merge(lIter->second.getOutputSet());
		lNodeSet.merge(lIter->second.getHiddenSet());
	}

	//   // Choose module to mutate based on generation.
	//   GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	//   unsigned int lNumHelpers = lGraphState->getNumberHelpers();
	//   unsigned int lHelperId = lGraphState->getFirstHelper() + ioContext.getGeneration()%(lNumHelpers+1);
	//   unsigned int lModuleKey = (lHelperId > lGraphState->getLastHelper()) ? UINT_MAX : lHelperId;
	//   const NN::Module& lModule = lGraph.getModule(lModuleKey);
	//   NN::NodeSet lNodeSet(lModule.getOutputSet());
	//   lNodeSet.merge(lModule.getHiddenSet());

	float lSingleMutationProba = mSingleMutationProba->getWrappedValue();
	float lGaussSigma = mMutateGaussSigma->getWrappedValue();
	for (NN::NodeSet::const_iterator lIter = lNodeSet.begin(); lIter != lNodeSet.end(); ++lIter) {
		NN::Node& lNode = lGraph.getNode(*lIter);
		for (NN::Node::iterator lNodeIter = lNode.begin(); lNodeIter != lNode.end(); ++lNodeIter) {
			if (lRandomizer.rollUniform(0.0, 1.0) < lSingleMutationProba) {
				NN::Edge& lEdge = lNodeIter->second;
				double lNewWeight = lRandomizer.rollGaussian(lEdge.getWeight(), lGaussSigma);
				lEdge.setWeight(lNewWeight);

				lOSS << " (" << lEdge.getSourceIdentifier() << ", " << lNode.getIdentifier() << ")";
				lMutated = true;
			}
		}
	}

	if (lMutated) {
		// Mutation successful, log messages and return
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationEdgeWeightOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationEdgeWeightOp",
		    std::string("Individual after NN edge-weight mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationEdgeWeightOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::MutationEdgeWeightOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}
