
/*!
 *  \file   beagle/NN/src/MutationInsertModuleOp.cpp
 *  \brief  Implementation of the type NN::MutationInsertModuleOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct insert module mutation operator for the NN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
NN::MutationInsertModuleOp::MutationInsertModuleOp(std::string inMutationPbName,
                                                   std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName),
		mModSelectionOp(new SelectRandomOp)
{ }


/*!
 *  \brief Register the parameters of the insert module mutation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::MutationInsertModuleOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Insert module mutation probability for an individual. ";
		lOSS << "A module is inserted in the individual, possibly replacing existing ones.";
		Register::Description lDescription(
		    "Individual insert module mutation prob.",
		    "Float",
		    "0.0",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.0f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Minimum value of initial edge weight",
		    "Double",
		    "-10.0",
		    "Edge weight is constrained to be greater than or equal to this value when a new edge is created."
		);
		mMinEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmin", new Double(-10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum value of initial edge weight",
		    "Double",
		    "10.0",
		    "Edge weight is constrained to be less than this value when a new edge is created."
		);
		mMaxEdgeWeight = castHandleT<Double>(
		                     ioSystem.getRegister().insertEntry("nn.edge.initwtmax", new Double(10.0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::MutationInsertModuleOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply insert module mutation to an NN individual.
 *  \param ioIndividual NN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool NN::MutationInsertModuleOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(ioIndividual);

	// This mutation operator is defined only for deme 0 when there is more than
	// one deme.  In other cases, it does nothing.
	if (ioContext.getDemeIndex() > 0 || ioContext.getVivarium().size() == 1) {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationInsertModuleOp",
		    std::string("Mutation not defined; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationInsertModuleOp",
	    std::string("Mutating individual with NN::MutationInsertModuleOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationInsertModuleOp",
	    std::string("Individual before NN insert module mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::MutationInsertModuleOp not implemented for individuals of size > 1");

	NN::Graph& lGraph = *lIndividual[0];
	GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	// Select a deme and an individual from that deme to extract the module.
	Beagle::Vivarium& lVivarium = ioContext.getVivarium();
	unsigned int lModDemeIndex = lRandomizer.rollInteger(1, lVivarium.size()-1);
	Beagle::Deme& lModDeme = *lVivarium[lModDemeIndex];
	unsigned int lModIndivIndex = mModSelectionOp->selectIndividual(lModDeme, ioContext);
	const NN::Individual& lModIndiv = castObjectT<NN::Individual&>(*lModDeme[lModIndivIndex]);
	const NN::Graph& lModGraph = *lModIndiv[0];

	if (insert(lModGraph, lGraph, ioContext)) {
		// Make sure all helper nodes in the graph have outgoing edges to all outputs.
		//     for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++) {
		//       NN::Node& lOutputNode = lGraph.getNode(i);
		//       for (unsigned int j = lGraphState->getFirstHelper(); j <= lGraphState->getLastHelper(); j++) {
		//         if (lGraph.hasNode(j) && !lOutputNode.hasEdge(j)) {
		//           double lWeight = lRandomizer.rollUniform(mMinEdgeWeight->getWrappedValue(), mMaxEdgeWeight->getWrappedValue());
		//           lOutputNode.addEdge(NN::Edge(j, lWeight, true));
		//           lGraph.getModule(lGraph.getNode(j).getModuleKey()).addExport(j, i);
		//         }
		//       }
		//     }

		// Add the individual that is inserted as a module as an endosymbiont.
		//     lIndividual.addEndosymbiont(castHandleT<NN::Individual>(lModDeme[lModIndivIndex]));

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationInsertModuleOp",
		    std::string("Individual after NN insert module mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationInsertModuleOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::MutationInsertModuleOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Insert modules from a one graph to another
 *  \param inGraph1 The graph from which modules are crossed over
 *  \param ioGraph2 The graph to which modules are crossed over
 *  \return True if at least one module was crossed over, false if not
 */
bool NN::MutationInsertModuleOp::insert(const Graph& inGraph1, Graph& ioGraph2, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	std::ostringstream lOSS;
	bool lInserted = false;

	// Insertion of a graph1 module into graph2 is possible only if the graph1 module
	// can satisfy all node references by external edges in the graph2 modules that
	// will be replaced as a result of the insertion.
	const MapT<Module>& lModules1 = inGraph1.getModules();
	for (MapT<Module>::const_iterator lModIter1 = lModules1.begin(); lModIter1 != lModules1.end(); ++lModIter1) {
		const Module& lModule1 = lModIter1->second;
		const NodeSet& lOutSet1 = lModule1.getOutputSet();
		const NodeSet& lHidSet1 = lModule1.getHiddenSet();
		MapT<NodeSet> lExpMap2;
		for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
			if (ioGraph2.hasNode(*lOutIter1)) {
				unsigned int lModKey2 = ioGraph2.getNode(*lOutIter1).getModuleKey();
				const MapT<NodeSet>& lModExp2 = ioGraph2.getModule(lModKey2).getExportMap();
				for (MapT<NodeSet>::const_iterator lExpIter2 = lModExp2.begin(); lExpIter2 != lModExp2.end(); ++lExpIter2) {
					lExpMap2.addData(lExpIter2->first, lExpIter2->second);
				}
			}
		}

		if (lModule1.provides(lExpMap2)) {
			lOSS << "Module [ " << lModule1.getKey() << " ] replaced modules [ ";

			// First remove the graph2 modules that would have to be replaced.
			for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
				if (ioGraph2.hasNode(*lOutIter1)) {
					unsigned int lModKey2 = ioGraph2.getNode(*lOutIter1).getModuleKey();
					const Module& lModule2 = ioGraph2.getModule(lModKey2);
					const NodeSet& lOutSet2 = lModule2.getOutputSet();
					const NodeSet& lHidSet2 = lModule2.getHiddenSet();
					for (NodeSet::const_iterator lOutIter2 = lOutSet2.begin(); lOutIter2 != lOutSet2.end(); ++lOutIter2) {
						ioGraph2.delNode(*lOutIter2);
					}
					for (NodeSet::const_iterator lHidIter2 = lHidSet2.begin(); lHidIter2 != lHidSet2.end(); ++lHidIter2) {
						ioGraph2.delNode(*lHidIter2);
					}
					ioGraph2.delModule(lModule2);

					lOSS << lModKey2 << " ";
				}
			}
			lOSS << "]";

			// Now add the module from graph1 to graph2, and replace the export map
			// of the module with the merged export map of the graph2 modules.
			for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
				ioGraph2.addNode(inGraph1.getNode(*lOutIter1));
			}
			for (NodeSet::const_iterator lHidIter1 = lHidSet1.begin(); lHidIter1 != lHidSet1.end(); ++lHidIter1) {
				ioGraph2.addNode(inGraph1.getNode(*lHidIter1));
			}
			ioGraph2.addModule(lModule1).setExportMap(lExpMap2);

			lInserted = true;
		}
	}

	if (lInserted) {
		// Log message of module insertion and return
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationInsertModuleOp",
		    lOSS.str()
		);

		return true;
	} else {
		return false;
	}

	Beagle_StackTraceEndM("bool NN::MutationInsertModuleOp::insert(const Graph& inGraph1, Graph& ioGraph2, Beagle::Context& ioContext)");
}


/*!
 *  \brief Initialize this operator.
 *  \param ioSystem System of the evolution.
 */
void NN::MutationInsertModuleOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	mModSelectionOp->init(ioSystem);
	Beagle_StackTraceEndM("void NN::MutationInsertModuleOp::init(Beagle::System&)");
}


/*!
 *  \brief Read a mutation operator from XML subtree.
 *  \param inIter XML iterator to use to read crossover operator.
 *  \param ioSystem Evolutionary system.
 */
void NN::MutationInsertModuleOp::readWithSystem(PACC::XML::ConstIterator inIter, Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!=getName())) {
		std::ostringstream lOSS;
		lOSS << "tag <" << getName() << "> expected!" << std::flush;
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}
	std::string lMutationPbReadName = inIter->getAttribute("mutationpb");
	if(lMutationPbReadName.empty() == false) mMutationPbName = lMutationPbReadName;

	if (inIter->isDefined("modselect")) {
		std::string lSelectOpName = inIter->getAttribute("modselect");
        Operator::Alloc::Handle lSelectOpAlloc =
            castHandleT<Operator::Alloc>(ioSystem.getFactory().getAllocator(lSelectOpName));
        if(lSelectOpAlloc == NULL) {
            std::ostringstream lOSS;
            lOSS << "Operator named '" << lSelectOpName << "' is unknown ";
            lOSS << "to the object factory!";
            throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
        }
        mModSelectionOp = castHandleT<SelectionOp>(lSelectOpAlloc->allocate());
        mModSelectionOp->setName(lSelectOpName);
    }

	Beagle_StackTraceEndM("void NN::MutationInsertModuleOp::readWithSystem(PACC::XML::ConstIterator, Beagle::System&)");
}


/*!
 *  \brief Write an operator into an XML stream.
 *  \param ioStreamer XML stream in which the operator is written.
 *  \param inIndent Whether XML output should be indented.
 */
void NN::MutationInsertModuleOp::writeContent(PACC::XML::Streamer &ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();
	Beagle::MutationOp::writeContent(ioStreamer, inIndent);
	if (mModSelectionOp!=NULL) {
		ioStreamer.insertAttribute("modselect", mModSelectionOp->getName());
	}
	Beagle_StackTraceEndM("void NN::MutationInsertModuleOp::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}

