
/*!
 *  \file   beagle/NN/src/MutationMergeModuleOp.cpp
 *  \brief  Implementation of the type NN::MutationMergeModuleOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct merge module mutation operator for the NN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
NN::MutationMergeModuleOp::MutationMergeModuleOp(std::string inMutationPbName,
                                                 std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the merge module mutation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::MutationMergeModuleOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Merge module mutation probability for an individual. ";
		lOSS << "A merge module mutation merges two modules together to form a single module.";
		Register::Description lDescription(
		    "Individual merge module mutation prob.",
		    "Float",
		    "0.0",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.0f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	Beagle_StackTraceEndM("void NN::MutationMergeModuleOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply merge module mutation to an NN individual.
 *  \param ioIndividual NN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool NN::MutationMergeModuleOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationMergeModuleOp",
	    std::string("Mutating individual with NN::MutationMergeModuleOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationMergeModuleOp",
	    std::string("Individual before NN merge module mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::MutationMergeModuleOp not implemented for individuals of size > 1");

	NN::Graph& lGraph = *lIndividual[0];

	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	// Select two random modules to merge.
	const Module& lModule1 = lGraph.getRandomModule(lRandomizer);
	const Module& lModule2 = lGraph.getRandomModule(lRandomizer);
	unsigned int lKey1 = lModule1.getKey();
	unsigned int lKey2 = lModule2.getKey();

	if (lKey1 != lKey2) {
		Module lMrgModule(lModule1, lModule2);
		unsigned int lMrgKey = lMrgModule.getKey();

		// Relabel module nodes with the new module key.
		const NodeSet& lMrgOutSet = lMrgModule.getOutputSet();
		const NodeSet& lMrgHidSet = lMrgModule.getHiddenSet();
		for (NodeSet::const_iterator lOutIter = lMrgOutSet.begin(); lOutIter != lMrgOutSet.end(); ++lOutIter) {
			lGraph.getNode(*lOutIter).setModuleKey(lMrgKey);
		}
		for (NodeSet::const_iterator lHidIter = lMrgHidSet.begin(); lHidIter != lMrgHidSet.end(); ++lHidIter) {
			lGraph.getNode(*lHidIter).setModuleKey(lMrgKey);
		}

		// Remove the separate modules and add the merged module to the graph.
		lGraph.delModule(lModule1);
		lGraph.delModule(lModule2);
		lGraph.addModule(lMrgModule);

		// Mutation successful, log messages and return
		std::ostringstream lOSS;
		lOSS << "Merged modules [ " << lKey1 << " " << lKey2 << " ] into a single module [ " << lMrgKey << " ].";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationMergeModuleOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationMergeModuleOp",
		    std::string("Individual after NN merge module mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationMergeModuleOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::MutationMergeModuleOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}

