
/*!
 *  \file   beagle/NN/src/MutationNodeSlopeOp.cpp
 *  \brief  Implementation of the type NN::MutationNodeSlopeOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct an node-slope mutation operator for the NN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
NN::MutationNodeSlopeOp::MutationNodeSlopeOp(std::string inMutationPbName,
                                             std::string inMutateGaussSigmaName,
                                             std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName),
		mMutateGaussSigmaName(inMutateGaussSigmaName)
{ }


/*!
 *  \brief Register the parameters of the node-slope mutation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::MutationNodeSlopeOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		std::ostringstream lOSS;
		lOSS << "Node-slope mutation probability for an individual. ";
		lOSS << "An node-slope mutation applies Gaussian perturbation ";
		lOSS << "to the slope of a randomly selected node of the NN genotype graph.";
		Register::Description lDescription(
		    "Individual node-slope mutation prob.",
		    "Float",
		    "0.05",
		    lOSS.str()
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	{
		std::ostringstream lOSS;
		lOSS << "Probability that the slope of any single node in the graph will be mutated.";
		Register::Description lDescription(
		    "Single node-slope mutation prob.",
		    "Float",
		    "0.05",
		    lOSS.str()
		);
		mSingleMutationProba = castHandleT<Float>(
		                           ioSystem.getRegister().insertEntry("nn.mutnodeslope.sngpb", new Float(0.05f), lDescription));
	}
	{
		Register::Description lDescription(
		    "Sigma of Gaussian mutation.",
		    "Float",
		    "1.0",
		    "The standard deviation of Gaussian noise applied to the node slope during mutation."
		);
		mMutateGaussSigma = castHandleT<Float>(
		                        ioSystem.getRegister().insertEntry(mMutateGaussSigmaName, new Float(1.0), lDescription));
	}
	Beagle_StackTraceEndM("void NN::MutationNodeSlopeOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply node-slope mutation to an NN individual.
 *  \param ioIndividual NN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 *
 *  The slope of a randomly chosen node is mutated by applying Gaussian noise.
 *  The sigma of the Gaussian is specified as a parameter of evolution, while
 *  the mean is the current value of the slope.
 */
bool NN::MutationNodeSlopeOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	NN::Individual& lIndividual = castObjectT<NN::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationNodeSlopeOp",
	    std::string("Mutating individual with NN::MutationNodeSlopeOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::NN::MutationNodeSlopeOp",
	    std::string("Individual before NN node-slope mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::MutationNodeSlopeOp not implemented for individuals of size > 1");

	NN::Graph& lGraph = *lIndividual[0];
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	bool lMutated = false;
	std::ostringstream lOSS;
	lOSS << "Mutated slope of nodes:";

	//   // Choose non-input nodes to mutate.  In deme 0, the nodes are chosen from
	//   // the main module, while in the other demes the nodes are chosen from the
	//   // module the individuals in the deme contain.
	//   const NN::Module& lModule = (ioContext.getDemeIndex() == 0) ? lGraph.getMain() : lGraph.getModules().begin()->second;
	//   NN::NodeSet lNodeSet(lModule.getOutputSet());
	//   lNodeSet.merge(lModule.getHiddenSet());

	// Get all non-input nodes in the graph.
	NN::NodeSet lNodeSet(lGraph.getMain().getOutputSet());
	lNodeSet.merge(lGraph.getMain().getHiddenSet());
	for (MapT<Module>::const_iterator lIter = lGraph.getModules().begin(); lIter != lGraph.getModules().end(); ++lIter) {
		lNodeSet.merge(lIter->second.getOutputSet());
		lNodeSet.merge(lIter->second.getHiddenSet());
	}

	//   // Choose module to mutate based on generation.
	//   GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	//   unsigned int lNumHelpers = lGraphState->getNumberHelpers();
	//   unsigned int lHelperId = lGraphState->getFirstHelper() + ioContext.getGeneration()%(lNumHelpers+1);
	//   unsigned int lModuleKey = (lHelperId > lGraphState->getLastHelper()) ? UINT_MAX : lHelperId;
	//   const NN::Module& lModule = lGraph.getModule(lModuleKey);
	//   NN::NodeSet lNodeSet(lModule.getOutputSet());
	//   lNodeSet.merge(lModule.getHiddenSet());

	float lSingleMutationProba = mSingleMutationProba->getWrappedValue();
	float lGaussSigma = mMutateGaussSigma->getWrappedValue();
	for (NN::NodeSet::const_iterator lIter = lNodeSet.begin(); lIter != lNodeSet.end(); ++lIter) {
		if (lRandomizer.rollUniform(0.0, 1.0) < lSingleMutationProba) {
			NN::Node& lNode = lGraph.getNode(*lIter);
			double lNewSlope = lRandomizer.rollGaussian(lNode.getSlope(), lGaussSigma);
			lNode.setSlope(lNewSlope);

			lOSS << " " << lNode.getIdentifier();
			lMutated = true;
		}
	}

	if (lMutated) {
		// Mutation successful, log messages and return
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationNodeSlopeOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationNodeSlopeOp",
		    std::string("Individual after NN node-slope mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::NN::MutationNodeSlopeOp",
		    std::string("Mutation attempt failed; ")+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::MutationNodeSlopeOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}
