
/*!
 *  \file   beagle/NN/src/Node.cpp
 *  \brief  Implementation of the type NN::Node.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct an NN graph node.
 *  \param inIdentifier Identifier of the node.
 */
NN::Node::Node(unsigned int inIdentifier, double inBias, double inSlope) :
		NN::MapT<Edge>("Node"),
		mIdentifier(inIdentifier),
		mModuleKey(UINT_MAX),
		mBias(inBias),
		mSlope(inSlope)
{ }


/*!
 *  \brief Compare equality of two graph nodes.
 *  \param inRightNode Node to be compared to.
 *  \return True if the nodes are equal, false otherwise.
 *  \throw Beagle::BadCastException If compared node objects are not of the same type.
 *
 *  Map equality tests only equality of keys.  So two nodes are equal if they
 *  have the same identifier and keys.
 */
bool NN::Node::isEqual(const Node& inRightNode) const
{
	Beagle_StackTraceBeginM();
	return (mIdentifier == inRightNode.mIdentifier) && NN::MapT<Edge>::isEqual(inRightNode);
	Beagle_StackTraceEndM("bool NN::Node::isEqual(const Node& inRightNode) const");
}


/*!
 *  \brief Check if this node is less than the other.
 *  \param inRightNode Node to be compared to.
 *  \return True if this node is less than the right one, false otherwise.
 *  \throw Beagle::BadCastException If compared node objects are not of the same type.
 *
 *  Map comparison tests only if the set of keys in this node is a proper subset
 *  of the set of keys in the right node.  So this node is less than the right 
 *  node if its identifier is less and if the keys are a proper subset.
 */
bool NN::Node::isLess(const Node& inRightNode) const
{
	Beagle_StackTraceBeginM();
	return (mIdentifier < inRightNode.mIdentifier) && NN::MapT<Edge>::isLess(inRightNode);
	Beagle_StackTraceEndM("bool NN::Node::isLess(const Node& inRightNode) const");
}


/*!
 *  \brief Read the attributes of the node from an XML subtree.
 *  \param inIter XML iterator to use to read the attributes from.
 */
void NN::Node::readAttributes(PACC::XML::ConstIterator inIter)
{
	Beagle_StackTraceBeginM();
	std::string lIdentifierText = inIter->getAttribute("key");
	if(lIdentifierText.empty()==false)
		mIdentifier = str2uint(lIdentifierText);
	std::string lModuleText = inIter->getAttribute("module");
	if(lModuleText.empty()==false)
		mModuleKey = str2uint(lModuleText);
	else
		mModuleKey = UINT_MAX;
	std::string lBiasText = inIter->getAttribute("bias");
	if(lBiasText.empty()==false)
		mBias = str2dbl(lBiasText);
	std::string lSlopeText = inIter->getAttribute("slope");
	if(lSlopeText.empty()==false)
		mSlope = str2dbl(lSlopeText);
	Beagle_StackTraceEndM("void NN::Node::readAttributes(PACC::XML::ConstIterator inIter)");
}


/*!
 *  \brief Write the attributes of the node into the XML streamer.
 *  \param ioStreamer XML streamer to write the attributes to.
 *  \param inIndent Whether XML output should be indented.
 *  \param inKey Key attribute if this node is used as an entry in a map.
 */
void NN::Node::writeAttributes(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const
{
	Beagle_StackTraceBeginM();
	if (inKey < UINT_MAX)
		ioStreamer.insertAttribute("key", uint2str(inKey));
	else
		ioStreamer.insertAttribute("key", uint2str(mIdentifier));
	if (mModuleKey < UINT_MAX)
		ioStreamer.insertAttribute("module", uint2str(mModuleKey));
	ioStreamer.insertAttribute("bias", dbl2str(mBias));
	ioStreamer.insertAttribute("slope", dbl2str(mSlope));
	Beagle_StackTraceEndM("void NN::Node::writeAttributes(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const");
}


