
/*!
 *  \file   beagle/NN/src/NodeSet.cpp
 *  \brief  Implementation of the type NN::NodeSet.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct a set of NN graph node ids.
 *  \param inType The type of nodes stored in this set.
 */
NN::NodeSet::NodeSet(const std::string& inName) :
		Beagle::NamedObject(inName)
{ }


/*!
 *  \brief Compare equality of two sets of graph node ids.
 *  \param inRightObj NodeSet to be compared to.
 *  \return True if the node sets are equal, false otherwise.
 *  \throw Beagle::BadCastException If compared node objects are not of the same type.
 */
bool NN::NodeSet::isEqual(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const NN::NodeSet& lRightNodeSet = castObjectT<const NN::NodeSet&>(inRightObj);
	bool lEqual = (size() == lRightNodeSet.size());
	for(const_iterator lLeftIter = begin(); lEqual && (lLeftIter != end()); lLeftIter++) {
		lEqual = lRightNodeSet.hasKey(*lLeftIter);
	}
	return lEqual;
	Beagle_StackTraceEndM("bool NN::NodeSet::isEqual(const Object& inRightObj) const");
}


/*!
 *  \brief Check if this node set is a proper subset of the other.
 *  \param inRightObj NodeSet to be compared to.
 *  \return True if this node set is a proper subset of the right one, false otherwise.
 *  \throw Beagle::BadCastException If compared node objects are not of the same type.
 */
bool NN::NodeSet::isLess(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const NN::NodeSet& lRightNodeSet = castObjectT<const NN::NodeSet&>(inRightObj);
	bool lLess = (size() < lRightNodeSet.size());
	for(const_iterator lLeftIter = begin(); lLess && (lLeftIter != end()); lLeftIter++) {
		lLess = lRightNodeSet.hasKey(*lLeftIter);
	}
	return lLess;
	Beagle_StackTraceEndM("bool NN::NodeSet::isLess(const Object& inRightObj) const");
}


/*!
 *  \brief Check if this node set is a subset of the other.
 *  \param inRightNodeSet NodeSet to be compared to.
 *  \return True if this node set is a subset of the right one, false otherwise.
 *  \throw Beagle::BadCastException If compared node objects are not of the same type.
 */
bool NN::NodeSet::isLessEqual(const NodeSet& inRightNodeSet) const
{
	Beagle_StackTraceBeginM();
	bool lLessEqual = (size() <= inRightNodeSet.size());
	for(const_iterator lLeftIter = begin(); lLessEqual && (lLeftIter != end()); lLeftIter++) {
		lLessEqual = inRightNodeSet.hasKey(*lLeftIter);
	}
	return lLessEqual;
	Beagle_StackTraceEndM("bool NN::NodeSet::isLessEqual(const NodeSet& inRightNodeSet) const");
}


/*!
 *  \return The minimum of the node ids in the node set.
 */
unsigned int NN::NodeSet::getMin() const
{
	Beagle_StackTraceBeginM();
	unsigned int lMin = UINT_MAX;
	for(const_iterator lIter = begin(); lIter != end(); lIter++) {
		if (*lIter < lMin)
			lMin = *lIter;
	}
	return lMin;
	Beagle_StackTraceEndM("unsigned int NN::NodeSet::getMin() const");
}


/*!
 *  \brief Read an NN graph node set from an XML subtree.
 *  \param inIter XML iterator to use to read the node set from.
 *  \throw Beagle::IOException If reading from the stream failed.
 */
void NN::NodeSet::read(PACC::XML::ConstIterator inIter)
{
	Beagle_StackTraceBeginM();

	if((inIter->getType() != PACC::XML::eData) || (inIter->getValue() != getName())) {
		std::ostringstream lOSS;
		lOSS << "tag <" << getName() << "> expected!";
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}

	clear();

	// Get size attribute and resize set for performance.
#if defined(BEAGLE_HAVE_STD_HASHSET) | defined(BEAGLE_HAVE_GNUCXX_HASHSET) | defined(BEAGLE_HAVE_STDEXT_HASHSET)
	std::string lSizeText = inIter->getAttribute("size");
	if(lSizeText.empty()==false)
		resize(str2uint(lSizeText));
#endif

	PACC::XML::ConstIterator lChild = inIter->getFirstChild();
	if (lChild) {
		if(lChild->getType()!=PACC::XML::eString)
			throw Beagle_IOExceptionNodeM(*lChild, "expected string to read NodeSet!");

		std::istringstream lISS(lChild->getValue());
		while(lISS.good()) {
			unsigned int lValue;
			lISS >> lValue;
			insert(lValue);
			if (lISS.good())
				lISS.get();
		}
	}

	Beagle_StackTraceEndM("void NN::NodeSet::read(PACC::XML::ConstIterator inIter)");
}


/*!
 *  \brief Write a NN graph node set into an XML streamer.
 *  \param ioStreamer XML streamer to write the node set into.
 *  \param inIndent Whether XML output should be indented.
 */
void NN::NodeSet::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();
	write(ioStreamer, inIndent, UINT_MAX);
	Beagle_StackTraceEndM("void NN::NodeSet::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}


/*!
 *  \brief Write a NN graph node set into an XML streamer.
 *  \param ioStreamer XML streamer to write the node set into.
 *  \param inIndent Whether XML output should be indented.
 *  \param inKey Key attribute to be used if this set is used as an entry in a map.
 */
void NN::NodeSet::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const
{
	Beagle_StackTraceBeginM();

	ioStreamer.openTag(getName(), inIndent);
	if (inKey < UINT_MAX)
		ioStreamer.insertAttribute("key", uint2str(inKey));
	ioStreamer.insertAttribute("size", uint2str(size()));
	std::ostringstream lOSS;
	const_iterator lIter = begin();
	while (lIter != end()) {
		lOSS << *lIter;
		++lIter;
		if (lIter != end())
			lOSS << ',';
	}
	ioStreamer.insertStringContent(lOSS.str());
	ioStreamer.closeTag();

	Beagle_StackTraceEndM("void NN::NodeSet::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inKey) const");
}


/*!
 *  \brief Merge the elements of the given node set with this node set.
 *  \param inNodeSet The node set to merge.
 */
void NN::NodeSet::merge(const NodeSet& inNodeSet)
{
	Beagle_StackTraceBeginM();
	for(const_iterator lIter = inNodeSet.begin(); lIter != inNodeSet.end(); lIter++) {
		insert(*lIter);
	}
	Beagle_StackTraceEndM("void NN::NodeSet::merge(const NodeSet& inNodeSet)");
}
