
/*!
 *  \file   beagle/NN/src/PackageBase.cpp
 *  \brief  Source code of class NN::PackageBase.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/Beagle.hpp"

using namespace Beagle;


/*!
 *  \brief Construct package with basic NN objects.
 */
NN::PackageBase::PackageBase() :
		Beagle::Package("NN-PackageBase"),
		mGraphState(new NN::GraphState)
{ }


/*!
 *  \brief Configure system by adding necessary components into it.
 *  \param ioSystem System to configure.
 */
void NN::PackageBase::configure(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Factory& lFactory = ioSystem.getFactory();

	// Add available basic types to the factory
	lFactory.insertAllocator("Beagle::NN::ChangeFitnessIndexOp", new NN::ChangeFitnessIndexOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::CrossoverModuleOp", new NN::CrossoverModuleOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::FitnessIndexed", new NN::FitnessIndexed::Alloc);
	lFactory.insertAllocator("Beagle::NN::GenerationalOp", new NN::GenerationalOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::Graph", new NN::Graph::Alloc);
	lFactory.insertAllocator("Beagle::NN::Individual", new NN::Individual::Alloc);
	lFactory.insertAllocator("Beagle::NN::InitControllerOp", new NN::InitControllerOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::InitFullyConnectedOp", new NN::InitFullyConnectedOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::MutationAddEdgeOp", new NN::MutationAddEdgeOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::MutationAddModuleOp", new NN::MutationAddModuleOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::MutationAddNodeOp", new NN::MutationAddNodeOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::MutationEdgeWeightOp", new NN::MutationEdgeWeightOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::MutationInsertModuleOp", new NN::MutationInsertModuleOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::MutationMergeModuleOp", new NN::MutationMergeModuleOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::MutationNodeBiasOp", new NN::MutationNodeBiasOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::MutationNodeSlopeOp", new NN::MutationNodeSlopeOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::PackageBase", new NN::PackageBase::Alloc);
	lFactory.insertAllocator("Beagle::NN::ReplacementModuleOp", new NN::ReplacementModuleOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::StatsCalcFitnessIndexedOp", new NN::StatsCalcFitnessIndexedOp::Alloc);
	lFactory.insertAllocator("Beagle::NN::TermMaxFitnessOp", new NN::TermMaxFitnessOp::Alloc);

	// Set aliases
	lFactory.aliasAllocator("Beagle::NN::ChangeFitnessIndexOp", "NN-ChangeFitnessIndexOp");
	lFactory.aliasAllocator("Beagle::NN::CrossoverModuleOp", "NN-CrossoverModuleOp");
	lFactory.aliasAllocator("Beagle::NN::FitnessIndexed", "NN-FitnessIndexed");
	lFactory.aliasAllocator("Beagle::NN::GenerationalOp", "NN-GenerationalOp");
	lFactory.aliasAllocator("Beagle::NN::Graph", "NN-Graph");
	lFactory.aliasAllocator("Beagle::NN::Individual", "NN-Individual");
	lFactory.aliasAllocator("Beagle::NN::InitControllerOp", "NN-InitControllerOp");
	lFactory.aliasAllocator("Beagle::NN::InitFullyConnectedOp", "NN-InitFullyConnectedOp");
	lFactory.aliasAllocator("Beagle::NN::MutationAddEdgeOp", "NN-MutationAddEdgeOp");
	lFactory.aliasAllocator("Beagle::NN::MutationAddModuleOp", "NN-MutationAddModuleOp");
	lFactory.aliasAllocator("Beagle::NN::MutationAddNodeOp", "NN-MutationAddNodeOp");
	lFactory.aliasAllocator("Beagle::NN::MutationEdgeWeightOp", "NN-MutationEdgeWeightOp");
	lFactory.aliasAllocator("Beagle::NN::MutationInsertModuleOp", "NN-MutationInsertModuleOp");
	lFactory.aliasAllocator("Beagle::NN::MutationMergeModuleOp", "NN-MutationMergeModuleOp");
	lFactory.aliasAllocator("Beagle::NN::MutationNodeBiasOp", "NN-MutationNodeBiasOp");
	lFactory.aliasAllocator("Beagle::NN::MutationNodeSlopeOp", "NN-MutationNodeSlopeOp");
	lFactory.aliasAllocator("Beagle::NN::PackageBase", "NN-PackageBase");
	lFactory.aliasAllocator("Beagle::NN::ReplacementModuleOp", "NN-ReplacementModuleOp");
	lFactory.aliasAllocator("Beagle::NN::StatsCalcFitnessIndexedOp", "NN-StatsCalcFitnessIndexedOp");
	lFactory.aliasAllocator("Beagle::NN::TermMaxFitnessOp", "NN-TermMaxFitnessOp");

	// Set some basic concept-type associations in factory
	lFactory.setConcept("CrossoverOp", "NN-CrossoverModuleOp");
	lFactory.setConcept("Fitness", "NN-FitnessIndexed");
	lFactory.setConcept("Genotype", "NN-Graph");
	lFactory.setConcept("Individual", "NN-Individual");
	lFactory.setConcept("InitializationOp", "NN-InitFullyConnectedOp");
	lFactory.setConcept("MutationOp", "NN-MutationEdgeWeightOp");
	lFactory.setConcept("StatsCalculateOp", "NN-StatsCalcFitnessIndexedOp");

	// Add NN components
	ioSystem.addComponent(mGraphState);

	// Set some default parameters
	ioSystem.getRegister().setEntryDefaultValue("ec.pop.size",
	        new UIntArray(1,500),
	        "500");
	ioSystem.getRegister().setEntryDefaultValue("ec.sel.tournsize",
	        new UInt(2),
	        "2");

	Beagle_StackTraceEndM("void NN::PackageBase::configure(System&)");
}


/*!
 *  \brief List packages on which current package depends.
 *  \return Bag with the packages instanciation on which current package depends.
 *
 *  NN::PackageBase depends on Beagle::PackageBase.
 */
Beagle::Package::Bag NN::PackageBase::listDependencies(void)
{
	Beagle_StackTraceBeginM();
	Package::Bag lDependencies;
	lDependencies.push_back(new Beagle::PackageBase());
	return lDependencies;
	Beagle_StackTraceEndM("Beagle::Package::Bag NN::PackageBase::listDependencies(void)");
}
