
/*!
 *  \file   beagle/NN/src/ReplacementModuleOp.cpp
 *  \brief  Source code of class ReplacementModuleOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Build generational module operator.
 *  \param inName Name of the generational module operator.
 */
NN::ReplacementModuleOp::ReplacementModuleOp(std::string inName) :
		ReplacementStrategyOp(inName)
{ }


/*!
 *  \brief Register the parameters of this operator.
 *  \param ioSystem Reference to the evolutionary system.
 */
void NN::ReplacementModuleOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	ReplacementStrategyOp::registerParams(ioSystem);
	Register::Description lDescription(
	    "Elitism keep size for modules",
	    "UInt",
	    "1",
	    "Number of individuals with modules to keep as is with strong n-elitism."
	);
	mElitismKeepSize = castHandleT<UInt>(
	                       ioSystem.getRegister().insertEntry("nn.elite.modkeepsize", new UInt(1), lDescription));
	Beagle_StackTraceEndM("void NN::ReplacementModuleOp::registerParams(System&)");
}


/*!
 *  \brief Apply the operation on a deme in the given context.
 *  \param ioDeme Reference to the deme on which the operation takes place.
 *  \param ioContext Evolutionary context of the operation.
 */
void NN::ReplacementModuleOp::operate(Beagle::Deme& ioDeme, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	Beagle_NonNullPointerAssertM(getRootNode());
	Beagle_NonNullPointerAssertM(mElitismKeepSize);
	Beagle_ValidateParameterM(mElitismKeepSize->getWrappedValue() <= ioDeme.size(),
	                          "nn.elite.modkeepsize",
	                          "The module elistism keepsize must be less than the deme size!");

	Beagle_LogTraceM(
	    ioContext.getSystem().getLogger(),
	    "replacement-strategy", "Beagle::ReplacementModuleOp",
	    std::string("Processing using generational module replacement strategy the ")+
	    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
	);
	Beagle_LogObjectM(
	    ioContext.getSystem().getLogger(),
	    Logger::eTrace,
	    "replacement-strategy", "Beagle::ReplacementModuleOp",
	    (*this)
	);

	History::Handle lHistory = castHandleT<History>(ioContext.getSystem().haveComponent("History"));

	RouletteT<unsigned int> lRoulette;
	buildRoulette(lRoulette, ioContext);

	// Iterate through all the helper nodes and perform module replacement in
	// individuals independent of the other helper nodes.  In each iteration,
	// the fitness index is set corresponding to that helper node.
	GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	std::vector<unsigned int> lHelpers;
	for (unsigned int k = lGraphState->getFirstHelper(); k <= lGraphState->getLastHelper(); k++) {
		lHelpers.push_back(k);
	}
	random_shuffle(lHelpers.begin(), lHelpers.end(), ioContext.getSystem().getRandomizer());
	for (unsigned int p = 0; p < lHelpers.size(); p++) {
		unsigned int lHelperId = lHelpers[p];
		FitnessIndexed::setIndex(lHelperId+1);

		Beagle::Individual::Bag lOffsprings;
        const Factory& lFactory = ioContext.getSystem().getFactory();
		if(mElitismKeepSize->getWrappedValue() > 0) {
			std::make_heap(ioDeme.begin(), ioDeme.end(), IsLessPointerPredicate());
			for(unsigned int i=0; i<mElitismKeepSize->getWrappedValue(); ++i) {
                std::string lIndividualType = ioDeme[0]->getType();
                Beagle::Individual::Alloc::Handle lIndividualAlloc =
                    castHandleT<Beagle::Individual::Alloc>(lFactory.getAllocator(lIndividualType));
                Beagle::Individual::Handle lEliteIndiv = castHandleT<Beagle::Individual>(lIndividualAlloc->allocate());
                lEliteIndiv->copy(*ioDeme[0], ioContext.getSystem());
				lOffsprings.push_back(lEliteIndiv);
				if (lHistory != NULL) {
                    std::vector<HistoryID> lParent;
                    HistoryID::Handle lHID = castHandleT<HistoryID>(ioDeme[0]->getMember("HistoryID"));
                    if(lHID != NULL) lParent.push_back(*lHID);
					lHistory->incrementHistoryVar(*lEliteIndiv);
					lHistory->trace(ioContext, lParent, lEliteIndiv, getName(), "elitism");
				}
				std::pop_heap(ioDeme.begin(), (ioDeme.end()-i), IsLessPointerPredicate());
			}
		}

		random_shuffle(ioDeme.begin(),
		               ioDeme.end()-mElitismKeepSize->getWrappedValue(),
		               ioContext.getSystem().getRandomizer());

		for(unsigned int i=0; i<(ioDeme.size()-mElitismKeepSize->getWrappedValue()); ++i) {
			unsigned int lIndexBreeder = lRoulette.select(ioContext.getSystem().getRandomizer());
			BreederNode::Handle lSelectedBreeder=getRootNode();
			for(unsigned int j=0; j<lIndexBreeder; ++j)
				lSelectedBreeder=lSelectedBreeder->getNextSibling();
			Beagle_NonNullPointerAssertM(lSelectedBreeder);
			Beagle_NonNullPointerAssertM(lSelectedBreeder->getBreederOp());
			Beagle::Individual::Handle lBredIndiv =
			    lSelectedBreeder->getBreederOp()->breed(ioDeme, lSelectedBreeder->getFirstChild(), ioContext);
			Beagle_NonNullPointerAssertM(lBredIndiv);
            std::string lIndividualType = ioDeme[i]->getType();
            Beagle::Individual::Alloc::Handle lIndividualAlloc =
                castHandleT<Beagle::Individual::Alloc>(lFactory.getAllocator(lIndividualType));
            Beagle::Individual::Handle lDemeIndiv = castHandleT<Beagle::Individual>(lIndividualAlloc->allocate());
            lDemeIndiv->copy(*ioDeme[i], ioContext.getSystem());

			bool lReplaced = replace(*lBredIndiv, *lDemeIndiv, lHelperId, ioContext);
			if(lReplaced && (lDemeIndiv->getFitness() != NULL)) {
				lDemeIndiv->getFitness()->setInvalid();
				if (lHistory != NULL) {
                    std::vector<HistoryID> lParents;
                    HistoryID::Handle lHID1 = castHandleT<HistoryID>(ioDeme[i]->getMember("HistoryID"));
                    if(lHID1 != NULL) lParents.push_back(*lHID1);
                    HistoryID::Handle lHID2 = castHandleT<HistoryID>(lBredIndiv->getMember("HistoryID"));
                    if(lHID2 != NULL) lParents.push_back(*lHID2);
					lHistory->incrementHistoryVar(*lDemeIndiv);
					lHistory->trace(ioContext, lParents, lDemeIndiv, getName(), "replacement");
				}
			}
			lOffsprings.push_back(lDemeIndiv);
		}

		for(unsigned int j=0; j<lOffsprings.size(); ++j)
			ioDeme[j] = lOffsprings[j];
	}

	// Set the fitness index back to zero before we return.
	FitnessIndexed::setIndex(0);

	Beagle_StackTraceEndM("void NN::ReplacementModuleOp::operate(Deme& ioDeme, Context& ioContext)");
}


/*!
 *  \brief Replace module in one individual with a module from another
 *  \param inIndiv1 The individual from which module is taken
 *  \param ioIndiv2 The individual whose module is replaced
 *  \return True if the module was replaced, false if not
 */
bool NN::ReplacementModuleOp::replace(const Beagle::Individual& inIndiv1, Beagle::Individual& ioIndiv2,
                                      unsigned int inHelperId, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	const NN::Individual& lIndiv1 = castObjectT<const NN::Individual&>(inIndiv1);
	NN::Individual& lIndiv2 = castObjectT<NN::Individual&>(ioIndiv2);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "replacement", "Beagle::NN::ReplacementModuleOp",
	    std::string("The individual before module replacement: ")+
	    lIndiv2.serialize()
	);

	if (lIndiv1.size() > 1 || lIndiv2.size() > 1)
		throw Beagle_RunTimeExceptionM("NN::ReplacementModuleOp not implemented for individuals of size > 1");

	const Graph& lGraph1 = *lIndiv1[0];
	Graph& lGraph2 = *lIndiv2[0];

	bool lReplaced = false;
	std::ostringstream lOSS;
	if (lGraph1.hasNode(inHelperId) && lGraph2.hasNode(inHelperId)) {
		// Insertion of a graph1 module into graph2 is possible only if the graph1 module
		// can satisfy all node references by external edges in the graph2 modules that
		// will be replaced as a result of the insertion.
		unsigned int lModKey1 = lGraph1.getNode(inHelperId).getModuleKey();
		const Module& lModule1 = lGraph1.getModule(lModKey1);
		const NodeSet& lOutSet1 = lModule1.getOutputSet();
		const NodeSet& lHidSet1 = lModule1.getHiddenSet();
		MapT<NodeSet> lExpMap2;
		for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
			if (lGraph2.hasNode(*lOutIter1)) {
				unsigned int lModKey2 = lGraph2.getNode(*lOutIter1).getModuleKey();
				const MapT<NodeSet>& lModExp2 = lGraph2.getModule(lModKey2).getExportMap();
				for (MapT<NodeSet>::const_iterator lExpIter2 = lModExp2.begin(); lExpIter2 != lModExp2.end(); ++lExpIter2) {
					lExpMap2.addData(lExpIter2->first, lExpIter2->second);
				}
			}
		}

		if (lModule1.provides(lExpMap2)) {
			lOSS << "Module [ " << lModule1.getKey() << " ] replaced modules [ ";

			// First remove the graph2 modules being replaced.
			for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
				if (lGraph2.hasNode(*lOutIter1)) {
					unsigned int lModKey2 = lGraph2.getNode(*lOutIter1).getModuleKey();
					const Module& lModule2 = lGraph2.getModule(lModKey2);
					const NodeSet& lOutSet2 = lModule2.getOutputSet();
					const NodeSet& lHidSet2 = lModule2.getHiddenSet();
					for (NodeSet::const_iterator lOutIter2 = lOutSet2.begin(); lOutIter2 != lOutSet2.end(); ++lOutIter2) {
						lGraph2.delNode(*lOutIter2);
					}
					for (NodeSet::const_iterator lHidIter2 = lHidSet2.begin(); lHidIter2 != lHidSet2.end(); ++lHidIter2) {
						lGraph2.delNode(*lHidIter2);
					}
					lGraph2.delModule(lModule2);

					lOSS << lModKey2 << " ";
				}
			}
			lOSS << "]";

			// Now add the module from graph1 to graph2, and replace the export map
			// of the module with the merged export map of the graph2 modules.
			for (NodeSet::const_iterator lOutIter1 = lOutSet1.begin(); lOutIter1 != lOutSet1.end(); ++lOutIter1) {
				lGraph2.addNode(lGraph1.getNode(*lOutIter1));
			}
			for (NodeSet::const_iterator lHidIter1 = lHidSet1.begin(); lHidIter1 != lHidSet1.end(); ++lHidIter1) {
				lGraph2.addNode(lGraph1.getNode(*lHidIter1));
			}
			lGraph2.addModule(lModule1).setExportMap(lExpMap2);

			lReplaced = true;
		}
	}

	if (lReplaced) {
		// Log message of module replacement and return
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "replacement", "Beagle::NN::ReplacementModuleOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "replacement", "Beagle::NN::ReplacementModuleOp",
		    std::string("The individual after module replacement: ")+
		    lIndiv2.serialize()
		);

		return true;
	} else {
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "replacement", "Beagle::NN::ReplacementModuleOp",
		    std::string("Replacement attempt failed; ")+
		    std::string("resuming from replacement without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool NN::ReplacementModuleOp::replace(const Beagle::Individual& inIndiv1, Beagle::Individual& ioIndiv2, unsigned int inHelperId, Beagle::Context& ioContext)");
}

