
/*!
 *  \file   beagle/NN/src/SearchReachable.cpp
 *  \brief  Implementation of the type NN::SearchReachable.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Determine the nodes from which a set of target nodes are reachable.
 *  \param ioGraph The graph to search in.
 *  \param inTargetSet The ids of the target set of nodes.
 *  \return A handle to the set of nodes from which the target node set is reachable.
 */
NN::NodeSet::Handle NN::SearchReachable::reachable(Graph& ioGraph, const NodeSet& inTargetSet)
{
	Beagle_StackTraceBeginM();

	mContinue = true;
	mSourceId = UINT_MAX;
	mReachableSet = new NN::NodeSet;

	search(ioGraph, inTargetSet);

	return mReachableSet;

	Beagle_StackTraceEndM("NN::NodeSet::Handle NN::SearchReachable::reachable(NN::Graph& ioGraph, unsigned int inTargetId)");
}


/*!
 *  \brief Determine the nodes from which a target node are reachable.
 *  \param ioGraph The graph to search in.
 *  \param inTargetId The target node id.
 *  \return A handle to the set of nodes from which the target node is reachable.
 */
NN::NodeSet::Handle NN::SearchReachable::reachable(Graph& ioGraph, unsigned int inTargetId)
{
	Beagle_StackTraceBeginM();

	mContinue = true;
	mSourceId = UINT_MAX;
	mReachableSet = new NN::NodeSet;

	NodeSet lTargetSet;
	lTargetSet.insert(inTargetId);
	search(ioGraph, lTargetSet);

	return mReachableSet;

	Beagle_StackTraceEndM("NN::NodeSet::Handle NN::SearchReachable::reachable(NN::Graph& ioGraph, unsigned int inTargetId)");
}


/*!
 *  \brief Determine if the target node is reachable from the source node.
 *  \param ioGraph The graph to search in.
 *  \param inTargetId The target node id.
 *  \param inSourceId The source node id.
 *  \return True if a path exists from the source to the target, false otherwise.
 */
bool NN::SearchReachable::reachable(Graph& ioGraph, unsigned int inTargetId, unsigned int inSourceId)
{
	Beagle_StackTraceBeginM();

	// First make sure the source node is in the graph.
	if (!ioGraph.hasNode(inSourceId)) {
		std::ostringstream lOSS;
		lOSS << "Invalid source node [" << inSourceId << "] for checking reachability!";
		throw Beagle_RunTimeExceptionM(lOSS.str());
	}

	mContinue = true;
	mSourceId = inSourceId;
	mReachableSet = NULL;

	NodeSet lTargetSet;
	lTargetSet.insert(inTargetId);
	search(ioGraph, lTargetSet);

	// If the source node is found during search, the continue flag would be set to false.
	return !mContinue;

	Beagle_StackTraceEndM("bool NN::SearchReachable::reachable(Graph& ioGraph, unsigned int inTargetId, unsigned int inSourceId)");
}
