
/*!
 *  \file   beagle/NN/src/StatsCalcFitnessIndexedOp.cpp
 *  \brief  Source code of class StatsCalcFitnessIndexedOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

#include <cmath>
#include <sstream>

using namespace Beagle;


/*!
 *  \brief Construct a calculate stats operator.
 *  \param inName Name of the operator.
 */
NN::StatsCalcFitnessIndexedOp::StatsCalcFitnessIndexedOp(std::string inName) :
		StatsCalculateOp(inName)
{ }


/*!
 *  \brief Calculate statistics of a given deme.
 *  \param outStats Evaluated statistics.
 *  \param ioDeme Deme to evalute the statistics.
 *  \param ioContext Context of the evolution.
 */
void NN::StatsCalcFitnessIndexedOp::calculateStatsDeme(Beagle::Stats& outStats,
        Beagle::Deme& ioDeme,
        Beagle::Context& ioContext) const
{
	Beagle_StackTraceBeginM();

	outStats.clear();
	outStats.clearItems();
	outStats.addItem("processed", ioContext.getProcessedDeme());
	outStats.addItem("total-processed", ioContext.getTotalProcessedDeme());

	if(ioDeme.size() == 0) {
		outStats.clear();
		outStats.setGenerationValues(std::string("deme")+uint2str(ioContext.getDemeIndex()+1),
		                             ioContext.getGeneration(), 0, true);
		return;
	}

	outStats.clear();
	outStats.setGenerationValues(std::string("deme")+uint2str(ioContext.getDemeIndex()+1),
	                             ioContext.getGeneration(), ioDeme.size(), true);

	const unsigned int lNumValues = castHandleT<FitnessIndexed>(ioDeme[0]->getFitness())->size();
	for(unsigned int j=0; j<lNumValues; ++j) {
		unsigned int lCount = 0;
		double lSum = 0.0;
		double lPow2Sum = 0.0;
		double lMax = -DBL_MAX;
		double lMin = DBL_MAX;

		for(unsigned int i=0; i<ioDeme.size(); i++) {
			FitnessIndexed::Handle lFitness = castHandleT<FitnessIndexed>(ioDeme[i]->getFitness());
			Beagle_AssertM(lNumValues == lFitness->size());
			const float lValue = lFitness->getValue(j);
			if (lValue > -FLT_MAX) {
				lCount++;
				lSum     += lValue;
				lPow2Sum += pow2Of<double>(lValue);
				lMax     =  maxOf<double>(lMax,lValue);
				lMin     =  minOf<double>(lMin,lValue);
			}
		}

		if(lCount == 0) {
			continue;
		} else if(lCount == 1) {
			outStats.push_back(Measure(std::string("value")+uint2str(j),
			                           lSum, 0.0, lMax, lMin));
		} else {
			double lAverage  = lSum / lCount;
			double lStdError = (lPow2Sum - (pow2Of<double>(lSum)/lCount)) / (lCount-1);
			lStdError        = sqrt(lStdError);
			outStats.push_back(Measure(std::string("value")+uint2str(j),
			                           lAverage, lStdError, lMax, lMin));
		}
	}

	Beagle_StackTraceEndM("void NN::StatsCalcFitnessIndexedOp::calculateStatsDeme(Stats& outStats, Deme& ioDeme, Context& ioContext) const");
}


/*!
 *  \brief Calculate the statistics of the current deme/generation.
 *  \param ioDeme Actual deme of the evolution.
 *  \param ioContext Context of the evolution.
 */
void NN::StatsCalcFitnessIndexedOp::operate(Beagle::Deme& ioDeme, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();
	Beagle_LogTraceM(
	    ioContext.getSystem().getLogger(),
	    "stats", "Beagle::NN::StatsCalcFitnessIndexedOp",
	    std::string("Calculating stats for the ")+
	    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
	);

	if(ioContext.getGeneration() != mGenerationCalculated) {
		mGenerationCalculated = ioContext.getGeneration();
		mNbDemesCalculated = 0;
	}

	if(ioDeme.getStats()->isValid() == false) {
		calculateStatsDeme(*ioDeme.getStats(), ioDeme, ioContext);
		ioDeme.getStats()->setValid();
	}
	Beagle_LogObjectM(
	    ioContext.getSystem().getLogger(),
	    Logger::eStats,
	    "stats", "Beagle::NN::StatsCalcFitnessIndexedOp",
	    *ioDeme.getStats()
	);

	// Don't calculate vivarium statistics because the demes could have
	// different stats sizes.

	Beagle_StackTraceEndM("void NN::StatsCalcFitnessIndexedOp::operate(Deme& ioDeme, Context& ioContext)");
}

