
/*!
 *  \file   beagle/NN/src/TermMaxFitnessOp.cpp
 *  \brief  Source code of class TermMaxFitnessOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a termination check operator based on a maximum fitness value.
 *  \param inMaxFitness Default maximum fitness value.
 *  \param inName Name of the operator.
 */
NN::TermMaxFitnessOp::TermMaxFitnessOp(float inMaxFitness, std::string inName) :
		TerminationOp(inName),
		mMaxFitnessDefault(inMaxFitness)
{ }


/*!
 *  \brief Register the parameters of this operator.
 *  \param ioSystem System to use to initialize the operator.
 */
void NN::TermMaxFitnessOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	TerminationOp::registerParams(ioSystem);
	Register::Description lDescription(
	    "Max fitness term criterion",
	    "Float",
	    dbl2str(mMaxFitnessDefault),
	    "Fitness value to reach before stopping evolution."
	);
	mMaxFitness = castHandleT<Float>(
	                  ioSystem.getRegister().insertEntry("nn.term.maxfitness", new Float(mMaxFitnessDefault), lDescription));

	Beagle_StackTraceEndM("void NN::TermMaxFitnessOp::registerParams(Beagle::System&)");
}


/*!
 *  \brief Read a max fitness termination operator from a XML subtree.
 *  \param inIter XML iterator from which the operator is read.
 *  \param inOpMap
 *  \throw IOException If a reading error occurs.
 */
void NN::TermMaxFitnessOp::readWithMap(PACC::XML::ConstIterator inIter, OperatorMap& inOpMap)
{
	Beagle_StackTraceBeginM();
	if((inIter->getType()!=PACC::XML::eData) || (inIter->getValue()!=getName())) {
		std::ostringstream lOSS;
		lOSS << "tag <" << getName() << "> expected!" << std::flush;
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}
	std::string lFitnessStr = inIter->getAttribute("fitness");
	if(lFitnessStr.empty() == false) {
		mMaxFitnessDefault = str2dbl(lFitnessStr);
		if(mMaxFitness != NULL)
			mMaxFitness->getWrappedValue() = mMaxFitnessDefault;
	}
	Beagle_StackTraceEndM("void NN::TermMaxFitnessOp::readWithMap(PACC::XML::ConstIterator inIter, OperatorMap& inOpMap)");
}


/*!
 *  \brief Check if the maximum fitness value is reached.
 *  \param inDeme Actual deme of the evolution.
 *  \param ioContext Actual evolution context.
 *  \return True if the ending criterion is reached, false if not.
 */
bool NN::TermMaxFitnessOp::terminate(const Beagle::Deme& inDeme, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();
	for(unsigned int i=0; i<inDeme.size(); ++i) {
		FitnessIndexed::Handle lFitness =
		    castHandleT<FitnessIndexed>(inDeme[i]->getFitness());
		if(lFitness->isValid() && lFitness->getValue() >= mMaxFitness->getWrappedValue()) {
			Beagle_LogInfoM(
			    ioContext.getSystem().getLogger(),
			    "termination", "Beagle::NN::TermMaxFitnessOp",
			    std::string("Maximum fitness value (") +
			    dbl2str(mMaxFitness->getWrappedValue()) +
			    std::string(") termination criterion reached by the ") +
			    uint2ordinal(i+1) + std::string(" individual (") +
			    dbl2str(lFitness->getValue()) + std::string(")")
			);
			ioContext.setTerminationSuccessful();
			return true;
		}
	}

	Beagle_LogTraceM(
	    ioContext.getSystem().getLogger(),
	    "termination", "Beagle::NN::TermMaxFitnessOp",
	    std::string("Maximum fitness value (") +
	    dbl2str(mMaxFitness->getWrappedValue()) +
	    std::string(") termination criterion not reached ")
	);
	return false;
	Beagle_StackTraceEndM("bool NN::TermMaxFitnessOp::terminate(const Beagle::Deme& inDeme, Beagle::Context& ioContext)");
}


/*!
 *  \brief Write a max fitness termination operator into a XML stream.
 *  \param ioStreamer XML stream in which the operator is written.
 *  \param inIndent Whether XML output should be indented.
 */
void NN::TermMaxFitnessOp::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();
	ioStreamer.openTag(getName(), inIndent);
	if(mMaxFitness != NULL)
		ioStreamer.insertAttribute("fitness", dbl2str(mMaxFitness->getWrappedValue()));
	else
		ioStreamer.insertAttribute("fitness", dbl2str(mMaxFitnessDefault));
	ioStreamer.closeTag();
	Beagle_StackTraceEndM("void NN::TermMaxFitnessOp::write(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}
