
/*!
 *  \file   beagle/SN/EstimateReconstructOp.hpp
 *  \brief  Definition of the type SN::EstimateReconstructOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_SN_EstimateReconstructOp_hpp
#define Beagle_SN_EstimateReconstructOp_hpp

#include <map>

#include "PACC/Util.hpp"

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/AllocatorT.hpp"
#include "beagle/PointerT.hpp"
#include "beagle/ContainerT.hpp"
#include "beagle/Operator.hpp"
#include "beagle/SN/MutationAddComparatorOp.hpp"


namespace Beagle
{
namespace SN
{


/*!
 *  \class EstimateReconstructOp beagle/SN/EstimateReconstructOp.hpp "beagle/SN/EstimateReconstructOp.hpp"
 *  \brief The class for estimating distribution of comparators based on
 *      the state of the network and using that distribution to construct
 *      new networks.
 *  \ingroup SNF
 */
class EstimateReconstructOp : public Operator
{

public:

	//! SN::EstimateReconstructOp allocator type.
	typedef AllocatorT<EstimateReconstructOp,Operator::Alloc>
	Alloc;
	//! SN::EstimateReconstructOp handle type.
	typedef PointerT<EstimateReconstructOp,Operator::Handle>
	Handle;
	//! SN::EstimateReconstructOp bag type.
	typedef ContainerT<EstimateReconstructOp,Operator::Bag>
	Bag;

    // Type for comparator frequencies for a given state of the sorting network.
    // The network state is an array of integers, each integer indicating the
    // binary state of the numsets array of a network (dual) function.
    typedef std::map<std::vector<unsigned int>, std::vector<std::vector<unsigned int> > > FrequenciesMap;

	EstimateReconstructOp(std::string inName="SN-EstimateReconstructOp");
	virtual ~EstimateReconstructOp()
	{ }

	virtual void registerParams(Beagle::System& ioSystem);
    virtual void init(Beagle::System& ioSystem);
	virtual void operate(Deme& ioDeme, Context& ioContext);

protected:

	UInt::Handle                mNumInputs;                     //!< Number of inputs of the sorting network.
	Float::Handle               mEstimateFraction;              //!< Fraction of population to use in estimating distribution.
	Float::Handle               mEstimateProbability;           //!< Probability of using estimated distribution to create new comparators.
	Bool::Handle                mGaussianTruncate;              //!< Whether to choose truncation point based on Gaussian or uniform distribution.
	FrequenciesMap              mFrequenciesMap;                //!< The frequency map for collecting statistics.
    MutationAddComparatorOp::Handle  mMutationOp;               //!< Mutation operator to use for adding comparators.

    void                clearFrequenciesMap();
    void                incrementFrequency(const std::vector<unsigned int>& inState, unsigned int inJoin, unsigned int inMeet);
    void                findComparatorRoulette(const std::vector<unsigned int>& inState, std::vector<unsigned int>& outIndexPairs, RouletteT<unsigned int>& outRoulette) const;
	void                estimate(const Beagle::Individual& inIndividual, Beagle::Context& ioContext);
	void                reconstruct(Beagle::Individual& ioIndividual, const Beagle::Individual& inParent, Beagle::Context& ioContext);

};

}
}


/*!
 *  \return Clear the frequency map.
 */
inline void Beagle::SN::EstimateReconstructOp::clearFrequenciesMap()
{
	Beagle_StackTraceBeginM();
    mFrequenciesMap.clear();
	Beagle_StackTraceEndM("void SN::EstimateReconstructOp::clearFrequenciesMap()");
}

#endif // Beagle_SN_EstimateReconstructOp_hpp
