
/*!
 *  \file   beagle/SN/Function.hpp
 *  \brief  Definition of the type SN::Function.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_SN_Function_hpp
#define Beagle_SN_Function_hpp

#include <boost/dynamic_bitset.hpp>

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/Container.hpp"
#include "beagle/SN/Utilities.hpp"


namespace Beagle
{
namespace SN
{

/*!
 *  \class Function beagle/SN/Function.hpp "beagle/SN/Function.hpp"
 *  \brief Data structure representing a monotone Boolean function.
 *  \ingroup SNF
 *  
 *  The horizontal lines of the sorting network represent monotone Boolean
 *	functions on the input variables or points in the free distributive lattice
 *	generated by the input variables.  The comparators compute the disjunction (join)
 *	and conjunction (meet) of pairs of input functions.  Each function is
 *	equivalent to a down set of the Boolean lattice of subsets of the input variables,
 *	and can therefore be represented as the set of points on the Boolean lattice
 *	where the function is true, i.e. we store the truth table of all 2^n possible
 *	inputs.  Each level k of the Boolean lattice contains all subsets containing
 *	k variables, i.e. all terms in which k variables are true.  The functions at the
 *	input of the network are down sets with a single-variable boundary term, i.e.
 *	all levels except the bottom level are only partially full.  On the other hand,
 *	the functions at the output of the network have only full or empty levels: level
 *	0 is empty and levels 1 to n are full for the first output, levels 0 and 1 are
 *	empty and levels 2 to n are full for the second output, and so on.  Because of
 *	this level-wise structure, we store the 2^n truth values of the function as
 *	separate levels.  Bit-wise AND and OR operations on the function representations
 *	are equivalent to intersection and union operations on the respective down sets.
 *
 *	The same function can also represent its dual, i.e. the function obtained by
 *	flipping all variables and the function value.
 */
class Function
{

public:

    typedef boost::dynamic_bitset<unsigned int> bitset_type;  // bitset implementation can be changed here.

	explicit Function() : mNumSets(), mTruthTable()
    { }
	virtual ~Function()
	{ }

	virtual void read(PACC::XML::ConstIterator inIter);
	virtual void write(PACC::XML::Streamer& ioStreamer, bool inIndent=true, unsigned int inIndex=UINT_MAX) const;

    bool isEqual(const Function& inRight) const;
    bool isLessEqual(const Function& inRight) const;
    bool isComparable(const Function& inRight) const;

    static void initialize(unsigned int inNumLevels);
    static unsigned int getNumLevels();
    static unsigned int getMaxSetsInLevel(unsigned int inLevel);
    static unsigned int getInputState();
    static unsigned int getInputStateDual();

    unsigned int getNumSets() const;
    unsigned int getNumSetsInLevel(unsigned int inLevel) const;
    unsigned int getNumSetsInLevelDual(unsigned int inLevel) const;
    unsigned int getFirstNonEmptyLevel() const;
    unsigned int getFirstNonEmptyLevelDual() const;
    unsigned int getNumNonEmptyLevels() const;
    unsigned int getNumNonEmptyLevelsDual() const;
    unsigned int getNumFullLevels() const;
    unsigned int getNumFullLevelsDual() const;
    unsigned int getNumFullLevelsBoth() const;
    bool isLevelFull(unsigned int inLevel) const;
    bool isLevelFullDual(unsigned int inLevel) const;
    unsigned int getState() const;
    unsigned int getStateDual() const;

    void setInput(unsigned int inLine);
    void setOutput(unsigned int inLine);

    void invalidate();
    bool isValid();

    void printNumSets() const;

private:

    static unsigned int                     smNumLevels;    //!< Number of levels is one more than the number of variables.
    static std::vector<unsigned int>        smMaxSets;      //!< Maximum number of sets in each level.
    std::vector<unsigned int>               mNumSets;       //!< Number of sets in each level.
    std::vector<bitset_type>                mTruthTable;    //!< Truth table of the function stored by level.

    // For efficiency, mTruthTable[i].size() > 0 only when mNumSets[i] > 0 and < smMaxSets[i]
    // because otherwise entries in mTruthTable[i] are either all 0s or all 1s respectively.

    friend void join(const Function& inLeft, const Function& inRight, Function& outJoin);
    friend void meet(const Function& inLeft, const Function& inRight, Function& outMeet);
    friend void join(Function& ioLeft, const Function& inRight);
    friend void meet(Function& ioLeft, const Function& inRight);

};

}
}



/*!
 *  \return The number of levels of the function.
 */
inline unsigned int Beagle::SN::Function::getNumLevels()
{
	Beagle_StackTraceBeginM();
    Beagle_AssertM(smNumLevels > 0);
    return smNumLevels;
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumLevels()");
}


/*!
 *  \return The maximum number of sets in the given level.
 */
inline unsigned int Beagle::SN::Function::getMaxSetsInLevel(unsigned int inLevel)
{
	Beagle_StackTraceBeginM();
    Beagle_AssertM(inLevel < smMaxSets.size() && smMaxSets[inLevel] > 0);
    return smMaxSets[inLevel];
	Beagle_StackTraceEndM("unsigned int SN::Function::getMaxSetsInLevel(unsigned int inLevel)");
}


/*!
 *  \return The total number of sets in all level.
 */
inline unsigned int Beagle::SN::Function::getNumSets() const
{
	Beagle_StackTraceBeginM();
    unsigned int lCount = 0;
    for (unsigned int i = 0; i < smNumLevels; ++i) lCount += mNumSets[i];
    return lCount;
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumSets() const");
}


/*!
 *  \return The number of sets in the given level.
 */
inline unsigned int Beagle::SN::Function::getNumSetsInLevel(unsigned int inLevel) const
{
	Beagle_StackTraceBeginM();
    Beagle_AssertM(inLevel < mNumSets.size());
    return mNumSets[inLevel];
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumSetsInLevel(unsigned int inLevel) const");
}


/*!
 *  \return The number of sets in the given level for the dual function.
 */
inline unsigned int Beagle::SN::Function::getNumSetsInLevelDual(unsigned int inLevel) const
{
	Beagle_StackTraceBeginM();
    Beagle_AssertM(inLevel < mNumSets.size());
    return (smMaxSets[inLevel] - mNumSets[inLevel]);
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumSetsInLevelDual(unsigned int inLevel) const");
}


/*!
 *  \return The first non-empty level.
 */
inline unsigned int Beagle::SN::Function::getFirstNonEmptyLevel() const
{
	Beagle_StackTraceBeginM();
    unsigned int lLevel = 0;
    while (mNumSets[lLevel] == 0) lLevel++;
    return lLevel;
	Beagle_StackTraceEndM("unsigned int SN::Function::getFirstNonEmptyLevel() const");
}


/*!
 *  \return The first non-empty level of dual function.
 */
inline unsigned int Beagle::SN::Function::getFirstNonEmptyLevelDual() const
{
	Beagle_StackTraceBeginM();
    unsigned int lLevel = smNumLevels-1;
    while (mNumSets[lLevel] == smMaxSets[lLevel]) lLevel--;
    return lLevel;
	Beagle_StackTraceEndM("unsigned int SN::Function::getFirstNonEmptyLevelDual() const");
}


/*!
 *  \return The number of non-empty levels.
 */
inline unsigned int Beagle::SN::Function::getNumNonEmptyLevels() const
{
	Beagle_StackTraceBeginM();
    unsigned int lCount = 0;
    for (unsigned int i = 0; i < smNumLevels; ++i) if (mNumSets[i] > 0) lCount++;
    return lCount;
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumNonEmptyLevels() const");
}


/*!
 *  \return The number of non-empty levels of dual function.
 */
inline unsigned int Beagle::SN::Function::getNumNonEmptyLevelsDual() const
{
	Beagle_StackTraceBeginM();
    unsigned int lCount = 0;
    for (unsigned int i = 0; i < smNumLevels; ++i) if (mNumSets[i] < smMaxSets[i]) lCount++;
    return lCount;
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumNonEmptyLevelsDual() const");
}


/*!
 *  \return The number of full levels.
 */
inline unsigned int Beagle::SN::Function::getNumFullLevels() const
{
	Beagle_StackTraceBeginM();
    unsigned int lCount = 0;
    for (unsigned int i = 0; i < smNumLevels; ++i) if (mNumSets[i] == smMaxSets[i]) lCount++;
    return lCount;
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumFullLevels() const");
}


/*!
 *  \return The number of full levels of dual function.
 */
inline unsigned int Beagle::SN::Function::getNumFullLevelsDual() const
{
	Beagle_StackTraceBeginM();
    unsigned int lCount = 0;
    for (unsigned int i = 0; i < smNumLevels; ++i) if (mNumSets[i] == 0) lCount++;
    return lCount;
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumFullLevelsDual() const");
}


/*!
 *  \return The number of full levels of dual function.
 */
inline unsigned int Beagle::SN::Function::getNumFullLevelsBoth() const
{
	Beagle_StackTraceBeginM();
    unsigned int lCount = 0;
    for (unsigned int i = 0; i < smNumLevels; ++i) if (mNumSets[i] == smMaxSets[i] || mNumSets[i] == 0) lCount++;
    return lCount;
	Beagle_StackTraceEndM("unsigned int SN::Function::getNumFullLevelsBoth() const");
}


/*!
 *  \return Whether there is maximum number of sets in the given level.
 */
inline bool Beagle::SN::Function::isLevelFull(unsigned int inLevel) const
{
	Beagle_StackTraceBeginM();
    Beagle_AssertM(inLevel < mNumSets.size());
    return (mNumSets[inLevel] == smMaxSets[inLevel]);
	Beagle_StackTraceEndM("bool SN::Function::isLevelFull(unsigned int inLevel) const");
}


/*!
 *  \return Whether there is maximum number of sets in the given level for the dual function.
 */
inline bool Beagle::SN::Function::isLevelFullDual(unsigned int inLevel) const
{
	Beagle_StackTraceBeginM();
    Beagle_AssertM(inLevel < mNumSets.size());
    return (mNumSets[inLevel] == 0);
	Beagle_StackTraceEndM("bool SN::Function::isLevelFullDual(unsigned int inLevel) const");
}


/*!
 *  \return The state represented by the function.
 */
inline unsigned int Beagle::SN::Function::getState() const
{
	Beagle_StackTraceBeginM();
    unsigned int lState = 0;
    for (unsigned int i = 0; i < smNumLevels; ++i) if (mNumSets[i] > 0) lState += 1<<i;
	return lState;
	Beagle_StackTraceEndM("unsigned int SN::Function::getState() const");
}


/*!
 *  \return The state represented by the dual function.
 */
inline unsigned int Beagle::SN::Function::getStateDual() const
{
	Beagle_StackTraceBeginM();
    unsigned int lState = 0;
    for (unsigned int i = 0; i < smNumLevels; ++i) if (mNumSets[i] < smMaxSets[i]) lState += 1<<i;
	return lState;
	Beagle_StackTraceEndM("unsigned int SN::Function::getStateDual() const");
}


/*!
 *  \return The state represented by the function at any of the network inputs.
 */
inline unsigned int Beagle::SN::Function::getInputState()
{
	Beagle_StackTraceBeginM();
	return (1<<smNumLevels)-2;
	Beagle_StackTraceEndM("unsigned int SN::Function::getInputState()");
}


/*!
 *  \return The state represented by the dual function at any of the network inputs.
 */
inline unsigned int Beagle::SN::Function::getInputStateDual()
{
	Beagle_StackTraceBeginM();
	return (1<<(smNumLevels-1))-1;
	Beagle_StackTraceEndM("unsigned int SN::Function::getInputStateDual()");
}


/*!
 *  \return Invalidate the function and clear memory.
 */
inline void Beagle::SN::Function::invalidate()
{
	Beagle_StackTraceBeginM();
    mTruthTable.clear();
	Beagle_StackTraceEndM("void SN::Function::invalidate()");
}


/*!
 *  \return Whether the function is valid.
 */
inline bool Beagle::SN::Function::isValid()
{
	Beagle_StackTraceBeginM();
    return (mTruthTable.size() > 0);
	Beagle_StackTraceEndM("bool SN::Function::isValid()");
}


#endif // Beagle_SN_Function_hpp
