
/*!
 *  \file   beagle/SN/Network.hpp
 *  \brief  Definition of the type SN::Network.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_SN_Network_hpp
#define Beagle_SN_Network_hpp

#include <string>
#include <ext/slist>

#include "beagle/config.hpp"
#include "beagle/macros.hpp"
#include "beagle/Object.hpp"
#include "beagle/Allocator.hpp"
#include "beagle/Pointer.hpp"
#include "beagle/Container.hpp"
#include "beagle/Genotype.hpp"
#include "beagle/SN/Function.hpp"

namespace Beagle
{
namespace SN
{

/*!
 *  \class Network beagle/SN/Network.hpp "beagle/SN/Network.hpp"
 *  \brief Genotype class for sorting networks, stored as a vector of the input
 *         indices of the comparators.
 *  \ingroup SNF
 */
class Network : public Genotype, public std::vector<unsigned int>
{

public:

	//! SN::Network allocator type.
	typedef AllocatorT<Network, Genotype::Alloc>
	Alloc;
	//! SN::Network handle type.
	typedef PointerT<Network, Genotype::Handle>
	Handle;
	//! SN::Network bag type.
	typedef ContainerT<Network, Genotype::Bag>
	Bag;

	//! Heuristic method to use in deciding where to add comparators.
	enum HeuristicMethod {eUniformHeuristic = 0, eDominatingComponents = 1, eNewSorted = 2, eNewSets = 3,
                          eWeightedNewSets = 4, eTargetDistanceReduction = 5, eNewSetsPotential = 6,
                          eMigrationCostReduction = 7, eSameWeightNewSets = 8, eBottomUpBinaryTree = 9,
                          eBottomUpBinaryTreeDual = 10, eComparatorUtility = 11, eNextLineUtility = 12,
                          eNextNextLineUtility = 13, eDeltaZeroBitCountLines = 14, eAllLinesUtility = 15,
                          eAllLinesLookAheadUtility = 16, eEffectNextLineComparators = 17, eNonZeroMeetLine = 18,
                          eMaxMatchingCardinality = 19, eSymmetricLinesMeasure = 20, eCombinedLinesUtility = 21,
                          eMeetProximity = 22, eAverageLookAheadUtility = 23, eFirstNonEmptyMeetIndex = 24,
                          eFirstNonEmptyIndexVariance = 25, eMatchMeetWithLines = 26, eSymmetricFunctionsBefore = 27,
                          eSymmetricFunctionsAfter = 28, eFilledLevels = 29, eDestinationLine = 30, eDisagreement = 31,
                          eAutoAllLinesUtility = 100};

	explicit Network(unsigned int inSize=0, unsigned int inModel=0);
	virtual ~Network()
	{ }

	virtual void                copy(const Member& inOriginal, System& ioSystem);
	virtual unsigned int        getSize() const;
	virtual const std::string&  getType() const;

	virtual bool                isEqual(const Object& inRightObj) const;
	virtual bool                isLess(const Object& inRightObj) const;
	virtual void                readWithContext(PACC::XML::ConstIterator inIter, Beagle::Context& ioContext);
	virtual void                writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent=true) const;

    void                        allocateDataStructures();

    void                        addComparator(unsigned int inJoin, unsigned int inMeet);

    unsigned int                getNumInputs() const;
    void                        setNumInputs(unsigned int inNumInputs);

    unsigned int                getNumComparators() const;
    unsigned int                getNumStages() const;
    unsigned int                getCurrentLine() const;
    unsigned int                getFilterStartIndex() const;
    bool                        isLineSorted(unsigned int inIndex) const;
    bool                        isLineSortedDual(unsigned int inIndex) const;
    bool                        isSorted() const;

    std::vector<unsigned int>&          getStages();
    const std::vector<unsigned int>&    getStages() const;

    std::vector<Function>&          getFunctions();
    const std::vector<Function>&    getFunctions() const;

    std::vector<HeuristicMethod>&     getHeuristicMethods();
    const std::vector<HeuristicMethod>&     getHeuristicMethods() const;

    const Function&             getJoin(unsigned int i, unsigned int j);
    const Function&             getMeet(unsigned int i, unsigned int j);

    std::vector<std::vector<std::vector<double> > >&          getHeuristic();
    const std::vector<std::vector<std::vector<double> > >&    getHeuristic() const;

    const std::vector<unsigned int>&           getTriplet(unsigned int i, unsigned int j) const;

    __gnu_cxx::slist<unsigned int>&            getAsymmetricComparators();
    const __gnu_cxx::slist<unsigned int>&      getAsymmetricComparators() const;

    std::vector<unsigned int>&          getLineStates();
    const std::vector<unsigned int>&    getLineStates() const;

    void                        adjustHeuristicForTriplets();
    void                        updateHeuristic(unsigned int i, unsigned int j);
    void                        calculateHeuristic();
    double                      calculateHeuristic(unsigned int i, unsigned int j, HeuristicMethod inHeuristicMethod);
    void					    nondomHeuristic(std::vector<unsigned int>& outIndexPairs) const;
	double					    maxHeuristic(unsigned int h) const;
	void					    maxHeuristic(std::vector<unsigned int>& outIndexPairs, unsigned int h) const;
	double					    maxHeuristicFilter(std::vector<unsigned int>& ioIndexPairs, unsigned int h) const;
	void					    nonzeroHeuristic(std::vector<unsigned int>& outIndexPairs, unsigned int h) const;
	void					    nonzeroHeuristicFilter(std::vector<unsigned int>& outIndexPairs, unsigned int h) const;
    void					    findCandidateComparators(std::vector<unsigned int>& outIndexPairs,
                                                         const std::vector<unsigned int>& inFilterGroup = std::vector<unsigned int>()) const;
    void					    findCandidateComparators(std::vector<unsigned int>& outIndexPairs, RouletteT<unsigned int>& outRoulette,
                                                         const std::vector<unsigned int>& inFilterGroup = std::vector<unsigned int>()) const;
    void					    lookAheadHeuristicFilter(std::vector<unsigned int>& ioIndexPairs) const;
    void					    maxMatchingFilter(std::vector<unsigned int>& ioIndexPairs) const;
    void					    findSymmetricComparators(const std::vector<unsigned int>& inIndexPairs, std::vector<unsigned int>& outIndexPairs) const;
    void                        findSymmetricComparatorIndices(const std::vector<unsigned int>& inIndexPairs, std::vector<unsigned int>& outIndices) const;

    unsigned int    findNumSorted() const;
    unsigned int    findNumSets() const;
    unsigned int    findNumFullLevels() const;
    unsigned int    findNumLinesCompleted() const;

    void            trace();
    void            printHeuristic() const;

protected:

	friend class InitializationOp;

    unsigned int                    mNumInputs;         //!< Number of inputs of the sorting network.        
    std::vector<unsigned int>       mStages;            //!< Indices into the comparator vector denoting parallel stages.
    std::vector<Function>           mFunctions;         //!< The current monotone Boolean functions for each horizontal line.
    std::vector<HeuristicMethod>    mHeuristicMethods;  //!< Heuristic methods used to decide comparator placement.
    unsigned int                    mFilterStartIndex;  //!< Index into the heuristic methods array, indicating the start of filters.
    std::vector<std::vector<Function> > mJoins;         //!< Array for storing joins for heuristic calculation.
    std::vector<std::vector<Function> > mMeets;         //!< Array for storing meets for heuristic calculation.
    std::vector<std::vector<std::vector<double> > > mHeuristic; //!< Values of heuristic for each pair of lines.
    std::vector<std::vector<std::vector<unsigned int> > > mTriplets;  //!< Array for storing comparator triplets with better utility than couplets.
    __gnu_cxx::slist<unsigned int>  mAsymmetricComparators;     //!< Index list of asymmetric comparators.
    unsigned int                    mNumAllLines;       //!< The number of lines actually used in eAutoAllLinesUtility heuristic.
    std::vector<unsigned int>       mLineStates;        //!< States of the lines as new comparators are added.

private:

	template <class T> void         readVectorContent(PACC::XML::ConstIterator inIter, std::vector<T>& outVector);
	template <class T> void         writeVectorContent(PACC::XML::Streamer& ioStreamer, const std::vector<T>& inVector) const;

    bool findComparatorTriplet(unsigned int p1, unsigned int q1, unsigned int inCurrentLine, unsigned int inDualCurrentLine,
                               unsigned int inUtilityIndex, double inUtility1, double inUtility2, double inUtility3);
    unsigned int checkComparatorTriplet(unsigned int a, unsigned int b, unsigned int c, unsigned int d,
                                        unsigned int& p1, unsigned int& q1, unsigned int& p2, unsigned int& q2);

    inline static bool gtrank(const std::pair<double, unsigned int>& a, const std::pair<double, unsigned int>& b) {return a.first > b.first;}

};

}
}


/*!
 *  \return Size of the vector representation of the sorting network.
 */
inline unsigned int Beagle::SN::Network::getSize() const
{
	Beagle_StackTraceBeginM();
	return size();
	Beagle_StackTraceEndM("unsigned int SN::Network::getSize() const");
}


/*!
 *  \return Number of inputs of the sorting network.
 */
inline unsigned int Beagle::SN::Network::getNumInputs() const
{
	Beagle_StackTraceBeginM();
	return mNumInputs;
	Beagle_StackTraceEndM("unsigned int SN::Network::getNumInputs() const");
}


/*!
 *  \brief Set the number of inputs of the sorting network.
 */
inline void Beagle::SN::Network::setNumInputs(unsigned int inNumInputs)
{
	Beagle_StackTraceBeginM();
	mNumInputs = inNumInputs;
	Beagle_StackTraceEndM("void SN::Network::setNumInputs(unsigned int inNumInputs)");
}


/*!
 *  \return Number of comparators in the sorting network.
 */
inline unsigned int Beagle::SN::Network::getNumComparators() const
{
	Beagle_StackTraceBeginM();
	return size()/2;
	Beagle_StackTraceEndM("unsigned int SN::Network::getNumComparators() const");
}


/*!
 *	\return The starting index of filter heuristic methods in the heuristic array.
 */
inline unsigned int Beagle::SN::Network::getFilterStartIndex() const
{
	Beagle_StackTraceBeginM();
	return mFilterStartIndex;
	Beagle_StackTraceEndM("unsigned int SN::Network::getFilterStartIndex() const");
}


/*!
 *	\return True if the line is already sorted; false otherwise.
 */
inline bool Beagle::SN::Network::isLineSorted(unsigned int inIndex) const
{
	Beagle_StackTraceBeginM();
    unsigned int lLevel = mFunctions[inIndex].getFirstNonEmptyLevel();
    return (lLevel == inIndex+1) && (mFunctions[inIndex].getNumSetsInLevel(lLevel) == Function::getMaxSetsInLevel(lLevel));
	Beagle_StackTraceEndM("bool SN::Network::isLineSorted(unsigned int inIndex) const");
}


/*!
 *	\return True if the line is already sorted; false otherwise (dual version).
 *          This function will return identical result as the primal version, but
 *          is more efficient if index > numinputs/2.
 */
inline bool Beagle::SN::Network::isLineSortedDual(unsigned int inIndex) const
{
	Beagle_StackTraceBeginM();
    unsigned int lLevel = mFunctions[inIndex].getFirstNonEmptyLevelDual();
    return (lLevel == inIndex) && (mFunctions[inIndex].getNumSetsInLevelDual(lLevel) == Function::getMaxSetsInLevel(lLevel));
	Beagle_StackTraceEndM("bool SN::Network::isLineSortedDual(unsigned int inIndex) const");
}


/*!
 *	\return True if all lines of the network are sorted; false otherwise.
 */
inline bool Beagle::SN::Network::isSorted() const
{
	Beagle_StackTraceBeginM();
    bool lSorted = true;
    for (unsigned int i = 0; lSorted && i < mNumInputs; ++i) {
        lSorted = lSorted && isLineSorted(i);
    }
	return lSorted;
	Beagle_StackTraceEndM("bool SN::Network::isSorted() const");
}


/*!
 *	\return The join between lines i and j (i < j)
 */
inline const Beagle::SN::Function& Beagle::SN::Network::getJoin(unsigned int i, unsigned int j)
{
	Beagle_StackTraceBeginM();
    unsigned int x = i;
    unsigned int y = j-i-1;
    if (mJoins[x][y].isValid() == false) {
        join(mFunctions[i], mFunctions[j], mJoins[x][y]);
    }
	return mJoins[x][y];
	Beagle_StackTraceEndM("const SN::Function& SN::Network::getJoin(unsigned int i, unsigned int j)");
}


/*!
 *	\return The meet between lines i and j (i < j)
 */
inline const Beagle::SN::Function& Beagle::SN::Network::getMeet(unsigned int i, unsigned int j)
{
	Beagle_StackTraceBeginM();
    unsigned int x = i;
    unsigned int y = j-i-1;
    if (mMeets[x][y].isValid() == false) {
        meet(mFunctions[i], mFunctions[j], mMeets[x][y]);
    }
	return mMeets[x][y];
	Beagle_StackTraceEndM("const SN::Function& SN::Network::getMeet(unsigned int i, unsigned int j)");
}


/*!
 *  \return Number of parallel stages in the sorting network.
 */
inline unsigned int Beagle::SN::Network::getNumStages() const
{
	Beagle_StackTraceBeginM();
	return mStages.size();
	Beagle_StackTraceEndM("unsigned int SN::Network::getNumStages() const");
}


/*!
 *  \return Parallel stages of the sorting network.
 */
inline std::vector<unsigned int>& Beagle::SN::Network::getStages()
{
	Beagle_StackTraceBeginM();
	return mStages;
	Beagle_StackTraceEndM("std::vector<unsigned int>& SN::Network::getStages()");
}


/*!
 *  \return Parallel stages of the sorting network.
 */
inline const std::vector<unsigned int>& Beagle::SN::Network::getStages() const
{
	Beagle_StackTraceBeginM();
	return mStages;
	Beagle_StackTraceEndM("const std::vector<unsigned int>& SN::Network::getStages() const");
}


/*!
 *  \return The monotone Boolean functions of the sorting network.
 */
inline std::vector<Beagle::SN::Function>& Beagle::SN::Network::getFunctions()
{
	Beagle_StackTraceBeginM();
	return mFunctions;
	Beagle_StackTraceEndM("std::vector<Beagle::SN::Function>& SN::Network::getFunctions()");
}


/*!
 *  \return The monotone Boolean functions of the sorting network.
 */
inline const std::vector<Beagle::SN::Function>& Beagle::SN::Network::getFunctions() const
{
	Beagle_StackTraceBeginM();
	return mFunctions;
	Beagle_StackTraceEndM("const std::vector<Beagle::SN::Function>& SN::Network::getFunctions() const");
}


/*!
 *  \return The heuristic methods used in the sorting network.
 */
inline std::vector<Beagle::SN::Network::HeuristicMethod>& Beagle::SN::Network::getHeuristicMethods()
{
	Beagle_StackTraceBeginM();
	return mHeuristicMethods;
	Beagle_StackTraceEndM("std::vector<Beagle::SN::Network::HeuristicMethod>& Beagle::SN::Network::getHeuristicMethods()");
}


/*!
 *  \return The heuristic methods used in the sorting network.
 */
inline const std::vector<Beagle::SN::Network::HeuristicMethod>& Beagle::SN::Network::getHeuristicMethods() const
{
	Beagle_StackTraceBeginM();
	return mHeuristicMethods;
	Beagle_StackTraceEndM("const std::vector<Beagle::SN::Network::HeuristicMethod>& Beagle::SN::Network::getHeuristicMethods() const");
}


/*!
 *  \return The array containing heuristic values.
 */
inline std::vector<std::vector<std::vector<double> > >& Beagle::SN::Network::getHeuristic()
{
	Beagle_StackTraceBeginM();
	return mHeuristic;
	Beagle_StackTraceEndM("std::vector<std::vector<std::vector<double> > >& SN::Network::getHeuristic()");
}


/*!
 *  \return The array containing heuristic values.
 */
inline const std::vector<std::vector<std::vector<double> > >& Beagle::SN::Network::getHeuristic() const
{
	Beagle_StackTraceBeginM();
	return mHeuristic;
	Beagle_StackTraceEndM("const std::vector<std::vector<std::vector<double> > >& SN::Network::getHeuristic() const");
}


/*!
 *  \return The entry in the triplets matrix corresponding to comparator (i, j)
 */
inline const std::vector<unsigned int>& Beagle::SN::Network::getTriplet(unsigned int i, unsigned int j) const
{
	Beagle_StackTraceBeginM();
	return mTriplets[i][j-i-1];
	Beagle_StackTraceEndM("const std::vector<unsigned int>& SN::Network::getTriplet(unsigned int i, unsigned int j) const");
}


/*!
 *  \return The comparators that are asymmetric.
 */
inline __gnu_cxx::slist<unsigned int>& Beagle::SN::Network::getAsymmetricComparators()
{
	Beagle_StackTraceBeginM();
	return mAsymmetricComparators;
	Beagle_StackTraceEndM("__gnu_cxx::slist<unsigned int>& SN::Network::getAsymmetricComparators()");
}


/*!
 *  \return The comparators that are asymmetric.
 */
inline const __gnu_cxx::slist<unsigned int>& Beagle::SN::Network::getAsymmetricComparators() const
{
	Beagle_StackTraceBeginM();
	return mAsymmetricComparators;
	Beagle_StackTraceEndM("const __gnu_cxx::slist<unsigned int>& SN::Network::getAsymmetricComparators() const");
}


/*!
 *  \return Line states of the sorting network.
 */
inline std::vector<unsigned int>& Beagle::SN::Network::getLineStates()
{
	Beagle_StackTraceBeginM();
	return mLineStates;
	Beagle_StackTraceEndM("std::vector<unsigned int>& SN::Network::getLineStates()");
}


/*!
 *  \return Line states of the sorting network.
 */
inline const std::vector<unsigned int>& Beagle::SN::Network::getLineStates() const
{
	Beagle_StackTraceBeginM();
	return mLineStates;
	Beagle_StackTraceEndM("const std::vector<unsigned int>& SN::Network::getLineStates() const");
}


#endif // Beagle_SN_Network_hpp
