
/*!
 *  \file   beagle/SN/Utilities.hpp
 *  \brief  Utilities for evolving sorting networks.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef Beagle_SN_Utilities_hpp
#define Beagle_SN_Utilities_hpp



namespace Beagle
{
namespace SN
{

    class Function;
    class Network;

	//! Color values used in breadth first search algorithm.
	enum Color {eWhite, eGray, eBlack};

	//! Partition values used in breadth first search algorithm.
	enum Partition {eNone, eLeft, eRight};

    // Functions for generating bit combinations.
    unsigned int nextComb(unsigned int w);
    unsigned int firstComb(unsigned int k);
    unsigned int lastComb(unsigned int k, unsigned int n);

    unsigned int binomial(unsigned int n, unsigned int k);

    void join(const Function& inLeft, const Function& inRight, Function& outJoin);
    void meet(const Function& inLeft, const Function& inRight, Function& outMeet);
    void join(Function& ioLeft, const Function& inRight);
    void meet(Function& ioLeft, const Function& inRight);

    unsigned int findUniformUtility(const Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    unsigned int countNewSorted(const Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    int findComparatorUtility(const Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    unsigned int findComparatorUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, unsigned int inLine);
    unsigned int findNonZeroMeetLineUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    unsigned int findNextLineUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, unsigned int inNext=1);
    unsigned int findCombinedLinesUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, bool inSimple=true);
    unsigned int findCombinedLinesUtility(const Function& inLeft, const Function& inRight, unsigned int inLeftIndex,
                                          unsigned int inRightIndex, unsigned int inNumInputs, bool inSimple=true);
    unsigned int findMeetProximity(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    double findAverageLookAheadUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    unsigned int findFirstNonEmptyMeetIndex(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    unsigned int findFirstNonEmptyIndexVariance(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    int matchMeetWithLines(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    int findAllLinesUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    int findAllLinesLookAheadUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    int utility(unsigned int inBeforeCount, unsigned int inAfterCount);
    unsigned int findSymmetricFunctionsUtilityBefore(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    unsigned int findSymmetricFunctionsUtilityAfter(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    int findFilledLevelsUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    unsigned int findDestinationLineUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);
    unsigned int findDisagreementUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex);

}
}


/*!
 *	\brief Return the smallest integer greater w with the same number of bits set.
 *
 *	 When used repeatedly, this function generates all bit string combinations
 *	 with the same number of bits set.  It is taken from Torsten Sillke's bitmani.h
 *	 available at http://www.mathematik.uni-bielefeld.de/~sillke/ALGORITHMS/
 *	
 *	  Examples (binary notation):
 *
 *	1 -> 10 -> 100 -> 1000 -> 10000 -> 100000 -> 1000000 -> ...
 *	11 -> 101 -> 110 -> 1001 -> 1010 -> 1100 -> 10001 -> ...
 *	111 -> 1011 -> 1101 -> 1110 -> 10011 -> 10101 -> 10110 -> ...
 *
 *	  Special cases:
 *
 *		0 -> 0
 *		if the number is too large then 0 is returned.
 *
 *	  The bit pattern to change at the end of the word is:
 *
 *	0 1	  0	  ->	1 0	  1
 *	   a   b		   b+1 a-1
 *
 *	  Example: (the bits marked by B remain fixed)
 *	w		= BBBB01111100000000
 *	z		= 000000000100000000   (the lowest set bit)
 *	w+z		= BBBB10000000000000
 *	(w+z)^w = 000011111100000000
 *	Next	= BBBB10000000001111
 *
 *	  Reference: HAKMEM Item 175
 *	In two's compelment arithmetic: z = w & -w;
 *
 *  Other implementations can be found on Jörg Arndt's bit wizardry page:
 *  http://www.jjj.de/bitwizardry/files/bitcombcolex.h
 */
inline unsigned int Beagle::SN::nextComb(unsigned int w)
{
	Beagle_StackTraceBeginM();

    unsigned int ww = w, z;
    if (ww) {
        z  = w & -w;
        w += z;
        if (w) w += (w^ww)/z>>2;
    }
    return w;

	Beagle_StackTraceEndM("unsigned int SN::nextComb(unsigned int w)");
}


/*!
 *	\brief Return the first combination of (smallest word with) k
 *	    bits, i.e. 00..001111..1 (k low bits set)
 *	
 *	Adapted from Jörg Arndt's bit wizardry page:
 *  http://www.jjj.de/bitwizardry/files/bitcombcolex.h
 */
inline unsigned int Beagle::SN::firstComb(unsigned int k)
{
	Beagle_StackTraceBeginM();

    return ~(UINT_MAX << k);

	Beagle_StackTraceEndM("unsigned int SN::firstComb(unsigned int k)");
}


/*!
 *	\brief Return the last combination of (biggest n-bit word with) k
 *	    bits, i.e. 1111..100..00 (k high bits set)
 *	
 *	Adapted from Jörg Arndt's bit wizardry page:
 *  http://www.jjj.de/bitwizardry/files/bitcombcolex.h
 */
inline unsigned int Beagle::SN::lastComb(unsigned int k, unsigned int n)
{
	Beagle_StackTraceBeginM();

    return firstComb(k) << (n - k);

	Beagle_StackTraceEndM("unsigned int SN::lastComb(unsigned int k, unsigned int n)");
}


#endif  // Beagle_SN_Utilities_hpp
