
/*!
 *  \file   beagle/SN/src/EstimateReconstructOp.cpp
 *  \brief  Source code of class SN::EstimateReconstructOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */


#include "beagle/SN.hpp"

using namespace Beagle;


/*!
 *  \brief Default constructor of SN estimate reconstruct operator.
 */
SN::EstimateReconstructOp::EstimateReconstructOp(std::string inName) :
        Operator(inName),
		mFrequenciesMap(),
        mMutationOp(new SN::MutationAddComparatorOp)
{ }


/*!
 *  \brief Register the parameters of the generic SN intialization operator.
 *  \param ioSystem System of the evolution.
 */
void SN::EstimateReconstructOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Operator::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of inputs of the sorting network",
		    "UInt",
		    "0",
		    "Number of inputs of the sorting network."
		);
		mNumInputs = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("sn.init.numinputs", new UInt(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Fraction of population to use in estimating distribution",
		    "Float",
		    "0.5",
		    "Fraction of population to use in estimating distribution; the remainder of the population is replaced."
		);
		mEstimateFraction = castHandleT<Float>(
		                 ioSystem.getRegister().insertEntry("sn.estimate.frac", new Float(0.5f), lDescription));
	}
	{
		Register::Description lDescription(
		    "Probability of using estimated distribution to create new comparators",
		    "Float",
		    "0.5",
		    "Probability of using estimated distribution to create new comparators; remaining comparators are created using uniform heuristic."
		);
		mEstimateProbability = castHandleT<Float>(
		                 ioSystem.getRegister().insertEntry("sn.estimate.prob", new Float(0.5f), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to choose truncation point based on Gaussian or uniform distribution",
		    "Bool",
		    "1",
		    "If true, choose truncation point based on Gaussian distribution; otherwise use uniform distribution."
		);
		mGaussianTruncate = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.estimate.gtrun", new Bool(1), lDescription));
	}
    if (mMutationOp != NULL) {
        mMutationOp->registerParams(ioSystem);
    }
	Beagle_StackTraceEndM("void SN::EstimateReconstructOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize operator.
 */
void SN::EstimateReconstructOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
    if (mMutationOp != NULL) {
        mMutationOp->init(ioSystem);
    }
	Beagle::Operator::init(ioSystem);
	Beagle_StackTraceEndM("void SN::EstimateReconstructOp::init(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply the operation on a deme in the given context.
 *  \param ioDeme Reference to the deme on which the operation takes place.
 *  \param ioContext Evolutionary context of the operation.
 */
void SN::EstimateReconstructOp::operate(Deme& ioDeme, Context& ioContext)
{
	Beagle_StackTraceBeginM();

	Beagle_ValidateParameterM(mEstimateFraction->getWrappedValue() <= 1,
	                          "sn.estimate.frac",
	                          "The estimate fraction must be less than or equal to 1!");
	Beagle_ValidateParameterM(mEstimateFraction->getWrappedValue() >= 0,
	                          "sn.estimate.frac",
	                          "The estimate fraction must be greater than or equal to 0!");

	Beagle_LogTraceM(
	    ioContext.getSystem().getLogger(),
	    "estimate-reconstruct", "Beagle::SN::EstimateReconstructOp",
	    std::string("Processing using estimate reconstruct operator the ")+
	    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
	);
	Beagle_LogObjectM(
	    ioContext.getSystem().getLogger(),
	    Logger::eTrace,
	    "estimate-reconstruct", "Beagle::SN::EstimateReconstructOp",
	    (*this)
	);

    // Clear the current frequencies map since we will be building a new one.
    mFrequenciesMap.clear();

	SN::Individual& lIndividual = castObjectT<SN::Individual&>(*ioDeme[0]);
	SN::Network& lNetwork = *lIndividual[0];

    // The top individuals are used for estimating distribution and are not reconstructed.
    // Estimate distribution only if the network already has comparators.
    unsigned int lEstimateSize = (lNetwork.getNumComparators() > 0) ? (unsigned int)(((float)ioDeme.size())*mEstimateFraction->getWrappedValue()) : 0;
	if(lEstimateSize > 0) {
		make_heap(ioDeme.begin(), ioDeme.end(), IsLessPointerPredicate());
		for(unsigned int i=0; i<lEstimateSize; ++i) {
            estimate(*ioDeme[0], ioContext);
			pop_heap(ioDeme.begin(), (ioDeme.end()-i), IsLessPointerPredicate());
		}
	}

    // Reconstruct remaining individuals (the above individuals become parents except
    // in the 0th generation).
	for(unsigned int i=0; i<(ioDeme.size()-lEstimateSize); ++i) {
        unsigned int lParentIndex = (lEstimateSize > 0) ? (ioDeme.size()-1-(i%lEstimateSize)) : i;
        reconstruct(*ioDeme[i], *ioDeme[lParentIndex], ioContext);
	}

	Beagle_StackTraceEndM("void SN::EstimateReconstructOp::operate(Deme& ioDeme, Context& ioContext)");
}


/*!
 *  \brief Estimate comparator addition of an individual
 *  \param inIndividual Individual to estimate.
 *  \param ioContext Context of the evolution.
 */
void SN::EstimateReconstructOp::estimate(const Beagle::Individual& inIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	const SN::Individual& lIndividual = castObjectT<const SN::Individual&>(inIndividual);

	if (inIndividual.size() > 1) {
		throw Beagle_RunTimeExceptionM("estimate() not implemented for individuals of size > 1");
    }

	const SN::Network& lNetwork = *lIndividual[0];
    unsigned int lNumInputs = lNetwork.getNumInputs();
    const std::vector<unsigned int>& lLineStates = lNetwork.getLineStates();

    // Initialize state of the network.
    std::vector<unsigned int> lState(lNumInputs*2);
    for (unsigned int i = 0; i < lNumInputs; i++) {
        lState[2*i] = Function::getInputState();
        lState[2*i+1] = Function::getInputStateDual();
    }

    for (unsigned int i=0; i < lNetwork.getNumComparators(); ++i) {
        unsigned int lJoin = lNetwork[2*i];
		unsigned int lMeet = lNetwork[2*i+1];
        incrementFrequency(lState, lJoin, lMeet);

        // Update state of the network.
        lState[2*lJoin] = lLineStates[4*i];
        lState[2*lJoin+1] = lLineStates[4*i+1];
        lState[2*lMeet] = lLineStates[4*i+2];
        lState[2*lMeet+1] = lLineStates[4*i+3];
    }

	Beagle_StackTraceEndM("void SN::EstimateReconstructOp::estimate(const Beagle::Individual& inIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Reconstruct an SN individual by adding comparators.
 *  \param ioIndividual SN individual to reconstruct.
 *  \param inParent Parent from which the individual will be reconstructed.
 *  \param ioContext Context of the evolution.
 */
void SN::EstimateReconstructOp::reconstruct(Beagle::Individual& ioIndividual, const Beagle::Individual& inParent, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	SN::Individual& lIndividual = castObjectT<SN::Individual&>(ioIndividual);
	SN::Network& lNetwork = *lIndividual[0];
    unsigned int lNumInputs = lNetwork.getNumInputs();
	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

	if (ioIndividual.size() > 1) {
		throw Beagle_RunTimeExceptionM("reconstruct() not implemented for individuals of size > 1");
    }

    unsigned int lUniformIndex = 0;
    const std::vector<Network::HeuristicMethod>&  lHeuristicMethods = lNetwork.getHeuristicMethods();
    for (; lUniformIndex < lHeuristicMethods.size(); ++lUniformIndex) {
        if (lHeuristicMethods[lUniformIndex] == Network::eUniformHeuristic) {
            break;
        }
    }

	const SN::Individual& lParent = castObjectT<const SN::Individual&>(inParent);
	const SN::Network& lParentNetwork = *lParent[0];

    // Find the index to truncate the parent network.
    unsigned int lNumOldComparators = lParentNetwork.size()/2;
    int lTruncateIndex = mGaussianTruncate->getWrappedValue() ? 
        lRandomizer.rollGaussian(lNumOldComparators/2, lNumOldComparators/4) : lRandomizer.rollInteger(0, lNumOldComparators);
    if (lTruncateIndex < 0) lTruncateIndex = 0;  // all comparators removed
    if (lTruncateIndex > lNumOldComparators) lTruncateIndex = lNumOldComparators;  // no comparators removed
    std::vector<unsigned int> lKeepComparators(lTruncateIndex*2);
    for (unsigned int i = 0; i < lKeepComparators.size(); ++i) {
        lKeepComparators[i] = lParentNetwork[i];
    }

    lNetwork.clear();
	lNetwork.getStages().clear();
	lNetwork.getLineStates().clear();
    lNetwork.allocateDataStructures();

    // Initialize the monotone Boolean functions of the network.
    std::vector<Function>& lFunctions = lNetwork.getFunctions();
    for (unsigned int i = 0; i < lNumInputs; i++) {
        lFunctions[i].setInput(i);
    }

    lNetwork.calculateHeuristic();

    // Initialize state of the network.
    std::vector<unsigned int> lState(lNumInputs*2);
    for (unsigned int i = 0; i < lNumInputs; i++) {
        lState[2*i] = Function::getInputState();
        lState[2*i+1] = Function::getInputStateDual();
    }

    // Add the comparators that we want to keep in the network.
    const std::vector<unsigned int>& lLineStates = lNetwork.getLineStates();
    __gnu_cxx::slist<unsigned int>& lAsymmetricComparators = lNetwork.getAsymmetricComparators();
    for (unsigned int i = 0; i < lKeepComparators.size()/2; ++i) {
        unsigned int lJoin = lKeepComparators[2*i];
        unsigned int lMeet = lKeepComparators[2*i+1];

        // If a matching comparator was not found in the list of asymmetric
        // comparators, and if the comparator is itself not symmetric, then
        // add it to the list.
        if (lMeet != lNumInputs-1-lJoin && mMutationOp->removeSymmetricComparator(lNetwork, lAsymmetricComparators, lJoin, lMeet) == false) {
            lAsymmetricComparators.push_front(lNetwork.getNumComparators());
        }
        lNetwork.addComparator(lJoin, lMeet);

        // Update state of the network.
        lState[2*lJoin] = lLineStates[4*i];
        lState[2*lJoin+1] = lLineStates[4*i+1];
        lState[2*lMeet] = lLineStates[4*i+2];
        lState[2*lMeet+1] = lLineStates[4*i+3];
    }

    while (lNetwork.isSorted() == false) {
        unsigned int lOldNetworkSize = lNetwork.size();
        std::vector<unsigned int> lIndexPairs;
        RouletteT<unsigned int> lRoulette;
        findComparatorRoulette(lState, lIndexPairs, lRoulette);

        // Use heuristic to add comparator if model doesn't suggest anything.
        if (lIndexPairs.size() == 0) {
            mMutationOp->mutate(lNetwork, ioContext);
        }
        // Use model to add comparator probabilistically.
        else if (lRandomizer.rollUniform(0.0, 1.0) < mEstimateProbability->getWrappedValue() || lUniformIndex == lHeuristicMethods.size()) {
            unsigned int k = lRoulette.select(lRandomizer);
            unsigned int lJoin = lIndexPairs[2*k];
            unsigned int lMeet = lIndexPairs[2*k+1];
            if (findUniformUtility(lNetwork, lJoin, lMeet) > 0) {
                lNetwork.addComparator(lJoin, lMeet);
            }
            // Use heuristics if model suggested useless comparator.
            else {
                mMutationOp->mutate(lNetwork, ioContext);
            }
        }
        else {
            std::vector<unsigned int> lIndexPairs;
            lNetwork.nonzeroHeuristic(lIndexPairs, lUniformIndex);
            if (lIndexPairs.size() > 0) {
                unsigned int k = lRandomizer.rollInteger(0, lIndexPairs.size()/2-1);
                unsigned int lJoin = lIndexPairs[2*k];
                unsigned int lMeet = lIndexPairs[2*k+1];
                lNetwork.addComparator(lJoin, lMeet);
            }
        }

        // Update state of the network.
        unsigned int lNewNetworkSize = lNetwork.size();
        if (lNewNetworkSize > lOldNetworkSize) {
            unsigned int lJoin = lNetwork[lNewNetworkSize-2];
            unsigned int lMeet = lNetwork[lNewNetworkSize-1];
            lState[2*lJoin] = lFunctions[lJoin].getState();
            lState[2*lJoin+1] = lFunctions[lJoin].getStateDual();
            lState[2*lMeet] = lFunctions[lMeet].getState();
            lState[2*lMeet+1] = lFunctions[lMeet].getStateDual();
        }
        else {
            throw Beagle_RunTimeExceptionM("Could not add comparator!");
        }
    }

    // Since we don't need the functions anymore, invalidate them to free memory.
    for (unsigned int i = 0; i < lNumInputs; i++) {
        lFunctions[i].invalidate();
    }

    if (ioIndividual.getFitness() != NULL) ioIndividual.getFitness()->setInvalid();

	Beagle_StackTraceEndM("void SN::EstimateReconstructOp::reconstruct(Beagle::Individual& ioIndividual, const Beagle::Individual& inParent, Beagle::Context& ioContext)");
}


/*!
 *  \brief Increment the frequency of comparator (inJoin, inMeet) for state inState
 */
void SN::EstimateReconstructOp::incrementFrequency(const std::vector<unsigned int>& inState, unsigned int inJoin, unsigned int inMeet)
{
	Beagle_StackTraceBeginM();

    std::pair<const std::vector<unsigned int>, std::vector<std::vector<unsigned int> > > lValue(inState, std::vector<std::vector<unsigned int> >());
    std::pair<FrequenciesMap::iterator, bool> lStatus = mFrequenciesMap.insert(lValue);

    std::vector<std::vector<unsigned int> >& lFrequencies = (*(lStatus.first)).second;

    // If the value was actually inserted, then inState was not in the map before,
    // so we need to initialize the frequencies array.
    if (lStatus.second) {
        unsigned int lNumInputs = mNumInputs->getWrappedValue();
        lFrequencies.resize(lNumInputs);
        for (unsigned int i = 0; i < lNumInputs; i++) {
            lFrequencies[i].resize(lNumInputs-i-1, 0);
        }
    }

    unsigned int x = inJoin;
    unsigned int y = inMeet-inJoin-1;
    lFrequencies[x][y]++;

	Beagle_StackTraceEndM("void SN::EstimateReconstructOp::incrementFrequency(const std::vector<unsigned int>& inState, unsigned int inJoin, unsigned int inMeet)");
}


/*!
 *  \brief Find all index pairs of comparators with non-zero frequencies
 *  \param outIndexPairs The index pairs of comparators with non-zero frequencies
 *  \param outRoulette Roulette wheel for probabilistic selection of the index pairs.
 */
void SN::EstimateReconstructOp::findComparatorRoulette(const std::vector<unsigned int>& inState, std::vector<unsigned int>& outIndexPairs, RouletteT<unsigned int>& outRoulette) const
{
	Beagle_StackTraceBeginM();

    outIndexPairs.clear();
    outRoulette.clear();

    FrequenciesMap::const_iterator lIter = mFrequenciesMap.find(inState);
    if (lIter == mFrequenciesMap.end()) {  // inState not found in the frequencies map
        return;
    }

    const std::vector<std::vector<unsigned int> >& lFrequencies = (*lIter).second;
    for (unsigned int i = 0; i < lFrequencies.size(); ++i) {
        for (unsigned int j = 0; j < lFrequencies[i].size(); ++j) {
            if (lFrequencies[i][j] > 0) {
                outRoulette.insert(outIndexPairs.size()/2, lFrequencies[i][j]);
                outIndexPairs.push_back(i);
                outIndexPairs.push_back(j+i+1);
            }
        }
    }

	Beagle_StackTraceEndM("void SN::EstimateReconstructOp::findComparatorRoulette(const std::vector<unsigned int>& inState, std::vector<unsigned int>& outIndexPairs, RouletteT<unsigned int>& outRoulette) const");
}

