
/*!
 *  \file   beagle/SN/src/EvaluationOp.cpp
 *  \brief  Source code of class SN::EvaluationOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/SN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a evaluation operator.
 *  \param inMutationOp Mutation operator to use before evaluation.
 *  \param inName Name of the operator.
 */
SN::EvaluationOp::EvaluationOp(const std::string& inName) :
		Beagle::EvaluationOp(inName),
        mMutationOp(new SN::MutationAddComparatorOp)
{ }


/*!
 *  \brief Register the parameters of the evaluation operator.
 *  \param ioSystem System of the evolution.
 */
void SN::EvaluationOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Beagle::EvaluationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of inputs of the sorting network",
		    "UInt",
		    "0",
		    "Number of inputs of the sorting network."
		);
		mNumInputs = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("sn.init.numinputs", new UInt(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to trace the operation of the network",
		    "Bool",
		    "0",
		    "If true, the intermediate monotone Boolean functions of the network will be printed out."
		);
		mTrace = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.eval.trace", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to use number of comparators in the network as fitness",
		    "Bool",
		    "0",
		    "If true, the number of comparators in the network is used as fitness."
		);
		mComparatorFitness = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.eval.compfit", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to use mutation operator to add comparators",
		    "Bool",
		    "0",
		    "If true, the add-comparator mutation operator is used to add comparators and make a valid sorting network."
		);
		mUseMutationOp = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.eval.mutate", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Probability to choose comparator for mutation using a roulette wheel",
		    "Float",
		    "0.0",
		    "Probability that the comparator for mutation is chosen using a roulette wheel."
		);
		mApplyRouletteProb = castHandleT<Float>(
		                  ioSystem.getRegister().insertEntry("sn.mutation.roulette", new Float(0.0), lDescription));
	}
    if (mMutationOp != NULL) {
        mMutationOp->registerParams(ioSystem);
    }
	Beagle_StackTraceEndM("void SN::EvaluationOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize operator.
 */
void SN::EvaluationOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
    // Initialize SN::Function here because this function is guaranteed to be called
    // whether beagle is called in evolving or testing modes.
    SN::Function::initialize(mNumInputs->getWrappedValue()+1);
    if (mMutationOp != NULL) {
        mMutationOp->init(ioSystem);
    }
	Beagle_StackTraceEndM("void SN::EvaluationOp::init(Beagle::System& ioSystem)");
}


/*!
 *  \brief Evaluate the fitness of the given SN individual.
 *  \param inIndividual Current individual to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness value of the SN individual.
 */
Fitness::Handle SN::EvaluationOp::evaluate(Beagle::Individual& inIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	SN::Individual& lIndividual = castObjectT<SN::Individual&>(inIndividual);
    SN::Network& lNetwork = *lIndividual[0];

    // Apply mutation operation to complete a copy of the network to estimate fitness.
    if (mUseMutationOp->getWrappedValue()) {
        SN::Network lNewNetwork = lNetwork;

        // Disable roulette wheel selection of comparator for mutation and restore its
        // original value before returning.
        float lApplyRouletteProb = mApplyRouletteProb->getWrappedValue();
        mApplyRouletteProb->setWrappedValue(0.0);
        while(mMutationOp->mutate(lNewNetwork, ioContext));
        mApplyRouletteProb->setWrappedValue(lApplyRouletteProb);

        return evaluate(lNewNetwork, ioContext);
    }

    return evaluate(lNetwork, ioContext);

	Beagle_StackTraceEndM("Fitness::Handle SN::EvaluationOp::evaluate(Beagle::Individual& inIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Evaluate the fitness of the given sorting network.
 *  \param ioNetwork Current sorting network to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness value of the sorting network.
 */
Fitness::Handle SN::EvaluationOp::evaluate(SN::Network& ioNetwork, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

    //unsigned int lNumSorted = ioNetwork.findNumSorted();
    // Maximize the number of sorted inputs and minimize the number of comparators.
    //unsigned int lFitness = lNumSorted + lNumInputs*lNumInputs - ioNetwork.getNumComparators();
    // Maximize the number of sorted inputs.
    //unsigned int lFitness = lNumSorted;

    //unsigned int lDistance = ioNetwork.findHammingDistance();
    //int lFitness = -1*((int)lDistance);

    if (mTrace->getWrappedValue()) {
        ioNetwork.trace();
    }

    // If mutation operator is non-null, we are evolving valid networks for
    // minimal number of comparators, so use that as fitness; otherwise use
    // a fitness measure that increases with the network's sorting ability.
    float lFitness;
    if (mComparatorFitness->getWrappedValue() || mUseMutationOp->getWrappedValue()) {
        lFitness = -1*(float)ioNetwork.getNumComparators();
    }
    else {
        //lFitness = ioNetwork.findNumSorted();
        //lFitness = ioNetwork.findNumSets();
        lFitness = ioNetwork.findNumFullLevels();
    }

    return new FitnessSimple((float)lFitness);

	Beagle_StackTraceEndM("Fitness::Handle SN::EvaluationOp::evaluate(SN::Network& ioNetwork, Beagle::Context& ioContext)");
}
