
/*!
 *  \file   beagle/SN/src/Function.cpp
 *  \brief  Implementation of the type SN::Function.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include <algorithm>
#include <cfloat>

#include "beagle/SN.hpp"

using namespace Beagle;

unsigned int                    SN::Function::smNumLevels;
std::vector<unsigned int>       SN::Function::smMaxSets;


/*!
 *  \brief Read a function from an XML subtree.
 *  \param inIter XML iterator from which the function is read.
 *  \throw Beagle::IOException If a reading error occurred (generally bad file format)!
 */
void SN::Function::read(PACC::XML::ConstIterator inIter)
{
	Beagle_StackTraceBeginM();
	if((inIter->getType() != PACC::XML::eData) || (inIter->getValue() != "Function")) {
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Function> expected!");
	}

    mNumSets.clear();
    mNumSets.resize(smNumLevels, 0);
    mTruthTable.clear();
    mTruthTable.resize(smNumLevels);

	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if(lChild->getType() == PACC::XML::eData && lChild->getValue() == "BitSet") {
            std::string lLevelText = inIter->getAttribute("level");
            unsigned int lLevel = (lLevelText.empty()==false) ? str2uint(lLevelText) : UINT_MAX;

            std::string lNumSetsText = inIter->getAttribute("numSets");
            mNumSets[lLevel] = (lNumSetsText.empty()==false) ? str2uint(lNumSetsText) : UINT_MAX;

            PACC::XML::ConstIterator lBitSet = inIter->getFirstChild();
            if(lBitSet->getType()!=PACC::XML::eString)
                throw Beagle_IOExceptionNodeM(*lBitSet,"Expected bitset!");

            std::istringstream lISS(lBitSet->getValue());
            lISS >> mTruthTable[lLevel];
        }
    }

	Beagle_StackTraceEndM("void SN::Function::read(PACC::XML::ConstIterator)");
}


/*!
 *  \brief Write a function into a XML streamer.
 *  \param ioStreamer XML streamer into which the function is written.
 *  \param inIndent Whether output should be indented.
 *  \param inIndex Index of this function in a vector of functions.
 */
void SN::Function::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inIndex) const
{
	Beagle_StackTraceBeginM();

    ioStreamer.openTag("Function", inIndent);
    ioStreamer.insertAttribute("index", uint2str(inIndex));
    for (unsigned int i = 0; i < smNumLevels; ++i) {
        ioStreamer.openTag("BitSet", inIndent);
        ioStreamer.insertAttribute("level", uint2str(i));
        ioStreamer.insertAttribute("numSets", uint2str(mNumSets[i]));
        std::ostringstream lOSS;
        lOSS << mTruthTable[i];
        ioStreamer.insertStringContent(lOSS.str());
        ioStreamer.closeTag();
    }
    ioStreamer.closeTag();

	Beagle_StackTraceEndM("void SN::Function::write(PACC::XML::Streamer& ioStreamer, bool inIndent, unsigned int inIndex) const");
}


/*!
 *  \brief Check if this function (left) is equal to the given function (right)
 *  \param inRight The function to compare with
 *  \return True if the functions are equal; false otherwise
 */
bool SN::Function::isEqual(const SN::Function& inRight) const
{
	Beagle_StackTraceBeginM();

	const Function& lLeft = *this;
	const Function& lRight = inRight;

    bool lEqual = true;
    for (unsigned int i = 0; lEqual && i < smNumLevels; ++i) {
        lEqual = (lLeft.mNumSets[i] == lRight.mNumSets[i]);
    }

    // If both functions have the same number of sets in each level,
    // check if the bit sets are actually the same.
    for (unsigned int i = 0; lEqual && i < smNumLevels; ++i) {
        lEqual = (lLeft.mTruthTable[i] == lRight.mTruthTable[i]);
    }

	return lEqual;

	Beagle_StackTraceEndM("bool SN::Function::isEqual(const SN::Function& inRight) const");
}


/*!
 *  \brief Check if this function (left) is less than or equal to the given function (right)
 *  \param inRight The function to compare with
 *  \return True if the left function is less than or equal to the right; false otherwise
 *  
 *  Since the functions represent lower sets, the left function is less than or
 *  equal to the right function if the left is contained in the right.
 */
bool SN::Function::isLessEqual(const SN::Function& inRight) const
{
	Beagle_StackTraceBeginM();

	const Function& lLeft = *this;
	const Function& lRight = inRight;

    bool lLessEqual = true;
    for (unsigned int i = 0; lLessEqual && i < smNumLevels; ++i) {
        lLessEqual = (lLeft.mNumSets[i] <= lRight.mNumSets[i]);
    }

    for (unsigned int i = 0; lLessEqual && i < smNumLevels; ++i) {
        lLessEqual = (lLeft.mNumSets[i] == 0) || (lRight.mNumSets[i] == smMaxSets[i]) || lLeft.mTruthTable[i].is_subset_of(lRight.mTruthTable[i]);
    }

	return lLessEqual;

	Beagle_StackTraceEndM("bool SN::Function::isLessEqual(const SN::Function& inRight) const");
}


/*!
 *  \brief Check if this function (left) can be compared to the given function (right)
 *  \param inRight The function to compare with
 *  \return True if the left function is comparable to the right; false otherwise
 */
bool SN::Function::isComparable(const SN::Function& inRight) const
{
	Beagle_StackTraceBeginM();

	const Function& lLeft = *this;
	const Function& lRight = inRight;

    bool lSubset = true;
    bool lSuperset = true;
    bool lComparable = lSubset || lSuperset;
    for (unsigned int i = 0; lComparable && i < smNumLevels; ++i) {
        lSubset = lSubset && (lLeft.mNumSets[i] <= lRight.mNumSets[i]);
        lSuperset = lSuperset && (lLeft.mNumSets[i] >= lRight.mNumSets[i]);
        lComparable = lSubset || lSuperset;
    }

    for (unsigned int i = 0; lComparable && i < smNumLevels; ++i) {
        lSubset = lSubset && ((lLeft.mNumSets[i] == 0) || (lRight.mNumSets[i] == smMaxSets[i]) || lLeft.mTruthTable[i].is_subset_of(lRight.mTruthTable[i]));
        lSuperset = lSuperset && ((lRight.mNumSets[i] == 0) || (lLeft.mNumSets[i] == smMaxSets[i]) || lRight.mTruthTable[i].is_subset_of(lLeft.mTruthTable[i]));
        lComparable = lSubset || lSuperset;
    }

	return lComparable;

	Beagle_StackTraceEndM("bool SN::Function::isComparable(const SN::Function& inRight) const");
}


/*!
 *  \brief Initialize the static members of the Function class.
 */
void SN::Function::initialize(unsigned int inNumLevels)
{
	Beagle_StackTraceBeginM();

    smNumLevels = inNumLevels;
    smMaxSets.resize(inNumLevels);
    for (unsigned int i = 0; i < inNumLevels; ++i) {
        smMaxSets[i] = binomial(inNumLevels-1, i);
    }

	Beagle_StackTraceEndM("void SN::Function::initialize(unsigned int inNumLevels)");
}


/*!
 *  \brief Set the input function of the given sorting network line.
 */
void SN::Function::setInput(unsigned int inLine)
{
	Beagle_StackTraceBeginM();

    mNumSets.clear();
    mNumSets.resize(smNumLevels, 0);
    mTruthTable.clear();
    mTruthTable.resize(smNumLevels);

    mNumSets[0] = 0;  // Function false at null set of variables (truth table not stored).
    unsigned int lMask = 1 << inLine;
    for (unsigned int i = 1; i < smNumLevels-1; ++i) {
        mTruthTable[i].resize(smMaxSets[i], false);
        for (unsigned int j = 0, lComb = firstComb(i); j < smMaxSets[i]; ++j, lComb = nextComb(lComb)) {
            if (lComb & lMask) {
                mTruthTable[i][j] = true;
                mNumSets[i]++;
            }
        }
    }
    mNumSets[smNumLevels-1] = smMaxSets[smNumLevels-1];  // Function true at set of all variables (truth table not stored).

	Beagle_StackTraceEndM("void SN::Function::setInput(unsigned int inLine)");
}



/*!
 *  \brief Set the output function of the given sorting network line.
 */
void SN::Function::setOutput(unsigned int inLine)
{
	Beagle_StackTraceBeginM();

    mNumSets.clear();
    mNumSets.resize(smNumLevels, 0);
    mTruthTable.clear();
    mTruthTable.resize(smNumLevels);

    for (unsigned int i = inLine+1; i < smNumLevels; ++i) {
        mNumSets[i] = smMaxSets[i];
    }

	Beagle_StackTraceEndM("void SN::Function::setOutput(unsigned int inLine)");
}



/*!
 *  \brief Print the number of sets in the function.
 *  \param inWidth Width of the bit-field.
 */
void SN::Function::printNumSets() const
{
	Beagle_StackTraceBeginM();

    bool lFirst = true;
    for (unsigned int i = 0; i < mNumSets.size(); ++i) {
        if (mNumSets[i] > 0) {
            if (!lFirst) std::cout << ", ";
            std::cout << i << " -> " << mNumSets[i];
            lFirst = false;
        }
    }

	Beagle_StackTraceEndM("void SN::Function::printNumSets() const");
}
