
/*!
 *  \file   beagle/SN/src/InitializationOp.cpp
 *  \brief  Source code of class SN::InitializationOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/SN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a plain new initialization operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register.
 *  \param inName Name of the initialization operator.
 */
SN::InitializationOp::InitializationOp(std::string inReproProbaName, std::string inName) :
		Beagle::InitializationOp(inReproProbaName, inName)
{ }


/*!
 *  \brief Register the parameters of the generic SN intialization operator.
 *  \param ioSystem System of the evolution.
 */
void SN::InitializationOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	Beagle::InitializationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of inputs of the sorting network",
		    "UInt",
		    "0",
		    "Number of inputs of the sorting network."
		);
		mNumInputs = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("sn.init.numinputs", new UInt(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to initialize the monotone Boolean functions",
		    "Bool",
		    "0",
		    "If true, the monotone Boolean functions are initialized; otherwise they are not initialized."
		);
		mInitFuncs = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.init.initfuncs", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to perform heuristic search for a valid sorting network",
		    "Bool",
		    "0",
		    "If true, the specified heuristics are used to search for a valid sorting network."
		);
		mHeuristicSearch = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.init.search", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "The depth bound for heuristic search",
		    "UInt",
		    "0",
		    "The depth bound for heuristic search; value of 0 indicates no bound."
		);
		mSearchDepthBound = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("sn.init.bound", new UInt(0), lDescription));
	}
	{
        std::ostringstream lOSS;
        lOSS << "Heuristic methods used to compute non-dominating set of candidate comparators. ";
		lOSS << "The format of an UIntArray is H1:H2:...:Hn, where Hi is the ith value. ";
        lOSS << "Permissible values are 0: uniform heuristic, 1: dominating components, 2: new sorted, ";
        lOSS << "3: new sets, 4: weighted new sets, 5: target distance reduction, 6: new sets potential, ";
        lOSS << "7: migration cost reduction, 8: same weight new sets, 9: bottom-up binary tree, ";
        lOSS << "10: bottom-up binary tree dual, 11: comparator utility, 12: next line utility, ";
        lOSS << "13: next next line utility, 14: delta zero bit count lines, 15: all lines utility, ";
        lOSS << "16: all lines look ahead utility, 17: effect next line comparators, 18: non-zero meet line, ";
        lOSS << "19: maximum matching cardinality, 20: symmetric lines measure, 21: combined lines utility, ";
        lOSS << "22: meet proximity, 23: average look ahead utility, 24: first non-empty meet index, ";
        lOSS << "25: first non-empty index variance, 26: match meet with lines, 27: symmetric functions before, ";
        lOSS << "28: symmetric functions after, 29: filled levels, 30: destination line, 31: disagreement, ";
        lOSS << "100: auto all lines utility";
		Register::Description lDescription(
		    "Heuristic methods used to compute non-dominating set of candidate comparators",
		    "UIntArray",
		    "",
		    lOSS.str()
		);
		mNondomMethods = castHandleT<UIntArray>(
                          ioSystem.getRegister().insertEntry("sn.init.nondom", new UIntArray, lDescription));
	}
	{
        std::ostringstream lOSS;
        lOSS << "Heuristic methods used to filter the non-dominating set of candidate comparators. ";
		lOSS << "The format of an UIntArray is H1:H2:...:Hn, where Hi is the ith value. ";
        lOSS << "Permissible values are same as that for the parameter sn.init.nondom";
		Register::Description lDescription(
		    "Heuristic methods used to filter the set of candidate comparators",
		    "UIntArray",
		    "",
		    lOSS.str()
		);
		mFilterMethods = castHandleT<UIntArray>(
                          ioSystem.getRegister().insertEntry("sn.init.filter", new UIntArray, lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to randomly choose the number of lines used for heuristics in eAutoAllLinesUtility",
		    "Bool",
		    "0",
		    "If true, the number of lines used for heuristics in eAutoAllLinesUtility is randomly chosen; otherwise all lines are used."
		);
		mRandomNumAllLines = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.init.randnumlines", new Bool(0), lDescription));
	}
	Beagle_StackTraceEndM("void SN::InitializationOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize the genotype of an individual.
 *  \param outIndividual Individual to initialize.
 *  \param ioContext Evolution context.
 */
void SN::InitializationOp::initIndividual(Beagle::Individual& outIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	SN::Individual& lIndividual = castObjectT<SN::Individual&>(outIndividual);

	// We just have one genotype in the individual.
    const Factory& lFactory = ioContext.getSystem().getFactory();
    Genotype::Alloc::Handle lGenotypeAlloc =
        castHandleT<Genotype::Alloc>(lFactory.getConceptAllocator("Genotype"));
    lIndividual.clear();
    lIndividual.push_back(castHandleT<Genotype>(lGenotypeAlloc->allocate()));

	Network::Handle lOldNetworkHandle = castHandleT<Network>(ioContext.getGenotypeHandle());
	unsigned int lOldNetworkIndex = ioContext.getGenotypeIndex();

	for(unsigned int i=0; i<lIndividual.size(); ++i) {
		ioContext.setGenotypeHandle(lIndividual[i]);
		ioContext.setGenotypeIndex(i);
		initNetwork(*lIndividual[i], ioContext);
	}

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "initialization", "Beagle::SN::InitializationOp",
	    std::string("Initialized individual: ")+lIndividual.serialize()
	);

	ioContext.setGenotypeIndex(lOldNetworkIndex);
	ioContext.setGenotypeHandle(lOldNetworkHandle);

	Beagle_StackTraceEndM("void SN::InitializationOp::initIndividual(Beagle::Individual& outIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Initialize a genotype network.
 *  \param outNetwork Network to initialize.
 *  \param ioContext Evolution context.
 */
void SN::InitializationOp::initNetwork(SN::Network& outNetwork, Beagle::Context& ioContext) const
{
	Beagle_StackTraceBeginM();

    unsigned int lNumInputs = mNumInputs->getWrappedValue();
    bool lInitFuncs = mInitFuncs->getWrappedValue();
    bool lRandomNumAllLines = mRandomNumAllLines->getWrappedValue();

	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    outNetwork.mNumInputs = lNumInputs;
    outNetwork.mNumAllLines = lRandomNumAllLines ? lRandomizer.rollInteger(1, lNumInputs/2) : lNumInputs/2;

    if (mNondomMethods->size() > 0) {
        throw Beagle_RunTimeExceptionM("Calculation of nondom heuristics is broken due to optimization in Network::updateHeuristics().");
    }

    outNetwork.mHeuristicMethods.clear();
    for (unsigned int i = 0; i < mNondomMethods->size(); ++i) {
        unsigned int lHeuristicMethod = (*mNondomMethods)[i];
        if (lHeuristicMethod == SN::Network::eAutoAllLinesUtility) {
            unsigned int lNumLines = lNumInputs/2;
            for (unsigned int j = 0; j < lNumLines; ++j) {
                outNetwork.mHeuristicMethods.push_back((SN::Network::HeuristicMethod)(lHeuristicMethod+j));
            }
        }
        else {
            outNetwork.mHeuristicMethods.push_back((SN::Network::HeuristicMethod)lHeuristicMethod);
        }
    }
    outNetwork.mFilterStartIndex = outNetwork.mHeuristicMethods.size();
    for (unsigned int i = 0; i < mFilterMethods->size(); ++i) {
        unsigned int lHeuristicMethod = (*mFilterMethods)[i];
        if (lHeuristicMethod == SN::Network::eAutoAllLinesUtility) {
            unsigned int lNumLines = lNumInputs/2;
            for (unsigned int j = 0; j < lNumLines; ++j) {
                outNetwork.mHeuristicMethods.push_back((SN::Network::HeuristicMethod)(lHeuristicMethod+j));
            }
        }
        else {
            outNetwork.mHeuristicMethods.push_back((SN::Network::HeuristicMethod)lHeuristicMethod);
        }
    }

    outNetwork.clear();
	outNetwork.mStages.clear();
	outNetwork.mLineStates.clear();
    outNetwork.allocateDataStructures();

    if (lInitFuncs) {
        Beagle_AssertM((UINT_MAX << lNumInputs) > 0);

        // Compute functions at the inputs of the network.
        std::vector<SN::Function>& lInputFunctions = outNetwork.getFunctions();
        for (unsigned int i = 0; i < lNumInputs; i++) {
            lInputFunctions[i].setInput(i);
        }

        outNetwork.calculateHeuristic();
    }

	Beagle_LogVerboseM(
	    ioContext.getSystem().getLogger(),
	    "initialization", "Beagle::SN::InitializationOp",
	    std::string("Initializing the ")+
	    uint2ordinal(ioContext.getGenotypeIndex()+1)+std::string(" genotype network.")
	);

    if (mHeuristicSearch->getWrappedValue()) {
        outNetwork.addComparator(1, 4);
        outNetwork.addComparator(3, 6);
        outNetwork.addComparator(2, 5);
        outNetwork.addComparator(5, 6);
        outNetwork.addComparator(2, 3);
        outNetwork.addComparator(0, 1);
        outNetwork.addComparator(1, 4);
        outNetwork.addComparator(0, 2);
        outNetwork.addComparator(4, 6);
        outNetwork.addComparator(3, 5);
        outNetwork.addComparator(2, 3);
        //outNetwork.addComparator(4, 5);
        //outNetwork.addComparator(1, 2);
        //outNetwork.addComparator(2, 5);
        //outNetwork.addComparator(2, 3);
        //outNetwork.addComparator(2, 4);
        //outNetwork.addComparator(3, 4);
        heuristicSearch(outNetwork, ioContext);
    }

	Beagle_StackTraceEndM("void SN::InitializationOp::initNetwork(SN::Network& outNetwork, Beagle::Context& ioContext) const");
}

/*!
 *  \brief Use heuristics to search for a sorting network.
 *  \param ioNetwork Network to search.
 */
void SN::InitializationOp::heuristicSearch(SN::Network& ioNetwork, Beagle::Context& ioContext) const
{
	Beagle_StackTraceBeginM();

    unsigned int lNumInputs = ioNetwork.getNumInputs();
    if (ioNetwork.isSorted()) {
        Beagle_LogBasicM(
            ioContext.getSystem().getLogger(),
            "initialization", "Beagle::SN::InitializationOp",
            std::string("Sorting network already valid!")
        );
        return;
    }

    unsigned int lDepthBound = mSearchDepthBound->getWrappedValue();

    std::vector<unsigned int> lIndexPairs;
    ioNetwork.findCandidateComparators(lIndexPairs);
    unsigned int lNumPairs = lIndexPairs.size()/2;
    bool lSucceeded = false;
    for (unsigned i = 0; lSucceeded == false && i < lNumPairs; ++i) {
		Beagle_LogBasicM(
		    ioContext.getSystem().getLogger(),
		    "initialization", "Beagle::InitializationOp",
		    std::string("Trying comparator (")+uint2str(lIndexPairs[2*i])+", "+uint2str(lIndexPairs[2*i+1])+") in "
            +uint2str(i+1)+" of "+uint2str(lNumPairs)+" iterations "+"at the first level of heuristic search."
		);
        // Value of 0 for lDepthBound indicates no bound because lDepthBound-1 = UINT_MAX
        lSucceeded = DFSVisit(ioNetwork, lIndexPairs[2*i], lIndexPairs[2*i+1], lDepthBound-1);
    }

    if (lSucceeded) {
        Beagle_LogBasicM(
            ioContext.getSystem().getLogger(),
            "initialization", "Beagle::SN::InitializationOp",
            std::string("Heuristic search succeeded!")
        );
    }
    else {
        Beagle_LogBasicM(
            ioContext.getSystem().getLogger(),
            "initialization", "Beagle::SN::InitializationOp",
            std::string("Heuristic search failed!")
        );
        std::cout << std::endl;
        exit(1);
    }

	Beagle_StackTraceEndM("void SN::InitializationOp::heuristicSearch(SN::Network& ioNetwork, Beagle::Context& ioContext) const");
}


/*!
 *  \brief Use heuristics and DFS to search for a valid sorting network.
 *  \param ioNetwork Network to search.
 *  \param inJoin The join line index of comparator to add at this level of the search.
 *  \param inMeet The meet line index of comparator to add at this level of the search.
 *  \param inDepthBound Maximum depth to which search is continued.
 *  \return True if search succeeded; false otherwise.
 */
bool SN::InitializationOp::DFSVisit(SN::Network& ioNetwork, unsigned int inJoin, unsigned int inMeet, unsigned int inDepthBound) const
{
	Beagle_StackTraceBeginM();

    std::vector<Function>& lFunctions = ioNetwork.getFunctions();

    // Save function values before modifying them so that they can be restored if search fails.
    Function lJoinFunc = lFunctions[inJoin];
    Function lMeetFunc = lFunctions[inMeet];

    // Now find the new values for the functions.
    Function lNewJoinFunc, lNewMeetFunc;
    join(lFunctions[inJoin], lFunctions[inMeet], lNewJoinFunc);
    meet(lFunctions[inJoin], lFunctions[inMeet], lNewMeetFunc);
    lFunctions[inJoin] = lNewJoinFunc;
    lFunctions[inMeet] = lNewMeetFunc;

    // Add comparator to the network.
    ioNetwork.push_back(inJoin);
    ioNetwork.push_back(inMeet);

    // Save copy of heuristic matrix and update it.
    std::vector<std::vector<std::vector<double> > > lHeuristic = ioNetwork.mHeuristic;
    ioNetwork.updateHeuristic(inJoin, inMeet);

    unsigned int lNumInputs = ioNetwork.getNumInputs();
    if (ioNetwork.isSorted()) { // Found valid sorting network.
        return true;
    }
    else if (inDepthBound > 0) {
        std::vector<unsigned int> lIndexPairs;
        ioNetwork.findCandidateComparators(lIndexPairs);
        unsigned int lNumPairs = lIndexPairs.size()/2;
        for (unsigned i = 0; i < lNumPairs; ++i) {
            if (DFSVisit(ioNetwork, lIndexPairs[2*i], lIndexPairs[2*i+1], inDepthBound-1)) {
                return true;
            }
        }
    }

    // Restore state of the network before returning false.
    lFunctions[inJoin] = lJoinFunc;
	lFunctions[inMeet] = lMeetFunc;
    ioNetwork.pop_back();
    ioNetwork.pop_back();
    ioNetwork.mHeuristic = lHeuristic;

    return false;

	Beagle_StackTraceEndM("bool SN::InitializationOp::DFSVisit(SN::Network& ioNetwork, unsigned int inJoin, unsigned int inMeet, unsigned int inDepthBound) const");
}
