
/*!
 *  \file   beagle/SN/src/MutationAddComparatorOp.cpp
 *  \brief  Implementation of the type SN::MutationAddComparatorOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/SN.hpp"

#include <string>
#include <cfloat>

using namespace Beagle;


/*!
 *  \brief Construct add-comparator mutation operator for the SN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
SN::MutationAddComparatorOp::MutationAddComparatorOp(std::string inMutationPbName,
                                                     std::string inName) :
		Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the add-comparator mutation operator.
 *  \param ioSystem System of the evolution.
 */
void SN::MutationAddComparatorOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		Register::Description lDescription(
		    "Individual add-comparator mutation probability",
		    "Float",
		    "0.05",
            "Add-comparator mutation probability for an individual."
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	{
		Register::Description lDescription(
		    "Probability to add comparators to make network symmetric",
		    "Float",
		    "0.0",
		    "Probability that symmetric comparators are added in the add-comparator mutation operator to make the network symmetric."
		);
		mMakeSymmetricProb = castHandleT<Float>(
		                  ioSystem.getRegister().insertEntry("sn.mutation.mksym", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to filter comparators using look-ahead heuristics",
		    "Bool",
		    "0",
		    "If true, the look-ahead filter heuristics is applied to narrow the choices of candidate comparators for addition."
		);
		mApplyLookAheadFilter = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.mutation.lookahead", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Probability to filter comparators using maximum cardinality matching",
		    "Float",
		    "0.0",
		    "Probability that maximum cardinality matching is applied to narrow the choices of candidate comparators for addition."
		);
		mApplyMaxMatchingFilterProb = castHandleT<Float>(
		                  ioSystem.getRegister().insertEntry("sn.mutation.maxmatch", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Probability to choose comparator for mutation using a roulette wheel",
		    "Float",
		    "0.0",
		    "Probability that the comparator for mutation is chosen using a roulette wheel."
		);
		mApplyRouletteProb = castHandleT<Float>(
		                  ioSystem.getRegister().insertEntry("sn.mutation.roulette", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether using triplets of comparators instead of singles is allowed",
		    "Bool",
		    "0",
		    "If true, the utility of triplets of comparators is evaluated and used if they are better than two single comparators touching the same lines."
		);
		mUseComparatorTriplets = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.mutation.triplets", new Bool(0), lDescription));
	}
	{
        std::ostringstream lOSS;
        lOSS << "Groups of filter heuristic methods to select comparators. The contents of this array are ";
        lOSS << "indexes into the filter heuristics array and each group begins with the index 0.  ";
		lOSS << "The format of an UIntArray is H1:H2:...:Hn, where Hi is the ith value.";
		Register::Description lDescription(
		    "Groups of filter heuristic methods to select comparators",
		    "UIntArray",
		    "",
		    lOSS.str()
		);
		mFilterGroups = castHandleT<UIntArray>(
                          ioSystem.getRegister().insertEntry("sn.mutation.fltgrps", new UIntArray, lDescription));
	}
	{
		Register::Description lDescription(
		    "Probabilities of the filter groups",
		    "FloatArray",
		    "",
		    "Probabilities of choosing each group specified in the filter groups array."
		);
		mFilterGroupProbs = castHandleT<FloatArray>(
                          ioSystem.getRegister().insertEntry("sn.mutation.fltgrppbs", new FloatArray, lDescription));
	}
	Beagle::MutationOp::registerParams(ioSystem);
	Beagle_StackTraceEndM("void SN::MutationAddComparatorOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Initialize operator.
 */
void SN::MutationAddComparatorOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
    
    // Make the filter groups vector.
    mFilterGroupVectors.clear();
    for (unsigned int i = 0; i < mFilterGroups->size(); i++) {
        if ((*mFilterGroups)[i] == 0) {
            mFilterGroupVectors.resize(mFilterGroupVectors.size()+1);
        }
        if (mFilterGroupVectors.size() > 0) {
            mFilterGroupVectors.back().push_back((*mFilterGroups)[i]);
        }
        else {
            throw Beagle_RunTimeExceptionM("Filter groups must start with the index 0.");
        }
    }

    // Make a roulette out of the filter group probabilities.
    if (mFilterGroupVectors.size() != mFilterGroupProbs->size()) {
        throw Beagle_RunTimeExceptionM("Filter group probabilities must be matched one-to-one with filter groups.");
    }
    mFilterGroupRoulette.clear();
    for (unsigned int i = 0; i < mFilterGroupVectors.size(); i++) {
        mFilterGroupRoulette.insert(i, (*mFilterGroupProbs)[i]);
    }

	Beagle::MutationOp::init(ioSystem);
	Beagle_StackTraceEndM("void SN::MutationAddComparatorOp::init(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply add-comparator mutation to an SN individual.
 *  \param ioIndividual SN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool SN::MutationAddComparatorOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	SN::Individual& lIndividual = castObjectT<SN::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::SN::MutationAddComparatorOp",
	    std::string("Mutating individual with SN::MutationAddComparatorOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::SN::MutationAddComparatorOp",
	    std::string("Individual before SN add-comparator mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("SN::MutationAddComparatorOp not implemented for individuals of size > 1");

	SN::Network& lNetwork = *lIndividual[0];

    unsigned int lOldSize = lNetwork.getNumComparators();
	if (mutate(lNetwork, ioContext)) {
		// Mutation successful, log messages and return.
        unsigned int lNewSize = lNetwork.getNumComparators();
		std::ostringstream lOSS;
		lOSS << "Added " << lNewSize - lOldSize << " comparators to the network.";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::SN::MutationAddComparatorOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::SN::MutationAddComparatorOp",
		    std::string("Individual after SN add-comparator mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		std::ostringstream lOSS;
		lOSS << "Attempt to add a new comparator failed; ";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::SN::MutationAddComparatorOp",
		    lOSS.str()+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool SN::MutationAddComparatorOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}


/*!
 *  \brief Apply add-comparator mutation to the given network.
 *  \param ioNetwork Network to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the network is effectively mutated, false if not.
 */
bool SN::MutationAddComparatorOp::mutate(SN::Network& ioNetwork, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();
    std::vector<SN::Function>& lFunctions = ioNetwork.getFunctions();
    unsigned int lNumInputs = ioNetwork.getNumInputs();

    // If filter groups are specified, then pick the group to use in this mutation.
    const std::vector<unsigned int>& lFilterGroup = (mFilterGroupRoulette.size() > 0) ?
        mFilterGroupVectors[mFilterGroupRoulette.select(lRandomizer)] : std::vector<unsigned int>();

    unsigned int lJoin;
    unsigned int lMeet;
    bool lMutated = false;
    std::vector<unsigned int> lIndexPairs;
    RouletteT<unsigned int> lRoulette;
    bool lApplyRoulette = lRandomizer.rollUniform(0.0, 1.0) < mApplyRouletteProb->getWrappedValue();
    if (lApplyRoulette) {
        ioNetwork.findCandidateComparators(lIndexPairs, lRoulette, lFilterGroup);
    }
    else {
        ioNetwork.findCandidateComparators(lIndexPairs, lFilterGroup);
        if (mApplyLookAheadFilter->getWrappedValue()) {
            ioNetwork.lookAheadHeuristicFilter(lIndexPairs);
        }
        if (lRandomizer.rollUniform(0.0, 1.0) < mApplyMaxMatchingFilterProb->getWrappedValue()) {
            ioNetwork.maxMatchingFilter(lIndexPairs);
        }
    }
    unsigned int lNumPairs = lIndexPairs.size()/2;
    if (lNumPairs > 0) {
        // Prefer symmetric comparators probabilistically.
        __gnu_cxx::slist<unsigned int>& lAsymmetricComparators = ioNetwork.getAsymmetricComparators();
        if (lRandomizer.rollUniform(0.0, 1.0) < mMakeSymmetricProb->getWrappedValue()) {
            // If the network is not symmetric, and if a comparator that is symmetric
            // is among the index pairs obtained above, then add that comparator.
            if (lAsymmetricComparators.empty() == false) {
                __gnu_cxx::slist<unsigned int>::iterator lPrevIter;
                for (__gnu_cxx::slist<unsigned int>::iterator lIter = lAsymmetricComparators.begin(); lMutated == false && lIter != lAsymmetricComparators.end();) {
                    unsigned int k = *lIter;
                    unsigned int lAsymJoin = ioNetwork[2*k];
                    unsigned int lAsymMeet = ioNetwork[2*k+1];
                    lJoin = lNumInputs-1-lAsymMeet;
                    lMeet = lNumInputs-1-lAsymJoin;
                    for (unsigned int i = 0; lMutated == false && i < lNumPairs; ++i) {
                        if (lJoin == lIndexPairs[2*i] && lMeet == lIndexPairs[2*i+1]) {
                            ioNetwork.addComparator(lJoin, lMeet);
                            lMutated = true;
                        }
                    }
                    // Erase the list element if a symmetric comparator was found and increment
                    // the list iterator.
                    if (lMutated) {
                        // Use erase_after() instead of erase() for constant time deletion.
                        if (lIter == lAsymmetricComparators.begin()) {
                            lAsymmetricComparators.erase(lIter);
                            lIter = lAsymmetricComparators.begin();
                        }
                        else {
                            ++lIter;  // Increment lIter before it gets erased.
                            lAsymmetricComparators.erase_after(lPrevIter);
                            // Now, *lPrevIter is the element before *lIter
                        }
                    }
                    else {
                        lPrevIter = lIter;
                        ++lIter;
                    }
                }
            }
            else {
                // Look for symmetric comparator pairs, and if found, restrict the
                // choices to such comparators.
                std::vector<unsigned int> lSymmetricPairs;
                ioNetwork.findSymmetricComparators(lIndexPairs, lSymmetricPairs);
                if (lSymmetricPairs.size() > 0) {
                    lIndexPairs = lSymmetricPairs;
                    lNumPairs = lIndexPairs.size()/2;
                    unsigned int k = lRandomizer.rollInteger(0, lNumPairs-1);
                    lJoin = lIndexPairs[2*k];
                    lMeet = lIndexPairs[2*k+1];
                    if (lMeet != lNumInputs-1-lJoin) { // The comparator itself is not symmetric and so makes the network asymmetric.
                        lAsymmetricComparators.push_front(ioNetwork.getNumComparators());
                    }
                    ioNetwork.addComparator(lJoin, lMeet);
                    lMutated = true;
                }
            }
        }

        // If we haven't found a comparator to add, choose a random comparator.
        if (lMutated == false) {
            unsigned int k = lApplyRoulette ? lRoulette.select(lRandomizer) : lRandomizer.rollInteger(0, lNumPairs-1);
            lJoin = lIndexPairs[2*k];
            lMeet = lIndexPairs[2*k+1];
            // If the comparator is itself not symmetric, and a matching comparator was not found in the list of asymmetric
            // comparators, then add it to the asymmetric list.
            if (lMeet != lNumInputs-1-lJoin && removeSymmetricComparator(ioNetwork, lAsymmetricComparators, lJoin, lMeet) == false) {
                lAsymmetricComparators.push_front(ioNetwork.getNumComparators());
            }
            ioNetwork.addComparator(lJoin, lMeet);
            lMutated = true;
        }

        if (lMutated && mUseComparatorTriplets->getWrappedValue()) {
            // If the added comparator is part of a triplet, add the remaining comparators also.
            const std::vector<unsigned int>& lTriplet = ioNetwork.getTriplet(lJoin, lMeet);
            for (unsigned int i = 0; i < lTriplet.size()/2; ++i) {
                lJoin = lTriplet[2*i];
                lMeet = lTriplet[2*i+1];

                // If the comparator is itself not symmetric, and a matching comparator was not found in the list of asymmetric
                // comparators, then add it to the asymmetric list.
                if (lMeet != lNumInputs-1-lJoin && removeSymmetricComparator(ioNetwork, lAsymmetricComparators, lJoin, lMeet) == false) {
                    lAsymmetricComparators.push_front(ioNetwork.getNumComparators());
                }
                ioNetwork.addComparator(lJoin, lMeet);
            }

            ioNetwork.adjustHeuristicForTriplets();
        }

    }

    return lMutated;

	Beagle_StackTraceEndM("bool SN::MutationAddComparatorOp::mutate(SN::Network& ioNetwork, Beagle::Context& ioContext)");
}


/*!
 *  \brief If a comparator symmetric to (inJoin, inMeet) is found in the given list of comparators,
 *      that comparator is removed from that list.
 *  \return True if a comparator was found and removed, false otherwise.
 */
bool SN::MutationAddComparatorOp::removeSymmetricComparator(const SN::Network& inNetwork, __gnu_cxx::slist<unsigned int>& ioAsymmetricComparators, unsigned int inJoin, unsigned int inMeet) const
{
	Beagle_StackTraceBeginM();

    unsigned int lNumInputs = inNetwork.getNumInputs();

    // If this comparator is symmetric to one of the asymmetric comparators,
    // then remove that comparator from the list of asymmetric comparators.
    bool lFound = false;
    if (ioAsymmetricComparators.empty() == false) {
        __gnu_cxx::slist<unsigned int>::iterator lPrevIter;
        for (__gnu_cxx::slist<unsigned int>::iterator lIter = ioAsymmetricComparators.begin(); lIter != ioAsymmetricComparators.end();) {
            unsigned int k = *lIter;
            unsigned int lAsymJoin = inNetwork[2*k];
            unsigned int lAsymMeet = inNetwork[2*k+1];
            // Erase the list element if a symmetric comparator is found.
            if (inJoin == lNumInputs-1-lAsymMeet && inMeet == lNumInputs-1-lAsymJoin) {
                // Use erase_after() instead of erase() for constant time deletion.
                if (lIter == ioAsymmetricComparators.begin()) {
                    ioAsymmetricComparators.erase(lIter);
                }
                else {
                    ioAsymmetricComparators.erase_after(lPrevIter);
                }
                lFound = true;
                break;
            }
            else {
                lPrevIter = lIter;
                ++lIter;
            }
        }
    }

    return lFound;
    
	Beagle_StackTraceEndM("bool SN::MutationAddComparatorOp::removeSymmetricComparator(const SN::Network& inNetwork, __gnu_cxx::slist<unsigned int>& ioAsymmetricComparators, unsigned int inJoin, unsigned int inMeet) const");
}
