
/*!
 *  \file   beagle/SN/src/MutationTruncateConstructOp.cpp
 *  \brief  Implementation of the type SN::MutationTruncateConstructOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/SN.hpp"

#include <string>
#include <cfloat>

using namespace Beagle;


/*!
 *  \brief Construct truncate and construct mutation operator for the SN genotype graph.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inName Name of the operator.
 */
SN::MutationTruncateConstructOp::MutationTruncateConstructOp(std::string inMutationPbName,
                                                             std::string inName) :
		SN::MutationAddComparatorOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Register the parameters of the truncate and construct mutation operator.
 *  \param ioSystem System of the evolution.
 */
void SN::MutationTruncateConstructOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	{
		Register::Description lDescription(
		    "Individual truncate and construct mutation probability",
		    "Float",
		    "0.05",
            "Truncate and construct mutation probability for an individual."
		);
		mMutationProba = castHandleT<Float>(
		                     ioSystem.getRegister().insertEntry(mMutationPbName, new Float(0.05f), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to construct a valid network after truncating it",
		    "Bool",
		    "1",
		    "If true, add comparators to construct a valid network after truncating it."
		);
		mConstructValid = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("sn.muttruncons.construct", new Bool(1), lDescription));
	}
	SN::MutationAddComparatorOp::registerParams(ioSystem);
	Beagle_StackTraceEndM("void SN::MutationTruncateConstructOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Apply truncate and construct mutation to an SN individual.
 *  \param ioIndividual SN individual to mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool SN::MutationTruncateConstructOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	SN::Individual& lIndividual = castObjectT<SN::Individual&>(ioIndividual);

	Beagle_LogDetailedM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::SN::MutationTruncateConstructOp",
	    std::string("Mutating individual with SN::MutationTruncateConstructOp")
	);

	Beagle_LogDebugM(
	    ioContext.getSystem().getLogger(),
	    "mutation", "Beagle::SN::MutationTruncateConstructOp",
	    std::string("Individual before SN truncate and construct mutation: ")+
	    ioIndividual.serialize()
	);

	if (ioIndividual.size() > 1)
		throw Beagle_RunTimeExceptionM("SN::MutationTruncateConstructOp not implemented for individuals of size > 1");

	SN::Network& lNetwork = *lIndividual[0];
    unsigned int lNumInputs = lNetwork.getNumInputs();
    unsigned int lNumOldComparators = lNetwork.size()/2;

	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    // Find the index to truncate the network.
    int lTruncateIndex = lRandomizer.rollGaussian(lNumOldComparators/2, lNumOldComparators/4);
    if (lTruncateIndex < 0) lTruncateIndex = 0;  // all comparators removed
    if (lTruncateIndex > lNumOldComparators) lTruncateIndex = lNumOldComparators;  // no comparators removed
    std::vector<unsigned int> lKeepComparators(lTruncateIndex*2);
    for (unsigned int i = 0; i < lKeepComparators.size(); ++i) {
        lKeepComparators[i] = lNetwork[i];
    }

    lNetwork.clear();
	lNetwork.getStages().clear();
	lNetwork.getLineStates().clear();
    lNetwork.allocateDataStructures();

    // Initialize the monotone Boolean functions.
    std::vector<Function>& lFunctions = lNetwork.getFunctions();
    for (unsigned int i = 0; i < lNumInputs; i++) {
        lFunctions[i].setInput(i);
    }

    lNetwork.calculateHeuristic();

    // Add the comparators that we want to keep back to the network.
    __gnu_cxx::slist<unsigned int>& lAsymmetricComparators = lNetwork.getAsymmetricComparators();
    for (unsigned int i = 0; i < lKeepComparators.size()/2; ++i) {
        unsigned int lJoin = lKeepComparators[2*i];
        unsigned int lMeet = lKeepComparators[2*i+1];

        // If a matching comparator was not found in the list of asymmetric
        // comparators, and if the comparator is itself not symmetric, then
        // add it to the list.
        if (lMeet != lNumInputs-1-lJoin && removeSymmetricComparator(lNetwork, lAsymmetricComparators, lJoin, lMeet) == false) {
            lAsymmetricComparators.push_front(lNetwork.getNumComparators());
        }
        lNetwork.addComparator(lJoin, lMeet);
    }

    // Add new comparators and construct a valid network.
    unsigned int lNumNewComparators = 0;
    if (mConstructValid->getWrappedValue()) {
        while (MutationAddComparatorOp::mutate(lNetwork, ioContext)) {
            lNumNewComparators++;
        }
    }

    // Since we don't need the functions anymore, invalidate them to free memory.
    for (unsigned int i = 0; i < lNumInputs; i++) {
        lFunctions[i].invalidate();
    }

	if (lNumOldComparators - lTruncateIndex > 0) {
		// Mutation successful, log messages and return.
		std::ostringstream lOSS;
		lOSS << "Removed " << lNumOldComparators - lTruncateIndex << " comparators ";
        lOSS << "and added " << lNumNewComparators << " new comparators";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::SN::MutationTruncateConstructOp",
		    lOSS.str()
		);

		Beagle_LogDebugM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::SN::MutationTruncateConstructOp",
		    std::string("Individual after SN truncate and construct mutation: ")+
		    ioIndividual.serialize()
		);

		return true;
	} else {
		std::ostringstream lOSS;
		lOSS << "Attempt to truncate network and add a new comparators failed; ";
		Beagle_LogTraceM(
		    ioContext.getSystem().getLogger(),
		    "mutation", "Beagle::SN::MutationTruncateConstructOp",
		    lOSS.str()+
		    std::string("resuming from mutation without modifying the individual")
		);

		return false;
	}

	Beagle_StackTraceEndM("bool SN::MutationTruncateConstructOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)");
}
