
/*!
 *  \file   beagle/SN/src/Network.cpp
 *  \brief  Implementation of the type SN::Network.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/SN.hpp"

#include <cfloat>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/max_cardinality_matching.hpp>

using namespace Beagle;


/*!
 *  \brief Construct a sorting network.
 */
SN::Network::Network(unsigned int inSize, unsigned int inModel) :
		std::vector<unsigned int>(inSize, inModel),
        mNumInputs(0),
        mStages(),
        mFunctions(),
        mHeuristicMethods(),
        mFilterStartIndex(0),
        mJoins(),
        mMeets(),
        mHeuristic(),
        mTriplets(),
        mAsymmetricComparators(),
        mNumAllLines(0),
        mLineStates()
{ }


/*!
 *  \brief Copy sorting network into current.
 *  \param inOriginal Sorting network to copy.
 *  \param ioSystem Evolutionary system to use for making the copy.
 */
void SN::Network::copy(const Member& inOriginal, System& ioSystem)
{
	Beagle_StackTraceBeginM();
	const SN::Network& lOriginal = castObjectT<const SN::Network&>(inOriginal);
	(*this) = lOriginal;
	Beagle_StackTraceEndM("void SN::Network::copy(const Member&,System&)");
}


/*!
 *  \brief Get exact type of the genotype.
 *  \return Exact type of the genotype.
 */
const std::string& SN::Network::getType() const
{
	Beagle_StackTraceBeginM();
	const static std::string lType("SN-Network");
	return lType;
	Beagle_StackTraceEndM("const std::string& SN::Network::getType() const")
}


/*!
 *  \brief  Test if two networks are equals.
 *  \param  inRightObj Second network used for the comparison.
 *  \return True if networks are equal and of same size, false if not.
 */
bool SN::Network::isEqual(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const SN::Network& lRightNetwork = castObjectT<const SN::Network&>(inRightObj);
	if(size()!=lRightNetwork.size()) return false;
	const std::vector<unsigned int>& lLeft = *this;
	const std::vector<unsigned int>& lRight = lRightNetwork;
	return (lLeft == lRight);
	Beagle_StackTraceEndM("bool SN::Network::isEqual(const Object&) const");
}


/*!
 *  \brief  Test the ordering of two networks.
 *  \param  inRightObj Second network used for the ordering test.
 *  \return True if the actual network is before the second, false if not.
 */
bool SN::Network::isLess(const Object& inRightObj) const
{
	Beagle_StackTraceBeginM();
	const SN::Network& lRightNetwork = castObjectT<const SN::Network&>(inRightObj);
	if(size()!=lRightNetwork.size()) return false;
	const std::vector<unsigned int>& lLeft = *this;
	const std::vector<unsigned int>& lRight = lRightNetwork;
	return (lLeft < lRight);
	Beagle_StackTraceEndM("bool SN::Network::isLess(const Object&) const");
}


/*!
 *  \brief Read a network from an XML subtree.
 *  \param inIter XML iterator from which the network is read.
 *  \param ioContext Evolutionary context.
 *  \throw Beagle::IOException If a reading error occurred (generally bad file format)!
 */
void SN::Network::readWithContext(PACC::XML::ConstIterator inIter, Context& ioContext)
{
	Beagle_StackTraceBeginM();
	if((inIter->getType() != PACC::XML::eData) || (inIter->getValue() != "Genotype")) {
		throw Beagle_IOExceptionNodeM(*inIter, "tag <Genotype> expected!");
	}
	std::string lGenotypeType = inIter->getAttribute("type");
	if((lGenotypeType.empty() == false) && (lGenotypeType != getType())) {
		std::ostringstream lOSS;
		lOSS << "type given '" << lGenotypeType << "' does not match type of the genotype '";
		lOSS << getType() << "'!";
		throw Beagle_IOExceptionNodeM(*inIter, lOSS.str());
	}

	std::string lNumInputsText = inIter->getAttribute("numInputs");
	if(lNumInputsText.empty()==false)
		mNumInputs = str2uint(lNumInputsText);

	std::string lNumAllLinesText = inIter->getAttribute("numAllLines");
	if(lNumAllLinesText.empty()==false)
		mNumAllLines = str2uint(lNumAllLinesText);

	for(PACC::XML::ConstIterator lChild=inIter->getFirstChild(); lChild; ++lChild) {
		if(lChild->getType() == PACC::XML::eData) {
            if (lChild->getValue() == "Comparators") {
                readVectorContent(lChild->getFirstChild(), *this);
            }
            else if (lChild->getValue() == "Stages") {
                readVectorContent(lChild->getFirstChild(), mStages);
            }
            else if (lChild->getValue() == "HeuristicMethods") {
                std::string lFilterStartIndexText = lChild->getAttribute("filterStart");
                if(lFilterStartIndexText.empty()==false)
                    mFilterStartIndex = str2uint(lFilterStartIndexText);
                readVectorContent(lChild->getFirstChild(), mHeuristicMethods);
            }
            // Uncomment the following code to read the monotone Boolean functions
            //else if (lChild->getValue() == "Functions") {
            //    mFunctions.clear();
            //    mFunctions.resize(mNumInputs);
            //    unsigned int i=0;
            //    for(PACC::XML::ConstIterator lGrandChild=lChild->getFirstChild(); lGrandChild; ++lGrandChild) {
            //        mFunctions[i].read(lGrandChild);
            //        i++;
            //    }
            //}
        }
    }

	Beagle_StackTraceEndM("void SN::Network::readWithContext(PACC::XML::ConstIterator,Context&)");
}


/*!
 *  \brief Read a string of unsigned int as vector of type T from an XML subtree.
 *  \param inIter XML iterator to use to read the vector from.
 *  \param outVector The vector to read into.
 */
template <class T>
void SN::Network::readVectorContent(PACC::XML::ConstIterator inIter, std::vector<T>& outVector)
{
	Beagle_StackTraceBeginM();

    if(inIter->getType()!=PACC::XML::eString)
        throw Beagle_IOExceptionNodeM(*inIter,"Expected vector of unsigned integers!");

    outVector.clear();
    std::istringstream lISS(inIter->getValue());
    while(lISS.good()) {
        unsigned int lValue;
        lISS >> lValue;
        outVector.push_back((T)lValue);
        if(lISS.good()==false) break;
        int lDelim=lISS.get();
        if((lISS.good()==false) || (lDelim==-1)) break;
    }

	Beagle_StackTraceEndM("void SN::Network::readVectorContent(PACC::XML::ConstIterator inIter, std::vector<T>& outVector)");
}


/*!
 *  \brief Write a network into a XML streamer.
 *  \param ioStreamer XML streamer into which the network is written.
 *  \param inIndent Whether output should be indented.
 */
void SN::Network::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const
{
	Beagle_StackTraceBeginM();

    ioStreamer.insertAttribute("numInputs", uint2str(mNumInputs));
    ioStreamer.insertAttribute("numAllLines", uint2str(mNumAllLines));

    if (size() > 0) {
        ioStreamer.openTag("Comparators", false);
        ioStreamer.insertAttribute("size", uint2str(size()));
        writeVectorContent(ioStreamer, *this);
        ioStreamer.closeTag();
    }

    if (mStages.size() > 0) {
        ioStreamer.openTag("Stages", false);
        ioStreamer.insertAttribute("size", uint2str(mStages.size()));
        writeVectorContent(ioStreamer, mStages);
        ioStreamer.closeTag();
    }

    if (mHeuristicMethods.size() > 0) {
        ioStreamer.openTag("HeuristicMethods", false);
        ioStreamer.insertAttribute("size", uint2str(mHeuristicMethods.size()));
        ioStreamer.insertAttribute("filterStart", uint2str(mFilterStartIndex));
        writeVectorContent(ioStreamer, mHeuristicMethods);
        ioStreamer.closeTag();
    }

    // Uncomment the following code to write the monotone Boolean functions
    //ioStreamer.openTag("Functions", inIndent);
    //ioStreamer.insertAttribute("size", uint2str(mFunctions.size()));
    //for(unsigned int i=0; i<mFunctions.size(); ++i) {
    //    mFunctions[i].write(ioStreamer, inIndent, i);
    //}
    //ioStreamer.closeTag();

	Beagle_StackTraceEndM("void SN::Network::writeContent(PACC::XML::Streamer& ioStreamer, bool inIndent) const");
}


/*!
 *  \brief Write an vector of type T as unsigned int into a XML streamer.
 *  \param ioStreamer XML streamer to write the vector into.
 *  \param inVector The vector to be written.
 */
template <class T>
void SN::Network::writeVectorContent(PACC::XML::Streamer& ioStreamer, const std::vector<T>& inVector) const
{
	Beagle_StackTraceBeginM();

    std::ostringstream lOSS;
    for(unsigned int i=0; i<inVector.size(); ++i) {
        lOSS << (unsigned int)inVector[i];
        if(i != (inVector.size()-1)) lOSS << ';';
    }
    ioStreamer.insertStringContent(lOSS.str());

	Beagle_StackTraceEndM("void SN::Network::writeVectorContent(PACC::XML::Streamer& ioStreamer, const std::vector<T>& inVector) const");
}


/*!
 *	\return The current line of the network that we are working on.
 */
unsigned int SN::Network::getCurrentLine() const
{
	Beagle_StackTraceBeginM();

    // We have completed a line if all sets for that line have been collected
    // to that line and it does not contain sets for any other line.
    unsigned int lLine = 0;
    unsigned int lDualLine = mNumInputs-1;
    while (lLine < lDualLine && isLineSorted(lLine) && isLineSortedDual(lDualLine)) {
        lLine++;
        lDualLine--;
    }

    return lLine;

	Beagle_StackTraceEndM("unsigned int SN::Network::getCurrentLine() const");
}


/*!
 *  \brief Allocate memory for the network data structures
 */
void SN::Network::allocateDataStructures()
{
	Beagle_StackTraceBeginM();

    mFunctions.clear();
    mFunctions.resize(mNumInputs);

    // The join, meet, and heuristic arrays are symmetric with 0 in the diagonal;
    // therefore, only the upper-triangular part is stored.
    mJoins.clear();
    mJoins.resize(mNumInputs);
    mMeets.clear();
    mMeets.resize(mNumInputs);
    mHeuristic.clear();
    mHeuristic.resize(mNumInputs);
    mTriplets.clear();
    mTriplets.resize(mNumInputs);
    for (unsigned int i = 0; i < mNumInputs; i++) {
        unsigned int lNumColumns = mNumInputs-i-1;
        mJoins[i].resize(lNumColumns);
        mMeets[i].resize(lNumColumns);
        mHeuristic[i].resize(lNumColumns);
        mTriplets[i].resize(lNumColumns);
        for (unsigned int j = 0; j < lNumColumns; j++) {
            mHeuristic[i][j].resize(mHeuristicMethods.size(), 0);
        }
    }

    mAsymmetricComparators.clear();

	Beagle_StackTraceEndM("void SN::Network::allocateDataStructures()");
}


/*!
 *  \brief Add a comparator to the network.
 */
void SN::Network::addComparator(unsigned int inJoin, unsigned int inMeet)
{
	Beagle_StackTraceBeginM();

    Function lTempFunc = mFunctions[inJoin];
    join(mFunctions[inJoin], mFunctions[inMeet]);
    meet(mFunctions[inMeet], lTempFunc);

    push_back(inJoin);
    push_back(inMeet);

    updateHeuristic(inJoin, inMeet);

	mLineStates.push_back(mFunctions[inJoin].getState());
	mLineStates.push_back(mFunctions[inJoin].getStateDual());
	mLineStates.push_back(mFunctions[inMeet].getState());
	mLineStates.push_back(mFunctions[inMeet].getStateDual());

	Beagle_StackTraceEndM("void SN::Network::addComparator(unsigned int inJoin, unsigned int inMeet)");
}


/*!
 *  \brief Adjust the heuristic values by checking if using comparator
 *      triplets instead of couplets will give better utilities.
 */
void SN::Network::adjustHeuristicForTriplets()
{
	Beagle_StackTraceBeginM();

    unsigned int lUtilityIndex = 0;
    for (; lUtilityIndex < mHeuristicMethods.size(); ++lUtilityIndex) {
        if (mHeuristicMethods[lUtilityIndex] == eCombinedLinesUtility) {
            break;
        }
    }
    if (lUtilityIndex == mHeuristicMethods.size()) {
        return;
    }

    // Find lines that are already sorted so that we can skip them.
    unsigned int lLine = 0;
    unsigned int lDualLine = mNumInputs-1;
    while (lLine < lDualLine && isLineSorted(lLine)) {
        lLine++;
        lDualLine--;
    }
    unsigned int lCurrentLine = lLine;

    lLine = 0;				 
    lDualLine = mNumInputs-1;
    while (lLine < lDualLine && isLineSorted(lDualLine)) {
        lLine++;
        lDualLine--;
    }
    unsigned int lDualCurrentLine = lDualLine;

    // See if we can form a comparator triplet with (p1, q1) as the first comparator.
    for (unsigned int p1 = lCurrentLine; p1 < lDualCurrentLine; ++p1) {
        for (unsigned int q1 = p1+1; q1 <= lDualCurrentLine; ++q1) {
            if (mFunctions[p1].getNumNonEmptyLevels() == 1 && mFunctions[p1].getNumNonEmptyLevelsDual() == 1 && 
                mFunctions[q1].getNumNonEmptyLevels() == 1 && mFunctions[q1].getNumNonEmptyLevelsDual() == 1) {
                unsigned int x1 = p1;
                unsigned int y1 = q1-p1-1;

                mTriplets[x1][y1].clear();  // reset the entry in the triplets array.

                unsigned int lUtility1 = findCombinedLinesUtility(*this, p1, q1, true);
                if (lUtility1 == 1) {
                    //findComparatorTriplet(p1, q1, lCurrentLine, lDualCurrentLine, lUtilityIndex, 1, 2, 1);
                }
                else if (lUtility1 == 0) {
                    if (findComparatorTriplet(p1, q1, lCurrentLine, lDualCurrentLine, lUtilityIndex, 0, 1, 1)) {
                        // Since the utility value was changed from 0 to non-zero, update
                        // all heuristic above it if they need to be updated.
                        for (unsigned int h = lUtilityIndex+1; h < mHeuristicMethods.size(); ++h) {
                            if (mHeuristic[x1][y1][h] == -DBL_MAX) {
                                mHeuristic[x1][y1][h] = calculateHeuristic(p1, q1, mHeuristicMethods[h]);
                            }
                        }
                        // Invalidate joins and meets to free space.
                        mJoins[x1][y1].invalidate();
                        mMeets[x1][y1].invalidate();
                    }
                }
            }
        }
    }

	Beagle_StackTraceEndM("void SN::Network::adjustHeuristicForTriplets()");
}


/*!
 *  \brief Find a comparator triplet if possible and update the heuristic
 *      matrix if such a triplet is found.  (p1, q1) is the first comparator
 *      of the triplet.  Utility 1, 2, and 3 are the respective utilities
 *      of comparators 1, 2, and 3.
 *  \return True if a triplet is found, false otherwise.
 */
bool SN::Network::findComparatorTriplet(unsigned int p1, unsigned int q1, unsigned int inCurrentLine, unsigned int inDualCurrentLine,
                                        unsigned int inUtilityIndex, double inUtility1, double inUtility2, double inUtility3)
{
	Beagle_StackTraceBeginM();

    unsigned int x1 = p1;
    unsigned int y1 = q1-p1-1;

    // We are working with four distinct points, p1, q1, i, and j in the
    // following code.  The i and j loops produce every possible valid
    // combination of comparator couplets that share an end point with
    // comparator (p1, q1).
    for (unsigned int i = inCurrentLine; i <= inDualCurrentLine; ++i) {
        if (i == p1 || i == q1) continue;

        unsigned int p2 = minOf(i, p1);
        unsigned int q2 = maxOf(i, p1);
        unsigned int x2 = p2;
        unsigned int y2 = q2-p2-1;

        unsigned int lUtility2 = findCombinedLinesUtility(*this, p2, q2, true);

        for (unsigned int j = inCurrentLine; j <= inDualCurrentLine; ++j) {
            if (j == p1 || j == q1 || j == i) continue;

            unsigned int p3 = minOf(j, q1);
            unsigned int q3 = maxOf(j, q1);
            unsigned int x3 = p3;
            unsigned int y3 = q3-p3-1;

            unsigned int lUtility3 = findCombinedLinesUtility(*this, p3, q3, true);

            // Check for the condition in which a comparator triplet will be useful.
            if ((lUtility2 == inUtility2 && lUtility3 == inUtility3) || (lUtility2 == inUtility3 && lUtility3 == inUtility2)) {

                // The comparator triplet we are looking for includes comparator 1, and
                // two comparators, each sharing one end of comparator 1 and the ends of
                // comparators 2 and 3 that they don't share with comparator 1.
                unsigned int r = (p2 != p1 && p2 != q1) ? p2 : q2;
                unsigned int s = (p3 != p1 && p3 != q1) ? p3 : q3;
                unsigned int p4, q4, p5, q5;

                unsigned int lTripletUtility = checkComparatorTriplet(p1, q1, r, s, p4, q4, p5, q5);

                // Using the stricter condition below reduces the frequency of obtaining
                // 39-comparator networks for n=12, but improves for other values of n.
                unsigned int lUtility4 = findCombinedLinesUtility(*this, p4, q4, true);
                unsigned int lUtility5 = findCombinedLinesUtility(*this, p5, q5, true);
                if (lTripletUtility > 0 && lUtility4 == 0 && lUtility5 == 0) {
                // if (lTripletUtility > 0) {
                    // Adjust utilities of comparators by averaging them.
                    double lAverage1 = (mHeuristic[x1][y1][inUtilityIndex] + lTripletUtility)/3;
                    double lAverage2 = (mHeuristic[x2][y2][inUtilityIndex] + mHeuristic[x3][y3][inUtilityIndex])/2;
                    mHeuristic[x1][y1][inUtilityIndex] = lAverage1;
                    mHeuristic[x2][y2][inUtilityIndex] = lAverage2;
                    mHeuristic[x3][y3][inUtilityIndex] = lAverage2;
                    // std::cout << p1 << "," << q1 << "," << p2 << "," << q2 << "," << p3 << "," << q3 << "," << p4 << "," << q4 << "," << p5 << "," << q5 << std::endl;
                    // std::cout << inUtility1 << "," << inUtility2 << "," << inUtility3 << "," << lTripletUtility << std::endl;
                    // mHeuristic[x1][y1][inUtilityIndex] = lTripletUtility;

                    // No need to store the first comparator because its indexes are the
                    // same as the entry in the mTriplets array.
                    mTriplets[x1][y1].push_back(p4);
                    mTriplets[x1][y1].push_back(q4);
                    mTriplets[x1][y1].push_back(p5);
                    mTriplets[x1][y1].push_back(q5);

                    return true;
                }
            }
        }
    }

    return false;

	Beagle_StackTraceEndM("bool SN::Network::findComparatorTriplet((unsigned int p1, unsigned int q1, unsigned int inCurrentLine, unsigned int inDualCurrentLine, unsigned int inUtilityIndex, double inUtility1, double inUtility2, double inUtility3)");
}


/*!
 *  \brief Find two comparators each with utility 2 such that one of their
 *      end points comes from (a, b) and the other end point comes
 *      from (c, d) with no repetition of end points.  (a, b) is a
 *      comparator with a lower utility that we might want to add.
 *  \return True if such comparators are found, false otherwise. If true,
 *      the new comparators are returned in (p1, q1), (p2, q2).
 */
unsigned int SN::Network::checkComparatorTriplet(unsigned int a, unsigned int b, unsigned int c, unsigned int d,
                                                 unsigned int& p1, unsigned int& q1, unsigned int& p2, unsigned int& q2)
{
	Beagle_StackTraceBeginM();

    // Effect of adding comparator (a, b).
    const Function& lFunctiona = getJoin(a, b);
    const Function& lFunctionb = getMeet(a, b);

    // There are two possible combinations of comparators:
    // Combination 1: (a, c), (b, d)
    // Combination 2: (a, d), (b, c)

    // Combination 1, Comparator 1
    unsigned int lSimpleUtility1, lUtility1;
    if (a < c) {
        p1 = a;
        q1 = c;
        lSimpleUtility1 = findCombinedLinesUtility(lFunctiona, mFunctions[c], a, c, mNumInputs, true);
        lUtility1 = findCombinedLinesUtility(lFunctiona, mFunctions[c], a, c, mNumInputs, false);
    }
    else {
        p1 = c;
        q1 = a;
        lSimpleUtility1 = findCombinedLinesUtility(mFunctions[c], lFunctiona, c, a, mNumInputs, true);
        lUtility1 = findCombinedLinesUtility(mFunctions[c], lFunctiona, c, a, mNumInputs, false);
    }

    // Combination 1, Comparator 2
    unsigned int lSimpleUtility2, lUtility2;
    if (b < d) {
        p2 = b;
        q2 = d;
        lSimpleUtility2 = findCombinedLinesUtility(lFunctionb, mFunctions[d], b, d, mNumInputs, true);
        lUtility2 = findCombinedLinesUtility(lFunctionb, mFunctions[d], b, d, mNumInputs, false);
    }
    else {
        p2 = d;
        q2 = b;
        lSimpleUtility2 = findCombinedLinesUtility(mFunctions[d], lFunctionb, d, b, mNumInputs, true);
        lUtility2 = findCombinedLinesUtility(mFunctions[d], lFunctionb, d, b, mNumInputs, false);
    }

    // If Combination 1 produced utility 2 for both comparators, return true.
    if (lSimpleUtility1 == 2 and lSimpleUtility2 == 2) {
        return lUtility1 + lUtility2;
        //return maxOf(lUtility1, lUtility2);
    }
    else {
        // Combination 2, Comparator 2
        if (a < d) {
            p1 = a;
            q1 = d;
            lSimpleUtility1 = findCombinedLinesUtility(lFunctiona, mFunctions[d], a, d, mNumInputs, true);
            lUtility1 = findCombinedLinesUtility(lFunctiona, mFunctions[d], a, d, mNumInputs, false);
        }
        else {
            p1 = d;
            q1 = a;
            lSimpleUtility1 = findCombinedLinesUtility(mFunctions[d], lFunctiona, d, a, mNumInputs, true);
            lUtility1 = findCombinedLinesUtility(mFunctions[d], lFunctiona, d, a, mNumInputs, false);
        }

        // Combination 2, Comparator 2
        if (b < c) {
            p2 = b;
            q2 = c;
            lSimpleUtility2 = findCombinedLinesUtility(lFunctionb, mFunctions[c], b, c, mNumInputs, true);
            lUtility2 = findCombinedLinesUtility(lFunctionb, mFunctions[c], b, c, mNumInputs, false);
        }
        else {
            p2 = c;
            q2 = b;
            lSimpleUtility2 = findCombinedLinesUtility(mFunctions[c], lFunctionb, c, b, mNumInputs, true);
            lUtility2 = findCombinedLinesUtility(mFunctions[c], lFunctionb, c, b, mNumInputs, false);
        }

        if (lSimpleUtility1 == 2 and lSimpleUtility2 == 2) {
            return lUtility1 + lUtility2;
            //return maxOf(lUtility1, lUtility2);
        }
        else {
            return 0;
        }
    }    

	Beagle_StackTraceEndM("unsigned int SN::Network::checkComparatorTriplet(unsigned int a, unsigned int b, unsigned int c, unsigned int d, unsigned int& p1, unsigned int& q1, unsigned int& p2, unsigned int& q2)");
}


/*!
 *  \brief Update the heuristic values after adding a comparator between two lines.
 *  \param i Index of top line
 *  \param j Index of bottom line
 */
void SN::Network::updateHeuristic(unsigned int i, unsigned int j)
{
	Beagle_StackTraceBeginM();

    // We need to update all entries for some heuristic methods (e.g.
    // eBottomUpBinaryTree, eComparatorUtility, etc.) if the last
    // comparator added matches the only current non-zero entry.
    // The following array is used to store this condition.
    std::vector<bool> lUpdateAll(mHeuristicMethods.size(), false);
    for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
        if (mHeuristicMethods[h] == eComparatorUtility) {
            std::vector<unsigned int> lIndexPairs;
            for (unsigned int p = 0; lIndexPairs.size() < 3 && p < mHeuristic.size(); ++p) {
                for (unsigned int q = 0; lIndexPairs.size() < 3 && q < mHeuristic[p].size(); ++q) {
                    if (mHeuristic[p][q][h] > 0) {
                        lIndexPairs.push_back(p);
                        lIndexPairs.push_back(q+p+1);
                    }
                }
            }
            if (lIndexPairs.size() == 2 && lIndexPairs[0] == i && lIndexPairs[1] == j) {
                lUpdateAll[h] = true;
            }
        }
    }

    // Update-all of some heuristics (e.g. eNextLineUtility) depends
    // on whether a similar update is performed for another heuristic
    // (e.g. eComparatorUtility).
    for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
        if (mHeuristicMethods[h] == eNextLineUtility ||
            mHeuristicMethods[h] == eNextNextLineUtility ||
            mHeuristicMethods[h] == eEffectNextLineComparators) {
            for (unsigned int g = 0; g < mHeuristicMethods.size(); ++g) {
                if (mHeuristicMethods[g] == eComparatorUtility && lUpdateAll[g]) {
                    lUpdateAll[h] = true;
                    break;
                }
            }
        }
    }

    // Invalidate joins and meets in the affected row and column.
    for (unsigned int k = 0; k < mHeuristic.size(); ++k) {
        if (i < k) {
            unsigned int x = i;
            unsigned int y = k-i-1;
            mJoins[x][y].invalidate();
            mMeets[x][y].invalidate();
        }
        else if (i > k) {
            unsigned int x = k;
            unsigned int y = i-k-1;
            mJoins[x][y].invalidate();
            mMeets[x][y].invalidate();
        }
        // Since [i][j] entry is invalidated above, no need to do it again.
        if (i != k) { // ignore (i, j)
            if (j < k) {  
                unsigned int x = j;
                unsigned int y = k-j-1;
                mJoins[x][y].invalidate();
                mMeets[x][y].invalidate();
            }
            else if (j > k) {  
                unsigned int x = k;
                unsigned int y = j-k-1;
                mJoins[x][y].invalidate();
                mMeets[x][y].invalidate();
            }
        }
    }

    // Update heuristics.  Since the heuristic matrices are used as filters,
    // we can optimize by not computing the entries that are excluded by the
    // first matrix.  Use the value -DBL_MAX to indicate that these entries
    // are invalid.  Ideally, the heuristic for the first matrix does not
    // require computing the meet (e.g. eComparatorUtility) so that it can
    // be calculated efficiently.  We also assume that this heuristic has
    // values greater than 0 for selected comparators.  Note that this
    // optimization works because the first heuristic is updated before any
    // of the other heuristics.

    // Update the entire heuristic matrix.  We iterate over the heuristic
    // methods in the innermost loop to optimize memory required to store
    // meets and joins (we can free the memory to store meets and joins for
    // a particular p,q combination after computing all heuristics for that
    // combination).  As a result, the condition to determine if entire
    // matrix is updated is inside the innermost loop.
    for (unsigned int p = 0; p < mHeuristic.size(); ++p) {
        for (unsigned int q = 0; q < mHeuristic[p].size(); ++q) {
            for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
                bool lUpdateMatrix = lUpdateAll[h] || mHeuristicMethods[h] == eAllLinesLookAheadUtility || mHeuristicMethods[h] == eNonZeroMeetLine;
                if (lUpdateMatrix == true) {
                    mHeuristic[p][q][h] = (h == 0 || mHeuristic[p][q][0] > 0) ? calculateHeuristic(p, q+p+1, mHeuristicMethods[h]) : -DBL_MAX;
                }
            }
            // Invalidate joins and meets to free space.
            mJoins[p][q].invalidate();
            mMeets[p][q].invalidate();
        }
    }

    // Otherwise, update only the affected row and column.  We need to
    // check again whether the entire matrix was updated above.
    for (unsigned int k = 0; k < mHeuristic.size(); ++k) {
        if (i < k) {
            unsigned int x = i;
            unsigned int y = k-i-1;
            for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
                bool lUpdateMatrix = lUpdateAll[h] || mHeuristicMethods[h] == eAllLinesLookAheadUtility || mHeuristicMethods[h] == eNonZeroMeetLine;
                if (lUpdateMatrix == false) {
                    mHeuristic[x][y][h] = (h == 0 || mHeuristic[x][y][0] > 0) ? calculateHeuristic(i, k, mHeuristicMethods[h]) : -DBL_MAX;
                }
            }
            // Invalidate joins and meets to free space.
            mJoins[x][y].invalidate();
            mMeets[x][y].invalidate();
        }
        else if (i > k) {
            unsigned int x = k;
            unsigned int y = i-k-1;
            for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
                bool lUpdateMatrix = lUpdateAll[h] || mHeuristicMethods[h] == eAllLinesLookAheadUtility || mHeuristicMethods[h] == eNonZeroMeetLine;
                if (lUpdateMatrix == false) {
                    mHeuristic[x][y][h] = (h == 0 || mHeuristic[x][y][0] > 0) ? calculateHeuristic(k, i, mHeuristicMethods[h]) : -DBL_MAX;
                }
            }
            // Invalidate joins and meets to free space.
            mJoins[x][y].invalidate();
            mMeets[x][y].invalidate();
        }
        // Since mHeuristic[i][j] is calculated above, don't calculate it again.
        if (i != k) { // ignore (i, j)
            if (j < k) {  
                unsigned int x = j;
                unsigned int y = k-j-1;
                for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
                    bool lUpdateMatrix = lUpdateAll[h] || mHeuristicMethods[h] == eAllLinesLookAheadUtility || mHeuristicMethods[h] == eNonZeroMeetLine;
                    if (lUpdateMatrix == false) {
                        mHeuristic[x][y][h] = (h == 0 || mHeuristic[x][y][0] > 0) ? calculateHeuristic(j, k, mHeuristicMethods[h]) : -DBL_MAX;
                    }
                }
                // Invalidate joins and meets to free space.
                mJoins[x][y].invalidate();
                mMeets[x][y].invalidate();
            }
            else if (j > k) {  
                unsigned int x = k;
                unsigned int y = j-k-1;
                for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
                    bool lUpdateMatrix = lUpdateAll[h] || mHeuristicMethods[h] == eAllLinesLookAheadUtility || mHeuristicMethods[h] == eNonZeroMeetLine;
                    if (lUpdateMatrix == false) {
                        mHeuristic[x][y][h] = (h == 0 || mHeuristic[x][y][0] > 0) ? calculateHeuristic(k, j, mHeuristicMethods[h]) : -DBL_MAX;
                    }
                }
                // Invalidate joins and meets to free space.
                mJoins[x][y].invalidate();
                mMeets[x][y].invalidate();
            }
        }
    }

    // Also update any invalid entries that are selected due to the entire
    // first matrix being updated.
    if (lUpdateAll[0]) { 
        for (unsigned int p = 0; p < mHeuristic.size(); ++p) {
            for (unsigned int q = 0; q < mHeuristic[p].size(); ++q) {
                for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
                    bool lUpdateMatrix = lUpdateAll[h] || mHeuristicMethods[h] == eAllLinesLookAheadUtility || mHeuristicMethods[h] == eNonZeroMeetLine;
                    if (lUpdateMatrix == false && mHeuristic[p][q][0] > 0 && mHeuristic[p][q][h] == -DBL_MAX) {
                        mHeuristic[p][q][h] = calculateHeuristic(p, q+p+1, mHeuristicMethods[h]);
                    }
                }
                // Invalidate joins and meets to free space.
                mJoins[p][q].invalidate();
                mMeets[p][q].invalidate();
            }
        }
    }

	Beagle_StackTraceEndM("void SN::Network::updateHeuristic(unsigned int i, unsigned int j)");
}


/*!
 *  \brief Calculate all the heuristic values.
 */
void SN::Network::calculateHeuristic()
{
	Beagle_StackTraceBeginM();

    // The heuristic array is symmetric with 0 values in the diagonal;
    // therefore, only the upper-triangular part is stored.
    for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
        for (unsigned int j = 0; j < mHeuristic[i].size(); ++j) {
            // Invalidate the join and meet arrays to erase any previous results.
            mJoins[i][j].invalidate();
            mMeets[i][j].invalidate();
            for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
                mHeuristic[i][j][h] = calculateHeuristic(i, j+i+1, mHeuristicMethods[h]);
            }
            // Invalidate joins and meets to free space.
            mJoins[i][j].invalidate();
            mMeets[i][j].invalidate();
        }
    }

	Beagle_StackTraceEndM("void SN::Network::calculateHeuristic()");
}


/*!
 *  \brief Calculate the heuristic value for adding a comparator between two lines.
 *  \param i Index of top line
 *  \param j Index of bottom line
 */
double SN::Network::calculateHeuristic(unsigned int i, unsigned int j, HeuristicMethod inHeuristicMethod)
{
	Beagle_StackTraceBeginM();

    if (isLineSorted(i) || isLineSorted(j)) {
        return 0;
    }

    switch (inHeuristicMethod) {
    case eUniformHeuristic:
        return findUniformUtility(*this, i, j);
    case eNewSorted:
        return countNewSorted(*this, i, j);
    case eComparatorUtility:
        return findComparatorUtility(*this, i, j);
    case eNextLineUtility:
        return findNextLineUtility(*this, i, j);
    case eNextNextLineUtility:
        return findNextLineUtility(*this, i, j, 2);
    case eAllLinesUtility:
        return findAllLinesUtility(*this, i, j);
    case eAllLinesLookAheadUtility:
        return findAllLinesLookAheadUtility(*this, i, j);
    case eNonZeroMeetLine:
        return findNonZeroMeetLineUtility(*this, i, j);
    case eCombinedLinesUtility:
        return findCombinedLinesUtility(*this, i, j, false);
    case eMeetProximity:
        return findMeetProximity(*this, i, j);
    case eAverageLookAheadUtility:
        return findAverageLookAheadUtility(*this, i, j);
    case eFirstNonEmptyMeetIndex:
        return findFirstNonEmptyMeetIndex(*this, i, j);
    case eFirstNonEmptyIndexVariance:
        return findFirstNonEmptyIndexVariance(*this, i, j);
    case eMatchMeetWithLines:
        return matchMeetWithLines(*this, i, j);
    case eSymmetricFunctionsBefore:
        return findSymmetricFunctionsUtilityBefore(*this, i, j);
    case eSymmetricFunctionsAfter:
        return findSymmetricFunctionsUtilityAfter(*this, i, j);
    case eFilledLevels:
        return findFilledLevelsUtility(*this, i, j);
    case eDestinationLine:
        return findDestinationLineUtility(*this, i, j);
    case eDisagreement:
        return findDisagreementUtility(*this, i, j);
    default:
        if (inHeuristicMethod >= eAutoAllLinesUtility) {
            unsigned int lLine = inHeuristicMethod - eAutoAllLinesUtility;
            if (lLine < getCurrentLine() + mNumAllLines) {
                return findComparatorUtility(*this, i, j, lLine);
            }
            else {
                return 0;
            }
        }
        else {
            return 0;
        }
    }

	Beagle_StackTraceEndM("double SN::Network::calculateHeuristic(unsigned int i, unsigned int j, HeuristicMethod inHeuristicMethod)");
}


/*!
 *  \brief Find the non-dominated heuristics and the corresponding index pairs
 *  \param outIndexPairs The index pairs for which heuristic is non-dominated
 */
void SN::Network::nondomHeuristic(std::vector<unsigned int>& outIndexPairs) const
{
	Beagle_StackTraceBeginM();

    // Find the pairs (i, j) of indices for which mHeuristic[i][j] is not dominated.
    std::vector<unsigned int> lNondomPairs;
    for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
        for (unsigned int j = 0; j < mHeuristic[i].size(); ++j) {
            // Determine whether the current value is comparable to another value we have
            // already seen.  If not comparable, we add (i, j) to the index pairs.
            bool lComparable = false;
            for (unsigned int k = 0; lComparable == false && k < lNondomPairs.size()/2; ++k) {
                unsigned int di = lNondomPairs[2*k];
                unsigned int dj = lNondomPairs[2*k+1]-di-1;
                bool lCurrentDominates = true;
                for (unsigned int h = 0; lCurrentDominates && h < mFilterStartIndex; ++h) {
                    lCurrentDominates = mHeuristic[i][j][h] >= mHeuristic[di][dj][h];
                }
                if (lCurrentDominates) {
                    lNondomPairs[2*k] = i;
                    lNondomPairs[2*k+1] = j+i+1;
                    lComparable = true;
                }
                else {
                    bool lPreviousDominates = true;
                    for (unsigned int h = 0; lPreviousDominates && h < mFilterStartIndex; ++h) {
                        lPreviousDominates = mHeuristic[i][j][h] <= mHeuristic[di][dj][h];
                    }
                    if (lPreviousDominates) {
                        lComparable = true;
                    }
                }
            }
            if (lComparable == false) {
                lNondomPairs.push_back(i);
                lNondomPairs.push_back(j+i+1);
            }
        }
    }

    // Find all pairs (i, j) of indices for which mHeuristic[i][j] is equal to entries
    // in lNondomPairs.
    outIndexPairs.clear();
    for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
        for (unsigned int j = 0; j < mHeuristic[i].size(); ++j) {
            // Determine whether the current value is equal to a value in lNondomPairs.
            bool lFound = false;
            for (unsigned int k = 0; lFound == false && k < lNondomPairs.size()/2; ++k) {
                unsigned int di = lNondomPairs[2*k];
                unsigned int dj = lNondomPairs[2*k+1]-di-1;
                bool lEqual = true;
                for (unsigned int h = 0; lEqual && h < mFilterStartIndex; ++h) {
                    lEqual = mHeuristic[i][j][h] == mHeuristic[di][dj][h];
                }
                if (lEqual) {
                    outIndexPairs.push_back(i);
                    outIndexPairs.push_back(j+i+1);
                    lFound = true;
                }
            }
        }
    }

	Beagle_StackTraceEndM("void SN::Network::nondomHeuristic(std::vector<unsigned int>& outIndexPairs) const");
}


/*!
 *  \brief Find the maximum value of the heuristic
 */
double SN::Network::maxHeuristic(unsigned int h) const
{
	Beagle_StackTraceBeginM();

    double lMaxValue = -DBL_MAX;
    for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
        for (unsigned int j = 0; j < mHeuristic[i].size(); ++j) {
            lMaxValue = maxOf(lMaxValue, mHeuristic[i][j][h]);
        }
    }

    return lMaxValue;

	Beagle_StackTraceEndM("double SN::Network::maxHeuristic(unsigned int h) const");
}


/*!
 *  \brief Find the maximum value of the heuristic and the corresponding index pairs
 *  \param outIndexPairs The index pairs for which heuristic attains max value
 */
void SN::Network::maxHeuristic(std::vector<unsigned int>& outIndexPairs, unsigned int h) const
{
	Beagle_StackTraceBeginM();

    double lMaxValue = -DBL_MAX;
    for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
        for (unsigned int j = 0; j < mHeuristic[i].size(); ++j) {
            lMaxValue = maxOf(lMaxValue, mHeuristic[i][j][h]);
        }
    }

    outIndexPairs.clear();
    if (lMaxValue > 0) {
        for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
            for (unsigned int j = 0; j < mHeuristic[i].size(); ++j) {
                if (mHeuristic[i][j][h] == lMaxValue) {
                    outIndexPairs.push_back(i);
                    outIndexPairs.push_back(j+i+1);
                }
            }
        }
    }

	Beagle_StackTraceEndM("void SN::Network::maxHeuristic(std::vector<unsigned int>& outIndexPairs, unsigned int h) const");
}


/*!
 *  \brief Filter the given index pairs based on the maximum value of the heuristic
 *  \param ioIndexPairs The index pairs for which heuristic attains max value
 *  \return the max value
 */
double SN::Network::maxHeuristicFilter(std::vector<unsigned int>& ioIndexPairs, unsigned int h) const
{
	Beagle_StackTraceBeginM();

    std::vector<unsigned int> lMaxIndexPairs;
    double lMaxValue = -DBL_MAX;
    for (unsigned i = 0; i < ioIndexPairs.size()/2; ++i) {
        unsigned int x = ioIndexPairs[2*i];
        unsigned int y = ioIndexPairs[2*i+1];
        double lValue = mHeuristic[x][y-x-1][h];
        if (lValue > lMaxValue) {
            lMaxValue = lValue;
            lMaxIndexPairs.clear();
            lMaxIndexPairs.push_back(ioIndexPairs[2*i]);
            lMaxIndexPairs.push_back(ioIndexPairs[2*i+1]);
        }
        else if (lValue == lMaxValue) {
            lMaxIndexPairs.push_back(ioIndexPairs[2*i]);
            lMaxIndexPairs.push_back(ioIndexPairs[2*i+1]);
        }
    }
    ioIndexPairs = lMaxIndexPairs;

    return lMaxValue;

	Beagle_StackTraceEndM("double SN::Network::maxHeuristicFilter(std::vector<unsigned int>& ioIndexPairs, unsigned int h) const");
}


/*!
 *  \brief Find all index pairs having non-zero value of the heuristic
 *  \param outIndexPairs The index pairs for which heuristic attains max value
 */
void SN::Network::nonzeroHeuristic(std::vector<unsigned int>& outIndexPairs, unsigned int h) const
{
	Beagle_StackTraceBeginM();

    outIndexPairs.clear();
    for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
        for (unsigned int j = 0; j < mHeuristic[i].size(); ++j) {
            if (mHeuristic[i][j][h] > 0) {
                outIndexPairs.push_back(i);
                outIndexPairs.push_back(j+i+1);
            }
        }
    }

	Beagle_StackTraceEndM("void SN::Network::nonzeroHeuristic(std::vector<unsigned int>& outIndexPairs, unsigned int h) const");
}


/*!
 *  \brief Find all index pairs having non-zero value of the heuristic
 *  \param outIndexPairs The index pairs for which heuristic attains max value
 */
void SN::Network::nonzeroHeuristicFilter(std::vector<unsigned int>& ioIndexPairs, unsigned int h) const
{
	Beagle_StackTraceBeginM();

    std::vector<unsigned int> lIndexPairs;
    for (unsigned i = 0; i < ioIndexPairs.size()/2; ++i) {
        unsigned int x = ioIndexPairs[2*i];
        unsigned int y = ioIndexPairs[2*i+1];
        if (mHeuristic[x][y-x-1][h] > 0) {
            lIndexPairs.push_back(ioIndexPairs[2*i]);
			lIndexPairs.push_back(ioIndexPairs[2*i+1]);
        }
    }
    ioIndexPairs = lIndexPairs;

	Beagle_StackTraceEndM("void SN::Network::nonzeroHeuristicFilter(std::vector<unsigned int>& ioIndexPairs, unsigned int h) const");
}


/*!
 *  \brief Find all index pairs of comparators satisfying the current heuristic specification
 *  \param outIndexPairs The index pairs of candidate comparators
 */
void SN::Network::findCandidateComparators(std::vector<unsigned int>& outIndexPairs, const std::vector<unsigned int>& inFilterGroup) const
{
	Beagle_StackTraceBeginM();

    outIndexPairs.clear();
    if (mHeuristicMethods.size() > 0 && mHeuristicMethods.back() == eAllLinesLookAheadUtility) {
        maxHeuristic(outIndexPairs, mHeuristicMethods.size()-1);
    }
    else {
        if (mFilterStartIndex > 0) {
            nondomHeuristic(outIndexPairs);
            unsigned int g = 0;
            for (unsigned int h = mFilterStartIndex; h < mHeuristicMethods.size(); ++h) {
                if (inFilterGroup.size() == 0 || (g < inFilterGroup.size() && inFilterGroup[g]+mFilterStartIndex == h)) {
                    maxHeuristicFilter(outIndexPairs, h);
                    g++;
                }
            }
        }
        else if (mHeuristicMethods.size() > 0) {
            maxHeuristic(outIndexPairs, 0);
            unsigned int g = 1;
            for (unsigned int h = 1; h < mHeuristicMethods.size(); ++h) {
                if (inFilterGroup.size() == 0 || (g < inFilterGroup.size() && inFilterGroup[g] == h)) {
                    maxHeuristicFilter(outIndexPairs, h);
                    g++;
                }
            }
        }
    }

	Beagle_StackTraceEndM("void SN::Network::findCandidateComparators(std::vector<unsigned int>& outIndexPair, const std::vector<unsigned int>& inFilterGroup) const");
}


/*!
 *  \brief Find all index pairs of comparators satisfying the current heuristic information.
 *  \param outIndexPairs The index pairs of candidate comparators
 *  \param outRoulette Roulette wheel for probabilistic selection of the index pairs.
 */
void SN::Network::findCandidateComparators(std::vector<unsigned int>& outIndexPairs, RouletteT<unsigned int>& outRoulette, const std::vector<unsigned int>& inFilterGroup) const
{
	Beagle_StackTraceBeginM();

    outIndexPairs.clear();
    outRoulette.clear();
    nonzeroHeuristic(outIndexPairs, 0);
    std::vector<double> lWeights(outIndexPairs.size()/2);

    std::vector< std::pair<double, unsigned int> > lRankVector(outIndexPairs.size()/2);
    for (unsigned i = 0; i < lRankVector.size(); ++i) {
        unsigned int x = outIndexPairs[2*i];
        unsigned int y = outIndexPairs[2*i+1];
        lRankVector[i].first = mHeuristic[x][y-x-1][0];
        lRankVector[i].second = i;
    }

    stable_sort(lRankVector.begin(), lRankVector.end(), gtrank);
    unsigned int lRank = 0;
    for (unsigned i = 0; i < lRankVector.size(); ++i) {
        if (i > 0 && lRankVector[i-1] > lRankVector[i]) lRank++;
        lWeights[lRankVector[i].second] = 1.0/(double)(1 << lRank);
    }
    
    unsigned int g = 1;
    std::vector<unsigned int> lFilterPairs = outIndexPairs;
    for (unsigned int h = 1; h < mHeuristicMethods.size(); ++h) {
        if (inFilterGroup.size() == 0 || (g < inFilterGroup.size() && inFilterGroup[g] == h)) {
            nonzeroHeuristicFilter(lFilterPairs, h);
            unsigned int j = 0;
            lRankVector.resize(lFilterPairs.size()/2);
            for (unsigned i = 0; i < lRankVector.size(); ++i) {
                unsigned int x = lFilterPairs[2*i];
                unsigned int y = lFilterPairs[2*i+1];
                lRankVector[i].first = mHeuristic[x][y-x-1][h];
                while (x != outIndexPairs[2*j] || y != outIndexPairs[2*j+1]) j++;
                lRankVector[i].second = j;
                j++;
            }

            stable_sort(lRankVector.begin(), lRankVector.end(), gtrank);
            lRank = 0;
            for (unsigned i = 0; i < lRankVector.size(); ++i) {
                if (i > 0 && lRankVector[i-1] > lRankVector[i]) lRank++;
                lWeights[lRankVector[i].second] += 1.0/(double)(1 << (lRank+h));
            }

            g++;
        }
    }

    for (unsigned i = 0; i < lWeights.size(); ++i) {
        outRoulette.insert(i, lWeights[i]);
    }

	Beagle_StackTraceEndM("void SN::Network::findCandidateComparators(std::vector<unsigned int>& outIndexPair, RouletteT<unsigned int>& outRoulette, const std::vector<unsigned int>& inFilterGroup) const");
}


/*!
 *  \brief Filter the given index pairs of comparators using look-ahead of the first filter heuristics
 *  \param ioIndexPairs The index pairs of candidate comparators to filter
 */
void SN::Network::lookAheadHeuristicFilter(std::vector<unsigned int>& ioIndexPairs) const
{
	Beagle_StackTraceBeginM();

    // If we have more than one candidate, narrow the choices further by looking
    // ahead one comparator addition.
    // This heuristic favors those comparators that produce non-zero entries
    // in the same locations of the eComparatorUtility and the eNextLineUtility
    // heuristics.
    if (ioIndexPairs.size() > 2 && mHeuristicMethods.size() == 2) {
        std::vector<unsigned int> lMaxIndexPairs;
        double lMaxValue = 0;
        for (unsigned i = 0; i < ioIndexPairs.size()/2; ++i) {
            // Compute the heuristic if this comparator is added to the network.
            unsigned int x = ioIndexPairs[2*i];
            unsigned int y = ioIndexPairs[2*i+1];
            Network lLookAheadNetwork = *this;
            lLookAheadNetwork.addComparator(x, y);

            // Find the candidate comparators in this new network.
            std::vector<unsigned int> lLookAheadPairs;
            lLookAheadNetwork.findCandidateComparators(lLookAheadPairs);

            // If there is overlap of non-zero eComparatorUtility and eNextLineUtility
            // entries, then comparator i gets the value 1 for the look ahead heuristic;
            // otherwise it gets the value 0.
            double lValue = 0;
            if (lLookAheadPairs.size() > 0) {
                unsigned int p = lLookAheadPairs[0];
                unsigned int q = lLookAheadPairs[1];
                std::vector<std::vector<std::vector<double> > >& lLookAheadHeuristic = lLookAheadNetwork.getHeuristic();
                lValue = (lLookAheadHeuristic[p][q-p-1][0] > 0 && lLookAheadHeuristic[p][q-p-1][1] > 0) ? 1 : 0;
            }

            if (lValue > lMaxValue) {
                lMaxValue = lValue;
                lMaxIndexPairs.clear();
                lMaxIndexPairs.push_back(ioIndexPairs[2*i]);
                lMaxIndexPairs.push_back(ioIndexPairs[2*i+1]);
            }
            else if (lValue == lMaxValue) {
                lMaxIndexPairs.push_back(ioIndexPairs[2*i]);
                lMaxIndexPairs.push_back(ioIndexPairs[2*i+1]);
            }
        }
        ioIndexPairs = lMaxIndexPairs;
    }

	Beagle_StackTraceEndM("void SN::Network::lookAheadHeuristicFilter(std::vector<unsigned int>& ioIndexPairs) const");
}


/*!
 *  \brief Filter the given index pairs of comparators using maximum cardinality matching.
 *  \param ioIndexPairs The index pairs of candidate comparators to filter
 */
void SN::Network::maxMatchingFilter(std::vector<unsigned int>& ioIndexPairs) const
{
	Beagle_StackTraceBeginM();

    // If we have more than one candidate, narrow the choices further by applying
    // maximum cardinality matching, i.e. a set of independent edges that have no
    // common vertices when taken pairwise.

    // unsigned int lHeuristicIndexStart = 0;
    // for (; lHeuristicIndexStart < mHeuristicMethods.size(); ++lHeuristicIndexStart) {
    //     if (mHeuristicMethods[lHeuristicIndexStart] == eComparatorUtility ||
	// 		mHeuristicMethods[lHeuristicIndexStart] == eAutoAllLinesUtility) {
    //         break;
    //     }
    // }
    // if (lHeuristicIndexStart == mHeuristicMethods.size()) {
    //     throw Beagle_RunTimeExceptionM("The maximum matching filter can be specified only with eComparatorUtility or eAutoAllLinesUtility.");
    // }

    // double lMaxValue = 0;
    // for (unsigned int i = lHeuristicIndexStart; i < mHeuristicMethods.size(); ++i) {
    //     lMaxValue = maxHeuristic(i);
    //     if (lMaxValue > 0) {
    //         break;
    //     }
    // }

    // if (ioIndexPairs.size() > 2 && lMaxValue == 2) {
    if (ioIndexPairs.size() > 2) {
        using namespace boost;
        typedef adjacency_list<vecS, vecS, undirectedS> GraphType;

        GraphType lGraph(mNumInputs);
        for (unsigned i = 0; i < ioIndexPairs.size()/2; ++i) {
            add_edge(ioIndexPairs[2*i], ioIndexPairs[2*i+1], lGraph);
        }

        std::vector<graph_traits<GraphType>::vertex_descriptor> lMate(mNumInputs);
        if (checked_edmonds_maximum_cardinality_matching(lGraph, &lMate[0]) == false) {
            throw Beagle_RunTimeExceptionM("Failed to find maximum cardinality matching.");
        }

        std::vector<unsigned int> lMaxIndexPairs;
        graph_traits<GraphType>::vertex_iterator vi, vi_end;
        for (tie(vi, vi_end) = vertices(lGraph); vi != vi_end; ++vi) {
            if (lMate[*vi] != graph_traits<GraphType>::null_vertex() && *vi < lMate[*vi]) {
                lMaxIndexPairs.push_back(*vi);
                lMaxIndexPairs.push_back(lMate[*vi]);
            }
        }

        ioIndexPairs = lMaxIndexPairs;
    }

	Beagle_StackTraceEndM("void SN::Network::maxMatchingFilter(std::vector<unsigned int>& ioIndexPairs) const");
}


/*!
 *  \brief Find all comparators that have a symmetric comparator in the given list
 *  \param inIndexPairs The index pairs of candidate comparators
 *  \param outIndexPairs The index pairs of symmetric comparators
 */
void SN::Network::findSymmetricComparators(const std::vector<unsigned int>& inIndexPairs, std::vector<unsigned int>& outIndexPairs) const
{
	Beagle_StackTraceBeginM();

    // Check all pairs of comparators to see if they are symmetric.
    // This check can be made O(nlogn) by finding the symmetric comparators
    // first and then sorting and comparing them.
    outIndexPairs.clear();
    unsigned int lNumPairs = inIndexPairs.size()/2;
    for (unsigned int i = 0; i < lNumPairs; ++i) {
        unsigned int lJoin = inIndexPairs[2*i];
        unsigned int lMeet = inIndexPairs[2*i+1];
        unsigned int lSymJoin = mNumInputs-1-lMeet;
        unsigned int lSymMeet = mNumInputs-1-lJoin;
        for (unsigned int j = i; j < lNumPairs; ++j) {
            if (lSymJoin == inIndexPairs[2*j] && lSymMeet == inIndexPairs[2*j+1]) {
                outIndexPairs.push_back(lJoin);
                outIndexPairs.push_back(lMeet);
            }
        }
    }

	Beagle_StackTraceEndM("void SN::Network::findSymmetricComparators(const std::vector<unsigned int>& inIndexPairs, std::vector<unsigned int>& outIndexPairs) const");
}


/*!
 *  \brief Find all comparators that have a symmetric comparator in the given list
 *  \param inIndexPairs The index pairs of candidate comparators
 *  \param outComparatorIndices The indices of symmetric comparators in the input
 */
void SN::Network::findSymmetricComparatorIndices(const std::vector<unsigned int>& inIndexPairs, std::vector<unsigned int>& outIndices) const
{
	Beagle_StackTraceBeginM();

    // Check all pairs of comparators to see if they are symmetric.
    // This check can be made O(nlogn) by finding the symmetric comparators
    // first and then sorting and comparing them.
    outIndices.clear();
    unsigned int lNumPairs = inIndexPairs.size()/2;
    for (unsigned int i = 0; i < lNumPairs; ++i) {
        unsigned int lJoin = inIndexPairs[2*i];
        unsigned int lMeet = inIndexPairs[2*i+1];
        unsigned int lSymJoin = mNumInputs-1-lMeet;
        unsigned int lSymMeet = mNumInputs-1-lJoin;
        for (unsigned int j = i; j < lNumPairs; ++j) {
            if (lSymJoin == inIndexPairs[2*j] && lSymMeet == inIndexPairs[2*j+1]) {
                outIndices.push_back(i);
                outIndices.push_back(j);
            }
        }
    }

	Beagle_StackTraceEndM("void SN::Network::findSymmetricComparators(const std::vector<unsigned int>& inIndexPairs, std::vector<unsigned int>& outIndices) const");
}


/*!
 *  \brief Find the number of binary inputs sorted by the network.
 */
unsigned int SN::Network::findNumSorted() const
{
	Beagle_StackTraceBeginM();

    // To prevent unsigned int overflow, the maximum number of input lines allowed is 31.
    if (mNumInputs > 31) {
        std::ostringstream lOSS;
        lOSS << "Number of inputs (" << mNumInputs << ") cannot exceed 31.";
        throw Beagle_RunTimeExceptionM(lOSS.str());
    }
    
    // Check if the network can sort all 2^mNumInputs binary combinations.
    unsigned int lNumSorted = 0;
    for (unsigned int i = 0; i < ((unsigned int)1 << mNumInputs); i++) {
        unsigned int lBits = i;
        for (unsigned int j = 0; j < getNumComparators(); j++) {
            unsigned int lInd1;
            unsigned int lInd2;
            // The j-th comparator is between input lines with indices (*this)[2*j]
            // and (*this)[2*j+1].
            lInd1 = (*this)[2*j];
            lInd2 = (*this)[2*j+1];

            // Extract the two bits from lBits.
            unsigned int lBit1 = lBits & (1 << lInd1);
            unsigned int lBit2 = lBits & (1 << lInd2);

            // If the two bits are different (i.e. one of them is zero and the other is
            // non-zero), we may have to swap them.
            if (lBit1 > 0 && lBit2 == 0 && lInd1 < lInd2) {
                // Set bit1 of lBits to 0 and bit2 of lBits to 1.
                lBits = ~(1 << lInd1) & lBits;
                lBits = (1 << lInd2) | lBits;
            }
            else if (lBit1 == 0 && lBit2 > 0 && lInd1 > lInd2) {
                // Set bit1 of lBits to 1 and bit2 of lBits to 0.
                lBits = (1 << lInd1) | lBits;
                lBits = ~(1 << lInd2) & lBits;
            }
        }

        // If the bits of lBits are sorted, then all the ones must be left-aligned,
        // starting at bit-position mNumInputs-1.
        unsigned int lNumOnes = 0;
        unsigned int lSortedBits = 0;
        for (unsigned int j = 0; j < mNumInputs; j++) {
            if (((i >> j) & 1) == 1) {
                lNumOnes++;
                lSortedBits |= 1 << (mNumInputs - lNumOnes);
            }
        }
        if (lSortedBits == lBits) lNumSorted++;
    }

    return lNumSorted;

	Beagle_StackTraceEndM("unsigned int SN::Network::findNumSorted() const");
}


/*!
 *  \brief Find the number of sets in the functions of the network.
 */
unsigned int SN::Network::findNumSets() const
{
	Beagle_StackTraceBeginM();

    // Since the elements of the functions are guaranteed to be disjoint,
    // we can just add up their sizes.
    unsigned int lNumSets = 0;
    for (unsigned int i = 0; i < mFunctions.size(); ++i) {
        lNumSets += mFunctions[i].getNumSetsInLevel(i+1);
    }

    return lNumSets;

	Beagle_StackTraceEndM("unsigned int SN::Network::findNumSets() const");
}


/*!
 *  \brief Find the number of levels that are full in the functions of the network.
 */
unsigned int SN::Network::findNumFullLevels() const
{
	Beagle_StackTraceBeginM();

    unsigned int lNumFull = 0;
    for (unsigned int i = 0; i < mFunctions.size(); ++i) {
        lNumFull += mFunctions[i].getNumFullLevelsBoth();
    }

    return lNumFull;

	Beagle_StackTraceEndM("unsigned int SN::Network::findNumFullLevels() const");
}


/*!
 *  \brief Find the number of lines of the network that are already sorted.
 */
unsigned int SN::Network::findNumLinesCompleted() const
{
	Beagle_StackTraceBeginM();

    unsigned int lCount = 0;
    for (unsigned int i = 0; i < mNumInputs; i++) {
        if (isLineSorted(i)) lCount++;
    }

    return lCount;

	Beagle_StackTraceEndM("unsigned int SN::Network::findNumLinesCompleted() const");
}


/*!
 *  \brief Trace the operation of the network, i.e. print the monotone Boolean
 *      functions produced as a result of each comparator addition.
 */
void SN::Network::trace()
{
	Beagle_StackTraceBeginM();

    allocateDataStructures();

    std::cout << std::endl;

    // Initialize functions.
    for (unsigned int i = 0; i < mNumInputs; i++) {
        mFunctions[i].setInput(i);
        std::cout << i << ": ";
        mFunctions[i].printNumSets();
        std::cout << std::endl;
    }
    std::cout << std::endl;

    calculateHeuristic();
    for (unsigned int j = 0; j < getNumComparators(); j++) {
        printHeuristic();

        std::vector<unsigned int> lIndexPairs;
        RouletteT<unsigned int> lRoulette;
        findCandidateComparators(lIndexPairs, lRoulette);

        unsigned int lPrintWidth = 5;
        std::cout << "roulette:" << std::endl;
        std::cout.width(lPrintWidth);
        std::cout << " ";
        for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
            std::cout.width(lPrintWidth);
            std::cout << i;
        }
        std::cout << std::endl;
        unsigned int r = 0;
        for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
            std::cout.width(lPrintWidth);
            std::cout << i;
            for (unsigned int k = 0; k < i+1; ++k) {
                std::cout.width(lPrintWidth);
                std::cout << " ";
            }
            for (unsigned int k = 0; k < mHeuristic[i].size(); ++k) {
                double lWeight = 0.0;
                if (r < lIndexPairs.size()/2 && i == lIndexPairs[2*r] && k+i+1 == lIndexPairs[2*r+1]) {
                    lWeight = (r > 0) ? (lRoulette[r].first - lRoulette[r-1].first) : lRoulette[r].first;
                    r++;
                }
                std::cout.width(lPrintWidth);
                std::cout << dbl2str(lWeight);
            }
            std::cout << std::endl;
        }

        // The j-th comparator is between input lines with indices (*this)[2*j]
        // and (*this)[2*j+1].
        unsigned int lJoin = (*this)[2*j];
        unsigned int lMeet = (*this)[2*j+1];

        Function lTempFunc = mFunctions[lJoin];
        join(mFunctions[lJoin], mFunctions[lMeet]);
        meet(mFunctions[lMeet], lTempFunc);

        std::cout << lJoin << ": ";
        mFunctions[lJoin].printNumSets();
        std::cout << std::endl;
        std::cout << lMeet << ": ";
        mFunctions[lMeet].printNumSets();
        std::cout << std::endl;
        std::cout << std::endl;

        updateHeuristic(lJoin, lMeet);
        adjustHeuristicForTriplets();
    }

	Beagle_StackTraceEndM("void SN::Network::trace()");
}


/*!
 *  \brief Print the values in the heuristic matrix.
 */
void SN::Network::printHeuristic() const
{
	Beagle_StackTraceBeginM();

    unsigned int lPrintWidth = 5;

    // The heuristic array is symmetric with 0 values in the diagonal;
    // therefore, only the upper-triangular part is stored.
    std::vector<unsigned int> lIndexPairs;
    for (unsigned int h = 0; h < mHeuristicMethods.size(); ++h) {
        std::cout << "heuristic " << mHeuristicMethods[h] << ":" << std::endl;
        std::cout.width(lPrintWidth);
        std::cout << " ";
        for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
            std::cout.width(lPrintWidth);
            std::cout << i;
        }
        std::cout << std::endl;
        for (unsigned int i = 0; i < mHeuristic.size(); ++i) {
            std::cout.width(lPrintWidth);
            std::cout << i;
            for (unsigned int j = 0; j < i+1; ++j) {
                std::cout.width(lPrintWidth);
                std::cout << " ";
            }
            for (unsigned int j = 0; j < mHeuristic[i].size(); ++j) {
                std::cout.width(lPrintWidth);
                std::cout << ((mHeuristic[i][j][h] > -DBL_MAX) ? dbl2str(mHeuristic[i][j][h]) : "x");
            }
            std::cout << std::endl;
        }

        if (mFilterStartIndex == 0 || h >= mFilterStartIndex-1) {
            if (mFilterStartIndex == 0 && h == 0) {
                maxHeuristic(lIndexPairs, 0);
            }
            else if (h == mFilterStartIndex-1) {
                nondomHeuristic(lIndexPairs);
            }
            else {
                maxHeuristicFilter(lIndexPairs, h);
            }
            std::cout << "candidates: ";
            unsigned int lNumPairs = lIndexPairs.size()/2;
            for (unsigned i = 0; i < lNumPairs; ++i) {
                unsigned int lInd1 = lIndexPairs[2*i];
                unsigned int lInd2 = lIndexPairs[2*i+1];
                std::cout << "(" << lInd1 << ", " << lInd2 << ")";
                if (i < lNumPairs-1) std::cout << ", ";
            }
            std::cout << std::endl;
        }
    }

    // std::cout << "Applying maximum cardinality matching filter..." << std::endl;
    // maxMatchingFilter(lIndexPairs);
    // std::cout << "candidates: ";
    // unsigned int lNumPairs = lIndexPairs.size()/2;
    // for (unsigned i = 0; i < lNumPairs; ++i) {
    //     unsigned int lInd1 = lIndexPairs[2*i];
    //     unsigned int lInd2 = lIndexPairs[2*i+1];
    //     std::cout << "(" << lInd1 << ", " << lInd2 << ")";
    //     if (i < lNumPairs-1) std::cout << ", ";
    // }
    // std::cout << std::endl;

	Beagle_StackTraceEndM("void SN::Network::printHeuristic() const");
}
