
/*!
 *  \file   beagle/SN/src/TermSortedOp.cpp
 *  \brief  Source code of class TermSortedOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/SN.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a termination check operator based on sorting correctness.
 *  \param inName Name of the operator.
 */
SN::TermSortedOp::TermSortedOp(std::string inName) :
		TerminationOp(inName)
{ }


/*!
 *  \brief Register the parameters of this operator.
 *  \param ioSystem System to use to initialize the operator.
 */
void SN::TermSortedOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	TerminationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Number of inputs of the sorting network",
		    "UInt",
		    "0",
		    "Number of inputs of the sorting network."
		);
		mNumInputs = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("sn.init.numinputs", new UInt(0), lDescription));
	}
	Beagle_StackTraceEndM("void SN::TermSortedOp::registerParams(Beagle::System&)");
}


/*!
 *  \brief Check if the network correctly sorts all values.
 *  \param inDeme Actual deme of the evolution.
 *  \param ioContext Actual evolution context.
 *  \return True if the ending criterion is reached, false if not.
 */
bool SN::TermSortedOp::terminate(const Beagle::Deme& inDeme, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

	for(unsigned int i=0; i<inDeme.size(); ++i) {
        const SN::Individual& lIndividual = castObjectT<const SN::Individual&>(*inDeme[i]);
        const SN::Network& lNetwork = *lIndividual[0];

        // In addition to checking that the network is sorted, make sure
        // that we have run evolution for number of generations at least
        // the number of comparators in the network, which ensures that we
        // find networks that are constructed slower by adding one comparator
        // at time and has smaller number of comparators.

        // unsigned int lNumInputs = lNetwork.getNumInputs();
        // unsigned int lNumSets = lNetwork.findNumSets();
		// if (lNumSets+1 == (unsigned int)(1<<lNumInputs)) {
        if (lNetwork.isSorted() && ioContext.getGeneration() >= lNetwork.getNumComparators()) {
			Beagle_LogInfoM(
			    ioContext.getSystem().getLogger(),
			    "termination", "Beagle::SN::TermSortedOp",
			    std::string("Termination criterion reached by the ") +
			    uint2ordinal(i+1) + std::string(" individual (") +
			    uint2str(lNetwork.getNumComparators()) + std::string(" comparators)")
			);
			ioContext.setTerminationSuccessful();
			return true;
		}
	}

	Beagle_LogTraceM(
	    ioContext.getSystem().getLogger(),
	    "termination", "Beagle::SN::TermSortedOp",
	    std::string("Termination criterion not reached ")
	);
	return false;

	Beagle_StackTraceEndM("bool SN::TermSortedOp::terminate(const Beagle::Deme& inDeme, Beagle::Context& ioContext)");
}
