
/*!
 *  \file   beagle/SN/src/Utilities.cpp
 *  \brief  Utilities for evolving sorting networks.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/SN.hpp"

using namespace Beagle;


/*!
 *  \brief Find the binomial coefficient n choose k using the recurrence relation
 *      binomial(n, k) = (n/k)*binomial(n-1, k-1).
 *      Taken from http://blog.plover.com/math/choose.html
 */
unsigned int SN::binomial(unsigned int n, unsigned int k)
{
	Beagle_StackTraceBeginM();

    if (k > n) return 0;

    unsigned int r = 1;
    for (unsigned int d = 1; d <= k; d++) {
        r *= n--;
        r /= d;
    }
    return r;

	Beagle_StackTraceEndM("unsigned int SN::binomial(unsigned int n, unsigned int k)");
}


/*!
 *  \brief Compute the join of the given functions
 */
void SN::join(const SN::Function& inLeft, const SN::Function& inRight, SN::Function& outJoin)
{
	Beagle_StackTraceBeginM();

    unsigned int lNumLevels = Function::getNumLevels();

    outJoin.mNumSets.clear();
	outJoin.mNumSets.resize(lNumLevels, 0);
	outJoin.mTruthTable.clear();
	outJoin.mTruthTable.resize(lNumLevels);
    
    for (unsigned int i = 0; i < lNumLevels; ++i) {
        if (inLeft.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i) || inRight.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i)) {
            outJoin.mNumSets[i] = Function::getMaxSetsInLevel(i);
        }
        else if (inLeft.getNumSetsInLevel(i) == 0 && inRight.getNumSetsInLevel(i) == 0) {
            outJoin.mNumSets[i] = 0;
        }
        else if (inLeft.getNumSetsInLevel(i) == 0) {
            outJoin.mTruthTable[i] = inRight.mTruthTable[i];
            outJoin.mNumSets[i] = inRight.getNumSetsInLevel(i);
        }
        else if (inRight.getNumSetsInLevel(i) == 0) {
            outJoin.mTruthTable[i] = inLeft.mTruthTable[i];
            outJoin.mNumSets[i] = inLeft.getNumSetsInLevel(i);
        }
        else {
            outJoin.mTruthTable[i] = inLeft.mTruthTable[i];
            outJoin.mTruthTable[i] |= inRight.mTruthTable[i];
            outJoin.mNumSets[i] = outJoin.mTruthTable[i].count();
            if (outJoin.mNumSets[i] == Function::getMaxSetsInLevel(i)) {  // cannot be 0
                outJoin.mTruthTable[i].clear();
            }
        }
    }

	Beagle_StackTraceEndM("void SN::join(const SN::Function& inLeft, const SN::Function& inRight, SN::Function& outJoin)");
}


/*!
 *  \brief Compute the meet of the given functions
 */
void SN::meet(const SN::Function& inLeft, const SN::Function& inRight, SN::Function& outMeet)
{
	Beagle_StackTraceBeginM();

    unsigned int lNumLevels = Function::getNumLevels();

    outMeet.mNumSets.clear();
	outMeet.mNumSets.resize(lNumLevels, 0);
	outMeet.mTruthTable.clear();
	outMeet.mTruthTable.resize(lNumLevels);
    
    for (unsigned int i = 0; i < lNumLevels; ++i) {
        if (inLeft.getNumSetsInLevel(i) == 0 || inRight.getNumSetsInLevel(i) == 0) {
            outMeet.mNumSets[i] = 0;
        }
        else if (inLeft.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i) && inRight.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i)) {
            outMeet.mNumSets[i] = Function::getMaxSetsInLevel(i);
        }
        else if (inLeft.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i)) {
            outMeet.mTruthTable[i] = inRight.mTruthTable[i];
            outMeet.mNumSets[i] = inRight.getNumSetsInLevel(i);
        }
        else if (inRight.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i)) {
            outMeet.mTruthTable[i] = inLeft.mTruthTable[i];
            outMeet.mNumSets[i] = inLeft.getNumSetsInLevel(i);
        }
        else {
            outMeet.mTruthTable[i] = inLeft.mTruthTable[i];
            outMeet.mTruthTable[i] &= inRight.mTruthTable[i];
            outMeet.mNumSets[i] = outMeet.mTruthTable[i].count();
            if (outMeet.mNumSets[i] == 0) {  // cannot be Function::getMaxSetsInLevel(i)
                outMeet.mTruthTable[i].clear();
            }
        }
    }

	Beagle_StackTraceEndM("void SN::meet(const SN::Function& inLeft, const SN::Function& inRight, SN::Function& outMeet)");
}


/*!
 *  \brief Compute the join of the given functions
 */
void SN::join(SN::Function& ioLeft, const SN::Function& inRight)
{
	Beagle_StackTraceBeginM();

    unsigned int lNumLevels = Function::getNumLevels();
    for (unsigned int i = 0; i < lNumLevels; ++i) {
        if (ioLeft.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i) || inRight.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i)) {
            ioLeft.mTruthTable[i].clear();
            ioLeft.mNumSets[i] = Function::getMaxSetsInLevel(i);
        }
        else if (ioLeft.getNumSetsInLevel(i) == 0 && inRight.getNumSetsInLevel(i) == 0) {
        }
        else if (ioLeft.getNumSetsInLevel(i) == 0) {
            ioLeft.mTruthTable[i] = inRight.mTruthTable[i];
            ioLeft.mNumSets[i] = inRight.getNumSetsInLevel(i);
        }
        else if (inRight.getNumSetsInLevel(i) == 0) {
        }
        else {
            ioLeft.mTruthTable[i] |= inRight.mTruthTable[i];
            ioLeft.mNumSets[i] = ioLeft.mTruthTable[i].count();
            if (ioLeft.mNumSets[i] == Function::getMaxSetsInLevel(i)) {  // cannot be 0
                ioLeft.mTruthTable[i].clear();
            }
        }
    }

	Beagle_StackTraceEndM("void SN::join(SN::Function& ioLeft, const SN::Function& inRight)");
}


/*!
 *  \brief Compute the meet of the given functions
 */
void SN::meet(SN::Function& ioLeft, const SN::Function& inRight)
{
	Beagle_StackTraceBeginM();

    unsigned int lNumLevels = Function::getNumLevels();
    for (unsigned int i = 0; i < lNumLevels; ++i) {
        if (ioLeft.getNumSetsInLevel(i) == 0 || inRight.getNumSetsInLevel(i) == 0) {
            ioLeft.mTruthTable[i].clear();
            ioLeft.mNumSets[i] = 0;
        }
        else if (ioLeft.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i) && inRight.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i)) {
        }
        else if (ioLeft.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i)) {
            ioLeft.mTruthTable[i] = inRight.mTruthTable[i];
            ioLeft.mNumSets[i] = inRight.getNumSetsInLevel(i);
        }
        else if (inRight.getNumSetsInLevel(i) == Function::getMaxSetsInLevel(i)) {
        }
        else {
            ioLeft.mTruthTable[i] &= inRight.mTruthTable[i];
            ioLeft.mNumSets[i] = ioLeft.mTruthTable[i].count();
            if (ioLeft.mNumSets[i] == 0) {  // cannot be Function::getMaxSetsInLevel(i)
                ioLeft.mTruthTable[i].clear();
            }
        }
    }

	Beagle_StackTraceEndM("void SN::meet(SN::Function& ioLeft, const SN::Function& inRight)");
}


/*!
 *  \brief Find if there is any utility in adding a comparator, i.e. if adding the
 *      comparator will sort part of any previously unsorted input.
 */
unsigned int SN::findUniformUtility(const SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];

    // If the lines are already sorted, there no point in adding a comparator.
    if (inNetwork.isLineSorted(inLeftIndex) || inNetwork.isLineSortedDual(inRightIndex)) {
        return 0;
    }
    else {
        return (lRight.isLessEqual(lLeft) == false);
    }

	Beagle_StackTraceEndM("unsigned int SN::findUniformHeuristic(const SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Count the number of binary inputs newly sorted as a result of
 *      adding a comparator between the left and right functions.
 */
unsigned int SN::countNewSorted(const SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    unsigned int lNumSortedOrig = inNetwork.findNumSorted();
    Network lTempNetwork = inNetwork;
    lTempNetwork.push_back(inLeftIndex);
    lTempNetwork.push_back(inRightIndex);
    unsigned int lNumSortedTemp = lTempNetwork.findNumSorted();
    
    return lNumSortedTemp - lNumSortedOrig;

	Beagle_StackTraceEndM("unsigned int SN::countNewSorted(const SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the utility of adding the specified comparator in terms of its
 *      ability to collect sets for both primal and dual functions.
 */
int SN::findComparatorUtility(const SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];

    // First find out which line we are currently working on.
    unsigned int lLine = inNetwork.getCurrentLine();
    unsigned int lDualLine = lNumInputs-1-lLine;
    if (lLine > lDualLine) {  // Sorting network construction complete
        return 0;
    }

    // If we have already finished with the left index or the right index,
    // then there is no point in adding a comparator.
    if (lLine > inLeftIndex || lLine > inRightIndex || lDualLine < inLeftIndex || lDualLine < inRightIndex) {
        return 0;
    }

    // We accumulate sets only if both left and right counts are non-zero or if
    // the comparator moves some of these sets to the current line.
    unsigned int lLeftCount = lLeft.getNumSetsInLevel(lLine+1);
    unsigned int lRightCount = lRight.getNumSetsInLevel(lLine+1);
    unsigned int lUtility = ((lLeftCount > 0 || inLeftIndex == lLine) && lRightCount > 0) ? 1 : 0;

    unsigned int lDualLeftCount = lLeft.getNumSetsInLevelDual(lDualLine);
    unsigned int lDualRightCount = lRight.getNumSetsInLevelDual(lDualLine);
    lUtility += (lDualLeftCount > 0 && (lDualRightCount > 0 || inRightIndex == lDualLine)) ? 1 : 0;

    return lUtility;

	Beagle_StackTraceEndM("int SN::findComparatorUtility(const SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the utility of adding the specified comparator in terms of its
 *      ability to collect sets for both primal and dual functions.
 */
unsigned int SN::findComparatorUtility(SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, unsigned int inLine)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    unsigned int lLine = inLine;
    unsigned int lDualLine = lNumInputs-1-lLine;

    // Since the heuristics are used as filters, a particular entry in the
    // heuristic matrix can have an effect in the output of the filter stack
    // only if all corresponding entries in previous matrices in the stack
    // are non-zero, i.e. the meet count or dual meet count should be 0 for
    // all previous lines.

    unsigned int lLeftCount = lLeft.getNumSetsInLevel(lLine+1);
	unsigned int lRightCount = lRight.getNumSetsInLevel(lLine+1);
    unsigned int lMeetCount = lMeet.getNumSetsInLevel(lLine+1);
	unsigned int lUtility = ((lMeetCount == 0 && (lLeftCount > 0 || inLeftIndex == lLine) && lRightCount > 0)) ? 1 : 0;
                             //(lJoinCount == 0 && lLeftCount > 0 && (lRightCount > 0 || inRightIndex == lLine))) ? 1 : 0;
                             //((lMeetClear && (lLeftCount > 0 || inLeftIndex <= lLine) && lRightCount > 0) ||
                             //(lMeetClear && lLeftMaxLine < UINT_MAX && lRightMaxLine < UINT_MAX && lLine <= lLeftMaxLine+lRightMaxLine+1 && lLeftCount == 0 && lRightCount == 0))) ||
                             //(lJoinCount == 0 && inLeftIndex < lLine && inRightIndex >= lLine && lLeftCount > 0 && (lRightCount > 0 || inRightIndex == lLine))) ? 1 : 0;

    unsigned int lDualLeftCount = lLeft.getNumSetsInLevelDual(lDualLine);
	unsigned int lDualRightCount = lRight.getNumSetsInLevelDual(lDualLine);
	unsigned int lDualMeetCount = lJoin.getNumSetsInLevelDual(lDualLine);
    lUtility += ((lDualMeetCount == 0 && lDualLeftCount > 0 && (lDualRightCount > 0 || inRightIndex == lDualLine))) ? 1 : 0;
                 //(lDualJoinCount == 0 && (lDualLeftCount > 0 || inLeftIndex == lDualLine) && lDualRightCount > 0)) ? 1 : 0;
                 //((lDualMeetClear && lDualLeftCount > 0 && (lDualRightCount > 0 || inRightIndex >= lDualLine)) ||
                 //(lDualMeetClear && lDualLeftMaxLine < UINT_MAX && lDualRightMaxLine < UINT_MAX && lLine <= lDualLeftMaxLine+lDualRightMaxLine+1 && lDualLeftCount == 0 && lDualRightCount == 0))) ||
                 //(lDualJoinCount == 0 && inRightIndex > lDualLine && inLeftIndex <= lDualLine && (lDualLeftCount > 0 || inLeftIndex == lDualLine) && lDualRightCount > 0)) ? 1 : 0;

    return lUtility;

	Beagle_StackTraceEndM("unsigned int SN::findComparatorUtility(SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, unsigned int inLine)");
}


/*!
 *  \brief Find the value of adding a comparator based on the first line containing
 *      non-zero sets in the meet.
 */
unsigned int SN::findNonZeroMeetLineUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    // First find out which line we are currently working on.
    unsigned int lLine = inNetwork.getCurrentLine();
    unsigned int lDualLine = lNumInputs-1-lLine;
    if (lLine > lDualLine) {  // Sorting network construction complete
        return 0;
    }

    // If we have already finished with the left index or the right index,
    // then there is no point in adding a comparator.
    if (lLine > inLeftIndex || lLine > inRightIndex || lDualLine < inLeftIndex || lDualLine < inRightIndex) {
        return 0;
    }

    // We accumulate sets only if both left and right counts are non-zero or if
    // the comparator moves some of these sets to the current line.
    unsigned int lLeftCount = lLeft.getNumSetsInLevel(lLine+1);
    unsigned int lRightCount = lRight.getNumSetsInLevel(lLine+1);
    bool lPrimal = (lLeftCount > 0 || inLeftIndex == lLine) && lRightCount > 0;

    unsigned int lDualLeftCount = lLeft.getNumSetsInLevelDual(lDualLine);
    unsigned int lDualRightCount = lRight.getNumSetsInLevelDual(lDualLine);
    bool lDual = lDualLeftCount > 0 && (lDualRightCount > 0 || inRightIndex == lDualLine);

    // Find the first non-empty line in both meets.
    unsigned int lMeetLine = lMeet.getFirstNonEmptyLevel()-1;
    unsigned int lDualMeetLine = lNumInputs-1-lJoin.getFirstNonEmptyLevelDual();

    if (lPrimal && lDual) {
        //return maxOf(lMeetLine, lDualMeetLine);
        //return minOf(lMeetLine, lDualMeetLine);
        //return lMeetLine+lDualMeetLine;
        return 0;
    }
    else if (lPrimal) {
        return lNumInputs-lDualMeetLine;
    }
    else if (lDual) {
        return lNumInputs-lMeetLine;
    }
    else {
         return 0;
    }

	Beagle_StackTraceEndM("unsigned int SN::findNonZeroMeetLineUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the utility of adding the specified comparator in terms of its
 *      ability to collect sets for both primal and dual functions with
 *      respect to the next line.
 */
unsigned int SN::findNextLineUtility(SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, unsigned int inNext)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    // First find out which line we are currently working on.
    unsigned int lLine = inNetwork.getCurrentLine();
    unsigned int lDualLine = lNumInputs-1-lLine;
    if (lLine > lDualLine) {  // Sorting network construction complete
        return 0;
    }

    // If we have already finished with the left index or the right index,
    // then there is no point in adding a comparator.
    if (lLine > inLeftIndex || lLine > inRightIndex || lDualLine < inLeftIndex || lDualLine < inRightIndex) {
        return 0;
    }

    unsigned int lNextLine = lLine+inNext;
    unsigned int lDualNextLine = lDualLine-inNext;

    // We accumulate sets only if both left and right counts are non-zero to begin
    // with and meet count is zero or if the comparator moves some of these sets
    // to the appropriate line or if no both left and right counts are zero and no
    // new sets of the same cardinality are created (this last condition signifies
    // keeping both the join and the meet lines free of the sets corresponding to
    // lNextLine even when the inputs to the comparator contain sets corresponding
    // to previous lines, i.e. even when the potential to create such sets exists).
    unsigned int lLeftCount = lLeft.getNumSetsInLevel(lNextLine+1);
    unsigned int lRightCount = lRight.getNumSetsInLevel(lNextLine+1);
    unsigned int lMeetCount = lMeet.getNumSetsInLevel(lNextLine+1);
    unsigned int lUtility = (lMeetCount == 0 && (((lLeftCount > 0 || inLeftIndex == lNextLine) && lRightCount > 0) ||
                                                 (lLeftCount == 0 && lRightCount == 0))) ? 1 : 0;

    unsigned int lDualLeftCount = lLeft.getNumSetsInLevelDual(lDualNextLine);
	unsigned int lDualRightCount = lRight.getNumSetsInLevelDual(lDualNextLine);
	unsigned int lDualMeetCount = lJoin.getNumSetsInLevelDual(lDualNextLine);
    lUtility += (lDualMeetCount == 0 && ((lDualLeftCount > 0 && (lDualRightCount > 0 || inRightIndex == lDualNextLine)) ||
                                         (lDualLeftCount == 0 && lDualRightCount == 0))) ? 1 : 0;

    return lUtility;

	Beagle_StackTraceEndM("unsigned int SN::findNextLineUtility(SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, unsigned int inNext)");
}


/*!
 *  \brief Find the utility of adding the specified comparator by considering
 *      all lines together.
 */
unsigned int SN::findCombinedLinesUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, bool inSimple)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];

    unsigned int lUtility = findCombinedLinesUtility(lLeft, lRight, inLeftIndex, inRightIndex, lNumInputs, inSimple);

    return lUtility;

	Beagle_StackTraceEndM("unsigned int SN::findCombinedLinesUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex, bool inSimple)");
}


/*!
 *  \brief Find the utility of adding the specified comparator by considering
 *      all lines together.
 */
unsigned int SN::findCombinedLinesUtility(const Function& inLeft, const Function& inRight, unsigned int inLeftIndex,
                                          unsigned int inRightIndex, unsigned int inNumInputs, bool inSimple)
{
	Beagle_StackTraceBeginM();

    // Find the indexes of the first non-zero entries in the functions.
    unsigned int lLeftLine = inLeft.getFirstNonEmptyLevel()-1;
    unsigned int lRightLine = inRight.getFirstNonEmptyLevel()-1;
    unsigned int lDualLeftLine = inNumInputs-1-inLeft.getFirstNonEmptyLevelDual();
    unsigned int lDualRightLine = inNumInputs-1-inRight.getFirstNonEmptyLevelDual();

    // We accumulate sets only if both left and right counts are non-zero or if
    // the comparator moves some of these sets to the current line.
    // unsigned int lUtility = (lLeftLine == lRightLine || (lLeftLine > lRightLine && inLeftIndex == lRightLine)) ? 1 : 0;
    // lUtility += (lDualLeftLine == lDualRightLine || (lDualRightLine > lDualLeftLine && inRightIndex == inNumInputs-1-lDualLeftLine)) ? 1 : 0;

    // unsigned int lUtility = (lLeftLine == lRightLine || (lLeftLine > lRightLine && inLeftIndex == lRightLine)) ?
    //     (1<<(inNumInputs-1-lRightLine)) : (1<<(inNumInputs-1-maxOf(lLeftLine, lRightLine)));
    // lUtility += (lDualLeftLine == lDualRightLine || (lDualRightLine > lDualLeftLine && inRightIndex == inNumInputs-1-lDualLeftLine)) ?
    //     (1<<(inNumInputs-1-lDualLeftLine)) : (1<<(inNumInputs-1-maxOf(lDualLeftLine, lDualRightLine)));

    // unsigned int lUtility = (lLeftLine == lRightLine || (lLeftLine > lRightLine && inLeftIndex == lRightLine)) ?
    //     (1<<(inNumInputs-1-lRightLine)) : 0;
    // lUtility += (lDualLeftLine == lDualRightLine || (lDualRightLine > lDualLeftLine && inRightIndex == inNumInputs-1-lDualLeftLine)) ?
    //     (1<<(inNumInputs-1-lDualLeftLine)) : 0;

    bool lAccumulates = lLeftLine == lRightLine || (lLeftLine > lRightLine && inLeftIndex == lRightLine);
    bool lDualAccumulates = lDualLeftLine == lDualRightLine || (lDualRightLine > lDualLeftLine && inRightIndex == inNumInputs-1-lDualLeftLine);

    unsigned int lUtility;
    if (inSimple) {
        lUtility = (unsigned int)lAccumulates + (unsigned int)lDualAccumulates;
    }
    else {
        if (lAccumulates && lDualAccumulates) {
            lUtility = (1<<(inNumInputs-1-lRightLine)) + (1<<(inNumInputs-1-lDualLeftLine));
        }
        else if (lAccumulates) {
            lUtility = (1<<(inNumInputs-1-lRightLine)) + (1<<(inNumInputs-1-maxOf(lDualLeftLine, lDualRightLine)));                            
        }
        else if (lDualAccumulates) {
            lUtility = (1<<(inNumInputs-1-lDualLeftLine)) + (1<<(inNumInputs-1-maxOf(lLeftLine, lRightLine)));                                 
        }
        else {
            lUtility = 0;
            //lUtility = (1<<(inNumInputs-1-maxOf(lLeftLine, lRightLine))) + (1<<(inNumInputs-1-maxOf(lDualLeftLine, lDualRightLine)));
        }
    }

    return lUtility;

	Beagle_StackTraceEndM("unsigned int SN::findCombinedLinesUtility(const Function& inLeft, const Function& inRight, unsigned int inLeftIndex, unsigned int inRightIndex, unsigned int inNumInputs, bool inSimple))");
}


/*!
 *  \brief Find the utility of adding the specified comparator by considering
 *      all lines together.
 */
unsigned int SN::findMeetProximity(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    unsigned int lProximity = lMeet.getFirstNonEmptyLevel() - maxOf(lLeft.getFirstNonEmptyLevel(), lRight.getFirstNonEmptyLevel());
	unsigned int lDualProximity = minOf(lLeft.getFirstNonEmptyLevelDual(), lRight.getFirstNonEmptyLevelDual()) - lMeet.getFirstNonEmptyLevelDual();

	return maxOf(lProximity, lDualProximity);
	//return lProximity + lDualProximity;

    // unsigned int lProximity = (inRightIndex < lNumInputs/2) ? (lMeet.getFirstNonEmptyIndex() - lRight.getFirstNonEmptyIndex()) :
    //     ((inLeftIndex >= (lNumInputs/2 + lNumInputs%2)) ? (lDualMeet.getFirstNonEmptyIndex() - lDualLeft.getFirstNonEmptyIndex()) : 0);

    // return lProximity;

	Beagle_StackTraceEndM("unsigned int SN::findMeetProximity(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the average utility of adding the specified comparator by considering
 *      the next two comparators that touch it.
 */
double SN::findAverageLookAheadUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    unsigned int lUtility = findCombinedLinesUtility(lLeft, lRight, inLeftIndex, inRightIndex, lNumInputs);

    // Find the maximum utility possible for comparators touching the end lines
    // of this comparator if this comparator has not already sorted those lines.
    unsigned int lLeftLevel = lJoin.getFirstNonEmptyLevel();
    bool lLeftLineSorted = (lLeftLevel == inLeftIndex+1) && (lJoin.getNumSetsInLevel(lLeftLevel) == Function::getMaxSetsInLevel(lLeftLevel));
    unsigned int lRightLevel = lMeet.getFirstNonEmptyLevel();
    bool lRightLineSorted = (lRightLevel == inRightIndex+1) && (lMeet.getNumSetsInLevel(lRightLevel) == Function::getMaxSetsInLevel(lRightLevel));

    std::vector<unsigned int> lLeftMaxIndexes;
    unsigned int lLeftMaxUtility = 0;
    if (!lLeftLineSorted) {
        unsigned int lOldMaxValue = 0;
        unsigned int lNewMaxValue = 0;
        for (unsigned int i = 0; i < lNumInputs; ++i) {
            if (i < inLeftIndex) {
                lOldMaxValue = maxOf(lOldMaxValue, findCombinedLinesUtility(lFunctions[i], lLeft, i, inLeftIndex, lNumInputs));
                unsigned int lNewValue = findCombinedLinesUtility(lFunctions[i], lJoin, i, inLeftIndex, lNumInputs);
                if (lNewValue > lNewMaxValue) {
                    lNewMaxValue = lNewValue;
                    lLeftMaxIndexes.clear();
                    lLeftMaxIndexes.push_back(i);
                }
                else if (lNewValue == lNewMaxValue) {
                    lLeftMaxIndexes.push_back(i);
                }
            }
            else if (i > inLeftIndex) {
                lOldMaxValue = maxOf(lOldMaxValue, findCombinedLinesUtility(lLeft, lFunctions[i], inLeftIndex, i, lNumInputs));
                unsigned int lNewValue = findCombinedLinesUtility(lJoin, lFunctions[i], inLeftIndex, i, lNumInputs);
                if (lNewValue > lNewMaxValue) {
                    lNewMaxValue = lNewValue;
                    lLeftMaxIndexes.clear();
                    lLeftMaxIndexes.push_back(i);
                }
                else if (lNewValue == lNewMaxValue) {
                    lLeftMaxIndexes.push_back(i);
                }
            }
        }

        lLeftMaxUtility = (lNewMaxValue > lOldMaxValue) ? (lNewMaxValue - lOldMaxValue) : 0;
    }

    std::vector<unsigned int> lRightMaxIndexes;
    unsigned int lRightMaxUtility = 0;
    if (!lRightLineSorted) {
        unsigned int lOldMaxValue = 0;
        unsigned int lNewMaxValue = 0;
        for (unsigned int i = 0; i < lNumInputs; ++i) {
            if (i < inRightIndex) {
                lOldMaxValue = maxOf(lOldMaxValue, findCombinedLinesUtility(lFunctions[i], lRight, i, inRightIndex, lNumInputs));
                unsigned int lNewValue = findCombinedLinesUtility(lFunctions[i], lMeet, i, inRightIndex, lNumInputs);
                if (lNewValue > lNewMaxValue) {
                    lNewMaxValue = lNewValue;
                    lRightMaxIndexes.clear();
                    lRightMaxIndexes.push_back(i);
                }
                else if (lNewValue == lNewMaxValue) {
                    lRightMaxIndexes.push_back(i);
                }
            }
            else if (i > inRightIndex) {
                lOldMaxValue = maxOf(lOldMaxValue, findCombinedLinesUtility(lRight, lFunctions[i], inRightIndex, i, lNumInputs));
                unsigned int lNewValue = findCombinedLinesUtility(lMeet, lFunctions[i], inRightIndex, i, lNumInputs);
                if (lNewValue > lNewMaxValue) {
                    lNewMaxValue = lNewValue;
                    lRightMaxIndexes.clear();
                    lRightMaxIndexes.push_back(i);
                }
                else if (lNewValue == lNewMaxValue) {
                    lRightMaxIndexes.push_back(i);
                }
            }
        }

        lRightMaxUtility = (lNewMaxValue > lOldMaxValue) ? (lNewMaxValue - lOldMaxValue) : 0;
    }

    unsigned int lNumComparators = 1;
    if (lLeftMaxIndexes.size() == 1 && lRightMaxIndexes.size() == 1 && lLeftMaxIndexes[0] == lRightMaxIndexes[0]) {
        unsigned int lMaxUtility = maxOf(lLeftMaxUtility, lRightMaxUtility);
        if (lMaxUtility > 0) {
            lNumComparators += 1;
            lUtility += lMaxUtility;
        }
    }
    else {
        if (lLeftMaxUtility > 0) {
            lNumComparators += 1;
            lUtility += lLeftMaxUtility;
        }
        if (lRightMaxUtility > 0) {
            lNumComparators += 1;
            lUtility += lRightMaxUtility;
        }
    }

    double lAverageUtility = ((double)lUtility)/((double)lNumComparators);

    return lUtility;

	Beagle_StackTraceEndM("double SN::findAverageLookAheadUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the first index of the meet that is not empty.
 */
unsigned int SN::findFirstNonEmptyMeetIndex(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    return minOf(lMeet.getFirstNonEmptyLevel()-1, lNumInputs-1-lMeet.getFirstNonEmptyLevelDual());

	Beagle_StackTraceEndM("unsigned int SN::findFirstNonEmptyMeetIndex(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Measure based on the variance of the first non-empty indexes
 *      before and after adding a comparator.
 */
unsigned int SN::findFirstNonEmptyIndexVariance(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    // A function is considered pure if it contains only one non-empty index.
    // The effect of the comparator is considered below only if the variables below are true.
    bool lPure = lLeft.getNumNonEmptyLevels() == 1 && lRight.getNumNonEmptyLevels() == 1 && lMeet.getNumNonEmptyLevels() == 1;
    bool lDualPure = lLeft.getNumNonEmptyLevelsDual() == 1 && lRight.getNumNonEmptyLevelsDual() == 1 && lMeet.getNumNonEmptyLevelsDual() == 1;

    std::vector<int> lCounts(lNumInputs, 0);
	std::vector<int> lDualCounts(lNumInputs, 0);
    for (unsigned int i = 0; i < lNumInputs; ++i) {
        unsigned int lIndex, lDualIndex;
        if (i == inLeftIndex) {
            unsigned int lJoinFirstNonEmptyIndex = minOf(lFunctions[inLeftIndex].getFirstNonEmptyLevel()-1, lFunctions[inRightIndex].getFirstNonEmptyLevel()-1);
            lIndex = lPure ? lJoinFirstNonEmptyIndex : (lFunctions[i].getFirstNonEmptyLevel()-1);
            lDualIndex = lDualPure ? (lNumInputs-1-lMeet.getFirstNonEmptyLevelDual()) : (lNumInputs-1-lFunctions[i].getFirstNonEmptyLevelDual());
        }
        else if (i == inRightIndex) {
            unsigned int lDualJoinFirstNonEmptyIndex = lNumInputs-1-maxOf(lFunctions[inLeftIndex].getFirstNonEmptyLevelDual(), lFunctions[inRightIndex].getFirstNonEmptyLevelDual());
            lIndex = lPure ? (lMeet.getFirstNonEmptyLevel()-1) : (lFunctions[i].getFirstNonEmptyLevel()-1);
            lDualIndex = lDualPure ? lDualJoinFirstNonEmptyIndex : (lNumInputs-1-lFunctions[i].getFirstNonEmptyLevelDual());
        }
        else {
            lIndex = lFunctions[i].getFirstNonEmptyLevel()-1;
            lDualIndex = lNumInputs-1-lFunctions[i].getFirstNonEmptyLevelDual();
        }
        lCounts[lIndex]++;
        lDualCounts[lDualIndex]++;
    }

    // The average of the counts is 1, since summing the elements of the
    // counts array gives the value lNumInputs.  We are interested in the
    // variance of the elements of the arrays, which is maximum at the input
    // of the sorting network (lCounts[0] = lNumInputs; all other elements
    // are 0) and minimum at the output (all elements are 1).
    unsigned int lSumOfSquares = 0;
    unsigned int lDualSumOfSquares = 0;
    for (unsigned int i = 0; i < lNumInputs; ++i) {
        int lDistFromMean = lCounts[i] - 1;
        int lDualDistFromMean = lDualCounts[i] - 1;
        lSumOfSquares += lDistFromMean*lDistFromMean;
        lDualSumOfSquares += lDualDistFromMean*lDualDistFromMean;
    }

    unsigned int lMaxValue = (lNumInputs-1)*(lNumInputs-1);

    return maxOf(lMaxValue - lSumOfSquares, lMaxValue - lDualSumOfSquares);
    //return lMaxValue - lSumOfSquares + lMaxValue - lDualSumOfSquares;

	Beagle_StackTraceEndM("unsigned int SN::findFirstNonEmptyIndexVariance(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Score the given comparator based on how it contributes to achieving
 *      the required non-empty first index at the outputs.
 */
int SN::matchMeetWithLines(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    int lDistance = (lMeet.getFirstNonEmptyLevel()-1) - inRightIndex;
	int lDualDistance = inLeftIndex - lJoin.getFirstNonEmptyLevelDual();

    //return maxOf(lDistance, lDualDistance);
    return minOf(lDistance, lDualDistance);

	Beagle_StackTraceEndM("int SN::matchMeetWithLines(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the utility of adding the specified comparator in terms of its
 *      ability to collect sets for both primal and dual functions.
 */
int SN::findAllLinesUtility(SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    int lUtility = 0;
    unsigned int lLine = 0;
    unsigned int lDualLine = lNumInputs-1;
    while (lLine < lDualLine) {
        unsigned int lLeftCount = lLeft.getNumSetsInLevel(lLine+1);
        unsigned int lRightCount = lRight.getNumSetsInLevel(lLine+1);
        unsigned int lMeetCount = lMeet.getNumSetsInLevel(lLine+1);
        lUtility += ((lMeetCount == 0 && (lLeftCount > 0 || inLeftIndex == lLine) && lRightCount > 0)) ? 1 : 0;

        unsigned int lDualLeftCount = lLeft.getNumSetsInLevelDual(lDualLine);
        unsigned int lDualRightCount = lRight.getNumSetsInLevelDual(lDualLine);
        unsigned int lDualMeetCount = lJoin.getNumSetsInLevelDual(lDualLine);
        lUtility += ((lDualMeetCount == 0 && lDualLeftCount > 0 && (lDualRightCount > 0 || inRightIndex == lDualLine))) ? 1 : 0;

        lLine++;
        lDualLine--;
    }

    return lUtility;

	Beagle_StackTraceEndM("int SN::findAllLinesUtility(SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the utility of adding the specified comparator in terms of its
 *      ability to collect sets for both primal and dual functions by looking
 *      ahead one comparator addition.
 */
int SN::findAllLinesLookAheadUtility(SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const std::vector<Network::HeuristicMethod>& lHeuristicMethods = inNetwork.getHeuristicMethods();
    if (lHeuristicMethods.size() == 0 || lHeuristicMethods[0] != Network::eAllLinesUtility) {
        throw Beagle_RunTimeExceptionM("The first heuristic method must be eAllLinesUtility.");
    }

    int lUtility = findAllLinesUtility(inNetwork, inLeftIndex, inRightIndex);

    if (lUtility > 0) {
        // Find the maximum utility possible if this comparator is added.
        // Remove the last heuristic (i.e. eAllLinesLookAheadUtility) before
        // adding the comparator to prevent recursion when Network::updateHeuristic()
        // is called by Network::addComparator().
        Network lLookAheadNetwork = inNetwork;
        lLookAheadNetwork.getHeuristicMethods().pop_back();
        lLookAheadNetwork.getHeuristic().pop_back();
        lLookAheadNetwork.addComparator(inLeftIndex, inRightIndex);
        std::vector<std::vector<std::vector<double> > >& lLookAheadHeuristic = lLookAheadNetwork.getHeuristic();
        double lMaxValue = -DBL_MAX;
        for (unsigned int i = 0; i < lLookAheadHeuristic.size(); ++i) {
            for (unsigned int j = 0; j < lLookAheadHeuristic[i].size(); ++j) {
                lMaxValue = maxOf(lMaxValue, lLookAheadHeuristic[i][j][0]);
            }
        }

        lUtility += (int)lMaxValue;
    }

    return lUtility;

	Beagle_StackTraceEndM("int SN::findAllLinesLookAheadUtility(SN::Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the utility given the bit counts before and after adding a comparator.
 */
int SN::utility(unsigned int inBeforeCount, unsigned int inAfterCount)
{
	Beagle_StackTraceBeginM();

    if (inBeforeCount == 0 and inAfterCount == 0)
        return 0;
    else if (inBeforeCount == 0 and inAfterCount > 0)
        return -1;
    else if (inBeforeCount > 0 and inAfterCount == 0)
        return 1;
    else // (inBeforeCount > 0 and inAfterCount > 0)
        return 0;

	Beagle_StackTraceEndM("int SN::utility(unsigned int inBeforeCount, unsigned int inAfterCount)");
}


/*!
 *  \brief Find the utility of adding the specified comparator based on whether the
 *      line functions are symmetric before the comparator is added.
 */
unsigned int SN::findSymmetricFunctionsUtilityBefore(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];

    bool lSymmetric = true;
    for (unsigned int i = 0; lSymmetric && (i < lNumInputs-1); ++i) {
        lSymmetric = (lLeft.getNumSetsInLevel(i+1) == lRight.getNumSetsInLevel(i+1));
    }

    unsigned int lUtility = 0;
    if (lSymmetric) {
        unsigned int lShift = lNumInputs - 2*maxOf(lLeft.getFirstNonEmptyLevel()-1, lNumInputs-1-lLeft.getFirstNonEmptyLevelDual());
        lUtility = 1 << lShift;
    }
    else {
        unsigned int lLeftMax = maxOf(lLeft.getFirstNonEmptyLevel()-1, lNumInputs-1-lLeft.getFirstNonEmptyLevelDual());
        unsigned int lRightMax = maxOf(lRight.getFirstNonEmptyLevel()-1, lNumInputs-1-lRight.getFirstNonEmptyLevelDual());
        unsigned int lShift = lNumInputs - 2*maxOf(lLeftMax, lRightMax);
        lUtility = 1 << (lShift-1);
    }
    return lUtility;

	Beagle_StackTraceEndM("unsigned int SN::findSymmetricFunctionsUtilityBefore(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find the utility of adding the specified comparator based on whether the
 *      line functions are symmetric after the comparator is added.
 */
unsigned int SN::findSymmetricFunctionsUtilityAfter(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    unsigned int lUtility = 0;

    bool lSymmetric = false;
    for (unsigned int i = 0; (lSymmetric == false) && (i < lNumInputs); ++i) {
        const Function& lTemp = (i == inLeftIndex) ? lJoin : ((i == inRightIndex) ? lMeet : lFunctions[i]);
        lSymmetric = true;
        for (unsigned int j = 0; lSymmetric && (j < lNumInputs-1); ++j) {
            lSymmetric = (lJoin.getNumSetsInLevel(j+1) == lTemp.getNumSetsInLevelDual(lNumInputs-1-j));
        }
    }
    if (lSymmetric) lUtility++;

    lSymmetric = false;
    for (unsigned int i = 0; (lSymmetric == false) && (i < lNumInputs); ++i) {
        const Function& lTemp = (i == inLeftIndex) ? lJoin : ((i == inRightIndex) ? lMeet : lFunctions[i]);
        lSymmetric = true;
        for (unsigned int j = 0; lSymmetric && (j < lNumInputs-1); ++j) {
            lSymmetric = (lTemp.getNumSetsInLevel(j+1) == lMeet.getNumSetsInLevelDual(lNumInputs-1-j));
        }
    }
    if (lSymmetric) lUtility++;

    return lUtility;

	Beagle_StackTraceEndM("unsigned int SN::findSymmetricFunctionsUtilityAfter(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find out how many levels are filled up by adding the comparator.
 */
int SN::findFilledLevelsUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    unsigned int lUtility = (lJoin.getNumFullLevels() - lLeft.getNumFullLevels()) + (lMeet.getNumFullLevelsDual() - lRight.getNumFullLevelsDual());

    return lUtility;

	Beagle_StackTraceEndM("int SN::findFilledLevelsUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find if sets belonging to either line are created by adding the comparator.
 */
unsigned int SN::findDestinationLineUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    unsigned int lUtility = 0;
    lUtility += (lMeet.getFirstNonEmptyLevel()-1 == inRightIndex) ? 1 : 0;
    lUtility += (lJoin.getFirstNonEmptyLevelDual() == inLeftIndex) ? 1 : 0;

    return lUtility;

	Beagle_StackTraceEndM("unsigned int SN::findDestinationLineUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}


/*!
 *  \brief Find out the disagreement between the functions for the given comparator.
 */
unsigned int SN::findDisagreementUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)
{
	Beagle_StackTraceBeginM();

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    const std::vector<Function>& lFunctions = inNetwork.getFunctions();
    const Function& lLeft = lFunctions[inLeftIndex];
    const Function& lRight = lFunctions[inRightIndex];
    const Function& lJoin = inNetwork.getJoin(inLeftIndex, inRightIndex);
    const Function& lMeet = inNetwork.getMeet(inLeftIndex, inRightIndex);

    // Find the first level for which there is overlap of the functions.
    const unsigned int lNumLevels = Function::getNumLevels();
    unsigned int lOverlapLevel = 0;
    for (unsigned int i = 1; lOverlapLevel == 0 && i < lNumLevels; ++i) {
        if (lLeft.getNumSetsInLevel(i) + lRight.getNumSetsInLevel(i) > lJoin.getNumSetsInLevel(i)) lOverlapLevel = i;
    }
    unsigned int lOverlapLevelDual = lNumLevels-1;
    for (unsigned int i = 1; lOverlapLevelDual == lNumLevels-1 && i < lNumLevels; ++i) {
        unsigned int j = lNumLevels-1-i;
        if (lLeft.getNumSetsInLevelDual(j) + lRight.getNumSetsInLevelDual(j) > lMeet.getNumSetsInLevelDual(j)) lOverlapLevelDual = j;
    }

    unsigned int lFirstLevel = maxOf(lLeft.getFirstNonEmptyLevel(), lRight.getFirstNonEmptyLevel());
	unsigned int lFirstLevelDual = minOf(lLeft.getFirstNonEmptyLevelDual(), lRight.getFirstNonEmptyLevelDual());
    unsigned int lUtility = (lOverlapLevel - lFirstLevel) + (lFirstLevelDual - lOverlapLevelDual);

    return lUtility;

	Beagle_StackTraceEndM("unsigned int SN::findDisagreementUtility(Network& inNetwork, unsigned int inLeftIndex, unsigned int inRightIndex)");
}
