/*
 *  Co-evolutionary Symbolic Regression (coev_symbreg):
 *  Mixed real-valued GA - GP co-evolution example with Open BEAGLE
 *
 *  Copyright (C) 2003-2006 by Christian Gagne and Jiachuan Wang
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   TrainSetThread.cpp
 *  \brief  Implementation of the class TrainSetThread.
 *  \author Jiachuan Wang <jiacwang@ecs.umass.edu>
 *  \author Christian Gagne <cgagne@gmail.com>
 *  $Revision: 1.16 $
 *  $Date: 2007/08/17 18:09:14 $
 */

#include "beagle/Coev.hpp"
#include "CoSymEvalOp.hpp"
#include "TrainSetEvalOp.hpp"
#include "TrainSetThread.hpp"
#include "SymGPEvalOp.hpp"
#include "SymGPThread.hpp"

#include <cmath>
#include <cstdlib>
#include <iostream>
#include <stdexcept>
#include <vector>
#include <numeric>

using namespace Beagle;

/*!
 *  \brief Construct an training set execution thread for the
 *    co-evolutionary symbolic regression.
 *  \param inConfigFilename Name of the configuration file to use.
 */
TrainSetThread::TrainSetThread(std::string inConfigFilename) :
		mConfigFilename(inConfigFilename)
{ }


/*!
 *  \brief Destruct an execution thread for the TrainSet problem.
 */
TrainSetThread::~TrainSetThread()
{
	wait();
}


/*!
 *  \brief Evolver the TrainSet problem.
 */
void TrainSetThread::main()
{
	try {
	
		// Build the system
		System::Handle lSystem = new System;

		// Add the GA float vector and coevolution packages
		const unsigned int lVectorSize=20;
		lSystem->addPackage(new GA::PackageFloatVector(lVectorSize));
		lSystem->addPackage(new Coev::Package);

		// Set fitness evaluation operator allocator
		lSystem->setEvaluationOp("TrainSetEvalOp", new TrainSetEvalOp::Alloc);

		// Initialize the evolver
		Evolver::Handle lEvolver = new Evolver;
		lEvolver->initialize(lSystem, mConfigFilename);

		// Create population
		Vivarium::Handle lVivarium = new Vivarium;

		// Launch evolution
		lEvolver->evolve(lVivarium, lSystem);

	} catch(Exception& inException) {
		inException.terminate(std::cerr);
	} catch(std::exception& inException) {
		std::cerr << "Standard exception catched:" << std::endl << std::flush;
		std::cerr << inException.what() << std::endl << std::flush;
	}
}
