/*
 *  Iterated Prisoner's Dilemma (ipd):
 *  Co-evolution GA example
 *
 *  Copyright (C) 2003-2007 by Christian Gagne and Marc Parizeau
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   IPDThread.cpp
 *  \brief  Implementation of the class IPDThread.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.18 $
 *  $Date: 2007/08/17 18:09:15 $
 */

#include "beagle/Coev.hpp"
#include "IPDThread.hpp"
#include "IPDEvalOp.hpp"

using namespace Beagle;


/*!
 *  \brief Construct an execution thread for the IPD problem.
 *  \param inConfigFilename Name of the configuration file to use.
 */
IPDThread::IPDThread(std::string inConfigFilename) :
		mConfigFilename(inConfigFilename)
{ }


/*!
 *  \brief Destruct an execution thread for the IPD problem.
 */
IPDThread::~IPDThread()
{
	wait();
}


/*!
 *  \brief Evolver the IPD problem.
 */
void IPDThread::main()
{
	try {

		// Build the system.
		System::Handle lSystem = new System;

		// Adding the GA bit string and coevolution packages
		const unsigned int lNumberOfBits=3;
		lSystem->addPackage(new GA::PackageBitString(lNumberOfBits));
		lSystem->addPackage(new Coev::Package);
		
		// Set fitness evaluation operator allocator
		lSystem->setEvaluationOp("IPDEvalOp", new IPDEvalOp::Alloc);

		// Initialize the evolver
		Evolver::Handle lEvolver = new Evolver;
		lEvolver->initialize(lSystem, mConfigFilename);

		// Create population
		Vivarium::Handle lVivarium = new Vivarium;

		// Launch evolution
		lEvolver->evolve(lVivarium, lSystem);

	} catch(Exception& inException) {
		inException.terminate(std::cerr);
	} catch(std::exception& inException) {
		std::cerr << "Standard exception catched:" << std::endl << std::flush;
		std::cerr << inException.what() << std::endl << std::flush;
	}
}
