/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2007 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   OptFctFloatMain.cpp
 *  \brief  Main routine for the real-valued function optimization problems.
 *  \author Christian Gagne
 *  $Revision: 1.9 $
 *  $Date: 2007/08/17 18:09:22 $
 */

#include "Functions.hpp"
#include "OptFctFloatEvalOpT.hpp"

#include <cstdlib>
#include <iostream>
#include <stdexcept>
#include <vector>
#include <numeric>

using namespace std;
using namespace Beagle;


/*!
 *  \brief Main routine for the real-valued function optimization problems.
 *  \param argc Number of arguments on the command-line.
 *  \param argv Arguments on the command-line.
 *  \return Return value of the program.
 */
int main(int argc, char** argv)
{
	try {

		// Build the system and set with CMA-ES package
		System::Handle lSystem = new System;
		lSystem->addPackage(new GA::PackageCMAES(5));

		// Add fitness evaluation operator allocators to the factory
		Factory& lFactory = lSystem->getFactory();
		lFactory.insertAllocator("SphereEvalOp", new OptFctFloatEvalOpT<SphereFunct>::Alloc);
		lFactory.insertAllocator("SchwefelEvalOp", new OptFctFloatEvalOpT<SchwefelFunct>::Alloc);
		lFactory.insertAllocator("F3EvalOp", new OptFctFloatEvalOpT<F3Funct>::Alloc);
		lFactory.insertAllocator("RosenbrockEvalOp", new OptFctFloatEvalOpT<RosenbrockFunct>::Alloc);
		lFactory.insertAllocator("RastriginEvalOp", new OptFctFloatEvalOpT<RastriginFunct>::Alloc);
		lFactory.setConcept("EvaluationOp", "SphereEvalOp");

		// Make use of FitnessSimpleMin fitness
		lFactory.setConcept("Fitness", "FitnessSimpleMin");

		// Initialize the evolver and population, and do the evolution
		Evolver::Handle lEvolver = new Evolver;
		lEvolver->initialize(lSystem, argc, argv);
		Vivarium::Handle lVivarium = new Vivarium;
		lEvolver->evolve(lVivarium, lSystem);

		// Log best-of-run solution
		if(lVivarium->getHallOfFame() != NULL) {
			Beagle_LogObjectM(
				lSystem->getLogger(),
				Beagle::Logger::eBasic,
				"main",
				"main",
				*lVivarium->getHallOfFame()
			);
		}

	} catch(Exception& inException) {
		inException.terminate(cerr);
	} catch(std::exception& inException) {
		cerr << "Standard exception catched:" << endl << flush;
		cerr << inException.what() << endl << flush;
		return 1;
	}
	return 0;
}
