/*
 *  Symbolic regression test code.
 *  Copyright (C) 2001-2006 by Christian Gagne and Marc Parizeau
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   Functions.hpp
 *  \brief  Available functions for testing GP symbolic regression.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.2 $
 *  $Date: 2007/08/01 19:49:30 $
 */

#ifndef Functions_hpp
#define Functions_hpp

#include "PACC/Util.hpp"


/*!
 *  Basic power 4 monovariate polynomial.
 */
class Power4Function
{

public:

	Power4Function()
	{ }

	/*!
	 *  \brief Return value of function at given sample.
	 *  \param inSample Sampled point for which function should be evaluated.
	 *  \return Value from function evaluated at given point.
	 */
	inline double evaluate(const std::vector<double>& inSample) const
	{
		const double lX = inSample[0];
		return (lX * (lX * (lX * (lX + 1.) + 1.) + 1.));
	}

	/*!
	 *  \brief Make a sample of a point to regress.
	 *  \param ioRand Random number generator to use to generate sample.
	 *  \return Vector with values of sample.
	 */
	inline std::vector<double> sample(PACC::Randomizer& ioRand) const
	{
		std::vector<double> lSample(1);
		lSample[0] = ioRand.getFloat(-1.,1.);
		return lSample;
	}

};


/*!
 *  Basic quadratic monovariate polynomial.
 */
class QuadraticFunction
{

public:

	QuadraticFunction()
	{ }

	/*!
	 *  \brief Return value of function at given sample.
	 *  \param inSample Sampled point for which function should be evaluated.
	 *  \return Value from function evaluated at given point.
	 */
	inline double evaluate(const std::vector<double>& inSample) const
	{
		const double lX = inSample[0];
		return ((lX - 4.) * (lX - 4.)) - 3.;
	}

	/*!
	 *  \brief Make a sample of a point to regress.
	 *  \param ioRand Random number generator to use to generate sample.
	 *  \return Vector with values of sample.
	 */
	inline std::vector<double> sample(PACC::Randomizer& ioRand) const
	{
		std::vector<double> lSample(1);
		lSample[0] = ioRand.getFloat(-1.,1.);
		return lSample;
	}

};


/*!
 *  Sphere function of arbitrary dimension.
 */
template<int Dim>
class SphereFunction
{

public:

	SphereFunction<Dim>()
	{ }

	/*!
	 *  \brief Return value of function at given sample.
	 *  \param inSample Sampled point for which function should be evaluated.
	 *  \return Value from function evaluated at given point.
	 */
	inline double evaluate(const std::vector<double>& inSample) const
	{
		double lResult = 0.;
		for(unsigned int i=0; i<Dim; ++i) {
			lResult += (inSample[i] * inSample[i]);
		}
		return lResult;
	}

	/*!
	 *  \brief Make a sample of a point to regress.
	 *  \param ioRand Random number generator to use to generate sample.
	 *  \return Vector with values of sample.
	 */
	inline std::vector<double> sample(PACC::Randomizer& ioRand) const
	{
		std::vector<double> lSample(Dim);
		for(unsigned int i=0; i<Dim; ++i) {
			lSample[i] = ioRand.getFloat(-1.,1.);
		}
		return lSample;
	}

};


#endif // Functions_hpp
