/*
 *  Symbolic regression test code.
 *  Copyright (C) 2001-2006 by Christian Gagne and Marc Parizeau
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   SymbRegEvalOpT.hpp
 *  \brief  Template of fitness evaluation operator for GP symbolic regression.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.2 $
 *  $Date: 2007/08/01 19:49:30 $
 */

#ifndef SymbRegEvalOpT_hpp
#define SymbRegEvalOpT_hpp

#include "beagle/GP.hpp"
#include "Functions.hpp"
#include <string>
#include <vector>


/*!
 *  \class SymbRegEvalOpT SymbRegEvalOpT.hpp "SymbRegEvalOpT.hpp"
 *  \brief Fitness evaluation operator symbolic regression.
 */
template <class T>
class SymbRegEvalOpT : public Beagle::GP::EvaluationOp
{

public:

	//! SymbRegEvalOpT allocator type.
	typedef Beagle::AllocatorT<SymbRegEvalOpT<T>,Beagle::GP::EvaluationOp::Alloc>
	Alloc;
	//!< SymbRegEvalOpT handle type.
	typedef Beagle::PointerT<SymbRegEvalOpT<T>,Beagle::GP::EvaluationOp::Handle>
	Handle;
	//!< SymbRegEvalOpT bag type.
	typedef Beagle::ContainerT<SymbRegEvalOpT<T>,Beagle::GP::EvaluationOp::Bag>
	Bag;

	explicit SymbRegEvalOpT(std::string inName="SymbRegEvalOpT");

	virtual Beagle::Fitness::Handle evaluate(Beagle::GP::Individual& inIndividual,
	        Beagle::GP::Context& ioContext);
	virtual void init(Beagle::System& ioSystem);
	virtual void registerParams(Beagle::System& ioSystem);
	virtual void testIndividual(std::vector< std::vector<double> >& outSamples,
	                            std::vector< double >& outDesiredValues,
	                            std::vector< double >& outObtainedValues,
	                            Beagle::GP::Individual::Handle inIndividual,
	                            Beagle::GP::Context& ioContext) const;

	/*!
	 *  \return Whether operator has been initialized, so used for fitness evaluation.
	 */
	inline bool isInitialized() const
	{
		return mIsInitialized;
	}

protected:

	Beagle::UInt::Handle                       mNumberOfSamples;
	std::vector< std::vector<Beagle::Double> > mX;
	std::vector< Beagle::Double >              mF;
	bool                                       mIsInitialized;

};


/*!
 *  \brief Construct an evaluation operator for GP simbolic regression.
 *  \param inName Operator name.
 */
template <class T>
SymbRegEvalOpT<T>::SymbRegEvalOpT(std::string inName) :
		Beagle::GP::EvaluationOp(inName),
		mIsInitialized(false)
{ }


/*!
 *  \brief Evaluate fitness from GP individual for symbolic regression.
 *  \param inIndividual Individual for symbolic regression fitness evaluation.
 *  \param ioContext Evolutionary context.
 *  \return Fitness of evaluated individual.
 */
template <class T>
Beagle::Fitness::Handle SymbRegEvalOpT<T>::evaluate(Beagle::GP::Individual& inIndividual,
        Beagle::GP::Context& ioContext)
{
	double lSquareError = 0.;
	for(unsigned int i=0; i<mX.size(); ++i) {
		for(unsigned int j=0; j<mX[i].size(); ++j) {
			std::ostringstream lOSS;
			lOSS << "X" << j;
			setValue(lOSS.str(), mX[i][j], ioContext);
		}
		Beagle::Double lResult;
		inIndividual.run(lResult, ioContext);
		const double lError = mF[i].getWrappedValue() - lResult.getWrappedValue();
		lSquareError += (lError*lError);
	}
	const double lMSE = lSquareError / mX.size();
	const double lRMSE = std::sqrt(lMSE);
	const double lFitness = (1. / (lRMSE + 1.));
	return new Beagle::FitnessSimple(lFitness);
}


/*!
 *  \brief Initialize individual by sampling function to regress.
 *  \param ioSystem Evolutionary system.
 */
template <class T>
void SymbRegEvalOpT<T>::init(Beagle::System& ioSystem)
{
	Beagle::GP::EvaluationOp::init(ioSystem);
	T lFunction;
	mX.resize(mNumberOfSamples->getWrappedValue());
	mF.resize(mNumberOfSamples->getWrappedValue());
	for(unsigned int i=0; i<mNumberOfSamples->getWrappedValue(); ++i) {
		std::vector<double> lSamples = lFunction.sample(ioSystem.getRandomizer());
		mX[i].resize(lSamples.size());
		for(unsigned int j=0; j<lSamples.size(); ++j) {
			mX[i][j].getWrappedValue() = lSamples[j];
		}
		mF[i].getWrappedValue() = lFunction.evaluate(lSamples);
	}
	mIsInitialized  = true;
}


/*!
 *  \brief Register operator parameters.
 *  \param ioSystem Evolutionary system.
 */
template <class T>
void SymbRegEvalOpT<T>::registerParams(Beagle::System& ioSystem)
{
	Beagle::GP::EvaluationOp::registerParams(ioSystem);
	Beagle::Register::Description lDescription(
	    "Number of samples",
	    "UInt",
	    "20",
	    "Number of samples of function to regress used for fitness evalution."
	);
	mNumberOfSamples = Beagle::castHandleT<Beagle::UInt>(
	                       ioSystem.getRegister().insertEntry("sr.eval.nbsamples", new Beagle::UInt(20), lDescription));
}


/*!
 *  \brief Test the individual and return values on samples.
 *  \param outSamples X-values of samples.
 *  \param outDesiredValues Desired values from samples.
 *  \param outObtainedValues Obtained value from samples.
 *  \param inIndividual Individual to test.
 *  \param ioContext Evolutionary context to use.
 */
template <class T>
void SymbRegEvalOpT<T>::testIndividual(std::vector< std::vector<double> >& outSamples,
                                       std::vector< double >& outDesiredValues,
                                       std::vector< double >& outObtainedValues,
                                       Beagle::GP::Individual::Handle inIndividual,
                                       Beagle::GP::Context& ioContext) const
{
	Beagle::GP::Individual::Handle lOldIndividual = ioContext.getIndividualHandle();
	ioContext.setIndividualHandle(inIndividual);
	outSamples.resize(mX.size());
	outDesiredValues.resize(mX.size());
	outObtainedValues.resize(mX.size());
	for(unsigned int i=0; i<mX.size(); ++i) {
		outSamples[i].resize(mX[i].size());
		outDesiredValues[i] = mF[i].getWrappedValue();
		for(unsigned int j=0; j<mX[i].size(); ++j) {
			std::ostringstream lOSS;
			lOSS << "X" << j;
			setValue(lOSS.str(), mX[i][j], ioContext);
			outSamples[i][j] = mX[i][j];
		}
		Beagle::Double lResult;
		inIndividual->run(lResult, ioContext);
		outObtainedValues[i] = lResult.getWrappedValue();
	}
	ioContext.setIndividualHandle(lOldIndividual);
}


#endif // SymbRegEvalOpT_hpp
