/*
 *  Symbolic regression (symbreg):
 *  A simple GP example with Open BEAGLE
 *  Copyright (C) 2001-2006 by Christian Gagne and Marc Parizeau
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   SymbRegMain.cpp
 *  \brief  Implementation of the main routine for the symbolic regression problem.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.7 $
 *  $Date: 2007/08/17 18:09:22 $
 */

#include "beagle/GP.hpp"
#include "Functions.hpp"
#include "SymbRegEvalOpT.hpp"

#define _USE_MATH_DEFINES
#include <cmath>
#include <cstdlib>
#include <iostream>
#include <stdexcept>
#include <vector>
#include <numeric>

using namespace std;
using namespace Beagle;


template <class T>
void logIndividual(Beagle::GP::Individual::Handle inIndividual,
                   const typename SymbRegEvalOpT<T>::Handle inEvalOp,
                   System::Handle ioSystem);


/*!
 *  \brief Main routine for the function symbolic regression problem.
 *  \param argc Number of arguments on the command-line.
 *  \param argv Arguments on the command-line.
 *  \return Return value of the program.
 */
int main(int argc, char *argv[])
{

	try {

		// Build primitives
		GP::PrimitiveSet::Handle lSet = new GP::PrimitiveSet;
		lSet->insert(new GP::Add);
		lSet->insert(new GP::Subtract);
		lSet->insert(new GP::Multiply);
		lSet->insert(new GP::Divide);
		// lSet->insert(new GP::Sin);
		// lSet->insert(new GP::Cos);
		// lSet->insert(new GP::Exp);
		// lSet->insert(new GP::Log);
		lSet->insert(new GP::TokenT<Double>("X0"), 1.);
		for(unsigned int i=1; i<5; ++i) {
			std::ostringstream lOSS;
			lOSS << "X" << i;
			lSet->insert(new GP::TokenT<Double>(lOSS.str()), 0.);
		}
		lSet->insert(new GP::EphemeralDouble);

		// Build a system and add GP package
		System::Handle lSystem = new System();
		lSystem->addPackage(new GP::PackageBase(lSet));

		// Add fitness evaluation operator allocators to factory
		Factory& lFactory = lSystem->getFactory();
		lFactory.insertAllocator("Power4EvalOp",    new SymbRegEvalOpT< Power4Function >::Alloc);
		lFactory.insertAllocator("QuadraticEvalOp", new SymbRegEvalOpT< QuadraticFunction >::Alloc);
		lFactory.insertAllocator("Sphere2EvalOp",   new SymbRegEvalOpT< SphereFunction<2> >::Alloc);
		lFactory.insertAllocator("Sphere3EvalOp",   new SymbRegEvalOpT< SphereFunction<3> >::Alloc);
		lFactory.insertAllocator("Sphere5EvalOp",   new SymbRegEvalOpT< SphereFunction<5> >::Alloc);
		lFactory.setConcept("EvaluationOp", "Power4EvalOp");

		// Create evolver, population, and do evolution
		Evolver::Handle lEvolver = new Evolver;
		lEvolver->initialize(lSystem, argc, argv);
		Vivarium::Handle lVivarium = new Vivarium;
		lEvolver->evolve(lVivarium, lSystem);

		// Log information on best-of-run solutions.
		Beagle::HallOfFame::Handle lHoF = lVivarium->getHallOfFame();
		if(lHoF != NULL) {
			for(unsigned int i=0; i<lHoF->size(); ++i) {
				GP::Individual::Handle lIndividual =
					Beagle::castHandleT<GP::Individual>((*lHoF)[i].mIndividual);
				std::string lEvalOpName = lFactory.getConceptTypeName("EvaluationOp");

				if(lEvalOpName == "Power4EvalOp") {
					SymbRegEvalOpT< Power4Function >::Alloc::Handle lEvalOpAlloc =
						castHandleT< SymbRegEvalOpT< Power4Function >::Alloc >(lFactory.getConceptAllocator("EvaluationOp"));
					SymbRegEvalOpT< Power4Function >::Handle lEvalOp =
						castHandleT< SymbRegEvalOpT< Power4Function > >(lEvalOpAlloc->allocate());
					logIndividual<Power4Function>(lIndividual, lEvalOp, lSystem);
				} else if(lEvalOpName == "QuadraticEvalOp") {
					SymbRegEvalOpT< QuadraticFunction >::Alloc::Handle lEvalOpAlloc =
						castHandleT< SymbRegEvalOpT< QuadraticFunction >::Alloc >(lFactory.getConceptAllocator("EvaluationOp"));
					SymbRegEvalOpT< QuadraticFunction >::Handle lEvalOp =
						castHandleT< SymbRegEvalOpT< QuadraticFunction > >(lEvalOpAlloc->allocate());
					logIndividual<QuadraticFunction>(lIndividual, lEvalOp, lSystem);
				} else if(lEvalOpName == "Sphere2EvalOp") {
					SymbRegEvalOpT< SphereFunction<2> >::Alloc::Handle lEvalOpAlloc =
						castHandleT< SymbRegEvalOpT< SphereFunction<2> >::Alloc >(lFactory.getConceptAllocator("EvaluationOp"));
					SymbRegEvalOpT< SphereFunction<2> >::Handle lEvalOp =
						castHandleT< SymbRegEvalOpT< SphereFunction<2> > >(lEvalOpAlloc->allocate());
					logIndividual< SphereFunction<2> >(lIndividual, lEvalOp, lSystem);
				} else if(lEvalOpName == "Sphere3EvalOp") {
					SymbRegEvalOpT< SphereFunction<3> >::Alloc::Handle lEvalOpAlloc =
						castHandleT< SymbRegEvalOpT< SphereFunction<3> >::Alloc >(lFactory.getConceptAllocator("EvaluationOp"));
					SymbRegEvalOpT< SphereFunction<3> >::Handle lEvalOp =
						castHandleT< SymbRegEvalOpT< SphereFunction<3> > >(lEvalOpAlloc->allocate());
					logIndividual< SphereFunction<3> >(lIndividual, lEvalOp, lSystem);
				} else if(lEvalOpName == "Sphere5EvalOp") {
					SymbRegEvalOpT< SphereFunction<5> >::Alloc::Handle lEvalOpAlloc =
						castHandleT< SymbRegEvalOpT< SphereFunction<5> >::Alloc >(lFactory.getConceptAllocator("EvaluationOp"));
					SymbRegEvalOpT< SphereFunction<5> >::Handle lEvalOp =
						castHandleT< SymbRegEvalOpT< SphereFunction<5> > >(lEvalOpAlloc->allocate());
					logIndividual< SphereFunction<5> >(lIndividual, lEvalOp, lSystem);
				}
			}
		}

	} catch(Exception& inException) {
		inException.terminate();
	} catch(exception& inException) {
		cerr << "Standard exception caught:" << endl;
		cerr << inException.what() << endl << flush;
		return 1;
	} catch(...) {
		cerr << "Unknown exception caught!" << endl << flush;
		return 1;
	}
	return 0;
}


/*!
 *  \brief Log detailled information about an individual.
 *  \param inIndividual Individual to log.
 *  \param inEvalOp Fitness evaluation operator.
 *  \param ioSystem Evolutionary system.
 */
template <class T>
void logIndividual(Beagle::GP::Individual::Handle inIndividual,
                   const typename SymbRegEvalOpT<T>::Handle inEvalOp,
                   System::Handle ioSystem)
{
	GP::Context lGPContext;
	lGPContext.setSystemHandle(ioSystem);
	std::vector< std::vector<double> > lSamples;
	std::vector< double > lObtainedValues;
	std::vector< double > lDesiredValues;
	inEvalOp->testIndividual(lSamples,
	                         lObtainedValues,
	                         lDesiredValues,
	                         inIndividual,
	                         lGPContext);
	std::ostringstream lOSS1;
	lOSS1 << "Information on best-of-run individual";
	Beagle_LogBasicM(
	    ioSystem->getLogger(),
	    "main",
	    "main",
	    lOSS1.str()
	);
	Beagle_LogObjectM(
	    ioSystem->getLogger(),
	    Beagle::Logger::eBasic,
	    "main",
	    "main",
	    *inIndividual
	);
	std::ostringstream lOSS2;
	lOSS2 << "Samples: [";
	for(unsigned int i=0; i<lSamples.size(); ++i) {
		for(unsigned int j=0; j<lSamples[i].size(); ++j) {
			lOSS2 << lSamples[i][j];
			if(j != (lSamples[i].size()-1)) lOSS2 << ",";
		}
		if(i != (lSamples.size()-1)) lOSS2 << "; ";
	}
	lOSS2 << "]";
	Beagle_LogBasicM(
	    ioSystem->getLogger(),
	    "main",
	    "main",
	    lOSS2.str()
	);
	std::ostringstream lOSS3;
	lOSS3 << "Desired outputs: [";
	for(unsigned int i=0; i<lDesiredValues.size(); ++i) {
		lOSS3 << lDesiredValues[i];
		if(i != (lDesiredValues.size()-1)) lOSS3 << "; ";
	}
	lOSS3 << "]";
	Beagle_LogBasicM(
	    ioSystem->getLogger(),
	    "main",
	    "main",
	    lOSS3.str()
	);
	std::ostringstream lOSS4;
	lOSS4 << "Obtained outputs: [";
	for(unsigned int i=0; i<lObtainedValues.size(); ++i) {
		lOSS4 << lObtainedValues[i];
		if(i != (lObtainedValues.size()-1)) lOSS4 << "; ";
	}
	lOSS4 << "]";
	Beagle_LogBasicM(
	    ioSystem->getLogger(),
	    "main",
	    "main",
	    lOSS4.str()
	);
}
