
################################################################################
##
## Given the number of graph vertices, compute the orbital partitions (colors)
## corresponding to the elements of the subgroup lattice of the symmetric group
## of that degree, and write the lattice colors to the given file.
##
## Only those elements of the subgroup lattice that are useful for representing
## the edges of undirected graphs are included in the output, i.e. each element
## in the output will contain at least one color that represents a symmetric
## adjacency matrix.  The following function determines if a given group has this
## property.

IsUndirected := function(nvert, group)
    local eorbits, trans, i;
    
    # At least one edge orbit must be its own transpose.
    eorbits := OrbitsDomain(group, Arrangements([1..nvert], 2), OnPairs);
    for i in [1..Length(eorbits)] do
        trans := List(eorbits[i], a -> [a[2], a[1]]);
        if IsEqualSet(eorbits[i], trans) then
            return true;
        fi;
    od;
    
    return false;
    
end;


## The following function is adapted from the function DotFileLatticeSubgroups()
## from Alexander Hulpke <ahulpke@gmail.com>:
## http://mail.gap-system.org/pipermail/forum/2008/002250.html
##
## This function is called at the bottom of the file to generate the XML file.
## Modify the number of vertices and the file name in that call to generate
## lattice colors for a different number of graph vertices. Then call gap as
## follows:
##    gap < LatticeColorsUndirected.gap
##

LatticeColorsUndirected := function(nvert, filename)
    local nelem, fstream, indent, str, sstream, index, children, vorbits, eorbits,
          color, L, cls, len, size, max, rep, z, t, i, j, k;
    
    L := LatticeSubgroups(SymmetricGroup(nvert));
    cls:=ConjugacyClassesSubgroups(L);
    len:=[];
    size:=[];
    for i in cls do
        Add(len,Size(i));
        Add(size,Size(Representative(i)));
    od;
    
    # If lattice elements are arranged in a vector, find the index at which
    # each conjugacy class starts and the number of elements in the vector.
    nelem := 0;
    index:=[];
    for i in [1..Length(cls)] do
        nelem := nelem + len[i];
        Add(index, nelem);
    od;
    
    # Indent levels of the XML output file.
    str := "";
    indent := [];
    for i in [1..3] do
        sstream := OutputTextString(str, true);
        AppendTo(sstream, "  ");
        CloseStream(sstream);
        Add(indent, StructuralCopy(str));
    od;
    
    # Write to XML output file.
    fstream := OutputTextFile(filename, false);
    AppendTo(fstream, "<Lattice size=\"", nelem, "\" numVertices=\"", nvert, "\">\n");
            
    # Find the orbital partitions and maximal subgroup relations in the lattice.
    # Notice that the groups are arranged from small to big in cls; we reverse
    # this order in the XML output.
    max:=MaximalSubgroupsLattice(L);
    for i in Reversed([1..Length(cls)]) do
        for k in [1..len[i]] do
            if not IsUndirected(nvert, ClassElementLattice(cls[i], k)) then
                continue;
            fi;
            
            AppendTo(fstream, indent[1], "<Element index=\"", nelem-index[i]+k-1, "\" groupSize=\"", size[i], "\">\n");
                    
            # Find orbits of vertices and edges.
            vorbits := OrbitsDomain(ClassElementLattice(cls[i], k), [1..nvert], OnPoints);
            eorbits := OrbitsDomain(ClassElementLattice(cls[i], k), Arrangements([1..nvert], 2), OnPairs);
            
            # Encode the orbits as colors and write to XML file.
            AppendTo(fstream, indent[2], "<Colors size=\"", Length(vorbits)+Length(eorbits), "\">\n");
            for j in [1..Length(vorbits)] do
                AppendTo(fstream, indent[3], "<Color index=\"", j-1, "\">");
                color := List([1..nvert*nvert], a -> 0);
                for z in vorbits[j] do
                    color[(z-1)*nvert + z] := 1;
                od;
                for z in color do
                    AppendTo(fstream, z);
                od;
                AppendTo(fstream, "</Color>\n");
            od;
            for j in [1..Length(eorbits)] do
                AppendTo(fstream, indent[3], "<Color index=\"", Length(vorbits)+j-1, "\">");
                color := List([1..nvert*nvert], a -> 0);
                for z in eorbits[j] do
                    color[(z[1]-1)*nvert + z[2]] := 1;
                od;
                for z in color do
                    AppendTo(fstream, z);
                od;
                AppendTo(fstream, "</Color>\n");
            od;
            AppendTo(fstream, indent[2], "</Colors>\n");
            
            # Find maximal subgroup relations between the lattice elements.
            children := [];
            for j in max[i] do
                if k=1 then # no need to conjugate if class has only one element
                    z:=j[2];
                else
                    t:=cls[i]!.normalizerTransversal[k];
                    z:=ClassElementLattice(cls[j[1]],1); # force computation of transv.
                    z:=cls[j[1]]!.normalizerTransversal[j[2]]*t;
                    z:=PositionCanonical(cls[j[1]]!.normalizerTransversal,z);
                fi;
                if IsUndirected(nvert, ClassElementLattice(cls[j[1]], z)) then
                    Add(children, nelem-index[j[1]]+z-1);
                fi;
            od;
            
            if Length(children) > 0 then
                AppendTo(fstream, indent[2], "<Children size=\"", Length(children), "\">");
                for j in children do
                    if j = children[1] then
                        AppendTo(fstream, j);
                    else
                        AppendTo(fstream, ";", j);
                    fi;
                od;
                AppendTo(fstream, "</Children>\n");
            fi;
            
            AppendTo(fstream, indent[1], "</Element>\n");
        od;
    od;
    
    AppendTo(fstream,"</Lattice>\n");
    CloseStream(fstream);
    
end;


# Specify the number of graph vertices in the following call.
LatticeColorsUndirected(4, "LatticeColorsUndirected-4.xml");
