/*
 * CM-2+ Walking Tables Control Program
 * Date : May-Jun 2010
 * Author : Vinod Valsalam
 *
 * Based on the Machine Metabolism Control Program
 * by Jeremy Blum and Franz Nigl.
 */

 
#include "dynamixel.h"

//--- Mapping of AX12 Motor IDs to legs of the robot ---		
#define M_L1				107
#define M_L2				103
#define M_L3				103
#define M_R1				102
#define M_R2				104
#define M_R3				101

// Number of servos actually used.
#define NUM_SERVOS			4

// Absolute value of a number.
#define ABSOLUTE(x)         (((x) < 0) ? (-(x)) : (x))

// Parameters and settings used in simulation.
//uint16_t gAX12MaxVel = 900;  // max velocity setting to use for the motors
//double gOutMult = 6.015/(2*M_PI);   // max rotations per second, corresponding to the above setting
uint16_t gAX12MaxVel = 800;  // max velocity setting to use for the motors
double gOutMult = 5.35/(2*M_PI);   // max rotations per second, corresponding to the above setting


int main(void)
{
	uint8_t lServos[NUM_SERVOS] = {M_L1, M_R1, M_L2, M_R2};
    double lVelScale[NUM_SERVOS] = {1.0, 1.0, 1.0, 1.0};  // Scale to modify servo speeds.
	uint8_t i;  // Variable to iterate over the servos.
    double lMaxTime = 60.0;  // Duration for which to run the robot.
    uint16_t lVelLowerBound = gAX12MaxVel-100;  // Lower limit on servo angular velocity.
    uint16_t lVelUpperBound = gAX12MaxVel+100;  // Upper limit on servo angular velocity.

    // Initial positions of the servos for trot gait.  Assume M_L1 starts at 0 degrees.
    // Then L1 is actually at 30+45=75 degrees away from the closest point to the body
    // (i.e. the point at which the leg is vertical) because L1 is ahead of M_L1 by 45
    // degrees and M_L1 itself starts counting at an offset of 30 degrees from the
    // vertical.  Since the diagonal legs L1 and R2 must be vertical at the same time,
    // and R2 rotates in the opposite direction, R2 should at -75 degrees.  If M_R2 is
    // at its upper bound of 300 degrees, then R2 is at 15 degrees past the vertical.
    // So M_R2 should start at 300-15-75=210 degrees.  Since the legs on the other
    // diagonal are 180 degrees apart, M_L2 should start at 180 degrees and M_R1 should
    // start at 30 degrees.
    uint16_t lInitialPosition[NUM_SERVOS] = {0, 102, 614, 716};  // {0, 30, 180, 210} degrees
    //uint16_t lInitialPosition[NUM_SERVOS] = {0, 102, 0, 716};  // For initial error experiment

	Initialize();  // Initialize the controller
	
	// Put motors in servo mode and initialize their positions.
	TxDString("Initializing motor positions.\r\n");
	for (i = 0; i < NUM_SERVOS; i++) {
        // Disable torque before enabling servo mode to prevent the motor from moving
        // when servo mode is enabled.  Torque is automatically enabled by setting
        // position/speed.
        DisableTorque(lServos[i]);
		EnableServoMode(lServos[i]);
		SetGoalPositionWithSpeed(lServos[i], lInitialPosition[i], 100);
	}

	for (i = 0; i < NUM_SERVOS; i++) {
		while (IsMoving(lServos[i]));
		TxDString("Motor "); TxD32Dec(lServos[i]); TxDString(" position: "); TxD32Dec(GetPosition(lServos[i])); TxDString("\r\n");
	}
	
	// Put motors in continous mode.
	TxDString("Configuring Motors for continuous mode.\r\n");
	for (i = 0; i < NUM_SERVOS; i++) {
		DisableServoMode(lServos[i]);
	}

    // Variables for calculating position in the blind region of AX12 servos.
    int lServoReading[NUM_SERVOS];
    int lEstimatedReading[NUM_SERVOS];                           
    double lEstimatedVel[NUM_SERVOS];
    double lDesiredVel[NUM_SERVOS];                              
	uint8_t lBlindRegion[NUM_SERVOS];
	uint8_t lPrevBlindRegion[NUM_SERVOS];
    double lErrorIntegral[NUM_SERVOS];                              
    double lPrevError[NUM_SERVOS];                              
    for (i = 0; i < NUM_SERVOS; i++) {
        lServoReading[i] = lInitialPosition[i];
        lEstimatedReading[i] = lInitialPosition[i];
        lEstimatedVel[i] = 0.0;
        lDesiredVel[i] = 0.0;												 
		lBlindRegion[i] = 0;
		lPrevBlindRegion[i] = 0;
		lErrorIntegral[i] = 0.0;
		lPrevError[i] = 0.0;
    }
    // Exponential smoothing factor for velocity.							 
    double lAlpha = 0.5;

    // PID gain constants
    double lKp = 1.0;
    double lKi = 1.0;
    double lKd = 1.0;

    double lStepSize = 0.01;   // match step size with that in simulation.
    double lStepStartTime = time + lStepSize;
    double lStartTime = lStepStartTime + lStepSize;
	lMaxTime += lStepStartTime;
    while (time < lMaxTime) {
        DelayMilliSec((lStepStartTime-time)*1000);
        lStepStartTime = time + lStepSize;

        // Read all servos at about the same time.
		for (i = 0; i < NUM_SERVOS; i++) {
			lServoReading[i] = GetPosition(lServos[i]);
        }

        //TxDString("Trace "); TxD32Dec(time*1000); TxDString(" ");
		for (i = 0; i < NUM_SERVOS; i++) {
            // In order to deal with the blind region, we calculate the input to the	 
            // neural network from the estimated reading instead of the direct servo	 
            // reading.	 The estimated reading is calculated from the estimated velocity,
            // which is adjusted based on the servo reading outside the blind region.	 
            lEstimatedReading[i] = lEstimatedReading[i] + (int)lEstimatedVel[i];		 
            // Make sure the estimated reading is within range.							 
            lEstimatedReading[i] = (lEstimatedReading[i]+1228)%1228;					 
																									 
            // Update estimated velocity if servo is not in the blind region.			 
            // The AX12 servos give position feedback only for angles in range [0, 300],
            // i.e. for lNormAnlge < 300/360.  Moreover, they give integers in [0, 1023]
            // for this range.  If the angle is not in this range (blind region), the
			// servo reading gives 1023 or 0 except for some time in between.  We will
			// use this observation to detect being in the blind region.
            if (lBlindRegion[i] == 0 && lServoReading[i] < 1023 && lServoReading[i] > 0) {
                // Since we assume that the motor is turning at a slow enough rate to	 
                // allow taking more than two readings in a rotation, if the above		 
                // difference is greater than 1228/2, then the reading has wrapped around
                // the lower or upper bound, and we need to add/subtract 1228 to get	 
                // the actual difference.												 
                int lReadingDiff = lServoReading[i] - lEstimatedReading[i];
                if (ABSOLUTE(lReadingDiff) > 614) {
                    lReadingDiff = (lReadingDiff < 0) ? (lReadingDiff+1228) : (lReadingDiff-1228);
                }
                lEstimatedVel[i] = (1-lAlpha)*lEstimatedVel[i] + lAlpha*((double)(lReadingDiff));
            }
            else {
                //lPrevBlindRegion[i] = lBlindRegion[i];

				// Detect entering/leaving the blind region
				if (lEstimatedVel[i] > 0) {
					if (lServoReading[i] == 1023) lBlindRegion[i] = 1;  // entering blind region
					else if (lServoReading[i] == 0) lBlindRegion[i] = 0;  // leaving blind region
				}
				if (lEstimatedVel[i] < 0) {
					if (lServoReading[i] == 0) lBlindRegion[i] = 1;  // entering blind region
					else if (lServoReading[i] == 1023) lBlindRegion[i] = 0;  // leaving blind region
				}

                //if (lPrevBlindRegion[i] == 0 && lBlindRegion[i] == 1) {
                //    TxDString("Motor "); TxD32Dec(lServos[i]); TxDString(": "); TxD32Dec(time*1000); TxDString("\r\n");
				//}
             }


            // Calculate normalized angle from estimated reading.
            double lNormAngle = ((double)lEstimatedReading[i])/1228.0;	 // range [0, 1)

            // The leg is mounted 45 degrees with respect to the zero of the servo.
            lNormAngle += 0.125;

            // Since the zeros of the legs are initially 30 degrees from the vertical,
            // there is a total 60 degrees offset between the left and right sides.
            if (i%2 == 1) lNormAngle += 0.16666667;

            lNormAngle = lNormAngle - floor(lNormAngle);  // range [0, 1)

            // Flip sign of angle for right side legs.
            if (i%2 == 1) lNormAngle = 1.0 - lNormAngle;

            // Calculate the desired angle as a function of time.
            // Trot gait: diagonal legs are synchronous and 180 degrees apart.
            double lDesiredAngle = (lStepStartTime-lStartTime)*gOutMult;
            lDesiredAngle += 0.125;  // Add 45 degrees offset to match lNormAngle.
            if (i == 1 || i == 2) {
                lDesiredAngle = (lDesiredAngle+0.5) - floor(lDesiredAngle+0.5);
            }
            else {
                lDesiredAngle = lDesiredAngle - floor(lDesiredAngle);
            }

            // PID control
            double lError = lDesiredAngle - lNormAngle;
            if (ABSOLUTE(lError) > 0.5) {
                lError = (lError < 0) ? (lError+1.0) : (lError-1.0);
            }
            double lIntegral = lErrorIntegral[i] + lError*lStepSize;
            double lDerivative = (lError - lPrevError[i])/lStepSize;
            lDesiredVel[i] = lKp*lError + lKi*lIntegral + lKd*lDerivative;
            lPrevError[i] = lError;
            lErrorIntegral[i] = lIntegral;

			//TxD32Dec(lNormAngle*1000); TxDString(" ");
			//TxD32Dec(lDesiredAngle*1000); TxDString(" ");
        }

        // Set all servo speeds at about the same time.
		for (i = 0; i < NUM_SERVOS; i++) {
            // The AX12 servos have discrete velocity values in the range [0, 1023] in each
            // direction, and the direction is determined by the most significant bit.
            // The desired velocity calculated above is a correction to the base velocity
            // of +/-1.0.
            if (i%2 == 1) {  // right side legs; negative velocity
                uint16_t lDiscreteVel = (-1.0 - lDesiredVel[i])*gAX12MaxVel;
                if (lDiscreteVel < -lVelUpperBound) lDiscreteVel = -lVelUpperBound;
                if (lDiscreteVel > -lVelLowerBound) lDiscreteVel = -lVelLowerBound;
                SetGoalSpeed(lServos[i], -lDiscreteVel*lVelScale[i]+1024);
                //TxD32Dec(-lDiscreteVel+1024); TxDString(" ");
            }
            else { // left side legs; positive velocity
                uint16_t lDiscreteVel = (1.0 + lDesiredVel[i])*gAX12MaxVel;
                if (lDiscreteVel > lVelUpperBound) lDiscreteVel = lVelUpperBound;
                if (lDiscreteVel < lVelLowerBound) lDiscreteVel = lVelLowerBound;
                SetGoalSpeed(lServos[i], lDiscreteVel*lVelScale[i]);
                //TxD32Dec(lDiscreteVel); TxDString(" ");
            }
        }
		//TxDString("\r\n");
    }
	
	for (i = 0; i < NUM_SERVOS; i++) {
		SetGoalSpeed(lServos[i], 1);
		DisableTorque(lServos[i]);
	}
	
	TxDString("Motion Complete.\r\n");
	
	while (1); // wait for reset button to be pressed.
}
