/*
 * CM-2+ Walking Tables Control Program
 * Date : May-Jun 2010
 * Author : Vinod Valsalam
 *
 * Based on the Machine Metabolism Control Program
 * by Jeremy Blum and Franz Nigl.
 */

 
#include "dynamixel.h"
#include "netutils.h"

//--- Mapping of AX12 Motor IDs to legs of the robot ---		
#define M_L1				107
#define M_L2				103
#define M_L3				103
#define M_R1				102
#define M_R2				104
#define M_R3				101

// Number of servos actually used.
#define NUM_SERVOS			4

// Absolute value of a number.
#define ABSOLUTE(x)         (((x) < 0) ? (-(x)) : (x))

// Parameters and settings used in simulation.
extern uint8_t gInpCosSin;
extern uint16_t gAX12MaxVel;


int main(void)
{
	uint8_t lServos[NUM_SERVOS] = {M_L1, M_R1, M_L2, M_R2};
    double lVelScale[NUM_SERVOS] = {1.0, 1.0, 1.0, 1.0};  // Scale to modify servo speeds.
	uint8_t i;  // Variable to iterate over the servos.
    double lMaxTime = 60.0;  // Duration for which to run the robot.
	double lInputs[gNetwork.numInputs];
	double lOutputs[gNetwork.numOutputs];
    //uint16_t lInitialPosition = 614-153;  // Initial position of the servos (180-45 degrees).
    uint16_t lInitialPosition[NUM_SERVOS] = {614-153, 614-153, 614-153, 614-153};
    //uint16_t lInitialPosition[NUM_SERVOS] = {614, 614, 614, 614};
    //uint16_t lInitialPosition[NUM_SERVOS] = {0, 102, 0, 716};  // For initial error experiment
    //uint16_t lInitialPosition[NUM_SERVOS] = {0, 102, 614, 716};  // {0, 30, 180, 210} degrees

	for (i = 0; i < gNetwork.numInputs; i++) {
		lInputs[i] = 0.0;
	}
	for (i = 0; i < gNetwork.numOutputs; i++) {
		lOutputs[i] = 0.0;
	}

	Initialize();  // Initialize the controller
	
	// Put motors in servo mode and initialize their positions.
	TxDString("Initializing motor positions.\r\n");
	for (i = 0; i < NUM_SERVOS; i++) {
        // Disable torque before enabling servo mode to prevent the motor from moving
        // when servo mode is enabled.  Torque is automatically enabled by setting
        // position/speed.
        DisableTorque(lServos[i]);
		EnableServoMode(lServos[i]);
		SetGoalPositionWithSpeed(lServos[i], lInitialPosition[i], 100);
        //if (i == 0) SetGoalPositionWithSpeed(lServos[i], 0, 100);
        //if (i == 1) SetGoalPositionWithSpeed(lServos[i], 614, 100);
        //if (i == 2) SetGoalPositionWithSpeed(lServos[i], 614, 100);
        //if (i == 3) SetGoalPositionWithSpeed(lServos[i], 0, 100);
	}

	for (i = 0; i < NUM_SERVOS; i++) {
		while (IsMoving(lServos[i]));
		TxDString("Motor "); TxD32Dec(lServos[i]); TxDString(" position: "); TxD32Dec(GetPosition(lServos[i])); TxDString("\r\n");
	}
	
	// Put motors in continous mode and set their speed.
	TxDString("Configuring Motors for continuous mode.\r\n");
	for (i = 0; i < NUM_SERVOS; i++) {
		DisableServoMode(lServos[i]);
	}

    // Variables for calculating position in the blind region of AX12 servos.
    int lServoReading[NUM_SERVOS];
    int lEstimatedReading[NUM_SERVOS];                           
    double lEstimatedVel[NUM_SERVOS];
    double lDesiredVel[NUM_SERVOS];                              
	uint8_t lBlindRegion[NUM_SERVOS];
	uint8_t lPrevBlindRegion[NUM_SERVOS];
    for (i = 0; i < NUM_SERVOS; i++) {
        lServoReading[i] = lInitialPosition[i];
        lEstimatedReading[i] = lInitialPosition[i];
        lEstimatedVel[i] = 0.0;
        lDesiredVel[i] = 0.0;												 
		lBlindRegion[i] = 0;
		lPrevBlindRegion[i] = 0;
    }
    // Exponential smoothing factor for velocity.							 
    double lAlpha = 0.5;											 

    refreshNetwork(); // prepare the network for fresh activations.
    double lStepSize = 0.01;   // match step size with that in simulation.
    double lStepStartTime = time + lStepSize;
	lMaxTime += lStepStartTime;
    while (time < lMaxTime) {
        DelayMilliSec((lStepStartTime-time)*1000);
        lStepStartTime = time + lStepSize;

        // Read all servos at about the same time.
		for (i = 0; i < NUM_SERVOS; i++) {
			lServoReading[i] = GetPosition(lServos[i]);
        }

        //TxDString("Trace "); TxD32Dec(time*1000); TxDString(" ");
		for (i = 0; i < NUM_SERVOS; i++) {
            // In order to deal with the blind region, we calculate the input to the	 
            // neural network from the estimated reading instead of the direct servo	 
            // reading.	 The estimated reading is calculated from the estimated velocity,
            // which is adjusted based on the servo reading outside the blind region.	 
            lEstimatedReading[i] = lEstimatedReading[i] + (int)lEstimatedVel[i];		 
            // Make sure the estimated reading is within range.							 
            lEstimatedReading[i] = (lEstimatedReading[i]+1228)%1228;					 
																									 
            // Update estimated velocity if servo is not in the blind region.			 
            // The AX12 servos give position feedback only for angles in range [0, 300],
            // i.e. for lNormAnlge < 300/360.  Moreover, they give integers in [0, 1023]
            // for this range.  If the angle is not in this range (blind region), the
			// servo reading gives 1023 or 0 except for some time in between.  We will
			// use this observation to detect being in the blind region.
            if (lBlindRegion[i] == 0 && lServoReading[i] < 1023 && lServoReading[i] > 0) {
                // Since we assume that the motor is turning at a slow enough rate to	 
                // allow taking more than two readings in a rotation, if the above		 
                // difference is greater than 1228/2, then the reading has wrapped around
                // the lower or upper bound, and we need to add/subtract 1228 to get	 
                // the actual difference.												 
                int lReadingDiff = lServoReading[i] - lEstimatedReading[i];
                if (ABSOLUTE(lReadingDiff) > 614) {
                    lReadingDiff = (lReadingDiff < 0) ? (lReadingDiff+1228) : (lReadingDiff-1228);
                }
                lEstimatedVel[i] = (1-lAlpha)*lEstimatedVel[i] + lAlpha*((double)(lReadingDiff));
            }
            else {
                //lPrevBlindRegion[i] = lBlindRegion[i];

				// Detect entering/leaving the blind region
				if (lEstimatedVel[i] > 0) {
					if (lServoReading[i] == 1023) lBlindRegion[i] = 1;  // entering blind region
					else if (lServoReading[i] == 0) lBlindRegion[i] = 0;  // leaving blind region
				}
				if (lEstimatedVel[i] < 0) {
					if (lServoReading[i] == 0) lBlindRegion[i] = 1;  // entering blind region
					else if (lServoReading[i] == 1023) lBlindRegion[i] = 0;  // leaving blind region
				}

                //if (lPrevBlindRegion[i] == 0 && lBlindRegion[i] == 1) {
                //    TxDString("Motor "); TxD32Dec(lServos[i]); TxDString(": "); TxD32Dec(time*1000); TxDString("\r\n");
				//}
             }


            // Calculate normalized angle from estimated reading; range [0, 1)
            double lNormAngle = ((double)lEstimatedReading[i])/1228.0;						 

            // The leg is mounted 45 degrees with respect to the zero of the servo.
            lNormAngle += 0.125;

            // Since the zeros of the legs are initially 30 degrees from the vertical,
            // there is a total 60 degrees offset between the left and right sides.
            if (i%2 == 1) lNormAngle += 0.16666667;

            lNormAngle = lNormAngle - floor(lNormAngle);  // range [0, 1]

            // Inputs lie in the range  [-0.5, 0.5]
            lNormAngle -= 0.5;
            if (i%2 == 1) lNormAngle *= -1.0;  // flip sign of angle for right side legs.
            if (gInpCosSin) {
                double lAngle = lNormAngle*2*M_PI;
                lInputs[i] = cos(lAngle)/2;
                lInputs[i+NUM_SERVOS] = sin(lAngle)/2;
            }
            else {
                lInputs[i] = lNormAngle;
            }

			//TxD32Dec(lNormAngle*1000); TxDString(" ");
        }

        // Activate the network.
        activateNetwork(lInputs);

        // Get the network outputs and set the motors.
        getNetworkOutputs(lOutputs);

		for (i = 0; i < NUM_SERVOS; i++) {
            // The AX12 servos have discrete velocity values in the range [0, 1023] in each
            // direction, and the direction is determined by the most significant bit.
            lDesiredVel[i] = (1-lAlpha)*lDesiredVel[i] + lAlpha*(lOutputs[i] - 0.5);
            // Shift max velocity to middle of sigmoid for robust control.
            double lSmoothedVel = 0.5 - ABSOLUTE(lDesiredVel[i]);  // range [0, 0.5]
            //double lSmoothedVel = 0.5;
            if (i%2 == 1) lSmoothedVel *= -1.0;  // flip sign of velocity for right side legs.
            lSmoothedVel *= 2.0*lVelScale[i];  // range [-1, 0] or [0, 1]
			uint16_t lDiscreteVel = (lSmoothedVel < 0.0) ? (-lSmoothedVel*gAX12MaxVel+1024) : (lSmoothedVel*gAX12MaxVel);
            SetGoalSpeed(lServos[i], lDiscreteVel);
            //if (lBlindRegion[i] == 1) {
            //    if (lPrevBlindRegion[i] == 0) DisableServoMode(lServos[i]);  // disable servo mode the first time.
            //    SetGoalSpeed(lServos[i], lDiscreteVel);
            //}
            //else {
            //    if (lPrevBlindRegion[i] == 1) EnableServoMode(lServos[i]);  // enable servo mode the first time.
            //    uint16_t lPosition = (lSmoothedVel < 0.0) ? 0 : 1023;
            //    SetGoalPositionWithSpeed(lServos[i], lPosition, lDiscreteVel);
            //}
            //TxD32Dec(lDiscreteVel); TxDString(" ");
        }
		//TxDString("\r\n");
    }
	
	for (i = 0; i < NUM_SERVOS; i++) {
		SetGoalSpeed(lServos[i], 1);
		DisableTorque(lServos[i]);
	}
	
	TxDString("Motion Complete.\r\n");
	
	while (1); // wait for reset button to be pressed.
}
