
/*!
 *  \file   netutils.c
 *  \author Vinod Valsalam
 */

#include <float.h>
#include <math.h>
#include "netutils.h"


void activateOnce(const double* inActivations)
{
    uint8_t lNodeId;

    // Toggle activation index.
    gNetwork.activationIndex ^= 1;

	// Set the activations of the input nodes of the network.  Note that these activations
	// affect the activations of the other nodes only in the next call to this function
	// when the activation index is incremented.
	for (lNodeId = 0; lNodeId < gNetwork.numInputs; ++lNodeId) {
		setActivation(lNodeId, inActivations[lNodeId]);
	}

	// Activate all non-input nodes of the network.
	for (lNodeId = gNetwork.numInputs; lNodeId < gNetwork.numNodes; ++lNodeId) {
		Node* lNode = gNetwork.nodes[lNodeId];
        double lActivationSum = 0.0;
        uint8_t lActivateNode = 0;
        uint8_t lEdgeIndex;
        for (lEdgeIndex = 0; lEdgeIndex < lNode->numEdges; ++lEdgeIndex) {
            const double lSourceActivation = getPreviousActivation(lNode->edges[lEdgeIndex].source);
            if (lSourceActivation < DBL_MAX) {
                lActivationSum += lSourceActivation*lNode->edges[lEdgeIndex].weight;
                lActivateNode = 1;
            }
        }
        // The node is activated only if at least one of its sources was previously activated.
        if (lActivateNode) {
            double lBias = gNetwork.nodes[lNodeId]->bias;
            double lSlope = gNetwork.nodes[lNodeId]->slope;
            double lActivation = 1/(1+exp(-lSlope*(lActivationSum+lBias)));  // sigmoid
            setActivation(lNodeId, lActivation);
        }
	}
}


void activateNetwork(const double* inActivations)
{
	if (getActivation(0) < DBL_MAX) {
		// If this is not the first time activating the network, activate it only once.
		activateOnce(inActivations);
	} else {
		// In order to make sure all nodes are activated, we use the number of nodes
		// in the network as an estimate of the number of times to activate.
        uint8_t i;
		for (i = 0; i < gNetwork.numNodes; ++i) {
			activateOnce(inActivations);
		}
	}
}


void refreshNetwork(void)
{
    uint8_t lNodeId;
	for (lNodeId = 0; lNodeId < gNetwork.numNodes; ++lNodeId) {
        gNetwork.nodes[lNodeId]->activations[0] = DBL_MAX;
        gNetwork.nodes[lNodeId]->activations[1] = DBL_MAX;
	}
    gNetwork.activationIndex = 0;
}


void getNetworkOutputs(double* outActivations)
{
    uint8_t i;
	for (i = 0; i < gNetwork.numOutputs; ++i) {
        outActivations[i] = getActivation(gNetwork.numInputs+i);
    }
}
