
/*!
 *  \file   RobotEvalOp.cpp
 *  \brief  Implementation of the class RobotEvalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"
#include "RobotEvalOp.hpp"

#include <cmath>

using namespace Beagle;


opal::Simulator* NN::RobotEvalOp::mSimulator = 0;


/*!
 *  \brief Construct a new robot evaluation operator.
 *  \param inName Name of the operator.
 */
NN::RobotEvalOp::RobotEvalOp(const std::string& inName) :
		NN::EvaluationOp(inName),
        mAgent(0),
        mEnviron(0),
        mTime(0.0)
{}


NN::RobotEvalOp::~RobotEvalOp()
{
	Beagle_StackTraceBeginM();

    // The simulator must not be destroyed because it is a singleton.
//     if (mSimulator) {
//         mSimulator->destroy();
//     }

    removeEnviron();

	Beagle_StackTraceEndM("NN::RobotEvalOp::~RobotEvalOp()");
}


/*!
 *  \brief Register the parameters of the evaluation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::RobotEvalOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	NN::EvaluationOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "The X coordinate of starting position of the robot.",
		    "Float",
		    "0.0",
		    "The X coordinate of starting position of the robot from where evaluation starts."
		);
		mStartingX = castHandleT<Float>(
		               ioSystem.getRegister().insertEntry("nn.robot.startingx", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "The Z coordinate of starting position of the robot.",
		    "Float",
		    "0.0",
		    "The Z coordinate of starting position of the robot from where evaluation starts."
		);
		mStartingZ = castHandleT<Float>(
		               ioSystem.getRegister().insertEntry("nn.robot.startingz", new Float(0.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum time to run the simulation (seconds)",
		    "Float",
		    "60.0",
		    "The maximum real time to simulate the robot (seconds)."
		);
		mMaxTime = castHandleT<Float>(
		               ioSystem.getRegister().insertEntry("nn.robot.maxtime", new Float(60.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of simulation steps at which to apply control signals",
		    "UInt",
		    "1",
		    "Interval in terms of the number of simulation steps at which to apply control signals."
		);
		mControlInterval = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.robot.ctrlint", new UInt(1), lDescription));
	}
    {
        Register::Description lDescription(
            "Opal agent blueprint filename",
            "String",
            "",
            "Name of the opal blueprint file containing the agent."
        );
        mAgentFile = castHandleT<String>(
                             ioSystem.getRegister().insertEntry("nn.robot.blueprint", new String(""), lDescription));
    }
    {
        Register::Description lDescription(
            "Opal environment blueprint filename",
            "String",
            "../blueprints/ground-flat.xml",
            "Name of the opal blueprint file containing the environment."
        );
        mEnvironFile = castHandleT<String>(
                             ioSystem.getRegister().insertEntry("nn.environ.blueprint", new String("../blueprints/ground-flat.xml"), lDescription));
    }
	{
		Register::Description lDescription(
		    "Whether to create the environment programatically",
		    "Bool",
		    "0",
		    "The environment is created programatically if this parameter is set."
		);
		mEnvironProg = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.environ.program", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of modules in the network",
		    "UInt",
		    "4",
		    "Number of modules in the network."
		);
		mNumModules = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.number", new UInt(4), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of input nodes per module",
		    "UInt",
		    "2",
		    "Number of input nodes per module."
		);
		mNumModInputs = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.module.numin", new UInt(2), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of output nodes per module",
		    "UInt",
		    "1",
		    "Number of output nodes per module."
		);
		mNumModOutputs = castHandleT<UInt>(
		                  ioSystem.getRegister().insertEntry("nn.module.numout", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of rows into which the module inputs are arranged",
		    "UInt",
		    "1",
		    "Number of rows into which the module inputs are arranged."
		);
		mNumInpRows = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.input.numrow", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether alternate module input rows/columns are arranged backwards",
		    "Bool",
		    "0",
		    "Alternate module input rows/columns are arranged backwards if this parameter is set."
		);
		mInpAltBack = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.input.altback", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether module inputs are arranged column-wise",
		    "Bool",
		    "0",
		    "Module inputs are arranged column-wise if this parameter is set, otherwise they are arranged row-wise."
		);
		mInpColwise = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.input.colwise", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Number of rows into which the module outputs are arranged",
		    "UInt",
		    "1",
		    "Number of rows into which the module outputs are arranged."
		);
		mNumOutRows = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.output.numrow", new UInt(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether alternate module output rows/columns are arranged backwards",
		    "Bool",
		    "0",
		    "Alternate module output rows/columns are arranged backwards if this parameter is set."
		);
		mOutAltBack = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.output.altback", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether module outputs are arranged column-wise",
		    "Bool",
		    "0",
		    "Module outputs are arranged column-wise if this parameter is set, otherwise they are arranged row-wise."
		);
		mOutColwise = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.output.colwise", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "The network output multiplier (maximum angular velocity in radians per second).",
		    "Float",
		    "10.0",
		    "The network outputs are multiplied by this number(s) before applying to the robot simulation."
		);
		mOutMult = castHandleT<FloatArray>(
                          ioSystem.getRegister().insertEntry("nn.output.mult", new FloatArray(1, 10.0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Integer to select the fitness measure to be used",
		    "UInt",
		    "0",
		    "Integer to select the fitness measure to be used"
		);
		mFitnessSel = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.robot.fitsel", new UInt(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether cos and sin should be used instead of joint angle as inputs",
		    "Bool",
		    "0",
		    "The cos and sin of the joint angles will be used as input instead of the angles if this parameter is set."
		);
		mInpCosSin = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.input.cossin", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether simulating the AX12 Dynamixel servos using CM-2 board",
		    "Bool",
		    "0",
		    "True if simulating the AX12 Dynamixel servos using CM-2 board."
		);
		mAX12CM2 = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.robot.ax12cm2", new Bool(0), lDescription));
	}
	{
		Register::Description lDescription(
		    "Maximum velocity setting on the AX12; range [1, 1023]",
		    "UInt",
		    "1023",
		    "Maximum velocity setting on the AX12; range [1, 1023]"
		);
		mAX12MaxVel = castHandleT<UInt>(
		                 ioSystem.getRegister().insertEntry("nn.robot.ax12maxvel", new UInt(1023), lDescription));
	}
    {
        Register::Description lDescription(
            "Max generation term criterion",
            "UInt",
            "50",
            "Maximum number of generations for the evolution."
        );
        mMaxGeneration = castHandleT<UInt>(
            ioSystem.getRegister().insertEntry("ec.term.maxgen", new UInt(50), lDescription));
    }
	Beagle_StackTraceEndM("void NN::RobotEvalOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Evaluate the graph fitness for the robot walking problem.
 *  \param ioGraph Graph to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness measure.
 */
NN::FitnessIndexed::Handle NN::RobotEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)
{
	Beagle_StackTraceBeginM();

	float lStartingX = mStartingX->getWrappedValue();
	float lStartingZ = mStartingZ->getWrappedValue();

	float lMaxTime = mMaxTime->getWrappedValue();
    unsigned int lControlInterval = mControlInterval->getWrappedValue();

	GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
	NN::FitnessIndexed::Handle lFitness = new FitnessIndexed(1, 0.0);

	Graph::Activations lInputs;
	for (unsigned int i = lGraphState->getFirstInput(); i <= lGraphState->getLastInput(); i++) {
		lInputs.insert(Graph::Activations::value_type(i, 0.0));
	}
	unsigned int lFirstIn = lGraphState->getFirstInput();

	Graph::Activations lOutputs;
	for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++) {
		lOutputs.insert(Graph::Activations::value_type(i, 0.0));
	}
	unsigned int lFirstOut = lGraphState->getFirstOutput();

    unsigned int lNumModules = mNumModules->getWrappedValue();
    unsigned int lNumModInputs = mNumModInputs->getWrappedValue();
    unsigned int lNumModOutputs = mNumModOutputs->getWrappedValue();

    unsigned int lNumInpRows = mNumInpRows->getWrappedValue();
    unsigned int lNumInpCols = lNumModules/lNumInpRows;
    bool lInpAltBack = mInpAltBack->getWrappedValue();
    bool lInpColwise = mInpColwise->getWrappedValue();

    unsigned int lNumOutRows = mNumOutRows->getWrappedValue();
    unsigned int lNumOutCols = lNumModules/lNumOutRows;
    bool lOutAltBack = mOutAltBack->getWrappedValue();
    bool lOutColwise = mOutColwise->getWrappedValue();

    unsigned int lFitnessSel = mFitnessSel->getWrappedValue();
    bool lInpCosSin = mInpCosSin->getWrappedValue();
    bool lAX12CM2 = mAX12CM2->getWrappedValue();

    unsigned int lNumTrials = mEnvironProg->getWrappedValue() ? 5 : 1;
    for (unsigned int lTrial = 0; lTrial < lNumTrials; lTrial++) {
        if (mEnvironProg->getWrappedValue()) createEnviron(ioContext);
        createAgent(lStartingX, 0.02, lStartingZ, 1);
        opal::real lTempFitness = 0.0;

        // Define a volume solid to use with volume sensors.
        // opal::Solid* lVolume = mSimulator->createSolid();
        // lVolume->setEnabled(false);
        // opal::SphereShapeData lSphereData;
        // lSphereData.radius = 0.1;
        // lVolume->addShape(lSphereData);

        // Variables for calculating position in the blind region of AX12 servos.
        double lEstimatedVel[mAgent->getNumJoints()];
        int lEstimatedReading[mAgent->getNumJoints()];
        double lDesiredVel[mAgent->getNumJoints()];
        for (unsigned int i = 0; i < mAgent->getNumJoints(); i++) {
            lEstimatedVel[i] = 0.0;
            lEstimatedReading[i] = 614;
            lDesiredVel[i] = 0.0;
        }
        // Exponential smoothing factor for velocity.
        double lAlpha = 0.5;

        unsigned int lNumJoints = mAgent->getNumJoints();
        if (lAX12CM2 && mOutMult->size() < 2*lNumJoints) {
            // Expand the mOutMult array to also store the actual multipliers applied
            // to the network outputs; the original multiplies are moved to
            // the second half of the array.
            if (mOutMult->size() == 1) { // multipliers are the same for all outputs
                mOutMult->resize(2*lNumJoints, (*mOutMult)[0]);
            }
            else if (mOutMult->size() == lNumJoints) {  // a separate multiplier for each output
                mOutMult->resize(2*lNumJoints);
                for (unsigned int i = 0; i < lNumJoints; i++) (*mOutMult)[lNumJoints+i] = (*mOutMult)[i];
            }
            else {
                throw Beagle_RunTimeExceptionM("incompatible size for output multiplier array");
            }
        }

        unsigned int lVelSign[lGraphState->getNumberOutputs()];
        for (unsigned int i = 0; i < lGraphState->getNumberOutputs(); i++) {
            lVelSign[i] = 0;
        }

        //printf("\n"); 
        mTime = 0.0;
        opal::real dt = mSimulator->getStepSize()*lControlInterval;
        unsigned int lSteps = 0;
        while (mTime < lMaxTime) {
            //printf("%12.5f", mTime); 
            // Setup the network inputs from joint angles and sensors.
            unsigned int lNumAxes = mAgent->getJoint(0)->getNumAxes();
            unsigned int lNumAccInputs = lNumAxes;
            for (unsigned int i1 = 0; i1 < lNumJoints; i1++) {
                unsigned int lRowi, lColi;
                if (lInpColwise) {
                    lRowi = i1%lNumInpRows;
                    lColi = i1/lNumInpRows;
                    if (lInpAltBack && lColi%2 == 1) lRowi = (lNumInpRows - 1) - lRowi;
                }
                else {
                    lRowi = i1/lNumInpCols;
                    lColi = i1%lNumInpCols;
                    if (lInpAltBack && lRowi%2 == 1) lColi = (lNumInpCols - 1) - lColi;
                }
                unsigned int i2 = lRowi*lNumInpCols + lColi;  // Legs are arranged row-wise in the blueprint.
                for (unsigned int j = 0; j < lNumAccInputs; ++j) {
                    unsigned int k = j*lNumJoints + i1;
                    unsigned int l = i2*lNumAccInputs + j;
                    opal::Joint* lJoint = mAgent->getJoint(l/lNumAxes);
                    unsigned int m = l%lNumAxes;
                    opal::real lNormAngle = (lJoint->getAngle(m) - lJoint->getLowLimit(m))/(lJoint->getHighLimit(m) - lJoint->getLowLimit(m));
                    if (lAX12CM2) {
                        // The AX12 servos give position feedback only for angles in range [0, 300],
                        // i.e. for lNormAnlge < 300/360.  Moreover, they give integers in [0, 1023]
                        // for this range.  The integer range for the full 360 degrees is 1227.6.
                        // So reduce the resolution to the range of integers [0, 1228].
                        int lServoReading = (int)(lNormAngle*1227.6);

                        // In order to deal with the blind region, we calculate the input to the
                        // neural network from the estimated reading instead of the direct servo
                        // reading.  The estimated reading is calculated from the estimated velocity,
                        // which is adjusted based on the servo reading outside the blind region.
                        lEstimatedReading[i1] = lEstimatedReading[i1] + (int)lEstimatedVel[i1];
                        // Make sure the estimated reading is within range.
                        lEstimatedReading[i1] = (lEstimatedReading[i1]+1228)%1228;

                        // Update estimated velocity if servo is not in the blind region.
                        if (lServoReading < 1023) {
                            // Since we assume that the motor is turning at a slow enough rate to
                            // allow taking more than two readings in a rotation, if the above
                            // difference is greater than 1228/2, then the reading has wrapped around
                            // the lower or upper bound, and we need to add/subtract 1228 to get
                            // the actual difference.
                            int lReadingDiff = lServoReading - lEstimatedReading[i1];
                            unsigned int lAbsReadingDiff = absolute(lReadingDiff);
                            if (lAbsReadingDiff > 614) {
                                lReadingDiff = (lReadingDiff < 0) ? (lReadingDiff+1228) : (lReadingDiff-1228);
                            }
                            lEstimatedVel[i1] = (1-lAlpha)*lEstimatedVel[i1] + lAlpha*((double)(lReadingDiff));
                        }

                        // Calculate normalized angle from estimated reading; range [0, 1)
                        lNormAngle = ((double)lEstimatedReading[i1])/1228.0;

                        // Since the zeros of the legs are initially 30 degrees from the vertical,
                        // there is a total 60 degrees offset between the left and right sides.
                        if (i1%2 == 1) {
                            lNormAngle += 0.16666667;
                            lNormAngle = lNormAngle - std::floor(lNormAngle);  // range [0, 1]
                        }
                    }

                    if (lVelSign[k] == 0 and lNormAngle > 0.5) lVelSign[k] = 1;
                    else if (lVelSign[k] == 1 and lNormAngle > 0.95) lVelSign[k] = 2;
                    else if (lVelSign[k] == 2 and lNormAngle < 0.5) lVelSign[k] = 3;
                    else if (lVelSign[k] == 3 and lNormAngle < 0.05) lVelSign[k] = 0;

                    // Inputs lie in the range  [-0.5, 0.5]
                    lNormAngle -= 0.5;
                    if (lAX12CM2 && i1%2 == 1) lNormAngle *= -1.0;  // flip sign of angle for right side legs to account for motor orientation in physical robot.
                    if (lInpCosSin) {
                        opal::real lAngle = lNormAngle*2*M_PI;
                        lInputs[lFirstIn+k] = cos(lAngle)/2;
                        k += lNumAccInputs*lNumJoints;
                        lInputs[lFirstIn+k] = sin(lAngle)/2;
                    }
                    else {
                        lInputs[lFirstIn+k] = lNormAngle;
                    }
                    //printf("%12.5f", lNormAngle*360); 
                }
            }
            //printf("\n"); 
            //fflush(stdout); 

            // if (lInpCosSin) lNumAccInputs *= 2;
            // if (lNumModInputs > lNumAccInputs) {
            //     for (unsigned int i = 0; i < lNumJoints; ++i) {
            //         unsigned int k = lNumAccInputs*lNumJoints + i;
            //         opal::VolumeSensor* lVolumeSensor = (opal::VolumeSensor*)mAgent->getSensor(i);
            //         opal::VolumeQueryResult lVolumeResult = lVolumeSensor->queryVolume(lVolume);
            //         lInputs[lFirstIn+k] = (lVolumeResult.getNumSolids() > 0) ? 0.5 : 0.0;
            //     }
            //     lNumAccInputs += 1;
            // }

            // Activate the network.
            activate(ioGraph, lInputs, ioContext);

            // Get the network outputs and set the motors.
            ioGraph.getActivations(lOutputs);

            unsigned int lNumAccOutputs = lNumAxes;
            for (unsigned int i1 = 0; i1 < lNumJoints; i1++) {
                unsigned int lRowi, lColi;
                if (lOutColwise) {
                    lRowi = i1%lNumOutRows;
                    lColi = i1/lNumOutRows;
                    if (lOutAltBack && lColi%2 == 1) lRowi = (lNumOutRows - 1) - lRowi;
                }
                else {
                    lRowi = i1/lNumOutCols;
                    lColi = i1%lNumOutCols;
                    if (lOutAltBack && lRowi%2 == 1) lColi = (lNumOutCols - 1) - lColi;
                }
                unsigned int i2 = lRowi*lNumOutCols + lColi;  // Legs are arranged row-wise in the blueprint.
                for (unsigned int j = 0; j < lNumAccOutputs; ++j) {
                    unsigned int k = j*lNumJoints + i1;
                    unsigned int l = i2*lNumAccOutputs + j;
                    opal::ServoMotor* lMotor = (opal::ServoMotor*)mAgent->getMotor(l);
                    opal::real lSmoothedVel;
                    if (lAX12CM2) {
                        lDesiredVel[i1] = (1-lAlpha)*lDesiredVel[i1] + lAlpha*(lOutputs[lFirstOut+k] - 0.5);
                        // Shift max velocity to middle of sigmoid for robust control.
                        lSmoothedVel = 0.5 - absolute(lDesiredVel[i1]);  // range [0, 0.5]
                        lSmoothedVel *= 2.0;  // range [0, 1]
                        if (i1%2 == 1) lSmoothedVel *= -1.0;  // flip sign of velocity for right side legs; range [-1, 0] or [0, 1]

                        // The AX12 servos have discrete velocity values in the range [0, 1023] in each
                        // direction, and the direction is determined by the most significant bit.
                        unsigned int lAX12MaxVel = mAX12MaxVel->getWrappedValue();
                        lSmoothedVel = (double)((int)(lSmoothedVel*lAX12MaxVel))/((double)lAX12MaxVel);  // make discrete
                        lSmoothedVel += ioContext.getSystem().getRandomizer().rollGaussian(0.0, 0.025);  // add noise
                        //double lCurStdDev = 0.2*(double)(ioContext.getGeneration())/(double)(mMaxGeneration->getWrappedValue());  // shift mean
                        //if (lSteps%1000 == 0) (*mOutMult)[i1] = ((*mOutMult)[lNumJoints+i1])*ioContext.getSystem().getRandomizer().rollGaussian(1.0, lCurStdDev);
                        //if (lSteps%1200 == 0) (*mOutMult)[i1] = ((*mOutMult)[lNumJoints+i1])*ioContext.getSystem().getRandomizer().rollGaussian(1.0, lCurStdDev);
                        // The mean shift array below was used in the one leg disabled experiment
                        // instead of Gaussian noise.  Its entries were hand-designed to provide
                        // the same difficult variation in mean in each evaluation to prevent some
                        // evaluations from getting lucky.
                        static const double lMeanShift[4][7] = {{.1, .15, .1, .05, -.05, -.1, .0},
                                                                {.05, -.05, -.1, -.15, -.1, -.05, .05},
                                                                {-.1, -.15, -.1, -.05, .0, .05, -.05},
                                                                {.0, .05, .1, .15, .1, .15, .1}};
                        if (lSteps%1000 == 0) {
                            double lGenScale = (double)(ioContext.getGeneration())/(double)(mMaxGeneration->getWrappedValue());
                            (*mOutMult)[i1] = ((*mOutMult)[lNumJoints+i1])*(1. + lMeanShift[i1][lSteps/1000]*lGenScale);
                        }
                    }
                    else {
                        // Check whether the joint can rotate full 360 degrees.  If it cannot, then use velocity sign
                        // to help make the joint use the full range of movement.
                        opal::Joint* lJoint = mAgent->getJoint(l/lNumAxes);
                        unsigned int m = l%lNumAxes;
                        bool lCannotRotate360 = (lJoint->getHighLimit(m) - lJoint->getLowLimit(m)) < 360;
                        if (lCannotRotate360) lSmoothedVel = ((lVelSign[k] < 2) ? 1 : -1)*absolute(lOutputs[lFirstOut+k] - 0.5);  // range [-0.5, 0] or [0, 0.5]
                        else lSmoothedVel = lOutputs[lFirstOut+k] - 0.5;  // range [-0.5, 0.5]
                        lSmoothedVel *= 2.0;  // range [-1, 0] or [0, 1]
                    }
                    if (mOutMult->size() == 1) lSmoothedVel = lSmoothedVel*((*mOutMult)[0]);
                    else if (mOutMult->size() >= lNumJoints) lSmoothedVel = lSmoothedVel*((*mOutMult)[i1]);
                    else throw Beagle_RunTimeExceptionM("incompatible size for output multiplier array");
                    lMotor->setDesiredVel(lSmoothedVel);  // desired velocity is in radians per second.
                }
            }

            // Simulate the next time step, and stop simulation if an exception occurs.
            try {
                simulateAgent(dt);
            }
            catch (...) {
                Beagle_LogDebugM(
                    ioContext.getSystem().getLogger(),
                    "evaluate", "Beagle::NN::RobotEvalOp",
                    "exception occurred during simulation"
                );
                break;
            }

            // Stop simulation if something bad happens to the robot like it is in a maze and gets thrown up in the air.
            if (lFitnessSel == 0 && absolute<opal::real>(mOrgPos[1]-mCurPos[1]) > 1.5) {
                break;
            }
            else if (lFitnessSel == 1 && absolute<opal::real>(mOrgPos[1]-mCurPos[1]) > 0.8) {
                break;
            }

            // Fitness based on the robot moving forward (local Z direction)
            opal::Vec3r lVelCur = mAgent->getSolid(0)->getLocalLinearVel();
            if (lFitnessSel == 3) {
                lTempFitness += lVelCur.z*dt;
            }
            // Fitness based on the robot moving sideways (local X direction)
            else if (lFitnessSel == 4) {
                lTempFitness += lVelCur.x*dt;
            }

            opal::real lZ = mCurPos[2];
            opal::real lXActual = mCurPos[0];
            updateStep(mTime, lZ, 0.0, lXActual, ioGraph, ioContext);
            //printf("%12.6f%12.6f%12.6f%12.6f%12.6f\n", mTime, lInputs[0]*60, lInputs[1]*60, lInputs[2]*60, lInputs[3]*60);

            lSteps++;
        }

        // mSimulator->destroySolid(lVolume);

        removeAgent();
        if (mEnvironProg->getWrappedValue()) removeEnviron();

        // Fitness based on the robot moving forward (local Z direction) or sideways (local X direction).
        if (lFitnessSel == 3 || lFitnessSel == 4) {
            (*lFitness)[0] += absolute<opal::real>(lTempFitness);
        }
        else {
            (*lFitness)[0] += calculateFitness();
        }
    }
    (*lFitness)[0] /= lNumTrials;

	return lFitness;

	Beagle_StackTraceEndM("NN::FitnessIndexed::Handle NN::RobotEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)");
}


/*!
 *  \brief Simulate the agent for the given amount of time.
 *  \param inDt The amount of time to simulate.
 */
inline void NN::RobotEvalOp::simulateAgent(opal::real inDt)
{
	Beagle_StackTraceBeginM();

    // Run the simulation.
	mSimulator->simulate(inDt);

    // Increment time.
    mTime += inDt;

    // Update agent position.
    mCurPos = mAgent->getSolid(0)->getPosition();

	Beagle_StackTraceEndM("void NN::RobotEvalOp::simulateAgent(opal::real inDt)");
}


/*!
 * \brief Calculate fitness based on the original and current positions of the agent.
 */
opal::real NN::RobotEvalOp::calculateFitness()
{
	Beagle_StackTraceBeginM();

    unsigned int lFitnessSel = mFitnessSel->getWrappedValue();
    opal::real lFitness = 0.0;

    // Fitness based on max distance moved in the X or Z directions.
    if (lFitnessSel == 0) {
        lFitness = maxOf<opal::real>(absolute<opal::real>(mOrgPos[0]-mCurPos[0]), absolute<opal::real>(mOrgPos[2]-mCurPos[2]));
    }
    // Fitness based on sum of distance moved in the X and Z directions (e.g. in a maze).
    else if (lFitnessSel == 1) {
        lFitness = absolute<opal::real>(mOrgPos[0]-mCurPos[0]) + absolute<opal::real>(mOrgPos[2]-mCurPos[2]);
    }
    // Fitness based on distance moved in the XZ plane.
    else if (lFitnessSel == 2) {
        opal::Point3r lOrgPos(mOrgPos[0], 0.0, mOrgPos[2]);
        opal::Point3r lCurPos(mCurPos[0], 0.0, mCurPos[2]);
        lFitness = opal::distance(lOrgPos, lCurPos);
    }
    // Fitness based on absolute distance moved in the Z direction.
    else if (lFitnessSel == 5) {
        lFitness = absolute<opal::real>(mCurPos[2]-mOrgPos[2]);
    }
    // Fitness based on distance moved in the positive Z direction.
    else if (lFitnessSel == 6) {
        lFitness =  mCurPos[2] - mOrgPos[2];
    }
    // Fitness based on absolute distance moved in the Z direction, less the absolute distance moved in the X direction.
    else if (lFitnessSel == 7) {
        lFitness =  absolute<opal::real>(mCurPos[2]-mOrgPos[2]) - absolute<opal::real>(mCurPos[0]-mOrgPos[0]);
    }

    return lFitness;

	Beagle_StackTraceEndM("opal::real NN::RobotEvalOp::calculateFitness()");
}


/*!
 * \brief Initialize the operator by sampling the function to approximate.
 * \param ioSystem System to use to sample.
 */
void NN::RobotEvalOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	if (isInitialized())
		return;

	NN::EvaluationOp::init(ioSystem);

    // Create an opal simulator if it has not been done already.
    if (mSimulator == 0) {
        mSimulator = opal::createSimulator();
        mSimulator->setGravity(opal::Vec3r(0, (opal::real)-9.81, 0));
        mSimulator->setStepSize(0.01);

        // By default, opal uses the QuickStep solver of ODE at medium accuracy.
        // QuickStep is also non-deterministic.  Setting the solver accuracy to
        // very high chooses the WorldStep solver of ODE.
        mSimulator->setSolverAccuracy(opal::SOLVER_ACCURACY_VERY_HIGH);

        // Create the environment from blueprint file.
        if (!mEnvironProg->getWrappedValue()) {
            opal::Blueprint lEnvironBlueprint;
            opal::loadFile(lEnvironBlueprint, mEnvironFile->getWrappedValue());
            mEnviron = new opal::BlueprintInstance();
            mSimulator->instantiateBlueprint(*mEnviron, lEnvironBlueprint, opal::Matrix44r(), 1.0);
        }
    }

    // Load the agent blueprint from file.
    opal::loadFile(mAgentBlueprint, mAgentFile->getWrappedValue());

	Beagle_StackTraceEndM("void NN::RobotEvalOp::init(System& ioSystem)");
}


/*! 
 *  \brief Create an agent in the environment at a specified location and scale.
 *  \param inX The x coordinate of the agent.
 *  \param inY The y coordinate of the agent.
 *  \param inZ The z coordinate of the agent.
 *  \param inScale Scale to apply to the blueprint dimensions.
 */
void NN::RobotEvalOp::createAgent(opal::real inX, opal::real inY, opal::real inZ, opal::real inScale)
{
	Beagle_StackTraceBeginM();

    mAgent = new opal::BlueprintInstance();
    opal::Matrix44r lTransform;
    lTransform.translate(inX, inY, inZ);
    mSimulator->instantiateBlueprint(*mAgent, mAgentBlueprint, lTransform, inScale);

    // Reset any "leftover" time from previous calls to simulate()
    mSimulator->resetTimeBuffer();

    mOrgPos = mAgent->getSolid(0)->getPosition();
    mCurPos = mOrgPos;

	Beagle_StackTraceEndM("void NN::RobotEvalOp::createAgent(opal::real inX, opal::real inY, opal::real inZ, opal::real inScale)");
}


/*! 
 *  \brief Remove the agent from the environment.  
 */
void NN::RobotEvalOp::removeAgent()
{
	Beagle_StackTraceBeginM();

    // Return immediately if there is no agent.
    if (mAgent == 0)
        return;

    // Remove all sensors, motors, joints and solids of the blueprint instance
    // from the simulator.
    for (unsigned int i=0; i<mAgent->getNumSensors(); ++i) {
        opal::Sensor* sensor = mAgent->getSensor(i);
        mSimulator->destroySensor(sensor);
    }

    for (unsigned int i=0; i<mAgent->getNumMotors(); ++i) {
        opal::Motor* motor = mAgent->getMotor(i);
        mSimulator->destroyMotor(motor);
    }

    for (unsigned int i=0; i<mAgent->getNumJoints(); ++i) {
        opal::Joint* joint = mAgent->getJoint(i);
        mSimulator->destroyJoint(joint);
    }

    for (unsigned int i=0; i<mAgent->getNumSolids(); ++i) {
        opal::Solid* solid = mAgent->getSolid(i);
        mSimulator->destroySolid(solid);
    }

    delete mAgent;
    mAgent = 0;

	Beagle_StackTraceEndM("void NN::RobotEvalOp::removeAgent()");
}


/*! 
 *  \brief Create the environment.
 */
void NN::RobotEvalOp::createEnviron(Context& ioContext)
{
	Beagle_StackTraceBeginM();

    opal::Blueprint lEnvironBlueprint;

    opal::SolidData lGroundSolid;
    lGroundSolid.name = "ground";
    lGroundSolid.isStatic = true;
    lGroundSolid.linearDamping = 0.0;
    lGroundSolid.angularDamping = 0.0;
    lGroundSolid.transform.translate(0.0, -50.0, 0.0);
    opal::BoxShapeData lGroundShape;
    lGroundShape.dimensions.set(1000.0, 100.0, 1000.0);
    lGroundShape.material.hardness = 0.8;
    lGroundShape.material.friction = 0.8;
    lGroundShape.material.bounciness = 0.15;
    lGroundShape.material.density = 200.0;
    lGroundSolid.addShape(lGroundShape);
    lEnvironBlueprint.addSolid(&lGroundSolid);

	Randomizer& lRandomizer = ioContext.getSystem().getRandomizer();

    unsigned int lNumBumps = 5;
    opal::real lBumpWidthMin = 0.1;
    opal::real lBumpWidthMax = 2.5;
    opal::real lBumpHeightMin = 0.1;
    opal::real lBumpHeightMax = 0.4;
    opal::real lBumpSpacing = 5.0;

    opal::real lBumpDistance = 10.0;
    for (unsigned int i = 0; i < lNumBumps; i++) {
        opal::real lBumpWidth = lRandomizer.rollUniform(lBumpWidthMin, lBumpWidthMax);
        opal::real lBumpHeight = lRandomizer.rollUniform(lBumpHeightMin, lBumpHeightMax);

        opal::SolidData lBump0Solid;
        lBump0Solid.name = "bump" + uint2str(i) + "0";
        lBump0Solid.isStatic = true;
        lBump0Solid.linearDamping = 0.0;
        lBump0Solid.angularDamping = 0.0;
        lBump0Solid.transform.translate(0.0, lBumpHeight/2, lBumpDistance);
        opal::BoxShapeData lBump0Shape;
        lBump0Shape.contactGroup = 2;
        lBump0Shape.dimensions.set(lBumpDistance*2-lBumpWidth, lBumpHeight, lBumpWidth);
        lBump0Shape.material.hardness = 0.8;
        lBump0Shape.material.friction = 0.8;
        lBump0Shape.material.bounciness = 0.15;
        lBump0Shape.material.density = 200.0;
        lBump0Solid.addShape(lBump0Shape);
        lEnvironBlueprint.addSolid(&lBump0Solid);

        opal::SolidData lBump1Solid;
        lBump1Solid.name = "bump" + uint2str(i) + "1";
        lBump1Solid.isStatic = true;
        lBump1Solid.linearDamping = 0.0;
        lBump1Solid.angularDamping = 0.0;
        lBump1Solid.transform.translate(0.0, lBumpHeight/2, -lBumpDistance);
        opal::BoxShapeData lBump1Shape;
        lBump1Shape.contactGroup = 2;
        lBump1Shape.dimensions.set(lBumpDistance*2-lBumpWidth, lBumpHeight, lBumpWidth);
        lBump1Shape.material.hardness = 0.8;
        lBump1Shape.material.friction = 0.8;
        lBump1Shape.material.bounciness = 0.15;
        lBump1Shape.material.density = 200.0;
        lBump1Solid.addShape(lBump1Shape);
        lEnvironBlueprint.addSolid(&lBump1Solid);

        opal::SolidData lBump2Solid;
        lBump2Solid.name = "bump" + uint2str(i) + "2";
        lBump2Solid.isStatic = true;
        lBump2Solid.linearDamping = 0.0;
        lBump2Solid.angularDamping = 0.0;
        lBump2Solid.transform.translate(lBumpDistance, lBumpHeight/2, 0.0);
        opal::BoxShapeData lBump2Shape;
        lBump2Shape.contactGroup = 2;
        lBump2Shape.dimensions.set(lBumpWidth, lBumpHeight, lBumpDistance*2+lBumpWidth);
        lBump2Shape.material.hardness = 0.8;
        lBump2Shape.material.friction = 0.8;
        lBump2Shape.material.bounciness = 0.15;
        lBump2Shape.material.density = 200.0;
        lBump2Solid.addShape(lBump2Shape);
        lEnvironBlueprint.addSolid(&lBump2Solid);

        opal::SolidData lBump3Solid;
        lBump3Solid.name = "bump" + uint2str(i) + "3";
        lBump3Solid.isStatic = true;
        lBump3Solid.linearDamping = 0.0;
        lBump3Solid.angularDamping = 0.0;
        lBump3Solid.transform.translate(-lBumpDistance, lBumpHeight/2, 0.0);
        opal::BoxShapeData lBump3Shape;
        lBump3Shape.contactGroup = 2;
        lBump3Shape.dimensions.set(lBumpWidth, lBumpHeight, lBumpDistance*2+lBumpWidth);
        lBump3Shape.material.hardness = 0.8;
        lBump3Shape.material.friction = 0.8;
        lBump3Shape.material.bounciness = 0.15;
        lBump3Shape.material.density = 200.0;
        lBump3Solid.addShape(lBump3Shape);
        lEnvironBlueprint.addSolid(&lBump3Solid);

        lBumpDistance += lBumpSpacing;
    }
 
    lEnvironBlueprint.finalize();
 
    mEnviron = new opal::BlueprintInstance();
    mSimulator->instantiateBlueprint(*mEnviron, lEnvironBlueprint);

	Beagle_StackTraceEndM("void NN::RobotEvalOp::createEnviron(Context& ioContext)");
}


/*! 
 *  \brief Remove the environment.  
 */
void NN::RobotEvalOp::removeEnviron()
{
	Beagle_StackTraceBeginM();

    // Return immediately if there is no environment.
    if (mEnviron == 0)
        return;

    // Remove all solids of the blueprint instance from the simulator.
    for (unsigned int i=0; i<mEnviron->getNumSolids(); ++i) {
        opal::Solid* solid = mEnviron->getSolid(i);
        mSimulator->destroySolid(solid);
    }

    delete mEnviron;
    mEnviron = 0;

	Beagle_StackTraceEndM("void NN::RobotEvalOp::removeEnviron()");
}
