
/*!
 *  \file   RobotEvalOp.hpp
 *  \brief  Definition of the type RobotEvalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#ifndef RobotEvalOp_hpp
#define RobotEvalOp_hpp

#include "beagle/NN.hpp"
#include <string>
#include <vector>
#include "opal.h"

namespace Beagle
{
namespace NN
{


/*!
 *  \class RobotEvalOp RobotEvalOp.hpp "RobotEvalOp.hpp"
 *  \brief Double pole balancing evaluation operator for neural networks (Markov version)
 *  \ingroup NNF
 */
class RobotEvalOp : public NN::EvaluationOp
{

public:

	//! RobotEvalOp allocator type.
	typedef AllocatorT<RobotEvalOp,NN::EvaluationOp::Alloc>
	Alloc;
	//! RobotEvalOp handle type.
	typedef PointerT<RobotEvalOp,NN::EvaluationOp::Handle>
	Handle;
	//! RobotEvalOp bag type.
	typedef ContainerT<RobotEvalOp,NN::EvaluationOp::Bag>
	Bag;

	explicit RobotEvalOp(const std::string& inName="NN-RobotEvalOp");
	virtual ~RobotEvalOp();

	virtual void registerParams(Beagle::System& ioSystem);
	virtual FitnessIndexed::Handle evaluate(NN::Graph& ioGraph, Context& ioContext);
	virtual void init(Beagle::System& ioSystem);

protected:

	Float::Handle   mStartingX;         //!< The X coordinate of robot starting position.
	Float::Handle   mStartingZ;         //!< The Z coordinate of robot starting position.
	Float::Handle   mMaxTime;           //!< The maximum time to run the simulation.
	UInt::Handle    mControlInterval;   //!< The number of simulation steps at which to apply control.
	String::Handle  mAgentFile;         //!< Name of file containing the agent blueprint.
	String::Handle  mEnvironFile;       //!< Name of file containing the environment blueprint.
	Bool::Handle    mEnvironProg;       //!< Whether environment is to be created programatically.
	UInt::Handle    mNumModules;        //!< Number of modules.
	UInt::Handle    mNumModInputs;      //!< Number of input nodes per module.
	UInt::Handle    mNumModOutputs;     //!< Number of output nodes per module.
	UInt::Handle    mNumInpRows;        //!< Number of rows into which module inputs are arranged.
	Bool::Handle    mInpAltBack;        //!< Whether alternate module input rows/cols are arranged backwards.
	Bool::Handle    mInpColwise;        //!< Whether module inputs are arranged column-wise.
	UInt::Handle    mNumOutRows;        //!< Number of rows into which module outputs are arranged.
	Bool::Handle    mOutAltBack;        //!< Whether alternate module output rows/cols are arranged backwards.
	Bool::Handle    mOutColwise;        //!< Whether module outputs are arranged column-wise.
	FloatArray::Handle   mOutMult;      //!< The network outputs are multiplied by this number.
	UInt::Handle    mFitnessSel;        //!< The fitness measure to be used.
	Bool::Handle    mInpCosSin;         //!< Whether cos and sin should be used instead of angle as input.
	Bool::Handle    mAX12CM2;           //!< Whether simulating the AX12 Dynamixel servos using CM-2 board.
	UInt::Handle    mAX12MaxVel;        //!< Maximum velocity setting on the AX12; range [1, 1023]
	UInt::Handle    mMaxGeneration;     //!< Maximum number of generation of the evolution.

    static opal::Simulator*     mSimulator;         //!< Pointer to the opal simulator (a singleton).
	opal::Blueprint             mAgentBlueprint;    //!< The agent blueprint.
    opal::BlueprintInstance*    mAgent;             //!< Blueprint instance containing the agent.
    opal::BlueprintInstance*    mEnviron;           //!< Blueprint instance containing the environment.
    opal::real                  mTime;              //!< The current simulation time.

    opal::Point3r   mOrgPos;            //!< The original position of the agent.
    opal::Point3r   mCurPos;            //!< The current position of the agent.

    virtual void createAgent(opal::real inX, opal::real inY, opal::real inZ, opal::real inScale);
    virtual void removeAgent();
    virtual void createEnviron(Context& ioContext);
    virtual void removeEnviron();
    virtual void simulateAgent(opal::real inDt);

	virtual void updateStep(opal::real inTime,
                            opal::real inZ, opal::real inXTarget, opal::real inXActual,
                            const NN::Graph& inGraph, Context& ioContext);

    opal::real calculateFitness();

};

}
}


/*!
 *  \brief Update any data plots.
 *  \param inTime Time at the current simulation step.
 *  \param inZ Z coordinate of the robot.
 *  \param inXTarget Target X coordinate of the robot.
 *  \param inXActual Actual X coordinate of the robot.
 *  \param inGraph Graph to evaluate.
 *  \param ioContext Evolutionary context.
 */
inline void Beagle::NN::RobotEvalOp::updateStep(opal::real inTime,
                                                opal::real inZ, opal::real inXTarget, opal::real inXActual,
                                                const Beagle::NN::Graph& inGraph, Beagle::Context& ioContext)
{
}


#endif // RobotEvalOp_hpp
