
/*!
 *  \file   main.cpp
 *  \brief  Main routine for the robot walking problem using ENSO.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/GA.hpp"
#include "beagle/NN.hpp"
#include "beagle/ENSO.hpp"
#include "RobotEvalOp.hpp"

#define _USE_MATH_DEFINES
#include <cmath>
#include <cstdlib>
#include <iostream>
#include <stdexcept>
#include <vector>
#include <numeric>

namespace Beagle
{
namespace ENSO
{


/*!
 *  \brief Class for initializing the phenotype evaluation operator with the
 *         neural networks robot evaluation operator.
 */
class RobotEvalOp : public ENSO::NeuralNetworkEvalOp
{

public:

	//! RobotEvalOp allocator type.
	typedef AllocatorT<RobotEvalOp,ENSO::NeuralNetworkEvalOp::Alloc>
	Alloc;
	//! RobotEvalOp handle type.
	typedef PointerT<RobotEvalOp,ENSO::NeuralNetworkEvalOp::Handle>
	Handle;
	//! RobotEvalOp bag type.
	typedef ContainerT<RobotEvalOp,ENSO::NeuralNetworkEvalOp::Bag>
	Bag;

	explicit RobotEvalOp(Beagle::EvaluationOp::Handle inPhenotypeEvalOp=new NN::RobotEvalOp,
                         const std::string& inName="ENSO-RobotEvalOp") :
            ENSO::NeuralNetworkEvalOp(inPhenotypeEvalOp, inName)
    { }

	virtual ~RobotEvalOp()
    { }

};

}
}

using namespace std;
using namespace Beagle;


int main(int argc, char *argv[])
{
	try {
		// Build a system with NN package
		System::Handle lSystem = new System;
        lSystem->addPackage(new GA::PackageBitString);
        lSystem->addPackage(new GA::PackageFloatVector);
        lSystem->addPackage(new GA::PackageIntegerVector);
        lSystem->addPackage(new NN::PackageBase);
        lSystem->addPackage(new ENSO::PackageBase);

        // Add optional history component
        lSystem->addComponent(new History);

        // Add the neural network parameter properties component
        lSystem->addComponent(new ENSO::NeuralNetworkParams);

        // Add evaluation operator allocator
        lSystem->setEvaluationOp("ENSO-RobotEvalOp", new ENSO::RobotEvalOp::Alloc);

        // Initialize the evolver
        Evolver::Handle lEvolver = new Evolver;
        lEvolver->initialize(lSystem, argc, argv);

        // Create population
        Vivarium::Handle lVivarium = new Vivarium;

        // Launch evolution
        lEvolver->evolve(lVivarium, lSystem);

	} catch(Exception& inException) {
		inException.terminate();
	} catch(exception& inException) {
		cerr << "Standard exception caught:" << endl;
		cerr << inException.what() << endl << flush;
		return 1;
	} catch(...) {
		cerr << "Unknown exception caught!" << endl << flush;
		return 1;
	}
	return 0;
}

