
/*!
 *  \file   GuiRobotEvalOp.cpp
 *  \brief  Implementation of the class GuiRobotEvalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "GuiRobotEvalOp.hpp"

#include <cmath>

using namespace Beagle;


/*!
 *  \brief Construct a new robot evaluation operator.
 *  \param inName Name of the operator.
 */
NN::GuiRobotEvalOp::GuiRobotEvalOp(const std::string& inName) :
		NN::VizRobotEvalOp(inName),
        mInteraction(NULL),
        mRenderTimer(new QTimer(this))
{ }


/*!
 *  \brief Override the evaluate method to allow interaction with the GUI.
 *  \param ioGraph Graph to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness measure.
 */
NN::FitnessIndexed::Handle NN::GuiRobotEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)
{
	Beagle_StackTraceBeginM();
  
    // Clear plot data
    Explorer::NN::PlotData::Handle lPlotData = castHandleT<Explorer::NN::PlotData>(ioContext.getSystem().getComponent("NN-PlotData"));
    if (lPlotData != NULL) {
        QVector<Explorer::NN::PlotData::Data>& lDataVector = lPlotData->getDataVector();

        lDataVector[0].mX.clear();
        GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
        unsigned int j = 0;
        for (unsigned int i = lGraphState->getFirstInput(); i <= lGraphState->getLastInput(); i++, j++) {
            lDataVector[0].mY[j].clear();
        }
        for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++, j++) {
            lDataVector[0].mY[j].clear();
        }

        lDataVector[1].mX.clear();
        lDataVector[1].mY[0].clear();
        lDataVector[1].mY[1].clear();
    }

    mInteraction->update(ioGraph, ioContext);
    return NN::VizRobotEvalOp::evaluate(ioGraph, ioContext);

	Beagle_StackTraceEndM("NN::FitnessIndexed::Handle NN::GuiRobotEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)");
}


/*!
 *  \brief Update any data plots.
 *  \param inTime Time at the current simulation step.
 *  \param inZ Z coordinate of the robot.
 *  \param inXTarget Target X coordinate of the robot.
 *  \param inXActual Actual X coordinate of the robot.
 *  \param inGraph Graph to evaluate.
 *  \param ioContext Evolutionary context.
 */
void NN::GuiRobotEvalOp::updateStep(opal::real inTime,
                                    opal::real inZ, opal::real inXTarget, opal::real inXActual,
                                    const NN::Graph& inGraph, Context& ioContext)
{
	Beagle_StackTraceBeginM();

    // Update plot data
    Explorer::NN::PlotData::Handle lPlotData = castHandleT<Explorer::NN::PlotData>(ioContext.getSystem().getComponent("NN-PlotData"));
    if (lPlotData != NULL) {
        QVector<Explorer::NN::PlotData::Data>& lDataVector = lPlotData->getDataVector();

        lDataVector[0].mX.push_back(inTime);
        GraphState::Handle lGraphState = castHandleT<GraphState>(ioContext.getSystem().getComponent("NN-GraphState"));
        unsigned int j = 0;
        for (unsigned int i = lGraphState->getFirstInput(); i <= lGraphState->getLastInput(); i++, j++) {
            lDataVector[0].mY[j].push_back(inGraph.getActivation(i));
        }
        for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++, j++) {
            lDataVector[0].mY[j].push_back(inGraph.getActivation(i));
        }

        lDataVector[1].mX.push_back(inZ);
        lDataVector[1].mY[0].push_back(inXTarget);
        lDataVector[1].mY[1].push_back(inXActual);
    }

    mInteraction->update(inGraph, ioContext);

	Beagle_StackTraceEndM("void NN::GuiRobotEvalOp::updateStep(opal::real inTime, opal::real inZ, opal::real inXTarget, opal::real inXActual, const NN::Graph& inGraph, Context& ioContext)");
}


/*!
 * \brief Initialize the operator by sampling the function to approximate.
 * \param ioSystem System to use to sample.
 */
void NN::GuiRobotEvalOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	if (isInitialized())
		return;

	NN::VizRobotEvalOp::init(ioSystem);

    mRenderTimer->setInterval(100);
    connect(mRenderTimer, SIGNAL(timeout()), this, SLOT(renderEventLoop()));

    // Create an NN interaction object to interact with the GUI.
    mInteraction = new Explorer::NN::Interaction;

    // Get the base singleton interaction data instance and make connections.
    Explorer::InteractionData& lInteractionData = Explorer::InteractionData::instance();
	connect(&lInteractionData, SIGNAL(enteringEventLoop()), this, SLOT(enterEventLoop()));
	connect(&lInteractionData, SIGNAL(exitingEventLoop()), this, SLOT(exitEventLoop()));

    Explorer::NN::PlotData::Handle lPlotData = castHandleT<Explorer::NN::PlotData>(ioSystem.getComponent("NN-PlotData"));
    QVector<Explorer::NN::PlotData::Data>& lDataVector = lPlotData->getDataVector();
    lDataVector.resize(2);

    // First plot contains node activations.
    lDataVector[0].mTitle = "Node Activations";

    lDataVector[0].mLabelAxisX = "Time (s)";
    lDataVector[0].mMinX = 0;
    lDataVector[0].mMaxX = 60;

    lDataVector[0].mLabelAxisY = "Node Activation";
    lDataVector[0].mMinY = -1.5;
    lDataVector[0].mMaxY = 1.5;

    GraphState::Handle lGraphState = castHandleT<GraphState>(ioSystem.getComponent("NN-GraphState"));
    unsigned int lNumCurves = lGraphState->getNumberInputs() + lGraphState->getNumberOutputs();
    //    unsigned int lNumCurves = lGraphState->getNumberOutputs();
    lDataVector[0].mLabelCurve.resize(lNumCurves);
    lDataVector[0].mY.resize(lNumCurves);

    unsigned int j = 0;
    for (unsigned int i = lGraphState->getFirstInput(); i <= lGraphState->getLastInput(); i++, j++) {
        lDataVector[0].mLabelCurve[j] = "Node " + QString::number(i);
    }
    for (unsigned int i = lGraphState->getFirstOutput(); i <= lGraphState->getLastOutput(); i++, j++) {
        lDataVector[0].mLabelCurve[j] = "Node " + QString::number(i);
    }

    // Second plot contains path of the robot.
    lDataVector[1].mTitle = "Robot Path";

    lDataVector[1].mLabelAxisX = "Z Direction";
    lDataVector[1].mMinX = -6;
    lDataVector[1].mMaxX = 6;

    lDataVector[1].mLabelAxisY = "X Direction";
    lDataVector[1].mMinY = -6;
    lDataVector[1].mMaxY = 6;

    lDataVector[1].mLabelCurve.resize(2);
    lDataVector[1].mY.resize(2);

    lDataVector[1].mLabelCurve[0] = "Target Path";
    lDataVector[1].mLabelCurve[1] = "Actual Path";

	Beagle_StackTraceEndM("void NN::GuiRobotEvalOp::init(System& ioSystem)");
}


/*! 
 * \brief Render when timer goes off during pauses or breaks.
 */
void NN::GuiRobotEvalOp::renderEventLoop()
{
	Beagle_StackTraceBeginM();

    mRobotApp->render(-1.0, FLT_MAX);
    mRenderTimer->start();

	Beagle_StackTraceEndM("void NN::GuiRobotEvalOp::renderEventLoop()");
}


/*! 
 * \brief Prepare for entering the Qt event loop in Explorer.
 */
void NN::GuiRobotEvalOp::enterEventLoop()
{
	Beagle_StackTraceBeginM();

    mRenderTimer->start();

	Beagle_StackTraceEndM("void NN::GuiRobotEvalOp::enterEventLoop()");
}


/*! 
 * \brief Prepare for exiting the Qt event loop in Explorer.
 */
void NN::GuiRobotEvalOp::exitEventLoop()
{
	Beagle_StackTraceBeginM();

    mTimer.reset();
    mCurTime = -1.0;

    mRenderTimer->stop();

	Beagle_StackTraceEndM("void NN::GuiRobotEvalOp::exitEventLoop()");
}
