
/*!
 *  \file   main.cpp
 *  \brief  Main routine for the robot walking problem using ENSO.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/GA.hpp"
#include "explorer/ENSO.hpp"
#include "GuiRobotEvalOp.hpp"

#define _USE_MATH_DEFINES
#include <cmath>
#include <cstdlib>
#include <iostream>
#include <stdexcept>
#include <vector>
#include <numeric>

namespace Beagle
{
namespace ENSO
{


/*!
 *  \brief Class for initializing the phenotype evaluation operator with the
 *         neural networks robot evaluation operator.
 */
class GuiRobotEvalOp : public ENSO::NeuralNetworkEvalOp
{

public:

	//! GuiRobotEvalOp allocator type.
	typedef AllocatorT<GuiRobotEvalOp,ENSO::NeuralNetworkEvalOp::Alloc>
	Alloc;
	//! GuiRobotEvalOp handle type.
	typedef PointerT<GuiRobotEvalOp,ENSO::NeuralNetworkEvalOp::Handle>
	Handle;
	//! GuiRobotEvalOp bag type.
	typedef ContainerT<GuiRobotEvalOp,ENSO::NeuralNetworkEvalOp::Bag>
	Bag;

	explicit GuiRobotEvalOp(Beagle::EvaluationOp::Handle inPhenotypeEvalOp=new NN::GuiRobotEvalOp,
                            const std::string& inName="ENSO-GuiRobotEvalOp") :
            ENSO::NeuralNetworkEvalOp(inPhenotypeEvalOp, inName)
    { }

	virtual ~GuiRobotEvalOp()
    { }

	virtual Fitness::Handle evaluate(ENSO::Individual& ioIndividual, Beagle::Context& ioContext);

};

}
}


/*!
 *  \brief Evaluate the fitness of the ENSO individual.
 *  \param ioIndividual Individual to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness measure.
 */
Beagle::Fitness::Handle Beagle::ENSO::GuiRobotEvalOp::evaluate(Beagle::ENSO::Individual& ioIndividual, Beagle::Context& ioContext)
{
	Beagle_StackTraceBeginM();

    Explorer::ENSO::NNInteraction::Handle lInteraction = new Explorer::ENSO::NNInteraction;
    lInteraction->update(ioIndividual, ioContext);

    return Beagle::ENSO::NeuralNetworkEvalOp::evaluate(ioIndividual, ioContext);

	Beagle_StackTraceEndM("Fitness::Handle ENSO::GuiRobotEvalOp::evaluate(ENSO::Individual& ioIndividual, Context& ioContext)");
}


int main(int argc, char *argv[])
{
	Explorer::Application lApp(argc, argv);
    Q_INIT_RESOURCE(explorer);

	try {
		// Build a system with NN package
		Beagle::System::Handle lSystem = new Beagle::System;
        lSystem->addPackage(new Beagle::GA::PackageBitString);
        lSystem->addPackage(new Beagle::GA::PackageFloatVector);
        lSystem->addPackage(new Beagle::GA::PackageIntegerVector);
        lSystem->addPackage(new Beagle::NN::PackageBase);
        lSystem->addPackage(new Beagle::ENSO::PackageBase);
        lSystem->addPackage(new Explorer::NN::PackageBase);
        lSystem->addPackage(new Explorer::ENSO::PackageBase);

        // Add optional history component
        lSystem->addComponent(new Beagle::History);

        // Add the neural network parameter properties component
        lSystem->addComponent(new Beagle::ENSO::NeuralNetworkParams);

        // Component for data plots
		lSystem->addComponent(new Explorer::NN::PlotData);

        // Add evaluation operator allocator
        lSystem->setEvaluationOp("ENSO-GuiRobotEvalOp", new Beagle::ENSO::GuiRobotEvalOp::Alloc);

        // Setup a Qt main window.
		Explorer::MainWindow* lMainWindow = new Explorer::MainWindow(lSystem);
		lMainWindow->show();

	} catch (Beagle::Exception& inException) {
		inException.terminate(std::cerr);
	} catch (std::exception& inException) {
		std::cerr << "Standard exception caught:" << std::endl << std::flush;
		std::cerr << inException.what() << std::endl << std::flush;
		abort();
	} catch (...) {
        std::cerr << "Unknown exception caught!" << std::endl << std::flush;
		abort();
	}

	return lApp.exec();
}
