
/*!
 *  \file   PrintRobotEvalOp.cpp
 *  \brief  Implementation of the class PrintRobotEvalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"
#include "PrintRobotEvalOp.hpp"

#include <cmath>

extern "C" {
#include "netutils.h"
}


using namespace Beagle;


/*!
 *  \brief Construct a new print robot evaluation operator.
 *  \param inName Name of the operator.
 */
NN::PrintRobotEvalOp::PrintRobotEvalOp(const std::string& inName) :
		NN::RobotEvalOp(inName)
{ }


/*!
 *  \brief Register the parameters of the evaluation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::PrintRobotEvalOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	NN::RobotEvalOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Whether to convert or evaluate neural network",
		    "Bool",
		    "1",
		    "If true, convert neural network to print version of the robot; otherwise evaluate already converted network."
		);
		mConvertNetwork = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.prnrobot.convert", new Bool(1), lDescription));
	}
	Beagle_StackTraceEndM("void NN::PrintRobotEvalOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Override method to convert neural network to print version of the robot or to evaluate
 *      already converted network.
 *  \param ioGraph Graph to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness measure.
 */
NN::FitnessIndexed::Handle NN::PrintRobotEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)
{
	Beagle_StackTraceBeginM();

    if (mConvertNetwork->getWrappedValue()) {
        convertNetwork(ioGraph, ioContext);
        return new FitnessIndexed(1, 0.0);
    }
  
	float lStartingX = mStartingX->getWrappedValue();
	float lStartingZ = mStartingZ->getWrappedValue();

	float lMaxTime = mMaxTime->getWrappedValue();
    unsigned int lControlInterval = mControlInterval->getWrappedValue();
	float lOutMult = (*mOutMult)[0];

	NN::FitnessIndexed::Handle lFitness = new FitnessIndexed(1, 0.0);

	double lInputs[gNetwork.numInputs];
	for (unsigned int i = 0; i < gNetwork.numInputs; i++) {
		lInputs[i] = 0.0;
	}

	double lOutputs[gNetwork.numOutputs];
	for (unsigned int i = 0; i < gNetwork.numOutputs; i++) {
		lOutputs[i] = 0.0;
	}

    unsigned int lNumModules = mNumModules->getWrappedValue();
    unsigned int lNumModInputs = mNumModInputs->getWrappedValue();
    unsigned int lNumModOutputs = mNumModOutputs->getWrappedValue();

    unsigned int lNumInpRows = mNumInpRows->getWrappedValue();
    unsigned int lNumInpCols = lNumModules/lNumInpRows;
    bool lInpAltBack = mInpAltBack->getWrappedValue();
    bool lInpColwise = mInpColwise->getWrappedValue();

    unsigned int lNumOutRows = mNumOutRows->getWrappedValue();
    unsigned int lNumOutCols = lNumModules/lNumOutRows;
    bool lOutAltBack = mOutAltBack->getWrappedValue();
    bool lOutColwise = mOutColwise->getWrappedValue();

    unsigned int lFitnessSel = mFitnessSel->getWrappedValue();
    bool lInpCosSin = mInpCosSin->getWrappedValue();

    unsigned int lNumTrials = mEnvironProg->getWrappedValue() ? 5 : 1;
    for (unsigned int lTrial = 0; lTrial < lNumTrials; lTrial++) {
        if (mEnvironProg->getWrappedValue()) createEnviron(ioContext);
        createAgent(lStartingX, 0.01, lStartingZ, 1);
        opal::real lTempFitness = 0.0;

        // Define a volume solid to use with volume sensors.
        // opal::Solid* lVolume = mSimulator->createSolid();
        // lVolume->setEnabled(false);
        // opal::SphereShapeData lSphereData;
        // lSphereData.radius = 0.1;
        // lVolume->addShape(lSphereData);

        mTime = 0.0;
        opal::real dt = mSimulator->getStepSize()*lControlInterval;
        unsigned int lSteps = 0;
        refreshNetwork(); // prepare the network for fresh activations.
        while (mTime < lMaxTime) {
            // Setup the network inputs from joint angles and sensors.
            unsigned int lNumJoints = mAgent->getNumJoints();
            unsigned int lNumAxes = mAgent->getJoint(0)->getNumAxes();
            unsigned int lNumAccInputs = lNumAxes;
            for (unsigned int i1 = 0; i1 < lNumJoints; i1++) {
                unsigned int lRowi, lColi;
                if (lInpColwise) {
                    lRowi = i1%lNumInpRows;
                    lColi = i1/lNumInpRows;
                    if (lInpAltBack && lColi%2 == 1) lRowi = (lNumInpRows - 1) - lRowi;
                }
                else {
                    lRowi = i1/lNumInpCols;
                    lColi = i1%lNumInpCols;
                    if (lInpAltBack && lRowi%2 == 1) lColi = (lNumInpCols - 1) - lColi;
                }
                unsigned int i2 = lRowi*lNumInpCols + lColi;  // Legs are arranged row-wise in the blueprint.
                for (unsigned int j = 0; j < lNumAccInputs; ++j) {
                    unsigned int k = j*lNumJoints + i1;
                    unsigned int l = i2*lNumAccInputs + j;
                    opal::Joint* lJoint = mAgent->getJoint(l/lNumAxes);
                    unsigned int m = l%lNumAxes;
                    // Inputs lie in the range  [-0.5, 0.5]
                    if (lInpCosSin) {
                        opal::real lAngle = lJoint->getAngle(m)*M_PI/180;
                        lInputs[k] = cos(lAngle)/2;
                        k += lNumAccInputs*lNumJoints;
                        lInputs[k] = sin(lAngle)/2;
                    }
                    else {
                        opal::real lNormAngle = (lJoint->getAngle(m) - lJoint->getLowLimit(m))/(lJoint->getHighLimit(m) - lJoint->getLowLimit(m));
                        lNormAngle -= 0.5;
                        lInputs[k] = lNormAngle;
                    }
                }
            }

            if (lInpCosSin) lNumAccInputs *= 2;
            // if (lNumModInputs > lNumAccInputs) {
            //     for (unsigned int i = 0; i < lNumJoints; ++i) {
            //         unsigned int k = lNumAccInputs*lNumJoints + i;
            //         opal::VolumeSensor* lVolumeSensor = (opal::VolumeSensor*)mAgent->getSensor(i);
            //         opal::VolumeQueryResult lVolumeResult = lVolumeSensor->queryVolume(lVolume);
            //         lInputs[k] = (lVolumeResult.getNumSolids() > 0) ? 0.5 : 0.0;
            //     }
            //     lNumAccInputs += 1;
            // }

            // Activate the network.
            activateNetwork(lInputs);

            // Get the network outputs and set the motors.
            getNetworkOutputs(lOutputs);

            unsigned int lNumAccOutputs = lNumAxes;
            for (unsigned int i1 = 0; i1 < lNumJoints; i1++) {
                unsigned int lRowi, lColi;
                if (lOutColwise) {
                    lRowi = i1%lNumOutRows;
                    lColi = i1/lNumOutRows;
                    if (lOutAltBack && lColi%2 == 1) lRowi = (lNumOutRows - 1) - lRowi;
                }
                else {
                    lRowi = i1/lNumOutCols;
                    lColi = i1%lNumOutCols;
                    if (lOutAltBack && lRowi%2 == 1) lColi = (lNumOutCols - 1) - lColi;
                }
                unsigned int i2 = lRowi*lNumOutCols + lColi;  // Legs are arranged row-wise in the blueprint.
                for (unsigned int j = 0; j < lNumAccOutputs; ++j) {
                    unsigned int k = j*lNumJoints + i1;
                    unsigned int l = i2*lNumAccOutputs + j;
                    opal::ServoMotor* lMotor = (opal::ServoMotor*)mAgent->getMotor(l);
                    opal::real lDesiredVel = (opal::real)(lOutputs[k] - 0.5);
                    lDesiredVel = lDesiredVel*lOutMult;
                    lMotor->setDesiredVel(lDesiredVel);
                }
            }

            // Simulate the next time step, and stop simulation if an exception occurs.
            try {
                simulateAgent(dt);
            }
            catch (...) {
                Beagle_LogDebugM(
                    ioContext.getSystem().getLogger(),
                    "evaluate", "Beagle::NN::RobotEvalOp",
                    "exception occurred during simulation"
                );
                break;
            }

            // Stop simulation if something bad happens to the robot like it is in a maze and gets thrown up in the air.
            if (lFitnessSel == 0 && absolute<opal::real>(mOrgPos[1]-mCurPos[1]) > 1.5) {
                break;
            }
            else if (lFitnessSel == 1 && absolute<opal::real>(mOrgPos[1]-mCurPos[1]) > 0.8) {
                break;
            }

            // Fitness based on the robot moving forward (local Z direction)
            opal::Vec3r lVelCur = mAgent->getSolid(0)->getLocalLinearVel();
            if (lFitnessSel == 3) {
                lTempFitness += lVelCur.z*dt;
            }
            // Fitness based on the robot moving sideways (local X direction)
            else if (lFitnessSel == 4) {
                lTempFitness += lVelCur.x*dt;
            }

            opal::real lZ = mCurPos[2];
            opal::real lXActual = mCurPos[0];
            updateStep(mTime, lZ, 0.0, lXActual, ioGraph, ioContext);

            lSteps++;
        }

        // mSimulator->destroySolid(lVolume);

        removeAgent();
        if (mEnvironProg->getWrappedValue()) removeEnviron();

        // Fitness based on the robot moving forward (local Z direction) or sideways (local X direction).
        if (lFitnessSel == 3 || lFitnessSel == 4) {
            (*lFitness)[0] += absolute<opal::real>(lTempFitness);
        }
        else {
            (*lFitness)[0] += calculateFitness();
        }
    }
    (*lFitness)[0] /= lNumTrials;

	return lFitness;

	Beagle_StackTraceEndM("NN::FitnessIndexed::Handle NN::PrintRobotEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)");
}


/*!
 *  \brief Convert neural network to that for print version of the robot.
 *  \param inGraph Graph to evaluate.
 *  \param inContext Evolutionary context.
 */
void NN::PrintRobotEvalOp::convertNetwork(const NN::Graph& inGraph, const Beagle::Context& inContext)
{
	Beagle_StackTraceBeginM();

    // Use dbl2str() for printing floating point numbers to get the same precision that is
    // used in the XML files.

	NN::GraphState::Handle lGraphState = Beagle::castHandleT<NN::GraphState>(inContext.getSystem().getComponent("NN-GraphState"));

	std::ofstream lOFS("network.c");

    lOFS << std::endl;
    lOFS << "/*!" << std::endl;
    lOFS << " *  \\file   network.c" << std::endl;
    lOFS << " *  \\author Vinod Valsalam (automatically generated)" << std::endl;
    lOFS << " */" << std::endl;
    lOFS << std::endl;
    lOFS << "#include \"network.h\"" << std::endl;
    lOFS << std::endl;

    // Iterate through all the nodes and construct them.
    unsigned int lNodeCount = 0;
	for (NN::Graph::const_iterator lGraphIter = inGraph.begin(); lGraphIter != inGraph.end(); ++lGraphIter) {
		const NN::Node& lNode = lGraphIter->second;
		unsigned int lNodeId = lNode.getIdentifier();

        if (lNodeId != lNodeCount) {
            // This could happen if NN::Graph is implemented as hash_map instead of a map.
			throw Beagle_RunTimeExceptionM("Network conversion failed: nodes not in sequence!");
        }

        lOFS << "Node n" << lNodeCount << " = {" << std::endl;
        lOFS << "    " << dbl2str(lNode.getBias()) << ", // bias" << std::endl;
		lOFS << "    " << dbl2str(lNode.getSlope()) << ", // slope" << std::endl;
		lOFS << "    {0.0, 0.0}, // previous and current activations" << std::endl;
		lOFS << "    " << lNode.getNumberEdges() << ", // number of edges" << std::endl;
        lOFS << "    { // array of edges" << std::endl;

        // Iterate through all the edges and construct them.
        unsigned int lEdgeCount = 0;
        for (NN::Node::const_iterator lNodeIter = lNode.begin(); lNodeIter != lNode.end(); ++lNodeIter) {
            const NN::Edge& lEdge = lNodeIter->second;

            lOFS << "        { // edge " << lEdgeCount << std::endl;
			lOFS << "          " << lEdge.getSourceIdentifier() << ", // source node" << std::endl;
			lOFS << "          " << dbl2str(lEdge.getWeight()) << " // weight" << std::endl;
			lOFS << "        }" << ((lEdgeCount < lNode.getNumberEdges()-1) ? "," : "") << " // edge " << lEdgeCount << std::endl;

            lEdgeCount++;
        }

        lOFS << "    } // array of edges" << std::endl;
        lOFS << "}; // n" << lNodeCount << std::endl;
        lOFS << std::endl;

        lNodeCount++;
	}

    // Construct the network.  We cannot initialize the nodes inside the network structure below
    // as we did for edges inside nodes because that kind of nested initialization is not allowed
    // by the compiler.
    lOFS << "Network gNetwork = {" << std::endl;
    lOFS << "    " << lGraphState->getNumberInputs() << ", // number of inputs" << std::endl;
    lOFS << "    " << lGraphState->getNumberOutputs() << ", // number of outputs" << std::endl;
    lOFS << "    0, // activation index" << std::endl;
    lOFS << "    " << inGraph.getNumberNodes() << ", // total number of nodes" << std::endl;

    lOFS << "    {";
    lNodeCount = 0;
	for (NN::Graph::const_iterator lGraphIter = inGraph.begin(); lGraphIter != inGraph.end(); ++lGraphIter) {
        lOFS << "&n" << lNodeCount << ((lNodeCount < inGraph.getNumberNodes()-1) ? ", " : "");
        lNodeCount++;
	}
    lOFS << "} // array of node pointers" << std::endl;

    lOFS << "}; // gNetwork" << std::endl;
    lOFS << std::endl;
    lOFS << "// Parameters and settings used in simulation." << std::endl;
    lOFS << "uint8_t gInpCosSin = " << mInpCosSin->getWrappedValue() << ";" << std::endl;
    lOFS << "uint16_t gAX12MaxVel = " << mAX12MaxVel->getWrappedValue() << ";" << std::endl;
    lOFS << std::endl;

	lOFS.close();

	Beagle_StackTraceEndM("void NN::PrintRobotEvalOp::convertNetwork(NN::Graph& ioGraph)");
}

