/*************************************************************************
 *                                                                       *
 * Open Physics Abstraction Layer                                        *
 * Copyright (C) 2004-2007                                               *
 * Alan Fischer  alan.fischer@gmail.com                                  *
 * Andres Reinot  andres@reinot.com                                      *
 * Tyler Streeter  tylerstreeter@gmail.com                               *
 * All rights reserved.                                                  *
 * Web: opal.sourceforge.net                                             *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of EITHER:                                  *
 *   (1) The GNU Lesser General Public License as published by the Free  *
 *       Software Foundation; either version 2.1 of the License, or (at  *
 *       your option) any later version. The text of the GNU Lesser      *
 *       General Public License is included with this library in the     *
 *       file license-LGPL.txt.                                          *
 *   (2) The BSD-style license that is included with this library in     *
 *       the file license-BSD.txt.                                       *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files    *
 * license-LGPL.txt and license-BSD.txt for more details.                *
 *                                                                       *
 *************************************************************************/

// This is a modified version of the BaseOpalApp.h sample file in the opal
// distribution.

#ifndef BASE_OPAL_APP_H
#define BASE_OPAL_APP_H

#include "ExampleApplication.h"
#include "ExampleFrameListener.h"
#include "PhysicalEntity.h"
#include "PhysicalCamera.h"
#include "opal.h"
#include <unistd.h>


class BaseOpalApp : public ExampleApplication, public ExampleFrameListener
{
public:
    BaseOpalApp()
        : ExampleApplication(), ExampleFrameListener(),
        mSimulator(NULL), mPhysicalCamera(NULL), mPaused(false), mSimTime(0.0),
        mRealTime(true), mShowFitness(true), mFitness(0.0)
    {
        // The simulator gets set in the initalize method.
        // The physical camera gets created later after the 
        // ExampleApplication has created its Ogre camera.
    }

    ~BaseOpalApp()
    {
        destroyScene();
        delete mPhysicalCamera;
        destroyAllPhysicalEntities();
    }

    bool initialize(opal::Simulator* inSimulator, bool inRealTime=true, bool inShowFitness=true)
    {
        mSimulator = inSimulator;
        mRealTime = inRealTime;
        mShowFitness = inShowFitness;
        return setup();
    }

    /// Method used to render the simulation.  Returns the number of frames
    /// that were rendered.
    unsigned int render(opal::real inSimTime=0.0, opal::real inFitness=0.0)
    {
        if (inSimTime >= 0.0) mSimTime = inSimTime;
        if (inFitness < FLT_MAX) mFitness = inFitness;
        unsigned int lCount = 0;
        do
        {
            Ogre::WindowEventUtilities::messagePump();
            mRoot->renderOneFrame();
            lCount++;
        }
        while (mPaused);

        return lCount;
    }

    /// Update the models in the world with control forces, etc.
    virtual void updateWorld() = 0;

    /// Creates a PhysicalEntity drawn as a box.  The OPAL Solid can
    /// be any shape, however.  This is useful for prototyping scenes
    /// when you don't have a specific visual mesh to use.  If the name
    /// string is empty, a unique name will be automatically generated;
    /// otherwise, the given name must be unique.
    void createPhysicalEntityBox(Ogre::SceneNode* parentNode, Ogre::Vector3 translationOffset,
                                 Ogre::Quaternion rotationOffset, const std::string& name,
                                 const std::string& materialName, Ogre::Vector3 dimensions,
                                 opal::Solid* s)
    {
        // Create an Ogre SceneNode for the Entity.
        Ogre::SceneNode* sn = parentNode->createChildSceneNode(name, translationOffset, rotationOffset);

        // Scale the object according to the given dimensions.
        sn->scale(dimensions[0], dimensions[1], dimensions[2]);

        // Create an Ogre Entity using a cube mesh.  This mesh must be
        // stored as a box with dimensions 1x1x1.
        Ogre::Entity* e = mSceneMgr->createEntity(name, "cube.mesh");
        e->setMaterialName(materialName);

        // Keep the normals normalized even after scaling. (No longer needed in Ogre 1.6)
        //e->setNormaliseNormals(true);

        // Attach the Entity to the SceneNode.
        sn->attachObject(e);
    }

    /// Creates a PhysicalEntity drawn as a sphere.  The OPAL Solid can
    /// be any shape, however.  This is useful for prototyping scenes
    /// when you don't have a specific visual mesh to use.  If the name
    /// string is empty, a unique name will be automatically generated;
    /// otherwise, the given name must be unique.
    void createPhysicalEntitySphere(Ogre::SceneNode* parentNode, Ogre::Vector3 translationOffset,
                                    Ogre::Quaternion rotationOffset, const std::string& name,
                                    const std::string& materialName, Ogre::Real radius,
                                    opal::Solid* s)
    {
        // Create an Ogre SceneNode for the Entity.
        Ogre::SceneNode* sn = parentNode->createChildSceneNode(name, translationOffset, rotationOffset);

        // Scale the object according to the given dimensions.
        sn->scale(radius, radius, radius);

        // Create an Ogre Entity using a sphere mesh.  This mesh must be
        // stored as a sphere with radius 1.
        Ogre::Entity* e = mSceneMgr->createEntity(name, "sphere.mesh");
        e->setMaterialName(materialName);

        // Keep the normals normalized even after scaling. (No longer needed in Ogre 1.6)
        //e->setNormaliseNormals(true);

        // Attach the Entity to the SceneNode.
        sn->attachObject(e);
    }

    /// Creates a PhysicalEntity drawn as a capsule.  The OPAL Solid can
    /// be any shape, however.  This is useful for prototyping scenes
    /// when you don't have a specific visual mesh to use.  If the name
    /// string is empty, a unique name will be automatically generated;
    /// otherwise, the given name must be unique.
    void createPhysicalEntityCapsule(Ogre::SceneNode* parentNode, Ogre::Vector3 translationOffset,
                                     Ogre::Quaternion rotationOffset, const std::string& name,
                                     const std::string& materialName, Ogre::Real radius,
                                     Ogre::Real length, opal::Solid* s)
    {
        // Create Ogre Entities using cylinder and sphere meshes.

        // Create an Ogre SceneNode for the cylinder Entity.
        std::string subObjectName = "cylinder" + name;
        Ogre::SceneNode* cylinderNode = parentNode->createChildSceneNode(subObjectName, translationOffset, rotationOffset);

        // Scale the object according to the given dimensions.  This
        // will also scale the transforms for the child nodes, but
        // we disable the "inherit scale" option for child nodes
        // here so the shapes themselves don't get scaled.
        cylinderNode->scale(radius, radius, length);

        // This mesh must be stored as a cylinder with length 1 and
        // radius 1.
        Ogre::Entity* e = mSceneMgr->createEntity(subObjectName, "cylinder.mesh");
        e->setMaterialName(materialName);
        //e->setNormaliseNormals(true);
        cylinderNode->attachObject(e);

        // The spheres must use separate scene nodes that are offset
        // from the cylinder's scene node.

        // This mesh must be stored as a sphere with radius 1.
        subObjectName = "sphere0" + name;
        Ogre::SceneNode* sn = cylinderNode->createChildSceneNode(subObjectName);
        sn->setInheritScale(false);
        sn->translate(0, 0, -0.5);
        sn->scale(radius, radius, radius);
        e = mSceneMgr->createEntity(subObjectName, "sphere.mesh");
        e->setMaterialName(materialName);
        //e->setNormaliseNormals(true);
        sn->attachObject(e);

        subObjectName = "sphere1" + name;
        sn = cylinderNode->createChildSceneNode(subObjectName);
        sn->setInheritScale(false);
        sn->translate(0, 0, 0.5);
        sn->scale(radius, radius, radius);
        e = mSceneMgr->createEntity(subObjectName, "sphere.mesh");
        e->setMaterialName(materialName);
        //e->setNormaliseNormals(true);
        sn->attachObject(e);
    }

    /// Creates a PhysicalEntity with a single mesh shape, drawn as a
    /// mesh.  If the name string is empty, a unique name will be
    /// automatically generated; otherwise, the given name must be
    /// unique.
    void createPhysicalEntityMesh(Ogre::SceneNode* parentNode, Ogre::Vector3 translationOffset,
                                  Ogre::Quaternion rotationOffset, const std::string& name,
                                  const std::string& materialName, const std::string& meshName,
                                  Ogre::Real scale, opal::Solid* s)
    {
        // Create an Ogre SceneNode for the Entity.
        Ogre::SceneNode* sn = parentNode->createChildSceneNode(name, translationOffset, rotationOffset);

        // Scale the object.
        sn->scale(scale, scale, scale);

        // Create an Ogre Entity using the given mesh.
        Ogre::Entity* e = mSceneMgr->createEntity(name, meshName);
        e->setMaterialName(materialName);

        // Keep the normals normalized even after scaling.
        //e->setNormaliseNormals(true);

        // Attach the Entity to the SceneNode.
        sn->attachObject(e);
    }

    /// Creates physical entities of solids in an Opal Blueprint instance
    std::vector<PhysicalEntity*>* createPhysicalEntityBlueprint(const std::string& materialName,
                                                                opal::BlueprintInstance& instance)
    {
        std::vector<PhysicalEntity*>* peVec = new std::vector<PhysicalEntity*>();

        for (unsigned int i=0; i<instance.getNumSolids(); ++i)
        {
            opal::Solid* s = instance.getSolid(i);

            // Create an Ogre SceneNode for the Entity.
            std::string nameStr = generateUniqueName();
            Ogre::SceneNode* sn = mSceneMgr->getRootSceneNode()->createChildSceneNode(nameStr);

            const opal::SolidData& data = s->getData();
            unsigned int j=0;
            for (j=0; j<data.getNumShapes(); ++j)
            {
                opal::ShapeData* shapeData = data.getShapeData(j);
                char shapeName[512];
                sprintf(shapeName, "%s_shape_%d", nameStr.c_str(), j);
                std::string shapeMaterialName = (shapeData->texture == "") ? materialName : shapeData->texture;
                
                opal::Point3r offsetPos = shapeData->offset.getPosition();
                Ogre::Vector3 translationOffset(offsetPos[0], offsetPos[1], offsetPos[2]);
                opal::Quaternion offsetQuat = shapeData->offset.getQuaternion();
                Ogre::Quaternion rotationOffset;
                rotationOffset.x = offsetQuat.x;
                rotationOffset.y = offsetQuat.y;
                rotationOffset.z = offsetQuat.z;
                rotationOffset.w = offsetQuat.w;
                
                switch(shapeData->getType())
                {
                case opal::BOX_SHAPE:
                    {
                        opal::Vec3r dim = ((opal::BoxShapeData*)shapeData)->dimensions;
                        Ogre::Vector3 boxDim(dim[0], dim[1], dim[2]);
                        createPhysicalEntityBox(sn, translationOffset, rotationOffset,
                                                shapeName, shapeMaterialName, boxDim, s);
                        break;
                    }
                case opal::SPHERE_SHAPE:
                    {
                        opal::real r = ((opal::SphereShapeData*)shapeData)->radius;
                        createPhysicalEntitySphere(sn, translationOffset, rotationOffset,
                                                   shapeName, shapeMaterialName, r, s);
                        break;
                    }
                case opal::CAPSULE_SHAPE:
                    {
                        opal::real r = ((opal::CapsuleShapeData*)shapeData)->radius;
                        opal::real l = ((opal::CapsuleShapeData*)shapeData)->length;
                        createPhysicalEntityCapsule(sn, translationOffset, rotationOffset,
                                                    shapeName, shapeMaterialName, r, l, s);
                        break;
                    }
                default:
                    assert(false);
                }
            }

            // Create a new Physical Entity.
            PhysicalEntity* pe = new PhysicalEntity(nameStr, sn, s);

            // Store the pointer in the PhysicalEntity list and the map of named
            // physical entities.
            mPhysicalEntityList.push_back(pe);
            mPhysicalEntityMap[nameStr] = pe;

            peVec->push_back(pe);
        }

        return peVec;
    }

    /// Create an Opal solid from an Ogre mesh.
    opal::Solid* createOpalMeshSolid(const std::string& meshName, opal::real scale)
    {
        Ogre::MeshPtr mesh = Ogre::MeshManager::getSingleton().
            load(meshName, Ogre::ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

        // We need to get vertex and triangle data from the Ogre
        // Entity for our OPAL mesh Shape.  Note: The vertex array
        // returned by the Entity's submesh contains some redundant
        // vertices.

        // The following code is adapted from code given by DWORD and
        // KingPrawnVindaloo in the Ogre forums in the thread entitled
        // "Vertex Data - Possible Bug?"

        // Note that this code assumes there is only one submesh in the
        // Entity.
        Ogre::SubMesh* subMesh = mesh->getSubMesh(0);

        Ogre::VertexData* vertexData = NULL;
        if (subMesh->useSharedVertices)
        {
            vertexData = mesh->sharedVertexData;
        }
        else
        {
            vertexData = subMesh->vertexData;
        }

        // Get the vertex and triangle counts.
        opal::MeshShapeData meshData;
        meshData.numVertices = (unsigned int)vertexData->vertexCount;
        meshData.numTriangles = (unsigned int)subMesh->indexData->indexCount / 3;

        // Allocate vertex and triangle arrays.
        meshData.vertexArray = new opal::real[3 * meshData.numVertices];
        meshData.triangleArray = new unsigned int[3 * meshData.numTriangles];

        const Ogre::VertexElement* elem = vertexData->vertexDeclaration->findElementBySemantic(Ogre::VES_POSITION);
        Ogre::HardwareVertexBufferSharedPtr vbuf = vertexData->vertexBufferBinding->getBuffer(elem->getSource());

        // The following buffer might contain normals, textured
        // coordinates, etc. in addition to vertex positions, so
        // we need to be careful when iterating through it.
        unsigned char* vertexPointer = static_cast<unsigned char*>
            (vbuf->lock(Ogre::HardwareBuffer::HBL_READ_ONLY));

        // Loop over the vertices in this submesh.
        size_t vertexSize = vbuf->getVertexSize();
        for (unsigned j = 0; j < meshData.numVertices; ++j)
        {
            // Get a pointer to the base address of the vertex.
            float* tempFloat;
            elem->baseVertexPointerToElement(vertexPointer, &tempFloat);

            meshData.vertexArray[j * 3] = tempFloat[0] * scale;
            meshData.vertexArray[j * 3 + 1] = tempFloat[1] * scale;
            meshData.vertexArray[j * 3 + 2] = tempFloat[2] * scale;

            // Move to the next vertex in the array.
            vertexPointer += vertexSize;
        }

        vbuf->unlock();

        Ogre::HardwareIndexBufferSharedPtr ibuf = subMesh->indexData->indexBuffer;
        unsigned short* indexData = new unsigned short[meshData.numTriangles * 3];

        ibuf->readData(0, ibuf->getSizeInBytes(), indexData);

        // Copy index data.  'indices' is a 1-dimensional array.
        for (unsigned int j = 0; j < meshData.numTriangles; ++j)
        {
            meshData.triangleArray[j * 3] = indexData[j * 3];
            meshData.triangleArray[j * 3 + 1] = indexData[j * 3 + 1];
            meshData.triangleArray[j * 3 + 2] = indexData[j * 3 + 2];
        }

        delete[] indexData;

        // Create an OPAL Solid for the mesh.
        opal::Solid* s = mSimulator->createSolid();

        // Add the mesh Shape to the Solid.
        s->addShape(meshData);

        return s;
    }

    /// Finds a PhysicalEntity pointer by name.  Returns NULL if the
    /// PhysicalEntity could not be found.
    PhysicalEntity* getPhysicalEntity(const std::string& name)const
    {
        std::map<std::string, PhysicalEntity*>::const_iterator iter =
            mPhysicalEntityMap.find(name);

        if (mPhysicalEntityMap.end() == iter)
        {
            // The named PhysicalEntity could not be found in the map.
            return NULL;
        }
        else
        {
            return (*iter).second;
        }
    }

    /// Destroys the given PhysicalEntity.
    void destroyPhysicalEntity(PhysicalEntity* pe)
    {
        if (!pe)
        {
            return;
        }

        // If the PhysicalEntity has a name, remove it from the
        // PhysicalEntity map.
        std::string name = pe->getName();
        if (!name.empty())
        {
            mPhysicalEntityMap.erase(name);
        }

        for(size_t i = 0; i<mPhysicalEntityList.size(); ++i)
        {
            if(mPhysicalEntityList[i] == pe)
            {
                // Destroy the PhysicalEntity's Ogre SceneNode and
                // attached Entities.
                Ogre::SceneNode* sn = pe->getSceneNode();

                // Detach and destroy all objects from the SceneNode;
                // in this case, there should only be a single entity
                // (which is why we can safely call 'removeEntity'
                // here).  TODO: We should also check child nodes and
                // destroy their Entities here; this isn't a huge
                // issue since Ogre will destroy all the Entities
                // when the app exits.
                while(sn->numAttachedObjects() > 0)
                {
                    Ogre::MovableObject* thisObject = sn->detachObject(static_cast<unsigned short>(0));
                    std::string s = thisObject->getName();
                    mSceneMgr->destroyEntity(thisObject->getName());
                }

                // Destroy the SceneNodes all of its children.
                sn->removeAndDestroyAllChildren();
                mSceneMgr->getRootSceneNode()->removeAndDestroyChild(sn->getName());
                sn = NULL;

                // Delete the PhysicalEntity.
                delete pe;
                pe = NULL;

                // Copy the last element over this one and pop the last
                // one off, removing it from the PhysicalEntity list.
                mPhysicalEntityList[i] = mPhysicalEntityList.back();
                mPhysicalEntityList.pop_back();

                // Return early since we've already found the desired
                // element.
                return;
            }
        }
    }

    /// Destroys all of the PhysicalEntities.
    void destroyAllPhysicalEntities()
    {
        while (!mPhysicalEntityList.empty())
        {
            destroyPhysicalEntity(mPhysicalEntityList.back());
        }

        // Every pointer in the mPhysicalEntityMap is a duplicate of
        // a pointer in mPhysicalEntityList, so we don't need to
        // delete the memory they point to.
        mPhysicalEntityMap.clear();
    }

protected:
    /// Create a PhysicalCamera that uses the ExampleApplication's
    /// Ogre camera.
    virtual void createPhysicalCamera()
    {
        mPhysicalCamera = new PhysicalCamera(mSimulator,
                                             ExampleApplication::mCamera, false);
    }

    /// Register the frame listener with the Ogre root so it will
    /// call the listener at the appropriate time.
    virtual void createFrameListener()
    {
        ExampleFrameListener::init(ExampleApplication::mWindow,
                                   ExampleApplication::mCamera);
        mRoot->addFrameListener(this);
    }

    /// Initial scene setup function.  Create and position lights, load
    /// models, create a camera, etc.
    virtual void createScene() = 0;

    /// Called once per frame after input and physics have been updated.
    /// This is a place for applications to do application-specific
    /// updates per frame.  It takes the dt since the last frame as a
    /// parameter.  This should return false if we should break out of
    /// the main loop.
    virtual bool appFrameStarted(Ogre::Real dt) = 0;

    /// Called at the start of each frame.  Returns false if we
    /// should break out of the main loop.  Applications should not
    /// override this; instead, they should override 'appFrameStarted'.
    virtual bool frameStarted(const Ogre::FrameEvent& evt)
    {
        if(ExampleApplication::mWindow->isClosed())
        {
            return false;
        }

        Ogre::Real dt = evt.timeSinceLastFrame;

        if (false == handleInput(dt))
        {
            return false;
        }

        if (!mPaused)
        {
            // Update the PhysicalEntities.
            size_t size = mPhysicalEntityList.size();
            for(size_t i = 0; i<size; ++i)
            {
                mPhysicalEntityList.at(i)->update(dt);
            }
        }

        // Update the PhysicalCamera.
        mPhysicalCamera->update(dt);

        // Notify the application about the new frame.
        if (false == appFrameStarted(dt))
        {
            return false;
        }

        return true;
    }

    /// Returns false if we should break out of the main loop.
    bool handleInput(Ogre::Real dt)
    {
        // This variable can be used to keep keys from repeating too fast.
        if (mTimeUntilNextToggle >= 0)
        {
            mTimeUntilNextToggle -= dt;
        }

        // Acquire data from input devices.
        mKeyboard->capture();
        mMouse->capture();

        // Handle keyboard input.
        if (processUnbufferedKeyInput(dt) == false)
        {
            return false;
        }

        // Handle mouse input.
        if (processUnbufferedMouseInput(dt) == false)
        {
            return false;
        }

        return true;
    }

    /// Process keyboard input here.  Returns false if we should break out
    /// of the main loop.
    virtual bool processUnbufferedKeyInput(Ogre::Real dt) = 0;

    /// Process mouse input here.  Returns false if we should break out of
    /// the main loop.
    virtual bool processUnbufferedMouseInput(Ogre::Real dt) = 0;

    /// Returns a unique name string.  Useful when creating lots of
    /// anonymous objects.
    std::string generateUniqueName()
    {
        static unsigned int namelessObjectCount = 0;

        // Make a unique name.
        char newName[20];
        sprintf(newName, "object%d", namelessObjectCount);
        std::string nameStr = newName;
        ++namelessObjectCount;

        return nameStr;
    }

    /// Pointer to the OPAL Simulator.
    opal::Simulator* mSimulator;

    /// Pointer to the PhysicalCamera.
    PhysicalCamera* mPhysicalCamera;

    /// True when the physics simulation is paused.
    bool mPaused;

    /// Map of named PhysicalEntities.  This is just used to find a
    /// PhysicalEntity by name.
    std::vector<PhysicalEntity*> mPhysicalEntityList;

    /// Map of named PhysicalEntities.
    std::map<std::string, PhysicalEntity*> mPhysicalEntityMap;

    /// Current time step taken by the simulator.
    opal::real mSimTime;
    
    /// Whether simulation is being run in real time.
    bool mRealTime;

    /// Whether to show the current fitness of the agent.
    bool mShowFitness;

    /// The current fitness of the agent.
    opal::real mFitness;

private:
};

#endif
