
/*!
 *  \file   ManualEvalOp.cpp
 *  \brief  Implementation of the class ManualEvalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"
#include "ManualEvalOp.hpp"

#include <cmath>

using namespace Beagle;


/*!
 *  \brief Construct a new robot evaluation operator.
 *  \param inName Name of the operator.
 */
NN::ManualEvalOp::ManualEvalOp(const std::string& inName) :
		VizRobotEvalOp(inName)
{}


/*!
 *  \brief Evaluate the graph fitness for the robot walking problem.
 *  \param ioGraph Graph to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness measure.
 */
NN::FitnessIndexed::Handle NN::ManualEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)
{
	Beagle_StackTraceBeginM();

	float lMaxTime = mMaxTime->getWrappedValue();

	NN::FitnessIndexed::Handle lFitness = new FitnessIndexed(1, 0.0);

	float lOutMult = ((*mOutMult)[0]);
    unsigned int lFitnessSel = mFitnessSel->getWrappedValue();
    bool lAX12CM2 = mAX12CM2->getWrappedValue();

    if (mEnvironProg->getWrappedValue()) createEnviron(ioContext);
    createAgent(0, 0.02, 0, 1);
    opal::Point3r lPos0 = mAgent->getSolid(0)->getPosition();
    opal::real lTempFitness = 0.0;

    // Define a volume solid to use with volume sensors.
    //opal::Solid* lVolume = mSimulator->createSolid();
    //lVolume->setEnabled(false);
    //opal::CapsuleShapeData lCapsuleData;
    //lCapsuleData.radius = 0.1;
    //lCapsuleData.length = 0.1;
    //lVolume->addShape(lCapsuleData);

    mTime = 0.0;
    opal::real dt = mSimulator->getStepSize();
    opal::Point3r lPosPrv = lPos0;
    opal::Point3r lPosCur = lPos0;
    while (mTime < lMaxTime) {
        unsigned int lNumJoints = mAgent->getNumJoints();
        unsigned int lNumAxes = mAgent->getJoint(0)->getNumAxes();
        for (unsigned int i = 0; i < lNumJoints; ++i) {
            if (lNumAxes == 1) {
                opal::ServoMotor* lMotor = (opal::ServoMotor*)mAgent->getMotor(i);
                opal::real lAngle = mTime*lOutMult;
                if (lAX12CM2) {
                    // Since the zeros of the legs are initially 30 degrees from the vertical,
                    // there is a total 60 degrees offset between the left and right sides.
                    if (i%2 == 1) lAngle += 0.16666667;
                }
                if (i%4 == 0 || i%4 == 3) {
                    lAngle = (lAngle+0.5) - std::floor(lAngle+0.5);
                }
                else {
                    lAngle = lAngle - std::floor(lAngle);
                }

                // Invert sign for right side legs
                if (lAX12CM2 && i%2 == 1) lAngle = 1.0 - lAngle;

                lMotor->setDesiredAngleNorm(lAngle);
            }
            else {
                opal::real lAngle = mTime*lOutMult;
                if (i%4 == 0 || i%4 == 3) {
                    lAngle = (lAngle+0.5) - std::floor(lAngle+0.5);
                }
                else {
                    lAngle = lAngle - std::floor(lAngle);
                }
                opal::real lA = lAngle;
                opal::real lB = 2*lA;
                opal::real lC = (lB < 1.0) ? lB : (2.0 - lB);
                opal::real lD = (lAngle+0.25) - std::floor(lAngle+0.25);
                opal::real lE = 2*lD;
                opal::real lF = (lE < 1.0) ? lE : (2.0 - lE);

                // opal::real lAngle = mTime*lOutMult;
                // opal::real lA = (lAngle+0.50) - std::floor(lAngle+0.50);
                // opal::real lB = 2*lA;
                // opal::real lC = (lB < 1.0) ? lB : (2.0 - lB);
                // opal::real lD = (lAngle+0.25) - std::floor(lAngle+0.25);
                // opal::real lE = 2*lD;
                // opal::real lF = (lE < 1.0) ? lE : (2.0 - lE);

                // // Invert sign for legs 1 and 2.
                // if (i%4 == 1 || i%4 == 2) {
                //     lC = 1.0 - lC;
                //     lF = 1.0 - lF;
                // }

                opal::ServoMotor* lMotor1 = (opal::ServoMotor*)mAgent->getMotor(2*i);
                lMotor1->setDesiredAngleNorm(sin(lC*M_PI/2));

                opal::ServoMotor* lMotor2 = (opal::ServoMotor*)mAgent->getMotor(2*i+1);
                lMotor2->setDesiredAngleNorm(sin(lF*M_PI/2));
            }
        }

        // for (unsigned int i = 0; i < lNumJoints/2; ++i) {
        //     opal::real lAngle = mTime*lOutMult;
        //     if (i%4 == 0 || i%4 == 3) {
        //         lAngle = (lAngle+0.5) - std::floor(lAngle+0.5);
        //     }
        //     else {
        //         lAngle = lAngle - std::floor(lAngle);
        //     }
        //     opal::real lA = lAngle;
        //     opal::real lB = 2*lA;
        //     opal::real lC = (lB < 1.0) ? lB : (2.0 - lB);
        //     opal::real lD = (lAngle+0.25) - std::floor(lAngle+0.25);
        //     opal::real lE = 2*lD;
        //     opal::real lF = (lE < 1.0) ? lE : (2.0 - lE);

        //     // Invert sign for right side legs.
        //     //if (i%2 == 1) {
        //     //    lC = 1.0 - lC;
        //     //    lF = 1.0 - lF;
        //     //}

        //     opal::ServoMotor* lMotor1 = (opal::ServoMotor*)mAgent->getMotor(i);
        //     //lMotor1->setDesiredAngleNorm(sin(lC*M_PI/2));
        //     lMotor1->setDesiredAngleNorm(lC);

        //     opal::ServoMotor* lMotor2 = (opal::ServoMotor*)mAgent->getMotor(lNumJoints/2+i);
        //     //lMotor2->setDesiredAngleNorm(sin(lF*M_PI/2));
        //     lMotor2->setDesiredAngleNorm(lF);
        // }

//         for (unsigned int i = 0; i < mAgent->getNumSensors(); ++i) {
//             opal::VolumeSensor* lVolumeSensor = (opal::VolumeSensor*)mAgent->getSensor(i);
//             opal::VolumeQueryResult lVolumeResult = lVolumeSensor->queryVolume(lVolume);
//             for (unsigned int j = 0; j < lVolumeResult.getNumSolids(); ++j) {
//                 opal::Solid* lResult = lVolumeResult.getSolid(j);
//                 std::cout << mTime << " " << i << " " << j << " " << lResult->getName() << std::endl;
//             }
//         }

        // Simulate the next time step.
        simulateAgent(dt);
        mTime += dt;

        // Update position.
        lPosPrv = lPosCur;
        lPosCur = mAgent->getSolid(0)->getPosition();

        // Stop simulation if something bad happens to the robot like it is in a maze and gets thrown up in the air.
        if (lFitnessSel == 0 && absolute<opal::real>(lPos0[1]-lPosCur[1]) > 1.5) {
            break;
        }
        else if (lFitnessSel == 1 && absolute<opal::real>(lPos0[1]-lPosCur[1]) > 0.8) {
            break;
        }

        // Fitness based on the robot moving forward (local Z direction)
        opal::Vec3r lVelCur = mAgent->getSolid(0)->getLocalLinearVel();
        if (lFitnessSel == 3) {
            lTempFitness += lVelCur.z*dt;
        }
        // Fitness based on the robot moving sideways (local X direction)
        else if (lFitnessSel == 4) {
            lTempFitness += lVelCur.x*dt;
        }
    }

    //mSimulator->destroySolid(lVolume);

    opal::Point3r lPos1 = mAgent->getSolid(0)->getPosition();
    removeAgent();
    if (mEnvironProg->getWrappedValue()) removeEnviron();
    
    // Fitness based on max distance moved in the X or Z directions.
    if (lFitnessSel == 0) {
        (*lFitness)[0] += maxOf<opal::real>(absolute<opal::real>(lPos0[0]-lPosCur[0]), absolute<opal::real>(lPos0[2]-lPosCur[2]));
    }
    // Fitness based on sum of distance moved in the X and Z directions.
    else if (lFitnessSel == 1) {
        (*lFitness)[0] += absolute<opal::real>(lPos0[0]-lPosCur[0]) + absolute<opal::real>(lPos0[2]-lPosCur[2]);
    }
    // Fitness based on distance moved in the XZ plane.
    else if (lFitnessSel == 2) {
        lPos0[1] = 0.0;
        lPosCur[1] = 0.0;
        (*lFitness)[0] += opal::distance(lPos0, lPosCur);
    }
    // Fitness based on the robot moving forward (local Z direction) or sideways (local X direction).
    else if (lFitnessSel == 3 || lFitnessSel == 4) {
        (*lFitness)[0] += absolute<opal::real>(lTempFitness);
    }
    // Fitness based on absolute distance moved in the Z direction.
    else if (lFitnessSel == 5) {
        (*lFitness)[0] += absolute<opal::real>(lPosCur[2]-lPos0[2]);
    }
    // Fitness based on distance moved in the positive Z direction.
    else if (lFitnessSel == 6) {
        (*lFitness)[0] += lPosCur[2] - lPos0[2];
    }
    // Fitness based on absolute distance moved in the Z direction, less the absolute distance moved in the X direction.
    else if (lFitnessSel == 7) {
        (*lFitness)[0] += absolute<opal::real>(lPosCur[2]-lPos0[2]) - absolute<opal::real>(lPosCur[0]-lPos0[0]);
    }

	return lFitness;

	Beagle_StackTraceEndM("NN::FitnessIndexed::Handle NN::ManualEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)");
}

