/***************************************************************************
 *   Copyright (C) 2006 by Vinod K. Valsalam   *
 *   vkv@cs.utexas.edu   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ROBOT_APP_H
#define ROBOT_APP_H

#include <time.h>
#include "BaseOpalApp.h"

class RobotApp : public BaseOpalApp
{
public:
    RobotApp(unsigned int inFrameRate=30, opal::real inCameraPeriod=15)
        : BaseOpalApp(), mUseShadows(true), mEnvironEntities(NULL), mAgentEntities(NULL),
        mAgent(0), mFrameRate(inFrameRate)
	{
        mCameraAngVel = 2*M_PI/inCameraPeriod;
	}

	~RobotApp()
	{
	}

	/// Create a physical environment from the blueprint instance.
	void createEnviron(opal::BlueprintInstance* environ, const std::string& materialName = "Plastic/Gray")
	{
        // Create the graphical elements for it.
        mEnvironEntities = createPhysicalEntityBlueprint(materialName, *environ);
	}

	/// Remove the last environment that was added.
	void removeEnviron()
	{
		// Remove all physical entities associated with the environment.
        while (!mEnvironEntities->empty())
        {
            destroyPhysicalEntity(mEnvironEntities->back());
            mEnvironEntities->pop_back();
        }
        delete mEnvironEntities;
	}

	/// Create a physical agent in the environment from the blueprint instance.
	void createAgent(opal::BlueprintInstance* agent, const std::string& materialName = "Textured/Wood")
	{
        mAgent = agent;

        // Create the graphical elements for it.
        mAgentEntities = createPhysicalEntityBlueprint(materialName, *agent);
	}

	/// Remove the last agent that was added from the environment.
	void removeAgent()
	{
		// Remove all physical entities associated with the agent.
        while (!mAgentEntities->empty())
        {
            destroyPhysicalEntity(mAgentEntities->back());
            mAgentEntities->pop_back();
        }
        delete mAgentEntities;

        mAgent = 0;
	}

	virtual void updateWorld()
	{
	}

	/// Save a screenshot of the contents of the display window.
	void dumpImage(char* fname)
	{
		ExampleApplication::mWindow->writeContentsToFile(fname);
	}

protected:
	/// Create the background objects in the scene.
	void setupSceneBackground()
	{
		// Setup shadows.
		if (mUseShadows)
		{
			mSceneMgr->setShadowTechnique(Ogre::SHADOWTYPE_STENCIL_ADDITIVE);
		}
		else
		{
			mSceneMgr->setShadowTechnique(Ogre::SHADOWTYPE_NONE);
		}

		// Set the ambient light level.
		mSceneMgr->setAmbientLight(Ogre::ColourValue(0.7, 0.7, 0.7));

		// Create a light source.
		Ogre::Light* light = mSceneMgr->createLight("light0");
		light->setType(Ogre::Light::LT_POINT);
		light->setDiffuseColour(1.0, 1.0, 1.0);
		light->setSpecularColour(0.8, 0.8, 0.8);
		light->setPosition(100.0, 300.0, 100.0);

		// Create another light source.
		light = mSceneMgr->createLight("light1");
		light->setType(Ogre::Light::LT_POINT);
		light->setDiffuseColour(0.5, 0.5, 0.5);
		light->setSpecularColour(0.5, 0.5, 0.5);
		light->setPosition(-100.0, 300.0, 100.0);
		light->setVisible(false);

		// Setup the initial camera position.
        mPhysicalCamera->setFixedYawAxis(true);  // so that the camera pans properly when following agent
		mPhysicalCamera->setPosition(opal::Point3r(0.8, 0.2, 0));
		mPhysicalCamera->lookAt(opal::Point3r(0, 0, 0));

		// Create the SkyBox
		mSceneMgr->setSkyBox(true, "Skyboxes/Clouds");

		// Make a sphere used to show the grasp position.
		Ogre::SceneNode* sphereNode = mSceneMgr->getRootSceneNode()->
				createChildSceneNode("pickingSphereNode");
		sphereNode->scale(0.005, 0.005, 0.005);
		Ogre::Entity* sphereEntity = mSceneMgr->createEntity("pickingSphere",
				"sphere.mesh");
		sphereEntity->setMaterialName("Unlit/Yellow");
		sphereEntity->setCastShadows(false);
		sphereEntity->setVisible(false);
		sphereNode->attachObject(sphereEntity);

		// Make a "line" (scaled cylinder) to show the grasping spring.
		Ogre::SceneNode* lineNode = mSceneMgr->getRootSceneNode()->
				createChildSceneNode("springLineNode");
		Ogre::Entity* lineEntity = mSceneMgr->createEntity("springLine",
				"cylinder.mesh");
		lineEntity->setMaterialName("Unlit/White");
		lineEntity->setCastShadows(false);
		lineEntity->setVisible(false);
		lineNode->attachObject(lineEntity);

		// Make a crosshairs for picking.
		Ogre::Overlay* crosshair =
				Ogre::OverlayManager::getSingleton().getByName("General/Crosshair");
		crosshair->show();

		// Show fitness if its display is enabled.
        if (mShowFitness) {
            Ogre::Overlay* fitness =
				Ogre::OverlayManager::getSingleton().getByName("General/Fitness");
            fitness->show();
        }
	}

	/// Create the physical environment for the world from all the solids in the simulator.
	void setupPhysicalEnvironment(const std::string& materialName = "Plastic/Gray")
	{
        for (unsigned int i=0; i<mSimulator->getNumSolids(); ++i)
        {
            opal::Solid* s = mSimulator->getSolid(i);

            // Create an Ogre SceneNode for the Entity.
            std::string nameStr = generateUniqueName();
            Ogre::SceneNode* sn = mSceneMgr->getRootSceneNode()->createChildSceneNode(nameStr);

            const opal::SolidData& data = s->getData();
            unsigned int j=0;
            for (j=0; j<data.getNumShapes(); ++j)
            {
                opal::ShapeData* shapeData = data.getShapeData(j);
                char shapeName[512];
                sprintf(shapeName, "%s_shape_%d", nameStr.c_str(), j);
                std::string shapeMaterialName = (shapeData->texture == "") ? materialName : shapeData->texture;

                opal::Point3r offsetPos = shapeData->offset.getPosition();
                Ogre::Vector3 translationOffset(offsetPos[0], offsetPos[1], offsetPos[2]);
                opal::Quaternion offsetQuat = shapeData->offset.getQuaternion();
                Ogre::Quaternion rotationOffset;
                rotationOffset.x = offsetQuat.x;
                rotationOffset.y = offsetQuat.y;
                rotationOffset.z = offsetQuat.z;
                rotationOffset.w = offsetQuat.w;

                switch(shapeData->getType())
                {
                case opal::BOX_SHAPE:
                    {
                        opal::Vec3r dim = ((opal::BoxShapeData*)shapeData)->dimensions;
                        Ogre::Vector3 boxDim(dim[0], dim[1], dim[2]);
                        createPhysicalEntityBox(sn, translationOffset, rotationOffset,
                                                shapeName, shapeMaterialName, boxDim, s);
                        break;
                    }
                case opal::SPHERE_SHAPE:
                    {
                        opal::real r = ((opal::SphereShapeData*)shapeData)->radius;
                        createPhysicalEntitySphere(sn, translationOffset, rotationOffset,
                                                   shapeName, shapeMaterialName, r, s);
                        break;
                    }
                case opal::CAPSULE_SHAPE:
                    {
                        opal::real r = ((opal::CapsuleShapeData*)shapeData)->radius;
                        opal::real l = ((opal::CapsuleShapeData*)shapeData)->length;
                        createPhysicalEntityCapsule(sn, translationOffset, rotationOffset,
                                                    shapeName, shapeMaterialName, r, l, s);
                        break;
                    }
                default:
                    assert(false);
                }
            }

            // Create a new Physical Entity.
            PhysicalEntity* pe = new PhysicalEntity(nameStr, sn, s);

            // Store the pointer in the PhysicalEntity list and the map of named
            // physical entities.
            mPhysicalEntityList.push_back(pe);
            mPhysicalEntityMap[nameStr] = pe;
        }

        //drawLine();
	}


    // Draw a line along the Z axis.
    void drawLine()
    {
        Ogre::ManualObject* lLineObject = mSceneMgr->createManualObject("manualLine");
        Ogre::SceneNode* lLineObjectNode = mSceneMgr->getRootSceneNode()->createChildSceneNode("manualLineNode");

        lLineObject->begin("Plastic/Red", Ogre::RenderOperation::OT_TRIANGLE_STRIP);
        lLineObject->position(-1, 0.01, 100);
        lLineObject->position(1, 0.01, 100);
        lLineObject->position(-1, 0.01, -100);
        lLineObject->position(1, 0.01, -100);
        lLineObject->end();

        lLineObjectNode->attachObject(lLineObject);
    }


	virtual void createScene()
	{
        setupSceneBackground();
		setupPhysicalEnvironment();
	}


	virtual bool appFrameStarted(Ogre::Real dt)
	{
		// Do per-frame application-specific things here.

		// Update the grasping spring line.
		if (mPhysicalCamera->isGrasping())
		{
			Ogre::Entity* pickingSphere = mSceneMgr->getEntity("pickingSphere");
			if (!pickingSphere->isVisible())
			{
				pickingSphere->setVisible(true);
			}

			Ogre::Entity* springLine = mSceneMgr->getEntity("springLine");
			if (!springLine->isVisible())
			{
				springLine->setVisible(true);
			}

			opal::Point3r desiredPos = mPhysicalCamera->getGraspGlobalPos();
			Ogre::Vector3 point0(desiredPos[0], desiredPos[1], desiredPos[2]);

			opal::Point3r attachPos = mPhysicalCamera->getAttachGlobalPos();
			Ogre::Vector3 point1(attachPos[0], attachPos[1], attachPos[2]);

			pickingSphere->getParentSceneNode()->setPosition(point1);

			Ogre::Vector3 lineVec = point0 - point1;
			if (!lineVec.isZeroLength())
			{
				Ogre::SceneNode* springLineNode = springLine->getParentSceneNode();
				springLineNode->setPosition(0.5 * (point0 + point1));

				springLineNode->setDirection(lineVec, Ogre::Node::TS_WORLD);
				springLineNode->setScale(0.01, 0.01, lineVec.length());
			}
			else
			{
				springLine->setVisible(false);
			}
		}
		else
		{
			Ogre::Entity* pickingSphere = mSceneMgr->getEntity("pickingSphere");
			if (pickingSphere->isVisible())
			{
				pickingSphere->setVisible(false);
			}

			Ogre::Entity* springLine = mSceneMgr->getEntity("springLine");
			if (springLine->isVisible())
			{
				springLine->setVisible(false);
			}
		}

        // Follow the agent with the camera.
        if (mAgent != 0) {
        //if (false) {
            opal::real lCameraRadius = 0.525;
            opal::Point3r lAgentPos = mAgent->getSolid(0)->getPosition();
            mPhysicalCamera->setPosition(opal::Point3r(lAgentPos[0] + lCameraRadius*cos(mCameraAngVel*mSimTime), 0.2, lAgentPos[2] + lCameraRadius*sin(mCameraAngVel*mSimTime)));
            mPhysicalCamera->lookAt(opal::Point3r(lAgentPos[0], lAgentPos[1], lAgentPos[2]));  // y component set to 0 to reduce shaking
        }

        if (mShowFitness) {
            static Ogre::String fitness = "Fitness: ";

            Ogre::OverlayElement* guiFitness = Ogre::OverlayManager::getSingleton().getOverlayElement("General/CurrentFitness");
            char buffer [50];
            sprintf(buffer, "% 6.2f", mFitness);
            guiFitness->setCaption(fitness + buffer);
        }

        // Take snapshots for making movies if simulation is not running in real time.
        if (!mRealTime) {
            // Turn off overlays before taking screenshots
            Ogre::OverlayManager::getSingleton().getByName("General/Crosshair")->hide();
            showDebugOverlay(false);

            while (mNumScreenShots <= mSimTime*mFrameRate) {
                char lTemp[20];
                sprintf(lTemp, "screenshot-%d.png", mNumScreenShots++);
                ExampleApplication::mWindow->writeContentsToFile(lTemp);
            }
        }

		// Return true to continue looping.
		return true;
	}

	virtual bool processUnbufferedKeyInput(Ogre::Real dt)
	{
		// Check if we should quit looping.
		if(mKeyboard->isKeyDown(OIS::KC_ESCAPE) || mKeyboard->isKeyDown(OIS::KC_Q))
		{
			return false;
		}

		// Check if we should pause physics.
		if(mKeyboard->isKeyDown(OIS::KC_P) && mTimeUntilNextToggle <= 0)
		{
			mPaused = !mPaused;

			// Reset the timer for toggle keys.
			mTimeUntilNextToggle = 0.5;
		}

		// The following code updates the camera's position.
		opal::Vec3r cameraDir(0, 0, 0);
		bool cameraMoved = false;

		if (mKeyboard->isKeyDown(OIS::KC_LEFT) 
			|| mKeyboard->isKeyDown(OIS::KC_A))
		{
			// Move camera left.
			cameraDir[0] -= (dt * mMoveSpeed);
			cameraMoved = true;
		}

		if (mKeyboard->isKeyDown(OIS::KC_RIGHT) 
			|| mKeyboard->isKeyDown(OIS::KC_D))
		{
			// Move camera right.
			cameraDir[0] += (dt * mMoveSpeed);
			cameraMoved = true;
		}

		if (mKeyboard->isKeyDown(OIS::KC_UP) 
			|| mKeyboard->isKeyDown(OIS::KC_W))
		{
			// Move camera forward.
			cameraDir[2] -= (dt * mMoveSpeed);
			cameraMoved = true;
		}

		if (mKeyboard->isKeyDown(OIS::KC_DOWN) 
			|| mKeyboard->isKeyDown(OIS::KC_S))
		{
			// Move camera backward.
			cameraDir[2] += (dt * mMoveSpeed);
			cameraMoved = true;
		}

		if (!cameraMoved)
		{
			// Slow physical camera motion if necessary.
		}

		// Use the camera dir vector to translate the camera.
		mPhysicalCamera->moveRelative(cameraDir);

		Ogre::Degree rotY(0.0);
		Ogre::Degree rotX(0.0);
		bool cameraRotated = false;

		if (mKeyboard->isKeyDown(OIS::KC_J))
		{
			// Rotate camera left.
			rotY += (dt * mRotateSpeed);
			cameraRotated = true;
		}

		if (mKeyboard->isKeyDown(OIS::KC_L))
		{
			// Rotate camera right.
			rotY -= (dt * mRotateSpeed);
			cameraRotated = true;
		}

		if (mKeyboard->isKeyDown(OIS::KC_I))
		{
			// Rotate camera up.
			rotX += (dt * mRotateSpeed);
			cameraRotated = true;
		}

		if (mKeyboard->isKeyDown(OIS::KC_K))
		{
			// Rotate camera down.
			rotX -= (dt * mRotateSpeed);
			cameraRotated = true;
		}

		if (!cameraRotated)
		{
			// Slow physical camera rotation if necessary.
		}

		// Use the rotY and rotX angles to rotate the camera.
		mPhysicalCamera->yawRelative(rotY.valueDegrees());
		mPhysicalCamera->pitchRelative(rotX.valueDegrees());

		// Toggle shadows.
		if(mKeyboard->isKeyDown(OIS::KC_H) && mTimeUntilNextToggle <= 0)
		{
			mUseShadows = !mUseShadows;

			if (mUseShadows)
			{
				mSceneMgr->setShadowTechnique(Ogre::SHADOWTYPE_STENCIL_ADDITIVE);
			}
			else
			{
				mSceneMgr->setShadowTechnique(Ogre::SHADOWTYPE_NONE);
			}

			// Reset the timer for toggle keys.
			mTimeUntilNextToggle = 0.5;
		}

		// Toggle second light source.
		if(mKeyboard->isKeyDown(OIS::KC_B) && mTimeUntilNextToggle <= 0)
		{
			Ogre::Light* light1 = mSceneMgr->getLight("light1");

			if (light1->isVisible())
			{
				light1->setVisible(false);
			}
			else
			{
				light1->setVisible(true);
			}

			// Reset the timer for toggle keys.
			mTimeUntilNextToggle = 0.5;
		}

		// Toggle GUI.
		if (mKeyboard->isKeyDown(OIS::KC_G) && mTimeUntilNextToggle <= 0)
		{
			mStatsOn = !mStatsOn;
			showDebugOverlay(mStatsOn);
			mTimeUntilNextToggle = 0.5;
		}

		// Toggle Crosshair.
		if (mKeyboard->isKeyDown(OIS::KC_C) && mTimeUntilNextToggle <= 0)
		{
            Ogre::Overlay* overlay =
				Ogre::OverlayManager::getSingleton().getByName("General/Crosshair");
            if (overlay->isVisible()) {
                overlay->hide();
            }
            else {
                overlay->show();
            }
			mTimeUntilNextToggle = 0.5;
        }

		// Toggle Fitness.
		if (mKeyboard->isKeyDown(OIS::KC_F) && mTimeUntilNextToggle <= 0 && mShowFitness)
		{
            Ogre::Overlay* overlay =
				Ogre::OverlayManager::getSingleton().getByName("General/Fitness");
            if (overlay->isVisible()) {
                overlay->hide();
            }
            else {
                overlay->show();
            }
			mTimeUntilNextToggle = 0.5;
        }

		// Return true to continue looping.
		return true;
	}

	virtual bool processUnbufferedMouseInput(Ogre::Real dt)
	{
		// The following code checks how far the mouse has move since 
		// the last poll.  This data can be used to rotate the camera 
		// around its X and Y axes (pitch and yaw, respectively).
		const OIS::MouseState &ms = mMouse->getMouseState();
//    	Ogre::Degree rotY = -ms.X.rel * mRotateSpeed;
//    	Ogre::Degree rotX = -ms.Y.rel * mRotateSpeed;
// 
//    	// Use the relative mouse motion to rotate the camera.
//    	mPhysicalCamera->yawRelative(rotY.valueDegrees());
//    	mPhysicalCamera->pitchRelative(rotX.valueDegrees());

		// Check mouse button states.
		if (ms.buttonDown(OIS::MB_Left))
		{
			// The left mouse button is down.
			mPhysicalCamera->grasp();
		}
		else
		{
			// The left mouse button is up.
			mPhysicalCamera->release();
		}

		// Return true to continue looping.
		return true;
	}

private:
	/// Used to toggle shadows on and off.
	bool mUseShadows;

	/// Physical entities associated with the environment.
	std::vector<PhysicalEntity*>* mEnvironEntities;

	/// Physical entities associated with the robot agent.
	std::vector<PhysicalEntity*>* mAgentEntities;

    /// The blueprint instance of the current agent.
    opal::BlueprintInstance* mAgent;

    /// Frame rate for making movies.
    unsigned int mFrameRate;

    /// Camera angular velocity as it follows the agent.
    opal::real mCameraAngVel;

};

#endif
