
/*!
 *  \file   VizRobotEvalOp.cpp
 *  \brief  Implementation of the class VizRobotEvalOp.
 *  \author Vinod Valsalam
 *  $Revision: $
 *  $Date: $
 */

#include "beagle/NN.hpp"
#include "VizRobotEvalOp.hpp"

#include <cmath>

using namespace Beagle;


RobotApp* NN::VizRobotEvalOp::mRobotApp = 0;


/*!
 *  \brief Construct a new robot evaluation operator.
 *  \param inName Name of the operator.
 */
NN::VizRobotEvalOp::VizRobotEvalOp(const std::string& inName) :
		NN::RobotEvalOp(inName),
        mTimer(false),
        mPrvTime(0.0),
        mCurTime(-1.0),
        mAccTime(0.0)
{ }


/*!
 *  \brief Register the parameters of the evaluation operator.
 *  \param ioSystem System of the evolution.
 */
void NN::VizRobotEvalOp::registerParams(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();
	NN::RobotEvalOp::registerParams(ioSystem);
	{
		Register::Description lDescription(
		    "Whether to run visualization of the simulation in real time",
		    "Bool",
		    "1",
		    "The simulation is run in real time if this parameter is set."
		);
		mRealTime = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.vizrobot.realtime", new Bool(1), lDescription));
	}
	{
		Register::Description lDescription(
		    "Whether to show current fitness of the agent in the visualization",
		    "Bool",
		    "1",
		    "The current fitness of the agent is shown in the visualization if this parameter is set."
		);
		mShowFitness = castHandleT<Bool>(
		                  ioSystem.getRegister().insertEntry("nn.vizrobot.showfitness", new Bool(1), lDescription));
	}
	Beagle_StackTraceEndM("void NN::VizRobotEvalOp::registerParams(Beagle::System& ioSystem)");
}


/*!
 *  \brief Override the evaluate method to allow interaction with the GUI.
 *  \param ioGraph Graph to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness measure.
 */
NN::FitnessIndexed::Handle NN::VizRobotEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)
{
	Beagle_StackTraceBeginM();
  
    // Reset timer before evaluating.
    mTimer.reset();
    mCurTime = -1.0;
    return NN::RobotEvalOp::evaluate(ioGraph, ioContext);

	Beagle_StackTraceEndM("NN::FitnessIndexed::Handle NN::VizRobotEvalOp::evaluate(NN::Graph& ioGraph, Context& ioContext)");
}


/*!
 * \brief Initialize the operator by sampling the function to approximate.
 * \param ioSystem System to use to sample.
 */
void NN::VizRobotEvalOp::init(Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	if (isInitialized())
		return;

	NN::RobotEvalOp::init(ioSystem);

    // mSimulator was initialized by the call above.
    if (mRobotApp == 0) {
        mRobotApp = new RobotApp();
        mRobotApp->initialize(mSimulator, mRealTime->getWrappedValue(), mShowFitness->getWrappedValue());
        mRobotApp->render();
    }

	Beagle_StackTraceEndM("void NN::VizRobotEvalOp::init(System& ioSystem)");
}


/*! 
 *  \brief Create an agent in the environment at a specified location and scale.
 *  \param inX The x coordinate of the agent.
 *  \param inY The y coordinate of the agent.
 *  \param inZ The z coordinate of the agent.
 *  \param inScale Scale to apply to the blueprint dimensions.
 */
void NN::VizRobotEvalOp::createAgent(opal::real inX, opal::real inY, opal::real inZ, opal::real inScale)
{
	Beagle_StackTraceBeginM();

    NN::RobotEvalOp::createAgent(inX, inY, inZ, inScale);
    mRobotApp->createAgent(mAgent);
	mRobotApp->render();

	Beagle_StackTraceEndM("void NN::VizRobotEvalOp::createAgent(opal::real inX, opal::real inY, opal::real inZ, opal::real inScale)");
}


/*! 
 *  \brief Remove the agent from the environment.  
 */
void NN::VizRobotEvalOp::removeAgent()
{
	Beagle_StackTraceBeginM();

    NN::RobotEvalOp::removeAgent();
    mRobotApp->removeAgent();
    opal::real lFitness = mShowFitness->getWrappedValue() ? calculateFitness() : FLT_MAX;  // Final fitness.
    mRobotApp->render(mTime, lFitness);

	Beagle_StackTraceEndM("void NN::VizRobotEvalOp::removeAgent()");
}


/*! 
 *  \brief Create the environment.
 */
void NN::VizRobotEvalOp::createEnviron(Context& ioContext)
{
	Beagle_StackTraceBeginM();

    NN::RobotEvalOp::createEnviron(ioContext);
    mRobotApp->createEnviron(mEnviron);
	mRobotApp->render();

	Beagle_StackTraceEndM("void NN::VizRobotEvalOp::createEnviron(Context& ioContext)");
}


/*! 
 *  \brief Remove the environment.  
 */
void NN::VizRobotEvalOp::removeEnviron()
{
	Beagle_StackTraceBeginM();

    NN::RobotEvalOp::removeEnviron();
    mRobotApp->removeEnviron();
	mRobotApp->render();

	Beagle_StackTraceEndM("void NN::VizRobotEvalOp::removeEnviron()");
}


/*!
 *  \brief Simulate the agent for the given amount of time.
 *  \param inDt The amount of time to simulate.
 *
 *  Reference on how to render using a fixed time step for physics:
 *  http://www.gaffer.org/game-physics/fix-your-timestep/
 *  The trick used here is to think of time as a resource produced
 *  by rendering and consumed by physics simulation.
 */
void NN::VizRobotEvalOp::simulateAgent(opal::real inDt)
{
	Beagle_StackTraceBeginM();

    if (mRealTime->getWrappedValue()) {
        if (mCurTime < 0.0) mCurTime = mTimer.getValue();

        NN::RobotEvalOp::simulateAgent(inDt);
        opal::real lFitness = mShowFitness->getWrappedValue() ? calculateFitness() : FLT_MAX;
        mAccTime -= inDt;

        while (mAccTime <= 0.0) {
            unsigned int lRenderCount = mRobotApp->render(mTime, lFitness);

            mPrvTime = mCurTime;
            mCurTime = mTimer.getValue();

            // The elapsed time is divided by the number of frames rendered to account
            // for the case when the simulation is paused by the user.
            double lElapsedTime = mCurTime - mPrvTime;
            mAccTime += lElapsedTime/lRenderCount;
        }
    }
    else {
        NN::RobotEvalOp::simulateAgent(inDt);
        opal::real lFitness = mShowFitness->getWrappedValue() ? calculateFitness() : FLT_MAX;
        mRobotApp->render(mTime, lFitness);
    }

	Beagle_StackTraceEndM("void NN::VizRobotEvalOp::simulateAgent(opal::real inDt)");
}
