
#include <QProcess>
#include <QGraphicsSvgItem>
#include <QGraphicsTextItem>
#include <QSettings>
#include <QMessageBox>

#include "beagle/GA.hpp"
#include "explorer/ENSO.hpp"

using namespace Explorer;


ENSO::GraphSettings::GraphSettings(QWidget* inParent, Qt::WFlags inFlags) :
		DisplaySettings(inParent, inFlags),
		mSvgRenderer(0)
{
	Beagle_StackTraceBeginM();

	mView->setWhatsThis(tr("<b>The phenotype graph being evaluated.</b>") +
	                    tr("<p>Vertices can be colored based on selected parameters. ") +
	                    tr("Edge color is based on the index of its genotype tree node.</p>"));

	mZoomSlider->setValue(mZoomSlider->maximum());

	Beagle_StackTraceEndM("GraphSettings::GraphSettings(QWidget* inParent, Qt::WFlags inFlags)");
}


/*!
 *  \brief Setup the network view of the given graph.
 */
void ENSO::GraphSettings::setup(const Beagle::ENSO::Individual& inIndividual, const Beagle::Context& inContext)
{
	Beagle_StackTraceBeginM();

    const std::vector<Beagle::ENSO::Parameters::Handle>& lGraph = inIndividual.getGraph();
    const Beagle::ENSO::Tree& lTree = *(inIndividual[0]);
    unsigned int lPrototypeLeafIndex = (*Beagle::castHandleT<Beagle::GA::IntegerVector>(lGraph[0]->find("index")->second))[0];
    mPrototypeObject = new Vertex(lTree[lPrototypeLeafIndex], lGraph[0], lTree);
	DisplaySettings::setup();

	QString lDotFile = "graph.dot";
	QString lSvgFile = "graph.svg";
	std::ofstream lDotOFS(lDotFile.toStdString().c_str());
	writeGraph(lDotOFS, inIndividual, inContext);
	lDotOFS.close();

	QSettings lSettings("OpenBEAGLE", "Explorer");
	QString lCircoExe = lSettings.value("circoexe", QString("/usr/bin/circo")).toString();
	QStringList lDotArguments;
	lDotArguments << "-Tsvg" << lDotFile << "-o" << lSvgFile;
	QProcess lDotProcess;
	lDotProcess.start(lCircoExe, lDotArguments);

	if (lDotProcess.waitForStarted() == false) {
		QMessageBox::warning(this, tr("Graphviz Circo Command Failure"),
		                     tr("Make sure the path to the circo command is set correctly via the Settings menu. ") +
		                     tr("The current path setting is %1").arg(lCircoExe));
		return;
	}

	if (lDotProcess.waitForFinished() == false) {
		QMessageBox::warning(this, tr("Graphviz Circo Command Failure"),
		                     tr("The circo command exited unsuccessfully."));
		return;
	}

	// Process the SVG file produced by graphviz to make it suitable for our use.
	std::ifstream lSvgIFS(lSvgFile.toStdString().c_str());
	PACC::XML::Document lParser;
	lParser.setNoParse("title");  // prevents messing up the edge title; content is a string anyway.
	lParser.parse(lSvgIFS, lSvgFile.toStdString());
	lSvgIFS.close();

	processSvg(lParser);

	std::ostringstream lSvgOSS;
	lParser.serialize(lSvgOSS);

	// Schedule the previous svg renderer for deletion before allocating a new one.
	if (mSvgRenderer)
		mSvgRenderer->deleteLater();
	mSvgRenderer = new QSvgRenderer(QByteArray(lSvgOSS.str().c_str()), this);

	// Schedule the previous scene for deletion before allocating a new one.
	if (mScene)
		mScene->deleteLater();
	mScene = new DisplayScene(this);

	mScene->setNodeHueIndex(mNodeHueCombo->currentIndex());
	mScene->setNodeSatIndex(mNodeSatCombo->currentIndex());
	mScene->setNodeValIndex(mNodeValCombo->currentIndex());

	for (PACC::XML::Iterator lRoot=lParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/g/g"); lGroup; lGroup = lFinder.findNext()) {
			if (lGroup->getAttribute("class")=="node") {
				QString lElementId = lGroup->getAttribute("id").c_str();
				QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);
                unsigned int lVertexId = lElementId.toUInt();
                unsigned int lGraphIndex = lVertexId*lTree.getNumVertices() + lVertexId;
                unsigned int lTreeIndex = (*Beagle::castHandleT<Beagle::GA::IntegerVector>(lGraph[lGraphIndex]->find("index")->second))[0];
                Vertex::Handle lVertex = new Vertex(lTree[lTreeIndex], lGraph[lGraphIndex], lTree);
                GraphVertex* lItem = new GraphVertex(lVertex, lVertexId);
                lItem->setToolTip(lVertex->getToolTip().c_str());
                lItem->setPos(lBoundingRect.topLeft());
                if (mScene->getNodeBoundingSize() < 0.0)
                    mScene->setNodeBoundingSize(lBoundingRect.width());
                mScene->addItem(lItem);
			} else if (lGroup->getAttribute("class")=="edge") {
				QString lElementId = lGroup->getAttribute("id").c_str();
				QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);
				QGraphicsSvgItem* lItem = new QGraphicsSvgItem;
                QStringList lVertices = lElementId.split("t");
                unsigned int lGraphIndex = lVertices[1].toUInt()*lTree.getNumVertices() + lVertices[0].toUInt();
                unsigned int lTreeIndex = (*Beagle::castHandleT<Beagle::GA::IntegerVector>(lGraph[lGraphIndex]->find("index")->second))[0];
                QString lToolTip = "Tree Index = " + QString::number(lTreeIndex);
                lToolTip += "\n";
                const Beagle::ENSO::Node& lNode = lTree[lTreeIndex];
                std::ostringstream lOSS;
                for(unsigned int i=0; i<lNode.mColor.size(); i++) {
                    lOSS << (lNode.mColor[i] ? '1' : '0');
                }
                lToolTip += "Color = " + QString::fromStdString(lOSS.str());
                lItem->setToolTip(lToolTip);
				lItem->setSharedRenderer(mSvgRenderer);
				lItem->setElementId(lElementId);
				lItem->setPos(lBoundingRect.topLeft());
				mScene->addItem(lItem);
			}
		}
	}

	mView->setScene(mScene);
    mNodeHueCombo->setCurrentIndex(1);  // Display the leaf node index as the node hue.

	Beagle_StackTraceEndM("void ENSO::GraphSettings::setup(const Beagle::ENSO::Individual& inIndividual, const Beagle::Context& inContext)");
}


/*!
 *  \brief Write the given phenotype graph in the graphviz dot format.
 *  \param outStream The stream to write the output to.
 *  \param inGraph The graph to write.
 *  \param inContext The context of evolution.
 */
void ENSO::GraphSettings::writeGraph(std::ostream& outStream, const Beagle::ENSO::Individual& inIndividual, const Beagle::Context& inContext) const
{
	Beagle_StackTraceBeginM();

    const std::vector<Beagle::ENSO::Parameters::Handle>& lGraph = inIndividual.getGraph();
    const Beagle::ENSO::Tree& lTree = *(inIndividual[0]);
    const unsigned int lNumVertices = lTree.getNumVertices();
    std::vector<unsigned int> lLeaves;
    lTree.getLeaves(lLeaves);

	std::ostringstream lOSS;
	lOSS << "digraph G {" << std::endl;
	lOSS << "center=true;" << std::endl;

    // Iterate through all the vertices and edges.
	for (unsigned int i=0; i<lNumVertices; i++) {
        for (unsigned int j=0; j<lNumVertices; j++) {
            unsigned int k = i*lNumVertices + j;
            if (i == j) {   // Vertex
                lOSS << i << " [";
                lOSS << "shape=ellipse";
                lOSS << ",style=bold";
                lOSS << ",width=0.50,height=0.50,label=\"\"];" << std::endl;
            }
            else if (lGraph[k] != NULL) {          // Edge
                unsigned int lTreeIndex = (*Beagle::castHandleT<Beagle::GA::IntegerVector>(lGraph[k]->find("index")->second))[0];
                unsigned int lLeafIndex = UINT_MAX;
                for (unsigned int x = 0; x < lLeaves.size() && lLeafIndex == UINT_MAX; x++) {
                    if (lLeaves[x] == lTreeIndex) lLeafIndex = x;
                }

                bool lIsEdge = (*Beagle::castHandleT<Beagle::GA::BitString>(lGraph[k]->find("isedge")->second))[0];

                lOSS << j << " -> " << i << " [style=";
                if (lIsEdge) {
                    lOSS << "solid";
                } else {
                    lOSS << "dashed";
                }
                lOSS << ",color=\"" << getColor((double)lLeafIndex, 0.0, (double)(lLeaves.size()-1)).name().toStdString() << "\"";
                lOSS << ",style=bold";
                lOSS << ",arrowsize=1.25";
                lOSS << "];" << std::endl;  // End of edge attributes.
            }
        }
    }

	lOSS << "}" << std::endl;
	outStream << lOSS.str();

	Beagle_StackTraceEndM("void ENSO::GraphSettings::writeGraph(std::ostream& outStream, const Beagle::ENSO::Individual& inIndividual, const Beagle::Context& inContext) const");
}


/*!
 *  \brief Process the SVG produced by graphviz
 */
void ENSO::GraphSettings::processSvg(PACC::XML::Document& ioParser) const
{
	Beagle_StackTraceBeginM();

	for (PACC::XML::Iterator lRoot=ioParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);

		// Find all groups that represent nodes and edges in the svg document.
		// Add the id attribute to them to enable access by ids.
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/g/g"); lGroup; lGroup = lFinder.findNext()) {
			if (lGroup->getAttribute("class")=="node") {
				std::string lTitleString;
				for (PACC::XML::Iterator lChild=lGroup->getFirstChild(); lChild; ++lChild) {
					if (lChild->getType()==PACC::XML::eData && lChild->getValue()=="title")
						lTitleString = lChild->getFirstChild()->getValue();
				}
				lGroup->setAttribute("id", lTitleString);
			} else if (lGroup->getAttribute("class")=="edge") {
				QString lTitleString;
				for (PACC::XML::Iterator lChild=lGroup->getFirstChild(); lChild; ++lChild) {
					if (lChild->getType()==PACC::XML::eData && lChild->getValue()=="title")
						lTitleString = lChild->getFirstChild()->getValue().c_str();
				}
                lTitleString.replace("&#45;&gt;", "t");   // replace the escaped "->" with "t"
                lGroup->setAttribute("id", lTitleString.toStdString());
			}
		}
	}

	Beagle_StackTraceEndM("void ENSO::GraphSettings::processSvg(PACC::XML::Document& ioParser) const");
}


/*!
 *  \brief Determine color based on the given value, and max and min values.
 */
QColor ENSO::GraphSettings::getColor(double inValue, double inMinValue, double inMaxValue) const
{
	Beagle_StackTraceBeginM();

	double lH = QColor(Qt::blue).hueF()*(inValue - inMinValue)/(inMaxValue - inMinValue);
	double lS = 1.0;
	double lL = 0.5;

	return hsl2rgb(lH, lS, lL);

	Beagle_StackTraceEndM("QColor ENSO::GraphSettings::getColor(double inValue, double inMinValue, double inMaxValue) const");
}


/*!
 *  \brief Update view by forcing update of all nodes.
 */
void ENSO::GraphSettings::updateView()
{
	Beagle_StackTraceBeginM();

	if (mScene == 0)
		return;

	foreach (QGraphicsItem* lItem, mScene->items()) {
		GraphVertex* lGraphVertex = qgraphicsitem_cast<GraphVertex*>(lItem);
		if (lGraphVertex) {
            lGraphVertex->getDisplayObject()->update();
            lGraphVertex->update();
        }
	}

	Beagle_StackTraceEndM("void ENSO::GraphSettings::updateView()");
}
