
#include "explorer/ENSO.hpp"

using namespace Explorer;


/*!
 *  \brief Returns the static singleton instance of the InteractionData class.
 */
ENSO::InteractionData::InteractionData& ENSO::InteractionData::instance()
{
	Beagle_StackTraceBeginM();
    static ENSO::InteractionData lInteractionData;
    return lInteractionData;
	Beagle_StackTraceEndM("ENSO::InteractionData::InteractionData& ENSO::InteractionData::instance()");
}


/*!
 *  \brief Construct an interaction object.
 */
ENSO::InteractionData::InteractionData() :
        mTreeWidget(0),
        mGraphWidget(0),
        mIndividual(NULL)
{ }


/*!
 *  \brief Setup the interaction object for this class.
 */
void ENSO::InteractionData::setup(Beagle::System::Handle ioSystem)
{
	Beagle_StackTraceBeginM();

    // Create and setup the tree and graph widgets.  Since we want them to be separate windows,
    // we don't make them children of another widget.
    mTreeWidget = new TreeWidget;
    mTreeWidget->setWindowTitle("Genotype Tree");

    mGraphWidget = new GraphWidget;
    mGraphWidget->setWindowTitle("Phenotype Graph");

	Beagle_StackTraceEndM("void ENSO::InteractionData::setup(Beagle::System::Handle ioSystem)");
}


/*!
 *  \brief Create and setup the evolution status indicators for this class.
 *  \param inParent The parent widget of the new status indicator widgets.
 *  \param outStatusIndicators The new status indicator widgets are returned in this array.
 */
void ENSO::InteractionData::setupStatusIndicators(QWidget* inParent, QVector< QVector<QWidget*> >& outStatusIndicators)
{ }


/*!
 *  \brief Create and setup the display window indicators for this class.
 *  \param inParent The parent widget of the new display indicator widgets.
 *  \param outDisplayIndicators The new display indicator widgets are returned in this array.
 */
void ENSO::InteractionData::setupDisplayIndicators(QWidget* inParent, QVector<QWidget*>& outDisplayIndicators)
{
	Beagle_StackTraceBeginM();

    mWidgetCheckBoxes.push_back(new QCheckBox(inParent));
    mWidgetCheckBoxes.back()->setObjectName(QString::fromUtf8("mWidgetCheckBoxes0"));
    mWidgetCheckBoxes.back()->setText(QApplication::translate("EvolutionTab", mTreeWidget->windowTitle().toStdString().c_str(), 0, QApplication::UnicodeUTF8));
	connect(mWidgetCheckBoxes.back(), SIGNAL(toggled(bool)), mTreeWidget, SLOT(setVisible(bool)));
	connect(mTreeWidget, SIGNAL(closed()), mWidgetCheckBoxes.back(), SLOT(toggle()));
    mTreeWidget->setVisible(mWidgetCheckBoxes.back()->isChecked());
    outDisplayIndicators.push_back(mWidgetCheckBoxes.back());

    mWidgetCheckBoxes.push_back(new QCheckBox(inParent));
    mWidgetCheckBoxes.back()->setObjectName(QString::fromUtf8("mWidgetCheckBoxes0"));
    mWidgetCheckBoxes.back()->setText(QApplication::translate("EvolutionTab", mGraphWidget->windowTitle().toStdString().c_str(), 0, QApplication::UnicodeUTF8));
	connect(mWidgetCheckBoxes.back(), SIGNAL(toggled(bool)), mGraphWidget, SLOT(setVisible(bool)));
	connect(mGraphWidget, SIGNAL(closed()), mWidgetCheckBoxes.back(), SLOT(toggle()));
    mGraphWidget->setVisible(mWidgetCheckBoxes.back()->isChecked());
    outDisplayIndicators.push_back(mWidgetCheckBoxes.back());

	Beagle_StackTraceEndM("void ENSO::InteractionData::setupDisplayIndicators(QWidget* inParent, QVector<QWidget*>& outDisplayIndicators)");
}


/*! 
 *  \brief Update the status of the widgets in the evolution tab.
 *  \param inIndividual The graph of the individual being evaluated.
 *  \param inContext The context of evolution.
 */
void ENSO::InteractionData::update(const Beagle::ENSO::Individual& inIndividual, const Beagle::Context& inContext)
{
    Beagle_StackTraceBeginM();

    // Hold a pointer to the individual so that it will be available for display
    // even after its evaluation is complete.
    mIndividual = &inIndividual;

    // Update the tree display
    mTreeWidget->reset();
    mTreeWidget->setup(inIndividual, inContext);

    // Update the graph display
    mGraphWidget->reset();
    mGraphWidget->setup(inIndividual, inContext);

    Beagle_StackTraceEndM("void ENSO::InteractionData::update(const Beagle::ENSO::Individual& inIndividual, const Beagle::Context& inContext)");
}


/*!
 *  \brief Process the close event.
 */
void ENSO::InteractionData::closeEvent(QCloseEvent *inEvent)
{
	Beagle_StackTraceBeginM();

    if (mTreeWidget != 0) mTreeWidget->close();
    if (mGraphWidget != 0) mGraphWidget->close();

	Beagle_StackTraceEndM("void ENSO::InteractionData::closeEvent(QCloseEvent *inEvent)");
}
