
#include "beagle/GA.hpp"
#include "explorer/ENSO.hpp"

using namespace Explorer;


std::vector<std::string>    ENSO::Node::smDisplayNames;
std::vector<float>          ENSO::Node::smMinDisplayValues;
std::vector<float>          ENSO::Node::smMaxDisplayValues;
std::vector<unsigned int>   ENSO::Node::smLeaves;

/*!
 *  \brief Construct an Explorer node corresponding to a Beagle::ENSO::Node
 *  \param inNode The ENSO node.
 *  \param inIndex The index of the node in the genotype tree.
 *  \param inTree The genotype tree.
 */
ENSO::Node::Node(const Beagle::ENSO::Node& inNode, unsigned int inIndex, const Beagle::ENSO::Tree& inTree) :
		mDisplayValues(0),
        mNode(inNode)
{
	Beagle_StackTraceBeginM();

    // Initialize static vector values.
    if (smDisplayNames.size() == 0) {
        smDisplayNames.push_back("Tree Index");
        smDisplayNames.push_back("Leaf Index");
        smDisplayNames.push_back("Edges");

        // Min and Max of tree index.
        smMinDisplayValues.push_back(0.0);
        smMaxDisplayValues.push_back((float)(inTree.size()-1));

        // Min and Max of leaf index.
        inTree.getLeaves(smLeaves);
        smMinDisplayValues.push_back(0.0);
        smMaxDisplayValues.push_back((float)(smLeaves.size()-1));

        // Min and Max of edge encoding (-1: vertex, 0: non-edge, 1: edge).
        smMinDisplayValues.push_back(-1.0);
        smMaxDisplayValues.push_back(1.0);
    }

    // The root node will have an undefined value for Edges becuase it represents both vertices and edges.
    float lIsEdge = FLT_MAX;
    if (inIndex != 0) {
        Beagle::ENSO::Parameters::const_iterator lIter = inNode.mParameters.find("isedge");
        if (lIter == inNode.mParameters.end()) {    // Vertex
            lIsEdge = -1.0;
        }
        else {    // Edge
            lIsEdge = ((*Beagle::castHandleT<Beagle::GA::BitString>(lIter->second))[0] == true) ? 1.0 : 0.0;
        }
    }

    // The index of node in the smLeaves vector; undefined if not a leaf.
    unsigned int lLeafIndex = UINT_MAX;
    if (mNode.mChildren.size() == 0) {
        for (unsigned int i = 0; i < smLeaves.size() && lLeafIndex == UINT_MAX; i++) {
            if (smLeaves[i] == inIndex) lLeafIndex = i;
        }
    }

    mDisplayValues.push_back((float)inIndex);
    mDisplayValues.push_back(lLeafIndex == UINT_MAX ? FLT_MAX : (float)lLeafIndex);
    mDisplayValues.push_back(lIsEdge);

	Beagle_StackTraceEndM("ENSO::Node::Node(const Beagle::ENSO::Node& inNode, const Beagle::ENSO::Tree& inTree)");
}


/*!
 *  \brief Reset information.
 */
void ENSO::Node::reset()
{
	Beagle_StackTraceBeginM();

	smDisplayNames.clear();
	smMinDisplayValues.clear();
	smMaxDisplayValues.clear();
	mDisplayValues.clear();

	Beagle_StackTraceEndM("void ENSO::Node::reset()");
}
