
#include <QProcess>
#include <QGraphicsSvgItem>
#include <QGraphicsTextItem>
#include <QSettings>
#include <QMessageBox>

#include "explorer/ENSO.hpp"

using namespace Explorer;


ENSO::TreeSettings::TreeSettings(QWidget* inParent, Qt::WFlags inFlags) :
		DisplaySettings(inParent, inFlags),
		mSvgRenderer(0)
{
	Beagle_StackTraceBeginM();

	mView->setWhatsThis(tr("<b>The ENSO genotype tree being evaluated.</b>") +
	                    tr("<p>Nodes can be colored based on selected parameters.</p>"));

	mZoomSlider->setValue(mZoomSlider->maximum()/2);

	Beagle_StackTraceEndM("TreeSettings::TreeSettings(QWidget* inParent, Qt::WFlags inFlags)");
}


/*!
 *  \brief Setup the genotype tree view of the given individual.
 */
void ENSO::TreeSettings::setup(const Beagle::ENSO::Individual& inIndividual, const Beagle::Context& inContext)
{
	Beagle_StackTraceBeginM();

    const Beagle::ENSO::Tree& lTree = *(inIndividual[0]);
    mPrototypeObject = new Node(lTree[0], 0, lTree);
	DisplaySettings::setup();

	QString lDotFile = "tree.dot";
	QString lSvgFile = "tree.svg";
	std::ofstream lDotOFS(lDotFile.toStdString().c_str());
	writeTree(lDotOFS, lTree, inContext);
	lDotOFS.close();

	QSettings lSettings("OpenBEAGLE", "Explorer");
	QString lDotExe = lSettings.value("dotexe", QString("/usr/bin/dot")).toString();
	QStringList lDotArguments;
	lDotArguments << "-Tsvg" << lDotFile << "-o" << lSvgFile;
	QProcess lDotProcess;
	lDotProcess.start(lDotExe, lDotArguments);

	if (lDotProcess.waitForStarted() == false) {
		QMessageBox::warning(this, tr("Graphviz Dot Command Failure"),
		                     tr("Make sure the path to the dot command is set correctly via the Settings menu. ") +
		                     tr("The current path setting is %1").arg(lDotExe));
		return;
	}

	if (lDotProcess.waitForFinished() == false) {
		QMessageBox::warning(this, tr("Graphviz Dot Command Failure"),
		                     tr("The dot command exited unsuccessfully."));
		return;
	}

	// Process the SVG file produced by graphviz to make it suitable for our use.
	std::ifstream lSvgIFS(lSvgFile.toStdString().c_str());
	PACC::XML::Document lParser;
	lParser.setNoParse("title");  // prevents messing up the edge title; content is a string anyway.
	lParser.parse(lSvgIFS, lSvgFile.toStdString());
	lSvgIFS.close();

	processSvg(lParser);

	std::ostringstream lSvgOSS;
	lParser.serialize(lSvgOSS);

	// Schedule the previous svg renderer for deletion before allocating a new one.
	if (mSvgRenderer)
		mSvgRenderer->deleteLater();
	mSvgRenderer = new QSvgRenderer(QByteArray(lSvgOSS.str().c_str()), this);

	// Schedule the previous scene for deletion before allocating a new one.
	if (mScene)
		mScene->deleteLater();
	mScene = new DisplayScene(this);

	mScene->setNodeHueIndex(mNodeHueCombo->currentIndex());
	mScene->setNodeSatIndex(mNodeSatCombo->currentIndex());
	mScene->setNodeValIndex(mNodeValCombo->currentIndex());

	for (PACC::XML::Iterator lRoot=lParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/g/g"); lGroup; lGroup = lFinder.findNext()) {
			if (lGroup->getAttribute("class")=="node") {
				QString lElementId = lGroup->getAttribute("id").c_str();
				QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);
                unsigned int lNodeId = lElementId.toUInt();
                Node::Handle lNode = new Node(lTree[lNodeId], lNodeId, lTree);
                TreeNode* lItem = new TreeNode(lNode, lNodeId);
                lItem->setToolTip(lNode->getToolTip().c_str());
                lItem->setPos(lBoundingRect.topLeft());
                if (mScene->getNodeBoundingSize() < 0.0)
                    mScene->setNodeBoundingSize(lBoundingRect.width());
                mScene->addItem(lItem);
			} else if (lGroup->getAttribute("class")=="edge") {
				QString lElementId = lGroup->getAttribute("id").c_str();
				QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);
				QGraphicsSvgItem* lItem = new QGraphicsSvgItem;
				lItem->setSharedRenderer(mSvgRenderer);
				lItem->setElementId(lElementId);
				lItem->setPos(lBoundingRect.topLeft());
				mScene->addItem(lItem);
			}
		}
	}

	mView->setScene(mScene);

	Beagle_StackTraceEndM("void ENSO::TreeSettings::setup(const Beagle::ENSO::Individual& inIndividual, const Beagle::Context& inContext)");
}


/*!
 *  \brief Write the given genotype tree in the graphviz dot format.
 *  \param outStream The stream to write the output to.
 *  \param inTree The tree to write.
 *  \param inContext The context of evolution.
 */
void ENSO::TreeSettings::writeTree(std::ostream& outStream, const Beagle::ENSO::Tree& inTree, const Beagle::Context& inContext) const
{
	Beagle_StackTraceBeginM();

	std::ostringstream lOSS;
	lOSS << "graph G {" << std::endl;
	lOSS << "rankdir=TB; ranksep=0.80; nodesep=0.80; center=true;" << std::endl;

    // Iterate through all the nodes.
	for (unsigned int i=0; i<inTree.size(); i++) {
		const Beagle::ENSO::Node& lNode = inTree[i];

        // Print the node.
		lOSS << i << " [";
        lOSS << "shape=box";
        lOSS << ",style=bold";
		lOSS << ",width=0.50,height=0.50,label=\"\"];" << std::endl;

        // Print edges for all children.
		for (unsigned int j=0; j<lNode.mChildren.size(); j++) {
			lOSS << i << " -- " << lNode.mChildren[j] << " [style=solid,style=bold];" << std::endl;
		}
	}

	lOSS << "}" << std::endl;
	outStream << lOSS.str();

	Beagle_StackTraceEndM("void ENSO::TreeSettings::writeTree(std::ostream& outStream, const Beagle::ENSO::Tree& inTree, const Beagle::Context& inContext) const");
}


/*!
 *  \brief Process the SVG produced by graphviz
 */
void ENSO::TreeSettings::processSvg(PACC::XML::Document& ioParser) const
{
	Beagle_StackTraceBeginM();

	for (PACC::XML::Iterator lRoot=ioParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);

		// Find all groups that represent nodes and edges in the svg document.
		// Add the id attribute to them to enable access by ids.
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/g/g"); lGroup; lGroup = lFinder.findNext()) {
			if (lGroup->getAttribute("class")=="node") {
				std::string lTitleString;
				for (PACC::XML::Iterator lChild=lGroup->getFirstChild(); lChild; ++lChild) {
					if (lChild->getType()==PACC::XML::eData && lChild->getValue()=="title")
						lTitleString = lChild->getFirstChild()->getValue();
				}
				lGroup->setAttribute("id", lTitleString);
			} else if (lGroup->getAttribute("class")=="edge") {
				QString lTitleString;
				for (PACC::XML::Iterator lChild=lGroup->getFirstChild(); lChild; ++lChild) {
					if (lChild->getType()==PACC::XML::eData && lChild->getValue()=="title")
						lTitleString = lChild->getFirstChild()->getValue().c_str();
				}
                lTitleString.replace("&#45;&#45;", "t");   // replace the escaped "--" with "t"
                lGroup->setAttribute("id", lTitleString.toStdString());
			}
		}
	}

	Beagle_StackTraceEndM("void ENSO::TreeSettings::processSvg(PACC::XML::Document& ioParser) const");
}


/*!
 *  \brief Update view by forcing update of all nodes.
 */
void ENSO::TreeSettings::updateView()
{
	Beagle_StackTraceBeginM();

	if (mScene == 0)
		return;

	foreach (QGraphicsItem* lItem, mScene->items()) {
		TreeNode* lTreeNode = qgraphicsitem_cast<TreeNode*>(lItem);
		if (lTreeNode) {
            lTreeNode->getDisplayObject()->update();
            lTreeNode->update();
        }
	}

	Beagle_StackTraceEndM("void ENSO::TreeSettings::updateView()");
}
