
#include "explorer/NN.hpp"

using namespace Explorer;


/*!
 *  \brief Returns the static singleton instance of the InteractionData class.
 */
NN::InteractionData::InteractionData& NN::InteractionData::instance()
{
	Beagle_StackTraceBeginM();
    static NN::InteractionData lInteractionData;
    return lInteractionData;
	Beagle_StackTraceEndM("NN::InteractionData::InteractionData& NN::InteractionData::instance()");
}


/*!
 *  \brief Construct an interaction object.
 */
NN::InteractionData::InteractionData() :
        mLabelActivationCount(0),
        mCurrentActivationCount(0),
        mBreakActivationCount(0),
        mNetworkWidget(0)
{ }


/*!
 *  \brief Setup the interaction object for this class.
 */
void NN::InteractionData::setup(Beagle::System::Handle ioSystem)
{
	Beagle_StackTraceBeginM();

    // Create and setup the NetworkWidget.  Since we want it to be a separate window,
    // we don't make it a child of another widget.
    mNetworkWidget = new NetworkWidget;
    mNetworkWidget->setWindowTitle("Neural Network");

    // Setup data plots.
    Explorer::NN::PlotData::Handle lPlotData = Beagle::castHandleT<Explorer::NN::PlotData>(ioSystem->getComponent("NN-PlotData"));
    if (lPlotData != NULL) {
        QVector<Explorer::NN::PlotData::Data>& lDataVector = lPlotData->getDataVector();

        for (unsigned int i = 0; i < lDataVector.size(); i++) {
            Plot* lPlot = new Plot(lDataVector[i]);
            lPlot->setWindowTitle(lDataVector[i].mTitle);
            lPlot->resize(600,400);
            mPlotWidgets.push_back(lPlot);
        }
    }

	Beagle_StackTraceEndM("void NN::InteractionData::setup(Beagle::System::Handle ioSystem)");
}


/*!
 *  \brief Create and setup the evolution status indicators for this class.
 *  \param inParent The parent widget of the new status indicator widgets.
 *  \param outStatusIndicators The new status indicator widgets are returned in this array.
 */
void NN::InteractionData::setupStatusIndicators(QWidget* inParent, QVector< QVector<QWidget*> >& outStatusIndicators)
{
	Beagle_StackTraceBeginM();

    outStatusIndicators.push_back(QVector<QWidget*>());

    mLabelActivationCount = new QLabel(inParent);
    mLabelActivationCount->setObjectName(QString::fromUtf8("mLabelActivationCount"));
    mLabelActivationCount->setText(QApplication::translate("EvolutionTab", "&Activation Count", 0, QApplication::UnicodeUTF8));
    outStatusIndicators.back().push_back(mLabelActivationCount);

    mCurrentActivationCount = new QLabel(inParent);
    mCurrentActivationCount->setObjectName(QString::fromUtf8("mCurrentActivationCount"));
    mCurrentActivationCount->setText(QString());
    outStatusIndicators.back().push_back(mCurrentActivationCount);

    mBreakActivationCount = new QSpinBox(inParent);
    mBreakActivationCount->setObjectName(QString::fromUtf8("mBreakActivationCount"));
    mBreakActivationCount->setMaximum(99999);
    outStatusIndicators.back().push_back(mBreakActivationCount);

    mLabelActivationCount->setBuddy(mBreakActivationCount);

	Beagle_StackTraceEndM("void NN::InteractionData::setupStatusIndicators(QWidget* inParent, QVector< QVector<QWidget*> >& outStatusIndicators)");
}


/*!
 *  \brief Create and setup the display window indicators for this class.
 *  \param inParent The parent widget of the new display indicator widgets.
 *  \param outDisplayIndicators The new display indicator widgets are returned in this array.
 */
void NN::InteractionData::setupDisplayIndicators(QWidget* inParent, QVector<QWidget*>& outDisplayIndicators)
{
	Beagle_StackTraceBeginM();

    mWidgetCheckBoxes.push_back(new QCheckBox(inParent));
    mWidgetCheckBoxes.back()->setObjectName(QString::fromUtf8("mWidgetCheckBoxes0"));
    mWidgetCheckBoxes.back()->setText(QApplication::translate("EvolutionTab", mNetworkWidget->windowTitle().toStdString().c_str(), 0, QApplication::UnicodeUTF8));
	connect(mWidgetCheckBoxes.back(), SIGNAL(toggled(bool)), mNetworkWidget, SLOT(setVisible(bool)));
	connect(mNetworkWidget, SIGNAL(closed()), mWidgetCheckBoxes.back(), SLOT(toggle()));
    mNetworkWidget->setVisible(mWidgetCheckBoxes.back()->isChecked());
    outDisplayIndicators.push_back(mWidgetCheckBoxes.back());

    for (unsigned int i = 0; i < mPlotWidgets.size(); i++) {
        mWidgetCheckBoxes.push_back(new QCheckBox(inParent));
        mWidgetCheckBoxes.back()->setObjectName(QString::fromUtf8("mWidgetCheckBoxes") + QString::number(i+1));
        mWidgetCheckBoxes.back()->setText(QApplication::translate("EvolutionTab", mPlotWidgets[i]->windowTitle().toStdString().c_str(), 0, QApplication::UnicodeUTF8));
        connect(mWidgetCheckBoxes.back(), SIGNAL(toggled(bool)), mPlotWidgets[i], SLOT(setVisible(bool)));
        connect(mPlotWidgets[i], SIGNAL(closed()), mWidgetCheckBoxes.back(), SLOT(toggle()));
        mPlotWidgets[i]->setVisible(mWidgetCheckBoxes.back()->isChecked());
        outDisplayIndicators.push_back(mWidgetCheckBoxes.back());
    }

	Beagle_StackTraceEndM("void NN::InteractionData::setupDisplayIndicators(QWidget* inParent, QVector<QWidget*>& outDisplayIndicators)");
}


/*!
 *  \brief Test if evolution should be paused based on breakpoint.
 *  \param inGraph The graph of the individual being evaluated.
 */
bool NN::InteractionData::testBreak(const Beagle::NN::Graph& inGraph)
{
	Beagle_StackTraceBeginM();

    unsigned int lActivationCount = inGraph.getActivationCount();
    int lBreakActivationCount = mBreakActivationCount->value();

    return (lActivationCount >= lBreakActivationCount);

	Beagle_StackTraceEndM("bool NN::InteractionData::testBreak(const Beagle::NN::Graph& inGraph)");
}


/*! 
 *  \brief Update the status of the widgets in the evolution tab.
 *  \param inGraph The graph of the individual being evaluated.
 *  \param inContext The context of evolution.
 */
void NN::InteractionData::update(const Beagle::NN::Graph& inGraph, const Beagle::Context& inContext)
{
    Beagle_StackTraceBeginM();

    // Update the network display
    if (inGraph.getActivationCount() == 0) {
        mNetworkWidget->reset();
        mNetworkWidget->setup(inGraph, inContext);
    }
    else {
        mNetworkWidget->updateView();
    }

    // Update the data plots
    Explorer::NN::PlotData::Handle lPlotData = Beagle::castHandleT<Explorer::NN::PlotData>(inContext.getSystem().getComponent("NN-PlotData"));
    if (lPlotData != NULL && inGraph.getActivationCount() > 0) {
        QVector<Explorer::NN::PlotData::Data>& lDataVector = lPlotData->getDataVector();

        for (unsigned int i = 0; i < lDataVector.size(); i++) {
            mPlotWidgets[i]->updateView(lDataVector[i]);
        }
    }

    mCurrentActivationCount->setText(QString::number(inGraph.getActivationCount()));

    Beagle_StackTraceEndM("void NN::InteractionData::update(const Beagle::NN::Graph& inGraph, const Beagle::Context& inContext)");
}


/*!
 *  \brief Process the close event.
 */
void NN::InteractionData::closeEvent(QCloseEvent *inEvent)
{
	Beagle_StackTraceBeginM();

    if (mNetworkWidget != 0) mNetworkWidget->close();
    for (unsigned int i = 0; i < mPlotWidgets.size(); i++) {
        mPlotWidgets[i]->close();
    }

	Beagle_StackTraceEndM("void NN::InteractionData::closeEvent(QCloseEvent *inEvent)");
}
