
#include <QProcess>
#include <QGraphicsSvgItem>
#include <QGraphicsTextItem>
#include <QSettings>
#include <QMessageBox>

#include "explorer/NN.hpp"

using namespace Explorer;


NN::NetworkSettings::NetworkSettings(QWidget* inParent, Qt::WFlags inFlags) :
		DisplaySettings(inParent, inFlags),
		mSvgRenderer(0)
{
	Beagle_StackTraceBeginM();

	mView->setWhatsThis(tr("<b>The network being evaluated.</b>") +
	                    tr("<p>Nodes can be colored based on selected parameters. ") +
	                    tr("Edge color is <font color=#0000FF>blue</font> if weight is negative, ") +
	                    tr("is <font color=#FF0000>red</font> if weight is positive, ") +
	                    tr("and intensity of the color indicates the magnitude of the weight.</p>"));

	mZoomSlider->setValue(mZoomSlider->maximum()/2);

	Beagle_StackTraceEndM("NetworkSettings::NetworkSettings(QWidget* inParent, Qt::WFlags inFlags)");
}


/*!
 *  \brief Setup the network view of the given graph.
 */
void NN::NetworkSettings::setup(const Beagle::NN::Graph& inGraph, const Beagle::Context& inContext)
{
	Beagle_StackTraceBeginM();

    mPrototypeObject = new Node(inGraph.begin()->second, inGraph);
	DisplaySettings::setup();

	QString lDotFile = "network.dot";
	QString lSvgFile = "network.svg";
	std::ofstream lDotOFS(lDotFile.toStdString().c_str());
	writeGraph(lDotOFS, inGraph, inContext);
	lDotOFS.close();

	QSettings lSettings("OpenBEAGLE", "Explorer");
	QString lDotExe = lSettings.value("dotexe", QString("/usr/bin/dot")).toString();
	QStringList lDotArguments;
	lDotArguments << "-Tsvg" << lDotFile << "-o" << lSvgFile;
	QProcess lDotProcess;
	lDotProcess.start(lDotExe, lDotArguments);

	if (lDotProcess.waitForStarted() == false) {
		QMessageBox::warning(this, tr("Graphviz Dot Command Failure"),
		                     tr("Make sure the path to the dot command is set correctly via the Settings menu. ") +
		                     tr("The current path setting is %1").arg(lDotExe));
		return;
	}

	if (lDotProcess.waitForFinished() == false) {
		QMessageBox::warning(this, tr("Graphviz Dot Command Failure"),
		                     tr("The dot command exited unsuccessfully."));
		return;
	}

	// Process the SVG file produced by graphviz to make it suitable for our use.
	std::ifstream lSvgIFS(lSvgFile.toStdString().c_str());
	PACC::XML::Document lParser;
	lParser.setNoParse("title");  // prevents messing up the edge title; content is a string anyway.
	lParser.parse(lSvgIFS, lSvgFile.toStdString());
	lSvgIFS.close();

	processSvg(lParser);

	std::ostringstream lSvgOSS;
	lParser.serialize(lSvgOSS);

	// Schedule the previous svg renderer for deletion before allocating a new one.
	if (mSvgRenderer)
		mSvgRenderer->deleteLater();
	mSvgRenderer = new QSvgRenderer(QByteArray(lSvgOSS.str().c_str()), this);

	// Schedule the previous scene for deletion before allocating a new one.
	if (mScene)
		mScene->deleteLater();
	mScene = new DisplayScene(this);

	mScene->setNodeHueIndex(mNodeHueCombo->currentIndex());
	mScene->setNodeSatIndex(mNodeSatCombo->currentIndex());
	mScene->setNodeValIndex(mNodeValCombo->currentIndex());

	for (PACC::XML::Iterator lRoot=lParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/g/g"); lGroup; lGroup = lFinder.findNext()) {
			if (lGroup->getAttribute("class")=="node") {
				QString lElementId = lGroup->getAttribute("id").c_str();
				QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);
                unsigned int lNodeId = lElementId.toUInt();
                Node::Handle lNode = new Node(inGraph.getNode(lNodeId), inGraph);
                NetworkNode* lItem = new NetworkNode(lNode);
                lItem->setToolTip(lNode->getToolTip().c_str());
                lItem->setPos(lBoundingRect.topLeft());
                if (mScene->getNodeBoundingSize() < 0.0)
                    mScene->setNodeBoundingSize(lBoundingRect.width());
                mScene->addItem(lItem);
			} else if (lGroup->getAttribute("class")=="edge") {
				QString lElementId = lGroup->getAttribute("id").c_str();
				QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);
				QGraphicsSvgItem* lItem = new QGraphicsSvgItem;
				lItem->setSharedRenderer(mSvgRenderer);
				lItem->setElementId(lElementId);
				lItem->setPos(lBoundingRect.topLeft());
				mScene->addItem(lItem);
			}
		}
	}

	mView->setScene(mScene);

	Beagle_StackTraceEndM("void NN::NetworkSettings::setup(const Beagle::NN::Graph& inGraph, const Beagle::Context& inContext)");
}


/*!
 *  \brief Write the given genotype graph in the graphviz dot format.
 *  \param outStream The stream to write the output to.
 *  \param inGraph The graph to write.
 *  \param inContext The context of evolution.
 */
void NN::NetworkSettings::writeGraph(std::ostream& outStream, const Beagle::NN::Graph& inGraph, const Beagle::Context& inContext) const
{
	Beagle_StackTraceBeginM();

	Beagle::NN::GraphState::Handle lGraphState = Beagle::castHandleT<Beagle::NN::GraphState>(inContext.getSystem().getComponent("NN-GraphState"));

    // Find the min and max edge weights.
	double lEdgeMaxVal = -DBL_MAX;
	double lEdgeMinVal = DBL_MAX;
	for (Beagle::NN::Graph::const_iterator lGraphIter = inGraph.begin(); lGraphIter != inGraph.end(); ++lGraphIter) {
		const Beagle::NN::Node& lNode = lGraphIter->second;
		unsigned int lNodeId = lNode.getIdentifier();
		for (Beagle::NN::Node::const_iterator lNodeIter = lNode.begin(); lNodeIter != lNode.end(); ++lNodeIter) {
			const Beagle::NN::Edge& lEdge = lNodeIter->second;
			unsigned int lSourceId = lEdge.getSourceIdentifier();
            double lWeight = inGraph.getNode(lNodeId).getEdge(lSourceId).getWeight();
            if (lWeight < lEdgeMinVal) lEdgeMinVal = lWeight;
            if (lWeight > lEdgeMaxVal) lEdgeMaxVal = lWeight;
        }
    }

	std::ostringstream lOSSHead;
	std::ostringstream lOSSBody;
	std::ostringstream lOSSTail;
	std::ostringstream lOSSSources;
	std::ostringstream lOSSSinks;

	lOSSHead << "digraph G {" << std::endl;
	lOSSHead << "rankdir=LR; ranksep=0.80; nodesep=0.80; center=true;" << std::endl;
	lOSSSources << "{ rank = source; ";
	lOSSSinks << "{ rank = sink; ";

    // Iterate through all the nodes.
	for (Beagle::NN::Graph::const_iterator lGraphIter = inGraph.begin(); lGraphIter != inGraph.end(); ++lGraphIter) {
		const Beagle::NN::Node& lNode = lGraphIter->second;
		unsigned int lNodeId = lNode.getIdentifier();

		// Shape of the node is circle (computational) or square (non-computational).
		lOSSBody << lNodeId << " [";
		if (lGraphState->isInput(lNodeId)) {  // Input node
			lOSSSources << lNodeId << "; ";
			lOSSBody << "shape=ellipse";
		} else if (lGraphState->isOutput(lNodeId)) {  // Output node
			lOSSSinks << lNodeId << "; ";
			lOSSBody << "shape=ellipse";
		} else {  // Hidden node
			lOSSBody << "shape=ellipse";
		}

        // Color of the node is based on its function.
        if (lGraphState->isInput(lNodeId))
			lOSSBody << ",color=\"#0000FF\"";
        else if (lGraphState->isOutput(lNodeId))
			lOSSBody << ",color=\"#FF0000\"";
		else if (lGraphState->isHelper(lNodeId))
			lOSSBody << ",color=\"#00FF00\"";

        lOSSBody << ",style=bold";
		lOSSBody << ",width=0.50,height=0.50,label=\"\"];" << std::endl;  // End of node attributes.

        // Iterate through all the edges.
		for (Beagle::NN::Node::const_iterator lNodeIter = lNode.begin(); lNodeIter != lNode.end(); ++lNodeIter) {
			const Beagle::NN::Edge& lEdge = lNodeIter->second;
			unsigned int lSourceId = lEdge.getSourceIdentifier();

			// Style of the edge depends on whether the edge is recurrent.
			lOSSBody << lSourceId << " -> " << lNodeId << " [style=";
			if (lEdge.isRecurrent()) {
				lOSSBody << "dashed";
			} else {
				lOSSBody << "solid";
			}

            double lWeight = inGraph.getNode(lNodeId).getEdge(lSourceId).getWeight();
			lOSSBody << ",color=\"" << getColor(lWeight, lEdgeMinVal, lEdgeMaxVal).name().toStdString() << "\"";
            lOSSBody << ",style=bold";
            lOSSBody << ",arrowsize=1.25";
			lOSSBody << "];" << std::endl;  // End of edge attributes.
		}
	}

	lOSSSources << "}" << std::endl;
	lOSSSinks << "}" << std::endl;
	lOSSTail << "}" << std::endl;

	// Write the different sections to the output stream.
	outStream << lOSSHead.str();
	outStream << lOSSSources.str();
	outStream << lOSSSinks.str();
	outStream << lOSSBody.str();
	outStream << lOSSTail.str();

	Beagle_StackTraceEndM("void NN::NetworkSettings::writeGraph(std::ostream& outStream, const Beagle::NN::Graph& inGraph, const Beagle::Context& inContext) const");
}


/*!
 *  \brief Process the SVG produced by graphviz
 */
void NN::NetworkSettings::processSvg(PACC::XML::Document& ioParser) const
{
	Beagle_StackTraceBeginM();

	for (PACC::XML::Iterator lRoot=ioParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);

		// Find all groups that represent nodes and edges in the svg document.
		// Add the id attribute to them to enable access by ids.
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/g/g"); lGroup; lGroup = lFinder.findNext()) {
			if (lGroup->getAttribute("class")=="node") {
				std::string lTitleString;
				for (PACC::XML::Iterator lChild=lGroup->getFirstChild(); lChild; ++lChild) {
					if (lChild->getType()==PACC::XML::eData && lChild->getValue()=="title")
						lTitleString = lChild->getFirstChild()->getValue();
				}
				lGroup->setAttribute("id", lTitleString);
			} else if (lGroup->getAttribute("class")=="edge") {
				QString lTitleString;
				for (PACC::XML::Iterator lChild=lGroup->getFirstChild(); lChild; ++lChild) {
					if (lChild->getType()==PACC::XML::eData && lChild->getValue()=="title")
						lTitleString = lChild->getFirstChild()->getValue().c_str();
				}
                lTitleString.replace("&#45;&gt;", "t");   // replace the escaped "->" with "t"
                lGroup->setAttribute("id", lTitleString.toStdString());
			}
		}
	}

	Beagle_StackTraceEndM("void NN::NetworkSettings::processSvg(PACC::XML::Document& ioParser) const");
}


/*!
 *  \brief Determine color based on the given value, and max and min values.
 */
QColor NN::NetworkSettings::getColor(double inValue, double inMinValue, double inMaxValue) const
{
	Beagle_StackTraceBeginM();

	double lNormL = (inValue < 0) ? inMinValue : inMaxValue;
	double lH = (inValue < 0) ? 0.66 : 0.0;
	double lS = 1.0;
	double lL = inValue/(2.0*lNormL);

	return hsl2rgb(lH, lS, lL);

	Beagle_StackTraceEndM("QColor NN::NetworkSettings::getColor(double inValue, double inMinValue, double inMaxValue) const");
}


/*!
 *  \brief Update view by forcing update of all nodes.
 */
void NN::NetworkSettings::updateView()
{
	Beagle_StackTraceBeginM();

	if (mScene == 0)
		return;

	foreach (QGraphicsItem* lItem, mScene->items()) {
		NetworkNode* lNetworkNode = qgraphicsitem_cast<NetworkNode*>(lItem);
		if (lNetworkNode) {
            lNetworkNode->getDisplayObject()->update();
            lNetworkNode->update();
        }
	}

	Beagle_StackTraceEndM("void NN::NetworkSettings::updateView()");
}
