
#include <QProcess>
#include <QGraphicsSvgItem>
#include <QGraphicsTextItem>
#include <QSettings>
#include <QMessageBox>

#include "explorer/SN.hpp"

using namespace Explorer;


SN::NetworkSettings::NetworkSettings(QWidget* inParent, Qt::WFlags inFlags) :
		DisplaySettings(inParent, inFlags),
		mSvgRenderer(0)
{
	Beagle_StackTraceBeginM();

	mView->setWhatsThis(tr("<b>The network being evaluated.</b>") +
	                    tr("<p>Horizontal lines are input lines and vertical lines are comparators. ") +
	                    tr("The columns of box arrays show the sets associated with the monotone Boolean functions ") +
	                    tr("at each stage of the network.  Each array has the same number of boxes as inputs and ") +
	                    tr("represents the set cardinalities of the function.  If the box is colored, then the ") +
	                    tr("corresponding set has non-zero elements; otherwise the set has zero elements.</p>"));

	mZoomSlider->setValue(mZoomSlider->maximum()/2);

	Beagle_StackTraceEndM("NetworkSettings::NetworkSettings(QWidget* inParent, Qt::WFlags inFlags)");
}


/*!
 *  \brief Scale the view by the value set by the zoom slider.
 */
void SN::NetworkSettings::scale(int inScale)
{
	Beagle_StackTraceBeginM();

	qreal lScale = 5*inScale/(qreal)(mZoomSlider->maximum());
	QMatrix lMatrix;
	lMatrix.scale(lScale, lScale);
	mView->setMatrix(lMatrix);

	Beagle_StackTraceEndM("void SN::NetworkSettings::scale(int inScale)");
}


/*!
 *  \brief Setup the network view of the given graph.
 */
void SN::NetworkSettings::setup(const Beagle::SN::Network& inNetwork, const Beagle::Context& inContext)
{
	Beagle_StackTraceBeginM();

    Beagle::SN::Function lPrototypeFunction;
    lPrototypeFunction.setInput(0);
    mPrototypeObject = new Function(lPrototypeFunction);
	DisplaySettings::setup();

	QString lXmlFile = "network.xml";
	QString lSvgFile = "network.svg";
	std::ofstream lXmlOFS(lXmlFile.toStdString().c_str());
    PACC::XML::Streamer lStreamer(lXmlOFS);
    inNetwork.write(lStreamer, true);
    lXmlOFS << std::endl;
	lXmlOFS.close();

    // The sorting network drawing program is based on the CPAN module Algorithm::Networksort
    // by John Gamble.
	QSettings lSettings("OpenBEAGLE", "Explorer");
	QString lDrawExe = lSettings.value("sndrawexe", QString("/usr/bin/sndraw")).toString();
	QStringList lDrawArguments;
	lDrawArguments << "--stroke_width=1" << "--vt_sep=20" << "--hz_sep=30" << "--indent=30" << lXmlFile;
	QProcess lDrawProcess;
    lDrawProcess.setStandardOutputFile(lSvgFile);
	lDrawProcess.start(lDrawExe, lDrawArguments);

	if (lDrawProcess.waitForStarted() == false) {
		QMessageBox::warning(this, tr("Sorting Network Draw Command Failure"),
		                     tr("Make sure the path to the sorting network draw command is set correctly via the Settings menu. ") +
		                     tr("The current path setting is %1").arg(lDrawExe));
		return;
	}

	if (lDrawProcess.waitForFinished() == false) {
		QMessageBox::warning(this, tr("Sorting Network Draw Command Failure"),
		                     tr("The sorting network draw command exited unsuccessfully."));
		return;
	}

	// Process the SVG file produced above to make it suitable for our use.
	std::ifstream lSvgIFS(lSvgFile.toStdString().c_str());
	PACC::XML::Document lParser;
	lParser.setNoParse("title");  // prevents messing up the edge title; content is a string anyway.
	lParser.parse(lSvgIFS, lSvgFile.toStdString());
	lSvgIFS.close();

	std::ostringstream lSvgOSS;
	lParser.serialize(lSvgOSS);

	// Schedule the previous svg renderer for deletion before allocating a new one.
	if (mSvgRenderer)
		mSvgRenderer->deleteLater();
	mSvgRenderer = new QSvgRenderer(QByteArray(lSvgOSS.str().c_str()), this);

	// Schedule the previous scene for deletion before allocating a new one.
	if (mScene)
		mScene->deleteLater();
	mScene = new DisplayScene(this);

	mScene->setNodeHueIndex(mNodeHueCombo->currentIndex());
	mScene->setNodeSatIndex(mNodeSatCombo->currentIndex());
	mScene->setNodeValIndex(mNodeValCombo->currentIndex());

    const unsigned int lNumInputs = inNetwork.getNumInputs();

    // Initialize functions for computing satellite data.
    std::vector<Beagle::SN::Function> lFunctions(lNumInputs);
    for (unsigned int i = 0; i < lNumInputs; i++) {
        lFunctions[i].setInput(i);
    }

    // Render the input lines.
    for (unsigned int i = 0; i < lNumInputs; i++) {
        QString lElementId = QString("i") + QString::number(i);
        QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);

        QGraphicsSvgItem* lItem = new QGraphicsSvgItem;
        lItem->setSharedRenderer(mSvgRenderer);
        lItem->setElementId(lElementId);
        lItem->setPos(lBoundingRect.topLeft());
        mScene->addItem(lItem);
    }

    // Render satellite data of the inputs.
    for (unsigned int i = 0; i < lNumInputs; i++) {
        QString lElementId = QString("i") + QString::number(i);
        QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);

        // Set size of the node for the satellite data.
        if (mScene->getNodeBoundingSize() < 0.0) {
            mScene->setNodeBoundingSize(lBoundingRect.height());
        }

        Function::Handle lDisplayFunc = new Function(lFunctions[i]);
        FunctionNode* lNodeItem = new FunctionNode(lDisplayFunc);
        lNodeItem->setToolTip(lDisplayFunc->getToolTip().c_str());
        lNodeItem->setPos(lBoundingRect.bottomLeft() + QPointF(lBoundingRect.height(), 0));
        mScene->addItem(lNodeItem);
    }

    // Render the comparators and satellite data.  The comparators are drawn by
    // columns, and when comparators on each column are rendered, we render the
    // satellite data on the other lines also.  For this, we need to iterate
    // through the comparators in the order they appear in the XML file, and
    // keep track of which lines the comparators in the current column touch.
    std::vector<bool> lCurrentLines(lNumInputs, false);
    qreal lCurrentx = mSvgRenderer->boundsOnElement(QString("c") + QString::number(0)).topLeft().x();
	for (PACC::XML::Iterator lRoot=lParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/use"); lGroup; lGroup = lFinder.findNext()) {
            QString lElementId = lGroup->getAttribute("id").c_str();
            // Proceed only if the element is a comparator.
            if (lElementId[0] == QChar('c')) {
                QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);

                // Render satellite data for the other lines in the previous column if we
                // are starting a new column.
                if (lCurrentx < lBoundingRect.topLeft().x()) {
                    for (unsigned int i = 0; i < lNumInputs; i++) {
                        if (lCurrentLines[i] == false) {
                            Function::Handle lDisplayFunc = new Function(lFunctions[i]);
                            FunctionNode* lNodeItem = new FunctionNode(lDisplayFunc);
                            lNodeItem->setToolTip(lDisplayFunc->getToolTip().c_str());
                            QRectF lLineBoundingRect = mSvgRenderer->boundsOnElement(QString("i") + QString::number(i));
                            qreal x = lCurrentx + lLineBoundingRect.height();
                            qreal y = lLineBoundingRect.bottomLeft().y();
                            lNodeItem->setPos(QPointF(x, y));
                            mScene->addItem(lNodeItem);
                        }
                    }

                    lCurrentLines.clear();
                    lCurrentLines.resize(lNumInputs, false);
                    lCurrentx = lBoundingRect.topLeft().x();
                }

                // Render the comparator from the SVG file.
                QGraphicsSvgItem* lItem = new QGraphicsSvgItem;
                lItem->setSharedRenderer(mSvgRenderer);
                lItem->setElementId(lElementId);
                lItem->setPos(lBoundingRect.topLeft());
                mScene->addItem(lItem);

                // Remove the character 'c' from the beginning of the id and convert
                // it to the comparator index.
                unsigned int lIndex = lElementId.remove(0, 1).toUInt();
                unsigned int lLine1 = inNetwork[2*lIndex];
                unsigned int lLine2 = inNetwork[2*lIndex+1];
                lCurrentLines[lLine1] = true;
                lCurrentLines[lLine2] = true;

                // Compute the meet and the join.
                Beagle::SN::Function lJoinFunc, lMeetFunc;
                join(lFunctions[lLine1], lFunctions[lLine2], lJoinFunc);
                meet(lFunctions[lLine1], lFunctions[lLine2], lMeetFunc);
                lFunctions[lLine1] = lJoinFunc;
                lFunctions[lLine2] = lMeetFunc;

                // Render satellite data of the comparator.
                Function::Handle lDisplayFunc = new Function(lFunctions[lLine1]);
                FunctionNode* lNodeItem = new FunctionNode(lDisplayFunc);
                lNodeItem->setToolTip(lDisplayFunc->getToolTip().c_str());
                lNodeItem->setPos(lBoundingRect.topLeft() + QPointF(lBoundingRect.width(), lBoundingRect.width()));
                mScene->addItem(lNodeItem);

                lDisplayFunc = new Function(lFunctions[lLine2]);
                lNodeItem = new FunctionNode(lDisplayFunc);
                lNodeItem->setToolTip(lDisplayFunc->getToolTip().c_str());
                lNodeItem->setPos(lBoundingRect.bottomLeft() + QPointF(lBoundingRect.width(), 0));
                mScene->addItem(lNodeItem);
            }
        }
    }

    // Render satellite data for the other lines in the last column.
    for (unsigned int i = 0; i < lNumInputs; i++) {
        if (lCurrentLines[i] == false) {
            Function::Handle lDisplayFunc = new Function(lFunctions[i]);
            FunctionNode* lNodeItem = new FunctionNode(lDisplayFunc);
            lNodeItem->setToolTip(lDisplayFunc->getToolTip().c_str());
            QRectF lLineBoundingRect = mSvgRenderer->boundsOnElement(QString("i") + QString::number(i));
            qreal x = lCurrentx + lLineBoundingRect.height();
            qreal y = lLineBoundingRect.bottomLeft().y();
            lNodeItem->setPos(QPointF(x, y));
            mScene->addItem(lNodeItem);
        }
    }

	mView->setScene(mScene);

	Beagle_StackTraceEndM("void SN::NetworkSettings::setup(const Beagle::SN::Graph& inGraph, const Beagle::Context& inContext)");
}


/*!
 *  \brief Update view by forcing update of all nodes.
 */
void SN::NetworkSettings::updateView()
{
	Beagle_StackTraceBeginM();

	if (mScene == 0)
		return;

	foreach (QGraphicsItem* lItem, mScene->items()) {
		FunctionNode* lFunctionNode = qgraphicsitem_cast<FunctionNode*>(lItem);
		if (lFunctionNode) {
            lFunctionNode->getDisplayObject()->update();
            lFunctionNode->update();
        }
	}

	Beagle_StackTraceEndM("void SN::NetworkSettings::updateView()");
}
