
#include <QtGui>

#include "designer/ConfigurationTree.hpp"


ConfigurationTree::ConfigurationTree(QWidget* inParent) :
		QTreeWidget(inParent)
{
    setColumnCount(2);

	QStringList lLabels;
	lLabels << QString() << QString();

    header()->setResizeMode(QHeaderView::Stretch);
	setHeaderLabels(lLabels);

	mElementIcon.addFile(QString::fromUtf8(":/images/tango/folder.svg"), QSize(), QIcon::Normal, QIcon::Off);
	mElementIcon.addFile(QString::fromUtf8(":/images/tango/folder-open.svg"), QSize(), QIcon::Normal, QIcon::On);
	mAttributeIcon.addFile(QString::fromUtf8(":/images/tango/document-properties.svg"));
}


bool ConfigurationTree::read(QIODevice* inDevice)
{
	QString lErrorStr;
	int lErrorLine;
	int lErrorColumn;

	if (!mDomDocument.setContent(inDevice, true, &lErrorStr, &lErrorLine, &lErrorColumn)) {
		QMessageBox::information(window(), tr("OpenBEAGLE Explorer"),
		                         tr("Parse error at line %1, column %2:\n%3")
		                         .arg(lErrorLine)
		                         .arg(lErrorColumn)
		                         .arg(lErrorStr));
		return false;
	}

	QDomElement lRoot = mDomDocument.documentElement();
	if (lRoot.tagName() != "Beagle") {
		QMessageBox::information(window(), tr("OpenBEAGLE Explorer"),
		                         tr("The file is not an OpenBEAGLE configuration file."));
		return false;
	}

	clear();

	disconnect(this, SIGNAL(itemChanged(QTreeWidgetItem*, int)),
	           this, SLOT(updateDomElement(QTreeWidgetItem*, int)));

	QDomElement lChild = lRoot.firstChildElement();
	while (!lChild.isNull()) {
		parseDomElement(lChild);
		lChild = lChild.nextSiblingElement();
	}

	connect(this, SIGNAL(itemChanged(QTreeWidgetItem*, int)),
	        this, SLOT(updateDomElement(QTreeWidgetItem*, int)));

	return true;
}


void ConfigurationTree::parseDomElement(const QDomElement &inElement, QTreeWidgetItem* inParentItem)
{
	QTreeWidgetItem* lElemItem = createItem(inElement, inParentItem);

	lElemItem->setFlags(lElemItem->flags() | Qt::ItemIsEditable);
	lElemItem->setIcon(0, mElementIcon);
	lElemItem->setText(0, inElement.tagName());

	// Elements with string content only.
	if (inElement.firstChildElement().isNull() && inElement.firstChild().isText()) {
		QString lText = inElement.text();
		lElemItem->setText(1, lText);
	}

	lElemItem->setExpanded(false);

	QDomNamedNodeMap lAttributes = inElement.attributes();
	for (unsigned int i=0; i < lAttributes.length(); ++i) {
		QDomAttr lAttribute = lAttributes.item(i).toAttr();
		QTreeWidgetItem* lAttrItem = createItem(lAttribute, lElemItem);
		lAttrItem->setFlags(lAttrItem->flags() | Qt::ItemIsEditable);
		lAttrItem->setIcon(0, mAttributeIcon);
		lAttrItem->setText(0, lAttribute.name());
		lAttrItem->setText(1, lAttribute.value());
	}

	QDomElement lChild = inElement.firstChildElement();
	while (!lChild.isNull()) {
		parseDomElement(lChild, lElemItem);
		lChild = lChild.nextSiblingElement();
	}
}


QTreeWidgetItem* ConfigurationTree::createItem(const QDomNode &inNode, QTreeWidgetItem* inParentItem)
{
	QTreeWidgetItem* lItem;
	if (inParentItem) {
		lItem = new QTreeWidgetItem(inParentItem);
	} else {
		lItem = new QTreeWidgetItem(this);
	}
	mDomNodeForItem.insert(lItem, inNode);

	return lItem;
}


bool ConfigurationTree::write(QIODevice* outDevice)
{
	const int lIndentSize = 2;

	QTextStream lOut(outDevice);
	mDomDocument.save(lOut, lIndentSize);

	return true;
}


void ConfigurationTree::updateDomElement(QTreeWidgetItem *ioItem, int inColumn)
{
	QDomNode lNode = mDomNodeForItem.value(ioItem);

	if (lNode.isNull())
		return;

	if (lNode.isElement()) {
		QDomElement lElement = lNode.toElement();
		if (inColumn == 0) {
			lElement.setTagName(ioItem->text(0));
		} else {
			// We allow the text to be added only if the element has no child elements.
			// We can modify the text directly if there was a text node already.
			// Otherwise, we will have to create a new text node.
			if (lElement.firstChildElement().isNull()) {
				if (lElement.firstChild().isText()) {
					lElement.firstChild().toText().setData(ioItem->text(1));
				} else {
					QDomText lText = mDomDocument.createTextNode(ioItem->text(1));
					lElement.appendChild(lText);
				}
			} else {
				ioItem->setText(1, QString());
			}
		}
	} else if (lNode.isAttr()) {
		QDomAttr lAttribute = lNode.toAttr();
		// Attribute name is not modifiable.
		if (inColumn == 0) {
			ioItem->setText(0, lAttribute.name());
		} else {
			lAttribute.setValue(ioItem->text(1));
		}
	}

    emit modified();
}

