
#include "explorer/Explorer.hpp"

using namespace Explorer;


DisplayNode::DisplayNode()
{
	Beagle_StackTraceBeginM();

	setFlags(ItemIsSelectable);

	Beagle_StackTraceEndM("DisplayNode::DisplayNode(HistoryEntry::Handle inHistoryEntry)");
}


QRectF DisplayNode::boundingRect() const
{
	Beagle_StackTraceBeginM();

	DisplayScene* lScene = qobject_cast<DisplayScene*>(scene());
	if (lScene)
		return QRectF(0, 0, lScene->getNodeBoundingSize(), lScene->getNodeBoundingSize());
	else
		return QRectF();

	Beagle_StackTraceEndM("QRectF DisplayNode::boundingRect() const");
}


/*!
 *  \brief The main color of the node.
 *
 *  Qt::transparent is returned if all node color indices are 0 representing "None".
 *  Qt::magenta is returned if color cannot be calculated because of invalidity.
 */
QColor DisplayNode::getOuterColor() const
{
	Beagle_StackTraceBeginM();

	DisplayScene* lScene = qobject_cast<DisplayScene*>(scene());
	if (lScene == 0)
		return QColor();

	int lNodeHueIndex = lScene->getNodeHueIndex();
	int lNodeSatIndex = lScene->getNodeSatIndex();
	int lNodeValIndex = lScene->getNodeValIndex();

    return getOuterColor(lNodeHueIndex, lNodeSatIndex, lNodeValIndex);

	Beagle_StackTraceEndM("QColor DisplayNode::getOuterColor() const");
}


/*!
 *  \brief The main color of the node.
 *
 *  Qt::transparent is returned if all node color indices are 0 representing "None".
 *  Qt::magenta is returned if color cannot be calculated because of invalidity.
 */
QColor DisplayNode::getOuterColor(int inNodeHueIndex, int inNodeSatIndex, int inNodeValIndex) const
{
	Beagle_StackTraceBeginM();

	if (inNodeHueIndex == 0 && inNodeSatIndex == 0 && inNodeValIndex == 0)
		return Qt::transparent;

	const DisplayObject::Handle lDisplayObject = getDisplayObject();
	if (lDisplayObject == NULL)
		return QColor(Qt::magenta);

	float lNodeHueValue = lDisplayObject->getDisplayValue(inNodeHueIndex);
	float lNodeSatValue = lDisplayObject->getDisplayValue(inNodeSatIndex);
	float lNodeValValue = lDisplayObject->getDisplayValue(inNodeValIndex);

	if (lNodeHueValue > -FLT_MAX && lNodeSatValue > -FLT_MAX && lNodeValValue > -FLT_MAX) {
		qreal lH = QColor(Qt::blue).hueF();
		qreal lS = (inNodeHueIndex == 0 && inNodeSatIndex == 0) ? 0.0 : 1.0;
		qreal lV = 1.0;

		if (inNodeHueIndex > 0) {
			qreal lMin = lDisplayObject->getMinDisplayValue(inNodeHueIndex);
			qreal lMax = lDisplayObject->getMaxDisplayValue(inNodeHueIndex);
			lH = (lMax > lMin) ? lH*(lNodeHueValue - lMin)/(lMax - lMin) : 0.0;
		}

		if (inNodeSatIndex > 0) {
			qreal lMin = lDisplayObject->getMinDisplayValue(inNodeSatIndex);
			qreal lMax = lDisplayObject->getMaxDisplayValue(inNodeSatIndex);
			lS = (lMax > lMin) ? (lNodeSatValue - lMin)/(lMax - lMin) : 0.0;
		}

		if (inNodeValIndex > 0) {
			qreal lMin = lDisplayObject->getMinDisplayValue(inNodeValIndex);
			qreal lMax = lDisplayObject->getMaxDisplayValue(inNodeValIndex);
			lV = (lMax > lMin) ? (lNodeValValue - lMin)/(lMax - lMin) : 0.0;
		}

        QColor lColor = QColor::fromHsvF(lH, lS, lV);
        if (lColor.isValid())
            return lColor;
        else
            return QColor(Qt::magenta);
	} else {
		return QColor(Qt::magenta);
	}

	Beagle_StackTraceEndM("QColor DisplayNode::getOuterColor(int inNodeHueIndex, int inNodeSatIndex, int inNodeValIndex) const");
}


/*!
 *  \brief The color of the inner sign based on the outer color.
 */
QColor DisplayNode::getInnerColor(const QColor& inOuterColor) const
{
	Beagle_StackTraceBeginM();

	DisplayScene* lScene = qobject_cast<DisplayScene*>(scene());
	if (lScene == 0)
		return QColor();

	int lNodeHueIndex = lScene->getNodeHueIndex();
	int lNodeSatIndex = lScene->getNodeSatIndex();
	int lNodeValIndex = lScene->getNodeValIndex();

    return getInnerColor(inOuterColor, lNodeHueIndex, lNodeSatIndex, lNodeValIndex);

	Beagle_StackTraceEndM("QColor DisplayNode::getInnerColor(const QColor& inOuterColor) const");
}


/*!
 *  \brief The color of the inner sign based on the outer color.
 */
QColor DisplayNode::getInnerColor(const QColor& inOuterColor, int inNodeHueIndex, int inNodeSatIndex, int inNodeValIndex) const
{
	Beagle_StackTraceBeginM();

	// The inner color is constructed as the complement of the outer color. The
	// complement is taken only of those color components that are used in the display.
	// Complement of hue is located diametrically opposite on the color wheel, while
	// that of saturation and value is the opposite of the closest bound.

	qreal lH = inOuterColor.hueF();
	if (inNodeHueIndex > 0) {
		lH = lH + 0.5;
		lH = lH - floor(lH);
	}

	qreal lS = inOuterColor.saturationF();
	if (inNodeSatIndex > 0) {
		lS = (lS > 0.5) ? 0.0 : 1.0;
	}

	qreal lV = inOuterColor.valueF();
	if (inNodeValIndex > 0) {
		lV = (lV > 0.5) ? 0.0 : 1.0;
	}

	return QColor::fromHsvF(lH, lS, lV);

	Beagle_StackTraceEndM("QColor DisplayNode::getInnerColor(const QColor& inOuterColor, int inNodeHueIndex, int inNodeSatIndex, int inNodeValIndex) const");
}
