
#include "explorer/Explorer.hpp"

using namespace Explorer;


DisplaySettings::DisplaySettings(QWidget* inParent, Qt::WFlags inFlags) :
		QWidget(inParent, inFlags),
		Ui::DisplaySettings(),
		mView(new QGraphicsView(this)),
		mScene(0)
{
	Beagle_StackTraceBeginM();

	setupUi(this);

	mView->setViewportUpdateMode(QGraphicsView::SmartViewportUpdate);
	mView->setRenderHints(QPainter::Antialiasing);

	connect(mZoomSlider, SIGNAL(valueChanged(int)), this, SLOT(scale(int)));

	mZoomSlider->setWhatsThis(tr("Adjust zoom level of the view."));
	mNodeColorGroup->setWhatsThis(tr("Specify components of node color. "
	                                 "Node is colored magenta if one or more components is invalid."));

	connect(mNodeHueCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(updateNodeHue(int)));
	connect(mNodeSatCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(updateNodeSat(int)));
	connect(mNodeValCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(updateNodeVal(int)));

	mNodeHueCombo->setWhatsThis(tr("Select parameter to use as the hue component of node color."));
	mNodeSatCombo->setWhatsThis(tr("Select parameter to use as the saturation component of node color."));
	mNodeValCombo->setWhatsThis(tr("Select parameter to use as the value component of node color."));

	qreal lHue0 = QColor(Qt::red).hueF();
	qreal lHue1 = QColor(Qt::blue).hueF();
	mNodeHueBar->setEndColors(QColor::fromHsvF(lHue0, 1.0, 1.0), QColor::fromHsvF(lHue1, 1.0, 1.0));
	mNodeSatBar->setEndColors(QColor::fromHsvF(lHue1, 0.0, 1.0), QColor::fromHsvF(lHue1, 1.0, 1.0));
	mNodeValBar->setEndColors(QColor::fromHsvF(lHue1, 0.0, 0.0), QColor::fromHsvF(lHue1, 0.0, 1.0));

	Beagle_StackTraceEndM("DisplaySettings::DisplaySettings(History::Handle inHistory, QWidget* inParent, Qt::WFlags inFlags)");
}


/*!
 *  \brief Reset display settings.
 */
void DisplaySettings::reset()
{
	Beagle_StackTraceBeginM();

	if (mPrototypeObject != NULL)
		mPrototypeObject->reset();
	mPrototypeObject = NULL;

	mNodeHueCombo->clear();
	mNodeSatCombo->clear();
	mNodeValCombo->clear();

	Beagle_StackTraceEndM("void DisplaySettings::reset()");
}


/*!
 *  \brief Setup display settings.
 */
void DisplaySettings::setup()
{
	Beagle_StackTraceBeginM();

	// If prototype is NULL, return with no setup.
	if (mPrototypeObject == NULL)
		return;

	// Setup the node color comboboxes using the prototype.
	int lNumEntries = mPrototypeObject->getDisplaySize();
	for (int k = 0; k < lNumEntries; ++k) {
		std::string lDisplayName = mPrototypeObject->getDisplayName(k);
		mNodeHueCombo->addItem(lDisplayName.c_str());
		mNodeSatCombo->addItem(lDisplayName.c_str());
		mNodeValCombo->addItem(lDisplayName.c_str());
	}

	Beagle_StackTraceEndM("void DisplaySettings::setup(Display::Handle inPrototype)");
}


/*!
 *  \brief Scale the view by the value set by the zoom slider.
 */
void DisplaySettings::scale(int inScale)
{
	Beagle_StackTraceBeginM();

	qreal lScale = inScale/(qreal)(mZoomSlider->maximum());
	QMatrix lMatrix;
	lMatrix.scale(lScale, lScale);
	mView->setMatrix(lMatrix);

	Beagle_StackTraceEndM("void DisplaySettings::scale(int inScale)");
}


/*!
 *  \brief Schedule nodes for update when color hue index changes.
 */
void DisplaySettings::updateNodeHue(int inNodeHueIndex)
{
	Beagle_StackTraceBeginM();

	if (mPrototypeObject != NULL) {
		QString lMinLabel, lMaxLabel;
		if (inNodeHueIndex > 0) {
			lMinLabel = QString::number(mPrototypeObject->getMinDisplayValue(inNodeHueIndex));
			lMaxLabel = QString::number(mPrototypeObject->getMaxDisplayValue(inNodeHueIndex));
		} else {
			lMinLabel = "--";
			lMaxLabel = "--";
		}
		mNodeHueEnd0->setText(lMinLabel);
		mNodeHueEnd1->setText(lMaxLabel);
	}

	if (mScene) {
		mScene->setNodeHueIndex(inNodeHueIndex);
		mScene->invalidate();
	}

	Beagle_StackTraceEndM("void DisplaySettings::updateNodeHue(int inNodeHueIndex)");
}


/*!
 *  \brief Schedule nodes for update when color saturation index changes.
 */
void DisplaySettings::updateNodeSat(int inNodeSatIndex)
{
	Beagle_StackTraceBeginM();

	if (mPrototypeObject != NULL) {
		QString lMinLabel, lMaxLabel;
		if (inNodeSatIndex > 0) {
			lMinLabel = QString::number(mPrototypeObject->getMinDisplayValue(inNodeSatIndex));
			lMaxLabel = QString::number(mPrototypeObject->getMaxDisplayValue(inNodeSatIndex));
		} else {
			lMinLabel = "--";
			lMaxLabel = "--";
		}
		mNodeSatEnd0->setText(lMinLabel);
		mNodeSatEnd1->setText(lMaxLabel);
	}

	if (mScene) {
		mScene->setNodeSatIndex(inNodeSatIndex);
		mScene->invalidate();
	}

	Beagle_StackTraceEndM("void DisplaySettings::updateNodeSat(int inNodeSatIndex)");
}


/*!
 *  \brief Schedule nodes for update when color value index changes.
 */
void DisplaySettings::updateNodeVal(int inNodeValIndex)
{
	Beagle_StackTraceBeginM();

	if (mPrototypeObject != NULL) {
		QString lMinLabel, lMaxLabel;
		if (inNodeValIndex > 0) {
			lMinLabel = QString::number(mPrototypeObject->getMinDisplayValue(inNodeValIndex));
			lMaxLabel = QString::number(mPrototypeObject->getMaxDisplayValue(inNodeValIndex));
		} else {
			lMinLabel = "--";
			lMaxLabel = "--";
		}
		mNodeValEnd0->setText(lMinLabel);
		mNodeValEnd1->setText(lMaxLabel);
	}

	if (mScene) {
		mScene->setNodeValIndex(inNodeValIndex);
		mScene->invalidate();
	}

	Beagle_StackTraceEndM("void DisplaySettings::updateNodeVal(int inNodeValIndex)");
}
