
#include <QVector>
#include <QFileDialog>
#include <QMessageBox>
#include <QSettings>
#include <QCloseEvent>
#include <QCoreApplication>

#include "explorer/Explorer.hpp"

using namespace Explorer;


EvolutionTab::EvolutionTab(Beagle::System::Handle ioSystem, QWidget* inParent, Qt::WFlags inFlags) :
		QWidget(inParent, inFlags),
		Ui::EvolutionTab(),
        mSystem(ioSystem),
        mEvolver(NULL)
{
	Beagle_StackTraceBeginM();

	setupUi(this);

    // Setup the configuration file editor inside the display widget.
    QHBoxLayout *lHBoxLayout = new QHBoxLayout(mDisplayWidget);
    mConfigurationTree = new ConfigurationTree(mDisplayWidget);
    lHBoxLayout->addWidget(mConfigurationTree);

	connect(mConfigurationTree, SIGNAL(modified()), this, SLOT(setConfigurationModified()));
	connect(mEvolveButton, SIGNAL(clicked()), this, SLOT(evolve()));

    // Open explorer.conf as configuration file if it exists.
    QString lFileName = (QCoreApplication::arguments().size() > 1) ? QCoreApplication::arguments().at(1) : "explorer.conf";
	QFile lFile(lFileName);
	if (lFile.open(QFile::ReadOnly | QFile::Text) && mConfigurationTree->read(&lFile)) {
        mConfigurationFile = lFileName;
		mConfigurationLine->setText(lFileName);
	}

    // Get the singleton data instance, and set up connections.
    InteractionData& lInteractionData = InteractionData::instance();
	connect(mBreakpointsCheckBox, SIGNAL(stateChanged(int)), &lInteractionData, SLOT(setBreakEnabled(int)));
	connect(mPauseButton, SIGNAL(clicked()), &lInteractionData, SLOT(setPauseEnabled()));
	connect(mContinueButton, SIGNAL(clicked()), &lInteractionData, SLOT(continueFromPauseOrBreak()));

	readSettings();

	Beagle_StackTraceEndM("EvolutionTab::EvolutionTab(Beagle::System::Handle ioSystem, QWidget* inParent, Qt::WFlags inFlags)");
}


void EvolutionTab::closeEvent(QCloseEvent *inEvent)
{
	Beagle_StackTraceBeginM();

    const Beagle::Factory& lFactory = mSystem->getFactory();
    Interaction::Alloc::Handle lInteractionAlloc =
        Beagle::castHandleT<Interaction::Alloc>(lFactory.getConceptAllocator("Interaction"));
    Interaction::Handle lInteraction = Beagle::castHandleT<Interaction>(lInteractionAlloc->allocate());
    lInteraction->closeEvent(inEvent);

	writeSettings();
	inEvent->accept();

	Beagle_StackTraceEndM("void EvolutionTab::closeEvent(QCloseEvent *inEvent)");
}


void EvolutionTab::openConfiguration()
{
	Beagle_StackTraceBeginM();

	QString lFileName = QFileDialog::getOpenFileName(this, tr("Open Configuration File"),
	                    QDir::currentPath(),
	                    tr("Configuration Files (*.conf)"));
	if (lFileName.isEmpty())
		return;

	QFile lFile(lFileName);
	if (!lFile.open(QFile::ReadOnly | QFile::Text)) {
		QMessageBox::warning(this, tr("OpenBEAGLE Explorer"),
		                     tr("Cannot read file %1:\n%2.")
		                     .arg(lFileName)
		                     .arg(lFile.errorString()));
		return;
	}

	if (mConfigurationTree->read(&lFile)) {
        mConfigurationFile = lFileName;
		mConfigurationLine->setText(lFileName);
	}

	Beagle_StackTraceEndM("void EvolutionTab::openConfiguration()");
}


void EvolutionTab::saveConfiguration()
{
	Beagle_StackTraceBeginM();

    QString lFileName = mConfigurationFile;
	if (lFileName.isEmpty())
		return saveConfigurationAs();

	QFile lFile(lFileName);
	if (!lFile.open(QFile::WriteOnly | QFile::Text)) {
		QMessageBox::warning(this, tr("OpenBEAGLE Explorer"),
		                     tr("Cannot write file %1:\n%2.")
		                     .arg(lFileName)
		                     .arg(lFile.errorString()));
		return;
	}

	if (mConfigurationTree->write(&lFile)) {
        mConfigurationFile = lFileName;
		mConfigurationLine->setText(lFileName);
	}

	Beagle_StackTraceEndM("void EvolutionTab::saveConfiguration()");
}


void EvolutionTab::saveConfigurationAs()
{
	Beagle_StackTraceBeginM();

	QString lFileName = QFileDialog::getSaveFileName(this, tr("Save Configuration File"),
	                    QDir::currentPath(),
	                    tr("Configuration Files (*.conf)"));
	if (lFileName.isEmpty())
		return;

	QFile lFile(lFileName);
	if (!lFile.open(QFile::WriteOnly | QFile::Text)) {
		QMessageBox::warning(this, tr("OpenBEAGLE Explorer"),
		                     tr("Cannot write file %1:\n%2.")
		                     .arg(lFileName)
		                     .arg(lFile.errorString()));
		return;
	}

	if (mConfigurationTree->write(&lFile)) {
        mConfigurationFile = lFileName;
		mConfigurationLine->setText(lFileName);
	}

	Beagle_StackTraceEndM("void EvolutionTab::saveConfigurationAs()");
}


void EvolutionTab::setConfigurationModified()
{
	Beagle_StackTraceBeginM();

    mConfigurationLine->setText(mConfigurationFile + "*");

	Beagle_StackTraceEndM("void EvolutionTab::setConfigurationModified()");
}


void EvolutionTab::evolve()
{
	Beagle_StackTraceBeginM();

    // Disable the evolve button while this evolution is running.
    // Also, Beagle doesn't allow starting another evolution with the same system.
    mEvolveButton->setEnabled(false);

    // Initialize the evolver before setting up the interaction component.
    mEvolver = new Beagle::Evolver;
    mEvolver->initialize(mSystem, mConfigurationFile.toStdString());

    // Create an interaction component and set it up.
    const Beagle::Factory& lFactory = mSystem->getFactory();
    Interaction::Alloc::Handle lInteractionAlloc =
        Beagle::castHandleT<Interaction::Alloc>(lFactory.getConceptAllocator("Interaction"));
    Interaction::Handle lInteraction = Beagle::castHandleT<Interaction>(lInteractionAlloc->allocate());
    lInteraction->setup(mSystem);

    QVector< QVector<QWidget*> > lStatusIndicators;
    lInteraction->setupStatusIndicators(mStatusBox, lStatusIndicators);
    for (unsigned int i=0; i<lStatusIndicators.size(); i++) {
        gridLayout->addWidget(lStatusIndicators[i][0], i+1, 0, 1, 1);
        gridLayout->addWidget(lStatusIndicators[i][1], i+1, 1, 1, 1);
        gridLayout->addWidget(lStatusIndicators[i][2], i+1, 2, 1, 1);
        connect(mBreakpointsCheckBox, SIGNAL(toggled(bool)), lStatusIndicators[i][2], SLOT(setEnabled(bool)));
        lStatusIndicators[i][2]->setEnabled(mBreakpointsCheckBox->isChecked());
    }

    QVector<QWidget*> lDisplayIndicators;
    lInteraction->setupDisplayIndicators(mControlCenter, lDisplayIndicators);
    for (unsigned int i=0; i<lDisplayIndicators.size(); i++) {
        vboxLayout1->addWidget(lDisplayIndicators[i]);
    }

    // Create population
    Beagle::Vivarium::Handle lVivarium = new Beagle::Vivarium;

    // Launch evolution
    mEvolver->evolve(lVivarium, mSystem);

	Beagle_StackTraceEndM("void EvolutionTab::evolve()");
}


void EvolutionTab::readSettings()
{
	Beagle_StackTraceBeginM();

	QSettings lSettings("OpenBEAGLE", "Explorer");

	lSettings.beginGroup("EvolutionTab");
	mSplitter->restoreState(lSettings.value("splitterSizes").toByteArray());
	lSettings.endGroup();

	Beagle_StackTraceEndM("void EvolutionTab::readSettings()");
}


void EvolutionTab::writeSettings()
{
	Beagle_StackTraceBeginM();

	QSettings lSettings("OpenBEAGLE", "Explorer");

	lSettings.beginGroup("EvolutionTab");
	lSettings.setValue("splitterSizes", mSplitter->saveState());
	lSettings.endGroup();

	Beagle_StackTraceEndM("void EvolutionTab::writeSettings()");
}
