
#include <QProcess>
#include <QGraphicsSvgItem>
#include <QGraphicsTextItem>
#include <QSettings>
#include <QMessageBox>

#include "explorer/Explorer.hpp"

using namespace Explorer;


GenealogySettings::GenealogySettings(History::Handle inHistory, QWidget* inParent, Qt::WFlags inFlags) :
		HistorySettings(inHistory, inParent, inFlags),
		mSvgRenderer(0)
{
	Beagle_StackTraceBeginM();

	mView->setWhatsThis(tr("<b>Genealogy view of selected nodes.</b>") +
	                    tr("<p>Select nodes to expand the genealogy tree further. ") +
	                    tr("Edges represent operations that created nodes from parents ") +
	                    tr("(<font color=#000000>Selection</font>, <font color=#FF0000>Elitism</font>, ") +
	                    tr("<font color=#00FF00>Mutation</font>, ") +
	                    tr("<font color=#0000FF>Crossover 1<sup>st</sup> Parent</font>, ") +
	                    tr("<font color=#00FFFF>Crossover 2<sup>nd</sup> Parent</font>).</p>"));

	Beagle_StackTraceEndM("GenealogySettings::GenealogySettings(History::Handle inHistory, QWidget* inParent, Qt::WFlags inFlags)");
}


/*!
 *  \brief Update view when selection or ancestry/descendancy depths change.
 */
void GenealogySettings::updateView()
{
	Beagle_StackTraceBeginM();

	if (mVivariumSelections == NULL)
		return;

	HistoryEntry::Bag::Handle lGenealogySelections = getSelectedHistoryEntries();

	HistoryEntry::Bag::Handle lSelections = new HistoryEntry::Bag;
	for (unsigned int i = 0; i < mVivariumSelections->size(); ++i) {
		lSelections->push_back(mVivariumSelections->at(i));
	}
	for (unsigned int i = 0; i < lGenealogySelections->size(); ++i) {
		lSelections->push_back(lGenealogySelections->at(i));
	}

	QString lDotFile = "genealogy.dot";
	QString lSvgFile = "genealogy.svg";
	std::ofstream lDotOFS(lDotFile.toStdString().c_str());
	mHistory->writeGenealogy(lDotOFS, *lSelections, mAncestrySpin->value(), mDescendancySpin->value());
	lDotOFS.close();

	QSettings lSettings("OpenBEAGLE", "Explorer");
	QString lDotExe = lSettings.value("dotexe", QString("/usr/bin/dot")).toString();
	QStringList lDotArguments;
	lDotArguments << "-Tsvg" << lDotFile << "-o" << lSvgFile;
	QProcess lDotProcess;
	lDotProcess.start(lDotExe, lDotArguments);

	if (lDotProcess.waitForStarted() == false) {
		QMessageBox::warning(this, tr("Graphviz Dot Command Failure"),
		                     tr("Make sure the path to the dot command is set correctly via the Settings menu. ") +
		                     tr("The current path setting is %1").arg(lDotExe));
		return;
	}

	if (lDotProcess.waitForFinished() == false) {
		QMessageBox::warning(this, tr("Graphviz Dot Command Failure"),
		                     tr("The dot command exited unsuccessfully."));
		return;
	}

	// Process the SVG file produced by graphviz to make it suitable for our use.
	std::ifstream lSvgIFS(lSvgFile.toStdString().c_str());
	PACC::XML::Document lParser;
	lParser.setNoParse("title");  // prevents messing up the edge title; content is a string anyway.
	lParser.parse(lSvgIFS, lSvgFile.toStdString());
	lSvgIFS.close();

	processSvg(lParser);

	std::ostringstream lSvgOSS;
	lParser.serialize(lSvgOSS);

	// Schedule the previous svg renderer for deletion before allocating a new one.
	if (mSvgRenderer)
		mSvgRenderer->deleteLater();
	mSvgRenderer = new QSvgRenderer(QByteArray(lSvgOSS.str().c_str()), this);

	// Schedule the previous scene for deletion before allocating a new one.
	if (mScene)
		mScene->deleteLater();
	mScene = new DisplayScene(this);

	mScene->setNodeHueIndex(mNodeHueCombo->currentIndex());
	mScene->setNodeSatIndex(mNodeSatCombo->currentIndex());
	mScene->setNodeValIndex(mNodeValCombo->currentIndex());

	for (PACC::XML::Iterator lRoot=lParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/g/g"); lGroup; lGroup = lFinder.findNext()) {
			if (lGroup->getAttribute("class")=="node") {
				QString lElementId = lGroup->getAttribute("id").c_str();
				QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);
				// Nodes representing individuals.
				if (lElementId.contains('v')) {
					unsigned int lHistoryId = lElementId.section('v', 0, 0).toUInt();
					unsigned int lHistoryVar = lElementId.section('v', -1).toUInt();
					HistoryEntry::Handle lHistoryEntry = mHistory->getEntry(lHistoryId, lHistoryVar);
					GenealogyNode* lItem = new GenealogyNode(lHistoryEntry);
					lItem->setToolTip(lHistoryEntry->getToolTip().c_str());
					lItem->setPos(lBoundingRect.topLeft());
					if (mScene->getNodeBoundingSize() < 0.0)
						mScene->setNodeBoundingSize(lBoundingRect.width());
					mScene->addItem(lItem);
					if (lHistoryEntry->isMember(lHistoryEntry, mVivariumSelections))
						lItem->setVivariumFlag(true);
					else if (lHistoryEntry->isMember(lHistoryEntry, lGenealogySelections))
						lItem->setSelected(true);
				}
				// Nodes representing generation labels.
				else {
					QGraphicsTextItem* lItem = new QGraphicsTextItem(lElementId);
					QRectF lTextRect = lItem->boundingRect();
					qreal lTextWidth = lTextRect.right() - lTextRect.left();
					qreal x = (lBoundingRect.left() + lBoundingRect.right() - lTextWidth)/2;
					qreal y = (lBoundingRect.top() + lBoundingRect.bottom())/2;
					lItem->setPos(x, y);
					mScene->addItem(lItem);
				}
			} else if (lGroup->getAttribute("class")=="edge") {
				QString lElementId = lGroup->getAttribute("id").c_str();
				QRectF lBoundingRect = mSvgRenderer->boundsOnElement(lElementId);
				QGraphicsSvgItem* lItem = new QGraphicsSvgItem;
				lItem->setSharedRenderer(mSvgRenderer);
				lItem->setElementId(lElementId);
				lItem->setPos(lBoundingRect.topLeft());
				mScene->addItem(lItem);
			}
		}
	}

	mView->setScene(mScene);

	connect(mScene, SIGNAL(nodeSelectionChanged()), this, SLOT(updateView()));

	Beagle_StackTraceEndM("void GenealogySettings::updateView()");
}


/*!
 *  \brief Process the SVG produced by graphviz
 */
void GenealogySettings::processSvg(PACC::XML::Document& ioParser) const
{
	Beagle_StackTraceBeginM();

	for (PACC::XML::Iterator lRoot=ioParser.getFirstRoot(); lRoot; ++lRoot) {
		PACC::XML::Finder lFinder(lRoot);

		// Find all groups that represent nodes and edges in the svg document.
		// Add the id attribute to them to enable access by ids.
		for (PACC::XML::Iterator lGroup = lFinder.find("/svg/g/g"); lGroup; lGroup = lFinder.findNext()) {
			if (lGroup->getAttribute("class")=="node") {
				std::string lTitleString;
				for (PACC::XML::Iterator lChild=lGroup->getFirstChild(); lChild; ++lChild) {
					if (lChild->getType()==PACC::XML::eData && lChild->getValue()=="title")
						lTitleString = lChild->getFirstChild()->getValue();
				}
				lGroup->setAttribute("id", lTitleString);
			} else if (lGroup->getAttribute("class")=="edge") {
				QString lTitleString;
				for (PACC::XML::Iterator lChild=lGroup->getFirstChild(); lChild; ++lChild) {
					if (lChild->getType()==PACC::XML::eData && lChild->getValue()=="title")
						lTitleString = lChild->getFirstChild()->getValue().c_str();
				}
				// Edge between nodes representing individuals.
				if (lTitleString.contains('v')) {
					lTitleString.replace("&#45;&gt;", "t");   // replace the escaped "->" with "t"
					lGroup->setAttribute("id", lTitleString.toStdString());
				}
				// Edge between nodes representing generation labels.
				else {
					ioParser.erase(lGroup);
				}
			}
		}
	}

	Beagle_StackTraceEndM("void GenealogySettings::processSvg(PACC::XML::Document& ioParser) const");
}


void GenealogySettings::setVivariumSelections(HistoryEntry::Bag::Handle inVivariumSelections)
{
	Beagle_StackTraceBeginM();

	mVivariumSelections = inVivariumSelections;

	Beagle_StackTraceEndM("void GenealogySettings::setVivariumSelections(HistoryEntry::Bag::Handle inVivariumSelections)");
}
