
#include "explorer/Explorer.hpp"

using namespace Explorer;


History::History() :
		mMinGeneration(UINT_MAX)
{}


/*!
 *  \brief Reset any previous information.
 */
void History::reset()
{
	Beagle_StackTraceBeginM();

	mEntries.clear();
	mMarkedEntries.clear();
	mMinGeneration = UINT_MAX;

	Beagle_StackTraceEndM("void History::reset()");
}


/*!
 *  \brief Setup using Beagle history.
 */
void History::setup(Beagle::History::Handle inHistory, Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	if (inHistory == NULL)
		return;

	const Beagle::HistoryTrace::Bag& lTraces = inHistory->getTraces();
	unsigned int lFirstId = inHistory->getFirstID();
	if (lFirstId >= mEntries.size())
		mEntries.resize(lTraces.size()+lFirstId);

	// Transfer information from Beagle history entries to Explorer history.
	// Entries are created in Explorer history when they are first referenced.
	for (unsigned int i = 0; i < lTraces.size(); ++i) {
		mEntries[i+lFirstId].resize(lTraces[i]->size());
		for (unsigned int j = 0; j < lTraces[i]->size(); ++j) {
			if (mEntries[i+lFirstId][j] == NULL)
				mEntries[i+lFirstId][j] = new HistoryEntry();
			HistoryEntry::Handle lEntry = mEntries[i+lFirstId][j];
			lEntry->setup((*lTraces[i])[j], ioSystem);
			lEntry->setHistoryId(i+lFirstId);

			// Marked individuals are collected in an array.
			if (lEntry->getMarkedFlag() == true) {
				unsigned int lGeneration = lEntry->getGeneration();
				if (lGeneration >= mMarkedEntries.size())
					mMarkedEntries.resize(lGeneration+1);
				mMarkedEntries[lGeneration].push_back(lEntry);
				if (lGeneration < mMinGeneration)
					mMinGeneration = lGeneration;
			}

			// Set parent and child relations.
			const std::vector<Beagle::HistoryID>& lParents = (*lTraces[i])[j]->getParents();
			for (unsigned int k = 0; k < lParents.size(); ++k) {
				const unsigned int lParentId = lParents[k].getID();
				const unsigned int lParentVar = lParents[k].getVar();
				if (lParentVar >= mEntries[lParentId].size())
					mEntries[lParentId].resize(lParentVar+1);
				if (mEntries[lParentId][lParentVar] == NULL)
					mEntries[lParentId][lParentVar] = new HistoryEntry();
				HistoryEntry::Handle lParentEntry = mEntries[lParentId][lParentVar];
				lEntry->addParent(lParentEntry);
				lParentEntry->addChild(lEntry);
			}
		}
	}

	// Now we can set the representative individuals for entries.  We have to
	// scan the entire list of entries to fill out any missing representatives
	// that are now available.
    for (unsigned int i = 0; i < mEntries.size(); ++i) {
        for (unsigned int j = 0; j < mEntries[i].size(); ++j) {
            HistoryEntry::Handle lEntry = mEntries[i][j];
            if (lEntry == NULL)
                continue;
            Individual::Handle lIndividual = lEntry->getIndividual();
            while (lIndividual == NULL && lEntry->getParents().size() > 0) {
                lEntry = (lEntry->getParents())[0];
                lIndividual = lEntry->getIndividual();
            }
            mEntries[i][j]->setIndividual(lIndividual);
        }
    }

	Beagle_StackTraceEndM("void History::setup(Beagle::History::Handle inHistory, Beagle::System& ioSystem)");
}


/*!
 *  \brief Find the ancestors and descendants of the given set of individuals.
 *  \param ioSelectedEntries The history entries whose ancestors/descendants to find.
 *  \param inAncestryDepth Depth to which ancestors are searched for.
 *  \param inDescendancyDepth Depth to which descendants are searched for.
 */
void History::findGenealogy(HistoryEntry::Bag& ioSelectedEntries, int inAncestryDepth, int inDescendancyDepth)
{
	Beagle_StackTraceBeginM();

	// Reset the genealogy flag of all individuals.
	for (unsigned int i = 0; i < mEntries.size(); ++i) {
		for (unsigned int j = 0; j < mEntries[i].size(); ++j) {
			if (mEntries[i][j] != NULL)
				mEntries[i][j]->setGenealogyFlag(false);
		}
	}

	int lAncestryDepth = (inAncestryDepth < 0) ? getMaxGeneration() : inAncestryDepth;
	int lDescendancyDepth = (inDescendancyDepth < 0) ? getMaxGeneration() : inDescendancyDepth;

	for (unsigned int i = 0; i < ioSelectedEntries.size(); ++i) {
		// Initialize the color of all nodes in the graph to white.
		for (unsigned int j = 0; j < mEntries.size(); ++j) {
			for (unsigned int k = 0; k < mEntries[j].size(); ++k) {
				if (mEntries[j][k] != NULL)
					mEntries[j][k]->setSearchColor(HistoryEntry::eWhite);
			}
		}

		genealogySearchBFS(ioSelectedEntries[i], lAncestryDepth, true);
		genealogySearchBFS(ioSelectedEntries[i], lDescendancyDepth, false);
	}

	Beagle_StackTraceEndM("void History::findGenealogy(HistoryEntry::Bag& ioSelectedEntries, int inAncestryDepth, int inDescendancyDepth)");
}


/*!
 *  \brief Search for the ancestors/descendants of the given individual using breadth-first-search.
 *  \param ioEntry The entry from which search is started.
 *  \param inDepth The maximum depth to which search is conducted.
 *  \param inAncestry If true, ancestry is searched, otherwise descendancy is searched.
 *
 *  The part of the graph that is searched must have the search color set to white before calling
 *  this method.
 */
void History::genealogySearchBFS(HistoryEntry::Handle ioEntry, int inDepth, bool inAncestry)
{
	Beagle_StackTraceBeginM();

	// Queue for managing the set of gray nodes.
	std::queue<HistoryEntry::Handle> lQ;

	// Discover and add the given individual to the queue.
	ioEntry->setSearchColor(HistoryEntry::eGray);
	lQ.push(ioEntry);

	unsigned int lReferenceGeneration = ioEntry->getGeneration();
	while (lQ.size() > 0) {
		// Examine the node at the front of the queue and remove it.
		HistoryEntry::Handle lCurrent = lQ.front();
		lQ.pop();
		lCurrent->setGenealogyFlag(true);

		// Examine all edges of the current node.
		std::vector<HistoryEntry::Handle>& lNext = inAncestry ? lCurrent->getParents() : lCurrent->getChildren();
		for (unsigned int i = 0; i < lNext.size(); ++i) {
			// Skip next node if it doesn't have a valid history id (i.e., the history entry
			// was not initialized by calling setup())
			if (lNext[i]->getHistoryId() == Beagle::HistoryID::eDummyID)
				continue;

			// If the edge to the next node is a tree edge, and the node is within the search
			// depth, then continue search by adding the node to the queue.
			int lDepth = std::abs((int)lReferenceGeneration - (int)(lNext[i]->getGeneration()));
			if ((lNext[i]->getSearchColor() == HistoryEntry::eWhite) && (lDepth <= inDepth)) {
				lNext[i]->setSearchColor(HistoryEntry::eGray);
				lQ.push(lNext[i]);
			}
		}

		// Finish examining the current node.
		lCurrent->setSearchColor(HistoryEntry::eBlack);
	}

	Beagle_StackTraceEndM("void History::genealogySearchBFS(HistoryEntry::Handle ioEntry, int inDepth, bool inAncestry)");
}


/*!
 *  \brief Write the history graph of flagged individuals in the graphviz dot format.
 *  \param outStream The stream to write the history graph to.
 *  \param ioSelectedEntries The history entries whose ancestors/descendants to find.
 *  \param inAncestryDepth Depth to which ancestors are searched for.
 *  \param inDescendancyDepth Depth to which descendants are searched for.
 */
void History::writeGenealogy(std::ostream& outStream, HistoryEntry::Bag& ioSelectedEntries, int inAncestryDepth, int inDescendancyDepth)
{
	Beagle_StackTraceBeginM();

	findGenealogy(ioSelectedEntries, inAncestryDepth, inDescendancyDepth);

	std::ostringstream lOSSHead;
	std::ostringstream lOSSBody;
	std::ostringstream lOSSTail;
	std::ostringstream lOSSSame;

	lOSSHead << "digraph G {" << std::endl;
	lOSSHead << "rankdir=LR; ranksep=0.40; nodesep=0.40; center=true;" << std::endl;
	lOSSHead << "{ node [shape=ellipse,width=0.50,height=0.50,label=\"\"]; ";

	// We iterate through all the individuals, choosing for display only those
	// individuals having their genealogy flag set by the findGenealogy() call
	// above.  Also, results of intermediate selection and flush operations are
	// not shown.

	unsigned int lCurrentGeneration = UINT_MAX;
	for (unsigned int i = 0; i < mEntries.size(); ++i) {
		for (unsigned int j = 0; j < mEntries[i].size(); ++j) {
			HistoryEntry::Handle lEntry = mEntries[i][j];
			if (lEntry == NULL)
				continue;
			if (lEntry->getGenealogyFlag() == false)
				continue;
			if ((lEntry->getMarkedFlag() == false) &&
			        (lEntry->getCreationAction() == HistoryEntry::eSelection || lEntry->getCreationAction() == HistoryEntry::eFlush)) {
				continue;
			}

			std::string lNodeId = "\"" + Beagle::uint2str(i) + "v" + Beagle::uint2str(j) + "\"";

			// Nodes in the same generation are assigned to the same rank.
			if (lEntry->getMarkedFlag() == true) {
				if (lCurrentGeneration != lEntry->getGeneration()) {
					if (lCurrentGeneration < UINT_MAX) {
						lOSSSame << "}" << std::endl;
						lOSSHead << " -> ";

						// If the gap between the current generation and next generation is more
						// than 1, then fill in that gap of generation numbers.
						for (unsigned int k = lCurrentGeneration+1; k < lEntry->getGeneration(); ++k) {
							std::string lGenerationId = "\"" + Beagle::uint2str(k) + "\"";
							lOSSHead << lGenerationId;
							lOSSHead << " -> ";
						}
					}

					lOSSSame << "{ rank = same; ";
					lCurrentGeneration = lEntry->getGeneration();
					std::string lGenerationId = "\"" + Beagle::uint2str(lCurrentGeneration) + "\"";
					lOSSHead << lGenerationId;
					lOSSSame << lGenerationId << "; ";
				}
				lOSSSame << lNodeId << "; ";
			}

			lOSSBody << lNodeId << " [";
			lOSSBody << "shape=box,width=0.50,height=0.50,label=\"\"";
			lOSSBody << "];" << std::endl;

			// Draw edges from those parents whose genealogy flag is also set.  If the parent
			// is an intermediate created by selection and, therefore, not shown, draw the
			// edge directly to the grandparent.
			const std::vector<HistoryEntry::Handle>& lParents = lEntry->getParents();
			for (unsigned int k = 0; k < lParents.size(); ++k) {
				// Determine color of edge to parents based on the creation action.
				std::string lEdgeColor;
				if (lEntry->getCreationAction() == HistoryEntry::eSelection)
					lEdgeColor = "#000000";
				else if (lEntry->getCreationAction() == HistoryEntry::eElitism)
					lEdgeColor = "#FF0000";
				else if (lEntry->getCreationAction() == HistoryEntry::eMutation)
					lEdgeColor = "#00FF00";
				else if (lEntry->getCreationAction() == HistoryEntry::eCrossover)
					lEdgeColor = (k == 0) ? "#0000FF" : "#00FFFF";
				else
					lEdgeColor = "#7F7F7F";

				HistoryEntry::Handle lSource = lParents[k];
				while ((lSource->getMarkedFlag() == false) &&
				        (lSource->getCreationAction() == HistoryEntry::eSelection || lSource->getCreationAction() == HistoryEntry::eFlush)) {
					lSource = (lSource->getParents())[0];
				}
				if (lSource->getGenealogyFlag() == true) {
					std::string lSourceId = "\"" + Beagle::uint2str(lSource->getHistoryId()) + "v" + Beagle::uint2str(lSource->getHistoryVar()) + "\"";

					lOSSBody << lSourceId << " -> " << lNodeId << " [";
					lOSSBody << "color=\"" << lEdgeColor << "\"";
					lOSSBody << ",style=bold";
					lOSSBody << ",arrowsize=1.25";
					lOSSBody << "];" << std::endl;
				}
			}
		}
	}

	lOSSHead << " [style=invis]; ";

	lOSSHead << "}" << std::endl;
	lOSSSame << "}" << std::endl;
	lOSSTail << "}" << std::endl;

	// Write the different sections to the output stream.
	outStream << lOSSHead.str();
	outStream << lOSSSame.str();
	outStream << lOSSBody.str();
	outStream << lOSSTail.str();

	Beagle_StackTraceEndM("void History::writeGenealogy(std::ostream& outStream, HistoryEntry::Bag& ioSelectedEntries, int inAncestryDepth, int inDescendancyDepth)");
}

