
#include <QSettings>
#include <QCloseEvent>

#include "explorer/Explorer.hpp"

using namespace Explorer;


HistoryTab::HistoryTab(Beagle::System::Handle ioSystem, QWidget* inParent, Qt::WFlags inFlags) :
		QWidget(inParent, inFlags),
		Ui::HistoryTab()
{
	Beagle_StackTraceBeginM();

	setupUi(this);

    const Beagle::Factory& lFactory = ioSystem->getFactory();
    History::Alloc::Handle lHistoryAlloc =
        Beagle::castHandleT<History::Alloc>(lFactory.getAllocator("Explorer-History"));
    mHistory = Beagle::castHandleT<History>(lHistoryAlloc->allocate());

	// Remove any widgets created by Qt Designer in the widget stacks or settings tabs.
	while (mViewStack->count() > 0)
		mViewStack->removeWidget(mViewStack->currentWidget());
	mSettingsTab->clear();

	mVivariumSettings = new VivariumSettings(mHistory, this);
	mGenealogySettings = new GenealogySettings(mHistory, this);

	// Add widgets to the widget stacks.
	mSettingsTab->addTab(mVivariumSettings, "Viva&rium");
	mViewStack->addWidget(mVivariumSettings->getView());
	mSettingsTab->addTab(mGenealogySettings, "&Genealogy");
	mViewStack->addWidget(mGenealogySettings->getView());

	// Make the connection between tabs and views.
	connect(mSettingsTab, SIGNAL(currentChanged(int)), this, SLOT(activateView(int)));

	readSettings();

	Beagle_StackTraceEndM("HistoryTab::HistoryTab(Beagle::System::Handle ioSystem, QWidget* inParent, Qt::WFlags inFlags)");
}


void HistoryTab::closeEvent(QCloseEvent *inEvent)
{
	Beagle_StackTraceBeginM();

	writeSettings();
	inEvent->accept();

	Beagle_StackTraceEndM("void HistoryTab::closeEvent(QCloseEvent *inEvent)");
}


void HistoryTab::readWithSystem(PACC::XML::ConstIterator inIter, Beagle::System& ioSystem)
{
	Beagle_StackTraceBeginM();

	Beagle::History::Handle lHistory = Beagle::castHandleT<Beagle::History>(ioSystem.haveComponent("History"));
	if (lHistory != NULL) {
		lHistory->readWithSystem(inIter, ioSystem);
		mHistory->setup(lHistory, ioSystem);
	}

	Beagle_StackTraceEndM("void HistoryTab::read(PACC::XML::ConstIterator inIter, Beagle::System& ioSystem)");
}


void HistoryTab::reset()
{
	Beagle_StackTraceBeginM();

	mHistory->reset();
	mVivariumSettings->reset();
	mGenealogySettings->reset();

	Beagle_StackTraceEndM("void HistoryTab::reset()");
}


void HistoryTab::setup()
{
	Beagle_StackTraceBeginM();

	mSettingsTab->setCurrentIndex(0);
	mVivariumSettings->setup();
	mGenealogySettings->setup();

	Beagle_StackTraceEndM("void HistoryTab::setup()");
}


/*!
 *  \brief Activate the appropriate view in response to a tab selection.
 */
void HistoryTab::activateView(int inTabIndex)
{
	Beagle_StackTraceBeginM();

	mViewStack->setCurrentIndex(inTabIndex);

	// Additional processing for selection of genealogy tab.
	if (inTabIndex == 1 && mGenealogyUpdateScheduled == true) {
		HistoryEntry::Bag::Handle lSelectedEntries = mVivariumSettings->getSelectedHistoryEntries();
		mGenealogySettings->setVivariumSelections(lSelectedEntries);
		mGenealogySettings->updateView();
		mGenealogyUpdateScheduled = false;
	}

	Beagle_StackTraceEndM("void HistoryTab::activateView(int inTabIndex)");
}


/*!
 *  \brief Sets the flag indicating whether genealogy view must be updated.
 */
void HistoryTab::scheduleGenealogyUpdate()
{
	Beagle_StackTraceBeginM();

	mGenealogyUpdateScheduled = true;

	Beagle_StackTraceEndM("void HistoryTab::scheduleGenealogyUpdate()");
}


void HistoryTab::readSettings()
{
	Beagle_StackTraceBeginM();

	QSettings lSettings("OpenBEAGLE", "Explorer");

	lSettings.beginGroup("HistoryTab");
	mSplitter->restoreState(lSettings.value("splitterSizes").toByteArray());
	lSettings.endGroup();

	Beagle_StackTraceEndM("void HistoryTab::readSettings()");
}


void HistoryTab::writeSettings()
{
	Beagle_StackTraceBeginM();

	QSettings lSettings("OpenBEAGLE", "Explorer");

	lSettings.beginGroup("HistoryTab");
	lSettings.setValue("splitterSizes", mSplitter->saveState());
	lSettings.endGroup();

	Beagle_StackTraceEndM("void HistoryTab::writeSettings()");
}
