
#include "explorer/Explorer.hpp"

using namespace Explorer;


/*!
 *  \brief Returns the static singleton instance of the InteractionData class.
 */
InteractionData::InteractionData& InteractionData::instance()
{
	Beagle_StackTraceBeginM();
    static InteractionData lInteractionData;
    return lInteractionData;
	Beagle_StackTraceEndM("InteractionData::InteractionData& InteractionData::instance()");
}


/*!
 *  \brief Construct an interaction data object.
 */
InteractionData::InteractionData() :
		QObject(),  // No parent specified, so Qt won't destroy this object.
        mLabelGeneration(0),
        mCurrentGeneration(0),
        mBreakGeneration(0),
        mLabelDemeIndex(0),
        mCurrentDemeIndex(0),
        mBreakDemeIndex(0),
        mLabelIndividualIndex(0),
        mCurrentIndividualIndex(0),
        mBreakIndividualIndex(0),
        mPauseEnabled(false),
        mBreakEnabled(false),
		mStatusUpdateTime(new QTime()),
		mStatusUpdateTimeInterval(100),
        mEventLoop(new QEventLoop())
{ }


/*!
 *  \brief Destructor for the interaction data object.
 */
InteractionData::~InteractionData()
{
	Beagle_StackTraceBeginM();
	delete mStatusUpdateTime;
    delete mEventLoop;
	Beagle_StackTraceEndM("InteractionData::~InteractionData()");
}


/*!
 *  \brief Setup the interaction data object.
 */
void InteractionData::setup(Beagle::System::Handle ioSystem)
{
	Beagle_StackTraceBeginM();
	mStatusUpdateTime->start();
	Beagle_StackTraceEndM("void InteractionData::setup(Beagle::System::Handle ioSystem)");
}


/*!
 *  \brief Create and setup the evolution status indicators.
 *  \param inParent The parent widget of the new status indicator widgets.
 *  \param outStatusIndicators The new status indicator widgets are returned in this array.
 */
void InteractionData::setupStatusIndicators(QWidget* inParent, QVector< QVector<QWidget*> >& outStatusIndicators)
{
	Beagle_StackTraceBeginM();

    outStatusIndicators.push_back(QVector<QWidget*>());

    mLabelGeneration = new QLabel(inParent);
    mLabelGeneration->setObjectName(QString::fromUtf8("mLabelGeneration"));
    mLabelGeneration->setText(QApplication::translate("EvolutionTab", "&Generation", 0, QApplication::UnicodeUTF8));
    outStatusIndicators.back().push_back(mLabelGeneration);

    mCurrentGeneration = new QLabel(inParent);
    mCurrentGeneration->setObjectName(QString::fromUtf8("mCurrentGeneration"));
    mCurrentGeneration->setText(QString());
    outStatusIndicators.back().push_back(mCurrentGeneration);

    mBreakGeneration = new QSpinBox(inParent);
    mBreakGeneration->setObjectName(QString::fromUtf8("mBreakGeneration"));
    mBreakGeneration->setMaximum(99999);
    outStatusIndicators.back().push_back(mBreakGeneration);

    outStatusIndicators.push_back(QVector<QWidget*>());

    mLabelDemeIndex = new QLabel(inParent);
    mLabelDemeIndex->setObjectName(QString::fromUtf8("mLabelDemeIndex"));
    mLabelDemeIndex->setText(QApplication::translate("EvolutionTab", "&Deme Index", 0, QApplication::UnicodeUTF8));
    outStatusIndicators.back().push_back(mLabelDemeIndex);

    mCurrentDemeIndex = new QLabel(inParent);
    mCurrentDemeIndex->setObjectName(QString::fromUtf8("mCurrentDemeIndex"));
    mCurrentDemeIndex->setText(QString());
    outStatusIndicators.back().push_back(mCurrentDemeIndex);

    mBreakDemeIndex = new QSpinBox(inParent);
    mBreakDemeIndex->setObjectName(QString::fromUtf8("mBreakDemeIndex"));
    mBreakDemeIndex->setMaximum(99999);
    outStatusIndicators.back().push_back(mBreakDemeIndex);

    outStatusIndicators.push_back(QVector<QWidget*>());

    mLabelIndividualIndex = new QLabel(inParent);
    mLabelIndividualIndex->setObjectName(QString::fromUtf8("mLabelIndividualIndex"));
    mLabelIndividualIndex->setText(QApplication::translate("EvolutionTab", "&Individual Index", 0, QApplication::UnicodeUTF8));
    outStatusIndicators.back().push_back(mLabelIndividualIndex);

    mCurrentIndividualIndex = new QLabel(inParent);
    mCurrentIndividualIndex->setObjectName(QString::fromUtf8("mCurrentIndividualIndex"));
    mCurrentIndividualIndex->setText(QString());
    outStatusIndicators.back().push_back(mCurrentIndividualIndex);

    mBreakIndividualIndex = new QSpinBox(inParent);
    mBreakIndividualIndex->setObjectName(QString::fromUtf8("mBreakIndividualIndex"));
    mBreakIndividualIndex->setMaximum(99999);
    outStatusIndicators.back().push_back(mBreakIndividualIndex);

    mLabelGeneration->setBuddy(mBreakGeneration);
    mLabelDemeIndex->setBuddy(mBreakDemeIndex);
    mLabelIndividualIndex->setBuddy(mBreakIndividualIndex);

	Beagle_StackTraceEndM("void InteractionData::setupStatusIndicators(QWidget* inParent, QVector< QVector<QWidget*> >& outStatusIndicators)");
}


/*!
 *  \brief Create and setup the display window indicators.
 *  \param inParent The parent widget of the new display indicator widgets.
 *  \param outDisplayIndicators The new display indicator widgets are returned in this array.
 */
void InteractionData::setupDisplayIndicators(QWidget* inParent, QVector<QWidget*>& outDisplayIndicators)
{ }


/*!
 *  \brief Test if pause/break should be applied to evolution.
 *  \param inContext The context of evolution.
 */
bool InteractionData::testPauseOrBreak(const Beagle::Context& inContext)
{
	Beagle_StackTraceBeginM();

    unsigned int lGeneration = inContext.getGeneration();
    unsigned int lDemeIndex = inContext.getDemeIndex();
    unsigned int lIndividualIndex = inContext.getIndividualIndex();

    int lBreakGeneration = mBreakGeneration->value();
    int lBreakDemeIndex = mBreakDemeIndex->value();
    int lBreakIndividualIndex = mBreakIndividualIndex->value();

    return (mPauseEnabled ||
            (mBreakEnabled && 
             ((lGeneration > lBreakGeneration) ||
              (lGeneration == lBreakGeneration && lDemeIndex > lBreakDemeIndex) ||
              (lGeneration == lBreakGeneration && lDemeIndex == lBreakDemeIndex && lIndividualIndex >= lBreakIndividualIndex))));

	Beagle_StackTraceEndM("bool InteractionData::testPauseOrBreak(const Beagle::Context& inContext)");
}


/*!
 *  \brief Test if pause is enabled, or break is enabled and we are past the breakpoints.
 *  \param inContext The context of evolution.
 */
bool InteractionData::testPauseOrPastBreak(const Beagle::Context& inContext)
{
	Beagle_StackTraceBeginM();

    unsigned int lGeneration = inContext.getGeneration();
    unsigned int lDemeIndex = inContext.getDemeIndex();
    unsigned int lIndividualIndex = inContext.getIndividualIndex();

    int lBreakGeneration = mBreakGeneration->value();
    int lBreakDemeIndex = mBreakDemeIndex->value();
    int lBreakIndividualIndex = mBreakIndividualIndex->value();

    return (mPauseEnabled ||
            (mBreakEnabled && 
             ((lGeneration > lBreakGeneration) ||
              (lGeneration == lBreakGeneration && lDemeIndex > lBreakDemeIndex) ||
              (lGeneration == lBreakGeneration && lDemeIndex == lBreakDemeIndex && lIndividualIndex > lBreakIndividualIndex))));

	Beagle_StackTraceEndM("bool InteractionData::testPauseOrPastBreak(const Beagle::Context& inContext)");
}


/*!
 *  \brief Test if break is enabled and we are at the break points.
 *  \param inContext The context of evolution.
 */
bool InteractionData::testAtBreak(const Beagle::Context& inContext)
{
	Beagle_StackTraceBeginM();

    unsigned int lGeneration = inContext.getGeneration();
    unsigned int lDemeIndex = inContext.getDemeIndex();
    unsigned int lIndividualIndex = inContext.getIndividualIndex();

    int lBreakGeneration = mBreakGeneration->value();
    int lBreakDemeIndex = mBreakDemeIndex->value();
    int lBreakIndividualIndex = mBreakIndividualIndex->value();

    return (mBreakEnabled && lGeneration == lBreakGeneration && lDemeIndex == lBreakDemeIndex && lIndividualIndex == lBreakIndividualIndex);

	Beagle_StackTraceEndM("bool InteractionData::testAtBreak(const Beagle::Context& inContext)");
}


/*!
 *  \brief Test if status update time has elapsed.
 */
bool InteractionData::testStatusUpdateTime()
{
	Beagle_StackTraceBeginM();
    return (mStatusUpdateTime->elapsed() > mStatusUpdateTimeInterval);
	Beagle_StackTraceEndM("bool InteractionData::testStatusUpdateTime()");
}


/*! 
 *  \brief Update the status of the widgets in the evolution tab.
 *  \param inContext The context of evolution.
 */
void InteractionData::update(const Beagle::Context& inContext)
{
    Beagle_StackTraceBeginM();

    mCurrentGeneration->setText(QString::number(inContext.getGeneration()));
    mCurrentDemeIndex->setText(QString::number(inContext.getDemeIndex()));
    mCurrentIndividualIndex->setText(QString::number(inContext.getIndividualIndex()));

	// We need to explicitly specify an immediate update because evaluations
	// are compute intensive.
	QCoreApplication::processEvents();

    // Reset timer.
    mStatusUpdateTime->start();

    Beagle_StackTraceEndM("void InteractionData::update(const Beagle::Context& inContext)");
}


/*! 
 * \brief Set whether the pause button is enabled.
 */
void InteractionData::setPauseEnabled()
{
	Beagle_StackTraceBeginM();
	mPauseEnabled = true;
	Beagle_StackTraceEndM("void InteractionData::setPauseEnabled()");
}


/*! 
 * \brief Set whether the break check box is enabled.
 */
void InteractionData::setBreakEnabled(int inEnabledState)
{
	Beagle_StackTraceBeginM();
	mBreakEnabled = inEnabledState == Qt::Checked;
	Beagle_StackTraceEndM("void InteractionData::setBreakEnabled(int inEnabledState)");
}


/*! 
 * \brief Continue evolution by exiting the Qt event loop.
 */
void InteractionData::continueFromPauseOrBreak()
{
	Beagle_StackTraceBeginM();
    exitEventLoop();
    mPauseEnabled = false;
	Beagle_StackTraceEndM("void InteractionData::continueFromPauseOrBreak()");
}
