
#include <QFileDialog>
#include <QMessageBox>
#include <QCloseEvent>
#include <QSettings>
#include <QWhatsThis>

#include "explorer/Explorer.hpp"

#include <fstream>

#ifdef BEAGLE_HAVE_LIBZ
#include <gzstream.h>
#endif // BEAGLE_HAVE_LIBZ


using namespace Explorer;


MainWindow::MainWindow(Beagle::System::Handle ioSystem, QWidget* inParent, Qt::WFlags inFlags) :
		QMainWindow(inParent, inFlags),
		Ui::MainWindow(),
		mSystem(ioSystem)
{
	Beagle_StackTraceBeginM();

	setupUi(this);
	connect(mActionOpenMilestone, SIGNAL(triggered()), this, SLOT(openMilestone()));
	connect(mActionQuit, SIGNAL(triggered()), this, SLOT(close()));
	connect(mActionConfigure, SIGNAL(triggered()), this, SLOT(showConfigureDialog()));
	connect(mActionWhatsThis, SIGNAL(triggered()), this, SLOT(enterWhatsThisMode()));
	connect(mActionAboutExplorer, SIGNAL(triggered()), this, SLOT(aboutExplorer()));

	// Remove any tabs created by Qt Designer.
	mTabWidget->clear();

    const Beagle::Factory& lFactory = ioSystem->getFactory();

    // Tab for running and controlling evolution.
    mEvolutionTab = new EvolutionTab(ioSystem, this);
	mTabWidget->addTab(mEvolutionTab, "&Evolution");
	mTabWidget->setTabEnabled(mTabWidget->indexOf(mEvolutionTab), mEvolutionTab->isEnabled());

	connect(mActionOpenConfiguration, SIGNAL(triggered()), mEvolutionTab, SLOT(openConfiguration()));
	connect(mActionSaveConfiguration, SIGNAL(triggered()), mEvolutionTab, SLOT(saveConfiguration()));
	connect(mActionSaveConfigurationAs, SIGNAL(triggered()), mEvolutionTab, SLOT(saveConfigurationAs()));

    // Make history tab if system contains history component.
    Beagle::History::Handle lHistory = Beagle::castHandleT<Beagle::History>(ioSystem->haveComponent("History"));
    if (lHistory != NULL) {
        mHistoryTab = new HistoryTab(ioSystem, this);
        mTabWidget->addTab(mHistoryTab, "&History");
        mTabWidget->setTabEnabled(mTabWidget->indexOf(mHistoryTab), mHistoryTab->isEnabled());
    }

    // Disable milestone file reading action if a specific individual allocator
    // has not been specified in the concept map.
    //if (lFactory.getConceptTypeName("Explorer-Individual") == "Explorer-Individual")
    //    mActionOpenMilestone->setEnabled(false);

	readSettings();

	Beagle_StackTraceEndM("MainWindow::MainWindow(Beagle::System::Handle ioSystem, QWidget* inParent, Qt::WFlags inFlags)");
}


void MainWindow::closeEvent(QCloseEvent *inEvent)
{
	Beagle_StackTraceBeginM();

	writeSettings();

    // Close the tabs so that their settings will also be written.
    for (int i=0; i<mTabWidget->count(); ++i) {
        mTabWidget->widget(i)->close();
    }

	inEvent->accept();

	Beagle_StackTraceEndM("void MainWindow::closeEvent(QCloseEvent *inEvent)");
}


void MainWindow::showConfigureDialog()
{
	Beagle_StackTraceBeginM();

	ConfigureDialog lConfigureDialog(this);
	lConfigureDialog.showDialog();

	Beagle_StackTraceEndM("void MainWindow::showConfigureDialog()");
}


void MainWindow::enterWhatsThisMode()
{
	Beagle_StackTraceBeginM();

	QWhatsThis::enterWhatsThisMode();

	Beagle_StackTraceEndM("void MainWindow::enterWhatsThisMode()");
}


void MainWindow::aboutExplorer()
{
	Beagle_StackTraceBeginM();

	QMessageBox lBox(this);
	lBox.setText(QString::fromLatin1("<center><h3>Open BEAGLE Explorer</h3><p>Version 0.1</p></center>"
	                                 "<p>Copyright (C) 2007 by Vinod Valsalam.</p>"
	                                 "<p>Open BEAGLE Explorer is a library based on the Open BEAGLE Evolutionary Computation framework "
	                                 "(<a href=\"http://beagle.gel.ulaval.ca\">http://beagle.gel.ulaval.ca</a>).</p>"
	                                 "<p>The graphical interface is developed using the Qt toolkit "
	                                 "(<a href=\"http://trolltech.com/products/qt\">http://trolltech.com/products/qt</a>).</p>"
	                                 "<p>The icon images are from the Tango Desktop Project "
	                                 "(<a href=\"http://tango.freedesktop.org\">http://tango.freedesktop.org</a>).</p>"
	                                 "<p>This library is distributed in the hope that it will be useful, "
	                                 "but WITHOUT ANY WARRANTY; without even the implied warranty of "
	                                 "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU "
	                                 "Lesser General Public License for more details.<p/>"));
	lBox.setWindowTitle(tr("Open BEAGLE Explorer"));
	lBox.setIcon(QMessageBox::NoIcon);
	lBox.exec();

	Beagle_StackTraceEndM("void MainWindow::aboutExplorer()");
}


void MainWindow::openMilestone()
{
	Beagle_StackTraceBeginM();

	QStringList lFileNames = QFileDialog::getOpenFileNames(this, tr("Open Milestone File(s)"),
	                         QDir::currentPath(),
	                         tr("Milestone Files (*.obm.gz)"));
	if (lFileNames.isEmpty())
		return;

    Beagle::History::Handle lHistory = Beagle::castHandleT<Beagle::History>(mSystem->haveComponent("History"));
    if (lHistory == NULL)
        return;

	mHistoryTab->reset();

	for (int i = 0; i < lFileNames.size(); ++i) {
		QString lFileName = lFileNames.at(i);

		if (!QFile::exists(lFileName)) {
			QMessageBox::warning(this, tr("Milestone File"),
			                     tr("File does not exist: %1").arg(lFileName));
			return;
		}

#ifdef BEAGLE_HAVE_LIBZ
		igzstream lStream(lFileName.toStdString().c_str());
#else // BEAGLE_HAVE_LIBZ

		std::ifstream lStream(lFileName.toStdString().c_str());
#endif // BEAGLE_HAVE_LIBZ

		if (lStream.good()) {
			PACC::XML::Document lDocument(lStream, lFileName.toStdString());
			PACC::XML::ConstFinder lFinder(lDocument.getFirstDataTag());

			PACC::XML::ConstIterator lPos = lFinder.find("/Beagle/System/History");
			if (lPos)
				mHistoryTab->readWithSystem(lPos, *mSystem);
		}

		lStream.close();
	}

	mHistoryTab->setup();
	statusBar()->showMessage(tr("Read history from file(s)."), 2000);
	mTabWidget->setTabEnabled(mTabWidget->indexOf(mHistoryTab), mHistoryTab->isEnabled());
	setCurrentTab(mHistoryTab);

	Beagle_StackTraceEndM("void MainWindow::openMilestone()");
}


void MainWindow::readSettings()
{
	Beagle_StackTraceBeginM();

	QSettings lSettings("OpenBEAGLE", "Explorer");

	lSettings.beginGroup("MainWindow");
	resize(lSettings.value("size", QSize(600, 400)).toSize());
	move(lSettings.value("pos", QPoint(200, 200)).toPoint());
	lSettings.endGroup();

	Beagle_StackTraceEndM("void MainWindow::readSettings()");
}


void MainWindow::writeSettings()
{
	Beagle_StackTraceBeginM();

	QSettings lSettings("OpenBEAGLE", "Explorer");

	lSettings.beginGroup("MainWindow");
	lSettings.setValue("pos", pos());
	lSettings.setValue("size", size());
	lSettings.endGroup();

	Beagle_StackTraceEndM("void MainWindow::writeSettings()");
}
