
#include "explorer/Explorer.hpp"

using namespace Explorer;


VivariumSettings::VivariumSettings(History::Handle inHistory, QWidget* inParent, Qt::WFlags inFlags) :
		HistorySettings(inHistory, inParent, inFlags),
        mHistoryTab(qobject_cast<HistoryTab*>(inParent))
{
	Beagle_StackTraceBeginM();

    mSortLabel = new QLabel(this);
    mSortLabel->setObjectName(QString::fromUtf8("mSortLabel"));
    mSortLabel->setText(QApplication::translate("VivariumSettings", "&Column Sort", 0, QApplication::UnicodeUTF8));
    gridLayout->addWidget(mSortLabel, 3, 0, 1, 1);

    mSortCombo = new QComboBox(this);
    mSortCombo->setObjectName(QString::fromUtf8("mSortCombo"));
    gridLayout->addWidget(mSortCombo, 3, 1, 1, 1);

    mSortLabel->setBuddy(mSortCombo);

	connect(mSortCombo, SIGNAL(currentIndexChanged(int)), this, SLOT(setupScene(int)));

	mSortCombo->setWhatsThis(tr("Specify the sort criterion for individuals in a generation."));

	mView->setWhatsThis(tr("<b>Vivarium view of evolutionary history.</b>") +
	                    tr("<p>Nodes in a column represent the individuals of a generation. ") +
	                    tr("Select nodes to highlight their ancestors and/or descendants. ") +
	                    tr("Switch to the <em>Genealogy</em> tab to see the genealogy tree of selected nodes.</p>"));

	Beagle_StackTraceEndM("VivariumSettings::VivariumSettings(History::Handle inHistory, QWidget* inParent, Qt::WFlags inFlags)");
}


/*!
 *  \brief Reset vivarium display settings.
 */
void VivariumSettings::reset()
{
	Beagle_StackTraceBeginM();

	HistorySettings::reset();

	mSortCombo->clear();

	Beagle_StackTraceEndM("void VivariumSettings::reset()");
}


/*!
 *  \brief Setup vivarium display settings.
 */
void VivariumSettings::setup()
{
	Beagle_StackTraceBeginM();

	HistorySettings::setup();

	// Setup the column sort combo box using the prototype display object.
	int lNumEntries = mPrototypeObject->getDisplaySize();
	for (int k = 0; k < lNumEntries; ++k) {
		std::string lDisplayName = mPrototypeObject->getDisplayName(k);
		mSortCombo->addItem(lDisplayName.c_str());
	}

    setupScene(mSortCombo->currentIndex());

	Beagle_StackTraceEndM("void VivariumSettings::setup()");
}


/*!
 *  \brief Setup the scene for the vivarium display.
 *  \param inSortIndex The index of the combo box by which to sort the columns.
 */
void VivariumSettings::setupScene(int inSortIndex)
{
	Beagle_StackTraceBeginM();

	// Schedule the previous scene for deletion before allocating a new one.
	if (mScene)
		mScene->deleteLater();
	mScene = new DisplayScene(this);

	mScene->setNodeHueIndex(mNodeHueCombo->currentIndex());
	mScene->setNodeSatIndex(mNodeSatCombo->currentIndex());
	mScene->setNodeValIndex(mNodeValCombo->currentIndex());

	qreal lBoundingSize = 40;
	qreal lGridSize = lBoundingSize + 15;
	mScene->setNodeBoundingSize(lBoundingSize);

	for (unsigned int i = mHistory->getMinGeneration(); i < mHistory->getMaxGeneration(); ++i) {
		// Place the generation label.
		QGraphicsTextItem* lItem = new QGraphicsTextItem(QString::number(i));
		QRectF lTextRect = lItem->boundingRect();
		qreal lTextWidth = lTextRect.right() - lTextRect.left();
		qreal x = i*lGridSize + (lBoundingSize - lTextWidth)/2;
		qreal y = lBoundingSize/2;
		lItem->setPos(x, y);
		mScene->addItem(lItem);

        std::vector<VivariumNode*> lNodes;
		for (unsigned int j = 0; j < mHistory->getVivariumSize(i); ++j) {
			// Create nodes for each marked history entry.
			HistoryEntry::Handle lHistoryEntry = mHistory->getMarkedEntry(i, j);
			VivariumNode* lItem = new VivariumNode(lHistoryEntry);
			lItem->setToolTip(lHistoryEntry->getToolTip().c_str());
            lNodes.push_back(lItem);
        }

        // Sort the nodes if a valid sort index is specified.
        if (inSortIndex > 0)
            std::stable_sort(lNodes.begin(), lNodes.end(), IsLessNodePredicate(inSortIndex));

		for (unsigned int j = 0; j < mHistory->getVivariumSize(i); ++j) {
            VivariumNode* lItem = lNodes[j];
			lItem->setPos(i*lGridSize, (j+1)*lGridSize);
			mScene->addItem(lItem);
		}
	}

	mView->setScene(mScene);

	connect(mScene, SIGNAL(nodeSelectionChanged()), this, SLOT(updateView()));
    mHistoryTab->scheduleGenealogyUpdate();
	connect(mScene, SIGNAL(nodeSelectionChanged()), mHistoryTab, SLOT(scheduleGenealogyUpdate()));

	Beagle_StackTraceEndM("void VivariumSettings::setupScene(int inSortIndex)");
}


/*!
 *  \brief Update view when selection or ancestry/descendancy depths change.
 */
void VivariumSettings::updateView()
{
	Beagle_StackTraceBeginM();

	if (mScene == 0)
		return;

	HistoryEntry::Bag::Handle lSelections = getSelectedHistoryEntries();
	mHistory->findGenealogy(*lSelections, mAncestrySpin->value(), mDescendancySpin->value());

	foreach (QGraphicsItem* lItem, mScene->items()) {
		VivariumNode* lVivariumNode = qgraphicsitem_cast<VivariumNode*>(lItem);
		if (lVivariumNode) {
			bool lGenealogyFlag = lVivariumNode->getHistoryEntry()->getGenealogyFlag();
			if (lGenealogyFlag != lVivariumNode->getGenealogyFlag()) {
				lVivariumNode->setGenealogyFlag(lGenealogyFlag);
				lVivariumNode->update();
			}
		}
	}

	Beagle_StackTraceEndM("void VivariumSettings::updateView()");
}
